/********************************************************************
 ** @version BRA10
 ** P: Support pointer data
 ** @author Alvaro Estebanez
 ** @date 1th March 2012
 *********************************************************************/

//  Main header file with all common variables
#include "specEngine.h"
//  speccode: OpenMP header file included
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>

// Mathematical header file, for the benefit of dynScheduler() function.
#include <math.h>

/**
 ** @version specload-P
 ** @author alvaro estebanez
 **/
//*********************************************************************
//   specload_pointer(): speculative load operation with forwarding
//*********************************************************************

/**
 *
 * @param address   - Address of the pointer to be loaded
 * @param byteSize  - Size of the data to be loaded
 * @param current   - The number of thread
 * @param value     - Address of the variable which loads data pointed
 * @return          - EXIT_SUCCESS if everything is right
 **/
int specload_pointer(void* address, unsigned int byteSize, unsigned int current, void* value) {

  int tidaux = modvector[current];

  // Check if thread is squashed
  if (wheel[tidaux] == SQUASHED) {
    return -1;
  }

#ifdef TIME_SPECLOAD
  double time_ini, time_end;
  int id = omp_get_thread_num();
  time_ini = get_time();
#endif

  // local variables
  int usedAddressFlag = false;
  int thread = current;
  baseType hashAddress = ((baseType) address >> 3) & MASK;
  int pos = 0;
  int lastData = 0;
  threadCopy_structure *currentThread;
  threadCopy_structure *searchThread;
#ifdef COMPLEX_STRUCTURES
  void *addressToSave;
  int offset;
#endif
  void *addressMostRecently;

  // currentThread should point to the current data matrix
  currentThread = wheel_data[tidaux].pointer;
  // searchThread should initially point to the current data matrix too
  searchThread = currentThread;

#ifdef MEASURE_COUNTERS
  int iden = omp_get_thread_num();
  totalCallsSpecload[iden]++;
#endif

  // check if the data to be loaded has been used by some thread
  //while ((!usedAddressFlag) && (thread >= wheel_ns)) {
  while (true) {
#ifdef MEASURE_COUNTERS
    totalAccessesSpecload[iden]++;
#endif
    // For each thread
    if ((searchThread->matriz[hashAddress][pos].origPointer != address)) {
      // Check if is the last position
      if (searchThread->matriz[hashAddress][pos].origPointer != NULL) {
        pos++;
#ifdef MAXDATA_DEBUG
        if (pos >= wheel_data[modvector[thread]].maxData) {
          // We need allocate more memory
          printf("ERROR-SPECLOAD: more memory is needed: thread %d,pos=%d,max=%d\n",modvector[thread],pos,wheel_data[modvector[thread]].maxData);
          exit(-1);
        }
#endif
      } else { // If is the last position, must looking for other thread
        if (thread == current) {
          // The first void position of this thread
          lastData = pos;
        }
        thread--;
        pos = 0;
        if (thread >= wheel_ns) {
#pragma omp flush(wheel_data)
          searchThread = wheel_data[modvector[thread]].pointer;
        } else {
          // usedAddressFlag = false;
          break;
        }
      }
    } else {
      usedAddressFlag = true;
      break;
      // thread variable will contain the thread which used the pointer
      // pos variable will contain the position of the pointer in its matrix
    }
  }

  // NOTE: thread variable will contain the thread that used the pointer
  //       pos variable will contain the position of the pointer in its matrix
  //       usedAddressFlag = true
  //  else if the threads dont use the pointer:
  //          thread = wheel_ns - 1;
  //          pos = 0;
  //          usedAddressFlag = false;

  // If the thread that wants to load data had used pointer before then
  // must load it from its matrix
  // flag usedAddressFlag == true
  if (current == thread) {

#ifdef COMPLEX_STRUCTURES
    // Checking for data size
    offset = currentThread->matriz[hashAddress][pos].copyOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);

    memcpy(value, addressMostRecently, currentThread->matriz[hashAddress][pos].size);
#else
    addressMostRecently = (void *) &(currentThread->matriz[hashAddress][pos].speculativeCopy);
    switch (byteSize) {
      case 1:
        *(char *) value = *(char *) addressMostRecently;
        break;
      case 2:
        *(short *) value = *(short *) addressMostRecently;
        break;
      case 4:
        *(int *) value = *(int *) addressMostRecently;
        break;
      case 8:
        * (double *) value = *(double *) addressMostRecently;
        break;
      default:
        printf("Error while copying temporal data, size = %d: SPECCODE->Specload: current==thread... Exit", byteSize);
        exit(-1);
    }
#endif
  } else { // current != thread

#ifdef MAXDATA_DEBUG
    // lastData has been asigned in the previous loop
    // Check if thread matrix is full (need to add data)
    //if (wheel_data[modvector[current]].lastData == wheel_data[modvector[current]].maxData)
    if (lastData >= wheel_data[modvector[current]].maxData) {
      //  reservarMasEspacio();
      printf("Error: lastData=%d >= wheel_data[modvector[current]].maxData=%d, current = %d,\n", lastData, wheel_data[modvector[current]].maxData, modvector[current]);
      fflush(stdout);
      exit(-1);
    }
#endif

#ifdef COMPLEX_STRUCTURES
    offset = wheel_data[tidaux].actualOffset;
    addressToSave = (void *) &(wheel_data[tidaux].temporalData[offset]);
#endif

    if (usedAddressFlag) { // If some thread used pointer, must load it from its matrix

#ifdef COMPLEX_STRUCTURES
      offset = searchThread->matriz[hashAddress][pos].copyOffset;
      addressMostRecently = (void *) &(wheel_data[modvector[thread]].temporalData[offset]);
#else
      addressMostRecently = (void *) &(searchThread->matriz[hashAddress][pos].speculativeCopy);
      currentThread->matriz[hashAddress][lastData].speculativeCopy = searchThread->matriz[hashAddress][pos].speculativeCopy;
      switch (byteSize) {
        case 1:
          *(char *) value = *(char *) addressMostRecently;
          break;
        case 2:
          *(short *) value = *(short *) addressMostRecently;
          break;
        case 4:
          *(int *) value = *(int *) addressMostRecently;
          break;
        case 8:
          * (double *) value = *(double *) addressMostRecently;
          break;
        default:
          printf("Error while copying temporal data, size = %d: SPECCODE (specload) -> line 217... Exit", byteSize);
          exit(-1);
      }
#endif
    } else {
      // No thread used the pointer so load value from memory

#ifdef COMPLEX_STRUCTURES
      addressMostRecently = address;
      // addressToSave        -> contain a direction to save data
      // addressMostRecently  -> contain the most recently version of the pointer
      // Copying Data...
      memcpy(addressToSave, addressMostRecently, byteSize);
      // Copying data to value variable
      memcpy(value, addressToSave, byteSize);
#else
      addressMostRecently = (void *) &(currentThread->matriz[hashAddress][lastData].speculativeCopy);
      switch (byteSize) {
        case 1:
          *(char *) addressMostRecently = *(char *) address;
          *(char *) value = *(char *) addressMostRecently;
          break;
        case 2:
          *(short *) addressMostRecently = *(short *) address;
          *(short *) value = *(short *) addressMostRecently;
          break;
        case 4:
          *(int *) addressMostRecently = *(int *) address;
          *(int *) value = *(int *) addressMostRecently;
          break;
        case 8:
          *(double *) addressMostRecently = *(double *) address;
          *(double *) value = *(double *) addressMostRecently;
          break;
        default:
          printf("Error while copying temporal data, size = %d: SPECCODE (specload) ->line 335.. Exit", byteSize);
          exit(-1);
      }
#endif
    }

    currentThread->matriz[hashAddress][lastData + 1].origPointer = NULL;
#ifdef COMPLEX_STRUCTURES
    currentThread->matriz[hashAddress][lastData].copyOffset = wheel_data[tidaux].actualOffset;
#endif
    currentThread->matriz[hashAddress][lastData].state = EXPLD;
    currentThread->matriz[hashAddress][lastData].size = byteSize;
    currentThread->matriz[hashAddress][lastData].origPointer = address;

#ifdef COMPLEX_STRUCTURES
    // Data added so must increase number of data in the matrix
    wheel_data[tidaux].actualOffset += byteSize;
#ifdef OFFSET_DEBUG
    if (wheel_data[modvector[current]].actualOffset > MAX_OFFSET) {
      printf("ERROR: Offset = %d: higher than maxoffset=%d,\n", wheel_data[modvector[current]].actualOffset, MAX_OFFSET);
      fflush(stdout);
      exit(-1);
    }
#endif
#endif
  }

#ifdef TIME_SPECLOAD
  time_end = get_time();
  time_specload[id] += (time_end - time_ini);
#endif

  return EXIT_SUCCESS;

} // END specload_pointer

/**
 ** @version specstore-P
 ** @author alvaro estebanez
 **/
//*********************************************************************
//   specstore_pointer(): speculative store operation
//*********************************************************************

/**
 *
 * @param address   - Address of the pointer to be saved
 * @param byteSize  - Size of the data to be saved
 * @param current   - The number of thread
 * @param value     - Address of the variable which contain the value to be saved
 * @return          - EXIT_SUCCESS if everything is right
 **/
void specstore_pointer(void* address, unsigned int byteSize, unsigned int current, void* value) {

#ifdef MEASURE_COUNTERS
  int iden = omp_get_thread_num();
  totalCallsSpecstore[iden]++;
#endif

  //// Check if thread is squashed
  //#pragma omp flush(wheel)
  //  if (wheel[modvector[current]] == SQUASHED) {
  //#ifdef SPECSTORE_DEBUG
  //  speccode: debug
  //  printf("ESTOY SQUASHED!.specstore: tid = %d, current = %d,\n",omp_get_thread_num(),current);
  //  fflush(stdout);
  //  speccode: end debug
  //#endif
  //        return -1;
  // }

#ifdef TIME_SPECSTORE
  double time_ini, time_end;
  int id = omp_get_thread_num();
  time_ini = get_time();
#endif

  // local variables
  int usedAddressFlag = false;
  int thread = current;
  baseType hashAddress = ((baseType) address >> 3) & MASK;
  int pos = 0;
  int lastData = -1;
  threadCopy_structure *searchThread;
  void *addressMostRecently;
#ifdef COMPLEX_STRUCTURES
  int offset;
#endif
  int tidaux = modvector[current];

  // searchThread should initially point to the current data matrix
  searchThread = wheel_data[tidaux].pointer;

  // check if the direction where I want to save data, has been used by current thread
  //while ((!usedAddressFlag) && (searchThread->matriz[T_ORIG_POINTER][row] != (baseType) NULL )) {
  while (searchThread->matriz[hashAddress][pos].origPointer != NULL) {
#ifdef MEASURE_COUNTERS
    totalAccessesSpecstore[iden]++;
#endif
    if (searchThread->matriz[hashAddress][pos].origPointer != address) {
      pos++;
#ifdef MAXDATA_DEBUG
        if (pos >= wheel_data[modvector[thread]].maxData) {
          // We need allocate more memory
          printf("ERROR-SPECSTORE: more memory is needed: thread %d,pos=%d,max=%d\n",tidaux,pos,wheel_data[modvector[thread]].maxData);
          exit(-1);
        }
#endif
    } else {
      usedAddressFlag = true;
#ifdef MEASURE_COUNTERS
      // se va a sumar un elemento no utilizado
      totalAccessesSpecstore[iden]--;
#endif
      break;
      // pos variable will contain the position of the pointer in its matrix
    }
  }
#ifdef MEASURE_COUNTERS
  totalAccessesSpecstore[iden]++;
#endif
  // NOTE: pos variable will contain the position of the pointer in its matrix

  // if current thread had used this value before then must override value
  // from its matrix
  if (usedAddressFlag) {
    // First of all, store the value
    // Checking for data size
#ifdef COMPLEX_STRUCTURES
    offset = searchThread->matriz[hashAddress][pos].copyOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);

    memcpy(addressMostRecently, value, searchThread->matriz[hashAddress][pos].size);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][pos].speculativeCopy);
    switch (byteSize) {
      case 1:
        *(char *) addressMostRecently = *(char *) value;
        break;
      case 2:
        *(short *) addressMostRecently = *(short *) value;
        break;
      case 4:
        *(int *) addressMostRecently = *(int *) value;
        break;
      case 8:
        *(double *) addressMostRecently = *(double *) value;
        break;
      default:
        printf("Error while copying temporal data, size = %d: SPECCODE (specstore) ->line 518.. Exit", byteSize);
        exit(-1);
    }
#endif

    if (searchThread->matriz[hashAddress][pos].state == EXPLD) {
      //  Mark this value as ELUP asap...
      searchThread->matriz[hashAddress][pos].state = ELUP;

#ifdef IV_DEBUG
      printf("SPST: Aumentando valores IV: IV[%d][%d] = %d,\n", modvector[thread], IV_tail[modvector[thread]], row);
      fflush(stdout);
#endif
      // Update IV
      IV[hashAddress][modvector[thread]][IV_tail[hashAddress][modvector[thread]]] = pos;
      IV_tail[hashAddress][modvector[thread]] = IV_tail[hashAddress][modvector[thread]] + 1;
    }
  } else { // current thread have not used the pointer
    lastData = pos;

#ifdef MAXDATA_DEBUG
    // Check if thread matrix is full (need to add data)
    //if (wheel_data[modvector[current]].lastData == wheel_data[modvector[current]].maxData)
#pragma omp flush(wheel_data)
    if (lastData >= wheel_data[modvector[current]].maxData) {
      //    reservarMasEspacio();
      printf("Error: lastData=%d >= wheel_data[modvector[current]].maxData=%d, current = %d,\n", lastData, wheel_data[modvector[current]].maxData, modvector[current]);
      fflush(stdout);
      exit(-1);
    }
#endif

#ifdef COMPLEX_STRUCTURES
    // First of all, store the value
    // addressMostRecently variables are needed because it is need to save values byte to byte
#pragma omp flush(wheel_data)
    offset = wheel_data[tidaux].actualOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);
    // Checking for data size
    memcpy(addressMostRecently, value, byteSize);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][lastData].speculativeCopy);
    switch (byteSize) {
      case 1:
        *(char *) addressMostRecently = *(char *) value;
        break;
      case 2:
        *(short *) addressMostRecently = *(short *) value;
        break;
      case 4:
        *(int *) addressMostRecently = *(int *) value;
        break;
      case 8:
        *(double *) addressMostRecently = *(double *) value;
        break;
      default:
        printf("Error while copying temporal data, size = %d: SPECCODE (specstore) ->line 518.. Exit", byteSize);
        exit(-1);
    }
#endif

    searchThread->matriz[hashAddress][lastData + 1].origPointer = NULL;
#ifdef COMPLEX_STRUCTURES
    searchThread->matriz[hashAddress][lastData].copyOffset = wheel_data[tidaux].actualOffset;
#endif
    searchThread->matriz[hashAddress][lastData].state = UPDATE;
    searchThread->matriz[hashAddress][lastData].size = byteSize;
    searchThread->matriz[hashAddress][lastData].origPointer = address;

#ifdef IV_DEBUG
    printf("SPST: Aumentando valores IV, th no used pointer: IV[%d][%d] = %d,\n", modvector[thread], IV_tail[modvector[thread]], lastData);
    fflush(stdout);
#endif
    // Update IV
    IV[hashAddress][modvector[thread]][IV_tail[hashAddress][modvector[thread]]] = lastData;
    IV_tail[hashAddress][modvector[thread]] = IV_tail[hashAddress][modvector[thread]] + 1;

#ifdef COMPLEX_STRUCTURES
    wheel_data[tidaux].actualOffset += byteSize;
#pragma omp flush(wheel_data, IV, IV_tail)
#else
#pragma omp flush(IV, IV_tail)
#endif
#ifdef COMPLEX_STRUCTURES
#ifdef OFFSET_DEBUG
    if (wheel_data[modvector[current]].actualOffset >= MAX_OFFSET) {
      printf("ERROR: Offset = %d: higher than maxoffset=%d,\n", wheel_data[modvector[current]].actualOffset, MAX_OFFSET);
      fflush(stdout);
      exit(-1);
    }
#endif
#endif
  }

  // Now, check that does not exist any successor thread that had used the
  // pointer, if this situation occurs then thread execution is discarded

  char localsquash = 'F';

  //  Look for possible RAW dependences.
  //  New since version 140: RAW dependences should be searched
  //  even if the thread is ms, since this situation could
  //  change during execution of this loop.

#ifdef SPECSTORE_DEBUG
  //  speccode: debug
  printf("Checking for squashes between %d and %d. SOY:current=%d,tid=%d.\n", current + 1, wheel_ms, current, omp_get_thread_num());
  for (thread = current + 1; thread <= wheel_ms; thread++)
    printf("\tESTADO(%d):%d\n", thread, wheel[thread]);
  fflush(stdout);
  //  speccode: end debug
#endif
  thread = current + 1;
e1122:
  //    El compilador no deja poner etiquetas justo antes de una instruccion omp
  if (1) {
#pragma omp flush(wheel_ms)
  }
  if (thread > wheel_ms)
    goto e2233;
  else {
    searchThread = wheel_data[modvector[thread]].pointer;
    pos = 0;
    while ((localsquash == 'F') && (searchThread->matriz[hashAddress][pos].origPointer != NULL)) {
#ifdef MEASURE_COUNTERS
      totalAccessesSpecstore[iden]++;
#endif
      if (searchThread->matriz[hashAddress][pos].origPointer != address) {
        pos++;
      } else {
        // Pointer exists, so it has been used and must be discarded
        if (searchThread->matriz[hashAddress][pos].state == UPDATE) {
#ifdef SPECSTORE_DEBUG
          //  speccode: debug
          printf("specstore: tid = %d, current = %d Shield found. Exiting DO loop\n", tid, current);
          fflush(stdout);
          //  speccode: end debug
#endif
          goto e2233;
        }
        if (searchThread->matriz[hashAddress][pos].state != NOTACC) {
          //      SQUASH!

          //***********************************************************************

#pragma omp critical
          {
#ifdef MEASURE_COUNTERS
            // Statistics
            SC_specstore++;
#endif

#ifdef SPECSTORE_DEBUG
            //  speccode: debug
            printf("specstore: tid = %d, current = %d, estado = %d\n", tid, current, wheel[tidaux]);
            fflush(stdout);
            //  speccode: end debug
#endif
            if ((wheel[tidaux] == RUNNING) && (thread <= wheel_ms)) {
              //          begin squash critical section

#ifdef MEASURE_COUNTERS
              //          We increment the spec_violations counter here
              spec_violations = spec_violations + 1;
#endif

              //          Squash "thread" and all its successors.
              //          If a sucessor is RUNNING, this is set as SQUASHED.
              //          If a sucessor is DONE, then is set as FREE.

#pragma omp flush(wheel_ms)

              for (lastData = thread; lastData <= wheel_ms; lastData++) {
#ifdef MEASURE_COUNTERS
                //            We increment the number of squashes performed.
                spec_squashes = spec_squashes + 1;
#endif

#pragma omp flush(wheel)
                if (wheel[modvector[lastData]] == RUNNING) {
#ifdef SPECSTORE_DEBUG
                  // speccode: debug
                  printf("Consumer %d RUNNING->SQUASHED, tid = %d\n", lastData, omp_get_thread_num());
                  fflush(stdout);
                  // speccode: end debug
#endif

                  wheel[modvector[lastData]] = SQUASHED;

                }

                if (wheel[modvector[lastData]] == DONE) {
#ifdef SPECSTORE_DEBUG
                  // speccode: debug
                  printf("Consumer %d DONE->FREE, tid = %d\n", lastData, omp_get_thread_num());
                  fflush(stdout);
                  // speccode: end debug
#endif
                  wheel[modvector[lastData]] = FREE;

                }

#pragma omp flush(wheel, wheel_ms)
              } // END FOR
#ifdef SPECSTORE_DEBUG
              // speccode: debug
              printf("Last consumer squashed= %d\n", wheel_ms);
              fflush(stdout);
              // speccode: end debug
#endif

              wheel_ms = thread - 1;
              //#pragma omp flush(wheel_ms, wheel_ns)

#ifdef SPECSTORE_DEBUG
              // speccode: debug
              printf("New most-spec thread= %d, Actual non-spec=%d.\n", wheel_ms, wheel_ns);
              fflush(stdout);
              // speccode: end debug
#endif
              localsquash = 'T';

            } //END IF((wheel[tidaux]==RUNNING)&&(thread<=wheel_ms))

          } // END OMP CRITICAL

          //***********************************************************************

          // speccode: end squash critical section.

        } // END IF(AV[index,modvector[thread]] != NOTACC)
        //else
        // PARA PONER UN DEBUG

        // Pointer located, so must break
        break;
      } // end else (searchThread->matriz[hashAddress][pos].origPointer != (baseType) address) 
      //row++;
    } // while ((localsquash == 'F') && (searchThread->matriz[hashAddress][pos].origPointer != (baseType) NULL)) 
#ifdef MEASURE_COUNTERS
    if (searchThread->matriz[hashAddress][pos].origPointer != NULL) {
      totalAccessesSpecstore[iden]++;
    }
#endif
    if (localsquash == 'T') {
      //      The squashes has been performed.
      //      We should exit the DO loop.
      goto e2233;
    }

  } // END IF(thread >= wheel_ms)

  //    v140 :  Now, we try to repeat the trick.
  thread = thread + 1;
  goto e1122;

  //    Once the searche is finished, label "e2233" is reached.
e2233:
  ;
  //   continue;

#ifdef TIME_SPECSTORE
  time_end = get_time();
  time_specstore[id] += (time_end - time_ini);
#endif

} // END specstore_pointer

/**
 ** @author alvaro estebanez
 **/
//***********************************************************************
//   specbegin(): operation to initialize blocks
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************
void specbegin(int max_iterations) {

#ifdef SPECBEGIN_DEBUG
  // Print the value of the four user parameters
  fprintf(stderr, "************************************************************\n");
  fprintf(stderr, "Speculative code executed with the following parameters:\n");
  fprintf(stderr, "Threads = %d\n", threads);
  fprintf(stderr, "Maximum number of iterations = %d\n", max_upper_limit);
  fprintf(stderr, "Block size = %d\n", blk);
  fprintf(stderr, "Initial capacity reserved to pointers in a matrix = %d\n", MAX_POINTER);
  fprintf(stderr, "************************************************************\n");
  fprintf(stderr, "\n");
#endif

  //  real*8   IEEE double-precision floating-point (Eight-byte)
  int specaux1;
  int tt;
  int maxblock;
  static int first_execution = true;

  /**
   **  initialize the number of iterations to execute in the loop
   */
  upper_limit = max_iterations;

  // Clean the number of attempts of re-executions for this execution of the block
  for (tt = 0; tt < upper_limit; tt++) {
    varblock[2][tt] = 0;
  }

  if (first_execution) {
    //  Initialize the modvector structure (only once).
    for (specaux1 = 0; specaux1 < (max_upper_limit + 8); specaux1++) {
      modvector[specaux1] = ((specaux1) % (wsize));
    }
  }

  // Initialize wheel structure for each position
  for (tt = 0; tt < wsize; tt++) {
    if (first_execution) {
      wheel_data[tt].pointer = (threadCopy_structure *) malloc(sizeof (threadCopy_structure));
      wheel_data[tt].maxData = MAX_POINTER;
    }

    for (specaux1 = 0; specaux1 < MASK + 1; specaux1++) {
      wheel_data[tt].pointer->matriz[specaux1][0].origPointer = NULL;

      // Initializing IV structures
      IV_tail[specaux1][tt] = 0;
      IV[specaux1][tt][IV_tail[specaux1][tt]] = -1;
    }

#ifdef COMPLEX_STRUCTURES
    if (first_execution) {
      wheel_data[tt].temporalData = (void *) malloc(sizeof (void) * MAX_OFFSET);
    }
    wheel_data[tt].actualOffset = 0;
#endif

    if (tt < threads) {
      wheel[tt] = RUNNING;
    } else {
      wheel[tt] = FREE;
    }

  }

  //  Initialize the wheel current limits
  wheel_ns = 0;
  wheel_ms = threads - 1;

  if (first_execution) {
    //  Depending on the value of blk, generate a fixed-block scheduling
    //  or read a variable one from disk.
    if (blk > 0) {
      //  Fixed block size: initialize vector with it.
      //  upper_limit & blk al ser enteros la division es entera
      for (maxblock = 0; maxblock <= (max_upper_limit / blk) + wsize; maxblock++) {

        //    First element: iteration to start
        varblock[0][maxblock] = blk * (maxblock);

        //    Second element: iteration to finish
        varblock[1][maxblock] = blk * (maxblock + 1) - 1;

        //    Third element: Number of attempts for this block
        //        varblock[2][maxblock] = 0;
#ifdef SPECBEGIN_DEBUG
        printf("New block: from %d to %d\n", varblock[0][maxblock], varblock[1][maxblock]);
#endif
        // Element with index 3: Number of total attempts for this block
        varblock[3][maxblock] = 0;
      }
    } else if (blk < -1) {
      for ( tt = 0, maxblock = 0; tt < threads; tt++ ) {
        varblock[0][tt] = maxblock;
        varblock[1][tt] = FIRSTCHUNK+varblock[0][tt];
        maxblock+=FIRSTCHUNK;
        maxblock++;
      }
      // Just initialize the entire squash counter column. The remaining initializations will 
      // be carried out in dynScheduler() and specstart().
      for ( tt = 0; tt < (max_upper_limit+8) ; tt++ ) {
        // Element with index 3: Number of total attempts for this block
        varblock[3][tt] = 0;
      }
    } else {
        // MESETA
        meseta(max_upper_limit, INITIAL_ZONE, FINAL_ZONE);
    }
  }

  // Initializating structures...
#ifdef TIME_SPECLOAD
  if (first_execution) {
    for (specaux1 = 0; specaux1 < threads; specaux1++) {
      time_specload[specaux1] = 0;
    }
  }
#endif
#ifdef TIME_SPECSTORE
  if (first_execution) {
    for (specaux1 = 0; specaux1 < threads; specaux1++) {
      time_specstore[specaux1] = 0;
    }
  }
#endif
#ifdef TIME_THREADEND
  if (first_execution) {
    for (specaux1 = 0; specaux1 < threads; specaux1++) {
      time_threadend[specaux1] = 0;
    }
  }
#endif
#ifdef TIME_COMMIT
  if (first_execution) {
    for (specaux1 = 0; specaux1 < threads; specaux1++) {
      time_commit[specaux1] = 0;
    }
  }
#endif
#ifdef TIME_DYNSCHEDULER
  if (first_execution) {
    for (specaux1 = 0; specaux1 < threads; specaux1++) {
      time_dynscheduler[specaux1] = 0;
    }
  }
#endif

#ifdef MEASURE_COUNTERS
  if (first_execution) {
    initializeCounters();
  }
#endif

#ifdef MEASURE_COUNTERS
  if (first_execution) {
    for (specaux1 = 0; specaux1 < 9; specaux1++) {
      tentZone[specaux1] = 0;
    }
  }
#endif

  first_execution = false;

} // end specbegin()

//***********************************************************************
//   specend(): operation to end speculation after every
//                speculative section/loop, is needed to free malloc structures
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************

void specend() {

  int thread;
  // Free structures
  for (thread = 0; thread < wsize; thread++) {
    free((void *) wheel_data[thread].pointer);
#ifdef COMPLEX_STRUCTURES
    free((void *) wheel_data[thread].temporalData);
#endif
  }

}

/**
 ** @version specstore-P
 ** @author alvaro estebanez
 **/
//*********************************************************************
//   threadend_pointer():   this operation is launch when a thread finish
//                          its block
//*********************************************************************

/**
 *
 * @param current           - Position of the thread
 * @param nonSpeculative    - To know if there are only one thread running
 * @return                  - Notify if there are some job to do
 **/
char threadend_pointer(int *current) {

#ifdef TIME_THREADEND
  double time_ini, time_end;
  int id = omp_get_thread_num();
  time_ini = get_time();
#endif

  //  Local work flags
  char auxdone;
  char auxjobtodo;
  int aux;
  char flag;
  int iter;

  //  Set work flags
  auxdone = 'F';
  auxjobtodo = 'F';
  flag = JOBTODO;

  //***********************************************************************
#pragma omp critical
  {
#ifdef MEASURE_COUNTERS
    //  Statistics
    SC_threadend1++;
#endif

#pragma omp flush(wheel)

    if (wheel[modvector[*current]] != SQUASHED) {
#pragma omp flush(wheel_ns, wheel_ms)

      // Now check if non-speculative
      if (*current == wheel_ns) {
        // No-speculative thread finished
        aux = commit();

        wheel_ns = aux + 1;

        if (varblock[1][aux] >= upper_limit - 1) {
          // Mark this thread as done
          auxdone = 'T';
        } else {
          // Move the non-speculative pointer
          //#pragma omp flush(wheel_ms, wheel_ns)
#pragma omp flush(wheel_ns)

          // If the block being executed by the most speculative
          // thread has an upper iteration that is greater than or
          // equal to the last iteration to perform, then we finish
          if (varblock[1][wheel_ms] >= upper_limit - 1)
            auxdone = 'T';
          else {
            // Only if the ms pointer plus one is pointing to a
            // FREE cell we can proceed
#pragma omp flush(wheel, wheel_ms)

            if (wheel[modvector[wheel_ms + 1]] == FREE) {
              // Clean the access vector of the cell to be used
              for (iter = 0; iter < MASK + 1; iter++) {
                wheel_data[modvector[wheel_ms + 1]].pointer->matriz[iter][0].origPointer = NULL;

                IV[iter][modvector[wheel_ms + 1]][0] = -1;
                IV_tail[iter][modvector[wheel_ms + 1]] = 0;
              }
#ifdef COMPLEX_STRUCTURES
              wheel_data[modvector[wheel_ms + 1]].actualOffset = 0;
#endif
              // Modify the status of the window cell to be used
              wheel[modvector[wheel_ms + 1]] = RUNNING;

#pragma omp flush(wheel, wheel_ms, wheel_data, IV, IV_tail)
              // Advance the wheel_ms pointer
              wheel_ms = wheel_ms + 1;
#pragma omp flush(wheel_ms)

              // Prepare the chunk for running
              auxjobtodo = 'T';
              *current = wheel_ms;

              // Calculate next chunk size
              dynScheduler(*current);

            } // END if ( wheel[modvector[wheel_ms+1]] == FREE )
          } // END else ( varblock[1,wheel_ms] >= upper_limit )
        } // END else ( varblock[1,tidaux] >= upper_limit )
      }// END if ( *current == wheel_ns )
      else {

        // Speculative thread finished
        // Set the current iteration as DONE
        wheel[modvector[*current]] = DONE;

#pragma omp flush(wheel,wheel_ms)

        // If the block being executed by the most speculative
        // thread has an upper iteration that is greater than or
        // equal to the last iteration to perform, then we finish

        if (varblock[1][wheel_ms] >= upper_limit - 1)
          auxdone = 'T';
        else {
          // Check if the cell pointed to by wheel_ms+1 is free
#pragma omp flush(wheel, wheel_ms)

          if (wheel[modvector[wheel_ms + 1]] == FREE) {
            // Clean the access vector of the cell to be used
            for (iter = 0; iter < MASK + 1; iter++) {
              wheel_data[modvector[wheel_ms + 1]].pointer->matriz[iter][0].origPointer = NULL;

              IV[iter][modvector[wheel_ms + 1]][0] = -1;
              IV_tail[iter][modvector[wheel_ms + 1]] = 0;
            }
#ifdef COMPLEX_STRUCTURES
            wheel_data[modvector[wheel_ms + 1]].actualOffset = 0;
#endif
            // Modify the status of the window cell to be used
            wheel[modvector[wheel_ms + 1]] = RUNNING;

#pragma omp flush(wheel, wheel_ms, wheel_data, IV, IV_tail)

            // Increment the most speculative pointer
            wheel_ms = wheel_ms + 1;

#pragma omp flush(wheel_ms)

            // Prepare the chunk for running
            auxjobtodo = 'T';
            *current = wheel_ms;

            // Calculate next chunk size
            dynScheduler(*current);

          } // END  if ( wheel[modvector[wheel_ms+1]] == FREE )
        } // END else ( varblock[1,wheel_ms] >= upper_limit)
      } // END else (current == wheel_ns)
    }// END  if ( wheel[modvector[current]] != SQUASHED )
    else {
      //    Squashed thread

      //    Mark the current chunk as FREE
      wheel[modvector[*current]] = FREE;

      auxdone = 'F';
      auxjobtodo = 'F';

#pragma omp flush(wheel)

      // (Note that the ms pointer was moved backwards by the thread that
      // performs the squash operation.)
    } // END else ( wheel[modvector[*current]] != SQUASHED )

  } // END  #pragma OMP CRITICAL

  //***********************************************************************

  // Now check what to do next

  if (auxdone == 'T') {
    flag = JOBDONE;
#ifdef TIME_THREADEND
    time_end = get_time();
    time_threadend[id] += (time_end - time_ini);
#endif
    return flag;
  }

  if (auxjobtodo == 'T') {
    flag = JOBTODO;
#ifdef TIME_THREADEND
    time_end = get_time();
    time_threadend[id] += (time_end - time_ini);
#endif
    return flag;
  }

  // If we are here, we cannot start new thread, so spin-wait
#pragma omp flush(wheel, wheel_ms)

e456:
  if (wheel[modvector[wheel_ms + 1]] != FREE) //SPIN1
  {
#pragma omp flush(wheel, wheel_ms)  //SPIN2
    goto e456;
    // SPIN3
  }

  //***********************************************************************
#pragma omp critical
  {
#ifdef MEASURE_COUNTERS
    //  Statistics
    SC_threadend2++;
#endif

    // Window moved, can start new thread now
    auxdone = 'F';

    // Double check that the new cell is really free and
    // that there are more iterations to run
    if ((wheel[modvector[wheel_ms + 1]] == FREE) && (varblock[1][wheel_ms] < upper_limit - 1)) {
      // Clean the access vector of the cell to be used
      for (iter = 0; iter < MASK + 1; iter++) {
        wheel_data[modvector[wheel_ms + 1]].pointer->matriz[iter][0].origPointer = NULL;

        IV[iter][modvector[wheel_ms + 1]][0] = -1;
        IV_tail[iter][modvector[wheel_ms + 1]] = 0;
      }
#ifdef COMPLEX_STRUCTURES
      wheel_data[modvector[wheel_ms + 1]].actualOffset = 0;
#endif

#pragma omp flush(wheel_data,wheel_ms,IV,IV_tail)
      // Increment the most speculative pointer
      wheel_ms = wheel_ms + 1;

#pragma omp flush(wheel_ms)

      // Prepare the chunk for running
      auxjobtodo = 'T';
      *current = wheel_ms;

      // Calculate next chunk size
      dynScheduler(*current);

      wheel[modvector[wheel_ms]] = RUNNING;

#pragma omp flush(wheel)
    }// END if
    else {
      // The new cell is not actually free, check out if done

      // If the block being executed by the most speculative
      // thread has an upper iteration that is greater than or
      // equal to the last iteration to perform, then we finish
      if (varblock[1][wheel_ms] >= upper_limit - 1) {
        auxdone = 'T';
        auxjobtodo = 'T';
      } else {
        auxjobtodo = 'F';
        auxdone = 'F';
      }
    }
  } // END #pragma OMP CRITICAL

  //***********************************************************************

  // Spin wait if no job to run

  if (auxjobtodo == 'F') {
#pragma omp flush(wheel, wheel_ms)
    goto e456;
  }

  // Jump to end if iterations finished

  if (auxdone == 'T') {
    flag = JOBDONE;
#ifdef TIME_THREADEND
    time_end = get_time();
    time_threadend[id] += (time_end - time_ini);
#endif
    return flag;
  }

#ifdef TIME_THREADEND
  time_end = get_time();
  time_threadend[id] += (time_end - time_ini);
#endif
  // GO to work
  flag = JOBTODO;
  return flag;

} // END threadend_pointer()

//***********************************************************************
//  commit(): commit operation for aggressive sliding window, without
//            checking for violations and with reduction support
//***********************************************************************

/**
 *
 * @return - the highest position of threads committed
 */
int commit() {

#ifdef TIME_COMMIT
  double time_ini, time_end;
  int id = omp_get_thread_num();
  time_ini = get_time();
#endif

  // Local variables
  int localindex, localtid, mytid;
  int specaux1, specaux2, specaux3;
  void *addressToSave;
  void *addressMostRecently;
  threadCopy_structure *doneThread;
#ifdef COMPLEX_STRUCTURES
  int offset;
#endif

#ifdef MEASURE_COUNTERS
  //  Statistics
  SC_commit++;
#endif

  // Mark the iteration as DONE
  wheel[modvector[wheel_ns]] = DONE;
#pragma omp flush(wheel)

  // Scan wheel to find most advanced DONE iteration
  for (mytid = wheel_ns; mytid <= (wheel_ms - 1); mytid++) {
    if ((wheel[modvector[mytid]] == DONE) && (wheel[modvector[mytid + 1]] != DONE)) {
      break;
    }
  }

  // Check for corner case: the most speculative can also be DONE
  if ((mytid == (wheel_ms - 1)) && (wheel[modvector[wheel_ms]] == DONE)) {
    mytid = mytid + 1;
  }

#ifdef COMMIT_DEBUG
  //  speccode: debug
  printf("commit: %d Committing iterations %d to %d\n", current, wheel_ns, mytid);
  fflush(stdout);
  //  speccode: end debug
#endif
#pragma omp flush(wheel_ns)
  for (specaux1 = wheel_ns; specaux1 <= mytid; specaux1++) {

    localtid = modvector[specaux1];
#pragma omp flush (wheel_data, IV, IV_tail)
    doneThread = wheel_data[localtid].pointer;
    for (specaux2 = 0; specaux2 < MASK + 1; specaux2++) {
      for (specaux3 = 0; specaux3 < IV_tail[specaux2][localtid]; specaux3++) {

        localindex = IV[specaux2][localtid][specaux3];

        // All the elements are distinct to NOTACC
#ifdef COMMIT_DEBUG
        //  speccode: debug
        printf("commit: %d Committing element %d of thread %d, pos=%d.\n", current, specaux2, specaux1, specaux3);
        fflush(stdout);
        //  speccode: end debug
#endif

        if ((doneThread->matriz[specaux2][localindex].state == ELUP) || (doneThread->matriz[specaux2][localindex].state == UPDATE)) {
          // Commit the updated version

#ifdef COMPLEX_STRUCTURES
          offset = doneThread->matriz[specaux2][localindex].copyOffset;
          addressMostRecently = (void *) &(wheel_data[localtid].temporalData[offset]);
          addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;
          memcpy(addressToSave, addressMostRecently, doneThread->matriz[specaux2][localindex].size);
#else
          addressMostRecently = (void *) &(doneThread->matriz[specaux2][localindex].speculativeCopy);
          addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;

          // Check data size
          switch (doneThread->matriz[specaux2][localindex].size) {
            case 1:
              *(char *) addressToSave = *(char *) addressMostRecently;
              break;
            case 2:
              *(short *) addressToSave = *(short *) addressMostRecently;
              break;
            case 4:
              *(int *) addressToSave = *(int *) addressMostRecently;
              break;
            case 8:
              *(double *) addressToSave = *(double *) addressMostRecently;
              break;
            default:
              printf("Error while committing temporal data: size =%d, SPECCODE (commit)->line 1316... Exit",
                  doneThread->matriz[specaux2][IV[specaux2][localtid][specaux3]].size);
              exit(-1);
          }
#endif
        }// if ((doneThread->matriz[specaux2][localindex].state == ELUP) || (doneThread->matriz[specaux2][localindex].state == UPDATE))
        else {

          if (doneThread->matriz[specaux2][localindex].state == REDADD) {
            // ADD reduction operation

#ifdef COMPLEX_STRUCTURES
            offset = doneThread->matriz[specaux2][localindex].copyOffset;
            addressMostRecently = (void *) &(wheel_data[localtid].temporalData[offset]);
            addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;
#else
            addressMostRecently = (void *) &(doneThread->matriz[specaux2][localindex].speculativeCopy);
            addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;
#endif
            // Check data size
            switch (doneThread->matriz[specaux2][localindex].size) {
              case 1:
                *(char *) addressToSave = *(char *) addressMostRecently + *(char *) addressToSave;
                break;
              case 2:
                *(short *) addressToSave = *(short *) addressMostRecently + *(short *) addressToSave;
                break;
              case 4:
                *(int *) addressToSave = *(int *) addressMostRecently + *(int *) addressToSave;
                break;
              case 8:
                *(double *) addressToSave = *(double *) addressMostRecently + *(double *) addressToSave;
                break;
              default:
                printf("Error in commit: Unknown size = %d\n", doneThread->matriz[specaux2][localindex].size);
                exit(-1);
            }
          } else {
            // MAX reduction operation?
            if (doneThread->matriz[specaux2][localindex].state == REDMAX) {
#ifdef COMPLEX_STRUCTURES
              offset = doneThread->matriz[specaux2][localindex].copyOffset;
              addressMostRecently = (void *) &(wheel_data[localtid].temporalData[offset]);
              addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;
#else
              addressMostRecently = (void *) &(doneThread->matriz[specaux2][localindex].speculativeCopy);
              addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;
#endif
              // addressToSave = max(addressToSave,addressMostRecently)
              // Check data size
              switch (doneThread->matriz[specaux2][localindex].size) {
                case 1:
                  if (*(char *) addressToSave < *(char *) addressMostRecently)
                    *(char *) addressToSave = *(char *) addressMostRecently;
                  break;
                case 2:
                  if (*(short *) addressToSave < *(short *) addressMostRecently)
                    *(short *) addressToSave = *(short *) addressMostRecently;
                  break;
                case 4:
                  if (*(int *) addressToSave < *(int *) addressMostRecently)
                    *(int *) addressToSave = *(int *) addressMostRecently;
                  break;
                case 8:
                  if (*(double *) addressToSave < *(double *) addressMostRecently)
                    *(double *) addressToSave = *(double *) addressMostRecently;
                  break;
                default:
                  fprintf(stderr,"Error in commit: Unknown size = %d\n", doneThread->matriz[specaux2][localindex].size);
                  exit(-1);
              }
            } else {
              // MIN reduction operation?
              if (doneThread->matriz[specaux2][localindex].state == REDMIN) {
#ifdef COMPLEX_STRUCTURES
                offset = doneThread->matriz[specaux2][localindex].copyOffset;
                addressMostRecently = (void *) &(wheel_data[localtid].temporalData[offset]);
                addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;
#else
                addressMostRecently = (void *) &(doneThread->matriz[specaux2][localindex].speculativeCopy);
                addressToSave = (void *) doneThread->matriz[specaux2][localindex].origPointer;
#endif
                // addressToSave = max(addressToSave,addressMostRecently)
                // Check data size
                switch (doneThread->matriz[specaux2][localindex].size) {
                  case 1:
                    if (*(char *) addressToSave > *(char *) addressMostRecently)
                      *(char *) addressToSave = *(char *) addressMostRecently;
                    break;
                  case 2:
                    if (*(short *) addressToSave > *(short *) addressMostRecently)
                      *(short *) addressToSave = *(short *) addressMostRecently;
                    break;
                  case 4:
                    if (*(int *) addressToSave > *(int *) addressMostRecently)
                      *(int *) addressToSave = *(int *) addressMostRecently;
                    break;
                  case 8:
                    if (*(double *) addressToSave > *(double *) addressMostRecently)
                      *(double *) addressToSave = *(double *) addressMostRecently;
                    break;
                  default:
                    fprintf(stderr,"Error in commit: Unknown size = %d\n", doneThread->matriz[specaux2][localindex].size);
                    exit(-1);
                }
              }
            }
          }
        }
      } // END for (specaux3)
    } // END for ( specaux2= 0 )
    //    Mark the freed cell as FREE
    wheel[localtid] = FREE;
#pragma omp flush(wheel)

    //      Accumulate number of re-executions of this block of iterations to accumulated results.
    varblock[3][specaux1] += varblock[2][specaux1];
    //      The number of re-executions of this block will be cleared at specstart()

    // Done with this thread, proceed with next.
  } // END for ( specaux1 = wheel_ns )

#ifdef TIME_COMMIT
  time_end = get_time();
  time_commit[id] += (time_end - time_ini);
#endif

  return mytid;

} // END commit()

#ifdef MEASURE_COUNTERS
//***********************************************************************
//   initializeCounters(): operation to initialize counters
//***********************************************************************
//    No need to use FLUSH: this is executed outside parallel region
//***********************************************************************

void initializeCounters() {
  int specaux1;

  spec_violations = 0;
  spec_squashes = 0;
  SC_commit = 0;
  SC_threadend1 = 0;
  SC_threadend2 = 0;
  SC_specstore = 0;
  for (specaux1 = 0; specaux1 < threads; specaux1++) {
    totalAccessesSpecload[specaux1] = 0;
    totalCallsSpecload[specaux1] = 0;
    totalAccessesSpecstore[specaux1] = 0;
    totalCallsSpecstore[specaux1] = 0;
  }
}
#endif

#ifdef MEASURE_COUNTERS
//***********************************************************************
//  specepilogue(): Show total number of violations and squashes.
//            (should be called as last instruction of the application)
//***********************************************************************

void specepilogue() {

  int i;

  fprintf(stderr,"------------------------------------------\n");
  fprintf(stderr,"Scheduling mechanism used: ");
  switch (blk) {
    case MESETA:
      fprintf(stderr,"MESETA function.\n"); 
      fprintf(stderr,"MAX_ITER = %d\n",MAX_ITER); 
      fprintf(stderr,"Percentage of iterations in the initial zone = %d\n",INITIAL_ZONE); 
      fprintf(stderr,"Percentage of iterations in the final zone = %d\n",FINAL_ZONE); 
      break;
    case DYN_PDP1:
      fprintf(stderr,"PDP(1) function, dynamic version.\n"); 
      break;
    case ADA_PDP1:
      fprintf(stderr,"PDP(1) function, adaptive version.\n"); 
      break;
    case DYN_PDP2:
      fprintf(stderr,"PDP(2) function, dynamic version.\n"); 
      break;
    case ADA_PDP2:
      fprintf(stderr,"PDP(2) function, adaptive version.\n"); 
      break;
    case DYN_TENT:
      fprintf(stderr,"TENT function, dynamic version.\n"); 
      fprintf(stderr,"First chunk size = %d\n", varblock[1][0]); 
      fprintf(stderr,"ALPHA = %f\n", ALPHA); 
      fprintf(stderr,"BETA = %f\n", BETA); 
      break;
    case ADA_TENT:
      fprintf(stderr,"TENT function, adaptive version.\n"); 
      fprintf(stderr,"First chunk size = %d\n", varblock[1][0]); 
      fprintf(stderr,"ALPHA = %f\n", ALPHA); 
      fprintf(stderr,"BETA = %f\n", BETA); 
      break;
    default:
      fprintf(stderr,"Fixed block size = %d.\n",blk); 
      break;
  }
  fprintf(stderr,"------------------------------------------\n");
  //fprintf(stderr,"Block number, block size, re-executions for last loop, accumulated re-executions\n");
  fprintf(stderr,"B_num\tb_ini\tb_fin\tb_size\tb_re-ex\tb_accumulated re-executions\n");

  i = 0;
  while (varblock[2][i] != 0) {
    fprintf(stderr,"%d\t%d\t%d\t%d\t%d\t%d\n",i, varblock[0][i],varblock[1][i],varblock[1][i]-varblock[0][i]+1, varblock[2][i], varblock[3][i]);
    i++;
  }

  fprintf(stderr,"------------------------------------------\n");

#ifdef MEASURE_COUNTERS
  fprintf(stderr,"Total accesses to different zones of TENT:\n");
  fprintf(stderr,"T128:\t%d\n",tentZone[T128]);
  fprintf(stderr,"T028:\t%d\n",tentZone[T028]);
  fprintf(stderr,"T023:\t%d\n",tentZone[T023]);
  fprintf(stderr,"T034:\t%d\n",tentZone[T034]);
  fprintf(stderr,"T046:\t%d\n",tentZone[T046]);
  fprintf(stderr,"T078:\t%d\n",tentZone[T078]);
  fprintf(stderr,"T067:\t%d\n",tentZone[T067]);
  fprintf(stderr,"T456:\t%d\n",tentZone[T456]);
  fprintf(stderr,"T35inf:\t%d\n",tentZone[T35inf]);
#endif

  fprintf(stderr,"------------------------------------------\n");

  fprintf(stderr,"Total violations: %d ; Total squashes: %d\n", spec_violations, spec_squashes);
  fprintf(stderr,"Access specstore SC: %d \n", SC_specstore);
  fprintf(stderr,"Access threadend1 SC: %d \n", SC_threadend1);
  fprintf(stderr,"Access threadend2 SC: %d \n", SC_threadend2);
  fprintf(stderr,"Access commit SC: %d \n", SC_commit);


  //    printf("------------------------------------------\n");
  //    printf("Block number\tblock size\tnumber of executions\n");
  //
  //    for (i = 0; i <= (upper_limit / blk) + 11; i++)
  //        printf("\t%d\t\t%d\t\t%d\n", i, varblock[1][i] - varblock[0][i] + 1, varblock[2][i]);
  //
  //    printf("------------------------------------------\n");
  //
  //    printf("Total violations: %d ; Total squashes: %d\n", spec_violations, spec_squashes);
  //    printf("Access specstore SC: %d \n", SC_specstore);
  //    printf("Access threadend1 SC: %d \n", SC_threadend1);
  //    printf("Access threadend2 SC: %d \n", SC_threadend2);
  //    printf("Access commit SC: %d \n", SC_commit);

} // END specepilogue()
#endif

/**************************************************************************/
/*  			Functions to get and print time 		  */

/**************************************************************************/
double get_time() {
  struct timeval tv;
  gettimeofday(&tv, (void *) 0);
  return (tv.tv_sec + 1.0e-6 * tv.tv_usec);
}

#ifdef TIME_SPECSTORE

void printSpecstoreTime() {
  int i;
  double cont = 0;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"Specstore time\n");
  fprintf(stderr,"------------------------------------\n");

  for (i = 0; i < threads; i++) {
    fprintf(stderr,"Specstore time for thread %d: %f\n", i, time_specstore[i]);
    cont += time_specstore[i];
  }

  fprintf(stderr,"------------------------------------\n");
  fprintf(stderr,"Specstore time global: %f\n", cont);
  fprintf(stderr,"Specstore time average: %f\n", (cont / threads));
  fprintf(stderr,"------------------------------------\n\n");

}
#endif

#ifdef TIME_THREADEND

void printThreadendTime() {
  int i;
  double cont = 0;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"Threadend time\n");
  fprintf(stderr,"------------------------------------\n");

  for (i = 0; i < threads; i++) {
    fprintf(stderr,"Threadend time for thread %d: %f\n", i, time_threadend[i]);
    cont += time_threadend[i];
  }

  fprintf(stderr,"------------------------------------\n");
  fprintf(stderr,"Threadend time global: %f\n", cont);
  fprintf(stderr,"Threadend time average: %f\n", (cont / threads));
  fprintf(stderr,"------------------------------------\n\n");

}
#endif

#ifdef TIME_COMMIT

void printCommitTime() {
  int i;
  double cont = 0;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"Commit time\n");
  fprintf(stderr,"------------------------------------\n");

  for (i = 0; i < threads; i++) {
    fprintf(stderr,"Commit time for thread %d: %f\n", i, time_commit[i]);
    cont += time_commit[i];
  }

  fprintf(stderr,"------------------------------------\n");
  fprintf(stderr,"Commit time global: %f\n", cont);
  fprintf(stderr,"Commit time average: %f\n", (cont / threads));
  fprintf(stderr,"------------------------------------\n\n");

}
#endif

#ifdef TIME_SPECLOAD

void printSpecloadTime() {
  int i;
  double cont = 0;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"Specload time\n");
  fprintf(stderr,"------------------------------------\n");

  for (i = 0; i < threads; i++) {
    fprintf(stderr,"Specload time for thread %d: %f\n", i, time_specload[i]);
    cont += time_specload[i];
  }

  fprintf(stderr,"------------------------------------\n");
  fprintf(stderr,"Specload time global: %f\n", cont);
  fprintf(stderr,"Specload time average: %f\n", (cont / threads));
  fprintf(stderr,"------------------------------------\n\n");

}
#endif

#ifdef TIME_DYNSCHEDULER

void printDynSchedulerTime() {
  int i;
  double cont = 0;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"dynScheduler time\n");
  fprintf(stderr,"------------------------------------\n");

  for (i = 0; i < threads; i++) {
    fprintf(stderr,"dynScheduler time for thread %d: %f\n", i, time_dynscheduler[i]);
    cont += time_dynscheduler[i];
  }

  fprintf(stderr,"------------------------------------\n");
  fprintf(stderr,"dynScheduler time global: %f\n", cont);
  fprintf(stderr,"dynScheduler time average: %f\n", (cont / threads));
  fprintf(stderr,"------------------------------------\n\n");

}
#endif

#ifdef MEASURE_COUNTERS

void printAccesses() {

  long long int calls = 0;
  long long int accesses = 0;
  int i;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"Specload() accesses\n");
  fprintf(stderr,"------------------------------------\n");
  for (i = 0; i < threads; i++) {
    fprintf(stderr,"Specload calls for thread %d, %lld\nSpecload accesses for thread %d: %lld\n\n", i, totalCallsSpecload[i], i,
        totalAccessesSpecload[i]);
    calls += totalCallsSpecload[i];
    accesses += totalAccessesSpecload[i];
  }
  float average = (float) accesses / (float) calls;
  fprintf(stderr,"Total accesses: %lld\tTotal calls: %lld\tAverage accesses: %f\n", accesses, calls, average);
  fprintf(stderr,"------------------------------------\n");
  calls = 0;
  accesses = 0;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"Specstore() accesses\n");
  fprintf(stderr,"------------------------------------\n");
  for (i = 0; i < threads; i++) {
    fprintf(stderr,"Specstore calls for thread %d, %lld\nSpecstore accesses for thread %d: %lld\n\n", i, totalCallsSpecstore[i], i,
        totalAccessesSpecstore[i]);
    calls += totalCallsSpecstore[i];
    accesses += totalAccessesSpecstore[i];
  }
  average = (float) accesses / (float) calls;
  fprintf(stderr,"Total accesses: %lld\tTotal calls: %lld\tAverage accesses: %f\n", accesses, calls, average);
  fprintf(stderr,"------------------------------------\n");
  calls = 0;
  accesses = 0;

  fprintf(stderr,"\n------------------------------------\n");
  fprintf(stderr,"Specredadd() accesses\n");
  fprintf(stderr,"------------------------------------\n");
  for (i = 0; i < threads; i++) {
    fprintf(stderr,"Specredadd calls for thread %d, %lld\nSpecredadd accesses for thread %d: %lld\n\n", i, totalCallsSpecredadd[i], i,
        totalAccessesSpecredadd[i]);
    calls += totalCallsSpecredadd[i];
    accesses += totalAccessesSpecredadd[i];
  }
  average = (float) accesses / (float) calls;
  fprintf(stderr,"Total accesses: %lld\tTotal calls: %lld\tAverage accesses: %f\n", accesses, calls, average);
  fprintf(stderr,"------------------------------------\n");
}
#endif

/* REDUCTION OPERATIONS */

//*********************************************************************
//   specredadd(): speculative ADD reduction operation
//*********************************************************************

void specredadd(void* address, unsigned int byteSize, unsigned int current, void* value) {

  // local variables
  int usedAddressFlag = false;
  int thread = current;
  baseType hashAddress = ((baseType) address >> 3) & MASK;
  int pos = 0;
  int lastData = -1;
  threadCopy_structure *searchThread;
  void *addressMostRecently;
#ifdef COMPLEX_STRUCTURES
  int offset;
#endif
  int tidaux = modvector[current];

  //  speccode: debug
  //  tid = omp_get_thread_num();
  //  printf("specredadd: tid = %d, current = %d Beginning ADD reduction\n",tid,current);
  //  speccode: end debug

#ifdef MEASURE_COUNTERS
  int iden = omp_get_thread_num();
  totalCallsSpecredadd[iden]++;
#endif

  // searchThread should initially point to the current data matrix
  searchThread = wheel_data[tidaux].pointer;

  // check if the direction where I want to save data, has been used by current thread
  while (searchThread->matriz[hashAddress][pos].origPointer != NULL) {
#ifdef MEASURE_COUNTERS
    totalAccessesSpecredadd[iden]++;
#endif
    if (searchThread->matriz[hashAddress][pos].origPointer != address) {
      pos++;
    } else {
      usedAddressFlag = true;
#ifdef MEASURE_COUNTERS
      // se va a sumar un elemento no utilizado
      totalAccessesSpecredadd[iden]--;
#endif
      break;
      // pos variable will contain the position of the pointer in its matrix
    }
  }
#ifdef MEASURE_COUNTERS
  totalAccessesSpecredadd[iden]++;
#endif
  // NOTE: pos variable will contain the position of the pointer in its matrix

  if (usedAddressFlag) {
    //  We assume here that the state is REDADD, thus the value,
#ifdef SPECREDADD_DEBUGG
    if (searchThread->matriz[hashAddress][pos].state != REDADD) {
      printf("ERROR in specredadd: action not implemented.\n");
      exit(-1);
    }
#endif

    //  IV and AV have been updated previously.

#ifdef SPECREDADD_DEBUG
    //  speccode: debug
    printf("specredadd: tid = %d, current = %d Simply accumulating the value\n", tid, current);
    fflush(stdout);
    //  speccode: end debug
#endif

    //  Thus, we only accumulate the value

    // First of all, load the value
    // Checking for data size
#ifdef COMPLEX_STRUCTURES
    offset = searchThread->matriz[hashAddress][pos].copyOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][pos].speculativeCopy);
#endif
    switch (byteSize) {
      case 1:
        *(char *) addressMostRecently = *(char *) addressMostRecently + *(char *) value;
        break;
      case 2:
        *(short *) addressMostRecently = *(short *) addressMostRecently + *(short *) value;
        break;
      case 4:
        *(int *) addressMostRecently = *(int *) addressMostRecently + *(int *) value;
        break;
      case 8:
        *(double *) addressMostRecently = *(double *) addressMostRecently + *(double *) value;
        break;
      default:
        printf("Error in specredadd: Unknown size = %d\n", byteSize);
        exit(-1);
    }
  } else {

    // current thread have not used the pointer
    lastData = pos;

#ifdef MAXDATA_DEBUG
    // Check if thread matrix is full (need to add data)
#pragma omp flush(wheel_data)
    if (lastData >= wheel_data[tidaux].maxData) {
      //    reservarMasEspacio();
      printf("Error in specredadd: lastData=%d >= wheel_data[modvector[current]].maxData=%d, current = %d,\n", lastData, wheel_data[modvector[current]].maxData, modvector[current]);
      fflush(stdout);
      exit(-1);
    }
#endif

#ifdef COMPLEX_STRUCTURES
    // First of all, store the value
    // addressMostRecently variables are needed because it is need to save values byte to byte
#pragma omp flush(wheel_data)
    offset = wheel_data[tidaux].actualOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);
    // Checking for data size
    memcpy(addressMostRecently, value, byteSize);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][lastData].speculativeCopy);
    switch (byteSize) {
      case 1:
        *(char *) addressMostRecently = *(char *) value;
        break;
      case 2:
        *(short *) addressMostRecently = *(short *) value;
        break;
      case 4:
        *(int *) addressMostRecently = *(int *) value;
        break;
      case 8:
        *(double *) addressMostRecently = *(double *) value;
        break;
      default:
        printf("Error in specredadd: while copying temporal data, unknown size = %d\n, Exit...\n", byteSize);
        exit(-1);
    }
#endif

    searchThread->matriz[hashAddress][lastData + 1].origPointer = NULL;
#ifdef COMPLEX_STRUCTURES
    searchThread->matriz[hashAddress][lastData].copyOffset = wheel_data[tidaux].actualOffset;
#endif
    searchThread->matriz[hashAddress][lastData].state = REDADD;
    searchThread->matriz[hashAddress][lastData].size = byteSize;
    searchThread->matriz[hashAddress][lastData].origPointer = address;

#ifdef IV_DEBUG
    printf("SPST: Aumentando valores IV, th no used pointer: IV[%d][%d] = %d,\n", modvector[thread], IV_tail[modvector[thread]], lastData);
    fflush(stdout);
#endif
    // Update IV
    IV[hashAddress][modvector[thread]][IV_tail[hashAddress][modvector[thread]]] = lastData;
    IV_tail[hashAddress][modvector[thread]] = IV_tail[hashAddress][modvector[thread]] + 1;

#ifdef COMPLEX_STRUCTURES
    wheel_data[tidaux].actualOffset += byteSize;
#pragma omp flush(wheel_data, IV, IV_tail)
#else
#pragma omp flush(IV, IV_tail)
#endif
#ifdef COMPLEX_STRUCTURES
#ifdef OFFSET_DEBUG
    if (wheel_data[modvector[current]].actualOffset >= MAX_OFFSET) {
      printf("ERROR in specredadd: Offset = %d: higher than maxoffset=%d,\n", wheel_data[modvector[current]].actualOffset, MAX_OFFSET);
      fflush(stdout);
      exit(-1);
    }
#endif
#endif

  }

#ifdef SPECREDADD_DEBUG
  //  speccode: debug
  printf("specredadd: tid = %d, current = %d Ending ADD reduction\n", tid, current);
  fflush(stdout);
  //  speccode: end debug
#endif

} // END specredadd

//*********************************************************************
//   specredmax(): speculative MAX reduction operation
//*********************************************************************

void specredmax(void* address, unsigned int byteSize, unsigned int current, void* value) {

  // local variables
  int usedAddressFlag = false;
  int thread = current;
  baseType hashAddress = ((baseType) address >> 3) & MASK;
  int pos = 0;
  int lastData = -1;
  threadCopy_structure *searchThread;
  void *addressMostRecently;
#ifdef COMPLEX_STRUCTURES
  int offset;
#endif
  int tidaux = modvector[current];

  //  speccode: debug
  //  tid = omp_get_thread_num();
  //  printf("specredmax: tid = %d, current = %d Beginning MAX reduction\n",tid,current);
  //  speccode: end debug

#ifdef MEASURE_COUNTERS
  int iden = omp_get_thread_num();
  totalCallsSpecredmax[iden]++;
#endif

  // searchThread should initially point to the current data matrix
  searchThread = wheel_data[tidaux].pointer;

  // check if the direction where I want to save data, has been used by current thread
  while (searchThread->matriz[hashAddress][pos].origPointer != NULL) {
#ifdef MEASURE_COUNTERS
    totalAccessesSpecredmax[iden]++;
#endif
    if (searchThread->matriz[hashAddress][pos].origPointer != address) {
      pos++;
    } else {
      usedAddressFlag = true;
#ifdef MEASURE_COUNTERS
      // se va a sumar un elemento no utilizado
      totalAccessesSpecredmax[iden]--;
#endif
      break;
      // pos variable will contain the position of the pointer in its matrix
    }
  }
#ifdef MEASURE_COUNTERS
  totalAccessesSpecredmax[iden]++;
#endif
  // NOTE: pos variable will contain the position of the pointer in its matrix

  if (usedAddressFlag) {
    //  We assume here that the state is REDMAX, thus the value,
#ifdef SPECREDADD_DEBUG
    if (searchThread->matriz[hashAddress][pos].state != REDMAX) {
      printf("ERROR in specredmax: action not implemented.\n");
      exit(-1);
    }
#endif

    //  IV and AV have been updated previously.

#ifdef SPECREDMAX_DEBUG
    //  speccode: debug
    printf("specredmax: tid = %d, current = %d Simply perform MAX operation\n", tid, current);
    fflush(stdout);
    //  speccode: end debug
#endif

    //  Thus, we perform MAX operation

    // First of all, load the value
    // Checking for data size
#ifdef COMPLEX_STRUCTURES
    offset = searchThread->matriz[hashAddress][pos].copyOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][pos].speculativeCopy);
#endif
    switch (byteSize) {
      case 1:
        if (*(char *) addressMostRecently < *(char *) value)
          *(char *) addressMostRecently = *(char *) value;
        break;
      case 2:
        if (*(short *) addressMostRecently < *(short *) value)
          *(short *) addressMostRecently = *(short *) value;
        break;
      case 4:
        if (*(int *) addressMostRecently < *(int *) value)
          *(int *) addressMostRecently = *(int *) value;
        break;
      case 8:
        if (*(double *) addressMostRecently < *(double *) value)
          *(double *) addressMostRecently = *(double *) value;
        break;
      default:
        fprintf(stderr,"Error in specredmax: Unknown size = %d\n", byteSize);
        exit(-1);
    }
  } else {

    // current thread have not used the pointer
    lastData = pos;

#ifdef MAXDATA_DEBUG
    // Check if thread matrix is full (need to add data)
#pragma omp flush(wheel_data)
    if (lastData >= wheel_data[tidaux].maxData) {
      //    reservarMasEspacio();
      printf("Error in specredadd: lastData=%d >= wheel_data[modvector[current]].maxData=%d, current = %d,\n", lastData, wheel_data[modvector[current]].maxData, modvector[current]);
      fflush(stdout);
      exit(-1);
    }
#endif

#ifdef COMPLEX_STRUCTURES
    // First of all, store the value
    // addressMostRecently variables are needed because it is need to save values byte to byte
#pragma omp flush(wheel_data)
    offset = wheel_data[tidaux].actualOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);
    // Checking for data size
    memcpy(addressMostRecently, value, byteSize);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][lastData].speculativeCopy);
    switch (byteSize) {
      case 1:
        *(char *) addressMostRecently = *(char *) value;
        break;
      case 2:
        *(short *) addressMostRecently = *(short *) value;
        break;
      case 4:
        *(int *) addressMostRecently = *(int *) value;
        break;
      case 8:
        *(double *) addressMostRecently = *(double *) value;
        break;
      default:
        printf("Error in specredadd: while copying temporal data, unknown size = %d\n, Exit...\n", byteSize);
        exit(-1);
    }
#endif

    searchThread->matriz[hashAddress][lastData + 1].origPointer = NULL;
#ifdef COMPLEX_STRUCTURES
    searchThread->matriz[hashAddress][lastData].copyOffset = wheel_data[tidaux].actualOffset;
#endif
    searchThread->matriz[hashAddress][lastData].state = REDMAX;
    searchThread->matriz[hashAddress][lastData].size = byteSize;
    searchThread->matriz[hashAddress][lastData].origPointer = address;

#ifdef IV_DEBUG
    printf("SPST: Aumentando valores IV, th no used pointer: IV[%d][%d] = %d,\n", modvector[thread], IV_tail[modvector[thread]], lastData);
    fflush(stdout);
#endif
    // Update IV
    IV[hashAddress][modvector[thread]][IV_tail[hashAddress][modvector[thread]]] = lastData;
    IV_tail[hashAddress][modvector[thread]] = IV_tail[hashAddress][modvector[thread]] + 1;

#ifdef COMPLEX_STRUCTURES
    wheel_data[tidaux].actualOffset += byteSize;
#pragma omp flush(wheel_data, IV, IV_tail)
#else
#pragma omp flush(IV, IV_tail)
#endif
#ifdef COMPLEX_STRUCTURES
#ifdef OFFSET_DEBUG
    if (wheel_data[modvector[current]].actualOffset >= MAX_OFFSET) {
      printf("ERROR in specredmax: Offset = %d: higher than maxoffset=%d,\n", wheel_data[modvector[current]].actualOffset, MAX_OFFSET);
      fflush(stdout);
      exit(-1);
    }
#endif
#endif

  }

#ifdef SPECREDMAX_DEBUG
  //  speccode: debug
  printf("specredmax: tid = %d, current = %d Ending MAX reduction\n", tid, current);
  fflush(stdout);
  //  speccode: end debug
#endif

} // END specredmax

//*********************************************************************
//   specredmin(): speculative MIN reduction operation
//*********************************************************************

void specredmin(void* address, unsigned int byteSize, unsigned int current, void* value) {

  // local variables
  int usedAddressFlag = false;
  int thread = current;
  baseType hashAddress = ((baseType) address >> 3) & MASK;
  int pos = 0;
  int lastData = -1;
  threadCopy_structure *searchThread;
  void *addressMostRecently;
#ifdef COMPLEX_STRUCTURES
  int offset;
#endif
  int tidaux = modvector[current];

  //  speccode: debug
  //  tid = omp_get_thread_num();
  //  printf("specredmin: tid = %d, current = %d Beginning MIN reduction\n",tid,current);
  //  speccode: end debug

#ifdef MEASURE_COUNTERS
  int iden = omp_get_thread_num();
  totalCallsSpecredmin[iden]++;
#endif

  // searchThread should initially point to the current data matrix
  searchThread = wheel_data[tidaux].pointer;

  // check if the direction where I want to save data, has been used by current thread
  while (searchThread->matriz[hashAddress][pos].origPointer != NULL) {
#ifdef MEASURE_COUNTERS
    totalAccessesSpecredmin[iden]++;
#endif
    if (searchThread->matriz[hashAddress][pos].origPointer != address) {
      pos++;
    } else {
      usedAddressFlag = true;
#ifdef MEASURE_COUNTERS
      // se va a sumar un elemento no utilizado
      totalAccessesSpecredmin[iden]--;
#endif
      break;
      // pos variable will contain the position of the pointer in its matrix
    }
  }
#ifdef MEASURE_COUNTERS
  totalAccessesSpecredmin[iden]++;
#endif
  // NOTE: pos variable will contain the position of the pointer in its matrix

  if (usedAddressFlag) {
    //  We assume here that the state is REDMIN, thus the value,
#ifdef SPECREDMIN_DEBUG
    if (searchThread->matriz[hashAddress][pos].state != REDMIN) {
      printf("ERROR in specredmin: action not implemented.\n");
      exit(-1);
    }
#endif

    //  IV and AV have been updated previously.

#ifdef SPECREDMIN_DEBUG
    //  speccode: debug
    printf("specredmin: tid = %d, current = %d Simply perform MIN operation\n", tid, current);
    fflush(stdout);
    //  speccode: end debug
#endif

    //  Thus, we perform MIN operation

    // First of all, load the value
    // Checking for data size
#ifdef COMPLEX_STRUCTURES
    offset = searchThread->matriz[hashAddress][pos].copyOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][pos].speculativeCopy);
#endif
    switch (byteSize) {
      case 1:
        if (*(char *) addressMostRecently > *(char *) value)
          *(char *) addressMostRecently = *(char *) value;
        break;
      case 2:
        if (*(short *) addressMostRecently > *(short *) value)
          *(short *) addressMostRecently = *(short *) value;
        break;
      case 4:
        if (*(int *) addressMostRecently > *(int *) value)
          *(int *) addressMostRecently = *(int *) value;
        break;
      case 8:
        if (*(double *) addressMostRecently > *(double *) value)
          *(double *) addressMostRecently = *(double *) value;
        break;
      default:
        fprintf(stderr,"Error in specredmin: Unknown size = %d\n", byteSize);
        exit(-1);
    }
  } else {

    // current thread have not used the pointer
    lastData = pos;

#ifdef MAXDATA_DEBUG
    // Check if thread matrix is full (need to add data)
#pragma omp flush(wheel_data)
    if (lastData >= wheel_data[tidaux].maxData) {
      //    reservarMasEspacio();
      printf("Error in specredadd: lastData=%d >= wheel_data[modvector[current]].maxData=%d, current = %d,\n", lastData, wheel_data[modvector[current]].maxData, modvector[current]);
      fflush(stdout);
      exit(-1);
    }
#endif

#ifdef COMPLEX_STRUCTURES
    // First of all, store the value
    // addressMostRecently variables are needed because it is need to save values byte to byte
#pragma omp flush(wheel_data)
    offset = wheel_data[tidaux].actualOffset;
    addressMostRecently = (void *) &(wheel_data[tidaux].temporalData[offset]);
    // Checking for data size
    memcpy(addressMostRecently, value, byteSize);
#else
    addressMostRecently = (void *) &(searchThread->matriz[hashAddress][lastData].speculativeCopy);
    switch (byteSize) {
      case 1:
        *(char *) addressMostRecently = *(char *) value;
        break;
      case 2:
        *(short *) addressMostRecently = *(short *) value;
        break;
      case 4:
        *(int *) addressMostRecently = *(int *) value;
        break;
      case 8:
        *(double *) addressMostRecently = *(double *) value;
        break;
      default:
        printf("Error in specredadd: while copying temporal data, unknown size = %d\n, Exit...\n", byteSize);
        exit(-1);
    }
#endif

    searchThread->matriz[hashAddress][lastData + 1].origPointer = NULL;
#ifdef COMPLEX_STRUCTURES
    searchThread->matriz[hashAddress][lastData].copyOffset = wheel_data[tidaux].actualOffset;
#endif
    searchThread->matriz[hashAddress][lastData].state = REDMIN;
    searchThread->matriz[hashAddress][lastData].size = byteSize;
    searchThread->matriz[hashAddress][lastData].origPointer = address;

#ifdef IV_DEBUG
    printf("SPST: Aumentando valores IV, th no used pointer: IV[%d][%d] = %d,\n", modvector[thread], IV_tail[modvector[thread]], lastData);
    fflush(stdout);
#endif
    // Update IV
    IV[hashAddress][modvector[thread]][IV_tail[hashAddress][modvector[thread]]] = lastData;
    IV_tail[hashAddress][modvector[thread]] = IV_tail[hashAddress][modvector[thread]] + 1;

#ifdef COMPLEX_STRUCTURES
    wheel_data[tidaux].actualOffset += byteSize;
#pragma omp flush(wheel_data, IV, IV_tail)
#else
#pragma omp flush(IV, IV_tail)
#endif
#ifdef COMPLEX_STRUCTURES
#ifdef OFFSET_DEBUG
    if (wheel_data[modvector[current]].actualOffset >= MAX_OFFSET) {
      printf("ERROR in specredmin: Offset = %d: higher than maxoffset=%d,\n", wheel_data[modvector[current]].actualOffset, MAX_OFFSET);
      fflush(stdout);
      exit(-1);
    }
#endif
#endif

  }

#ifdef SPECREDMIN_DEBUG
  //  speccode: debug
  printf("specredmin: tid = %d, current = %d Ending MIN reduction\n", tid, current);
  fflush(stdout);
  //  speccode: end debug
#endif

} // END specredmin




//
// DYNAMIC SCHEDULING:
//


//***********************************************************************
// slopeEconomia(): Function to recover the slope of the regression line 
//***********************************************************************

double slopeEconomia(int min, int max) {

  double a10=0, a01=0, a11=0, a20=0;
  double m11, m20, b;
  int i;
  double n;

  if (min == max)
    return 0.0;

  n = (double)(max - min + 1);

  // printf("   slope(): Valores recibidos: %d y %d\n",min,max);
  // printf("   slope(): n = %f\n",n);

  // Se calcula a01
  for (i=min; i<=max; i++) {
    a01 += (double) varblock[2][i];
    a10 += (double) i;
    a11 += (double) i * (double) varblock[2][i];
    a20 += (double) i * (double) i;
  }
  a01 /= n;
  a10 /= n;
  a11 /= n;
  a20 /= n;

  // printf("   slope(): a01 = %lf\n",a01);
  // printf("   slope(): a10 = %lf\n",a10);
  // printf("   slope(): a11 = %lf\n",a11);
  // printf("   slope(): a20 = %lf\n",a20);

  // Se calcula m11
  m11 = a11 - a10 * a01;

  // printf("   slope(): m11 = %lf\n",m11);

  // Se calcula m20
  m20 = a20 - a10 * a10;

  // printf("   slope(): m20 = %lf\n",m20);

  // Se calcula b
  b= m11/m20;

  // printf("   slope(): b = %lf\n",b);

  return ((double) b);
}


/***********************************************************************
 **  specRightTurn(): Returns a negative number if the turn is clockwise.
 **  @author Diego R. Llanos, David Orden
 **  @date 25th November 2012
 ************************************************************************/

float specRightTurn(float dx, float dy, float px, float py, float qx, float qy)
{
  return (dx*py + dy*qx + px*qy -qx*py -dx*qy-px*dy);
}


/***********************************************************************
 **  findZone(): Find the zone where the point is, according to Scheduling 2012
 **  @author Diego R. Llanos, David Orden
 **  @date 25th November 2012
 **  @updated 30th May 2014 by Alvaro Estebanez
 ************************************************************************/

//int findZone(float d, float meanH, float maxMeanH) 
int findZone(float x, float y, float maxMeanH) {   

  //float p0x = 0;
  //float p0y = goodMeanH;
  //float p1x = -1;
  //float p1y = 1;
  float p2x = -1;
  float p2y = goodMeanH;
  float p3x = -1;
  float p3y = maxMeanH;
  float p4x = 0;
  float p4y = maxMeanH;
  //float p5x = 1;
  //float p5y = maxMeanH;
  float p6x = 1;
  float p6y = goodMeanH;
  float p7x = 1;
  float p7y = 1;
  float p8x = 0;
  float p8y = 1;

  // Se mira a qué lado de la recta A estamos.
  if ( specRightTurn(x, y, p7x, p7y, p3x, p3y) > 0) {
    // Estamos en la zona definida por P7, P3 y P1.
    //  **********
    //  ********__
    //  ******____
    //  ***_______
    //  *_________
    if ( specRightTurn(x, y, p6x, p6y, p2x, p2y) > 0) {
      // Estamos en la zona definida por P7, P0, P2 y P1.
      //  *****_____
      //  *****_____
      //  *****_____
      //  ***_______
      //  *_________
      if ( specRightTurn(x, y, p8x, p8y, p4x, p4y) > 0) {
        // Estamos en la zona definida por P8, P0, P2 y P1.
        //  *****_____
        //  *****_____
        //  *****_____
        //  __________
        //  __________
        //  __________
        if ( specRightTurn(x, y, p8x, p8y, p2x, p2y) > 0) {
          //  *****_____
          //  ***_______
          //  *_________
          //  __________
          //  __________
          //  __________
          return (T128);
        } else {
          //  ____*_____
          //  __***_____
          //  *****_____
          //  __________
          //  __________
          //  __________
          return (T028);
        }
      }
      else {
        /*    if (y-1 <(goodMeanH-1)/2) {
        //  ____*_____
        //  __***_____
        //  *****_____
        //  __________
        //  __________
        //  __________
        return (T128);
        } else {
        //  __________
        //  __________
        //  __________
        //  *****_____
        //  ***_______
        //  *_________
        return (T078);
        }*/
        return (T078);
      }
    }
    else {
      //  _____*****
      //  _____***__
      //  _____*____
      //  __________
      //  __________
      //  __________
      return (T023);
    }
  }
  else {
    // Estamos en la zona definida por P7, P3 y P5.
    //  _________*
    //  _______***
    //  ____******
    //  __********
    //  **********
    if ( specRightTurn(x, y, p6x, p6y, p2x, p2y) > 0) {
      //  __________
      //  __________
      //  __________
      //  ____*_____
      //  __***_____
      //  *****_____
      return (T067);
    } else {
      // Estamos en la zona definida por P6, P0, P3 y P5.
      //  _________*
      //  _______***
      //  _____*****
      //  _____*****
      //  _____*****
      if ( specRightTurn(x, y, p8x, p8y, p4x, p4y) > 0) {
        //  _________*
        //  _______***
        //  _____*****
        //  __________
        //  __________
        //  __________
        return (T034);
      } else {
        // Estamos en la zona definida por P0, P4, P5 y P6.
        if ( specRightTurn(x, y, p6x, p6y, p4x, p4y) > 0) {
          //  __________
          //  __________
          //  __________
          //  _____*****
          //  _____***__
          //  _____*____
          return (T046);
        }
        else {
          //  __________
          //  __________
          //  __________
          //  _________*
          //  _______***
          //  _____*****
          return (T456);
        }
      }
    }
  }

}


//***********************************************************************
//  dynScheduler(): Function that calculates next chunk size
//***********************************************************************

void dynScheduler(int chunkIndex) {

#ifdef TIME_DYNSCHEDULER
  double time_ini, time_end;
  int id = omp_get_thread_num();
  time_ini = get_time();
#endif

  int i;
  float logChunkIndex, logTotal, meanH=0;
  int firstWindowChunk;
  int windowSize; 
  float mySlope;
  float d;
  float lastChunk;
  float maxMeanH;
  int zone;
  float nextChunk;
  float chunkMax;
#ifdef FSC_DEBUG
  fprintf(stderr,"blk:%d",blk);
#endif

#pragma omp flush(varblock)
  switch (blk) {

//    case DYN_TEST: // Dynamic scheduling, test case with 1000 iterations
//      // First element: iteration to start
//      varblock[0][chunkIndex] = 1000 * chunkIndex;
//      // Second element: iteration to finish
//      varblock[1][chunkIndex] = 1000*(chunkIndex+1)-1;
//      // The third element (number of attempts for this block) was initialized
//      // in specinit() and it is incremented in a macro in fixedStructures.h.
//      break;

    case MESETA:
      break;

      // ==================================================================================================
      // ==================================================================================================
    case DYN_PDP1: 

      // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function PDP(1): Setting characteristics of chunk %d.\n", chunkIndex);
#endif
      // ----------------------------------------------------------------------------------------------

      if (varblock[2][chunkIndex]!=0) {

        // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
        printf ("   dynScheduler(), function DYN_PDP(1): No need of recalculating. chunk %d = %d.\n"
            , chunkIndex, (varblock[1][chunkIndex]-varblock[0][chunkIndex]+1));
#endif
        // ----------------------------------------------------------------------------------------------
#ifdef TIME_DYNSCHEDULER
  time_end = get_time();
  time_dynscheduler[id] += (time_end - time_ini);
#endif
        return;
      }

      // Caution! If varblock[2][chunkIndex]==0, it proceeds as in the adaptive case.
      // This is why we do not put a "break" statement here.


      // ==================================================================================================
      // ==================================================================================================
    case ADA_PDP1: // Dynamic scheduling with function (1) of PDP 2008.

      // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function PDP(1), dyn/ada flavors: Setting characteristics of chunk %d.\n", chunkIndex);
#endif
      // ----------------------------------------------------------------------------------------------

      //Alvaro: Ahora se hace en specbegin.
      //if (chunkIndex == 0) {
      //  // First iteration of the block: 0.
      //  varblock[0][chunkIndex] = 0;
      //  // Last iteration: Lets say 1 (no way to calculate logChunkSize)
      //  varblock[1][chunkIndex] = 1;
      //  return;
      //}

      // First iteration of the block: the last iteration of predeccessor plus one.
      varblock[0][chunkIndex] = varblock[1][chunkIndex-1] + 1;

      // Last iteration of the block: the first one plus the chunk function, calculated with
      // function (1) of PDP 2008 article.
      logChunkIndex = (float) log( (float)chunkIndex );
      logTotal = (float) log( (float)max_upper_limit );

      // Calculating meanH. Note that we may not have H predeccessors.
      firstWindowChunk = chunkIndex - H;
      if (firstWindowChunk < 0)
        firstWindowChunk = 0;
      windowSize = 0; 
      meanH = 0;
      for (i=firstWindowChunk; i<chunkIndex; i++) {
        meanH = meanH + (float) varblock[2][i];
        windowSize++;
      }
      meanH = meanH / windowSize;

      varblock[1][chunkIndex] = varblock[0][chunkIndex] + (int) ( (logChunkIndex*logTotal)/meanH );

      // ----------------------------------------------------------------------------------------------
      // debug
#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function PDP1: Setting characteristics of chunk %d.\n", chunkIndex);
      printf ("   First iteration: %d\n",varblock[0][chunkIndex]);
      printf ("   Last iteration details:\n");
      printf ("      First chunk of the scheduling window: %d\n",firstWindowChunk);
      printf ("      Last chunk of the scheduling window: %d\n",chunkIndex-1);
      printf ("      Size of the scheduling window: %d\n",windowSize);
      printf ("      Re-executions so far inside of the scheduling window:\n");
      printf ("        ");
      for (i=firstWindowChunk; i<chunkIndex; i++) 
        printf("%d\t",varblock[2][i]);
      printf ("\n");
      printf ("      meanH: %f\n",meanH); 
      printf ("      logChunkIndex = log(%d) = %f\n",chunkIndex, logChunkIndex); 
      printf ("      logTotal = log(%d) = %f\n",max_upper_limit, logTotal); 
      printf ("   Last iteration: %d\n",varblock[1][chunkIndex]);
      printf ("   Iterations of chunk %d = %d\n",chunkIndex,(int) ( (logChunkIndex*logTotal)/meanH )+1);
      fflush(stdout);
#endif
      // end debug
      // ----------------------------------------------------------------------------------------------

      break;

      // ==================================================================================================
      // ==================================================================================================
    case DYN_PDP2: 

      // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function PDP2: Setting characteristics of chunk %d.\n", chunkIndex);
#endif
      // ----------------------------------------------------------------------------------------------

      if (varblock[2][chunkIndex]!=0) {

        // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
        printf ("   dynScheduler(), function DYN_PDP2: No need of recalculating. chunk %d = %d.\n"
            , chunkIndex, (varblock[1][chunkIndex]-varblock[0][chunkIndex]+1));
#endif
        // ----------------------------------------------------------------------------------------------
#ifdef TIME_DYNSCHEDULER
  time_end = get_time();
  time_dynscheduler[id] += (time_end - time_ini);
#endif
        return;
      }

      // Caution! If varblock[2][chunkIndex]==0, it proceeds as in the adaptive case.
      // This is why we do not put a "break" statement here.


      // ==================================================================================================
      // ==================================================================================================
    case ADA_PDP2: // Adaptive scheduling with function (2) of PDP 2008.

      // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function PDP(2), dyn/ada flavors: Setting characteristics of chunk %d.\n", chunkIndex);
#endif
      // ----------------------------------------------------------------------------------------------

      //Alvaro: Ahora se hace en specbegin.
      //if (chunkIndex == 0) {
      //  // First iteration of the block: 0.
      //  varblock[0][chunkIndex] = 0;
      //  // Last iteration: Lets say 1 (no way to calculate logChunkSize)
      //  varblock[1][chunkIndex] = 1;
      //  return;
      //}

      // First iteration of the block: the last iteration of predeccessor plus one.
      varblock[0][chunkIndex] = varblock[1][chunkIndex-1] + 1;

      // Last iteration of the block: the first one plus the chunk function, calculated with
      // function (1) of PDP 2008 article.
      logChunkIndex = (float) log( (float)chunkIndex );
      logTotal = (float) log( (float)max_upper_limit );

      // Calculating meanH. Note that we may not have H predeccessors.
      firstWindowChunk = chunkIndex - H;
      if (firstWindowChunk < 0)
        firstWindowChunk = 0;
      windowSize = 0; 
      meanH = 0; 
      for (i=firstWindowChunk; i<chunkIndex; i++) {
        meanH = meanH + (float) varblock[2][i];
        windowSize++;
      }
      meanH = meanH / windowSize;

      varblock[1][chunkIndex] = varblock[0][chunkIndex] + (int) ( (logChunkIndex*logChunkIndex*logTotal)/meanH );

      // ----------------------------------------------------------------------------------------------
      // debug
#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function PDP2: Setting characteristics of chunk %d.\n", chunkIndex);
      printf ("   First iteration: %d\n",varblock[0][chunkIndex]);
      printf ("   Last iteration details:\n");
      printf ("      First chunk of the scheduling window: %d\n",firstWindowChunk);
      printf ("      Last chunk of the scheduling window: %d\n",chunkIndex-1);
      printf ("      Size of the scheduling window: %d\n",windowSize);
      printf ("      Re-executions so far inside of the scheduling window:\n");
      printf ("        ");
      for (i=firstWindowChunk; i<chunkIndex; i++) 
        printf("%d\t",varblock[2][i]);
      printf ("\n");
      printf ("      meanH: %f\n",meanH); 
      printf ("      logChunkIndex = log(%d) = %f\n",chunkIndex, logChunkIndex); 
      printf ("      logTotal = log(%d) = %f\n",max_upper_limit, logTotal); 
      printf ("   Last iteration: %d\n",varblock[1][chunkIndex]);
      printf ("   Iterations of chunk %d = %d\n",chunkIndex,(int) ( (logChunkIndex*logTotal)/meanH )+1);
      fflush(stdout);
#endif 
      // ----------------------------------------------------------------------------------------------

      break;

      // ==================================================================================================
      // ==================================================================================================
    case DYN_TENT: 

      // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function TENT: Setting characteristics of chunk %d.\n", chunkIndex);
#endif
      // ----------------------------------------------------------------------------------------------

      if (varblock[2][chunkIndex]!=0) {

        // ----------------------------------------------------------------------------------------------
#ifdef debugScheduling
        printf ("   dynScheduler(), function DYN_TENT: No need of recalculating. chunk %d = %d.\n"
            , chunkIndex, (varblock[1][chunkIndex]-varblock[0][chunkIndex]+1));
#endif
        // ----------------------------------------------------------------------------------------------
        return;
      }

      // Caution! If varblock[2][chunkIndex]==0, it proceeds as in the adaptive case.
      // This is why we do not put a "break" statement here.

      // ==================================================================================================
      // ==================================================================================================
    case ADA_TENT: 

#ifdef debugScheduling
      printf ("-------------------------------------------------------\n");
      printf ("dynScheduler(), function TENT, dyn/ada flavors: Setting characteristics of chunk %d.\n", chunkIndex);
#endif

      //Alvaro: Ahora se hace en specbegin.
      //if (chunkIndex == 0) {
      // First iteration of the block: 0.
      //  varblock[0][chunkIndex] = 0;
      // Last iteration: Same as the FSC best chunk size
      //  varblock[1][chunkIndex] = FIRSTCHUNK;
      //#ifdef debugScheduling
      //        printf ("   ->First iteration!\n");
      //        printf ("   First iteration: %d\n",varblock[0][chunkIndex]);
      //       printf ("   Last iteration: %d\n",varblock[1][chunkIndex]);
      //        fflush(stdout);
      //#endif 

      //      return;
      //    }

      // First iteration of the block: the last iteration of predecessor plus one.
      varblock[0][chunkIndex] = varblock[1][chunkIndex-1] + 1;

      // Last iteration of the block: to be calculated with the TENT function (Scheduling 2012)
      // As the previous history, we will use the H predecessors of the non-spec thread.
      firstWindowChunk = wheel_ns - H;

      // Calculating meanH. Note that we may not have H predecessors.
      if (firstWindowChunk < 0) 
        firstWindowChunk = 0;
      windowSize = 0; 
      meanH = 0;

      // fprintf(stderr,"dynScheduler: En TENT, antes del for, wheel_ns=%i, meanH=%f\n",wheel_ns, meanH);

      if (firstWindowChunk >= wheel_ns) {
        // No committed threads inside the history window: we use FIRSTCHUNK iterations.
        varblock[1][chunkIndex] = varblock[0][chunkIndex]+FIRSTCHUNK;
#ifdef debugScheduling
        printf ("   ->No committed threads inside the history window!\n");
        printf ("   First iteration: %d\n",varblock[0][chunkIndex]);
        printf ("   Last iteration details:\n");
        printf ("      First chunk of the scheduling window: %d\n",firstWindowChunk);
        printf ("      Last chunk of the scheduling window: %d\n",wheel_ns-1);
        printf ("   Last iteration: %d\n",varblock[1][chunkIndex]);
        fflush(stdout);
#endif 
#ifdef TIME_DYNSCHEDULER
  time_end = get_time();
  time_dynscheduler[id] += (time_end - time_ini);
#endif
        return;
      }

      for (i=firstWindowChunk; i< wheel_ns; i++) {
        meanH = meanH + (float) varblock[2][i];
        windowSize++;
      }
      meanH = meanH / windowSize;

      // fprintf(stderr,"dynScheduler: En TENT, despues del for, wheel_ns=%i, meanH=%f\n",wheel_ns, meanH);

      // Calculate slope
      mySlope = slopeEconomia(firstWindowChunk, wheel_ns-1);

      // Calculate d=arctan(slope) / (PI/2)
      d = (float) (atan((double)mySlope)*2)/3.14159265358979323846;

      //fprintf (stderr, "d = %f meanH = %f\n",d,meanH);
      //fflush(stderr);

      // Calculate maxMeanH
      lastChunk = (float) (varblock[1][chunkIndex-1] - varblock[0][chunkIndex-1]);
      maxMeanH = (float) goodMeanH + ( (lastChunk-1) / (float) tan((double)BETA));

      // Calculate cMax
      chunkMax = (float) lastChunk + (float) ( tan((double)ALPHA)* (double) (goodMeanH-1) );

      // Alvaro: Optimizacion
      /*if (d <= 0.1 && meanH <= 1.1) {
        varblock[1][chunkIndex] = varblock[0][chunkIndex] + (int) rintf(chunkMax);
#pragma omp flush(varblock)
        return;
      }*/

      // If we are outside the tent (meanH > maxMeanH), chunk size is 1.
      if (meanH > maxMeanH) {
        varblock[1][chunkIndex] = varblock[0][chunkIndex] + 1;
#pragma omp flush(varblock)
#ifdef debugScheduling
        printf ("   -> Outside the tent (meanH > maxMeanH)\n");
        printf ("   First iteration: %d\n",varblock[0][chunkIndex]);
        printf ("   Last iteration details:\n");
        printf ("      First chunk of the scheduling window: %d\n",firstWindowChunk);
        printf ("      Last chunk of the scheduling window: %d\n",wheel_ns-1);
        printf ("      Size of the scheduling window: %d\n",windowSize);
        printf ("      Re-executions so far inside of the scheduling window:\n");
        printf ("        ");
        for (i=firstWindowChunk; i<wheel_ns; i++) 
          printf("%d\t",varblock[2][i]);
        printf ("\n");
        printf ("      meanH: %f\n",meanH); 
        printf ("      mySlope: %f\n",(float) mySlope); 
        printf ("      d: %f\n",(float) d); 
        printf ("      zone: %d\n",zone); 
        printf ("      nextChunk: %f\n",nextChunk); 
        printf ("   Last iteration: %d\n",varblock[1][chunkIndex]);
        // printf ("   Calculated chunk size: %d\n",varblock[1][chunkIndex]-varblock[0][chunkIndex]+1);
        fflush(stdout);
#endif 
#ifdef MEASURE_COUNTERS
        tentZone[T35inf]++;
#endif

#ifdef TIME_DYNSCHEDULER
  time_end = get_time();
  time_dynscheduler[id] += (time_end - time_ini);
#endif
        return;
      }

      // Call findZone() to get the zone.
      zone = findZone(d, meanH, maxMeanH);

      // Depending on the zone, generate "output" (new chunk)
#ifdef debugScheduling
      printf("   Before entering zone switch:\n");
      printf("Zone %i : lastChunk = %f\n",zone,lastChunk);
      printf("Zone %i : meanH = %f\n",zone,meanH);
      printf("Zone %i : goodMeanH = %i\n",zone,goodMeanH);
      printf("Zone %i : maxMeanH = %f\n",zone,maxMeanH);
      fflush(stdout);
#endif 

      switch (zone) {
        case T128:
          nextChunk = chunkMax; 
#ifdef debugScheduling
          printf("T128, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T128]++;
#endif
          break;

        case T028:
          nextChunk = ((lastChunk - chunkMax)*meanH 
              + ((lastChunk - chunkMax)*goodMeanH - lastChunk + chunkMax)*d 
              + chunkMax*goodMeanH - lastChunk)/(goodMeanH - 1);
#ifdef debugScheduling
          printf("T028, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T028]++;
#endif
          break;

        case T023:
          // Ajuste: ver correo de David del 21/2/2013
          if (goodMeanH == maxMeanH) 
            nextChunk = 1;
          else 
            nextChunk = -((lastChunk - chunkMax)*meanH 
                - (2*lastChunk - chunkMax)*goodMeanH - ((lastChunk - chunkMax)*goodMeanH 
                  - (lastChunk - chunkMax)*maxMeanH)*d + lastChunk*maxMeanH)/(goodMeanH - maxMeanH);
#ifdef debugScheduling
          printf("T023, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T023]++;
#endif
          break;

        case T034:
          // Ajuste: ver correo de David del 21/2/2013
          if (goodMeanH == maxMeanH)
            nextChunk = 1;
          else 
            nextChunk = ((lastChunk - 1)*meanH - ((lastChunk - 1)*goodMeanH 
                  - (lastChunk - 1)*maxMeanH)*d - lastChunk*maxMeanH + goodMeanH)/(goodMeanH - maxMeanH);
#ifdef debugScheduling
          printf("T034, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T034]++;
#endif
          break;

        case T078:
          nextChunk = ((lastChunk - chunkMax)*meanH 
              + ((lastChunk - chunkMax)*goodMeanH - lastChunk + chunkMax)*d 
              + chunkMax*goodMeanH - lastChunk)/(goodMeanH - 1);
#ifdef debugScheduling
          printf("T078, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T078]++;
#endif
          break;

        case T067:
          nextChunk = -((lastChunk - 1)*meanH - (2*lastChunk - 1)*goodMeanH 
              + ((lastChunk - 1)*goodMeanH - lastChunk + 1)*d 
              + lastChunk)/(goodMeanH - 1);
#ifdef debugScheduling
          printf("T067, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T067]++;
#endif
          break;

        case T046:
          // Ajuste: ver correo de David del 21/2/2013
          if (goodMeanH == maxMeanH) 
            nextChunk = 1;
          else 
            nextChunk = ((lastChunk - 1)*meanH - ((lastChunk - 1)*goodMeanH 
                  - (lastChunk - 1)*maxMeanH)*d - lastChunk*maxMeanH
                + goodMeanH)/(goodMeanH - maxMeanH);
#ifdef debugScheduling
          printf("T046, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T046]++;
#endif
          break;

        case T456:
          nextChunk = 1;
#ifdef debugScheduling
          printf("T456, nextChunk=%f, lastChunk = %f, meanH = %f, maxMeanH = %f, d = %f, it %d por %d, mySlope = %f.\n",
              nextChunk,lastChunk,meanH,maxMeanH,d,varblock[0][chunkIndex],chunkIndex,mySlope);
          fflush(stdout);
#endif
#ifdef MEASURE_COUNTERS
          tentZone[T456]++;
#endif
          break;
      }

      // Use nextChunk to update varblock[1][chunkIndex]
      varblock[1][chunkIndex] = varblock[0][chunkIndex] + (int) rintf(nextChunk);
#pragma omp flush(varblock)

      // ----------------------------------------------------------------------------------------------
      // debug
#ifdef debugScheduling
      printf ("   First iteration: %d\n",varblock[0][chunkIndex]);
      printf ("   Last iteration details:\n");
      printf ("      First chunk of the scheduling window: %d\n",firstWindowChunk);
      printf ("      Last chunk of the scheduling window: %d\n",wheel_ns-1);
      printf ("      Size of the scheduling window: %d\n",windowSize);
      printf ("      Re-executions so far inside of the scheduling window:\n");
      printf ("        ");
      for (i=firstWindowChunk; i<wheel_ns; i++) 
        printf("%d\t",varblock[2][i]);
      printf ("\n");
      printf ("      meanH: %f\n",meanH); 
      printf ("      mySlope: %f\n",(float) mySlope); 
      printf ("      d: %f\n",(float) d); 
      printf ("      zone: %d\n",zone); 
      printf ("      nextChunk: %f\n",nextChunk); 
      printf ("      chunkMax calculations:\n"); 
      printf ("         tan((double)ALPHA: %f\n",tan((double)ALPHA)); 
      printf ("         goodMeanH-1: %d\n",goodMeanH-1); 
      printf ("         chunkMax: %f\n",chunkMax); 
      printf ("   Last iteration: %d\n",varblock[1][chunkIndex]);
      printf ("   Chunk size: %d\n",varblock[1][chunkIndex]-varblock[0][chunkIndex]);
      // printf ("   Calculated chunk size: %d\n",varblock[1][chunkIndex]-varblock[0][chunkIndex]+1);
      fflush(stdout);
#endif 
      // end debug
      // ----------------------------------------------------------------------------------------------

#ifdef TIME_DYNSCHEDULER
  time_end = get_time();
  time_dynscheduler[id] += (time_end - time_ini);
#endif
      return;

    default: // FSC
#ifdef TIME_DYNSCHEDULER
  time_end = get_time();
  time_dynscheduler[id] += (time_end - time_ini);
#endif
#ifdef FSC_DEBUG
  fprintf(stderr,"Value returned FSC:%d",blk);
#endif
      return;
  }

#ifdef TIME_DYNSCHEDULER
  time_end = get_time();
  time_dynscheduler[id] += (time_end - time_ini);
#endif

} // END dynScheduler()


void meseta(int totalIterations, int percentageStart, int percentageEnd) {

    int initialZoneIterations = (int) ((percentageStart*totalIterations)/100);
    int finalZoneIterations = (int) ((percentageEnd*totalIterations)/100);
    int j;
    int numIter = 0;
    int i=0;

    // The following variable will store either increments or decrements of the iterations assigned
    int variation;

    // The following variables will store the rough length of each zone
    float lengthInitialZone, lengthFinalZone;

    // Supposing that the total iterations of the zone are similar to the area of a triangle
    // Area of a triangle = (b*h)/2
    // area = initialZoneIterations;
    // b = lengthInitialZone;
    // h = MAX_ITER - 1;
    lengthInitialZone = (initialZoneIterations*2)/(MAX_ITER-1);
    // in order to avoid a length lower than 1 that may cause problems
    lengthInitialZone++;

    // start of the Initial zone
    while (numIter<initialZoneIterations) {
        // Straight line equation:
        // f(x) = ( ( (x - x1)* (y2 - y1) ) / (x2 - x1) ) + y1;
        // f(x) = variation
        // x1 = 0;
        // x2 = lengthInitialZone;
        // y1 = 0;
        // y2 = MAX_ITER - 1;
        // x = i;
        variation = (int) floor( ( i * (MAX_ITER-1) ) / lengthInitialZone);

        if (variation < 1)
            variation = 1;
        varblock[0][i] = numIter;
        varblock[1][i] = numIter+variation;
#ifdef MESETA_DEBUG
        //printf("i:%d\tvar0=%d,var1=%d, \t variation=%d, lengthInitialZone=%f\n",i,varblock[0][i],varblock[1][i],variation,lengthInitialZone);
        printf("%d %d\n",i,variation);
#endif

        numIter += variation+1;
        i++;
    }
    // end of the Initial zone

    // start of the middle zone
    variation=MAX_ITER-1;
    while (numIter < (totalIterations - finalZoneIterations) ) {

        varblock[0][i] = numIter;
        varblock[1][i] = numIter+variation;
#ifdef MESETA_DEBUG
        //printf("i:%d\tvar0=%d,var1=%d, \t variation=%d\n",i,varblock[0][i],varblock[1][i],variation);
        printf("%d %d\n",i,variation);
#endif

        numIter += variation+1;
        i++;
    }
    // end of the middle zone

    // Supposing that the total iterations of the zone are similar to the area of a triangle
    // Area of a triangle = (b*h)/2
    // area = finalZoneIterations;
    // b = lengthFinalZone;
    // h = MAX_ITER - 1;
    lengthFinalZone = (finalZoneIterations*2)/(MAX_ITER-1);
    // in order to avoid a lenth lower than 1 that may cause problems
    lengthFinalZone++;

    j=i;
    // start of the final zone
    while (numIter < totalIterations) {
        // Straight line equation:
        // f(x) = ( ( (x - x1)* (y2 - y1) ) / (x2 - x1) ) + y1;
        // f(x) = variation
        // x1 = j;
        // x2 = lengthFinalZone+j;
        // y1 = MAX_ITER - 1;
        // y2 = 0;
        // x = i+1;
        variation = (int) floor( ( (((i+1)-j)*(-1*(MAX_ITER-1))) / lengthFinalZone) + (MAX_ITER-1) );

        varblock[0][i] = numIter;
        varblock[1][i] = numIter+variation;
        numIter += variation+1;
#ifdef MESETA_DEBUG
        //printf("i:%d\tvar0=%d,var1=%d, \t variation=%d, lengthFinalZone=%f\n",i,varblock[0][i],varblock[1][i],variation,lengthFinalZone);
        printf("%d %d\n",i,variation);
#endif
        i++;
    }
    // end of the final zone

    // Assign some more iterations in order to avoid errors (they will never be executed)
    for (j=0;j<wsize;j++) {
        varblock[0][i] = numIter;
        varblock[1][i] = numIter;
        numIter += 1;
        i++;
    }

#ifdef MESETA_DEBUG
    int sum = 0;
    for (j=0;j<i;j++) {
        sum += varblock[1][j] - varblock[0][j] + 1;
    }

    printf("iterations of the Initial Zone=%d, sum of the iterations =%d, number of Iterations =%d ... i=%d \n",initialZoneIterations,sum,numIter,i);
#endif

}

