/*
 * incr-2d: C version of 2-dimensional, incremental convex hull
 * Alvaro Estebanez Lopez, Diego R. Llanos
 * University of Valladolid
 */

#include <stdio.h>
#include <stdlib.h>

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file 
#include "specEngine.h"
//***********************************************************************

// N total number of points  (min. 3)
// For 2D-Hull, Kuzmin, 10M = 10.000.000
#define N 10000000 		//10.000.000
#define NITER 9999997		//9.999.997

#define source 1
#define target 2
#define succ   3
#define pred   4
#define inch   5

#define x 1
#define y 2

// TOP: Number of points in the Convex Hull
// For 2D-Hull, Kuzmin, 10M = 200
#define TOP 200

int hull[TOP + 1][5 + 1];
double input[2 + 1][N + 1];

int i, j, k;
int p, psource, ptarget;
int auxinch;
int lastedge;
int e, eprev, enext, ea;
int first;
double px, py, qx, qy, dx, dy;
double determ;
int outside;

// Function
double rightturn(double dx, double dy, double px, double py, double qx, double qy);

int main() {
  FILE *inputfile;

  /************************************************************************/
  // speccode: Stop profiling
  // collector_pause();
  /************************************************************************/

  //  Opens the input file
  if ( (inputfile = fopen("/home/Points/kuzmin_010M_limpio.dat", "r")) )
  {
    for (i = 1; i < N + 1; i++) {

      fscanf(inputfile, "%lf %lf", &input[x][i], &input[y][i]);
      //printf("New point: %lf %lf\n",input[x][i],input[y][i]);
    }

    fclose(inputfile);
  } else {
    printf("Error reading from file \n ");
    exit(0);
  }

  // Calculate the first determinant
  dx = input[x][1];
  dy = input[y][1];
  px = input[x][2];
  py = input[y][2];
  qx = input[x][3];
  qy = input[y][3];

  determ = rightturn(dx, dy, px, py, qx, qy);

  if (determ == 0) {
    printf("incremental: The first three points are aligned.");
    printf("incremental: We aree too lazy to proceed. Stop.");
    exit(0);
  }

  // If determinant < 0, reorder points two and three
  if (determ < 0) {
    input[x][2] = qx;
    input[y][2] = qy;
    input[x][3] = px;
    input[y][3] = py;
  }

  // Links the first edge
  hull[1][source] = 1;
  hull[1][target] = 2;
  hull[1][pred] = 3;
  hull[1][succ] = 2;
  hull[1][inch] = 1;

  // Links the second edge
  hull[2][source] = 2;
  hull[2][target] = 3;
  hull[2][pred] = 1;
  hull[2][succ] = 3;
  hull[2][inch] = 1;

  // Links the third edge
  hull[3][source] = 3;
  hull[3][target] = 1;
  hull[3][pred] = 2;
  hull[3][succ] = 1;
  hull[3][inch] = 1;

  hull[TOP][1] = 1;
  hull[TOP][5] = 3;
  // Put -1 in all remaining elements of hull, just in case.
  for (i = 4; i < TOP; i++) {
    for (j = 1; j < 6; j++) {
      //-1 can prompt a SEGV signal
      //hull[i][j] = -1;
      hull[i][j] = 0;
    }
  }

  // Main loop (iterating over remaining points)

  //***********************************************************************
  // speccode: OMP threading directive
  omp_set_num_threads(threads);
  // speccode: initialize speculation structrures
  specbegin(N + 1 - 4);
  //***********************************************************************

  /************************************************************************/
  // speccode: Start profiling
  double spec_start, spec_stop;
  double spec_time = 0;
  spec_start = get_time();
  /************************************************************************/

  //***********************************************************************
  // speccode: Start profiling
  // collector_resume();
  //***********************************************************************
  //***********************************************************************
  // speccode: Speculative loop
  int iteration, ini;
  int value;

#pragma omp parallel default(none) \
  private(outside,p,px,py,qx,qy,dx,dy,determ,i,k,lastedge, \
      ini,iteration,psource,ptarget,auxinch,e,eprev,enext,ea,value,  \
      current, tid, retflag)  \
  shared(input,hull, wheel_ns, wheel_ms,    \
      wheel, upper_limit, varblock)                            
  {
#pragma omp for \
    schedule(static)
    //for ( i = 4 ; i < N+1 ; i++)
    for (tid = 0; tid <= threads - 1; tid++) {

      ini = 4;
      current = tid;
      iteration = varblock[0][current] + ini;

      if (iteration > upper_limit - 1)
        goto labelSquash_1;

      varblock[2][current] = varblock[2][current] + 1;
      //printf("\t\tvarblock[0][current]:%d,current:%d,iteration:%d\n",varblock[0][current],current,iteration);
labelStartIteration_1: //printf("ENDLOOP,iteration:%d,current:%d\n",iteration,current);

      i = iteration;

      // Debug
      // printf("incr-2d: starting iteration %i, attempt %i\n",i, varblock[2][current]);
      // End Debug

      outside = 0;
      for (e = 1; e < 4; e++) {
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      psource = hull[e][source];
        if (specload_pointer((void *) &(hull[e][source]), sizeof (hull[e][source]), current, (void *) &value) == -1)
          earlySquash(1);
        psource = value;
        //***********************************************************************
        dx = input[x][psource];
        dy = input[y][psource];


        //***********************************************************************
        // speccode: speculative load. Original line:
        //      ptarget = hull[e][target];
        if (specload_pointer((void *) &(hull[e][target]), sizeof (hull[e][target]), current, (void *) &value) == -1)
          earlySquash(1);
        ptarget = value;
        //***********************************************************************
        px = input[x][ptarget];
        py = input[y][ptarget];

        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx, dy, px, py, qx, qy);
        if (determ < 0) {
          // The point is outside
          outside = 1;
          break;
        }
      }

e500:
      //***********************************************************************
      // speccode: speculative load. Original line:
      //      auxinch = hull[e][inch];
      if (specload_pointer((void *) &(hull[e][inch]), sizeof (hull[e][inch]), current, (void *) &value) == -1)
        earlySquash(1);
      auxinch = value;
      //***********************************************************************
      if ((auxinch == 0) && (outside)) {
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      eprev = hull[e][pred];
        if (specload_pointer((void *) &(hull[e][pred]), sizeof (hull[e][pred]), current, (void *) &value) == -1)
          earlySquash(1);
        eprev = value;
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      psource = hull[eprev][source];
        if (specload_pointer((void *) &(hull[eprev][source]), sizeof (hull[eprev][source]), current, (void *) &value) == -1)
          earlySquash(1);
        psource = value;
        //***********************************************************************

        dx = input[x][psource];
        dy = input[y][psource];

        //***********************************************************************
        // speccode: speculative load. Original line:
        //      ptarget = hull[eprev][target];
        if (specload_pointer((void *) &(hull[eprev][target]), sizeof (hull[eprev][target]), current, (void *) &value) == -1)
          earlySquash(1);
        ptarget = value;
        //***********************************************************************
        px = input[x][ptarget];
        py = input[y][ptarget];

        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx, dy, px, py, qx, qy);

        if (determ < 0)
          e = eprev;
        else {
          //***********************************************************************
          // speccode: speculative load. Original line:
          //        enext = hull[e][succ];
          if (specload_pointer((void *) &(hull[e][succ]), sizeof (hull[e][succ]), current, (void *) &value) == -1)
            earlySquash(1);
          enext = value;
          //***********************************************************************
          //***********************************************************************
          // speccode: speculative load. Original line:
          //        psource = hull[enext][source];
          if (specload_pointer((void *) &(hull[enext][source]), sizeof (hull[enext][source]), current, (void *) &value) == -1)
            earlySquash(1);
          psource = value;
          //***********************************************************************

          dx = input[x][psource];
          dy = input[y][psource];

          //***********************************************************************
          // speccode: speculative load. Original line:
          //        ptarget = hull[enext][target];
          if (specload_pointer((void *) &(hull[enext][target]), sizeof (hull[enext][target]), current, (void *) &value) == -1)
            earlySquash(1);
          ptarget = value;
          //***********************************************************************
          px = input[x][ptarget];
          py = input[y][ptarget];

          qx = input[x][i];
          qy = input[y][i];

          determ = rightturn(dx, dy, px, py, qx, qy);

          if (determ >= 0)
            outside = 0;
          else
            e = enext;
        }
        goto e500;
      }

      if (outside) {
        //Compute upper tangent
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      enext = hull[e][succ];
        if (specload_pointer((void *) &(hull[e][succ]), sizeof (hull[e][succ]), current, (void *) &value) == -1)
          earlySquash(1);
        enext = value;
        //***********************************************************************

e600:
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      psource = hull[enext][source];
        if (specload_pointer((void *) &(hull[enext][source]), sizeof (hull[enext][source]), current, (void *) &value) == -1)
          earlySquash(1);
        psource = value;
        //***********************************************************************
        dx = input[x][psource];
        dy = input[y][psource];

        //***********************************************************************
        // speccode: speculative load. Original line:
        //      ptarget = hull[enext][target];
        if (specload_pointer((void *) &(hull[enext][target]), sizeof (hull[enext][target]), current, (void *) &value) == -1)
          earlySquash(1);
        ptarget = value;
        //***********************************************************************
        px = input[x][ptarget];
        py = input[y][ptarget];

        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx, dy, px, py, qx, qy);

        //***********************************************************************
        // speccode: speculative load. Original line:
        //      lastedge = hull[TOP][5];
        if (specload_pointer((void *) &(hull[TOP][5]), sizeof (hull[TOP][5]), current, (void *) &value) == -1)
          earlySquash(1);
        lastedge = value;
        //***********************************************************************

        if (determ <= 0) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          //        hull[enext][inch] = 0;
          value = 0;
          specstore_pointer((void *) &(hull[enext][inch]), sizeof (hull[enext][inch]), current, (void *) &value);
          //***********************************************************************
          //***********************************************************************
          // speccode: speculative store. Original line:
          //        hull[enext][pred] = lastedge+1;
          value = lastedge + 1;
          specstore_pointer((void *) &(hull[enext][pred]), sizeof (hull[enext][pred]), current, (void *) &value);
          //***********************************************************************
          //***********************************************************************
          // speccode: speculative load. Original line:
          //        ea = hull[enext][succ];
          if (specload_pointer((void *) &(hull[enext][succ]), sizeof (hull[enext][succ]), current, (void *) &value) == -1)
            earlySquash(1);
          ea = value;
          //***********************************************************************
          //***********************************************************************
          // speccode: speculative store. Original line:
          //        hull[enext][succ] = lastedge+2;
          value = lastedge + 2;
          specstore_pointer((void *) &(hull[enext][succ]), sizeof (hull[enext][succ]), current, (void *) &value);
          //***********************************************************************

          enext = ea;
          goto e600;
        }
        // Compute lower tangent

        //***********************************************************************
        // speccode: speculative load. Original line:
        //      eprev = hull[e][pred];
        if (specload_pointer((void *) &(hull[e][pred]), sizeof (hull[e][pred]), current, (void *) &value) == -1)
          earlySquash(1);
        eprev = value;
        //***********************************************************************
e700:
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      psource = hull[eprev][source];
        if (specload_pointer((void *) &(hull[eprev][source]), sizeof (hull[eprev][source]), current, (void *) &value) == -1)
          earlySquash(1);
        psource = value;
        //***********************************************************************
        dx = input[x][psource];
        dy = input[y][psource];

        //***********************************************************************
        // speccode: speculative load. Original line:
        //      ptarget = hull[eprev][target];
        if (specload_pointer((void *) &(hull[eprev][target]), sizeof (hull[eprev][target]), current, (void *) &value) == -1)
          earlySquash(1);
        ptarget = value;
        //***********************************************************************
        px = input[x][ptarget];
        py = input[y][ptarget];

        qx = input[x][i];
        qy = input[y][i];

        determ = rightturn(dx, dy, px, py, qx, qy);

        if (determ <= 0) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          //        hull[eprev][inch] = 0;
          value = 0;
          specstore_pointer((void *) &(hull[eprev][inch]), sizeof (hull[eprev][inch]), current, (void *) &value);
          //***********************************************************************
          //***********************************************************************
          // speccode: speculative load. Original line:
          //        ea = hull[eprev][pred];
          if (specload_pointer((void *) &(hull[eprev][pred]), sizeof (hull[eprev][pred]), current, (void *) &value) == -1)
            earlySquash(1);
          ea = value;
          //***********************************************************************
          //***********************************************************************
          // speccode: speculative store. Original line:
          //        hull[eprev][pred] = lastedge+1;
          value = lastedge + 1;
          specstore_pointer((void *) &(hull[eprev][pred]), sizeof (hull[eprev][pred]), current, (void *) &value);
          //***********************************************************************
          //***********************************************************************
          // speccode: speculative store. Original line:
          //        hull[eprev][succ] = lastedge+2;
          value = lastedge + 2;
          specstore_pointer((void *) &(hull[eprev][succ]), sizeof (hull[eprev][succ]), current, (void *) &value);
          //***********************************************************************
          eprev = ea;
          goto e700;
        }
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[e][inch] = 0;
        value = 0;
        specstore_pointer((void *) &(hull[e][inch]), sizeof (hull[e][inch]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[e][pred] = lastedge+1;
        value = lastedge + 1;
        specstore_pointer((void *) &(hull[e][pred]), sizeof (hull[e][pred]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[e][succ] = lastedge+2;
        value = lastedge + 2;
        specstore_pointer((void *) &(hull[e][succ]), sizeof (hull[e][succ]), current, (void *) &value);
        //***********************************************************************

        // Add new edges to convex hull

        lastedge = lastedge + 1;
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[TOP][5] = lastedge;
        value = lastedge;
        specstore_pointer((void *) &(hull[TOP][5]), sizeof (hull[TOP][5]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      hull[lastedge][source] = hull[eprev][target];
        if (specload_pointer((void *) &(hull[eprev][target]), sizeof (hull[eprev][target]), current, (void *) &value) == -1)
          earlySquash(1);
        // speccode: speculative store. Original line:
        specstore_pointer((void *) &(hull[lastedge][source]), sizeof (hull[lastedge][source]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][target] = i;
        value = i;
        specstore_pointer((void *) &(hull[lastedge][target]), sizeof (hull[lastedge][target]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][succ] = lastedge + 1;
        value = lastedge + 1;
        specstore_pointer((void *) &(hull[lastedge][succ]), sizeof (hull[lastedge][succ]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][pred] = eprev;
        value = eprev;
        specstore_pointer((void *) &(hull[lastedge][pred]), sizeof (hull[lastedge][pred]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][inch] = 1;
        value = 1;
        specstore_pointer((void *) &(hull[lastedge][inch]), sizeof (hull[lastedge][inch]), current, (void *) &value);
        //***********************************************************************

        lastedge = lastedge + 1;
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[TOP][5] = lastedge;
        value = lastedge;
        specstore_pointer((void *) &(hull[TOP][5]), sizeof (hull[TOP][5]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][source] = i;
        value = i;
        specstore_pointer((void *) &(hull[lastedge][source]), sizeof (hull[lastedge][source]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative load. Original line:
        //      hull[lastedge][target] = hull[enext][source];
        if (specload_pointer((void *) &(hull[enext][source]), sizeof (hull[enext][source]), current, (void *) &value) == -1)
          earlySquash(1);
        // speccode: speculative store. Original line:
        specstore_pointer((void *) &(hull[lastedge][target]), sizeof (hull[lastedge][target]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][succ] = enext;
        value = enext;
        specstore_pointer((void *) &(hull[lastedge][succ]), sizeof (hull[lastedge][succ]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][pred] = lastedge - 1;
        value = lastedge - 1;
        specstore_pointer((void *) &(hull[lastedge][pred]), sizeof (hull[lastedge][pred]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[lastedge][inch] = 1;
        value = 1;
        specstore_pointer((void *) &(hull[lastedge][inch]), sizeof (hull[lastedge][inch]), current, (void *) &value);
        //***********************************************************************

        // Connect to old convex hull
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[enext][pred] = lastedge;
        value = lastedge;
        specstore_pointer((void *) &(hull[enext][pred]), sizeof (hull[enext][pred]), current, (void *) &value);
        //***********************************************************************
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[eprev][succ] = lastedge - 1;
        value = lastedge - 1;
        specstore_pointer((void *) &(hull[eprev][succ]), sizeof (hull[eprev][succ]), current, (void *) &value);
        //***********************************************************************

        //Update first
        //***********************************************************************
        // speccode: speculative store. Original line:
        //      hull[TOP][1] = lastedge;
        value = lastedge;
        specstore_pointer((void *) &(hull[TOP][1]), sizeof (hull[TOP][1]), current, (void *) &value);
        //***********************************************************************
      } //outside==0
      // End of process, proceed with next point

      /**********************************************************************		\
      // speccode: end of thread and commit						\
      //           First check if block of iterations is done **************/		\
      //labelEndIteration_1:
      if ((iteration != varblock[1][current] + ini) && (iteration < (N+1-3) - 1)) {
        iteration = iteration + 1;
        goto labelStartIteration_1;
      }
      /* Thread done, perform the commit */
labelSquash_1:
      retflag = threadend_pointer(&current);
      /* End if job done */
      if (retflag == JOBDONE) goto labelEndLoop_1;
      if (retflag == JOBTODO) {
        /*  Set loop variable */
        iteration = varblock[0][current] + ini;
        varblock[2][current] = varblock[2][current] + 1;
        goto labelStartIteration_1;
      }
labelEndLoop_1:
      ; /*freeWheel(current);*/
    } /*End for initLoop()*/

    } // END OMP PARALLEL

    /************************************************************************/
    // speccode: Stop measuring time
    spec_stop = get_time();
    spec_time += spec_stop - spec_start;
    /************************************************************************/

    //	speccode: free speculation structures
    specend();

    // Show the results

    e = hull[TOP][1];
    p = hull[e][source];
e400:
    printf("%f %f\n",input[x][p],input[y][p]);
    e = hull[e][succ];
    p = hull[e][source];
    if ( e != hull[TOP][1] ) 
      goto e400;

    printf("Total number of edges: %d\n", hull[TOP][5]);

    // speccode: statistics for target loop.
    //specepilogue(2);
#ifdef TIME_SPECLOAD
    printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
    printSpecstoreTime();
#endif
#ifdef MEASURE_COUNTERS
    specepilogue();
    // printAccesses();
#endif

    // -----------------------------------------------------------
    // speccode: print time results
    fprintf(stderr,"--------------------------------------------------\n");
    fprintf(stderr,"speccode: Time elapsed: %lf seconds.\n", spec_time);
    // -----------------------------------------------------------

    return 0;

  }

  double rightturn(double dx, double dy, double px, double py, double qx, double qy) {
    return (dx * py + dy * qx + px * qy - qx * py - dx * qy - px * dy);
  }

