// =====================================================================
// TREECODE: Fortran77 version of hierarchical N-body code (O(N log N)).

//     This program evolves a self-gravitating N-body system, using
//     a hierarchical O(N log N) algorithm to compute gravitational
//     forces (see Barnes, J.E. & Hut, P. 1986, Nature, 324, 446).
//     The computational system of units is determined by the input
//     data, with the assumption that G = 1.  Particles are not
//     required to have identical masses.

//     This code is based on a Fortran implementation of the
//     hierarchical algorithm developed and kindly provided by
//     Lars Hernquist.  Several additional refinments by Jun Makino
//     have been incorporated as well.

//     The present version of the code includes quadrupole moments
//     when calculating the gravitational fields of cells.  It also
//     incorporates an improved opening criterion which increases
//     the accuracy of the forces in certain exceptional situations.

//					Joshua Barnes
//					April 23, 1993
// =====================================================================


// ==============================================================
// TREECD: top-level N-body evolution program.  Its tasks are to:
//  *  read input parameters and initial conditions;
//  *  perform an initial force calculation, reporting
//     system diagnostics and time required;
//  *  advance the system for NSTEPS time-steps, periodically
//     saving a snapshot of the system;
//  *  terminate the simulation and close data files.
// ==============================================================

//      PROGRAM TREECD

//***********************************************************************
//       speccode: Directives for profiling
//$PRAGMA C(collector_resume,collector_pause)
//***********************************************************************

#include "treedefs.c"
#include <math.h>
int ndiag, n;

//***********************************************************************
//       speccode: Stop profiling
//       call collector_pause()
//***********************************************************************

int main() {
	//EQUIVALENCE (MID(1,1), POS(1,1))
	mid = (double **) &pos;
	//EQUIVALENCE (CLSIZE(INCELL), RCRIT2(INCELL))
	clsize = (double *) &rcrit2;
//       ---------------------
//       Initialize cpu timer.
//       ---------------------
	cput0 = cputimer();
//	----------------------------------------------------------------
//	Open data files, read input parameters and initial system state,
//       initialize system parameters, and start output.
//	----------------------------------------------------------------
        inparm();
	getbds();
	initcd();
	begout();
//       -----------------------------------------------
//	Generate tree and compute initial acceleration.
//	-----------------------------------------------
        mktree();
	accel();
//	------------------------------------------
//	Output timing data and system diagnostics.
//	------------------------------------------
	outtim();
	outlog();
	diags();
//       -----------------------------------------------
//       Set number of steps between diagnostic outputs.
//       -----------------------------------------------
	if (1 < nsteps/128)
		ndiag = nsteps / 128;
	else
		ndiag = 1;
//	=================================================================
//	Primary loop to advance system state for a given number of steps.
//	=================================================================
	for (n=0; n<nsteps; n++) { 
//	  ------------------------------------------------------
//	  Advance velocity by 1/2 step, positions by 1 timestep.
//	  ------------------------------------------------------
	  advvel(dtime/2.0);
	  advpos(dtime);
//	  ----------------------------
//	  Create tree, compute forces.
//	  ----------------------------
	  mktree();
	  accel();
//	  ----------------------------------------------------
//	  Advance velocity 1/2 step to synchronize tvel, tpos.
//	  ----------------------------------------------------
          advvel(dtime/2.0);
//	  ----------------------------------------------------
//	  At appropriate intervals, output system diagnostics.
//	  ----------------------------------------------------
          if (n%ndiag == 0) {
		  outlog();
		  diags();
	  }
//         ----------------------------------------------
//	  At appropriate intervals, output system state.
//         ----------------------------------------------
          if (n%nout == 0)
		  putbds(false);
	}
//	----------------------------------
//	Output final snapshot with masses.
//	----------------------------------
        putbds(true);
//	------------------------------------------------------------------
//	Stop timing, write timing data, close files, terminate simulation.
//	------------------------------------------------------------------
	outcpu();
	stpout();
}

// ----------------------------------------------------------
// INITCD: initialize system parameters that depend on either
// the input data or defined PARAMETERS.
// ----------------------------------------------------------

void initcd() {
 
//	---------------------------------------
//	Initialize position and velocity times.
//	---------------------------------------
        tpos = tnow;
	tvel = tnow;
//	---------------------------------
//	Initialize box coordinate system.
//	---------------------------------
	rsize = 4.0;
}

// ---------------------------------------------------------------
// ADVPOS: advance the positions of the bodies for a timestep DT.
// ---------------------------------------------------------------
 
void advpos(double dt) {

	int k,p;

//	------------------------------------------------- 
//	Loop over all spatial coordinates for all bodies.
//	------------------------------------------------- 
        for (k = 0; k<NDIM; k++) {
		for (p=0; p<nbody; p++) {
			pos[p][k] = pos[p][k] + dt*vel[p][k];
		}
	}
//	----------------------------------
//	Update position time, system time.
//	----------------------------------
        tpos = tpos+dt;
        tnow = tpos;
}

// ----------------------------------------------------------------
// ADVVEL: advance the velocities of the bodies for a timestep dt.
// ----------------------------------------------------------------
 
void advvel(double dt) {
        
	int k, p;

//	------------------------------------------------- 
//	Loop over all spatial coordinates for all bodies.
//	------------------------------------------------- 
        for (k = 0; k<NDIM; k++) {
		for (p=0; p<nbody; p++) {
			vel[p][k] = vel[p][k] + acc[p][k];
		}
	}
//	----------------------------------
//	Update velocity time, system time.
//	----------------------------------
        tvel = tvel+dt;
        tnow = tvel;
}

// ----------------------------------------------------------------
// DIAGS: compute diagnostics for the system: total energy, total
// kinetic energy, total potential energy, angular momentum, center
// of mass coordinates, and center of mass velocity.
// ----------------------------------------------------------------
 
void diags() {
 
	int k,p;

//       --------------------------------------------------------------
//       Check that tpos and tvel are synchronized to one part in 1000.
//       --------------------------------------------------------------
        if (abs(tpos-tvel) > dtime/1.0E3)
		terror(" DIAGS: TIMES NOT SYNCHRONIZED");
//       ---------------------------------------------
//       Zero the accumulators for system diagnostics.
//       ---------------------------------------------
        mtot = 0.0;
        ektot = 0.0;
        eptot = 0.0;
	for (k=0; k<NDIM; k++) {
          cmpos[k] = 0.0;
          cmvel[k] = 0.0;
          amvec[k] = 0.0;
	}
//       ----------------------------------------------------------
//       Loop over bodies to compute net mass and potential energy.
//       ----------------------------------------------------------
	for (p=0; p<nbody; p++) {
          mtot = mtot + mass[p];
          eptot = eptot + 0.5 * mass[p] * phi[p];
	}
//       -----------------------------------------------------------------
//       Compute net kinetic energy, center of mass position and velocity.
//       -----------------------------------------------------------------
	for (k=0; k<NDIM; k++) {
	  for (p=0; p<nbody; p++) {
            ektot = ektot + 0.5 * mass[p] * (vel[p][k]*vel[p][k]);
            cmpos[k] = cmpos[k] + mass[p] * pos[p][k];
            cmvel[k] = cmvel[k] + mass[p] * vel[p][k];
	  }
          cmvel[k] = cmvel[k] / mtot;
          cmpos[k] = cmpos[k] / mtot;
	}
//       ----------------------------
//       Compute total system energy.
//       ----------------------------
        etot = ektot + eptot;
//       ---------------------------------------
//       Compute angular momentum of the system.
//       ---------------------------------------
        for (p=0; p<nbody; p++) {
          amvec[0] += mass[p] * (pos[p][2]*vel[p][3] - pos[p][3]*vel[p][2]);
          amvec[1] += mass[p] * (pos[p][3]*vel[p][1] - pos[p][1]*vel[p][3]);
          amvec[2] += mass[p] * (pos[p][1]*vel[p][2] - pos[p][2]*vel[p][1]);
	}
//       ----------------------------------
//       Write diagnostics to the log file.
//       ----------------------------------
        outdag();
}

// -------------------------------------------------------------
// TERROR: terminate the program as the result of a fatal error,
// closing the output files and writing timing information.
// -------------------------------------------------------------

void terror(char *msg) {

//       ------------------------------------
//       Write error message to the log file.
//       ------------------------------------
	outerr(msg);
//       ---------------------------------------------------
//       Output timing data, close files, terminate the run.
//       ---------------------------------------------------
        outcpu();
	stpout();
	exit(-1);
}

// =============================================================
// TREEGRAV: routines to perform hierarchical force calculation.
// =============================================================



// --------------------------------------------------------
// ACCEL: routine to compute the acceleration and potential
// for all bodies in the system.
// --------------------------------------------------------

void accel() {

//***********************************************************************
//       speccode: Interface for profiling API directives
//       //$PRAGMA C(collector_pause,collector_resume)
//***********************************************************************

	double cput1, cput2;
		
	int p;
	int mxsptr = 256;
	double eps2, phi0, acc0[NDIM], pos0[NDIM];
	double dx, dy, dz, dr2, dr2inv, drinv, phim, dr5inv, phiq;
	int q, nbterm, ncterm, sptr, stack[mxsptr], k;
	int skpslf;

//       ----------------------------------------
//       Initialize timing for force calculation.
//       ----------------------------------------
        cput1 = cputimer();

//       --------------------------------------------
//       Initialize the force evaluation diagnostics.
//       --------------------------------------------
        nbtot = 0;
	nctot = 0;
        ntmax = 0;

//***********************************************************************
//       speccode: start profiling
//	call collector_resume()
//***********************************************************************

//	---------------------------------------------------
//	Loop over bodies, computing force for each in turn.
//	---------------------------------------------------
	for (p=0; p<nbody; p++) {
		eps2 = eps*eps;
//	----------------------------------------------------------
//	Zero potential and acceleration for subsequent summations.
//	----------------------------------------------------------
        phi0 = 0.0;
        acc0[0] = 0.0;
        acc0[1] = 0.0;
        acc0[2] = 0.0;
//       -----------------------------------------
//       Copy position of P for quicker reference.
//       -----------------------------------------
        pos0[0] = pos[p][0];
        pos0[1] = pos[p][1];
        pos0[2] = pos[p][2];
//	----------------------------------------------
//	Initialize counters for number of force terms.
//	----------------------------------------------
        nbterm = 0;
        ncterm = 0;
	skpslf = false;
//	----------------------------------
//	Push the root cell onto the stack.
//	----------------------------------
        sptr = 0;
        stack[sptr] = root;
//	-------------------------------------
//	Loop while nodes on stack to process.
//	-------------------------------------
label_10:
	if (sptr >= 0) {
//	  --------------------------
//	  Pop node off top of stack.
//	  --------------------------
          q = stack[sptr];
          sptr = sptr - 1;
//         ---------------------------------------------
//         Compute distance to center-of-mass of node Q.
//         ---------------------------------------------
          dx = pos0[0] - pos[q][0];
          dy = pos0[1] - pos[q][1];
          dz = pos0[2] - pos[q][2];
          dr2 = dx*dx + dy*dy + dz*dz;
//	  -------------------------------
//	  Classify Q as a body or a cell.
//	  -------------------------------
          if (q < INCELL) {
//	    -----------------------------------
//	    A body: check for self-interaction.
//	    -----------------------------------
	    if (q != p) {
//             ------------------------------
//             Compute body-body interaction.
//             ------------------------------
              dr2inv = 1.0 / (dr2 + eps2);
              phim = mass[q] * sqrt(dr2inv);
              phi0 = phi0 - phim;
              phim = phim * dr2inv;
              acc0[0] -= phim * dx;
              acc0[1] -= phim * dy;
              acc0[2] -= phim * dz;
	      nbterm = nbterm + 1;
	    } else {
//	      -------------------------------------------
//	      Remember that self-interaction was skipped.
//	      -------------------------------------------
	      skpslf = true;
	    }
	  } else {
//           --------------------------------------------
//           A cell: test if interaction can be accepted.
//           --------------------------------------------
	    if (dr2 >= rcrit2[q]) {
//             ----------------------------------------
//             Accepted: compute body-cell interaction.
//             ----------------------------------------
              dr2inv = 1.0 / (dr2 + eps2);
              drinv = sqrt(dr2inv);
              phim = mass[q] * drinv;
              phi0 = phi0 - phim;
              phim = phim * dr2inv;
              acc0[0] -= phim * dx;
              acc0[1] -= phim * dy;
              acc0[2] -= phim * dz;
//	      ------------------------------------
//	      Optionally include quadrupole terms.
//	      ------------------------------------
	      if (usquad) {
                dr5inv = dr2inv * dr2inv * drinv;
                phiq = dr5inv * (0.5 * ((dx*dx - dz*dz) * quad[q][0] + 
                            (dy*dy - dz*dz) * quad[q][3]) + 
                     dx*dy * quad[q][1] + dx*dz * quad[q][2] + 
                     dy*dz * quad[q][4]);
                phi0 = phi0 - phiq;
                phiq = 5.0 * phiq * dr2inv;
                acc0[0] -= phiq*dx + dr5inv * (dx*quad[q][0] + dy*quad[q][1] + dz*quad[q][2]);
                acc0[1] -= phiq*dy + dr5inv * (dx*quad[q][1] + dy*quad[q][3] + dz*quad[q][4]);
                acc0[2] -= phiq*dz + dr5inv * (dx*quad[q][2] + dy*quad[q][4] - dz*(quad[q][0] + quad[q][3]));
	      }
	      ncterm = ncterm + 1;
	    } else {
//	      -----------------------------------
//	      Rejected: examine children of cell.
//	      -----------------------------------
              for (k=0; k< NSUBC; k++) {
//		--------------------------------------
//		Push existing children onto the stack.
//		--------------------------------------
                if (subp[q-MXNODE][k] != NULL) {
		  if (sptr >= mxsptr) {
		    terror(" TRWALK: STACK OVERFLOW");
		  }
                  sptr = sptr + 1;
                  stack[sptr] = subp[q-MXNODE][k];
		}
	      }
	    }
	  }
          goto label_10;
	}
//       ---------------------------------------------------
//       Check that self-interaction was explicitly skipped.
//       ---------------------------------------------------
        if (!skpslf) {
	  terror(" TRWALK: MISSED SELF-INTERACTION");
	}
//       ----------------------------------------------
//       Copy total potential and acceleration to body.
//       ----------------------------------------------
        phi[p] = phi0;
        acc[p][0] = acc0[0];
        acc[p][1] = acc0[1];
        acc[p][2] = acc0[2];
//       -------------------------------------
//       Update force calculation diagnostics.
//       -------------------------------------
        nbtot = nbtot + nbterm;
        nctot = nctot + ncterm;
	if (ntmax < ncterm+nbterm){
	  ntmax = ncterm+nbterm;
	}
     }// for

//***********************************************************************
//       speccode: stop profiling
//       call collector_pause()
//***********************************************************************

//       -----------------------------------------------
//       Compute average number of force terms per body.
//       -----------------------------------------------
        ntavg = (nbtot + nctot) / nbody;
//       -------------------------------------
//       Compute timing for force calculation.
//       -------------------------------------
        cput2 = cputimer();
        cpuacc = cput2 - cput1;
}
