
#include "treedefs.h"
#ifdef COMPILAR
// ============================================================
// TREEDEFS: include file with common definitions for TREECODE.
// ============================================================
 
// --------------------------------------------------------------
// Fundamental constants for treecode, not to be altered lightly.
// --------------------------------------------------------------
 
// HDLINE      identification string for the run.
// THETA       accuracy parameter.
// EPS         potential softening parameter.
// DTIME       N-body integration timestep.
// USQUAD      option to include quadrupole terms.
// NSTEPS      total number of timesteps in the simulation.
// NOUT        frequency of system record outputs.
 
extern char hdline[32];
extern double theta, eps, dtime;
extern int usquad;
extern int nsteps, nout;
 
// ----------------------------------
// System status and diagnostic data.
// ----------------------------------
        
// TNOW        value of current system time.
// MTOT        total mass of system.
// EKTOT       total kinetic energy.
// EPTOT       total potential energy.
// ETOT        sum of KE and PE.
// CMPOS       position of center of mass.
// CMVEL       velocity of center of mass.
// AMVEC       total angular momentum vector.
// CPUMKT      cpu time used in last tree construction.
// CPUACC      cpu time used in last force calculation.
// CPUT0       baseline for computing total cpu time.
// NBTOT       total number of body-body interactions.
// NCTOT       total number of body-cell interactions.
// NTMAX       maximum number of interactions for any body.
// NTAVG       average number of interactions per body.
 
extern double tnow, mtot, ektot, eptot, etot, cmpos[NDIM], cmvel[NDIM], amvec[NDIM], cpumkt, cpuacc;
extern double cput0;
extern int nbtot, nctot, ntmax, ntavg;
 
// ---------------------------
// Tree structure data arrays.
// ---------------------------
 
// Bodies and cells are used to represent the tree.  The information for a
// given node is distributed across a number of arrays.  Quantities defined
// for both bodies and cells, such as mass and position, are stored in
// arrays with indicies running from 1 to MXNODE.  Quantities defined only
// for bodies are stored in arrays with indicies running from 1 to MXBODY,
// while those defined only for cells are stored in arrays with indicies
// running from INCELL to MXNODE.  With this convention, indicies can refer
// to either bodies or cells without conflict, and the type is deduced by
// comparing the index to the index of the first cell.  During tree
// construction, descendent indicies are stored in the SUBP arrays:
//
//          +-------------------------------------------------------+
// ROOT --> | CELL: MASS, POS, RCRIT2, QUAD, SUBP:[/,o,/,/,/,/,o,/] |
//          +----------------------------------------|---------|----+
//                                                   |         |
//     +---------------------------------------------+         |
//     |                                                       |
//     |    +--------------------------------+                 |
//     +--> | BODY: MASS, POS, VEL, ACC, PHI |                 |
//          +--------------------------------+                 |
//                                                             |
//     +-------------------------------------------------------+
//     |
//     |    +-------------------------------------------------------+
//     +--> | CELL: MASS, POS, RCRIT2, QUAD, SUBP:[o,/,/,o,/,/,o,/] |
//          +--------------------------------------|-----|-----|----+
//                                                etc   etc   etc
 
// MXBODY      maximum number of bodies allowed.
// MXCELL      maximum number of cells allowed.
// INCELL      index of first cell in arrays.
// MXNODE      maximum number of nodes (bodies + cells).

// NBODY       number of bodies in the system.
// NCELL       number of cells currently in use.
// MASS        mass of each node.
// POS         position of each node.
// VEL         velocity of each body.
// ACC         acceleration of each body
// PHI         potential of each body.
// RCRIT2      critical distances**2 of each cell.
// QUAD        quad moments of each cell.
// SUBP        descendent of each cell.
// TPOS        current position time.
// TVEL        current velocity time.
// RSIZE       side-length of root cell.
// ROOT        index of cell representing root.
 
extern int nbody, ncell;
extern double mass[MXNODE], pos[MXNODE][NDIM];
extern double vel[MXBODY][NDIM], acc[MXBODY][NDIM], phi[MXBODY];
//double rcrit2(incell:mxnode), quad(incell:mxnode,NQUAD)
extern double rcrit2[MXNODE-INCELL];
extern double quad[MXNODE-INCELL][NQUAD];
//int subp(incell:mxnode,NSUBC), root
extern int subp[MXNODE-INCELL][NSUBC];
extern int root;
double tpos, tvel, rsize;
 
// These definitions apply only during tree construction.

// MID         geometric center of each cell [equiv to POS].
// CLSIZE      size of each cell [equiv to RCRIT2].

//double mid[MXNODE][NDIM];
	//EQUIVALENCE (MID(1,1), POS(1,1))
double **mid;
//clsize(INCELL:MXNODE)
//double clsize[MXNODE-INCELL];
	//EQUIVALENCE (CLSIZE(INCELL), RCRIT2(INCELL))
double *clsize;


// --------------------------
// Input/output unit numbers.
// --------------------------
 
FILE *ulog, *ubodo, *ubodf;
FILE *uterm, *upars, *ubodi;

#include <sys/time.h>

double cputimer()
{
   struct timeval tv;
   gettimeofday(&tv, (void *)0);
   return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}
#endif
