// ==================================
// TREEIO: I/O routines for TREECODE.
// ==================================

#include "treedefs.c"

// --------------------------------------------------------------------
// INPARM: read in following input parameters:
//
//        hdline    : identification string for the run.
//        nsteps    : number of timesteps.
//        nout      : output system state once every nout steps.
//        dtime     : the timestep.
//        theta     : error tolerance; 0.0 => exact (PP) calculation.
//        eps       : potential softening parameter.
//        usquad    : option to include (.TRUE.) quadrupole terms.

// --------------------------------------------------------------------

void inparm() {
 
        char parsfl[8] = "treepars";
	int aux;
	FILE *inputfile;
 
//       -----------------------------------------------------
//       Open parameter file, read parameters, close the file.
//       -----------------------------------------------------
	if (inputfile = fopen("treepars","r")) {
          // READ (UPARS, '(A)') HDLINE
          if(fscanf(inputfile,"%s",hdline) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
          // READ (UPARS, *) NSTEPS, NOUT, DTIME, THETA, EPS, USQUAD
          if(fscanf(inputfile,"%d",&nsteps) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
          if(fscanf(inputfile,"%d",&nout) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
          if(fscanf(inputfile,"%lf",&dtime) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
          if(fscanf(inputfile,"%lf",&theta) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
          if(fscanf(inputfile,"%lf",&eps) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
          if(fscanf(inputfile,"%d",&usquad) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }

	} else { 
	  printf("Error reading from file \n");
	  exit(-1);
	}

	fclose(inputfile);
}

// ---------------------------------------------------------
// GETBDS: read in particle data, assumed to be of the form:
//
//       nbody		: number of bodies.
//       NDIM		: number of space dimensions.
//       time		: initial value of time.
//       mass(1)		: mass of first particle.
//       .......	
//       mass(n)		: mass of last particle.
//       x(1) y(1) z(1)	: position of first particle.
//       ..............
//       x(n) y(n) z(n)	: position of last particle.
//       u(1) v(1) w(1)	: velocity of first particle.
//       ..............
//       u(n) v(n) w(n)	: velocity of last particle.
// ---------------------------------------------------------

void getbds() {

        int indim,p,k;
	FILE *inputfile;

//       --------------------
//       Open body data file.
//       --------------------

        if (inputfile = fopen("treebodi","r")) {

//       ------------------
//       Read in body data.
//       ------------------

          if(fscanf(inputfile,"%d",&nbody) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
          if(fscanf(inputfile,"%d",&indim) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }

//       ----------------------------------------------------------
//       Terminate the simulation if NBODY or INDIM not consistent.
//       ----------------------------------------------------------
	  if (nbody <= 0 || nbody > MXBODY || indim != NDIM) {
	    terror(" GETBDS: INCONSISTENT OR ILLEGAL INPUTS");
	  }
//       --------------------------------------------
//       Read time, masses, positions and velocities.
//       --------------------------------------------

          if(fscanf(inputfile,"%lf",&tnow) < 0){
	    printf("Error reading from file\n");
            exit(-1);
	  }
        //DO 10 P = 1, NBODY
        //  READ (UBODI, *) MASS(P)
  //10    CONTINUE
	  for (p=0; p<nbody; p++) {
            if(fscanf(inputfile,"%lf",&mass[p]) < 0){
	      printf("Error reading from file\n");
              exit(-1);
	    }
	  }
    //    DO 20 P = 1, NBODY
    //      READ (UBODI, *) (POS(P,K), K = 1, NDIM)
  //20    CONTINUE
	  for (p=0; p<nbody; p++) {
	    for (k=0; k<NDIM; k++) {
              if(fscanf(inputfile,"%lf",&pos[p][k]) < 0) {
	        printf("Error reading from file\n");
                exit(-1);
	      }
	    }
	  }
    //    DO 30 P = 1, NBODY
    //      READ (UBODI, *) (VEL(P,K), K = 1, NDIM)
  //30    CONTINUE
	  for (p=0; p<nbody; p++) {
	    for (k=0; k<NDIM; k++) {
              if(fscanf(inputfile,"%lf",&vel[p][k]) < 0){
	        printf("Error reading from file\n");
                exit(-1);
	      }
	    }
	  }

	} else { 
	  printf("Error reading from file treebodi.\n");
	  exit(-1);
	}

//       --------------------
//       Close up input file.
//       --------------------
	fclose(inputfile);
}

// ----------------------------------------------------
// BEGOUT: initialize disk files for subsequent output.
// ----------------------------------------------------

void begout() {

//       --------------------------------
//       Open log file and output header.
//       --------------------------------
	if (ulog = fopen("treelogs","w")) {
	  outhdr();
//       ----------------------------
//       Open body data output files.
//       ----------------------------

	  if (ubodo = fopen("treebodo","w")) {
	    if (ubodo = fopen("treebodf","w")) {
	    } else { 
	      printf("Error opening from file treebodf.\n");
	      exit(-1);
	    }
	  } else { 
	    printf("Error opening from file treebodo.\n");
	    exit(-1);
	  }
	} else { 
	  printf("Error opening from file treelogs.\n");
	  exit(-1);
	}

}
 
// ----------------------------------------------------------
// OUTHDR: write header to log file listing input parameters.
// ----------------------------------------------------------

void outhdr() {
 
	char aux;
	if (usquad == true)
		aux = 'T';
	else
		aux = 'F';
	fprintf(ulog, "\n\n");
	fprintf(ulog, " ************************************************************************\n");
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *                   %s                   *\n",hdline);
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " ************************************************************************\n");
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *    NBODY =   %d    NSTEPS = %d    NOUT =   %d    DTIME =  %3f    *\n", nbody, nsteps, nout, dtime);
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *    USEQUAD = %c    EPS = %3f      THETA =  %2f                   *\n", aux, eps, theta);
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *    OPTIONS:  IMPROVED ACCEPTANCE CRITERION                           *\n");
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " ************************************************************************\n");

}

// ---------------------------------------------------
// OUTTIM: write force calculation timing to log file.
// ---------------------------------------------------
 
void outtim() {

	fprintf(ulog, "\n      CPU TIME USED:          MKTREE           ACCEL           TOTAL\n");
	fprintf(ulog, "        (SECONDS)               %2f            %2f            %2f\n", cpumkt, cpuacc, cpumkt+cpuacc);

}

// --------------------------------------------------------
// OUTLOG: write force calculation diagnostics to log file.
// --------------------------------------------------------

void outlog() {
 
        double cput;
 
//       -----------------------------
//       Get system-specific cpu time.
//       -----------------------------
        cput = cputimer();
	cput = cput - cput0;
//       -------------------------------------------------
//       OUTPUT CPU TIME AND FORCE EVALUATION DIAGNOSTICS.
//       -------------------------------------------------
	fprintf(ulog, "\n\n\n      TIME    NCELL     NBTOT     NCTOT   NTMAX   NTAVG  CPUTIME\n");
	fprintf(ulog, "     %3f     %d    %d    %d     %d     %d     %2f\n",tnow, ncell, nbtot, nctot, ntmax, ntavg, cput/60.0);

}

// -------------------------------------------------
// OUTDAG: write N-body diagnostic data to log file.
// -------------------------------------------------
 
void outdag() {

        //INTEGER K, IPRINT
        //DATA IPRINT /0/
	int iprint = 0;

        //WRITE (*, '(1X,I4,1X,6(1PE12.4))') IPRINT, &
        //            (CMPOS(K), K = 1, NDIM), (CMVEL(K), K = 1, NDIM)
        //IPRINT = IPRINT+1
//       ----------------------------------------------------------------
//       Write mass, energy, angular momentum, center of mass quantities.
//       ----------------------------------------------------------------
	fprintf(ulog, "\n\n             ETOT            EKIN            EPOT            MASS\n");
	fprintf(ulog, "      %8f      %8f     %8f      %8f\n",etot,ektot,eptot,mtot);
	fprintf(ulog, "\n\n                                X               Y               Z\n");
	fprintf(ulog, "             JTOT    %6E   %6E   %6E\n",amvec[0], amvec[1], amvec[2]);
	fprintf(ulog, "             CMPOS   %6E   %6E   %6E\n",cmpos[0], cmpos[1], cmpos[2]);
	fprintf(ulog, "             CMVEL   %6E   %6E   %6E\n",cmvel[0], cmvel[1], cmvel[2]);
}

// -----------------------------------------------------------
// OUTERR: output error messages to the log file and terminal.
// -----------------------------------------------------------

void outerr(char *msg) {

//       ------------------------------------------------------
//       Write the message, surrounded by stars for visibility.
//       ------------------------------------------------------
	fprintf(ulog, "\n ************************************************************************\n");
	fprintf(ulog, "%s",msg);
	fprintf(ulog, "\n ************************************************************************\n");
//       ---------------------------
//       Repeat message to terminal.
//	---------------------------
	fprintf(uterm, "\n ************************************************************************\n");
	fprintf(uterm, "%s",msg);
	fprintf(uterm, "\n ************************************************************************\n");
}

// -----------------------------------------------
// OUTCPU: output cpu timing data to the log file.
// -----------------------------------------------

void outcpu() {

	double cput;

//       -----------------------------
//       Get system-specific cpu time.
//       -----------------------------
	cput = cputimer();
	cput = cput - cput0;
	fprintf(ulog, "\n\n ELAPSED CPU TIME (SECONDS) : %16.2f",cput);
}

// ------------------------------------------------------
// PUTBDS: write body data to output file or final file.
// ------------------------------------------------------

void putbds(int final) {
 
//       ------------------------------------
//       Decide which file to use for output.
//       ------------------------------------
        if (final) {
//         -------------------------------------------
//         Output final coordinates, including masses.
//         -------------------------------------------

	  fprintf(ubodf, " %20d\n",nbody);
	  fprintf(ubodf, " %20d\n",NDIM);
	  fprintf(ubodf, " %20.12E\n",tnow);

	  putmas(ubodf);
	  putpos(ubodf);
	  putvel(ubodf);
	} else {
//	  ---------------------------------
//	  Output particle coordinates only.
//	  ---------------------------------

	  fprintf(ubodo, " %20d\n",nbody);
	  fprintf(ubodo, " %20d\n",NDIM);
	  fprintf(ubodo, " %20.12E\n",tnow);

	  putpos(ubodo);
	  putvel(ubodo);

	}
}

// -------------------------------
// PUTMAS: output body mass array.
// -------------------------------

void putmas(FILE *file) {

	int i;

	for (i=0; i<nbody; i++) {
	  fprintf(file, " %20.12E\n", mass[i]);
	}

}

// -----------------------------------
// PUTPOS: output body position array.
// -----------------------------------

void putpos(FILE *file) {

	int i,k;

	for (i=0; i<nbody; i++) {
	  for (k=0; k<NDIM; k++) {
	    fprintf(file, " %20.12E", pos[i][k]);
	  }
	  fprintf(file, "\n");
	}

}

// -----------------------------------
// PUTVEL: output body velocity array.
// -----------------------------------

void putvel(FILE *file) {

	int i,k;

	for (i=0; i<nbody; i++) {
	  for (k=0; k<NDIM; k++) {
	    fprintf(file, " %20.12E", vel[i][k]);
	  }
	  fprintf(file, "\n");
	}

}
 
// ----------------------------------------------
// STPOUT: terminate output and close open files.
// ----------------------------------------------

void stpout() {
 
	fprintf(ulog, "\n\n CALCULATION TERMINATED");
//       ---------------------
//       Close the open files.
//       ---------------------
	fclose(ubodo);
	fclose(ubodf);
	fclose(ulog);
}
