// ================================================================
// TREELOAD: routines to handle tree construction and verification.
// ================================================================

#include "treedefs.c"
#include <math.h>
 
// ----------------------------------------------------------------
// MKTREE: initialize the tree structure for the force calculation.
// ----------------------------------------------------------------
 
void mktree() {

        double cput1, cput2;
 
//       ----------------------------------
//       Start timing of tree construction.
//       ----------------------------------
        cput1 = cputimer();
//       --------------------------------------
//       Expand root volume to hold all bodies.
//       --------------------------------------
	expbox();
//	--------------------------
//       Load bodies into the tree.
//	--------------------------
	ldtree();
//       ------------------------------------------------
//       Compute masses, center of mass coordinates, etc.
//       ------------------------------------------------
        hackcm();
//       -------------------------------------
//       Compute timing for tree construction.
//       -------------------------------------
        cput2 = cputimer();
        cpumkt = cput2 - cput1;
}
 
// -----------------------------------------------------
// EXPBOX: enlarge system cube to include all particles.
// -----------------------------------------------------
 
void expbox() {
 
	double xyzmax;
	int i,j;
 
//       ------------------------------
//       Find maximum coordinate value.
//       ------------------------------
	xyzmax = 0;
	for (i=0; i<NDIM; i++) {
	  for (j=0; j<nbody; j++) {
	    if (xyzmax < abs(pos[i][j]))
	      xyzmax = abs(pos[i][j]);
	  }
	}
//	-----------------------------------------
//	Expand box by factors of 2 until it fits.
//	-----------------------------------------
label_30:
        if (xyzmax >= rsize/2.0) {
	  rsize = 2.0 * rsize;
	  goto label_30;
	}
}


// ---------------------------------------------------------------------
// LDTREE: construct tree body-by-body.  This phase initializes the SUBP
// array and loads the geometric midpoint into the MID of each cell.
// ---------------------------------------------------------------------

void ldtree() {

        int k, p;

//       ---------------------------------------
//       Deallocate current tree, begin new one.
//       ---------------------------------------
        ncell = 0;
        root = mkcell();
//	------------------------------------------
//	Initialize midpoint and size of root cell.
//	------------------------------------------
	for (k=0; k<NDIM; k++) {
	  mid[root][k] = 0.0;
	}
	clsize[root] = rsize;
//       ---------------------------------------------
//       Load bodies into the new tree, one at a time.
//       ---------------------------------------------
	for (p=0; p<nbody; k++) {
          ldbody(p);
	}
}
 
// ----------------------------------------
// LDBODY: load body P into tree structure.
// ----------------------------------------
 
void ldbody(int p) {

        int q, qind, c, k, p0;

//       ---------------------------------------------
//       Start Q,QIND pair in correct subcell of root.
//       ---------------------------------------------
        q = root;
        qind = sbindx(p,q);
//       -----------------------------------------------------
//       Loop descending tree until an empty subcell is found.
//       -----------------------------------------------------
label_20:
	if (subp[q-MXNODE][qind] != NULL) {
//         --------------------------------------
//         On reaching another body, extend tree.
//         --------------------------------------
          if (subp[q-MXNODE][qind] < INCELL) {
//           -------------------------------------------
//           Allocate an empty cell to hold both bodies.
//           -------------------------------------------
            c = mkcell();
//           ------------------------------------------------------
//           Locate midpoint of new cell wrt. parent, and set size.
//           ------------------------------------------------------
            for (k=0; k<NDIM; k++) {
	      if (pos[p][k] >= mid[q][k]) {
                mid[c][k] = mid[q][k] + clsize[q]/4.0;
	      } else {
                mid[c][k] = mid[q][k] - clsize[q]/4.0;
	      }
	    }
	    clsize[c] = clsize[q] / 2.0;
//           ------------------------------------------------------
//           Store old body in appropriate subcell within new cell.
//           ------------------------------------------------------
            p0 = subp[q-MXNODE][qind];
            subp[c-MXNODE][sbindx(p0,c)] = p0;
//           ---------------------------------------------
//           Link new cell into tree in place of old body.
//           ---------------------------------------------
            subp[q-MXNODE][qind] = c;
	  }
//         --------------------------------------------------------
//         At this point, the node indexed by Q,QIND is known to be
//	  a cell, so advance to the next level of tree, and loop.
//         --------------------------------------------------------
          q = subp[q-MXNODE][qind];
          qind = sbindx(p,q);
          goto label_20;
	}
//       ---------------------------------------------
//       Found place in tree for P, so store it there.
//       ---------------------------------------------
        subp[q-MXNODE][qind] = p;
}
 
// -------------------------------------------------------
// SBINDX: compute subcell index for node P within cell Q.
// -------------------------------------------------------
 
int sbindx(int p, int q) {

	int k;
	int ret_aux;
 
//       ---------------------------------------------------
//       Initialize subindex to point to lower left subcell.
//       ---------------------------------------------------
        ret_aux = 1;
//       ---------------------------------
//       Loop over all spatial dimensions.
//       ---------------------------------
        for (k=0; k<NDIM; k++) {
	  if (pos[p][k] >= mid[q][k])
	    ret_aux += pow(2,(3-k));
	}
	return ret_aux;
}
 
// ---------------------------------------------------------
// MKCELL: function to allocate a cell, returning its index.
// ---------------------------------------------------------
 
int mkcell() {

	int i;
	int ret_aux;
 
//       ----------------------------------------------------------
//       Terminate simulation if no remaining space for a new cell.
//       ----------------------------------------------------------
        if (ncell >= MXCELL)
	  terror(" MKCELL: NO MORE MEMORY");
//       ----------------------------------------------------
//       Increment cell counter, initialize new cell pointer.
//       ----------------------------------------------------
        ncell = ncell + 1;
        ret_aux = ncell + MXBODY;
//       --------------------------------------
//       Zero pointers to subcells of new cell.
//       --------------------------------------
        for (i=0; i<NSUBC; i++) {
          subp[ret_aux-MXNODE][i] = NULL;
	}
	return ret_aux;
}
 
// ----------------------------------------------------------------
// HACKCM: compute cell masses, c.m. positions, check tree structure,
// assign critical radii, and compute quadrupole moments.  Whew//
// ----------------------------------------------------------------
 
void hackcm() {

	int ind[MXCELL], p, q, i, j, k, l, m, n;
	double pos0[NDIM], dist2;
 
//       ---------------------------------------
//       List cells in order of decreasing size.
//       ---------------------------------------
         bflist(ind);
//       --------------------------------------------
//       Loop processing cells from smallest to root.
//       --------------------------------------------
        for (i=ncell-1; i>=0; i--) {
          p = ind[i];
//         --------------------------------------------------------------
//         Zero accumulators for this cell.  A temporary variable is used
//	  for the c. of m. so as to preserve the stored midpoints.
//         --------------------------------------------------------------
          mass[p] = 0.0;
	  for (k=0; k<NDIM; k++) {
            pos0[k] = 0.0;
	  }
//         -------------------------------------------------------------
//         Compute cell properties as sum of properties of its subcells.
//         -------------------------------------------------------------
          for (j=0; j<NSUBC; j++) {
            q = subp[p-MXNODE][j];
//           ------------------------------
//           Only access cells which exist.
//           ------------------------------
            if (q != NULL) {
//             -------------------------------------------------------
//             Sum properties of subcells to obtain values for cell P.
//             -------------------------------------------------------
              mass[p] += mass[q];
	      for (k=0; k<NDIM; k++) {
                pos0[k] += mass[q] * pos[q][k];
	      }
	    }
	  }
//         --------------------------------------------------------
//         Normalize center of mass coordinates by total cell mass.
//         --------------------------------------------------------
          for (k=0; k<NDIM; k++) {
            pos0[k] = pos0[k] / mass[p];
	  }
//         -----------------------------------------------------------------
//         Check tree, compute cm-to-mid distance, and assign cell position.
//         -----------------------------------------------------------------
          dist2 = 0.0;
	  for (k=0; k<NDIM; k++) {
            if (pos0[k] < mid[p][k] - clsize[p]/2.0 || 
      		pos0[k] >= mid[p][k] + clsize[p]/2.0) {
	      fprintf(ulog, "\n TREE ERROR %6d %6d %14.6lf %14.6lf %14.6lf\n",p,k,pos0[k],mid[p][k],clsize[p]);
              terror(" HACKCM: TREE STRUCTURE ERROR");
	    }
            dist2 += (pos0[k] - mid[p][k])*(pos0[k] - mid[p][k]);
//	    --------------------------------------------------------
//	    Copy cm position to cell.  This overwrites the midpoint.
//	    --------------------------------------------------------
	    pos[p][k] = pos0[k];
	  }

//	  ------------------------------------------------------------
//	  Assign critical radius for cell, adding offset from midpoint
//	  for more accurate forces.  This overwrites the cell size.
//	  ------------------------------------------------------------
          rcrit2[p] = (clsize[p] / theta + sqrt(dist2))*(clsize[p] / theta + sqrt(dist2));

	}
//	----------------------------------------
//	Compute quadrupole moments, if required.
//	----------------------------------------
	if (usquad) {
//         -------------------------------
//         Loop processing cells as above.
//         -------------------------------
          for (i=ncell-1; i>=0; i--) {
            p = ind[i];
//           --------------------------------------------
//           Zero accumulator for quad moments of cell P.
//           --------------------------------------------
            for (k=0; k<NQUAD; k++) {
              quad[p][k] = 0.0;
	    }
//           --------------------------------
//           Loop over descendents of cell P.
//           --------------------------------
            for (j=0; j<NSUBC; j++) {
	      q = subp[p-MXNODE][j];
	      if (q != NULL) {
//               --------------------------------------------------------
//               Sum properties of subcell Q to obtain values for cell P.
//               --------------------------------------------------------
                for (m=0; m<min(2,NDIM); m++) {
                  for (n=m; n<NDIM; n++) {
                    l = (m-1) * (NDIM-1) + n;
                    quad[p][l] += 3.0 * mass[q] * (pos[q][m] - pos[p][m]) * (pos[q][n] - pos[p][n]);
		    if (m == n) {
                      for (k=0; k<NDIM; k++) {
                        quad[p][l] = quad[p][l] - mass[q] * ((pos[q][k] - pos[p][k])*(pos[q][k] - pos[p][k]));
		      }
		    }
//                   -------------------------------------------
//                   If Q itself is a cell, add its moments too.
//                   -------------------------------------------
                    if (q >= INCELL)
                      quad[p][l] = quad[p][l] + quad[q][l];
		  }
		}
	      }
	    }
	  }
	}
}

// -----------------------------------------------------------------
// BFLIST: list cells in breadth-first order, from largest (root) to
// smallest.  Thanks to Jun Makino for this elegant routine.
// -----------------------------------------------------------------
 
void bflist(int *ind) {
 
	int facell, lacell, nacell, k, i;
 
//	-----------------------------------------
//	Start scan with root as only active cell.
//	-----------------------------------------
	ind[0] = root;
	facell = 0;
	lacell = 0;
//	-----------------------------------
//	Loop while active cells to process.
//	-----------------------------------
label_10:
	if (facell <= lacell) {
//	  ----------------------------------------------
//	  Start counting active cells in next iteration.
//	  ---------------------------------------------- 
	  nacell = lacell;
//	  ---------------------------------------
//	  Loop over subcells of each active cell.
//	  ---------------------------------------
	  for (k=0; k<NSUBC; k++) {
	    for (i=facell; i<lacell; i++) {
//	      -------------------------------------------
//	      Add all cells on next level to active list.
//	      -------------------------------------------
	      if (subp[ind[i]-MXNODE][k] >= INCELL) {
		nacell = nacell + 1;
		ind[nacell] = subp[ind[i]-MXNODE][k];
	      }
	    }
	  }
//	  ------------------------------------------------------
//	  Advance first and last active cell indicies, and loop.
//	  ------------------------------------------------------
	  facell = lacell + 1;
	  lacell = nacell;
	  goto label_10;
	}
//	--------------------------------------------------
//	Above loop should list all cells; check the count.
//	--------------------------------------------------
	if (nacell != ncell)
	  terror("  BFLIST: INCONSISTENT CELL COUNT");
}
