// ============================================================
// TREEDEFS: include file with common definitions for TREECODE.
// ============================================================
 
// --------------------------------------------------------------
// Fundamental constants for treecode, not to be altered lightly.
// --------------------------------------------------------------
 
// NULL        value denoting pointer to nothing.
// NDIM        number of spatial dimensions.
// NSUBC       number of descendents per cell.
// NQUAD       number of independent quadrupole components.
 
#include <stdlib.h>
#include <stdio.h>

#define NULO 0
#define NDIM 3
//#define NSUBC 2^NDIM
#define NSUBC 8
#define NQUAD 2*NDIM-1
 
#define MXSPTR 256
// -------------------------------------------------------
// Treecode control parameters, input from parameter file.
// -------------------------------------------------------
 
#define false 0
#define true 1
 
// MXBODY      maximum number of bodies allowed.
// MXCELL      maximum number of cells allowed.
// INCELL      index of first cell in arrays.
// MXNODE      maximum number of nodes (bodies + cells).
 
#define MXBODY 8192
#define MXCELL MXBODY
#define INCELL MXBODY + 1
#define MXNODE MXBODY + MXCELL
 
// Function declarations

void initcd();
void advpos(double dt);
void advvel(double dt);
void diags();
void terror(char *msg);
void accel();
void inparm();
void getbds();
void begout();
void outhdr();
void outtim();
void outlog();
void outdag();
void outerr(char *msg);
void outcpu();
void putbds(int final);
void putmas(FILE *file);
void putpos(FILE *file);
void putvel(FILE *file);
void stpout();
void mktree();
void expbox();
void ldtree();
void ldbody(int p);
void hackcm();
void bflist(int *ind);

int mkcell();
int sbindx(int p, int q);
double cputimer();
double get_time();


#ifdef SEQUENTIAL
/* Funtion to get the current time.  */
double get_time()
{
	 struct timeval tv;
	 gettimeofday(&tv, (void *)0);
	 return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}
#endif

// ============================================================
// TREEDEFS: include file with common definitions for TREECODE.
// ============================================================
 
// --------------------------------------------------------------
// Fundamental constants for treecode, not to be altered lightly.
// --------------------------------------------------------------
 
// HDLINE      identification string for the run.
// THETA       accuracy parameter.
// EPS         potential softening parameter.
// DTIME       N-body integration timestep.
// USQUAD      option to include quadrupole terms.
// NSTEPS      total number of timesteps in the simulation.
// NOUT        frequency of system record outputs.
 
char hdline[32];
double theta, eps, dtime;
int usquad;
int nsteps, nout;
 
// ----------------------------------
// System status and diagnostic data.
// ----------------------------------
				
// TNOW        value of current system time.
// MTOT        total mass of system.
// EKTOT       total kinetic energy.
// EPTOT       total potential energy.
// ETOT        sum of KE and PE.
// CMPOS       position of center of mass.
// CMVEL       velocity of center of mass.
// AMVEC       total angular momentum vector.
// CPUMKT      cpu time used in last tree construction.
// CPUACC      cpu time used in last force calculation.
// CPUT0       baseline for computing total cpu time.
// NBTOT       total number of body-body interactions.
// NCTOT       total number of body-cell interactions.
// NTMAX       maximum number of interactions for any body.
// NTAVG       average number of interactions per body.
 
double tnow, mtot, ektot, eptot, etot, cmpos[NDIM+1], cmvel[NDIM+1], amvec[NDIM+1], cpumkt, cpuacc;
double cput0;
int g_nbtot, g_nctot, g_ntmax; //global copies of nbtot, nctot and ntmax local variables in accel()
int ntavg;
 
// ---------------------------
// Tree structure data arrays.
// ---------------------------
 
// Bodies and cells are used to represent the tree.  The information for a
// given node is distributed across a number of arrays.  Quantities defined
// for both bodies and cells, such as mass and position, are stored in
// arrays with indicies running from 1 to MXNODE.  Quantities defined only
// for bodies are stored in arrays with indicies running from 1 to MXBODY,
// while those defined only for cells are stored in arrays with indicies
// running from INCELL to MXNODE.  With this convention, indicies can refer
// to either bodies or cells without conflict, and the type is deduced by
// comparing the index to the index of the first cell.  During tree
// construction, descendent indicies are stored in the SUBP arrays:
//
//          +-------------------------------------------------------+
// ROOT --> | CELL: MASS, POS, RCRIT2, QUAD, SUBP:[/,o,/,/,/,/,o,/] |
//          +----------------------------------------|---------|----+
//                                                   |         |
//     +---------------------------------------------+         |
//     |                                                       |
//     |    +--------------------------------+                 |
//     +--> | BODY: MASS, POS, VEL, ACC, PHI |                 |
//          +--------------------------------+                 |
//                                                             |
//     +-------------------------------------------------------+
//     |
//     |    +-------------------------------------------------------+
//     +--> | CELL: MASS, POS, RCRIT2, QUAD, SUBP:[o,/,/,o,/,/,o,/] |
//          +--------------------------------------|-----|-----|----+
//                                                etc   etc   etc
 
// MXBODY      maximum number of bodies allowed.
// MXCELL      maximum number of cells allowed.
// INCELL      index of first cell in arrays.
// MXNODE      maximum number of nodes (bodies + cells).

// NBODY       number of bodies in the system.
// NCELL       number of cells currently in use.
// MASS        mass of each node.
// POS         position of each node.
// VEL         velocity of each body.
// ACC         acceleration of each body
// PHI         potential of each body.
// RCRIT2      critical distances**2 of each cell.
// QUAD        quad moments of each cell.
// SUBP        descendent of each cell.
// TPOS        current position time.
// TVEL        current velocity time.
// RSIZE       side-length of root cell.
// ROOT        index of cell representing root.
 
int nbody, ncell;
double mass[MXNODE+1];
double vel[MXBODY+1][NDIM+1], acc[MXBODY+1][NDIM+1], phi[MXBODY+1];
//double rcrit2(incell:mxnode), quad(incell:mxnode,NQUAD)
double quad[MXNODE+1][NQUAD+1];
//int subp(incell:mxnode,NSUBC), root
int subp[MXNODE+1][NSUBC+1];
int root;
double tpos, tvel, rsize;
 
// These definitions apply only during tree construction.

// MID         geometric center of each cell [equiv to POS].
// CLSIZE      size of each cell [equiv to RCRIT2].

double pos[MXNODE+1][NDIM+1];
//double **pos;
//double mid[MXNODE][NDIM];
	//EQUIVALENCE (MID(1,1), POS(1,1))
#define midm(i,j) (mid[(i*(NDIM+1))+j])
double *mid;
//clsize(INCELL:MXNODE)
//double rcrit2[MXNODE-INCELL];
double *rcrit2;
//double clsize[MXNODE-INCELL];
	//EQUIVALENCE (CLSIZE(INCELL), RCRIT2(INCELL))
double *clsize;


// --------------------------
// Input/output unit numbers.
// --------------------------
 
FILE *ulog, *ubodo, *ubodf;
FILE *upars, *ubodi;

#include <sys/time.h>

double cputimer()
{
	 struct timeval tv;
	 gettimeofday(&tv, (void *)0);
	 return (tv.tv_sec + 1.0e-6*tv.tv_usec);
}
// ==================================
// TREEIO: I/O routines for TREECODE.
// ==================================

// --------------------------------------------------------------------
// INPARM: read in following input parameters:
//
//        hdline    : identification string for the run.
//        nsteps    : number of timesteps.
//        nout      : output system state once every nout steps.
//        dtime     : the timestep.
//        theta     : error tolerance; 0.0 => exact (PP) calculation.
//        eps       : potential softening parameter.
//        usquad    : option to include (.TRUE.) quadrupole terms.
// --------------------------------------------------------------------

void inparm() {
 
				char parsfl[8] = "treepars";
	int aux;
	FILE *inputfile;
 
//       -----------------------------------------------------
//       Open parameter file, read parameters, close the file.
//       -----------------------------------------------------
	if (inputfile = fopen("treepars","r")) {
					// READ (UPARS, '(A)') HDLINE
					//if(aux=fscanf(inputfile,"%s",hdline) < 0){
					if(aux=fread(hdline, sizeof(char), 32, inputfile) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}
					// READ (UPARS, *) NSTEPS, NOUT, DTIME, THETA, EPS, USQUAD
					if(aux=fscanf(inputfile,"%d\n",&nsteps) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}
					if(aux=fscanf(inputfile,"%d\n",&nout) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}
					if(aux=fscanf(inputfile,"%lf\n",&dtime) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}
					if(aux=fscanf(inputfile,"%lf\n",&theta) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}
					if(aux=fscanf(inputfile,"%lf\n",&eps) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}
					if(aux=fscanf(inputfile,"%d\n",&usquad) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}

	} else { 
		fprintf(stderr,"Error reading from file \n");
		exit(-1);
	}

	fclose(inputfile);
}

// ---------------------------------------------------------
// GETBDS: read in particle data, assumed to be of the form:
//
//       nbody		: number of bodies.
//       NDIM		: number of space dimensions.
//       time		: initial value of time.
//       mass(1)	: mass of first particle.
//       .......	
//       mass(n)	: mass of last particle.
//       x(1) y(1) z(1)	: position of first particle.
//       ..............
//       x(n) y(n) z(n)	: position of last particle.
//       u(1) v(1) w(1)	: velocity of first particle.
//       ..............
//       u(n) v(n) w(n)	: velocity of last particle.
// ---------------------------------------------------------

void getbds() {

				int indim,p,k;
	FILE *inputfile;

//       --------------------
//       Open body data file.
//       --------------------

				if (inputfile = fopen("treebodi","r")) {

//       ------------------
//       Read in body data.
//       ------------------

					if(fscanf(inputfile,"%d",&nbody) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}
					if(fscanf(inputfile,"%d",&indim) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}

//       ----------------------------------------------------------
//       Terminate the simulation if NBODY or INDIM not consistent.
//       ----------------------------------------------------------
		if (nbody <= 0 || nbody > MXBODY || indim != NDIM) {
			terror(" GETBDS: INCONSISTENT OR ILLEGAL INPUTS");
		}
//       --------------------------------------------
//       Read time, masses, positions and velocities.
//       --------------------------------------------

					if(fscanf(inputfile,"%lf",&tnow) < 0){
			fprintf(stderr,"Error reading from file\n");
						exit(-1);
		}

		for (p=1; p<=nbody; p++) {
						if(fscanf(inputfile,"%lf",&mass[p]) < 0){
				fprintf(stderr,"Error reading from file\n");
							exit(-1);
			}
		}

		for (p=1; p<=nbody; p++) {
			for (k=1; k<=NDIM; k++) {
							if(fscanf(inputfile,"%lf",&pos[p][k]) < 0) {
					fprintf(stderr,"Error reading from file\n");
								exit(-1);
				}
			}
		}

		for (p=1; p<=nbody; p++) {
			for (k=1; k<=NDIM; k++) {
							if(fscanf(inputfile,"%lf",&vel[p][k]) < 0){
					fprintf(stderr,"Error reading from file\n");
								exit(-1);
				}
			}
		}

	} else { 
		fprintf(stderr,"Error reading from file treebodi.\n");
		exit(-1);
	}

//      --------------------
//      Close up input file.
//      --------------------
	fclose(inputfile);
}

// ----------------------------------------------------
// BEGOUT: initialize disk files for subsequent output.
// ----------------------------------------------------

void begout() {

//       --------------------------------
//       Open log file and output header.
//       --------------------------------
	if (ulog = fopen("treelogs","w")) {
		outhdr();
//       ----------------------------
//       Open body data output files.
//       ----------------------------

		if (ubodo = fopen("treebodo","w")) {
			if (ubodf = fopen("treebodf","w")) {
			} else { 
				fprintf(stderr,"Error opening from file treebodf.\n");
				exit(-1);
			}
		} else { 
			fprintf(stderr,"Error opening from file treebodo.\n");
			exit(-1);
		}
	} else { 
		fprintf(stderr,"Error opening from file treelogs.\n");
		exit(-1);
	}

}
 
// ----------------------------------------------------------
// OUTHDR: write header to log file listing input parameters.
// ----------------------------------------------------------

void outhdr() {
 
	char aux;
	if (usquad == true)
		aux = 'T';
	else
		aux = 'F';
	fprintf(ulog, "\n\n");
	fprintf(ulog, " ************************************************************************\n");
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *                   %s                   *\n",hdline);
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " ************************************************************************\n");
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *    NBODY =%7d    NSTEPS =%4d    NOUT =%4d    DTIME =%7.3f    *\n", nbody, nsteps, nout, dtime);
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *    USEQUAD =%2c       EPS =%6.3f     THETA =%5.2f                    *\n", aux, eps, theta);
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " *    OPTIONS:  IMPROVED ACCEPTANCE CRITERION                           *\n");
	fprintf(ulog, " *                                                                      *\n");
	fprintf(ulog, " ************************************************************************\n");

}

// ---------------------------------------------------
// OUTTIM: write force calculation timing to log file.
// ---------------------------------------------------
 
void outtim() {

	fprintf(ulog, "\n      CPU TIME USED:          MKTREE           ACCEL           TOTAL\n");
	fprintf(ulog, "        (SECONDS)   %16.2f%16.2f%16.2f\n", cpumkt, cpuacc, cpumkt+cpuacc);

}

// --------------------------------------------------------
// OUTLOG: write force calculation diagnostics to log file.
// --------------------------------------------------------

void outlog() {
 
				double cput;
 
//       -----------------------------
//       Get system-specific cpu time.
//       -----------------------------
				cput = cputimer();
	cput = cput - cput0;
//       -------------------------------------------------
//       OUTPUT CPU TIME AND FORCE EVALUATION DIAGNOSTICS.
//       -------------------------------------------------
	fprintf(ulog, "\n\n\n      TIME    NCELL     NBTOT     NCTOT   NTMAX   NTAVG  CPUTIME\n");
	fprintf(ulog, " %9.3f%9d%10d%10d%8d%8d%9.2f\n",tnow, ncell, g_nbtot, g_nctot, g_ntmax, ntavg, cput/60.0);

}

// -------------------------------------------------
// OUTDAG: write N-body diagnostic data to log file.
// -------------------------------------------------
 
void outdag() {

				//INTEGER K, IPRINT
				//DATA IPRINT /0/
	static int iprint = 0;

	printf(" %4d %12.4E%12.4E%12.4E%12.4E%12.4E%12.4E\n",iprint, cmpos[1], cmpos[2], cmpos[3], cmvel[1], cmvel[2], cmvel[3]);
	iprint++;
//      ----------------------------------------------------------------
//      Write mass, energy, angular momentum, center of mass quantities.
//      ----------------------------------------------------------------
	fprintf(ulog, "\n\n             ETOT            EKIN            EPOT            MASS\n");
	fprintf(ulog, " %16.8f%16.8f%16.8f%16.8f\n",etot,ektot,eptot,mtot);
	fprintf(ulog, "\n %16s%16s%16s%16s\n"," ","X","Y","Z");
	fprintf(ulog, "             JTOT%16.6E%16.6E%16.6E\n",amvec[1], amvec[2], amvec[3]);
	fprintf(ulog, "            CMPOS%16.6E%16.6E%16.6E\n",cmpos[1], cmpos[2], cmpos[3]);
	fprintf(ulog, "            CMVEL%16.6E%16.6E%16.6E\n",cmvel[1], cmvel[2], cmvel[3]);
}

// -----------------------------------------------------------
// OUTERR: output error messages to the log file and terminal.
// -----------------------------------------------------------

void outerr(char *msg) {

//       ------------------------------------------------------
//       Write the message, surrounded by stars for visibility.
//       ------------------------------------------------------
	fprintf(ulog, "\n ************************************************************************\n");
	fprintf(ulog, "%s",msg);
	fprintf(ulog, "\n ************************************************************************");
//       ---------------------------
//       Repeat message to terminal.
//	---------------------------
	fprintf(stderr, "\n ************************************************************************\n");
	fprintf(stderr, "%s",msg);
	fprintf(stderr, "\n ************************************************************************");
}

// -----------------------------------------------
// OUTCPU: output cpu timing data to the log file.
// -----------------------------------------------

void outcpu() {

	double cput;

//       -----------------------------
//       Get system-specific cpu time.
//       -----------------------------
	cput = cputimer();
	cput = cput - cput0;
	fprintf(ulog, "\n\n ELAPSED CPU TIME (SECONDS) : %16.2f",cput);
}

// ------------------------------------------------------
// PUTBDS: write body data to output file or final file.
// ------------------------------------------------------

void putbds(int final) {
 
//       ------------------------------------
//       Decide which file to use for output.
//       ------------------------------------
				if (final) {
//         -------------------------------------------
//         Output final coordinates, including masses.
//         -------------------------------------------

		fprintf(ubodf, "%20d\n",nbody);
		fprintf(ubodf, "%20d\n",NDIM);
		fprintf(ubodf, "%20.12E\n",tnow);

		putmas(ubodf);
		putpos(ubodf);
		putvel(ubodf);
	} else {
//	  ---------------------------------
//	  Output particle coordinates only.
//	  ---------------------------------

		fprintf(ubodo, "%20d\n",nbody);
		fprintf(ubodo, "%20d\n",NDIM);
		fprintf(ubodo, "%20.12E\n",tnow);

		putpos(ubodo);
		putvel(ubodo);

	}
}

// -------------------------------
// PUTMAS: output body mass array.
// -------------------------------

void putmas(FILE *file) {

	int i;

	for (i=1; i<=nbody; i++) {
		fprintf(file, "%20.12E\n", mass[i]);
	}

}

// -----------------------------------
// PUTPOS: output body position array.
// -----------------------------------

void putpos(FILE *file) {

	int i;

	for (i=1; i<=nbody; i++) {
			fprintf(file, "%20.12E%20.12E%20.12E\n", pos[i][1], pos[i][2], pos[i][3]);
	}

}

// -----------------------------------
// PUTVEL: output body velocity array.
// -----------------------------------

void putvel(FILE *file) {

	int i;

	for (i=1; i<=nbody; i++) {
			fprintf(file, "%20.12E%20.12E%20.12E\n", vel[i][1], vel[i][2], vel[i][3]);
	}

}
 
// ----------------------------------------------
// STPOUT: terminate output and close open files.
// ----------------------------------------------

void stpout() {
 
	fprintf(ulog, "\n\n CALCULATION TERMINATED");
//       ---------------------
//       Close the open files.
//       ---------------------
	fclose(ubodo);
	fclose(ubodf);
	fclose(ulog);
}

// ================================================================
// TREELOAD: routines to handle tree construction and verification.
// ================================================================

#include <math.h>
 
// ----------------------------------------------------------------
// MKTREE: initialize the tree structure for the force calculation.
// ----------------------------------------------------------------
 
void mktree() {

				double cput1, cput2;
 
//       ----------------------------------
//       Start timing of tree construction.
//       ----------------------------------
				cput1 = cputimer();
//       --------------------------------------
//       Expand root volume to hold all bodies.
//       --------------------------------------
	expbox();
//	--------------------------
//       Load bodies into the tree.
//	--------------------------
	ldtree();
//      ------------------------------------------------
//      Compute masses, center of mass coordinates, etc.
//      ------------------------------------------------
				hackcm();
//      -------------------------------------
//      Compute timing for tree construction.
//      -------------------------------------
				cput2 = cputimer();
				cpumkt = cput2 - cput1;
}
 
// -----------------------------------------------------
// EXPBOX: enlarge system cube to include all particles.
// -----------------------------------------------------
 
void expbox() {
 
	double xyzmax;
	int i,j;
 
//      ------------------------------
//      Find maximum coordinate value.
//      ------------------------------
	xyzmax = 0.0;

	for (i=1; i<=nbody; i++) {
		for (j=1; j<=NDIM; j++) {
			if (xyzmax < fabs(pos[i][j])) {
				xyzmax = fabs(pos[i][j]);
			}
		}
	}
//	-----------------------------------------
//	Expand box by factors of 2 until it fits.
//	-----------------------------------------
label_30:
				if (xyzmax >= rsize/2.0) {
		rsize = 2.0 * rsize;
		goto label_30;
	}
}


// ---------------------------------------------------------------------
// LDTREE: construct tree body-by-body.  This phase initializes the SUBP
// array and loads the geometric midpoint into the MID of each cell.
// ---------------------------------------------------------------------

void ldtree() {

				int k, p;

//      ---------------------------------------
//      Deallocate current tree, begin new one.
//      ---------------------------------------
				ncell = 0;
				root = mkcell();
//	------------------------------------------
//	Initialize midpoint and size of root cell.
//	------------------------------------------
	for (k=1; k<=NDIM; k++) {
		midm(root,k) = 0.0;
	}
	clsize[root] = rsize;
//       ---------------------------------------------
//       Load bodies into the new tree, one at a time.
//       ---------------------------------------------
	for (p=1; p<=nbody; p++) {
					ldbody(p);
	}
}
 
// ----------------------------------------
// LDBODY: load body P into tree structure.
// ----------------------------------------
 
void ldbody(int p) {

				int q, qind, c, k, p0;

//      ---------------------------------------------
//      Start Q,QIND pair in correct subcell of root.
//      ---------------------------------------------
				q = root;
				qind = sbindx(p,q);
//      -----------------------------------------------------
//      Loop descending tree until an empty subcell is found.
//      -----------------------------------------------------
label_20:
	if (subp[q][qind] != NULO) {
//        --------------------------------------
//        On reaching another body, extend tree.
//        --------------------------------------
					if (subp[q][qind] < INCELL) {
//          -------------------------------------------
//          Allocate an empty cell to hold both bodies.
//          -------------------------------------------
						c = mkcell();
//          ------------------------------------------------------
//          Locate midpoint of new cell wrt. parent, and set size.
//          ------------------------------------------------------
						for (k=1; k<=NDIM; k++) {
				if (pos[p][k] >= midm(q,k)) {
								midm(c,k) = midm(q,k) + clsize[q]/4.0;
				} else {
								midm(c,k) = midm(q,k) - clsize[q]/4.0;
				}
			}
			clsize[c] = clsize[q] / 2.0;
//          ------------------------------------------------------
//          Store old body in appropriate subcell within new cell.
//          ------------------------------------------------------
						p0 = subp[q][qind];
						subp[c][sbindx(p0,c)] = p0;
//          ---------------------------------------------
//          Link new cell into tree in place of old body.
//          ---------------------------------------------
						subp[q][qind] = c;
		}
//        --------------------------------------------------------
//        At this point, the node indexed by Q,QIND is known to be
//	  a cell, so advance to the next level of tree, and loop.
//        --------------------------------------------------------
					q = subp[q][qind];
					qind = sbindx(p,q);
					goto label_20;
	}
//      ---------------------------------------------
//      Found place in tree for P, so store it there.
//      ---------------------------------------------
				subp[q][qind] = p;
}
 
// -------------------------------------------------------
// SBINDX: compute subcell index for node P within cell Q.
// -------------------------------------------------------
 
int sbindx(int p, int q) {

	int k;
	int ret_aux;
 
//       ---------------------------------------------------
//       Initialize subindex to point to lower left subcell.
//       ---------------------------------------------------
				ret_aux = 1;
//       ---------------------------------
//       Loop over all spatial dimensions.
//       ---------------------------------
				for (k=1; k<=NDIM; k++) {
		if (pos[p][k] >= midm(q,k))
			ret_aux += pow(2,(3-k));
	}
	return ret_aux;
}
 
// ---------------------------------------------------------
// MKCELL: function to allocate a cell, returning its index.
// ---------------------------------------------------------
 
int mkcell() {

	int i;
	int ret_aux;
 
//       ----------------------------------------------------------
//       Terminate simulation if no remaining space for a new cell.
//       ----------------------------------------------------------
				if (ncell >= MXCELL) {
		terror(" MKCELL: NO MORE MEMORY");
	}
//       ----------------------------------------------------
//       Increment cell counter, initialize new cell pointer.
//       ----------------------------------------------------
				ncell = ncell + 1;
				ret_aux = ncell + MXBODY;
//       --------------------------------------
//       Zero pointers to subcells of new cell.
//       --------------------------------------
				for (i=1; i<=NSUBC; i++) {
					subp[ret_aux][i] = NULO;
	}
	return ret_aux;
}
 
// ----------------------------------------------------------------
// HACKCM: compute cell masses, c.m. positions, check tree structure,
// assign critical radii, and compute quadrupole moments.  Whew//
// ----------------------------------------------------------------
 
void hackcm() {

	int ind[MXCELL], p, q, i, j, k, l, m, n;
	double pos0[NDIM], dist2;
	int min_aux;
 
//       ---------------------------------------
//       List cells in order of decreasing size.
//       ---------------------------------------
				 bflist(ind);
//       --------------------------------------------
//       Loop processing cells from smallest to root.
//       --------------------------------------------
				for (i=ncell; i>0; i--) {
					p = ind[i];
//         --------------------------------------------------------------
//         Zero accumulators for this cell.  A temporary variable is used
//	  for the c. of m. so as to preserve the stored midpoints.
//         --------------------------------------------------------------
					mass[p] = 0.0;
		for (k=1; k<=NDIM; k++) {
						pos0[k] = 0.0;
		}
//         -------------------------------------------------------------
//         Compute cell properties as sum of properties of its subcells.
//         -------------------------------------------------------------
					for (j=1; j<=NSUBC; j++) {
						q = subp[p][j];
//           ------------------------------
//           Only access cells which exist.
//           ------------------------------
						if (q != NULO) {
//             -------------------------------------------------------
//             Sum properties of subcells to obtain values for cell P.
//             -------------------------------------------------------
							mass[p] += mass[q];
				for (k=1; k<=NDIM; k++) {
								pos0[k] += mass[q] * pos[q][k];
				}
			}
		}
//         --------------------------------------------------------
//         Normalize center of mass coordinates by total cell mass.
//         --------------------------------------------------------
					for (k=1; k<=NDIM; k++) {
						pos0[k] = pos0[k] / mass[p];
		}
//         -----------------------------------------------------------------
//         Check tree, compute cm-to-mid distance, and assign cell position.
//         -----------------------------------------------------------------
					dist2 = 0.0;
		for (k=1; k<=NDIM; k++) {
						if (pos0[k] < midm(p,k) - clsize[p]/2.0 || 
					pos0[k] >= midm(p,k) + clsize[p]/2.0) {
				fprintf(ulog, "\n TREE ERROR %6d %6d %14.6lf %14.6lf %14.6lf\n",p,k,pos0[k],midm(p,k),clsize[p]);
							terror(" HACKCM: TREE STRUCTURE ERROR");
			}
						dist2 += (pos0[k] - midm(p,k))*(pos0[k] - midm(p,k));
//	    --------------------------------------------------------
//	    Copy cm position to cell.  This overwrites the midpoint.
//	    --------------------------------------------------------
			pos[p][k] = pos0[k];
		}

//	  ------------------------------------------------------------
//	  Assign critical radius for cell, adding offset from midpoint
//	  for more accurate forces.  This overwrites the cell size.
//	  ------------------------------------------------------------
					rcrit2[p] = (clsize[p] / theta + sqrt(dist2))*(clsize[p] / theta + sqrt(dist2));

	}
//	----------------------------------------
//	Compute quadrupole moments, if required.
//	----------------------------------------
	if (usquad) {
//        -------------------------------
//        Loop processing cells as above.
//        -------------------------------
					for (i=ncell; i>0; i--) {
						p = ind[i];
//          --------------------------------------------
//          Zero accumulator for quad moments of cell P.
//          --------------------------------------------
						for (k=1; k<=NQUAD; k++) {
							quad[p][k] = 0.0;
			}
//          --------------------------------
//          Loop over descendents of cell P.
//          --------------------------------
						for (j=1; j<=NSUBC; j++) {
				q = subp[p][j];
				if (q != NULO) {
//              --------------------------------------------------------
//              Sum properties of subcell Q to obtain values for cell P.
//              --------------------------------------------------------
								if (2 < NDIM)
			min_aux = 2; 
		else
			min_aux = NDIM;
								for (m=1; m<=min_aux; m++) {
									for (n=m; n<=NDIM; n++) {
										l = (m-1) * (NDIM-1) + n;
										quad[p][l] += 3.0 * mass[q] * (pos[q][m] - pos[p][m]) * (pos[q][n] - pos[p][n]);
				if (m == n) {
											for (k=1; k<=NDIM; k++) {
												quad[p][l] = quad[p][l] - mass[q] * ((pos[q][k] - pos[p][k])*(pos[q][k] - pos[p][k]));
					}
				}
//                  -------------------------------------------
//                  If Q itself is a cell, add its moments too.
//                  -------------------------------------------
										if (q >= INCELL)
											quad[p][l] = quad[p][l] + quad[q][l];
			} // for (n=m; n<=NDIM; n++)
		} // for (m=1; m<=min_aux; m++)
				} // if (q != NULO)
			} // for (j=1; j<=NSUBC; j++)
		} // for (i=ncell; i>0; i--)
	} // if (usquad)
}

// -----------------------------------------------------------------
// BFLIST: list cells in breadth-first order, from largest (root) to
// smallest.  Thanks to Jun Makino for this elegant routine.
// -----------------------------------------------------------------
 
void bflist(int *ind) {
 
	int facell, lacell, nacell, k, i;
 
//	-----------------------------------------
//	Start scan with root as only active cell.
//	-----------------------------------------
	ind[1] = root;
	facell = 1;
	lacell = 1;
//	-----------------------------------
//	Loop while active cells to process.
//	-----------------------------------
label_10:
	if (facell <= lacell) {
//	  ----------------------------------------------
//	  Start counting active cells in next iteration.
//	  ---------------------------------------------- 
		nacell = lacell;
//	  ---------------------------------------
//	  Loop over subcells of each active cell.
//	  ---------------------------------------
		for (k=1; k<=NSUBC; k++) {
			for (i=facell; i<=lacell; i++) {
//	      -------------------------------------------
//	      Add all cells on next level to active list.
//	      -------------------------------------------
				if (subp[ind[i]][k] >= INCELL) {
		nacell = nacell + 1;
		ind[nacell] = subp[ind[i]][k];
				}
			}
		}
//	  ------------------------------------------------------
//	  Advance first and last active cell indicies, and loop.
//	  ------------------------------------------------------
		facell = lacell + 1;
		lacell = nacell;
		goto label_10;
	}
//	--------------------------------------------------
//	Above loop should list all cells; check the count.
//	--------------------------------------------------
	if (nacell != ncell)
		terror("  BFLIST: INCONSISTENT CELL COUNT");
}
// =====================================================================
// TREECODE: Fortran77 version of hierarchical N-body code (O(N log N)).

//     This program evolves a self-gravitating N-body system, using
//     a hierarchical O(N log N) algorithm to compute gravitational
//     forces (see Barnes, J.E. & Hut, P. 1986, Nature, 324, 446).
//     The computational system of units is determined by the input
//     data, with the assumption that G = 1.  Particles are not
//     required to have identical masses.

//     This code is based on a Fortran implementation of the
//     hierarchical algorithm developed and kindly provided by
//     Lars Hernquist.  Several additional refinments by Jun Makino
//     have been incorporated as well.

//     The present version of the code includes quadrupole moments
//     when calculating the gravitational fields of cells.  It also
//     incorporates an improved opening criterion which increases
//     the accuracy of the forces in certain exceptional situations.

//					Joshua Barnes
//					April 23, 1993
// =====================================================================


// ==============================================================
// TREECD: top-level N-body evolution program.  Its tasks are to:
//  *  read input parameters and initial conditions;
//  *  perform an initial force calculation, reporting
//     system diagnostics and time required;
//  *  advance the system for NSTEPS time-steps, periodically
//     saving a snapshot of the system;
//  *  terminate the simulation and close data files.
// ==============================================================

//      PROGRAM TREECD

//***********************************************************************
//       speccode: Directives for profiling
//$PRAGMA C(collector_resume,collector_pause)
//***********************************************************************


//***********************************************************************
//       speccode: Stop profiling
//       call collector_pause()
//***********************************************************************
double spec_time;
#ifdef MEASURE_COUNTERS
int number_calls;
int max_iterations;
#endif

int main() {

	int i;
				int ndiag, n;
	//EQUIVALENCE (MID(1,1), POS(1,1))
	mid = (double *) &(pos);
	//mid = pos;
	rcrit2 = (double *) malloc ((MXNODE)*sizeof(double));
	//EQUIVALENCE (CLSIZE(INCELL), RCRIT2(INCELL))
	clsize = rcrit2;
//       ---------------------
//       Initialize cpu timer.
//       ---------------------
	cput0 = cputimer();
//	----------------------------------------------------------------
//	Open data files, read input parameters and initial system state,
//       initialize system parameters, and start output.
//	----------------------------------------------------------------
				inparm();
	getbds();
	initcd();
	begout();

//       -----------------------------------------------
//	Generate tree and compute initial acceleration.
//	-----------------------------------------------
				mktree();
	accel();
//	------------------------------------------
//	Output timing data and system diagnostics.
//	------------------------------------------
	outtim();
	outlog();
	diags();
//      -----------------------------------------------
//      Set number of steps between diagnostic outputs.
//      -----------------------------------------------
	if (nsteps/128 > 1)
		ndiag = nsteps / 128;
	else
		ndiag = 1;
//	=================================================================
//	Primary loop to advance system state for a given number of steps.
//	=================================================================
	for (n=1; n<=nsteps; n++) { 
//	  ------------------------------------------------------
//	  Advance velocity by 1/2 step, positions by 1 timestep.
//	  ------------------------------------------------------
		advvel(dtime/2.0);
		advpos(dtime);
//	  ----------------------------
//	  Create tree, compute forces.
//	  ----------------------------
		mktree();
		accel();
//	  ----------------------------------------------------
//	  Advance velocity 1/2 step to synchronize tvel, tpos.
//	  ----------------------------------------------------
					advvel(dtime/2.0);
//	  ----------------------------------------------------
//	  At appropriate intervals, output system diagnostics.
//	  ----------------------------------------------------
					if (n%ndiag == 0) {
			outlog();
			diags();
		}
//        ----------------------------------------------
//	  At appropriate intervals, output system state.
//        ----------------------------------------------
					if (n%nout == 0)
			putbds(false);
	}
//	----------------------------------
//	Output final snapshot with masses.
//	----------------------------------
				putbds(true);
//	------------------------------------------------------------------
//	Stop timing, write timing data, close files, terminate simulation.
//	------------------------------------------------------------------
	outcpu();
	stpout();

	// -----------------------------------------------------------
	// speccode: print time results
	fprintf (stderr, "--------------------------------------------------\n");
	fprintf (stderr, "speccode: Time elapsed: %lf seconds.\n", spec_time);
#ifdef MEASURE_COUNTERS
	fprintf (stderr, "Number_calls=%d, max_iterations = %d\n", number_calls, max_iterations);
#endif
	// -----------------------------------------------------------
	
	return 0;
}

// ----------------------------------------------------------
// INITCD: initialize system parameters that depend on either
// the input data or defined PARAMETERS.
// ----------------------------------------------------------

void initcd() {
 
//	---------------------------------------
//	Initialize position and velocity times.
//	---------------------------------------
				tpos = tnow;
	tvel = tnow;
//	---------------------------------
//	Initialize box coordinate system.
//	---------------------------------
	rsize = 4.0;
}

// ---------------------------------------------------------------
// ADVPOS: advance the positions of the bodies for a timestep DT.
// ---------------------------------------------------------------
 
void advpos(double dt) {

	int k,p;

//	------------------------------------------------- 
//	Loop over all spatial coordinates for all bodies.
//	------------------------------------------------- 
				for (k = 1; k<=NDIM; k++) {
		for (p=1; p<=nbody; p++) {
			pos[p][k] = pos[p][k] + dt*vel[p][k];
		}
	}
//	----------------------------------
//	Update position time, system time.
//	----------------------------------
				tpos = tpos+dt;
				tnow = tpos;
}

// ----------------------------------------------------------------
// ADVVEL: advance the velocities of the bodies for a timestep dt.
// ----------------------------------------------------------------
 
void advvel(double dt) {
				
	int k, p;

//	------------------------------------------------- 
//	Loop over all spatial coordinates for all bodies.
//	------------------------------------------------- 
				for (k = 1; k<=NDIM; k++) {
		for (p=1; p<=nbody; p++) {
			vel[p][k] = vel[p][k] + dt*acc[p][k];
		}
	}
//	----------------------------------
//	Update velocity time, system time.
//	----------------------------------
				tvel = tvel+dt;
				tnow = tvel;
}

// ----------------------------------------------------------------
// DIAGS: compute diagnostics for the system: total energy, total
// kinetic energy, total potential energy, angular momentum, center
// of mass coordinates, and center of mass velocity.
// ----------------------------------------------------------------
 
void diags() {
 
	int k,p;

//      --------------------------------------------------------------
//      Check that tpos and tvel are synchronized to one part in 1000.
//      --------------------------------------------------------------
				if (fabs(tpos-tvel) > dtime/1.0E3)
		terror(" DIAGS: TIMES NOT SYNCHRONIZED");
//      ---------------------------------------------
//      Zero the accumulators for system diagnostics.
//      ---------------------------------------------
				mtot = 0.0;
				ektot = 0.0;
				eptot = 0.0;
	for (k=1; k<=NDIM; k++) {
					cmpos[k] = 0.0;
					cmvel[k] = 0.0;
					amvec[k] = 0.0;
	}
//      ----------------------------------------------------------
//      Loop over bodies to compute net mass and potential energy.
//      ----------------------------------------------------------
	for (p=1; p<=nbody; p++) {
					mtot = mtot + mass[p];
					eptot = eptot + 0.5 * mass[p] * phi[p];
	}
//      -----------------------------------------------------------------
//      Compute net kinetic energy, center of mass position and velocity.
//      -----------------------------------------------------------------
	for (k=1; k<=NDIM; k++) {
		for (p=1; p<=nbody; p++) {
						ektot = ektot + 0.5 * mass[p] * (vel[p][k]*vel[p][k]);
						cmpos[k] = cmpos[k] + mass[p] * pos[p][k];
						cmvel[k] = cmvel[k] + mass[p] * vel[p][k];
		}
					cmvel[k] = cmvel[k] / mtot;
					cmpos[k] = cmpos[k] / mtot;
	}
//      ----------------------------
//      Compute total system energy.
//      ----------------------------
				etot = ektot + eptot;
//      ---------------------------------------
//      Compute angular momentum of the system.
//      ---------------------------------------
				for (p=1; p<=nbody; p++) {
					amvec[1] += mass[p] * (pos[p][2]*vel[p][3] - pos[p][3]*vel[p][2]);
					amvec[2] += mass[p] * (pos[p][3]*vel[p][1] - pos[p][1]*vel[p][3]);
					amvec[3] += mass[p] * (pos[p][1]*vel[p][2] - pos[p][2]*vel[p][1]);
	}
//       ----------------------------------
//       Write diagnostics to the log file.
//       ----------------------------------
				outdag();
}

// -------------------------------------------------------------
// TERROR: terminate the program as the result of a fatal error,
// closing the output files and writing timing information.
// -------------------------------------------------------------

void terror(char *msg) {

//       ------------------------------------
//       Write error message to the log file.
//       ------------------------------------
	outerr(msg);
//       ---------------------------------------------------
//       Output timing data, close files, terminate the run.
//       ---------------------------------------------------
				outcpu();
	stpout();
	exit(-1);
}

// =============================================================
// TREEGRAV: routines to perform hierarchical force calculation.
// =============================================================



// --------------------------------------------------------
// ACCEL: routine to compute the acceleration and potential
// for all bodies in the system.
// --------------------------------------------------------

void accel() {

//***********************************************************************
//       speccode: Interface for profiling API directives
//       //$PRAGMA C(collector_pause,collector_resume)
//***********************************************************************

	double cput1, cput2;
		
	int p;
	//int mxsptr = 256; //Defined as constant
	double eps2, phi0, acc0[NDIM+1], pos0[NDIM+1];
	double dx, dy, dz, dr2, dr2inv, drinv, phim, dr5inv, phiq;
	int q, nbterm, ncterm, sptr, stack[MXSPTR+1], k;
	int skpslf;

//      ----------------------------------------
//      Initialize timing for force calculation.
//      ----------------------------------------
				cput1 = cputimer();

//      --------------------------------------------
//      Initialize the force evaluation diagnostics.
//      --------------------------------------------
	int nbtot = 0;
	int nctot = 0;
	int ntmax = 0;

//***********************************************************************
//      speccode: start profiling
//	call collector_resume()
//***********************************************************************

	//imprimir();
	//exit(-1);
//	---------------------------------------------------
//	Loop over bodies, computing force for each in turn.
//	---------------------------------------------------
#ifdef MEASURE_COUNTERS
number_calls++;
if (nbody > max_iterations) {
	max_iterations = nbody;
	fprintf(stderr,"new max = %d\n",max_iterations);
}
#endif
			 // Time measurement variables
			 double spec_start, spec_stop;

			 // speccode: Start profiling
			 spec_start = get_time();
			 //***********************************************************************

	int iter = nbody;
#ifndef SEQUENTIAL
	specbegin(iter);
#endif
	
	#pragma omp parallel for default(none)  schedule(static)             \
	 private(p, skpslf, eps2, phi0, acc0, pos0, dx, dy, dz, dr2, dr2inv,  \
		drinv, phim, dr5inv, q, nbterm, ncterm, sptr, k)   \
	shared(iter, eps, pos, root, mass, rcrit2, usquad, phiq,  \
		quad, subp, phi, acc) \
	speculative(stack, nbtot, nctot, ntmax) 
	//nbtot, nctot, ntmax are speculative reductions (addition and max)
	for (p=1; p<=iter; p++) {
		eps2 = eps*eps;
		
		//----------------------------------------------------------
		// Zero potential and acceleration for subsequent summations.
		//----------------------------------------------------------
		phi0 = 0.0;
		acc0[1] = 0.0;
		acc0[2] = 0.0;
		acc0[3] = 0.0;
		
		//-----------------------------------------
		// Copy position of P for quicker reference.
		//-----------------------------------------
		pos0[1] = pos[p][1];
		pos0[2] = pos[p][2];
		pos0[3] = pos[p][3];

		//----------------------------------------------
		// Initialize counters for number of force terms.
		//----------------------------------------------
		nbterm = 0;
		ncterm = 0;
		skpslf = false;

		//----------------------------------
		//	Push the root cell onto the stack.
		//----------------------------------
		sptr = 1;
		stack[sptr] = root;

		//-------------------------------------
		//  Loop while nodes on stack to process.
		//-------------------------------------
		while (sptr > 0) {
			//--------------------------
			// Pop node off top of stack.
			//--------------------------
			q = stack[sptr];
			sptr = sptr - 1;
			
			//---------------------------------------------
			// Compute distance to center-of-mass of node Q.
			// ---------------------------------------------
			dx = pos0[1] - pos[q][1];
			dy = pos0[2] - pos[q][2];
			dz = pos0[3] - pos[q][3];
			dr2 = dx*dx + dy*dy + dz*dz;

			//-------------------------------
			// Classify Q as a body or a cell.
			//-------------------------------
			if (q < INCELL) {
				//-----------------------------------
				// A body: check for self-interaction.
				//-----------------------------------
				if (q != p) {
					//------------------------------
					// Compute body-body interaction.
					//------------------------------
					dr2inv = 1.0 / (dr2 + eps2);
					phim = mass[q] * sqrt(dr2inv);
					phi0 = phi0 - phim;
					phim = phim * dr2inv;
					acc0[1] -= phim * dx;
					acc0[2] -= phim * dy;
					acc0[3] -= phim * dz;
					nbterm = nbterm + 1;
				} else {
					//-------------------------------------------
					// Remember that self-interaction was skipped.
					//-------------------------------------------
					skpslf = true;
				}
			} else {
			//--------------------------------------------
			// A cell: test if interaction can be accepted.
			//--------------------------------------------
			
			if (dr2 >= rcrit2[q]) {
				//----------------------------------------
				// Accepted: compute body-cell interaction.
				//----------------------------------------
				dr2inv = 1.0 / (dr2 + eps2);
				drinv = sqrt(dr2inv);
				phim = mass[q] * drinv;
				phi0 = phi0 - phim;
				phim = phim * dr2inv;
				acc0[1] -= phim * dx;
				acc0[2] -= phim * dy;
				acc0[3] -= phim * dz;
				
				//------------------------------------
				// Optionally include quadrupole terms.
				//------------------------------------
				if (usquad) {
					dr5inv = dr2inv * dr2inv * drinv;
					phiq = dr5inv * (0.5 * ((dx*dx - dz*dz) * quad[q][1] + 
								 (dy*dy - dz*dz) * quad[q][4]) + 
								 dx*dy * quad[q][2] + dx*dz * quad[q][3] + 
								 dy*dz * quad[q][5]);
					phi0 = phi0 - phiq;
					phiq = 5.0 * phiq * dr2inv;
					acc0[1] -= phiq*dx + dr5inv * (dx*quad[q][1] + dy*quad[q][2] + dz*quad[q][3]);
					acc0[2] -= phiq*dy + dr5inv * (dx*quad[q][2] + dy*quad[q][4] + dz*quad[q][5]);
					acc0[3] -= phiq*dz + dr5inv * (dx*quad[q][3] + dy*quad[q][5] - dz*(quad[q][1] + quad[q][4]));
				}
				
				ncterm = ncterm + 1;
				
				} else {
					//-----------------------------------
					// Rejected: examine children of cell.
					//-----------------------------------
					for (k=1; k<=NSUBC; k++) {
						
						//--------------------------------------
						// Push existing children onto the stack.
						//--------------------------------------
						if (subp[q][k] != NULO) {
							if (sptr >= MXSPTR) {
								terror(" TRWALK: STACK OVERFLOW");
							}
							
							sptr = sptr + 1;
							stack[sptr] = subp[q][k];
						}
					}
				}
			}
		
		}

	
		//---------------------------------------------------
		// Check that self-interaction was explicitly skipped.
		//---------------------------------------------------
		if (!skpslf) {
			terror(" TRWALK: MISSED SELF-INTERACTION");
		}

		//----------------------------------------------
		// Copy total potential and acceleration to body.
		//----------------------------------------------
		phi[p] = phi0;
		acc[p][1] = acc0[1];
		acc[p][2] = acc0[2];
		acc[p][3] = acc0[3];

		//-------------------------------------
		// Update force calculation diagnostics.
		//-------------------------------------
		nbtot = nbtot + nbterm;
		nctot = nctot + ncterm;
		int aux = ncterm+nbterm;
		//if (ntmax < ncterm+nbterm)
		if (ntmax < aux) {
			//ntmax = ncterm+nbterm;
			ntmax = aux;
		}
	} // for (p=1; p<=nbody; p++)


			/************************************************************************/
			// speccode: Stop measuring time
			spec_stop = get_time();
			spec_time += spec_stop - spec_start;
			/************************************************************************/
#ifdef TIME_SPECLOAD
 printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
 printSpecstoreTime();
#endif
#ifdef MEASURE_COUNTERS
 specepilogue();
 printAccesses();
#endif
//***********************************************************************
//       speccode: stop profiling
//       call collector_pause()
//***********************************************************************

//
//  Save global values
//
			g_ntmax = ntmax;
			g_nbtot = nbtot;
			g_nctot = nctot;

//     -----------------------------------------------
//     Compute average number of force terms per body.
//     -----------------------------------------------
			 ntavg = (nbtot + nctot) / nbody;
//     -------------------------------------
//     Compute timing for force calculation.
//     -------------------------------------
			 cput2 = cputimer();
			 cpuacc = cput2 - cput1;
}
