/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* hmmcalibrate.c
 * SRE, Fri Oct 31 09:25:21 1997 [St. Louis]
 *
 * Score an HMM against random sequence data sets;
 * set histogram fitting parameters.
 *
 * CVS $Id: hmmcalibrate.c,v 1.27 2003/05/26 23:56:33 eddy Exp $
 */
/* sergio: definiciones para el compilador */
double spec_time = 0;
#include "compiler.h"

#include "config.h"		/* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <float.h>
#ifdef HMMER_THREADS
#include <pthread.h>
#endif
#ifdef HMMER_PVM
#include <pvm3.h>
#endif

#include "squid.h"		/* general sequence analysis library    */
#include "stopwatch.h"		/* process timings                      */
#include "structs.h"		/* data structures, macros, #define's   */
#include "funcs.h"		/* function declarations                */

#ifndef SPEC_CPU
#include "globals.h"		/* alphabet global variables            */
#endif

//***********************************************************************
// speccode: OpenMP header file included
#include <omp.h>
//***********************************************************************
//***********************************************************************
// speccode: Main header file with all common variables
#include "specEngine.h"
//***********************************************************************

/* sergio: Redefinicion de banner a bannercalibrate */
static char bannercalibrate[] =
    "hmmcalibrate -- calibrate HMM search statistics";

/* sergio: Redefinicion de usage a usagecalibrate */
static char usagecalibrate[] = "\
Usage: hmmcalibrate [-options] <hmmfile>\n\
Available options are:\n\
  -h             : print short usage and version info, then exit\n\
";

/* sergio: Redefinicion de experts a expertscalibrate */
static char expertscalibrate[] = "\
  --cpu <n>      : run <n> threads in parallel (if threaded)\n\
  --fixed <n>    : fix random sequence length at <n>\n\
  --histfile <f> : save histogram(s) to file <f>\n\
  --mean <x>     : set random seq length mean at <x> [350]\n\
  --num <n>      : set number of sampled seqs to <n> [5000]\n\
  --pvm          : run on a Parallel Virtual Machine (PVM)\n\
  --sd <x>       : set random seq length std. dev to <x> [350]\n\
  --seed <n>     : set random seed to <n> [time()]\n\
";
/* sergio: Redefinicion de OPTIONS a OPTIONScalibrate */
static struct opt_s OPTIONScalibrate[] = {
    {"-h", TRUE, sqdARG_NONE},
    {"--cpu", FALSE, sqdARG_INT},
    {"--fixed", FALSE, sqdARG_INT},
    {"--histfile", FALSE, sqdARG_STRING},
    {"--mean", FALSE, sqdARG_FLOAT},
    {"--num", FALSE, sqdARG_INT},
    {"--pvm", FALSE, sqdARG_NONE},
    {"--sd", FALSE, sqdARG_FLOAT},
    {"--seed", FALSE, sqdARG_INT},
};

/* sergio: Redefinicion de NOPTIONS a NOPTIONScalibrate */
#define NOPTIONScalibrate (sizeof(OPTIONScalibrate) / sizeof(struct opt_s))

/* sergio: Redefinicion de main_loop_serial a main_loop_serialcalibrate */
static void main_loop_serialcalibrate (struct plan7_s *hmm, int seed,
                                       int nsample, float lenmean,
                                       float lensd, int fixedlen,
                                       struct histogram_s **ret_hist,
                                       float *ret_max);

#ifdef HMMER_THREADS
/* A structure of this type is shared by worker threads in the POSIX
 * threads parallel version.
 */
struct workpool_s
{
    /* Static configuration:
     */
    struct plan7_s *hmm;		/* ptr to single HMM to search with    */
    int fixedlen;			/* if >0, fix random seq len to this   */
    float lenmean;		/* mean of Gaussian for random seq len */
    float lensd;			/* s.d. of Gaussian for random seq len */
    float *randomseq;		/* 0..Alphabet_size-1 i.i.d. probs     */
    int nsample;			/* number of random seqs to do         */

    /* Shared (mutex-protected) input:
     */
    int nseq;			/* current number of seqs searched     */

    /* Shared (mutex-protected) output:
     */
    struct histogram_s *hist;	/* histogram          */
    float max_score;		/* maximum score seen */
    Stopwatch_t watch;		/* Timings accumulated for threads */

    /* Thread pool information:
     */
    pthread_t *thread;		/* our pool of threads */
    int num_threads;		/* number of threads   */
    pthread_mutex_t input_lock;	/* a mutex protecting input fields */
    pthread_mutex_t output_lock;	/* a mutex protecting output fields */
};
static void main_loop_threaded (struct plan7_s *hmm, int seed, int nsample,
                                float lenmean, float lensd, int fixedlen,
                                int nthreads,
                                struct histogram_s **ret_hist, float *ret_max,
                                Stopwatch_t * twatch);
static struct workpool_s *workpool_start (struct plan7_s *hmm,
        float lenmean, float lensd,
        int fixedlen, float *randomseq,
        int nsample,
        struct histogram_s *hist,
        int num_threads);
static void workpool_stop (struct workpool_s *wpool);
static void workpool_free (struct workpool_s *wpool);
static void *worker_thread (void *ptr);
#endif /* HMMER_THREADS */

#ifdef HMMER_PVM
static void main_loop_pvm (struct plan7_s *hmm, int seed, int nsample,
                           int lumpsize,
                           float lenmean, float lensd, int fixedlen,
                           struct histogram_s **ret_hist, float *ret_max,
                           Stopwatch_t * extrawatch, int *ret_nslaves);
#endif /* HMMER_PVM */


int
#ifdef SPEC_CPU
hmmcalibrate (int argc, char **argv)
#else
main (int argc, char **argv)
#endif
{
    char *hmmfile;		/* HMM file to open                */
    char *tmpfile;		/* temporary calibrated HMM file   */
    HMMFILE *hmmfp;		/* opened hmm file pointer         */
    FILE *outfp;			/* for writing HMM(s) into tmpfile */
    char *mode;			/* write mode, "w" or "wb"         */
    struct plan7_s *hmm;		/* the hidden Markov model         */
    int idx;			/* counter over sequences          */
#ifndef SPEC_CPU
    sigset_t blocksigs;		/* list of signals to protect from */
#endif
    int nhmm;			/* number of HMMs calibrated       */

    struct histogram_s *hist;	/* a resulting histogram           */
    float max;			/* maximum score from an HMM       */
    char *histfile;		/* histogram save file             */
    FILE *hfp;			/* open file pointer for histfile  */

    Stopwatch_t stopwatch;	/* main stopwatch for process       */
    Stopwatch_t extrawatch;	/* stopwatch for threads/PVM slaves */

    float *mu;			/* array of EVD mu's for HMMs      */
    float *lambda;		/* array of EVD lambda's for HMMs  */
    int mu_lumpsize;		/* allocation lumpsize for mu, lambda */

    int nsample;			/* number of random seqs to sample */
    int seed;			/* random number seed              */
    int fixedlen;			/* fixed length, or 0 if unused    */
    float lenmean;		/* mean of length distribution     */
    float lensd;			/* std dev of length distribution  */
    int do_pvm;			/* TRUE to use PVM                 */
    int pvm_lumpsize;		/* # of seqs to do per PVM slave exchange */
    int pvm_nslaves;		/* number of slaves used in the PVM */


    char *optname;		/* name of option found by Getopt() */
    char *optarg;			/* argument found by Getopt()       */
    int optind;			/* index in argv[]                  */

    int num_threads;		/* number of worker threads */


    /***********************************************
     * Parse the command line
     ***********************************************/
#ifndef SPEC_CPU
    StopwatchStart (&stopwatch);
    StopwatchZero (&extrawatch);
#endif

    nsample = 5000;
    fixedlen = 0;
    lenmean = 325.;
    lensd = 200.;
    seed = (int) time ((time_t *) NULL);
    histfile = NULL;
    do_pvm = FALSE;
    pvm_lumpsize = 20;		/* 20 seqs/PVM exchange: sets granularity */
    mu_lumpsize = 100;
#ifdef HMMER_THREADS
    num_threads = ThreadNumber ();	/* only matters if we're threaded */
#else
    num_threads = 0;
#endif

    while (Getopt
            (argc, argv, OPTIONScalibrate, NOPTIONScalibrate, usagecalibrate,
             &optind, &optname, &optarg)) {
        if (strcmp (optname, "--cpu") == 0) {
            num_threads = atoi (optarg);
        }
        else if (strcmp (optname, "--fixed") == 0) {
            fixedlen = atoi (optarg);
        }
        else if (strcmp (optname, "--histfile") == 0) {
            histfile = optarg;
        }
        else if (strcmp (optname, "--mean") == 0) {
            lenmean = atof (optarg);
        }
        else if (strcmp (optname, "--num") == 0) {
            nsample = atoi (optarg);
        }
        else if (strcmp (optname, "--pvm") == 0) {
            do_pvm = TRUE;
        }
        else if (strcmp (optname, "--sd") == 0) {
            lensd = atof (optarg);
        }
        else if (strcmp (optname, "--seed") == 0) {
            seed = atoi (optarg);
        }
        else if (strcmp (optname, "-h") == 0) {
            HMMERBanner (stdout, bannercalibrate);
            puts (usagecalibrate);
            puts (expertscalibrate);
            exit (0);
        }
    }

    if (argc - optind != 1) {
        Die ("Incorrect number of arguments.\n%s\n", usagecalibrate);
    }
    hmmfile = argv[optind++];

#ifndef HMMER_PVM
    if (do_pvm) {
        Die ("PVM support is not compiled into HMMER; --pvm doesn't work.");
    }
#endif
#ifndef HMMER_THREADS
    if (num_threads)
        Die
        ("Posix threads support is not compiled into HMMER; --cpu doesn't have any effect");
#endif

    /***********************************************
     * Open our i/o file pointers, make sure all is well
     ***********************************************/

    /* HMM file */
    if ((hmmfp = HMMFileOpen (hmmfile, NULL)) == NULL) {
        Die ("failed to open HMM file %s for reading.", hmmfile);
    }

    /* histogram file */
    hfp = NULL;
    if (histfile != NULL) {
        if ((hfp = fopen (histfile, "w")) == NULL) {
            Die ("Failed to open histogram save file %s for writing\n", histfile);
        }
    }

    /* Generate calibrated HMM(s) in a tmp file in the current
     * directory. When we're finished, we delete the original
     * HMM file and rename() this one. That way, the worst
     * effect of a catastrophic failure should be that we
     * leave a tmp file lying around, but the original HMM
     * file remains uncorrupted. tmpnam() doesn't work portably here,
     * because it'll put the file in /tmp and we won't
     * necessarily be able to rename() it from there.
     */
    tmpfile = MallocOrDie (strlen (hmmfile) + 5);
    strcpy (tmpfile, hmmfile);
    strcat (tmpfile, ".new");	/* could be more inventive here... */
    if (FileExists (tmpfile))
#ifdef SPEC_CPU
        if (remove (tmpfile) != 0) {
            PANIC;
        }
#else
        Die ("temporary file %s already exists; please delete it first", tmpfile);
#endif
    if (hmmfp->is_binary) {
        mode = "wb";
    }
    else {
        mode = "w";
    }

    /***********************************************
     * Show the banner
     ***********************************************/

    HMMERBanner (stdout, bannercalibrate);
    printf ("HMM file:                 %s\n", hmmfile);
    if (fixedlen) {
        printf ("Length fixed to:          %d\n", fixedlen);
    }
    else {
        printf ("Length distribution mean: %.0f\n", lenmean);
        printf ("Length distribution s.d.: %.0f\n", lensd);
    }
    printf ("Number of samples:        %d\n", nsample);
    printf ("random seed:              %d\n", seed);
    printf ("histogram(s) saved to:    %s\n",
            histfile != NULL ? histfile : "[not saved]");
    if (do_pvm) {
        printf ("PVM:                      ACTIVE\n");
    }
    else if (num_threads > 0) {
        printf ("POSIX threads:            %d\n", num_threads);
    }
    printf
    ("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n\n");

    /***********************************************
     * Read the HMMs one at a time, and send them off
     * in probability form to one of the main loops.
     * The main loop functions are responsible for
     * synthesizing random sequences and returning
     * a score histogram for each HMM.
     ***********************************************/

    nhmm = 0;
    mu = MallocOrDie (sizeof (float) * mu_lumpsize);
    lambda = MallocOrDie (sizeof (float) * mu_lumpsize);

    while (HMMFileRead (hmmfp, &hmm)) {
        if (hmm == NULL) {
            Die ("HMM file may be corrupt or in incorrect format; parse failed");
        }

        if (!do_pvm && num_threads == 0)
            main_loop_serialcalibrate (hmm, seed, nsample, lenmean, lensd, fixedlen,
                                       &hist, &max);
#ifdef HMMER_PVM
        else if (do_pvm) {
            pvm_nslaves = 0;		/* solely to silence compiler warnings */
            main_loop_pvm (hmm, seed, nsample, pvm_lumpsize,
                           lenmean, lensd, fixedlen,
                           &hist, &max, &extrawatch, &pvm_nslaves);
        }
#endif
#ifdef HMMER_THREADS
        else if (num_threads > 0)
            main_loop_threaded (hmm, seed, nsample, lenmean, lensd, fixedlen,
                                num_threads, &hist, &max, &extrawatch);
#endif
        else {
            Die ("wait. that can't happen. I didn't do anything.");
        }


        /* Fit an EVD to the observed histogram.
         * The TRUE left-censors and fits only the right slope of the histogram.
         * The 9999. is an arbitrary high number that means we won't trim
         * outliers on the right.
         */
        if (!ExtremeValueFitHistogram (hist, TRUE, 9999.)) {
            Die ("fit failed; --num may be set too small?\n");
        }

        mu[nhmm] = hist->param[EVD_MU];
        lambda[nhmm] = hist->param[EVD_LAMBDA];
        nhmm++;
        if (nhmm % 100 == 0) {
            mu = ReallocOrDie (mu, sizeof (float) * (nhmm + mu_lumpsize));
            lambda = ReallocOrDie (lambda, sizeof (float) * (nhmm + mu_lumpsize));
        }

        /* Output
         */
        printf ("HMM    : %s\n", hmm->name);
        printf ("mu     : %12f\n", hist->param[EVD_MU]);
        printf ("lambda : %12f\n", hist->param[EVD_LAMBDA]);
        printf ("max    : %12f\n", max);
fprintf(stderr,"tiempo especulativo: %f\n",spec_time);
        printf ("//\n");

#ifdef TIME_SPECLOAD
  printSpecloadTime();
#endif
#ifdef TIME_SPECSTORE
  printSpecstoreTime();
#endif
#ifdef MEASURE_COUNTERS
  specepilogue();
  printAccesses();
#endif


        if (hfp != NULL) {
            fprintf (hfp, "HMM: %s\n", hmm->name);
            PrintASCIIHistogram (hfp, hist);
            fprintf (hfp, "//\n");
        }

        FreeHistogram (hist);
        FreePlan7 (hmm);
    }
    SQD_DPRINTF1 (("Main body believes it has calibrations for %d HMMs\n",
                   nhmm));

    /*****************************************************************
     * Rewind the HMM file for a second pass.
     * Write a temporary HMM file with new mu, lambda values in it
     *****************************************************************/

    HMMFileRewind (hmmfp);
    if (FileExists (tmpfile)) {
        Die ("Ouch. Temporary file %s appeared during the run.", tmpfile);
    }
    if ((outfp = fopen (tmpfile, mode)) == NULL) {
        Die ("Ouch. Temporary file %s couldn't be opened for writing.", tmpfile);
    }

    for (idx = 0; idx < nhmm; idx++) {
        /* Sanity checks
         */
        if (!HMMFileRead (hmmfp, &hmm)) {
            Die ("Ran out of HMMs too early in pass 2");
        }
        if (hmm == NULL) {
            Die ("HMM file %s was corrupted? Parse failed in pass 2", hmmfile);
        }

        /* Put results in HMM
         */
        hmm->mu = mu[idx];
        hmm->lambda = lambda[idx];
        hmm->flags |= PLAN7_STATS;
        Plan7ComlogAppend (hmm, argc, argv);

        /* Save HMM to tmpfile
         */
        if (hmmfp->is_binary) {
            WriteBinHMM (outfp, hmm);
        }
        else {
            WriteAscHMM (outfp, hmm);
        }

        FreePlan7 (hmm);
    }

    /*****************************************************************
     * Now, carefully remove original file and replace it
     * with the tmpfile. Note the protection from signals;
     * we wouldn't want a user to ctrl-C just as we've deleted
     * their HMM file but before the new one is moved.
     *****************************************************************/

    HMMFileClose (hmmfp);
    if (fclose (outfp) != 0) {
        PANIC;
    }

#ifndef SPEC_CPU
    if (sigemptyset (&blocksigs) != 0) {
        PANIC;
    }
    if (sigaddset (&blocksigs, SIGINT) != 0) {
        PANIC;
    }
    if (sigprocmask (SIG_BLOCK, &blocksigs, NULL) != 0) {
        PANIC;
    }
    if (remove (hmmfile) != 0) {
        PANIC;
    }
    if (rename (tmpfile, hmmfile) != 0) {
        PANIC;
    }
    if (sigprocmask (SIG_UNBLOCK, &blocksigs, NULL) != 0) {
        PANIC;
    }
#endif
    free (tmpfile);

    /***********************************************
     * Exit
     ***********************************************/

#ifndef SPEC_CPU
    StopwatchStop (&stopwatch);
#endif
#ifdef HMMER_PVM
    if (do_pvm > 0) {
        printf ("PVM processors used: %d\n", pvm_nslaves);
        StopwatchInclude (&stopwatch, &extrawatch);
    }
#endif
#ifdef PTHREAD_TIMES_HACK
    else if (num_threads > 0) {
        StopwatchInclude (&stopwatch, &extrawatch);
    }
#endif

    /*  StopwatchDisplay(stdout, "CPU Time: ", &stopwatch); */

    free (mu);
    free (lambda);
    if (hfp != NULL) {
        fclose (hfp);
    }
    SqdClean ();
    return 0;
}

/* Function: main_loop_serialcalibrate()
 * Date:     SRE, Tue Aug 18 16:18:28 1998 [St. Louis]
 *
 * Purpose:  Given an HMM and parameters for synthesizing random
 *           sequences; return a histogram of scores.
 *           (Serial version)
 *
 * Args:     hmm      - an HMM to calibrate.
 *           seed     - random number seed
 *           nsample  - number of seqs to synthesize
 *           lenmean  - mean length of random sequence
 *           lensd    - std dev of random seq length
 *           fixedlen - if nonzero, override lenmean, always this len
 *           ret_hist - RETURN: the score histogram
 *           ret_max  - RETURN: highest score seen in simulation
 *
 * Returns:  (void)
 *           hist is alloc'ed here, and must be free'd by caller.
 */
static void
main_loop_serialcalibrate (struct plan7_s *hmm, int seed, int nsample,
                           float lenmean, float lensd, int fixedlen,
                           struct histogram_s **ret_hist, float *ret_max)
{
    struct histogram_s *hist;
    struct dpmatrix_s *mx;
    float randomseq[MAXABET];
    float p1;
    float max;
    char *seq;
    char *dsq;
    float score;
    int sqlen;
    int idx;

    /* Initialize.
     * We assume we've already set the alphabet (safe, because
     * HMM input sets the alphabet).
     */
    sre_srandom (seed);
    P7Logoddsify (hmm, TRUE);
    P7DefaultNullModel (randomseq, &p1);
    hist = AllocHistogram (-200, 200, 100);
    mx = CreatePlan7Matrix (1, hmm->M, 25, 0);
    max = -FLT_MAX;

    for (idx = 0; idx < nsample; idx++) {
        /* choose length of random sequence */
        if (fixedlen) {
            sqlen = fixedlen;
        }
        else
            do {
                sqlen = (int) Gaussrandom (lenmean, lensd);
            }
            while (sqlen < 1);
        /* generate it */
        seq = RandomSequence (Alphabet, randomseq, Alphabet_size, sqlen);
        dsq = DigitizeSequence (seq, sqlen);

        if (P7ViterbiSize (sqlen, hmm->M) <= RAMLIMIT) {
            score = P7Viterbi (dsq, sqlen, hmm, mx, NULL);
        }
        else {
            score = P7SmallViterbi (dsq, sqlen, hmm, mx, NULL);
        }

        AddToHistogram (hist, score);
        if (score > max) {
            max = score;
        }

        free (dsq);
        free (seq);
    }

    FreePlan7Matrix (mx);
    *ret_hist = hist;
    *ret_max = max;
    return;
}


#ifdef HMMER_THREADS
/* Function: main_loop_threaded()
 * Date:     SRE, Wed Dec  1 12:43:09 1999 [St. Louis]
 *
 * Purpose:  Given an HMM and parameters for synthesizing random
 *           sequences; return a histogram of scores.
 *           (Threaded version.)
 *
 * Args:     hmm      - an HMM to calibrate.
 *           seed     - random number seed
 *           nsample  - number of seqs to synthesize
 *           lenmean  - mean length of random sequence
 *           lensd    - std dev of random seq length
 *           fixedlen - if nonzero, override lenmean, always this len
 *           nthreads - number of threads to start
 *           ret_hist - RETURN: the score histogram
 *           ret_max  - RETURN: highest score seen in simulation
 *           twatch   - RETURN: accumulation of thread times
 *
 * Returns:  (void)
 *           hist is alloc'ed here, and must be free'd by caller.
 */
static void
main_loop_threaded (struct plan7_s *hmm, int seed, int nsample,
                    float lenmean, float lensd, int fixedlen,
                    int nthreads,
                    struct histogram_s **ret_hist, float *ret_max,
                    Stopwatch_t * twatch)
{
    struct histogram_s *hist;
    float randomseq[MAXABET];
    float p1;
    struct workpool_s *wpool;	/* pool of worker threads  */

    /* Initialize.
     * We assume we've already set the alphabet (safe, because
     * HMM input sets the alphabet).
     */
    sre_srandom (seed);
    P7Logoddsify (hmm, TRUE);
    P7DefaultNullModel (randomseq, &p1);
    hist = AllocHistogram (-200, 200, 100);

    wpool = workpool_start (hmm, lenmean, lensd, fixedlen, randomseq, nsample,
                            hist, nthreads);
    workpool_stop (wpool);

    *ret_hist = hist;
    *ret_max = wpool->max_score;
    StopwatchInclude (twatch, &(wpool->watch));

    workpool_free (wpool);
    return;
}

/*****************************************************************
 * POSIX threads implementation.
 * API:
 *      workpool_start()   (makes a workpool_s structure. Starts calculations.)
 *      workpool_stop()    (waits for threads to finish.)
 *      [process histogram]
 *      workpool_free()    (destroys the structure)
 *
 * Threads:
 *      worker_thread()    (the actual parallelized worker thread).
 *****************************************************************/

/* Function: workpool_start()
 * Date:     SRE, Thu Jul 16 11:09:05 1998 [St. Louis]
 *
 * Purpose:  Initialize a workpool_s structure, and return it.
 *
 * Args:     hmm      - the HMM to calibrate
 *           fixedlen - 0, or a fixed length for seqs (bypass of Gaussian)
 *           lenmean  - mean sequence length
 *           lensd    - std. dev. for sequence length
 *           randomseq- i.i.d. frequencies for residues, 0..Alphabet_size-1
 *           nsample  - how many seqs to calibrate on
 *           hist     - histogram structure for storing results
 *           num_threads - how many processors to run on
 *
 * Returns:  ptr to struct workpool_s.
 *           Caller must wait for threads to finish with workpool_stop(),
 *           then free the structure with workpool_free().
 */
static struct workpool_s *
workpool_start (struct plan7_s *hmm, float lenmean, float lensd, int fixedlen,
                float *randomseq, int nsample, struct histogram_s *hist,
                int num_threads)
{
    struct workpool_s *wpool;
    pthread_attr_t attr;
    int i;
    int rtn;

    wpool = MallocOrDie (sizeof (struct workpool_s));
    wpool->thread = MallocOrDie (num_threads * sizeof (pthread_t));
    wpool->hmm = hmm;
    wpool->fixedlen = fixedlen;
    wpool->lenmean = lenmean;
    wpool->lensd = lensd;
    wpool->randomseq = randomseq;
    wpool->nsample = nsample;

    wpool->nseq = 0;
    wpool->hist = hist;
    wpool->max_score = -FLT_MAX;
    wpool->num_threads = num_threads;

    StopwatchZero (&(wpool->watch));

    if ((rtn = pthread_mutex_init (&(wpool->input_lock), NULL)) != 0) {
        Die ("pthread_mutex_init FAILED; %s\n", strerror (rtn));
    }
    if ((rtn = pthread_mutex_init (&(wpool->output_lock), NULL)) != 0) {
        Die ("pthread_mutex_init FAILED; %s\n", strerror (rtn));
    }

    /* Create slave threads.
     * Note the crazy machinations we have to go through to achieve concurrency.
     * You'd think that POSIX threads were portable... ha.
     * On IRIX 6.5, system scope threads are only available to root, or if
     *   /etc/capability has been configured specially, so to avoid strange
     *   permissions errors we can't set PTHREAD_SCOPE_SYSTEM for IRIX.
     * On IRIX pre-6.5, we can't get good concurrency, period. As of 6.5,
     *   SGI provides the nonportable pthread_setconcurrency() call.
     * On FreeBSD (3.0 snapshots), the pthread_attr_setscope() call isn't
     *   even provided, apparently on grounds of "if it doesn't do anything,
     *   why provide it?" Hello? POSIX compliance, perhaps?
     * On Sun Solaris, we need to set system scope to achieve concurrency.
     * Linux and DEC Digital UNIX seem to work fine in either process scope
     *   or system scope, without a pthread_setconcurrency call.
     */
    pthread_attr_init (&attr);
#ifndef __sgi
#ifdef HAVE_PTHREAD_ATTR_SETSCOPE
    pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
#endif
#endif
#ifdef HAVE_PTHREAD_SETCONCURRENCY
    pthread_setconcurrency (num_threads + 1);
#endif
    for (i = 0; i < num_threads; i++) {
        if ((rtn = pthread_create (&(wpool->thread[i]), &attr,worker_thread, (void *) wpool)) != 0) {
            Die ("Failed to create thread %d; return code %d\n", i, rtn);
        }
    }
    pthread_attr_destroy (&attr);

    return wpool;
}

/* Function: workpool_stop()
 * Date:     SRE, Thu Jul 16 11:20:16 1998 [St. Louis]
 *
 * Purpose:  Waits for threads in a workpool to finish.
 *
 * Args:     wpool -- ptr to the workpool structure
 *
 * Returns:  (void)
 */
static void
workpool_stop (struct workpool_s *wpool)
{
    int i;
    /* wait for threads to stop */
    for (i = 0; i < wpool->num_threads; i++) {
        if (pthread_join (wpool->thread[i], NULL) != 0) {
            Die ("pthread_join failed");
        }
    }
    return;
}

/* Function: workpool_free()
 * Date:     SRE, Thu Jul 16 11:26:27 1998 [St. Louis]
 *
 * Purpose:  Free a workpool_s structure, after the threads
 *           have finished.
 *
 * Args:     wpool -- ptr to the workpool.
 *
 * Returns:  (void)
 */
static void
workpool_free (struct workpool_s *wpool)
{
    free (wpool->thread);
    free (wpool);
    return;
}

/* Function: worker_thread()
 * Date:     SRE, Thu Jul 16 10:41:02 1998 [St. Louis]
 *
 * Purpose:  The procedure executed by the worker threads.
 *
 * Args:     ptr  - (void *) that is recast to a pointer to
 *                  the workpool.
 *
 * Returns:  (void *)
 */
void *
worker_thread (void *ptr)
{
    struct plan7_s *hmm;
    struct dpmatrix_s *mx;
    struct workpool_s *wpool;
    char *seq;
    char *dsq;
    int len;
    float sc;
    int rtn;
    Stopwatch_t thread_watch;

    StopwatchStart (&thread_watch);
    wpool = (struct workpool_s *) ptr;
    hmm = wpool->hmm;
    mx = CreatePlan7Matrix (1, hmm->M, 25, 0);
    for (;;) {
        /* 1. Synthesize a random sequence.
         *    The input sequence number is a shared resource,
         *    and sre_random() isn't thread-safe, so protect
         *    the whole section with mutex.
         */
        /* acquire a lock */
        if ((rtn = pthread_mutex_lock (&(wpool->input_lock))) != 0) {
            Die ("pthread_mutex_lock failure: %s\n", strerror (rtn));
        }
        /* generate a sequence */
        wpool->nseq++;
        if (wpool->nseq > wpool->nsample) {	/* we're done; release input lock, break loop */
            if ((rtn = pthread_mutex_unlock (&(wpool->input_lock))) != 0) {
                Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
            }
            break;
        }
        if (wpool->fixedlen) {
            len = wpool->fixedlen;
        }
        else
            do {
                len = (int) Gaussrandom (wpool->lenmean, wpool->lensd);
            }
            while (len < 1);
        seq = RandomSequence (Alphabet, wpool->randomseq, Alphabet_size, len);

        /* release the lock */
        if ((rtn = pthread_mutex_unlock (&(wpool->input_lock))) != 0) {
            Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
        }

        /* 2. Score the sequence against the model.
         */
        dsq = DigitizeSequence (seq, len);

        if (P7ViterbiSize (len, hmm->M) <= RAMLIMIT) {
            sc = P7Viterbi (dsq, len, hmm, mx, NULL);
        }
        else {
            sc = P7SmallViterbi (dsq, len, hmm, mx, NULL);
        }
        free (dsq);
        free (seq);

        /* 3. Save the output; hist and max_score are shared,
         *    so protect this section with the output mutex.
         */
        /* acquire lock on the output queue */
        if ((rtn = pthread_mutex_lock (&(wpool->output_lock))) != 0) {
            Die ("pthread_mutex_lock failure: %s\n", strerror (rtn));
        }
        /* save output */
        AddToHistogram (wpool->hist, sc);
        if (sc > wpool->max_score) {
            wpool->max_score = sc;
        }
        /* release our lock */
        if ((rtn = pthread_mutex_unlock (&(wpool->output_lock))) != 0) {
            Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
        }
    }

    StopwatchStop (&thread_watch);
    /* acquire lock on the output queue */
    if ((rtn = pthread_mutex_lock (&(wpool->output_lock))) != 0) {
        Die ("pthread_mutex_lock failure: %s\n", strerror (rtn));
    }
    /* accumulate cpu time into main stopwatch */
    StopwatchInclude (&(wpool->watch), &thread_watch);
    /* release our lock */
    if ((rtn = pthread_mutex_unlock (&(wpool->output_lock))) != 0) {
        Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
    }

    FreePlan7Matrix (mx);
    pthread_exit (NULL);
    return NULL;			/* solely to silence compiler warnings */
}
#endif /* HMMER_THREADS */



#ifdef HMMER_PVM
/* Function: main_loop_pvm()
 * Date:     SRE, Wed Aug 19 13:59:54 1998 [St. Louis]
 *
 * Purpose:  Given an HMM and parameters for synthesizing random
 *           sequences; return a histogram of scores.
 *           (PVM version)
 *
 * Args:     hmm     - an HMM to calibrate.
 *           seed    - random number seed
 *           nsample - number of seqs to synthesize
 *           lumpsize- # of seqs per slave exchange; controls granularity
 *           lenmean - mean length of random sequence
 *           lensd   - std dev of random seq length
 *           fixedlen- if nonzero, override lenmean, always this len
 *           hist       - RETURN: the score histogram
 *           ret_max    - RETURN: highest score seen in simulation
 *           extrawatch - RETURN: total CPU time spend in slaves.
 *           ret_nslaves- RETURN: number of PVM slaves run.
 *
 * Returns:  (void)
 *           hist is alloc'ed here, and must be free'd by caller.
 */
static void
main_loop_pvm (struct plan7_s *hmm, int seed, int nsample, int lumpsize,
               float lenmean, float lensd, int fixedlen,
               struct histogram_s **ret_hist, float *ret_max,
               Stopwatch_t * extrawatch, int *ret_nslaves)
{
    struct histogram_s *hist;
    int master_tid;
    int *slave_tid;
    int nslaves;
    int nsent;			/* # of seqs we've asked for so far       */
    int ndone;			/* # of seqs we've got results for so far */
    int packet;			/* # of seqs to have a slave do           */
    float max;
    int slaveidx;			/* id of a slave */
    float *sc;			/* scores returned by a slave */
    Stopwatch_t slavewatch;
    int i;

    StopwatchZero (extrawatch);
    hist = AllocHistogram (-200, 200, 100);
    max = -FLT_MAX;

    /* Initialize PVM
     */
    if ((master_tid = pvm_mytid ()) < 0) {
        Die ("pvmd not responding -- do you have PVM running?");
    }
#if DEBUGLEVEL >= 1
    pvm_catchout (stderr);	/* catch output for debugging */
#endif
    PVMSpawnSlaves ("hmmcalibrate-pvm", &slave_tid, &nslaves);

    /* Initialize the slaves
     */
    pvm_initsend (PvmDataDefault);
    pvm_pkfloat (&lenmean, 1, 1);
    pvm_pkfloat (&lensd, 1, 1);
    pvm_pkint (&fixedlen, 1, 1);
    pvm_pkint (&Alphabet_type, 1, 1);
    pvm_pkint (&seed, 1, 1);
    if (!PVMPackHMM (hmm)) {
        Die ("Failed to pack the HMM");
    }
    pvm_mcast (slave_tid, nslaves, HMMPVM_INIT);
    SQD_DPRINTF1 (("Initialized %d slaves\n", nslaves));

    /* Confirm slaves' OK status.
     */
    PVMConfirmSlaves (slave_tid, nslaves);
    SQD_DPRINTF1 (("Slaves confirm that they're ok...\n"));

    /* Load the slaves
     */
    nsent = ndone = 0;
    for (slaveidx = 0; slaveidx < nslaves; slaveidx++) {
        packet = (nsample - nsent > lumpsize ? lumpsize : nsample - nsent);

        pvm_initsend (PvmDataDefault);
        pvm_pkint (&packet, 1, 1);
        pvm_pkint (&slaveidx, 1, 1);
        pvm_send (slave_tid[slaveidx], HMMPVM_WORK);
        nsent += packet;
    }
    SQD_DPRINTF1 (("Loaded %d slaves\n", nslaves));

    /* Receive/send loop
     */
    sc = MallocOrDie (sizeof (float) * lumpsize);
    while (nsent < nsample) {
        /* integrity check of slaves */
        PVMCheckSlaves (slave_tid, nslaves);

        /* receive results */
        SQD_DPRINTF2 (("Waiting for results...\n"));
        pvm_recv (-1, HMMPVM_RESULTS);
        pvm_upkint (&slaveidx, 1, 1);
        pvm_upkint (&packet, 1, 1);
        pvm_upkfloat (sc, packet, 1);
        SQD_DPRINTF2 (("Got results.\n"));
        ndone += packet;

        /* store results */
        for (i = 0; i < packet; i++) {
            AddToHistogram (hist, sc[i]);
            if (sc[i] > max) {
                max = sc[i];
            }
        }
        /* send new work */
        packet = (nsample - nsent > lumpsize ? lumpsize : nsample - nsent);

        pvm_initsend (PvmDataDefault);
        pvm_pkint (&packet, 1, 1);
        pvm_pkint (&slaveidx, 1, 1);
        pvm_send (slave_tid[slaveidx], HMMPVM_WORK);
        SQD_DPRINTF2 (("Told slave %d to do %d more seqs.\n", slaveidx, packet));
        nsent += packet;
    }

    /* Wait for the last output to come in.
     */
    while (ndone < nsample) {
        /* integrity check of slaves */
        PVMCheckSlaves (slave_tid, nslaves);

        /* receive results */
        SQD_DPRINTF1 (("Waiting for final results...\n"));
        pvm_recv (-1, HMMPVM_RESULTS);
        pvm_upkint (&slaveidx, 1, 1);
        pvm_upkint (&packet, 1, 1);
        pvm_upkfloat (sc, packet, 1);
        SQD_DPRINTF2 (("Got some final results.\n"));
        ndone += packet;
        /* store results */
        for (i = 0; i < packet; i++) {
            AddToHistogram (hist, sc[i]);
            if (sc[i] > max) {
                max = sc[i];
            }
        }
    }

    /* Shut down the slaves: send -1,-1,-1.
     */
    pvm_initsend (PvmDataDefault);
    packet = -1;
    pvm_pkint (&packet, 1, 1);
    pvm_pkint (&packet, 1, 1);
    pvm_pkint (&packet, 1, 1);
    pvm_mcast (slave_tid, nslaves, HMMPVM_WORK);

    /* Collect stopwatch results; quit the VM; return.
     */
    for (i = 0; i < nslaves; i++) {
        pvm_recv (-1, HMMPVM_RESULTS);
        pvm_upkint (&slaveidx, 1, 1);
        StopwatchPVMUnpack (&slavewatch);

        SQD_DPRINTF1 (("Slave %d finished; says it used %.2f cpu, %.2f sys\n",
                       slaveidx, slavewatch.user, slavewatch.sys));

        StopwatchInclude (extrawatch, &slavewatch);
    }

    free (slave_tid);
    free (sc);
    pvm_exit ();
    *ret_hist = hist;
    *ret_max = max;
    *ret_nslaves = nslaves;
    return;
}
#endif /* HMMER_PVM */



/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* hmmio.c
 *
 * Input/output of HMMs.
 *
 * As of HMMER 2.0, HMMs are saved by default in a tabular ASCII format
 * as log-odds or log probabilities scaled to an integer. A binary save
 * file format is also available which is faster to access (a
 * consideration which might be important for HMM library applications).
 * HMMs can be concatenated into HMM libraries.
 *
 * A comment on loss of accuracy. Storing a number as a scaled log
 * probability guarantees us an error of about 0.035% or
 * less in the retrieved probability. We are relatively invulnerable
 * to the truncation errors which HMMER 1.8 was vulnerable to.
 *
 * Magic numbers (both for the ASCII and binary save formats) are used
 * to label save files with a major version number. This simplifies the task of
 * backwards compatibility as new versions of the program are created.
 * Reverse but not forward compatibility is guaranteed. I.e. HMMER 2.0
 * can read `1.7' save files, but not vice versa. Note that the major
 * version number in the save files is NOT the version of the software
 * that generated it; rather, the number of the last major version in which
 * save format changed.
 *
 ******************************************************************
 *
 * The HMM input API:
 *
 *       HMMFILE        *hmmfp;
 *       char           *hmmfile;
 *       struct plan7_s *hmm;
 *       char            env[] = "HMMERDB";  (a la BLASTDB)
 *
 *       hmmfp = HMMFileOpen(hmmfile, env)   NULL on failure
 *       while (HMMFileRead(hmmfp, &hmm))    0 if no more HMMs
 *          if (hmm == NULL) Die();          NULL on file parse failure
 *          whatever;
 *          FreeHMM(hmm);
 *       }
 *       HMMFileClose(hmmfp);
 *
 *****************************************************************
 *
 * The HMM output API:
 *
 *       FILE           *ofp;
 *       struct plan7_s *hmm;
 *
 *       WriteAscHMM(ofp, hmm);    to write/append an HMM to open file
 *   or  WriteBinHMM(ofp, hmm);    to write/append binary format HMM to open file
 *
 *****************************************************************
 *
 * V1.0: original implementation
 * V1.1: regularizers removed from model structure
 * V1.7: ref and cs annotation lines added from alignment, one
 *       char per match state 1..M
 * V1.9: null model and name added to HMM structure. ASCII format changed
 *       to compact tabular one.
 * V2.0: Plan7. Essentially complete rewrite.
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#if !defined(SPEC_CPU_WINDOWS)
#include <unistd.h>		/* to get SEEK_CUR definition on silly Suns */
#endif /* SPEC_CPU_WINDOWS */
#include "squid.h"
#include "ssi.h"
#include "structs.h"
#include "funcs.h"


/* Magic numbers identifying binary formats.
 * Do not change the old magics! Necessary for backwards compatibility.
 */
static unsigned int v10magic = 0xe8ededb1;	/* v1.0 binary: "hmm1" + 0x80808080 */
static unsigned int v10swap = 0xb1edede8;	/* byteswapped v1.0                 */
static unsigned int v11magic = 0xe8ededb2;	/* v1.1 binary: "hmm2" + 0x80808080 */
static unsigned int v11swap = 0xb2edede8;	/* byteswapped v1.1                 */
static unsigned int v17magic = 0xe8ededb3;	/* v1.7 binary: "hmm3" + 0x80808080 */
static unsigned int v17swap = 0xb3edede8;	/* byteswapped v1.7                 */
static unsigned int v19magic = 0xe8ededb4;	/* V1.9 binary: "hmm4" + 0x80808080 */
static unsigned int v19swap = 0xb4edede8;	/* V1.9 binary, byteswapped         */
static unsigned int v20magic = 0xe8ededb5;	/* V2.0 binary: "hmm5" + 0x80808080 */
static unsigned int v20swap = 0xb5edede8;	/* V2.0 binary, byteswapped         */

/* Old HMMER 1.x file formats.
 */
#define HMMER1_0B  1		/* binary HMMER 1.0     */
#define HMMER1_0F  2		/* flat ascii HMMER 1.0 */
#define HMMER1_1B  3		/* binary HMMER 1.1     */
#define HMMER1_1F  4		/* flat ascii HMMER 1.1 */
#define HMMER1_7B  5		/* binary HMMER 1.7     */
#define HMMER1_7F  6		/* flat ascii HMMER 1.7 */
#define HMMER1_9B  7		/* HMMER 1.9 binary     */
#define HMMER1_9F  8		/* HMMER 1.9 flat ascii */

static int read_asc20hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_bin20hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_asc19hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_bin19hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_asc17hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_bin17hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_asc11hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_bin11hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_asc10hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);
static int read_bin10hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm);

static void byteswap (char *swap, int nbytes);
static char *prob2ascii (float p, float null);
static float ascii2prob (char *s, float null);
static void write_bin_string (FILE * fp, char *s);
static int read_bin_string (FILE * fp, int doswap, char **ret_s);
static void multiline (FILE * fp, char *pfx, char *s);

static struct plan9_s *read_plan9_binhmm (FILE * fp, int version,
        int swapped);
static struct plan9_s *read_plan9_aschmm (FILE * fp, int version);

/*****************************************************************
 * HMM input API functions:
 *   HMMFileOpen()
 *   HMMFileRead()
 *   HMMFileClose()
 *   HMMFileRewind()
 *****************************************************************/

/* Function: HMMFileOpen()
 *
 * Purpose:  Open an HMM file for reading. The file may be either
 *           an index for a library of HMMs, or an HMM.
 *
 * Args:     hmmfile - name of file
 *           env     - NULL, or environment variable for HMM database.
 *
 * Return:   Valid HMMFILE *, or NULL on failure.
 */
HMMFILE *
HMMFileOpen (char *hmmfile, char *env)
{
    HMMFILE *hmmfp;
    unsigned int magic;
    char buf[512];
    char *ssifile;
    char *dir;			/* dir name in which HMM file was found */
    int status;

    hmmfp = (HMMFILE *) MallocOrDie (sizeof (HMMFILE));
    hmmfp->f = NULL;
    hmmfp->parser = NULL;
    hmmfp->is_binary = FALSE;
    hmmfp->byteswap = FALSE;
    hmmfp->is_seekable = TRUE;	/* always; right now, an HMM must always be in a file. */

    /* Open the file. Look in current directory.
     * If that doesn't work, check environment var for
     * a second possible directory (usually the location
     * of a system-wide HMM library).
     * Using dir name if necessary, construct correct SSI file name.
     */
    hmmfp->f = NULL;
    hmmfp->ssi = NULL;
    if ((hmmfp->f = fopen (hmmfile, "r")) != NULL) {
        ssifile = MallocOrDie (sizeof (char) * (strlen (hmmfile) + 5));
        sprintf (ssifile, "%s.ssi", hmmfile);

        if ((hmmfp->mode = SSIRecommendMode (hmmfile)) == -1) {
            Die ("SSIRecommendMode() failed");
        }
    }
    else if ((hmmfp->f = EnvFileOpen (hmmfile, env, &dir)) != NULL) {
        char *full;
        full = FileConcat (dir, hmmfile);

        ssifile =
            MallocOrDie (sizeof (char) * (strlen (full) + strlen (hmmfile) + 5));
        sprintf (ssifile, "%s.ssi", full);

        if ((hmmfp->mode = SSIRecommendMode (full)) == -1) {
            Die ("SSIRecommendMode() failed");
        }

        free (full);
        free (dir);
    }
    else {
        return NULL;
    }

    /* Open the SSI index file. If it doesn't exist, or it's corrupt, or
     * some error happens, hmmfp->ssi stays NULL.
     */
    SQD_DPRINTF1 (("Opening ssifile %s...\n", ssifile));
    SSIOpen (ssifile, &(hmmfp->ssi));
    free (ssifile);

    /* Initialize the disk offset stuff.
     */
    status = SSIGetFilePosition (hmmfp->f, hmmfp->mode, &(hmmfp->offset));
    if (status != 0) {
        Die ("SSIGetFilePosition() failed");
    }

    /* Check for binary or byteswapped binary format
     * by peeking at first 4 bytes.
     */
    if (!fread ((char *) &magic, sizeof (unsigned int), 1, hmmfp->f)) {
        HMMFileClose (hmmfp);
        return NULL;
    }
    rewind (hmmfp->f);

    if (magic == v20magic) {
        hmmfp->parser = read_bin20hmm;
        hmmfp->is_binary = TRUE;
        return hmmfp;
    }
    else if (magic == v20swap) {
        SQD_DPRINTF1 (("Opened a HMMER 2.0 binary file [byteswapped]\n"));
        hmmfp->parser = read_bin20hmm;
        hmmfp->is_binary = TRUE;
        hmmfp->byteswap = TRUE;
        return hmmfp;
    }
    else if (magic == v19magic) {
        hmmfp->parser = read_bin19hmm;
        hmmfp->is_binary = TRUE;
        return hmmfp;
    }
    else if (magic == v19swap) {
        hmmfp->parser = read_bin19hmm;
        hmmfp->is_binary = TRUE;
        hmmfp->byteswap = TRUE;
        return hmmfp;
    }
    else if (magic == v17magic) {
        hmmfp->parser = read_bin17hmm;
        hmmfp->is_binary = TRUE;
        return hmmfp;
    }
    else if (magic == v17swap) {
        hmmfp->parser = read_bin17hmm;
        hmmfp->is_binary = TRUE;
        hmmfp->byteswap = TRUE;
        return hmmfp;
    }
    else if (magic == v11magic) {
        hmmfp->parser = read_bin11hmm;
        hmmfp->is_binary = TRUE;
        return hmmfp;
    }
    else if (magic == v11swap) {
        hmmfp->parser = read_bin11hmm;
        hmmfp->is_binary = TRUE;
        hmmfp->byteswap = TRUE;
        return hmmfp;
    }
    else if (magic == v10magic) {
        hmmfp->parser = read_bin10hmm;
        hmmfp->is_binary = TRUE;
        return hmmfp;
    }
    else if (magic == v10swap) {
        hmmfp->parser = read_bin10hmm;
        hmmfp->is_binary = TRUE;
        hmmfp->byteswap = TRUE;
        return hmmfp;
    }
    /* else we fall thru; it may be an ASCII file. */

    /* If magic looks binary but we don't recognize it, choke and die.
     */
    if (magic & 0x80000000) {
        Warn ("\
%s appears to be a binary but format is not recognized\n\
It may be from a HMMER version more recent than yours,\n\
or may be a different kind of binary altogether.\n", hmmfile);
        HMMFileClose (hmmfp);
        return NULL;
    }

    /* Check for ASCII format by peeking at first word.
     */
    if (fgets (buf, 512, hmmfp->f) == NULL) {
        HMMFileClose (hmmfp);
        return NULL;
    }
    rewind (hmmfp->f);

    if (strncmp ("HMMER2.0", buf, 8) == 0) {
        hmmfp->parser = read_asc20hmm;
        return hmmfp;
    }
    else if (strncmp ("HMMER v1.9", buf, 10) == 0) {
        hmmfp->parser = read_asc19hmm;
        return hmmfp;
    }
    else if (strncmp ("# HMM v1.7", buf, 10) == 0) {
        hmmfp->parser = read_asc17hmm;
        return hmmfp;
    }
    else if (strncmp ("# HMM v1.1", buf, 10) == 0) {
        hmmfp->parser = read_asc11hmm;
        return hmmfp;
    }
    else if (strncmp ("# HMM v1.0", buf, 10) == 0) {
        hmmfp->parser = read_asc10hmm;
        return hmmfp;
    }

    /* If we haven't recognized it yet, it's bogus.
     */
    HMMFileClose (hmmfp);
    return NULL;
}

int
HMMFileRead (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    int status;
    /* Set the disk position marker. */
    if (hmmfp->is_seekable) {
        status = SSIGetFilePosition (hmmfp->f, hmmfp->mode, &(hmmfp->offset));
        if (status != 0) {
            Die ("SSIGetFilePosition() failed");
        }
    }
    /* Parse the HMM and return it. */
    return (*hmmfp->parser) (hmmfp, ret_hmm);
}

void
HMMFileClose (HMMFILE * hmmfp)
{
    if (hmmfp->f != NULL) {
        fclose (hmmfp->f);
    }
    if (hmmfp->ssi != NULL) {
        SSIClose (hmmfp->ssi);
    }
    free (hmmfp);
}

void
HMMFileRewind (HMMFILE * hmmfp)
{
    rewind (hmmfp->f);
}

int
HMMFilePositionByName (HMMFILE * hmmfp, char *name)
{
    SSIOFFSET offset;		/* offset in hmmfile, from SSI */
    int fh;			/* ignored.                    */

    if (hmmfp->ssi == NULL) {
        return 0;
    }
    if (SSIGetOffsetByName (hmmfp->ssi, name, &fh, &offset) != 0) {
        return 0;
    }
    if (SSISetFilePosition (hmmfp->f, &offset) != 0) {
        return 0;
    }
    return 1;
}

int
HMMFilePositionByIndex (HMMFILE * hmmfp, int idx)
{   /* idx runs from 0..nhmm-1 */
    int fh;			/* file handle is ignored; only one HMM file */
    SSIOFFSET offset;		/* file position of HMM */

    if (hmmfp->ssi == NULL) {
        return 0;
    }
    if (SSIGetOffsetByNumber (hmmfp->ssi, idx, &fh, &offset) != 0) {
        return 0;
    }
    if (SSISetFilePosition (hmmfp->f, &offset) != 0) {
        return 0;
    }
    return 1;
}

/*****************************************************************
 * HMM output API:
 *    WriteAscHMM()
 *    WriteBinHMM()
 *
 *****************************************************************/

/* Function: WriteAscHMM()
 *
 * Purpose:  Save an HMM in flat text ASCII format.
 *
 * Args:     fp        - open file for writing
 *           hmm       - HMM to save
 */
void
WriteAscHMM (FILE * fp, struct plan7_s *hmm)
{
    int k;			/* counter for nodes             */
    int x;			/* counter for symbols           */
    int ts;			/* counter for state transitions */

    fprintf (fp, "HMMER2.0  [%s]\n", PACKAGE_VERSION);	/* magic header */

    /* write header information
     */
    fprintf (fp, "NAME  %s\n", hmm->name);
    if (hmm->flags & PLAN7_ACC) {
        fprintf (fp, "ACC   %s\n", hmm->acc);
    }
    if (hmm->flags & PLAN7_DESC) {
        fprintf (fp, "DESC  %s\n", hmm->desc);
    }
    fprintf (fp, "LENG  %d\n", hmm->M);
    fprintf (fp, "ALPH  %s\n",
             (Alphabet_type == hmmAMINO) ? "Amino" : "Nucleic");
    fprintf (fp, "RF    %s\n", (hmm->flags & PLAN7_RF) ? "yes" : "no");
    fprintf (fp, "CS    %s\n", (hmm->flags & PLAN7_CS) ? "yes" : "no");
    fprintf (fp, "MAP   %s\n", (hmm->flags & PLAN7_MAP) ? "yes" : "no");
    multiline (fp, "COM   ", hmm->comlog);
    fprintf (fp, "NSEQ  %d\n", hmm->nseq);
    fprintf (fp, "DATE  %s\n", hmm->ctime);
    fprintf (fp, "CKSUM %d\n", hmm->checksum);
    if (hmm->flags & PLAN7_GA) {
        fprintf (fp, "GA    %.1f %.1f\n", hmm->ga1, hmm->ga2);
    }
    if (hmm->flags & PLAN7_TC) {
        fprintf (fp, "TC    %.1f %.1f\n", hmm->tc1, hmm->tc2);
    }
    if (hmm->flags & PLAN7_NC) {
        fprintf (fp, "NC    %.1f %.1f\n", hmm->nc1, hmm->nc2);
    }

    /* Specials
     */
    fputs ("XT     ", fp);
    for (k = 0; k < 4; k++) {
        for (x = 0; x < 2; x++) {
            fprintf (fp, "%6s ", prob2ascii (hmm->xt[k][x], 1.0));
        }
    }

    fputs ("\n", fp);

    /* Save the null model first, so HMM readers can decode
     * log odds scores on the fly. Save as log odds probabilities
     * relative to 1/Alphabet_size (flat distribution)
     */
    fprintf (fp, "NULT  ");
    fprintf (fp, "%6s ", prob2ascii (hmm->p1, 1.0));	/* p1 */
    fprintf (fp, "%6s\n", prob2ascii (1.0 - hmm->p1, 1.0));	/* p2 */
    fputs ("NULE  ", fp);
    for (x = 0; x < Alphabet_size; x++) {
        fprintf (fp, "%6s ", prob2ascii (hmm->null[x], 1 / (float) (Alphabet_size)));
    }
    fputs ("\n", fp);

    /* EVD statistics
     */
    if (hmm->flags & PLAN7_STATS) {
        fprintf (fp, "EVD   %10f %10f\n", hmm->mu, hmm->lambda);
    }

    /* Print header
     */
    fprintf (fp, "HMM      ");
    for (x = 0; x < Alphabet_size; x++) {
        fprintf (fp, "  %c    ", Alphabet[x]);
    }
    fprintf (fp, "\n");
    fprintf (fp, "       %6s %6s %6s %6s %6s %6s %6s %6s %6s\n",
             "m->m", "m->i", "m->d", "i->m", "i->i", "d->m", "d->d", "b->m",
             "m->e");

    /* Print HMM parameters (main section of the save file)
     */
    fprintf (fp, "       %6s %6s ", prob2ascii (1 - hmm->tbd1, 1.0), "*");
    fprintf (fp, "%6s\n", prob2ascii (hmm->tbd1, 1.0));
    for (k = 1; k <= hmm->M; k++) {
        /* Line 1: k, match emissions, map */
        fprintf (fp, " %5d ", k);
        for (x = 0; x < Alphabet_size; x++) {
            fprintf (fp, "%6s ", prob2ascii (hmm->mat[k][x], hmm->null[x]));
        }
        if (hmm->flags & PLAN7_MAP) {
            fprintf (fp, "%5d", hmm->map[k]);
        }
        fputs ("\n", fp);
        /* Line 2: RF and insert emissions */
        fprintf (fp, " %5c ", hmm->flags & PLAN7_RF ? hmm->rf[k] : '-');
        for (x = 0; x < Alphabet_size; x++) {
            fprintf (fp, "%6s ",
                     (k < hmm->M) ? prob2ascii (hmm->ins[k][x],
                                                hmm->null[x]) : "*");
        }
        fputs ("\n", fp);
        /* Line 3: CS and transition probs */
        fprintf (fp, " %5c ", hmm->flags & PLAN7_CS ? hmm->cs[k] : '-');
        for (ts = 0; ts < 7; ts++) {
            fprintf (fp, "%6s ",
                     (k < hmm->M) ? prob2ascii (hmm->t[k][ts], 1.0) : "*");
        }
        fprintf (fp, "%6s ", prob2ascii (hmm->begin[k], 1.0));
        fprintf (fp, "%6s ", prob2ascii (hmm->end[k], 1.0));

        fputs ("\n", fp);
    }
    fputs ("//\n", fp);
}

/* Function: WriteBinHMM()
 *
 * Purpose:  Write an HMM in binary format.
 */
void
WriteBinHMM (FILE * fp, struct plan7_s *hmm)
{
    int k;

    /* ye olde magic number */
    fwrite ((char *) &(v20magic), sizeof (unsigned int), 1, fp);

    /* header section
     */
    fwrite ((char *) &(hmm->flags), sizeof (int), 1, fp);
    write_bin_string (fp, hmm->name);
    if (hmm->flags & PLAN7_ACC) {
        write_bin_string (fp, hmm->acc);
    }
    if (hmm->flags & PLAN7_DESC) {
        write_bin_string (fp, hmm->desc);
    }
    fwrite ((char *) &(hmm->M), sizeof (int), 1, fp);
    fwrite ((char *) &(Alphabet_type), sizeof (int), 1, fp);
    if (hmm->flags & PLAN7_RF) {
        fwrite ((char *) hmm->rf, sizeof (char), hmm->M + 1, fp);
    }
    if (hmm->flags & PLAN7_CS) {
        fwrite ((char *) hmm->cs, sizeof (char), hmm->M + 1, fp);
    }
    if (hmm->flags & PLAN7_MAP) {
        fwrite ((char *) hmm->map, sizeof (int), hmm->M + 1, fp);
    }
    write_bin_string (fp, hmm->comlog);
    fwrite ((char *) &(hmm->nseq), sizeof (int), 1, fp);
    write_bin_string (fp, hmm->ctime);
    fwrite ((char *) &(hmm->checksum), sizeof (int), 1, fp);
    if (hmm->flags & PLAN7_GA) {
        fwrite ((char *) &(hmm->ga1), sizeof (float), 1, fp);
        fwrite ((char *) &(hmm->ga2), sizeof (float), 1, fp);
    }
    if (hmm->flags & PLAN7_TC) {
        fwrite ((char *) &(hmm->tc1), sizeof (float), 1, fp);
        fwrite ((char *) &(hmm->tc2), sizeof (float), 1, fp);
    }
    if (hmm->flags & PLAN7_NC) {
        fwrite ((char *) &(hmm->nc1), sizeof (float), 1, fp);
        fwrite ((char *) &(hmm->nc2), sizeof (float), 1, fp);
    }

    /* Specials */
    for (k = 0; k < 4; k++) {
        fwrite ((char *) hmm->xt[k], sizeof (float), 2, fp);
    }
    /* Null model */
    fwrite ((char *) &(hmm->p1), sizeof (float), 1, fp);
    fwrite ((char *) hmm->null, sizeof (float), Alphabet_size, fp);

    /* EVD stats */
    if (hmm->flags & PLAN7_STATS) {
        fwrite ((char *) &(hmm->mu), sizeof (float), 1, fp);
        fwrite ((char *) &(hmm->lambda), sizeof (float), 1, fp);
    }

    /* entry/exit probabilities
     */
    fwrite ((char *) &(hmm->tbd1), sizeof (float), 1, fp);
    fwrite ((char *) hmm->begin, sizeof (float), hmm->M + 1, fp);
    fwrite ((char *) hmm->end, sizeof (float), hmm->M + 1, fp);

    /* main model
     */
    for (k = 1; k <= hmm->M; k++) {
        fwrite ((char *) hmm->mat[k], sizeof (float), Alphabet_size, fp);
    }
    for (k = 1; k < hmm->M; k++) {
        fwrite ((char *) hmm->ins[k], sizeof (float), Alphabet_size, fp);
    }
    for (k = 1; k < hmm->M; k++) {
        fwrite ((char *) hmm->t[k], sizeof (float), 7, fp);
    }
}


/*****************************************************************
 *
 * Internal: HMM file parsers for various releases of HMMER.
 *
 * read_{asc,bin}xxhmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
 *
 * Upon return, *ret_hmm is an allocated Plan7 HMM.
 * Return 0 if no more HMMs in the file (normal).
 * Return 1 and *ret_hmm = something if we got an HMM (normal)
 * Return 1 if an error occurs (meaning "I tried to
 *   read something...") and *ret_hmm == NULL (meaning
 *   "...but it wasn't an HMM"). I know, this is a funny
 *   way to handle errors.
 *
 *****************************************************************/

static int
read_asc20hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    struct plan7_s *hmm;
    char buffer[512];
    char *s;
    int M;
    float p;
    int k, x;
    int atype;			/* alphabet type, hmmAMINO or hmmNUCLEIC */

    hmm = NULL;
    if (feof (hmmfp->f) || fgets (buffer, 512, hmmfp->f) == NULL) {
        return 0;
    }
    if (strncmp (buffer, "HMMER2.0", 8) != 0) {
        goto FAILURE;
    }

    /* Get the header information: tag/value pairs in any order,
     * ignore unknown tags, stop when "HMM" is reached (signaling
     * start of main model)
     */
    hmm = AllocPlan7Shell ();
    M = -1;
    while (fgets (buffer, 512, hmmfp->f) != NULL) {
        if (strncmp (buffer, "NAME ", 5) == 0) {
            Plan7SetName (hmm, buffer + 6);
        }
        else if (strncmp (buffer, "ACC  ", 5) == 0) {
            Plan7SetAccession (hmm, buffer + 6);
        }
        else if (strncmp (buffer, "DESC ", 5) == 0) {
            Plan7SetDescription (hmm, buffer + 6);
        }
        else if (strncmp (buffer, "LENG ", 5) == 0) {
            M = atoi (buffer + 6);
        }
        else if (strncmp (buffer, "NSEQ ", 5) == 0) {
            hmm->nseq = atoi (buffer + 6);
        }
        else if (strncmp (buffer, "ALPH ", 5) == 0) {	/* Alphabet type */
            s2upper (buffer + 6);
            if (strncmp (buffer + 6, "AMINO", 5) == 0) {
                atype = hmmAMINO;
            }
            else if (strncmp (buffer + 6, "NUCLEIC", 7) == 0) {
                atype = hmmNUCLEIC;
            }
            else {
                goto FAILURE;
            }

            if (Alphabet_type == hmmNOTSETYET) {
                SetAlphabet (atype);
            }
            else if (atype != Alphabet_type)
                Die
                ("Alphabet mismatch error.\nI thought we were working with %s, but tried to read a %s HMM.\n",
                 AlphabetType2String (Alphabet_type), AlphabetType2String (atype));
        }
        else if (strncmp (buffer, "RF   ", 5) == 0) {	/* Reference annotation present? */
            if (sre_toupper (*(buffer + 6)) == 'Y') {
                hmm->flags |= PLAN7_RF;
            }
        }
        else if (strncmp (buffer, "CS   ", 5) == 0) {	/* Consensus annotation present? */
            if (sre_toupper (*(buffer + 6)) == 'Y') {
                hmm->flags |= PLAN7_CS;
            }
        }
        else if (strncmp (buffer, "MAP  ", 5) == 0) {	/* Map annotation present? */
            if (sre_toupper (*(buffer + 6)) == 'Y') {
                hmm->flags |= PLAN7_MAP;
            }
        }
        else if (strncmp (buffer, "COM  ", 5) == 0) {	/* Command line log */
            StringChop (buffer + 6);
            if (hmm->comlog == NULL) {
                hmm->comlog = Strdup (buffer + 6);
            }
            else {
                hmm->comlog = ReallocOrDie (hmm->comlog, sizeof (char *) *
                                            (strlen (hmm->comlog) + 1 +
                                             strlen (buffer + 6)));
                strcat (hmm->comlog, "\n");
                strcat (hmm->comlog, buffer + 6);
            }
        }
        else if (strncmp (buffer, "DATE ", 5) == 0) {	/* Date file created */
            StringChop (buffer + 6);
            hmm->ctime = Strdup (buffer + 6);
        }
        else if (strncmp (buffer, "GA   ", 5) == 0) {
            if ((s = strtok (buffer + 6, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->ga1 = atof (s);
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->ga2 = atof (s);
            hmm->flags |= PLAN7_GA;
        }
        else if (strncmp (buffer, "TC   ", 5) == 0) {
            if ((s = strtok (buffer + 6, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->tc1 = atof (s);
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->tc2 = atof (s);
            hmm->flags |= PLAN7_TC;
        }
        else if (strncmp (buffer, "NC   ", 5) == 0) {
            if ((s = strtok (buffer + 6, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->nc1 = atof (s);
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->nc2 = atof (s);
            hmm->flags |= PLAN7_NC;
        }
        else if (strncmp (buffer, "XT   ", 5) == 0) {	/* Special transition section */
            if ((s = strtok (buffer + 6, " \t\n")) == NULL) {
                goto FAILURE;
            }
            for (k = 0; k < 4; k++) {
                for (x = 0; x < 2; x++) {
                    if (s == NULL) {
                        goto FAILURE;
                    }
                    hmm->xt[k][x] = ascii2prob (s, 1.0);
                    s = strtok (NULL, " \t\n");
                }
            }
        }
        else if (strncmp (buffer, "NULT ", 5) == 0) {	/* Null model transitions */
            if ((s = strtok (buffer + 6, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->p1 = ascii2prob (s, 1.);
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->p1 = hmm->p1 / (hmm->p1 + ascii2prob (s, 1.0));
        }
        else if (strncmp (buffer, "NULE ", 5) == 0) {	/* Null model emissions */
            if (Alphabet_type == hmmNOTSETYET) {
                Die ("ALPH must precede NULE in HMM save files");
            }
            s = strtok (buffer + 6, " \t\n");
            for (x = 0; x < Alphabet_size; x++) {
                if (s == NULL) {
                    goto FAILURE;
                }
                hmm->null[x] = ascii2prob (s, 1. / (float) Alphabet_size);
                s = strtok (NULL, " \t\n");
            }
        }
        else if (strncmp (buffer, "EVD  ", 5) == 0) {	/* EVD parameters */
            hmm->flags |= PLAN7_STATS;
            if ((s = strtok (buffer + 6, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->mu = atof (s);
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->lambda = atof (s);
        }
        else if (strncmp (buffer, "CKSUM", 5) == 0) {
            hmm->checksum = atoi (buffer + 6);
        }
        else if (strncmp (buffer, "HMM  ", 5) == 0) {
            break;
        }
    }

    /* partial check for mandatory fields */
    if (feof (hmmfp->f)) {
        goto FAILURE;
    }
    if (M < 1) {
        goto FAILURE;
    }
    if (hmm->name == NULL) {
        goto FAILURE;
    }
    if (Alphabet_type == hmmNOTSETYET) {
        goto FAILURE;
    }

    /* Main model section. Read as integer log odds, convert
     * to probabilities
     */
    AllocPlan7Body (hmm, M);
    /* skip an annotation line */
    if (fgets (buffer, 512, hmmfp->f) == NULL) {
        goto FAILURE;
    }
    /* parse tbd1 line */
    if (fgets (buffer, 512, hmmfp->f) == NULL) {
        goto FAILURE;
    }
    if ((s = strtok (buffer, " \t\n")) == NULL) {
        goto FAILURE;
    }
    p = ascii2prob (s, 1.0);
    if ((s = strtok (NULL, " \t\n")) == NULL) {
        goto FAILURE;
    }
    if ((s = strtok (NULL, " \t\n")) == NULL) {
        goto FAILURE;
    }
    hmm->tbd1 = ascii2prob (s, 1.0);
    hmm->tbd1 = hmm->tbd1 / (p + hmm->tbd1);

    /* main model */
    for (k = 1; k <= hmm->M; k++) {
        /* Line 1: k, match emissions, map */
        if (fgets (buffer, 512, hmmfp->f) == NULL) {
            goto FAILURE;
        }
        if ((s = strtok (buffer, " \t\n")) == NULL) {
            goto FAILURE;
        }
        if (atoi (s) != k) {
            goto FAILURE;
        }
        for (x = 0; x < Alphabet_size; x++) {
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->mat[k][x] = ascii2prob (s, hmm->null[x]);
        }
        if (hmm->flags & PLAN7_MAP) {
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            hmm->map[k] = atoi (s);
        }
        /* Line 2:  RF and insert emissions */
        if (fgets (buffer, 512, hmmfp->f) == NULL) {
            goto FAILURE;
        }
        if ((s = strtok (buffer, " \t\n")) == NULL) {
            goto FAILURE;
        }
        if (hmm->flags & PLAN7_RF) {
            hmm->rf[k] = *s;
        }
        if (k < hmm->M) {
            for (x = 0; x < Alphabet_size; x++) {
                if ((s = strtok (NULL, " \t\n")) == NULL) {
                    goto FAILURE;
                }
                hmm->ins[k][x] = ascii2prob (s, hmm->null[x]);
            }
        }
        /* Line 3: CS and transitions */
        if (fgets (buffer, 512, hmmfp->f) == NULL) {
            goto FAILURE;
        }
        if ((s = strtok (buffer, " \t\n")) == NULL) {
            goto FAILURE;
        }
        if (hmm->flags & PLAN7_CS) {
            hmm->cs[k] = *s;
        }
        for (x = 0; x < 7; x++) {
            if ((s = strtok (NULL, " \t\n")) == NULL) {
                goto FAILURE;
            }
            if (k < hmm->M) {
                hmm->t[k][x] = ascii2prob (s, 1.0);
            }
        }
        if ((s = strtok (NULL, " \t\n")) == NULL) {
            goto FAILURE;
        }
        hmm->begin[k] = ascii2prob (s, 1.0);
        if ((s = strtok (NULL, " \t\n")) == NULL) {
            goto FAILURE;
        }
        hmm->end[k] = ascii2prob (s, 1.0);

    }				/* end loop over main model */

    /* Advance to record separator
     */
    while (fgets (buffer, 512, hmmfp->f) != NULL)
        if (strncmp (buffer, "//", 2) == 0) {
            break;
        }

    Plan7Renormalize (hmm);	/* Paracel reported bug 6/11/99 */

    /* Set flags and return
     */
    hmm->flags |= PLAN7_HASPROB;	/* probabilities are valid */
    hmm->flags &= ~PLAN7_HASBITS;	/* scores are not valid    */

    *ret_hmm = hmm;
    return 1;

FAILURE:
    if (hmm != NULL) {
        FreePlan7 (hmm);
    }
    *ret_hmm = NULL;
    return 1;
}


static int
read_bin20hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    struct plan7_s *hmm;
    int k, x;
    int type;
    unsigned int magic;

    hmm = NULL;

    /* Header section
     */
    if (feof (hmmfp->f)) {
        return 0;
    }
    if (!fread ((char *) &magic, sizeof (unsigned int), 1, hmmfp->f)) {
        return 0;
    }

    if (hmmfp->byteswap) {
        byteswap ((char *) &magic, sizeof (unsigned int));
    }
    if (magic != v20magic) {
        goto FAILURE;
    }
    /* allocate HMM shell for header info */
    hmm = AllocPlan7Shell ();
    /* flags */
    if (!fread ((char *) &(hmm->flags), sizeof (int), 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (hmmfp->byteswap) {
        byteswap ((char *) &(hmm->flags), sizeof (int));
    }
    /* name */
    if (!read_bin_string (hmmfp->f, hmmfp->byteswap, &(hmm->name))) {
        goto FAILURE;
    }

    /* optional accession */
    if ((hmm->flags & PLAN7_ACC) &&
            !read_bin_string (hmmfp->f, hmmfp->byteswap, &(hmm->acc))) {
        goto FAILURE;
    }
    /* optional description */
    if ((hmm->flags & PLAN7_DESC) &&
            !read_bin_string (hmmfp->f, hmmfp->byteswap, &(hmm->desc))) {
        goto FAILURE;
    }
    /* length of model */
    if (!fread ((char *) &hmm->M, sizeof (int), 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (hmmfp->byteswap) {
        byteswap ((char *) &(hmm->M), sizeof (int));
    }
    /* alphabet type */
    if (!fread ((char *) &type, sizeof (int), 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (hmmfp->byteswap) {
        byteswap ((char *) &type, sizeof (int));
    }
    if (Alphabet_type == hmmNOTSETYET) {
        SetAlphabet (type);
    }
    else if (type != Alphabet_type)
        Die
        ("Alphabet mismatch error.\nI thought we were working with %s, but tried to read a %s HMM.\n",
         AlphabetType2String (Alphabet_type), AlphabetType2String (type));

    /* now allocate for rest of model */
    AllocPlan7Body (hmm, hmm->M);

    /* optional #=RF alignment annotation */
    if ((hmm->flags & PLAN7_RF) &&
            !fread ((char *) hmm->rf, sizeof (char), hmm->M + 1, hmmfp->f)) {
        goto FAILURE;
    }
    hmm->rf[hmm->M + 1] = '\0';
    /* optional #=CS alignment annotation */
    if ((hmm->flags & PLAN7_CS) &&
            !fread ((char *) hmm->cs, sizeof (char), hmm->M + 1, hmmfp->f)) {
        goto FAILURE;
    }
    hmm->cs[hmm->M + 1] = '\0';
    /* optional alignment map annotation */
    if ((hmm->flags & PLAN7_MAP) &&
            !fread ((char *) hmm->map, sizeof (int), hmm->M + 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (hmmfp->byteswap)
        for (k = 1; k <= hmm->M; k++) {
            byteswap ((char *) &(hmm->map[k]), sizeof (int));
        }
    /* command line log */
    if (!read_bin_string (hmmfp->f, hmmfp->byteswap, &(hmm->comlog))) {
        goto FAILURE;
    }
    /* nseq */
    if (!fread ((char *) &(hmm->nseq), sizeof (int), 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (hmmfp->byteswap) {
        byteswap ((char *) &(hmm->nseq), sizeof (int));
    }
    /* creation time */
    if (!read_bin_string (hmmfp->f, hmmfp->byteswap, &(hmm->ctime))) {
        goto FAILURE;
    }
    /* checksum */
    if (!fread ((char *) &(hmm->checksum), sizeof (int), 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (hmmfp->byteswap) {
        byteswap ((char *) &(hmm->checksum), sizeof (int));
    }

    /* Pfam gathering thresholds */
    if (hmm->flags & PLAN7_GA) {
        if (!fread ((char *) &(hmm->ga1), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }
        if (!fread ((char *) &(hmm->ga2), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }
        if (hmmfp->byteswap) {
            byteswap ((char *) &(hmm->ga1), sizeof (float));
            byteswap ((char *) &(hmm->ga2), sizeof (float));
        }
    }
    /* Pfam trusted cutoffs */
    if (hmm->flags & PLAN7_TC) {
        if (!fread ((char *) &(hmm->tc1), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }
        if (!fread ((char *) &(hmm->tc2), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }
        if (hmmfp->byteswap) {
            byteswap ((char *) &(hmm->tc1), sizeof (float));
            byteswap ((char *) &(hmm->tc2), sizeof (float));
        }
    }
    /* Pfam noise cutoffs */
    if (hmm->flags & PLAN7_NC) {
        if (!fread ((char *) &(hmm->nc1), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }
        if (!fread ((char *) &(hmm->nc2), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }
        if (hmmfp->byteswap) {
            byteswap ((char *) &(hmm->nc1), sizeof (float));
            byteswap ((char *) &(hmm->nc2), sizeof (float));
        }
    }

    /* specials */
    for (k = 0; k < 4; k++) {
        if (!fread ((char *) hmm->xt[k], sizeof (float), 2, hmmfp->f)) {
            goto FAILURE;
        }
        if (hmmfp->byteswap) {
            for (x = 0; x < 2; x++) {
                byteswap ((char *) &(hmm->xt[k][x]), sizeof (float));
            }
        }
    }

    /* null model */
    if (!fread ((char *) &(hmm->p1), sizeof (float), 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (!fread ((char *) hmm->null, sizeof (float), Alphabet_size, hmmfp->f)) {
        goto FAILURE;
    }

    /* EVD stats */
    if (hmm->flags & PLAN7_STATS) {
        if (!fread ((char *) &(hmm->mu), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }
        if (!fread ((char *) &(hmm->lambda), sizeof (float), 1, hmmfp->f)) {
            goto FAILURE;
        }

        if (hmmfp->byteswap) {
            byteswap ((char *) &(hmm->mu), sizeof (float));
            byteswap ((char *) &(hmm->lambda), sizeof (float));
        }
    }

    /* entry/exit probabilities
     */
    if (!fread ((char *) &(hmm->tbd1), sizeof (float), 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (!fread ((char *) hmm->begin, sizeof (float), hmm->M + 1, hmmfp->f)) {
        goto FAILURE;
    }
    if (!fread ((char *) hmm->end, sizeof (float), hmm->M + 1, hmmfp->f)) {
        goto FAILURE;
    }

    /* main model */
    for (k = 1; k <= hmm->M; k++) {
        if (!fread ((char *) hmm->mat[k], sizeof (float), Alphabet_size, hmmfp->f)) {
            goto FAILURE;
        }
    }
    for (k = 1; k < hmm->M; k++) {
        if (!fread ((char *) hmm->ins[k], sizeof (float), Alphabet_size, hmmfp->f)) {
            goto FAILURE;
        }
    }
    for (k = 1; k < hmm->M; k++) {
        if (!fread ((char *) hmm->t[k], sizeof (float), 7, hmmfp->f)) {
            goto FAILURE;
        }
    }
    /* byteswapping
     */
    if (hmmfp->byteswap) {
        for (x = 0; x < Alphabet_size; x++) {
            byteswap ((char *) &(hmm->null[x]), sizeof (float));
        }
        byteswap ((char *) &(hmm->p1), sizeof (float));
        byteswap ((char *) &(hmm->tbd1), sizeof (float));

        for (k = 1; k <= hmm->M; k++) {
            for (x = 0; x < Alphabet_size; x++) {
                byteswap ((char *) &(hmm->mat[k][x]), sizeof (float));
            }
            if (k < hmm->M)
                for (x = 0; x < Alphabet_size; x++) {
                    byteswap ((char *) &(hmm->ins[k][x]), sizeof (float));
                }
            byteswap ((char *) &(hmm->begin[k]), sizeof (float));
            byteswap ((char *) &(hmm->end[k]), sizeof (float));
            if (k < hmm->M)
                for (x = 0; x < 7; x++) {
                    byteswap ((char *) &(hmm->t[k][x]), sizeof (float));
                }
        }
    }


    /* set flags and return
     */
    hmm->flags |= PLAN7_HASPROB;	/* probabilities are valid  */
    hmm->flags &= ~PLAN7_HASBITS;	/* scores are not yet valid */
    *ret_hmm = hmm;
    return 1;

FAILURE:
    if (hmm != NULL) {
        FreePlan7 (hmm);
    }
    *ret_hmm = NULL;
    return 1;
}





/* Function: read_asc19hmm()
 * Date:     Tue Apr  7 17:11:29 1998 [StL]
 *
 * Purpose:  Read ASCII-format tabular (1.9 and later) save files.
 *
 *           HMMER 1.9 was only used internally at WashU, as far as
 *           I know, so this code shouldn't be terribly important
 *           to anyone.
 */
static int
read_asc19hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    struct plan7_s *hmm;
    FILE *fp;
    char buffer[512];
    char *s;
    int M;			/* length of model  */
    int k;			/* state number  */
    int x;			/* symbol number */
    int atype;			/* Alphabet type */

    hmm = NULL;
    fp = hmmfp->f;
    if (feof (fp) || fgets (buffer, 512, fp) == NULL) {
        return 0;
    }
    if (strncmp (buffer, "HMMER v1.9", 10) != 0) {
        goto FAILURE;
    }

    hmm = AllocPlan7Shell ();
    /* read M from first line */
    if ((s = Getword (fp, sqdARG_INT)) == NULL) {
        goto FAILURE;
    }
    M = atoi (s);			/* model length */
    if ((s = Getword (fp, sqdARG_INT)) == NULL) {
        goto FAILURE;    /* ignore alphabet size */
    }
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }
    Plan7SetName (hmm, s);	/* name */
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;    /* alphabet type */
    }
    s2upper (s);
    if (strcmp (s, "AMINO") == 0) {
        atype = hmmAMINO;
    }
    else if (strcmp (s, "NUCLEIC") == 0) {
        atype = hmmNUCLEIC;
    }
    else {
        goto FAILURE;
    }

    if (Alphabet_type == hmmNOTSETYET) {
        SetAlphabet (atype);
    }
    else if (atype != Alphabet_type)
        Die
        ("Alphabet mismatch error.\nI thought we were working with %s, but tried to read a %s HMM.\n",
         AlphabetType2String (Alphabet_type), AlphabetType2String (atype));

    /* read alphabet, make sure it's Plan7-compatible... */
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }
    if (strncmp (s, Alphabet, Alphabet_size) != 0) {
        goto FAILURE;
    }

    /* whether we have ref, cs info */
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }
    if (strcmp (s, "yes") == 0) {
        hmm->flags |= PLAN7_RF;
    }
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }
    if (strcmp (s, "yes") == 0) {
        hmm->flags |= PLAN7_CS;
    }

    /* null model. 1.9 has emissions only. invent transitions. */
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }
    if (strcmp (s, "null") != 0) {
        goto FAILURE;
    }
    for (x = 0; x < Alphabet_size; x++) {
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        hmm->null[x] = ascii2prob (s, 1.0);
    }
    hmm->p1 = (Alphabet_type == hmmAMINO) ? 350. / 351. : 1000. / 1001.;

    /* Done with header; check some stuff before proceeding
     */
    if (feof (hmmfp->f)) {
        goto FAILURE;
    }
    if (M < 1) {
        goto FAILURE;
    }
    if (hmm->name == NULL) {
        goto FAILURE;
    }
    if (Alphabet_type == hmmNOTSETYET) {
        goto FAILURE;
    }

    /* Allocate the model. Set up the probabilities that Plan9
     * doesn't set.
     */
    AllocPlan7Body (hmm, M);
    ZeroPlan7 (hmm);
    Plan7LSConfig (hmm);

    /* The zero row has: 4 or 20 unused scores for nonexistent M0 state
     * then: B->M, tbd1, a B->I that Plan7 doesn't have;
     *       three unused D-> transitions; then three I0 transitions that Plan7 doesn't have;
     *       then two unused rf, cs annotations.
     */
    if ((s = Getword (fp, sqdARG_INT)) == NULL) {
        goto FAILURE;    /* position index ignored */
    }
    for (x = 0; x < Alphabet_size; x++) {
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;    /* emissions ignored */
        }
    }
    if ((s = Getword (fp, sqdARG_INT)) == NULL) {
        goto FAILURE;
    }
    hmm->begin[1] = ascii2prob (s, 1.0);
    if ((s = Getword (fp, sqdARG_INT)) == NULL) {
        goto FAILURE;
    }
    hmm->tbd1 = ascii2prob (s, 1.0);
    /* renormalize */
    hmm->begin[1] = hmm->begin[1] / (hmm->begin[1] + hmm->tbd1);
    hmm->tbd1 = hmm->tbd1 / (hmm->begin[1] + hmm->tbd1);
    /* skip rest of line, seven integer fields, two char fields */
    for (x = 0; x < 7; x++) {
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
    }
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }

    /* main model: table of emissions, transitions, annotation */
    for (k = 1; k <= hmm->M; k++) {
        /* position index ignored */
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        /* match emissions */
        for (x = 0; x < Alphabet_size; x++) {
            if ((s = Getword (fp, sqdARG_INT)) == NULL) {
                goto FAILURE;
            }
            hmm->mat[k][x] = ascii2prob (s, hmm->null[x]);
        }
        /* nine transitions; two are ignored */
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        if (k < hmm->M) {
            hmm->t[k][TMM] = ascii2prob (s, 1.0);
        }
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        if (k < hmm->M) {
            hmm->t[k][TMD] = (k == hmm->M) ? 0.0 : ascii2prob (s, 1.0);
        }
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        if (k < hmm->M) {
            hmm->t[k][TMI] = ascii2prob (s, 1.0);
        }

        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        if (k < hmm->M) {
            hmm->t[k][TDM] = ascii2prob (s, 1.0);
        }
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        if (k < hmm->M) {
            hmm->t[k][TDD] = (k == hmm->M) ? 0.0 : ascii2prob (s, 1.0);
        }
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;    /* TDI ignored. */
        }

        /* no insert state at k == M, be careful */
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        if (k < hmm->M) {
            hmm->t[k][TIM] = ascii2prob (s, 1.0);
        }
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;    /* TID ignored. */
        }
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;
        }
        if (k < hmm->M) {
            hmm->t[k][TII] = ascii2prob (s, 1.0);
        }

        /* annotations */
        if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
            goto FAILURE;
        }
        if (hmm->flags & PLAN7_RF) {
            hmm->rf[k] = *s;
        }
        if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
            goto FAILURE;
        }
        if (hmm->flags & PLAN7_CS) {
            hmm->cs[k] = *s;
        }
    }
    /* table of insert emissions;
     * Plan7 has no insert state at 0 or M  */
    for (k = 0; k <= hmm->M; k++) {
        if ((s = Getword (fp, sqdARG_INT)) == NULL) {
            goto FAILURE;    /* position index ignored */
        }
        for (x = 0; x < Alphabet_size; x++) {
            if ((s = Getword (fp, sqdARG_INT)) == NULL) {
                goto FAILURE;
            }
            if (k > 0 && k < hmm->M) {
                hmm->ins[k][x] = ascii2prob (s, hmm->null[x]);
            }
        }
    }

    /* Set flags and return
     */
    hmm->flags |= PLAN7_HASPROB;	/* probabilities are valid */
    hmm->flags &= ~PLAN7_HASBITS;	/* scores are not valid    */
    Plan7Renormalize (hmm);
    hmm->comlog = Strdup ("[converted from an old Plan9 HMM]");
    Plan7SetCtime (hmm);
    *ret_hmm = hmm;
    return 1;

FAILURE:
    if (hmm != NULL) {
        FreePlan7 (hmm);
    }
    *ret_hmm = NULL;
    return 1;
}

static int
read_bin19hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    unsigned int magic;
    struct plan7_s *hmm;		/* plan7 HMM */
    struct plan9_s *p9hmm;	/* old style 1.x HMM */

    /* Read the magic number; if we don't see it, then we
     * must be out of data in the file.
     */
    if (feof (hmmfp->f)) {
        return 0;
    }
    if (!fread ((char *) &magic, sizeof (unsigned int), 1, hmmfp->f)) {
        return 0;
    }

    p9hmm = read_plan9_binhmm (hmmfp->f, HMMER1_9B, hmmfp->byteswap);
    if (p9hmm == NULL) {
        *ret_hmm = NULL;
        return 1;
    }

    Plan9toPlan7 (p9hmm, &hmm);

    hmm->comlog = Strdup ("[converted from an old Plan9 HMM]");
    Plan7SetCtime (hmm);

    P9FreeHMM (p9hmm);
    *ret_hmm = hmm;
    return 1;
}

static int
read_asc17hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    struct plan7_s *hmm;		/* plan7 HMM */
    struct plan9_s *p9hmm;	/* old style 1.x HMM */
    char buffer[512];

    /* Read the magic header; if we don't see it, then
     * we must be out of data in the file.
     */
    if (feof (hmmfp->f) || fgets (buffer, 512, hmmfp->f) == NULL) {
        return 0;
    }

    p9hmm = read_plan9_aschmm (hmmfp->f, HMMER1_7F);
    if (p9hmm == NULL) {
        *ret_hmm = NULL;
        return 1;
    }

    Plan9toPlan7 (p9hmm, &hmm);

    hmm->comlog = Strdup ("[converted from an old Plan9 HMM]");
    Plan7SetCtime (hmm);

    P9FreeHMM (p9hmm);
    Plan7Renormalize (hmm);
    *ret_hmm = hmm;
    return 1;
}

static int
read_bin17hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    unsigned int magic;
    struct plan7_s *hmm;		/* plan7 HMM */
    struct plan9_s *p9hmm;	/* old style 1.x HMM */

    /* Read the magic number; if we don't see it, then we
     * must be out of data in the file.
     */
    if (feof (hmmfp->f)) {
        return 0;
    }
    if (!fread ((char *) &magic, sizeof (unsigned int), 1, hmmfp->f)) {
        return 0;
    }

    p9hmm = read_plan9_binhmm (hmmfp->f, HMMER1_7B, hmmfp->byteswap);
    if (p9hmm == NULL) {
        *ret_hmm = NULL;
        return 1;
    }

    Plan9toPlan7 (p9hmm, &hmm);

    hmm->comlog = Strdup ("[converted from an old Plan9 HMM]");
    Plan7SetCtime (hmm);

    P9FreeHMM (p9hmm);
    *ret_hmm = hmm;
    return 1;
}

static int
read_asc11hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    Die ("1.1 ASCII HMMs unsupported");
    return 1;
}

static int
read_bin11hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    unsigned int magic;
    struct plan7_s *hmm;		/* plan7 HMM */
    struct plan9_s *p9hmm;	/* old style 1.x HMM */

    /* Read the magic number; if we don't see it, then we
     * must be out of data in the file.
     */
    if (feof (hmmfp->f)) {
        return 0;
    }
    if (!fread ((char *) &magic, sizeof (unsigned int), 1, hmmfp->f)) {
        return 0;
    }

    p9hmm = read_plan9_binhmm (hmmfp->f, HMMER1_1B, hmmfp->byteswap);
    if (p9hmm == NULL) {
        *ret_hmm = NULL;
        return 1;
    }

    Plan9toPlan7 (p9hmm, &hmm);

    hmm->comlog = Strdup ("[converted from an old Plan9 HMM]");
    Plan7SetCtime (hmm);

    P9FreeHMM (p9hmm);
    *ret_hmm = hmm;
    return 1;
}

static int
read_asc10hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    Die ("1.0 ASCII HMMs unsupported");
    return 1;
}

static int
read_bin10hmm (HMMFILE * hmmfp, struct plan7_s **ret_hmm)
{
    unsigned int magic;
    struct plan7_s *hmm;		/* plan7 HMM */
    struct plan9_s *p9hmm;	/* old style 1.x HMM */

    /* Read the magic number; if we don't see it, then we
     * must be out of data in the file.
     */
    if (feof (hmmfp->f)) {
        return 0;
    }
    if (!fread ((char *) &magic, sizeof (unsigned int), 1, hmmfp->f)) {
        return 0;
    }

    p9hmm = read_plan9_binhmm (hmmfp->f, HMMER1_0B, hmmfp->byteswap);
    if (p9hmm == NULL) {
        *ret_hmm = NULL;
        return 1;
    }

    Plan9toPlan7 (p9hmm, &hmm);

    hmm->comlog = Strdup ("[converted from an old Plan9 HMM]");
    Plan7SetCtime (hmm);

    P9FreeHMM (p9hmm);
    *ret_hmm = hmm;
    return 1;
}

/*****************************************************************
 * Some miscellaneous utility functions
 *****************************************************************/

/* Function: prob2ascii()
 *
 * Purpose:  Format a probability for output to an ASCII save
 *           file. Returns a ptr to a static internal buffer.
 *
 */
static char *
prob2ascii (float p, float null)
{
    static char buffer[8];

    if (p == 0.0) {
        return "*";
    }
    sprintf (buffer, "%6d", Prob2Score (p, null));
    return buffer;
}


/* Function: ascii2prob()
 *
 * Purpose:  Convert a saved string back to a probability.
 */
static float
ascii2prob (char *s, float null)
{
    return (*s == '*') ? 0. : Score2Prob (atoi (s), null);
}

/* Function: byteswap()
 *
 * Purpose:  Swap between big-endian and little-endian.
 *           For example:
 *               int foo = 0x12345678;
 *               byteswap((char *) &foo, sizeof(int));
 *               printf("%x\n", foo)
 *           gives 78563412.
 *
 *           I don't fully understand byte-swapping issues.
 *           However, I have tested this on chars through floats,
 *           on various machines:
 *               SGI IRIX 4.0.5, SunOS 4.1.3, DEC Alpha OSF/1, Alliant
 *
 *           Note: this is only a partial solution to the problem of
 *           binary file portability. 32 bit integers are assumed by HMMER,
 *           for instance. This should be true for all UNIX, VAX, and WinNT
 *           platforms, I believe.
 *
 * Date: Sun Feb 12 10:26:22 1995
 */
static void
byteswap (char *swap, int nbytes)
{
    int x;
    char byte;

    for (x = 0; x < nbytes / 2; x++) {
        byte = swap[nbytes - x - 1];
        swap[nbytes - x - 1] = swap[x];
        swap[x] = byte;
    }
}

/* Function: write_bin_string()
 * Date:     SRE, Wed Oct 29 13:49:27 1997 [TWA 721 over Canada]
 *
 * Purpose:  Write a string in binary save format: an integer
 *           for the string length (including \0), followed by
 *           the string.
 */
static void
write_bin_string (FILE * fp, char *s)
{
    int len;
    if (s != NULL) {
        len = strlen (s) + 1;
        fwrite ((char *) &len, sizeof (int), 1, fp);
        fwrite ((char *) s, sizeof (char), len, fp);
    }
    else {
        len = 0;
        fwrite ((char *) &len, sizeof (int), 1, fp);
    }
}

/* Function: read_bin_string()
 * Date:     SRE, Wed Oct 29 14:03:23 1997 [TWA 721]
 *
 * Purpose:  Read in a string from a binary file, where
 *           the first integer is the length (including '\0').
 *
 * Args:     fp       - FILE to read from
 *           doswap   - TRUE to byteswap
 *           ret_s    - string to read into
 *
 * Return:   0 on failure. ret_s is malloc'ed here.
 */
static int
read_bin_string (FILE * fp, int doswap, char **ret_s)
{
    char *s;
    int len;

    if (!fread ((char *) &len, sizeof (int), 1, fp)) {
        return 0;
    }
    if (doswap) {
        byteswap ((char *) &len, sizeof (int));
    }
    s = MallocOrDie (sizeof (char) * (len));
    if (!fread ((char *) s, sizeof (char), len, fp)) {
        free (s);
        return 0;
    }

    *ret_s = s;
    return 1;
}

/* Function: multiline()
 * Date:     Mon Jan  5 14:57:50 1998 [StL]
 *
 * Purpose:  Given a record (like the comlog) that contains
 *           multiple lines, print it as multiple lines with
 *           a given prefix. e.g.:
 *
 *           given:   "COM   ", "foo\nbar\nbaz"
 *           print:   COM   foo
 *                    COM   bar
 *                    COM   baz
 *
 *
 *           Used to print the command log to ASCII save files.
 *
 * Args:     fp:   FILE to print to
 *           pfx:  prefix for each line
 *           s:    line to break up and print; tolerates a NULL
 *
 * Return:   (void)
 */
static void
multiline (FILE * fp, char *pfx, char *s)
{
    char *buf;
    char *sptr;

    if (s == NULL) {
        return;
    }
    buf = Strdup (s);
    sptr = strtok (buf, "\n");
    while (sptr != NULL) {
        fprintf (fp, "%s%s\n", pfx, sptr);
        sptr = strtok (NULL, "\n");
    }
    free (buf);
}


/*****************************************************************
 * HMMER 1.x save file reading functions, modified from the
 * corpse of 1.9m.
 *****************************************************************/


/* Function: read_plan9_binhmm()
 *
 * Read old (Plan9) binary HMM save files from HMMER 1.9 and earlier.
 * V1.0 saved regularizer and sympvec info, which V1.1 ignores.
 * V1.7 and later may include optional ref, cs annotation lines.
 * V1.9 added name, null model.
 *
 * Returns pointer to the HMM on success; NULL
 * on failure. Sets global alphabet information based on
 * whether it reads 4 or 20 as alphabet size (don't rely
 * on ancient HMMER macro definitions).
 */
static struct plan9_s *
read_plan9_binhmm (FILE * fp, int version, int swapped)
{
    struct plan9_s *hmm;
    int M;			/* length of model  */
    int k;			/* state number  */
    int x;			/* symbol or transition number */
    int len;			/* length of variable length string */
    int asize;			/* alphabet size */
    int atype;			/* alphabet type (read but ignored) */
    char abet[20];		/* alphabet (read but ignored) */

    /* read M and alphabet size */
    if (!fread ((char *) &(M), sizeof (int), 1, fp)) {
        return NULL;
    }
    if (!fread ((char *) &asize, sizeof (int), 1, fp)) {
        return NULL;
    }
    if (swapped) {
        byteswap ((char *) &M, sizeof (int));
        byteswap ((char *) &asize, sizeof (int));
    }

    /* Set global alphabet information
     */
    if (asize == 4) {
        atype = hmmNUCLEIC;
    }
    else if (asize == 20) {
        atype = hmmAMINO;
    }
    else
        Die
        ("A nonbiological alphabet size of %d; so I can't convert plan9 to plan7",
         asize);
    if (Alphabet_type == hmmNOTSETYET) {
        SetAlphabet (atype);
    }
    else if (atype != Alphabet_type)
        Die
        ("Alphabet mismatch error.\nI thought we were working with %s, but tried to read a %s HMM.\n",
         AlphabetType2String (Alphabet_type), AlphabetType2String (atype));

    /* now, create space for hmm */
    if ((hmm = P9AllocHMM (M)) == NULL) {
        Die ("malloc failed for reading hmm in\n");
    }

    /* version 1.9+ files have a name */
    if (version == HMMER1_9B) {
        if (!fread ((char *) &len, sizeof (int), 1, fp)) {
            return NULL;
        }
        if (swapped) {
            byteswap ((char *) &len, sizeof (int));
        }
        hmm->name = (char *) ReallocOrDie (hmm->name, sizeof (char) * (len + 1));
        if (!fread ((char *) hmm->name, sizeof (char), len, fp)) {
            return NULL;
        }
        hmm->name[len] = '\0';
    }

    /* read alphabet_type and alphabet, but ignore: we've already set them */
    if (!fread ((char *) &atype, sizeof (int), 1, fp)) {
        return NULL;
    }
    if (!fread ((char *) abet, sizeof (char), Alphabet_size, fp)) {
        return NULL;
    }

    /* skip the random symbol frequencies in V1.0 */
    if (version == HMMER1_0B) {
        fseek (fp, (long) (sizeof (float) * Alphabet_size), SEEK_CUR);
    }

    /* Get optional info in V1.7 and later
     */
    if (version == HMMER1_7B || version == HMMER1_9B) {
        if (!fread ((char *) &(hmm->flags), sizeof (int), 1, fp)) {
            return NULL;
        }
        if (swapped) {
            byteswap ((char *) &hmm->flags, sizeof (int));
        }
        if ((hmm->flags & HMM_REF) &&
                !fread ((char *) hmm->ref, sizeof (char), hmm->M + 1, fp)) {
            return NULL;
        }
        hmm->ref[hmm->M + 1] = '\0';
        if ((hmm->flags & HMM_CS) &&
                !fread ((char *) hmm->cs, sizeof (char), hmm->M + 1, fp)) {
            return NULL;
        }
        hmm->cs[hmm->M + 1] = '\0';
    }

    /* Get the null model in V1.9 and later
     */
    if (version == HMMER1_9B) {
        if (!fread ((char *) hmm->null, sizeof (float), Alphabet_size, fp)) {
            return NULL;
        }
        if (swapped)
            for (x = 0; x < Alphabet_size; x++) {
                byteswap ((char *) &(hmm->null[x]), sizeof (float));
            }
    }
    else {
        P9DefaultNullModel (hmm->null);
    }

    /* everything else is states */
    for (k = 0; k <= hmm->M; k++) {
        /* get match state info */
        if (!fread ((char *) &(hmm->mat[k].t[MATCH]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) &(hmm->mat[k].t[DELETE]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) &(hmm->mat[k].t[INSERT]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) hmm->mat[k].p, sizeof (float), Alphabet_size, fp)) {
            return NULL;
        }
        if (swapped) {
            byteswap ((char *) &(hmm->mat[k].t[MATCH]), sizeof (float));
            byteswap ((char *) &(hmm->mat[k].t[DELETE]), sizeof (float));
            byteswap ((char *) &(hmm->mat[k].t[INSERT]), sizeof (float));
            for (x = 0; x < Alphabet_size; x++) {
                byteswap ((char *) &(hmm->mat[k].p[x]), sizeof (float));
            }
        }

        /* skip the regularizer info in V1.0 */
        if (version == HMMER1_0B) {
            fseek (fp, (long) (sizeof (float) * (3 + Alphabet_size)), SEEK_CUR);
        }

        /* get delete state info */
        if (!fread ((char *) &(hmm->del[k].t[MATCH]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) &(hmm->del[k].t[DELETE]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) &(hmm->del[k].t[INSERT]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (swapped) {
            byteswap ((char *) &(hmm->del[k].t[MATCH]), sizeof (float));
            byteswap ((char *) &(hmm->del[k].t[DELETE]), sizeof (float));
            byteswap ((char *) &(hmm->del[k].t[INSERT]), sizeof (float));
        }

        /* skip the regularizer info in V1.0 */
        if (version == HMMER1_0B) {
            fseek (fp, (long) (sizeof (float) * 3), SEEK_CUR);
        }

        /* get insert state info */
        if (!fread ((char *) &(hmm->ins[k].t[MATCH]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) &(hmm->ins[k].t[DELETE]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) &(hmm->ins[k].t[INSERT]), sizeof (float), 1, fp)) {
            return NULL;
        }
        if (!fread ((char *) hmm->ins[k].p, sizeof (float), Alphabet_size, fp)) {
            return NULL;
        }
        if (swapped) {
            byteswap ((char *) &(hmm->ins[k].t[MATCH]), sizeof (float));
            byteswap ((char *) &(hmm->ins[k].t[DELETE]), sizeof (float));
            byteswap ((char *) &(hmm->ins[k].t[INSERT]), sizeof (float));
            for (x = 0; x < Alphabet_size; x++) {
                byteswap ((char *) &(hmm->ins[k].p[x]), sizeof (float));
            }
        }

        /* skip the regularizer info in V1.0 */
        if (version == HMMER1_0B) {
            fseek (fp, (long) (sizeof (float) * (3 + Alphabet_size)), SEEK_CUR);
        }
    }
    P9Renormalize (hmm);
    return hmm;
}


/* Function: read_plan9_aschmm()
 *
 * Purpose:  Read ASCII-format save files from 1.8.4 and earlier.
 *           V1.0 contained sympvec and regularizers; these are ignored
 *                in V1.1 and later
 *           V1.7 and later contain ref and cs annotation.
 *
 * Args:     fp      - open save file, header has been read already
 *           version - HMMER1_7F, for instance
 *
 * Returns ptr to the (allocated) new HMM on success,
 * or NULL on failure.
 */
static struct plan9_s *
read_plan9_aschmm (FILE * fp, int version)
{
    struct plan9_s *hmm;
    int M;			/* length of model  */
    char buffer[512];
    char *statetype;
    char *s;
    int k;			/* state number  */
    int i;			/* symbol number */
    int asize;			/* Alphabet size */
    int atype;			/* Alphabet type */

    /* read M from first line */
    if (fgets (buffer, 512, fp) == NULL) {
        return NULL;
    }
    if ((s = strtok (buffer, " \t\n")) == NULL) {
        return NULL;
    }
    if (!isdigit ((int) (*s))) {
        return NULL;
    }
    M = atoi (s);
    /* read alphabet_length */
    if (fgets (buffer, 512, fp) == NULL) {
        return NULL;
    }
    if ((s = strtok (buffer, " \t\n")) == NULL) {
        return NULL;
    }
    if (!isdigit ((int) (*s))) {
        return NULL;
    }
    asize = atoi (s);

    /* Set global alphabet information
     */
    if (asize == 4) {
        atype = hmmNUCLEIC;
    }
    else if (asize == 20) {
        atype = hmmAMINO;
    }
    else
        Die
        ("A nonbiological alphabet size of %d; so I can't convert plan9 to plan7",
         asize);
    if (Alphabet_type == hmmNOTSETYET) {
        SetAlphabet (atype);
    }
    else if (atype != Alphabet_type)
        Die
        ("Alphabet mismatch error.\nI thought we were working with %s, but tried to read a %s HMM.\n",
         AlphabetType2String (Alphabet_type), AlphabetType2String (atype));

    /* now, create space for hmm */
    if ((hmm = P9AllocHMM (M)) == NULL) {
        Die ("malloc failed for reading hmm in\n");
    }

    /* read alphabet_type but ignore */
    if (fgets (buffer, 512, fp) == NULL) {
        return NULL;
    }
    if ((s = strtok (buffer, " \t\n")) == NULL) {
        return NULL;
    }
    if (!isdigit ((int) (*s))) {
        return NULL;
    }
    /* read alphabet but ignore */
    if (fgets (buffer, 512, fp) == NULL) {
        return NULL;
    }
    if ((s = strtok (buffer, " \t\n")) == NULL) {
        return NULL;
    }

    /* skip the random symbol frequencies in V1.0 files. now unused */
    if (version == HMMER1_0F)
        for (i = 0; i < Alphabet_size; i++) {
            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
        }

    /* V1.7 has lines for whether we have valid ref, cs info
     */
    if (version == HMMER1_7F) {
        if (fgets (buffer, 512, fp) == NULL) {
            return NULL;
        }
        if (strncmp (buffer, "yes", 3) == 0) {
            hmm->flags |= HMM_REF;
        }
        if (fgets (buffer, 512, fp) == NULL) {
            return NULL;
        }
        if (strncmp (buffer, "yes", 3) == 0) {
            hmm->flags |= HMM_CS;
        }
    }

    /* everything else is states */
    while (fgets (buffer, 512, fp) != NULL) {
        /* get state type and index info */
        if ((statetype = strtok (buffer, " \t\n")) == NULL) {
            return NULL;
        }
        if ((s = strtok ((char *) NULL, " \t\n")) == NULL) {
            return NULL;
        }
        if (!isdigit ((int) (*s))) {
            return NULL;
        }
        k = atoi (s);
        if (k < 0 || k > hmm->M + 1) {
            return NULL;
        }

        if (strcmp (statetype, "###MATCH_STATE") == 0) {
            /* V1.7: get ref, cs info:   */
            /* ###MATCH_STATE 16 (x) (H) */
            if (version == HMMER1_7F) {
                s = strtok (NULL, "\n");
                while (*s != '(' && *s != '\0') {
                    s++;
                }
                if (*s != '(') {
                    return NULL;
                }
                hmm->ref[k] = *(s + 1);
                while (*s != '(' && *s != '\0') {
                    s++;
                }
                if (*s != '(') {
                    return NULL;
                }
                hmm->cs[k] = *(s + 1);
            }

            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->mat[k].t[MATCH] = (float) atof (s);

            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->mat[k].t[DELETE] = (float) atof (s);

            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->mat[k].t[INSERT] = (float) atof (s);

            for (i = 0; i < Alphabet_size; i++) {
                if (fgets (buffer, 512, fp) == NULL) {
                    return NULL;
                }
                if ((s = strtok (buffer, " \t\n")) == NULL) {
                    return NULL;
                }
                hmm->mat[k].p[i] = (float) atof (s);
            }

            /* Skip all regularizer info for V1.0 */
            if (version == HMMER1_0F)
                for (i = 0; i < Alphabet_size + 3; i++) {
                    if (fgets (buffer, 512, fp) == NULL) {
                        return NULL;
                    }
                }
        }
        else if (strcmp (statetype, "###INSERT_STATE") == 0) {
            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->ins[k].t[MATCH] = (float) atof (s);

            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->ins[k].t[DELETE] = (float) atof (s);

            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->ins[k].t[INSERT] = (float) atof (s);

            for (i = 0; i < Alphabet_size; i++) {
                if (fgets (buffer, 512, fp) == NULL) {
                    return NULL;
                }
                if ((s = strtok (buffer, " \t\n")) == NULL) {
                    return NULL;
                }
                hmm->ins[k].p[i] = (float) atof (s);
            }

            /* Skip all regularizer info in V1.0 files */
            if (version == HMMER1_0F)
                for (i = 0; i < Alphabet_size + 3; i++) {
                    if (fgets (buffer, 512, fp) == NULL) {
                        return NULL;
                    }
                }
        }
        else if (strcmp (statetype, "###DELETE_STATE") == 0) {
            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->del[k].t[MATCH] = (float) atof (s);

            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->del[k].t[DELETE] = (float) atof (s);

            if (fgets (buffer, 512, fp) == NULL) {
                return NULL;
            }
            if ((s = strtok (buffer, " \t\n")) == NULL) {
                return NULL;
            }
            hmm->del[k].t[INSERT] = (float) atof (s);

            /* Skip all regularizer info in V1.0 files */
            if (version == HMMER1_0F)
                for (i = 0; i < 3; i++) {
                    if (fgets (buffer, 512, fp) == NULL) {
                        return NULL;
                    }
                }
        }
        else {
            return NULL;
        }
    }

    P9DefaultNullModel (hmm->null);
    P9Renormalize (hmm);
    return hmm;
}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* hmmsearch.c
 * SRE, Tue Jan  7 17:19:20 1997 [St. Louis]
 *
 * Search a sequence database with a profile HMM.
 * Conditionally includes PVM parallelization when HMMER_PVM is defined
 *    at compile time; hmmsearch --pvm runs the PVM version.
 *
 * CVS $Id: hmmsearch.c,v 1.40 2003/04/14 16:00:17 eddy Exp $
 */

#include "config.h"		/* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#ifdef HMMER_THREADS
#include <pthread.h>
#endif
#ifdef HMMER_PVM
#include <pvm3.h>
#endif

#include "squid.h"		/* general sequence analysis library    */
#include "structs.h"		/* data structures, macros, #define's   */
#include "funcs.h"		/* function declarations                */
#include "globals.h"		/* alphabet global variables            */

#ifdef SPEC_CPU
extern int hmmcalibrate (int argc, char **argv);
#endif
/* sergio: redefinicion de banner a bannersearch */
static char bannersearch[] =
    "hmmsearch - search a sequence database with a profile HMM";
/* sergio: redefinion de usage a usagesearch */
static char usagesearch[] = "\
Usage: hmmsearch [-options] <hmmfile> <sequence file or database>\n\
  Available options are:\n\
   -h        : help; print brief help on version and usage\n\
   -A <n>    : sets alignment output limit to <n> best domain alignments\n\
   -E <x>    : sets E value cutoff (globE) to <= x\n\
   -T <x>    : sets T bit threshold (globT) to >= x\n\
   -Z <n>    : sets Z (# seqs) for E-value calculation\n\
";
/* sergio: redefinion de experts a expertssearch */
static char expertssearch[] = "\
   --compat       : make best effort to use last version's output style\n\
   --cpu <n>      : run <n> threads in parallel (if threaded)\n\
   --cut_ga       : use Pfam GA gathering threshold cutoffs\n\
   --cut_nc       : use Pfam NC noise threshold cutoffs\n\
   --cut_tc       : use Pfam TC trusted threshold cutoffs\n\
   --domE <x>     : sets domain Eval cutoff (2nd threshold) to <= x\n\
   --domT <x>     : sets domain T bit thresh (2nd threshold) to >= x\n\
   --forward      : use the full Forward() algorithm instead of Viterbi\n\
   --informat <s> : sequence file is in format <s>\n\
   --null2        : turn OFF the post hoc second null model\n\
   --pvm          : run on a Parallel Virtual Machine (PVM)\n\
   --xnu          : turn ON XNU filtering of target protein sequences\n\
";

/* sergio: redefinion de OPTIONS a OPTIONSsearch */
static struct opt_s OPTIONSsearch[] = {
    {"-h", TRUE, sqdARG_NONE},
    {"-A", TRUE, sqdARG_INT},
    {"-E", TRUE, sqdARG_FLOAT},
    {"-T", TRUE, sqdARG_FLOAT},
    {"-Z", TRUE, sqdARG_INT},
    {"--compat", FALSE, sqdARG_NONE},
    {"--cpu", FALSE, sqdARG_INT},
    {"--cut_ga", FALSE, sqdARG_NONE},
    {"--cut_nc", FALSE, sqdARG_NONE},
    {"--cut_tc", FALSE, sqdARG_NONE},
    {"--domE", FALSE, sqdARG_FLOAT},
    {"--domT", FALSE, sqdARG_FLOAT},
    {"--forward", FALSE, sqdARG_NONE},
    {"--informat", FALSE, sqdARG_STRING},
    {"--null2", FALSE, sqdARG_NONE},
    {"--pvm", FALSE, sqdARG_NONE},
    {"--xnu", FALSE, sqdARG_NONE},

};

/* sergio: Redefinicion de NOPTIONS a NOPTIONSsearch */
#define NOPTIONSsearch (sizeof(OPTIONSsearch) / sizeof(struct opt_s))


/* sergio: redefinion de main_loop_serial a main_loop_serialsearch */
static void main_loop_serialsearch (struct plan7_s *hmm, SQFILE * sqfp,
                                    struct threshold_s *thresh,
                                    int do_forward, int do_null2, int do_xnu,
                                    struct histogram_s *histogram,
                                    struct tophit_s *ghit,
                                    struct tophit_s *dhit, int *ret_nseq);
static void main_loop_pvm (struct plan7_s *hmm, SQFILE * sqfp,
                           struct threshold_s *thresh, int do_forward,
                           int do_null2, int do_xnu,
                           struct histogram_s *histogram,
                           struct tophit_s *ghit, struct tophit_s *dhit,
                           int *ret_nseq);
static void main_loop_threaded (struct plan7_s *hmm, SQFILE * sqfp,
                                struct threshold_s *thresh, int do_forward,
                                int do_null2, int do_xnu, int num_threads,
                                struct histogram_s *histogram,
                                struct tophit_s *ghit, struct tophit_s *dhit,
                                int *ret_nseq);

#ifdef HMMER_THREADS
/* POSIX threads version:
 * the threads share a workpool_s structure amongst themselves,
 * for obtaining locks on input HMM file and output histogram and
 * tophits structures.
 */
struct workpool_s
{
    /* Shared configuration resources which don't change:
     */
    struct plan7_s *hmm;		/* HMM to search with              */
    int do_xnu;			/* TRUE to apply XNU filter  */
    int do_forward;		/* TRUE to score using Forward     */
    int do_null2;			/* TRUE to apply null2 ad hoc correction */
    struct threshold_s *thresh;	/* score/evalue threshold info     */

    /* Shared (mutex-protected) input resources:
     */
    SQFILE *sqfp;			/* ptr to open sequence file      */
    int nseq;			/* number of seqs searched so far */
    pthread_mutex_t input_lock;	/* mutex for locking input        */

    /* Shared (mutex-protected) output resources:
     */
    struct tophit_s *ghit;	/* per-sequence top hits */
    struct tophit_s *dhit;	/* per-domain top hits */
    struct histogram_s *hist;	/* histogram of scores */
    pthread_mutex_t output_lock;	/* mutex for locking output */

    /* Thread pool information
     */
    pthread_t *thread;		/* our pool of threads */
    int num_threads;		/* number of threads   */
};
static struct workpool_s *workpool_start (struct plan7_s *hmm, SQFILE * sqfp,
        int do_xnu, int do_forward,
        int do_null2,
        struct threshold_s *thresh,
        struct tophit_s *ghit,
        struct tophit_s *dhit,
        struct histogram_s *hist,
        int num_threads);
static void workpool_stop (struct workpool_s *wpool);
static void workpool_free (struct workpool_s *wpool);
static void *worker_thread (void *ptr);
#endif /* HMMER_THREADS */



int
main (int argc, char **argv)
{
    char *hmmfile;		/* file to read HMM(s) from                */
    HMMFILE *hmmfp;		/* opened hmmfile for reading              */
    char *seqfile;		/* file to read target sequence(s) from    */
    SQFILE *sqfp;			/* opened seqfile for reading              */
    int format;			/* format of seqfile                       */
    int i;
    struct plan7_s *hmm;		/* HMM to search with                      */
    struct histogram_s *histogram;	/* histogram of all scores                 */
    struct fancyali_s *ali;	/* displayed alignment info                */
    struct tophit_s *ghit;	/* list of top hits for whole sequences    */
    struct tophit_s *dhit;	/* list of top hits for domains            */

    float sc;			/* score of an HMM search                  */
    double pvalue;		/* pvalue of an HMM score                  */
    double evalue;		/* evalue of an HMM score                  */
    double motherp;		/* pvalue of a whole seq HMM score         */
    float mothersc;		/* score of a whole seq parent of domain   */
    int sqfrom, sqto;		/* coordinates in sequence                 */
    int hmmfrom, hmmto;		/* coordinate in HMM                       */
    char *name, *desc;		/* hit sequence name and description       */
    int sqlen;			/* length of seq that was hit              */
    int nseq;			/* number of sequences searched            */
    int Z;			/* # of seqs for purposes of E-val calc    */
    int domidx;			/* number of this domain                   */
    int ndom;			/* total # of domains in this seq          */
    int namewidth;		/* max width of sequence name              */
    int descwidth;		/* max width of description */
    int nreported;		/* # of hits reported in a list            */

    int Alimit;			/* A parameter limiting output alignments   */
    struct threshold_s thresh;	/* contains all threshold (cutoff) info     */

    char *optname;		/* name of option found by Getopt()         */
    char *optarg;			/* argument found by Getopt()               */
    int optind;			/* index in argv[]                          */
    int do_null2;			/* TRUE to adjust scores with null model #2 */
    int do_forward;		/* TRUE to use Forward() not Viterbi()      */
    int do_xnu;			/* TRUE to filter sequences thru XNU        */
    int do_pvm;			/* TRUE to run on Parallel Virtual Machine  */
    int be_backwards;		/* TRUE to be backwards-compatible in output */
    int num_threads;		/* number of worker threads                 */
    int threads_support;		/* TRUE if threads support compiled in      */
    int pvm_support;		/* TRUE if PVM support compiled in          */

#ifdef SPEC_CPU
    if (argc != 3) {
        return hmmcalibrate (argc, argv);
    }
#endif

    /***********************************************
     * Parse command line
     ***********************************************/

    format = SQFILE_UNKNOWN;	/* default: autodetect seq file format  */
    do_forward = FALSE;
    do_null2 = TRUE;
    do_xnu = FALSE;
    do_pvm = FALSE;
    Z = 0;
    be_backwards = FALSE;

    pvm_support = FALSE;
    threads_support = FALSE;
    num_threads = 0;
#ifdef HMMER_THREADS
    num_threads = ThreadNumber ();
    threads_support = TRUE;
#endif
#ifdef HMMER_PVM
    pvm_support = TRUE;
#endif

    Alimit = INT_MAX;		/* no limit on alignment output       */
    thresh.globE = 10.0;		/* use a reasonable Eval threshold;   */
    thresh.globT = -FLT_MAX;	/*   but no bit threshold,            */
    thresh.domT = -FLT_MAX;	/*   no domain bit threshold,         */
    thresh.domE = FLT_MAX;	/*   and no domain Eval threshold.    */
    thresh.autocut = CUT_NONE;	/*   and no Pfam cutoffs used         */
    thresh.Z = 0;			/* Z not preset; use actual # of seqs */

    while (Getopt (argc, argv, OPTIONSsearch, NOPTIONSsearch, usagesearch,
                   &optind, &optname, &optarg)) {
        if (strcmp (optname, "-A") == 0) {
            Alimit = atoi (optarg);
        }
        else if (strcmp (optname, "-E") == 0) {
            thresh.globE = atof (optarg);
        }
        else if (strcmp (optname, "-T") == 0) {
            thresh.globT = atof (optarg);
        }
        else if (strcmp (optname, "-Z") == 0) {
            thresh.Z = atoi (optarg);
        }
        else if (strcmp (optname, "--compat") == 0) {
            be_backwards = TRUE;
        }
        else if (strcmp (optname, "--cpu") == 0) {
            num_threads = atoi (optarg);
        }
        else if (strcmp (optname, "--cut_ga") == 0) {
            thresh.autocut = CUT_GA;
        }
        else if (strcmp (optname, "--cut_nc") == 0) {
            thresh.autocut = CUT_NC;
        }
        else if (strcmp (optname, "--cut_tc") == 0) {
            thresh.autocut = CUT_TC;
        }
        else if (strcmp (optname, "--domE") == 0) {
            thresh.domE = atof (optarg);
        }
        else if (strcmp (optname, "--domT") == 0) {
            thresh.domT = atof (optarg);
        }
        else if (strcmp (optname, "--forward") == 0) {
            do_forward = TRUE;
        }
        else if (strcmp (optname, "--null2") == 0) {
            do_null2 = FALSE;
        }
        else if (strcmp (optname, "--pvm") == 0) {
            do_pvm = TRUE;
        }
        else if (strcmp (optname, "--xnu") == 0) {
            do_xnu = TRUE;
        }
        else if (strcmp (optname, "--informat") == 0) {
            format = String2SeqfileFormat (optarg);
            if (format == SQFILE_UNKNOWN) {
                Die ("unrecognized sequence file format \"%s\"", optarg);
            }
        }
        else if (strcmp (optname, "-h") == 0) {
            HMMERBanner (stdout, bannersearch);
            puts (usagesearch);
            puts (expertssearch);
            exit (0);
        }
    }
    if (argc - optind != 2) {
        Die ("Incorrect number of arguments.\n%s\n", usagesearch);
    }

    hmmfile = argv[optind++];
    seqfile = argv[optind++];

    if (do_pvm && !pvm_support)
        Die
        ("PVM support is not compiled into your HMMER software; --pvm doesn't work.");
    if (num_threads && !threads_support)
        Die
        ("POSIX threads support is not compiled into HMMER; --cpu doesn't have any effect");

    /* Try to work around inability to autodetect from a pipe or .gz:
     * assume FASTA format
     */
    if (format == SQFILE_UNKNOWN &&
            (Strparse ("^.*\\.gz$", seqfile, 0) || strcmp (seqfile, "-") == 0)) {
        format = SQFILE_FASTA;
    }

    /***********************************************
     * Open sequence database (might be in BLASTDB or current directory)
     ***********************************************/

    if ((sqfp = SeqfileOpen (seqfile, format, "BLASTDB")) == NULL)
        Die ("Failed to open sequence database file %s\n%s\n", seqfile,
             usagesearch);

    /***********************************************
     * Open HMM file (might be in HMMERDB or current directory).
     * Read a single HMM from it. (Config HMM, if necessary).
     * Alphabet globals are set by reading the HMM.
     ***********************************************/

    if ((hmmfp = HMMFileOpen (hmmfile, "HMMERDB")) == NULL) {
        Die ("Failed to open HMM file %s\n%s", hmmfile, usagesearch);
    }
    if (!HMMFileRead (hmmfp, &hmm)) {
        Die ("Failed to read any HMMs from %s\n", hmmfile);
    }
    if (hmm == NULL) {
        Die ("HMM file %s corrupt or in incorrect format? Parse failed", hmmfile);
    }
    P7Logoddsify (hmm, !do_forward);

    if (do_xnu && Alphabet_type == hmmNUCLEIC)
        Die
        ("The HMM is a DNA model, and you can't use the --xnu filter on DNA data");

    /*****************************************************************
     * Set up optional Pfam score thresholds.
     * Can do this before starting any searches, since we'll only use 1 HMM.
     *****************************************************************/

    if (!SetAutocuts (&thresh, hmm))
        Die ("HMM %s did not contain the GA, TC, or NC cutoffs you needed",
             hmm->name);

    /***********************************************
     * Show the banner
     ***********************************************/

    HMMERBanner (stdout, bannersearch);
    printf ("HMM file:                   %s [%s]\n", hmmfile, hmm->name);
    printf ("Sequence database:          %s\n", seqfile);
    if (do_pvm) {
        printf ("PVM:                        ACTIVE\n");
    }
    printf ("per-sequence score cutoff:  ");
    if (thresh.globT == -FLT_MAX) {
        printf ("[none]\n");
    }
    else {
        printf (">= %.1f", thresh.globT);
        if (thresh.autocut == CUT_GA) {
            printf (" [GA1]\n");
        }
        else if (thresh.autocut == CUT_NC) {
            printf (" [NC1]\n");
        }
        else if (thresh.autocut == CUT_TC) {
            printf (" [TC1]\n");
        }
        else {
            printf ("\n");
        }
    }
    printf ("per-domain score cutoff:    ");
    if (thresh.domT == -FLT_MAX) {
        printf ("[none]\n");
    }
    else {
        printf (">= %.1f", thresh.domT);
        if (thresh.autocut == CUT_GA) {
            printf (" [GA2]\n");
        }
        else if (thresh.autocut == CUT_NC) {
            printf (" [NC2]\n");
        }
        else if (thresh.autocut == CUT_TC) {
            printf (" [TC2]\n");
        }
        else {
            printf ("\n");
        }
    }
    printf ("per-sequence Eval cutoff:   ");
    if (thresh.globE == FLT_MAX) {
        printf ("[none]\n");
    }
    else {
        printf ("<= %-10.2g\n", thresh.globE);
    }

    printf ("per-domain Eval cutoff:     ");
    if (thresh.domE == FLT_MAX) {
        printf ("[none]\n");
    }
    else {
        printf ("<= %10.2g\n", thresh.domE);
    }
    printf
    ("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");

    /***********************************************
     * Search HMM against each sequence
     ***********************************************/

    /* set up structures for storing output  */
    histogram = AllocHistogram (-200, 200, 100);	/* keeps full histogram */
    ghit = AllocTophits (200);	/* per-seq hits: 200=lumpsize */
    dhit = AllocTophits (200);	/* domain hits:  200=lumpsize */

    if (pvm_support && do_pvm)
        main_loop_pvm (hmm, sqfp, &thresh, do_forward, do_null2, do_xnu,
                       histogram, ghit, dhit, &nseq);
    else if (threads_support && num_threads)
        main_loop_threaded (hmm, sqfp, &thresh, do_forward, do_null2, do_xnu,
                            num_threads, histogram, ghit, dhit, &nseq);
    else
        main_loop_serialsearch (hmm, sqfp, &thresh, do_forward, do_null2, do_xnu,
                                histogram, ghit, dhit, &nseq);

    /***********************************************
     * Process hit lists, produce text output
     ***********************************************/

    /* Set the theoretical EVD curve in our histogram using
     * calibration in the HMM, if available.
     */
    if (hmm->flags & PLAN7_STATS)
        ExtremeValueSetHistogram (histogram, hmm->mu, hmm->lambda,
                                  histogram->lowscore, histogram->highscore, 0);
    if (!thresh.Z) {
        thresh.Z = nseq;    /* set Z for good now that we're done. */
    }

    /* Format and report our output
     */
    /* 1. Report overall sequence hits (sorted on E-value) */
    if (be_backwards) {
        printf ("\nQuery HMM: %s|%s|%s\n",
                hmm->name,
                hmm->flags & PLAN7_ACC ? hmm->acc : "",
                hmm->flags & PLAN7_DESC ? hmm->desc : "");
    }
    else {
        printf ("\nQuery HMM:   %s\n", hmm->name);
        printf ("Accession:   %s\n",
                hmm->flags & PLAN7_ACC ? hmm->acc : "[none]");
        printf ("Description: %s\n",
                hmm->flags & PLAN7_DESC ? hmm->desc : "[none]");
    }

    if (hmm->flags & PLAN7_STATS)
        printf
        ("  [HMM has been calibrated; E-values are empirical estimates]\n");
    else {
        printf ("  [No calibration for HMM; E-values are upper bounds]\n");
    }

    FullSortTophits (ghit);
    namewidth = MAX (8, TophitsMaxName (ghit));	/* cannot truncate name. */
    descwidth = MAX (52 - namewidth, 11);	/* may truncate desc, but need strlen("Description") */

    /* Kaivalya commented following output 6/5/03 */
#ifndef SPEC_CPU
    /*  ifdef PRODUCT_HMMER */
    printf ("\nScores for complete sequences (score includes all domains):\n");
    printf ("%-*s %-*s %7s %10s %3s\n", namewidth, "Sequence", descwidth,
            "Description", "Score", "E-value", " N ");
    printf ("%-*s %-*s %7s %10s %3s\n", namewidth, "--------", descwidth,
            "-----------", "-----", "-------", "---");
    for (i = 0, nreported = 0; i < ghit->num; i++) {
        char *safedesc;
        GetRankedHit (ghit, i, &pvalue, &sc, NULL, NULL, &name, NULL, &desc, NULL, NULL, NULL,	/* sequence positions */
                      NULL, NULL, NULL,	/* HMM positions      */
                      NULL, &ndom,	/* domain info        */
                      NULL);	/* alignment info     */
        evalue = pvalue * (double) thresh.Z;

        /* safedesc is a workaround for an apparent Linux printf()
         * bug with the *.*s format. dbmalloc crashes with a memchr() ptr out of bounds
         * flaw if the malloc'ed space for desc is short. The workaround
         * is to make sure the ptr for *.* has a big malloc space.
         */
        if (desc != NULL && strlen (desc) < 80) {
            safedesc = MallocOrDie (sizeof (char) * 80);
            strcpy (safedesc, desc);
        }
        else {
            safedesc = Strdup (desc);
        }

        if (evalue <= thresh.globE && sc >= thresh.globT) {
            printf ("%-*s %-*.*s %7.1f %10.2g %3d\n",
                    namewidth, name,
                    descwidth, descwidth, safedesc != NULL ? safedesc : "",
                    sc, evalue, ndom);
            nreported++;
        }
        free (safedesc);
    }

    if (nreported == 0) {
        printf ("\t[no hits above thresholds]\n");
    }
#endif
    /* Kaivalya above output removed 6/5/03 */


    /* 2. Report domain hits (also sorted on E-value) */
    FullSortTophits (dhit);
    namewidth = MAX (8, TophitsMaxName (dhit));

    /* Kaivalya commented the following output 6/5/03 */

#ifndef SPEC_CPU
    /* ifdef PRODUCT_HMMER */
    printf ("\nParsed for domains:\n");
    printf ("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
            namewidth, "Sequence", "Domain ", "seq-f", "seq-t", "hmm-f",
            "hmm-t", "score", "E-value");
    printf ("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n", namewidth, "--------",
            "-------", "-----", "-----", "-----", "-----", "-----", "-------");

    for (i = 0, nreported = 0; i < dhit->num; i++) {
        GetRankedHit (dhit, i, &pvalue, &sc, &motherp, &mothersc, &name, NULL, NULL, &sqfrom, &sqto, &sqlen,	/* seq position info  */
                      &hmmfrom, &hmmto, NULL,	/* HMM position info  */
                      &domidx, &ndom,	/* domain info        */
                      NULL);	/* alignment info     */
        evalue = pvalue * (double) thresh.Z;

        if (motherp * (double) thresh.Z > thresh.globE || mothersc < thresh.globT) {
            continue;
        }
        else if (evalue <= thresh.domE && sc >= thresh.domT) {
            printf ("%-*s %3d/%-3d %5d %5d %c%c %5d %5d %c%c %7.1f %8.2g\n",
                    namewidth, name,
                    domidx, ndom,
                    sqfrom, sqto,
                    sqfrom == 1 ? '[' : '.', sqto == sqlen ? ']' : '.',
                    hmmfrom, hmmto,
                    hmmfrom == 1 ? '[' : '.', hmmto == hmm->M ? ']' : '.',
                    sc, evalue);
            nreported++;
        }
    }
    if (nreported == 0) {
        printf ("\t[no hits above thresholds]\n");
    }
#endif
    /* Kaivalya the above output commented 6/5/03  */

    /* 3. Alignment output, also by domain.
     *    dhits is already sorted and namewidth is set, from above code.
     *    Number of displayed alignments is limited by Alimit parameter;
     *    also by domE (evalue threshold), domT (score theshold).
     */
    if (Alimit != 0) {
        printf ("\nAlignments of top-scoring domains:\n");
        for (i = 0, nreported = 0; i < dhit->num; i++) {
            if (nreported == Alimit) {
                break;    /* limit to Alimit output alignments */
            }
            GetRankedHit (dhit, i, &pvalue, &sc, &motherp, &mothersc, &name, NULL, NULL, &sqfrom, &sqto, &sqlen,	/* seq position info  */
                          &hmmfrom, &hmmto, NULL,	/* HMM position info  */
                          &domidx, &ndom,	/* domain info        */
                          &ali);	/* alignment info     */
            evalue = pvalue * (double) thresh.Z;

            if (motherp * (double) thresh.Z > thresh.globE
                    || mothersc < thresh.globT) {
                continue;
            }
            else if (evalue <= thresh.domE && sc >= thresh.domT) {
                printf ("%s: domain %d of %d, from %d to %d: score %.1f, E = %.2g\n",
                        name, domidx, ndom, sqfrom, sqto, sc, evalue);
                PrintFancyAli (stdout, ali);
                nreported++;
            }
        }
        if (nreported == 0) {
            printf ("\t[no hits above thresholds]\n");
        }
        if (nreported == Alimit) {
            printf ("\t[output cut off at A = %d top alignments]\n", Alimit);
        }
    }


    /* 4. Histogram output */

    /*   Kaivalya 060603 */
    /* ifdef PRODUCT_HMMER turn on later */
#ifndef SPEC_CPU
    printf ("\nHistogram of all scores:\n");
    PrintASCIIHistogram (stdout, histogram);
#endif
    /* #endif */

    /*   Kaivalya 061103 */
#ifndef SPEC_CPU
    /* ifdef PRODUCT_HMMER turn on later */

    /* 5. Tophits summaries, while developing...
     */
    printf ("\nTotal sequences searched: %d\n", nseq);
    printf ("\nWhole sequence top hits:\n");
    TophitsReport (ghit, thresh.globE, nseq);
    printf ("\nDomain top hits:\n");
    TophitsReport (dhit, thresh.domE, nseq);
    /*  #endif   Kaivalya turn on later */
#endif

    printf ("\nTotal sequences searched: %d\n", nseq);	/*Kaivalya 6-11-03 */
    /***********************************************
     * Clean-up and exit.
     ***********************************************/

    FreeHistogram (histogram);
    HMMFileClose (hmmfp);
    SeqfileClose (sqfp);
    FreeTophits (ghit);
    FreeTophits (dhit);
    FreePlan7 (hmm);
    SqdClean ();

    return 0;
}


/* Function: main_loop_serialsearch()
 * Date:     SRE, Wed Sep 23 10:20:49 1998 [St. Louis]
 *
 * Purpose:  Search an HMM against a sequence database.
 *           main loop for the serial (non-PVM, non-threads)
 *           version.
 *
 *           In:   HMM and open sqfile, plus options
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with.
 *           sqfp       - open SQFILE for sequence database
 *           thresh     - score/evalue threshold info
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *
 * Returns:  (void)
 */
static void
main_loop_serialsearch (struct plan7_s *hmm, SQFILE * sqfp,
                        struct threshold_s *thresh, int do_forward,
                        int do_null2, int do_xnu,
                        struct histogram_s *histogram, struct tophit_s *ghit,
                        struct tophit_s *dhit, int *ret_nseq)
{
    struct dpmatrix_s *mx;	/* DP matrix, growable                     */
    struct p7trace_s *tr;		/* traceback                               */
    char *seq;			/* target sequence                         */
    char *dsq;			/* digitized target sequence               */
    SQINFO sqinfo;		/* optional info for seq                   */
    float sc;			/* score of an HMM search                  */
    double pvalue;		/* pvalue of an HMM score                  */
    double evalue;		/* evalue of an HMM score                  */
    int nseq;			/* number of sequences searched            */

    /* Create a DP matrix; initially only two rows big, but growable;
     * we overalloc by 25 rows (L dimension) when we grow; not growable
     * in model dimension, since we know the hmm size
     */
    mx = CreatePlan7Matrix (1, hmm->M, 25, 0);

    nseq = 0;
    while (ReadSeq (sqfp, sqfp->format, &seq, &sqinfo)) {
        /* Silently skip length 0 seqs.
         * What, you think this doesn't occur? Welcome to genomics,
         * young grasshopper.
         */
        if (sqinfo.len == 0) {
            continue;
        }

        nseq++;
        dsq = DigitizeSequence (seq, sqinfo.len);

        if (do_xnu && Alphabet_type == hmmAMINO) {
            XNU (dsq, sqinfo.len);
        }

        /* 1. Recover a trace by Viterbi.
         *    In extreme cases, the alignment may be literally impossible;
         *    in which case, the score comes out ridiculously small (but not
         *    necessarily <= -INFTY, because we're not terribly careful
         *    about underflow issues), and tr will be returned as NULL.
         */
        if (P7ViterbiSize (sqinfo.len, hmm->M) <= RAMLIMIT) {
            sc = P7Viterbi (dsq, sqinfo.len, hmm, mx, &tr);
        }
        else {
            sc = P7SmallViterbi (dsq, sqinfo.len, hmm, mx, &tr);
        }

        /* 2. If we're using Forward scores, calculate the
         *    whole sequence score; this overrides anything
         *    PostprocessSignificantHit() is going to do to the per-seq score.
         */
        if (do_forward) {
            sc = P7Forward (dsq, sqinfo.len, hmm, NULL);
            if (do_null2) {
                sc -= TraceScoreCorrection (hmm, tr, dsq);
            }
        }

#if DEBUGLEVEL >= 2
        P7PrintTrace (stdout, tr, hmm, dsq);
#endif

        /* 2. Store score/pvalue for global alignment; will sort on score,
         *    which in hmmsearch is monotonic with E-value.
         *    Keep all domains in a significant sequence hit.
         *    We can only make a lower bound estimate of E-value since
         *    we don't know the final value of nseq yet, so the list
         *    of hits we keep in memory is >= the list we actually
         *    output.
         */
        pvalue = PValue (hmm, sc);
        evalue = thresh->Z ? (double) thresh->Z * pvalue : (double) nseq *pvalue;
        if (sc >= thresh->globT && evalue <= thresh->globE) {
            sc = PostprocessSignificantHit (ghit, dhit, tr, hmm, dsq, sqinfo.len, sqinfo.name, sqinfo.flags & SQINFO_ACC ? sqinfo.acc : NULL, sqinfo.flags & SQINFO_DESC ? sqinfo.desc : NULL, do_forward, sc, do_null2, thresh, FALSE);	/* FALSE-> not hmmpfam mode, hmmsearch mode */
        }
        SQD_DPRINTF2 (("AddToHistogram: %s\t%f\n", sqinfo.name, sc));
        AddToHistogram (histogram, sc);
        FreeSequence (seq, &sqinfo);
        P7FreeTrace (tr);
        free (dsq);
    }

    FreePlan7Matrix (mx);
    *ret_nseq = nseq;
    return;
}



#ifdef HMMER_PVM
/*****************************************************************
 * PVM specific functions
 ****************************************************************/

/* Function: main_loop_pvm()
 * Date:     SRE, Wed Sep 23 10:36:44 1998 [St. Louis]
 *
 * Purpose:  Search an HMM against a sequence database.
 *           main loop for the PVM version.
 *
 *           In:   HMM and open sqfile, plus options
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with. scoring form.
 *           sqfp       - open SQFILE for sequence database
 *           thresh     - score/evalue threshold information
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *
 * Returns:  (void)
 */
static void
main_loop_pvm (struct plan7_s *hmm, SQFILE * sqfp, struct threshold_s *thresh,
               int do_forward, int do_null2, int do_xnu,
               struct histogram_s *histogram, struct tophit_s *ghit,
               struct tophit_s *dhit, int *ret_nseq)
{
    char *seq;			/* target sequence */
    char *dsq;			/* digitized target seq */
    SQINFO sqinfo;		/* optional info about target seq */
    int master_tid;		/* master's (my) PVM TID */
    int *slave_tid;		/* array of slave TID's  */
    int nslaves;			/* number of slaves      */
    int code;			/* status code rec'd from a slave */
    int nseq;			/* number of sequences searched */
    int sent_trace;		/* TRUE if slave gave us a trace */
    char **dsqlist;		/* remember what seqs slaves are doing */
    char **namelist;		/* remember what seq names slaves are doing */
    char **acclist;		/* remember what seq accessions slaves are doing */
    char **desclist;		/* remember what seq desc's slaves are doing */
    int *lenlist;			/* remember lengths of seqs slaves are doing */
    int slaveidx;			/* counter for slaves */
    float sc;			/* score of an alignment */
    double pvalue;		/* P-value of a score of an alignment */
    struct p7trace_s *tr;		/* Viterbi traceback of an alignment */
    int i;			/* generic counter */

    /* Initialize PVM.
     */
    SQD_DPRINTF1 (("Requesting master TID...\n"));
    master_tid = pvm_mytid ();
#if DEBUGLEVEL >= 1
    pvm_catchout (stderr);	/* catch output for debugging */
#endif
    SQD_DPRINTF1 (("Spawning slaves...\n"));
    PVMSpawnSlaves ("hmmsearch-pvm", &slave_tid, &nslaves);
    SQD_DPRINTF1 (("Spawned a total of %d slaves...\n", nslaves));

    /* Initialize the slaves by broadcast.
     */
    SQD_DPRINTF1 (("Broadcasting to %d slaves...\n", nslaves));
    pvm_initsend (PvmDataDefault);
    pvm_pkfloat (&(thresh->globT), 1, 1);
    pvm_pkdouble (&(thresh->globE), 1, 1);
    pvm_pkint (&(thresh->Z), 1, 1);
    pvm_pkint (&do_forward, 1, 1);
    pvm_pkint (&do_null2, 1, 1);
    pvm_pkint (&Alphabet_type, 1, 1);
    PVMPackHMM (hmm);
    pvm_mcast (slave_tid, nslaves, HMMPVM_INIT);
    SQD_DPRINTF1 (("Slaves should be ready...\n"));

    /* Confirm slaves' OK status.
     */
    PVMConfirmSlaves (slave_tid, nslaves);
    SQD_DPRINTF1 (("Slaves confirm that they're ok...\n"));

    /* Alloc arrays for remembering what seq each
     * slave was working on.
     */
    namelist = MallocOrDie (sizeof (char *) * nslaves);
    acclist = MallocOrDie (sizeof (char *) * nslaves);
    desclist = MallocOrDie (sizeof (char *) * nslaves);
    dsqlist = MallocOrDie (sizeof (char *) * nslaves);
    lenlist = MallocOrDie (sizeof (int) * nslaves);

    /* Load the slaves.
     * Give them all a sequence number and a digitized sequence
     * to work on.
     * A side effect of the seq number is that we assign each slave
     * a number from 0..nslaves-1.
     */
    for (nseq = 0; nseq < nslaves; nseq++) {
        if (!ReadSeq (sqfp, sqfp->format, &seq, &sqinfo)) {
            break;
        }
        if (sqinfo.len == 0) {
            nseq--;
            continue;
        }

        dsq = DigitizeSequence (seq, sqinfo.len);
        if (do_xnu && Alphabet_type == hmmAMINO) {
            XNU (dsq, sqinfo.len);
        }

        pvm_initsend (PvmDataDefault);
        pvm_pkint (&nseq, 1, 1);
        pvm_pkint (&(sqinfo.len), 1, 1);
        pvm_pkbyte (dsq, sqinfo.len + 2, 1);
        pvm_send (slave_tid[nseq], HMMPVM_WORK);
        SQD_DPRINTF1 (("sent a dsq : %d bytes\n", sqinfo.len + 2));

        namelist[nseq] = Strdup (sqinfo.name);
        acclist[nseq] = (sqinfo.flags & SQINFO_ACC) ? Strdup (sqinfo.acc) : NULL;
        desclist[nseq] =
            (sqinfo.flags & SQINFO_DESC) ? Strdup (sqinfo.desc) : NULL;
        lenlist[nseq] = sqinfo.len;
        dsqlist[nseq] = dsq;

        FreeSequence (seq, &sqinfo);
    }
    SQD_DPRINTF1 (("%d slaves are loaded\n", nseq));

    /* main receive/send loop
     */
    while (ReadSeq (sqfp, sqfp->format, &seq, &sqinfo)) {
        if (sqinfo.len == 0) {
            continue;
        }
        nseq++;
        /* check slaves before blocking */
        PVMCheckSlaves (slave_tid, nslaves);

        /* receive output */
        SQD_DPRINTF1 (("Waiting for a slave to give me output...\n"));
        pvm_recv (-1, HMMPVM_RESULTS);
        pvm_upkint (&slaveidx, 1, 1);	/* # of slave who's sending us stuff */
        pvm_upkfloat (&sc, 1, 1);	/* score   */
        pvm_upkdouble (&pvalue, 1, 1);	/* P-value */
        pvm_upkint (&sent_trace, 1, 1);	/* TRUE if trace is coming */
        tr = (sent_trace) ? PVMUnpackTrace () : NULL;
        SQD_DPRINTF1 (("Slave %d finished %s for me...\n", slaveidx,
                       namelist[slaveidx]));

        /* send new work */
        dsq = DigitizeSequence (seq, sqinfo.len);
        if (do_xnu) {
            XNU (dsq, sqinfo.len);
        }

        pvm_initsend (PvmDataDefault);
        pvm_pkint (&nseq, 1, 1);
        pvm_pkint (&(sqinfo.len), 1, 1);
        pvm_pkbyte (dsq, sqinfo.len + 2, 1);
        pvm_send (slave_tid[slaveidx], HMMPVM_WORK);

        /* process output */
        if (sent_trace) {
            sc = PostprocessSignificantHit (ghit, dhit, tr, hmm, dsqlist[slaveidx], lenlist[slaveidx], namelist[slaveidx], acclist[slaveidx], desclist[slaveidx], do_forward, sc, do_null2, thresh, FALSE);	/* FALSE-> not hmmpfam mode, hmmsearch mode */
            P7FreeTrace (tr);
        }
        AddToHistogram (histogram, sc);

        /* record seq info for seq we just sent */
        free (namelist[slaveidx]);
        if (acclist[slaveidx] != NULL) {
            free (acclist[slaveidx]);
        }
        if (desclist[slaveidx] != NULL) {
            free (desclist[slaveidx]);
        }
        free (dsqlist[slaveidx]);

        dsqlist[slaveidx] = dsq;
        namelist[slaveidx] = Strdup (sqinfo.name);
        acclist[slaveidx] =
            (sqinfo.flags & SQINFO_ACC) ? Strdup (sqinfo.acc) : NULL;
        desclist[slaveidx] =
            (sqinfo.flags & SQINFO_DESC) ? Strdup (sqinfo.desc) : NULL;
        lenlist[slaveidx] = sqinfo.len;

        FreeSequence (seq, &sqinfo);
    }
    SQD_DPRINTF1 (("End of receive/send loop\n"));

    /* Collect the output. All n slaves are still working.
     */
    for (i = 0; i < nslaves && i < nseq; i++) {
        /* don't check slaves (they're exiting normally);
           window of vulnerability here to slave crashes */
        /* receive output */
        pvm_recv (-1, HMMPVM_RESULTS);
        pvm_upkint (&slaveidx, 1, 1);	/* # of slave who's sending us stuff */
        pvm_upkfloat (&sc, 1, 1);	/* score   */
        pvm_upkdouble (&pvalue, 1, 1);	/* P-value */
        pvm_upkint (&sent_trace, 1, 1);	/* TRUE if trace is coming */
        tr = (sent_trace) ? PVMUnpackTrace () : NULL;
        SQD_DPRINTF1 (("Slave %d finished %s for me...\n", slaveidx,
                       namelist[slaveidx]));

        /* process output */
        if (sent_trace) {
            sc = PostprocessSignificantHit (ghit, dhit, tr, hmm, dsqlist[slaveidx], lenlist[slaveidx], namelist[slaveidx], acclist[slaveidx], desclist[slaveidx], do_forward, sc, do_null2, thresh, FALSE);	/* FALSE-> not hmmpfam mode, hmmsearch mode */
            P7FreeTrace (tr);
        }
        SQD_DPRINTF2 (("AddToHistogram: %s\t%f\n", namelist[slaveidx], sc));
        AddToHistogram (histogram, sc);

        /* free seq info */
        free (namelist[slaveidx]);
        if (acclist[slaveidx] != NULL) {
            free (acclist[slaveidx]);
        }
        if (desclist[slaveidx] != NULL) {
            free (desclist[slaveidx]);
        }
        free (dsqlist[slaveidx]);

        /* send cleanup/shutdown flag to slave */
        pvm_initsend (PvmDataDefault);
        code = -1;
        pvm_pkint (&code, 1, 1);
        pvm_send (slave_tid[slaveidx], HMMPVM_WORK);
    }


    /* Cleanup; quit the VM; and return
     */
    free (slave_tid);
    free (dsqlist);
    free (namelist);
    free (acclist);
    free (desclist);
    free (lenlist);
    pvm_exit ();
    *ret_nseq = nseq;
    return;
}
#else /* HMMER_PVM off, no PVM support, dummy function: */
static void
main_loop_pvm (struct plan7_s *hmm, SQFILE * sqfp, struct threshold_s *thresh,
               int do_forward, int do_null2, int do_xnu,
               struct histogram_s *histogram, struct tophit_s *ghit,
               struct tophit_s *dhit, int *ret_nseq)
{
    Die ("No PVM support");
}
#endif /*HMMER_PVM */

#ifdef HMMER_THREADS
/*****************************************************************
 * POSIX threads implementation.
 *
 * API:
 *      workpool_start()   (makes a workpool_s structure. Starts calculations.)
 *      workpool_stop()    (waits for threads to finish.)
 *      workpool_free()    (destroys the structure)
 *
 * Threads:
 *      worker_thread()    (the actual parallelized worker thread).
 *****************************************************************/
/* Function: main_loop_threaded
 * Date:     SRE, Wed Feb 27 11:38:11 2002 [St. Louis]
 *
 * Purpose:  Search an HMM against a sequence database.
 *           main loop for the threaded version.
 *
 *           In:   HMM and open sqfile, plus options
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with.
 *           sqfp       - open SQFILE for sequence database
 *           thresh     - score/evalue threshold info
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           num_threads- number of worker threads to start, >=1
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *
 * Returns:  (void)
 */
static void
main_loop_threaded (struct plan7_s *hmm, SQFILE * sqfp,
                    struct threshold_s *thresh, int do_forward, int do_null2,
                    int do_xnu, int num_threads,
                    struct histogram_s *histogram, struct tophit_s *ghit,
                    struct tophit_s *dhit, int *ret_nseq)
{
    struct workpool_s *wpool;	/* pool of worker threads                  */
    int nseq;			/* number of sequences searched            */

    wpool = workpool_start (hmm, sqfp, do_xnu, do_forward, do_null2, thresh,
                            ghit, dhit, histogram, num_threads);
    workpool_stop (wpool);
    nseq = wpool->nseq;
    workpool_free (wpool);

    *ret_nseq = nseq;
    return;
}

/* Function: workpool_start()
 * Date:     SRE, Mon Oct  5 16:44:53 1998
 *
 * Purpose:  Initialize a workpool_s structure, and return it.
 *
 * Args:     sqfp       - open sequence file, at start
 *           do_xnu     - TRUE to apply XNU filter
 *           do_forward - TRUE to score using Forward
 *           do_null2   - TRUE to apply null2 ad hoc correction
 *           thresh     - score/evalue threshold info
 *           ghit       - per-seq hit list
 *           dhit       - per-domain hit list
 *           hist       - histogram (alloced but empty)
 *           num_threads- number of worker threads to run.
 *
 * Returns:  ptr to struct workpool_s.
 *           Caller must wait for threads to finish with workpool_stop(),
 *           then free the structure with workpool_free().
 */
static struct workpool_s *
workpool_start (struct plan7_s *hmm, SQFILE * sqfp, int do_xnu,
                int do_forward, int do_null2, struct threshold_s *thresh,
                struct tophit_s *ghit, struct tophit_s *dhit,
                struct histogram_s *hist, int num_threads)
{
    struct workpool_s *wpool;
    pthread_attr_t attr;
    int i;
    int rtn;

    wpool = MallocOrDie (sizeof (struct workpool_s));
    wpool->thread = MallocOrDie (num_threads * sizeof (pthread_t));
    wpool->hmm = hmm;

    wpool->do_xnu = do_xnu;
    wpool->do_forward = do_forward;
    wpool->do_null2 = do_null2;
    wpool->thresh = thresh;

    wpool->sqfp = sqfp;
    wpool->nseq = 0;
    if ((rtn = pthread_mutex_init (&(wpool->input_lock), NULL)) != 0) {
        Die ("pthread_mutex_init FAILED; %s\n", strerror (rtn));
    }

    wpool->ghit = ghit;
    wpool->dhit = dhit;
    wpool->hist = hist;
    if ((rtn = pthread_mutex_init (&(wpool->output_lock), NULL)) != 0) {
        Die ("pthread_mutex_init FAILED; %s\n", strerror (rtn));
    }

    wpool->num_threads = num_threads;

    /* Create slave threads. See comments in hmmcalibrate.c at this
     * step, regarding concurrency, system scope, and portability
     * amongst various UNIX implementations of pthreads.
     */
    pthread_attr_init (&attr);
#ifndef __sgi
#ifdef HAVE_PTHREAD_ATTR_SETSCOPE
    pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
#endif
#endif
#ifdef HAVE_PTHREAD_SETCONCURRENCY
    pthread_setconcurrency (num_threads + 1);
#endif
    /* pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM); */
    for (i = 0; i < num_threads; i++) {
        if ((rtn = pthread_create (&(wpool->thread[i]), &attr,
                                   worker_thread, (void *) wpool)) != 0) {
            Die ("Failed to create thread %d; return code %d\n", i, rtn);
        }
    }
    pthread_attr_destroy (&attr);
    return wpool;
}

/* Function: workpool_stop()
 * Date:     SRE, Thu Jul 16 11:20:16 1998 [St. Louis]
 *
 * Purpose:  Waits for threads in a workpool to finish.
 *
 * Args:     wpool -- ptr to the workpool structure
 *
 * Returns:  (void)
 */
static void
workpool_stop (struct workpool_s *wpool)
{
    int i;
    /* wait for threads to stop */
    for (i = 0; i < wpool->num_threads; i++) {
        if (pthread_join (wpool->thread[i], NULL) != 0) {
            Die ("pthread_join failed");
        }
    }
    return;
}

/* Function: workpool_free()
 * Date:     SRE, Thu Jul 16 11:26:27 1998 [St. Louis]
 *
 * Purpose:  Free a workpool_s structure, after the threads
 *           have finished.
 *
 * Args:     wpool -- ptr to the workpool.
 *
 * Returns:  (void)
 */
static void
workpool_free (struct workpool_s *wpool)
{
    free (wpool->thread);
    free (wpool);
    return;
}


/* Function: worker_thread()
 * Date:     SRE, Mon Sep 28 10:48:29 1998 [St. Louis]
 *
 * Purpose:  The procedure executed by the worker threads.
 *
 * Args:     ptr  - (void *) that is recast to a pointer to
 *                  the workpool.
 *
 * Returns:  (void *)
 */
void *
worker_thread (void *ptr)
{
    struct workpool_s *wpool;	/* our working threads structure   */
    char *seq;			/* target sequence                 */
    SQINFO sqinfo;		/* information assoc w/ seq        */
    char *dsq;			/* digitized sequence              */
    struct dpmatrix_s *mx;	/* growable DP matrix              */
    struct p7trace_s *tr;		/* traceback from an alignment     */
    float sc;			/* score of an alignment           */
    int rtn;			/* a return code from pthreads lib */
    double pvalue;		/* P-value of score                */
    double evalue;		/* E-value of score                */

    wpool = (struct workpool_s *) ptr;

    /* Init with a small DP matrix; we'll grow in the sequence dimension
     * overalloc'ing by 25 rows (residues).
     */
    mx = CreatePlan7Matrix (1, wpool->hmm->M, 25, 0);
    for (;;) {

        /* 1. acquire lock on sequence input, and get
         *    the next seq to work on.
         */
        /* acquire a lock */
        if ((rtn = pthread_mutex_lock (&(wpool->input_lock))) != 0) {
            Die ("pthread_mutex_lock failure: %s\n", strerror (rtn));
        }
        if (!ReadSeq (wpool->sqfp, wpool->sqfp->format, &seq, &sqinfo)) {	/* we're done. release lock, exit thread */
            if ((rtn = pthread_mutex_unlock (&(wpool->input_lock))) != 0) {
                Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
            }
            FreePlan7Matrix (mx);
            pthread_exit (NULL);
        }
        SQD_DPRINTF1 (("a thread is working on %s\n", sqinfo.name));
        wpool->nseq++;
        /* release the lock */
        if ((rtn = pthread_mutex_unlock (&(wpool->input_lock))) != 0) {
            Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
        }

        if (sqinfo.len == 0) {
            continue;    /* silent skip of len=0 seqs (wormpep!?!) */
        }

        dsq = DigitizeSequence (seq, sqinfo.len);
        if (wpool->do_xnu) {
            XNU (dsq, sqinfo.len);
        }

        /* 1. Recover a trace by Viterbi.
         */
        if (P7ViterbiSize (sqinfo.len, wpool->hmm->M) <= RAMLIMIT) {
            sc = P7Viterbi (dsq, sqinfo.len, wpool->hmm, mx, &tr);
        }
        else {
            sc = P7SmallViterbi (dsq, sqinfo.len, wpool->hmm, mx, &tr);
        }

        /* 2. If we're using Forward scores, do another DP
         *    to get it; else, we already have a Viterbi score
         *    in sc.
         */
        if (wpool->do_forward) {
            sc = P7Forward (dsq, sqinfo.len, wpool->hmm, NULL);
            if (wpool->do_null2) {
                sc -= TraceScoreCorrection (wpool->hmm, tr, dsq);
            }
        }

        /* 3. Save the output in tophits and histogram structures, after acquiring a lock
         */
        if ((rtn = pthread_mutex_lock (&(wpool->output_lock))) != 0) {
            Die ("pthread_mutex_lock failure: %s\n", strerror (rtn));
        }
        SQD_DPRINTF1 (("seq %s scores %f\n", sqinfo.name, sc));

        pvalue = PValue (wpool->hmm, sc);
        evalue =
            wpool->thresh->Z ? (double) wpool->thresh->Z *
            pvalue : (double) wpool->nseq * pvalue;

        if (sc >= wpool->thresh->globT && evalue <= wpool->thresh->globE) {
            sc = PostprocessSignificantHit (wpool->ghit, wpool->dhit, tr, wpool->hmm, dsq, sqinfo.len, sqinfo.name, sqinfo.flags & SQINFO_ACC ? sqinfo.acc : NULL, sqinfo.flags & SQINFO_DESC ? sqinfo.desc : NULL, wpool->do_forward, sc, wpool->do_null2, wpool->thresh, FALSE);	/* FALSE-> not hmmpfam mode, hmmsearch mode */
        }
        SQD_DPRINTF2 (("AddToHistogram: %s\t%f\n", sqinfo.name, sc));
        AddToHistogram (wpool->hist, sc);
        if ((rtn = pthread_mutex_unlock (&(wpool->output_lock))) != 0) {
            Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
        }

        P7FreeTrace (tr);
        FreeSequence (seq, &sqinfo);
        free (dsq);
    }				/* end 'infinite' loop over seqs in this thread */
}
#else /*HMMER_THREADS off; no threads support; dummy stub: */
static void
main_loop_threaded (struct plan7_s *hmm, SQFILE * sqfp,
                    struct threshold_s *thresh, int do_forward, int do_null2,
                    int do_xnu, int num_threads,
                    struct histogram_s *histogram, struct tophit_s *ghit,
                    struct tophit_s *dhit, int *ret_nseq)
{
    Die ("No threads support");
}
#endif /* HMMER_THREADS */

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/
/* @(#)qsort.c	4.2 (Berkeley) 3/9/83 */

/*
 * qsort.c:
 * Our own version of the system qsort routine which is faster by an average
 * of 25%, with lows and highs of 10% and 50%.
 * The THRESHold below is the insertion sort threshold, and has been adjusted
 * for records of size 48 bytes.
 * The MTHREShold is where we stop finding a better median.
 */

#define		THRESH		4	/* threshold for insertion */
#define		MTHRESH		6	/* threshold for median */

static int (*qcmp) ();		/* the comparison routine */
static int qsz;			/* size of each record */
static int thresh;		/* THRESHold in chars */
static int mthresh;		/* MTHRESHold in chars */

static void qst (char *, char *);	/* function prototype RRH */

/*
 * qsort:
 * First, set up some global parameters for qst to share.  Then, quicksort
 * with qst(), and then a cleanup insertion sort ourselves.  Sound simple?
 * It's not...
 */
#undef min
#undef max
void
specqsort (base, n, size, compar)
char *base;
int n;
int size;
int (*compar) ();
{
    register char c, *i, *j, *lo, *hi;
    char *min, *max;

    if (n <= 1) {
        return;
    }
    qsz = size;
    qcmp = compar;
    thresh = qsz * THRESH;
    mthresh = qsz * MTHRESH;
    max = base + n * qsz;
    if (n >= THRESH) {
        qst (base, max);
        hi = base + thresh;
    }
    else {
        hi = max;
    }
    /*
     * First put smallest element, which must be in the first THRESH, in
     * the first position as a sentinel.  This is done just by searching
     * the first THRESH elements (or the first n if n < THRESH), finding
     * the min, and swapping it into the first position.
     */
    for (j = lo = base; (lo += qsz) < hi;) {
        if ((*qcmp) (j, lo) > 0) {
            j = lo;
        }
    }
    if (j != base) {
        /* swap j into place */
        for (i = base, hi = base + qsz; i < hi;) {
            c = *j;
            *j++ = *i;
            *i++ = c;
        }
    }
    /*
     * With our sentinel in place, we now run the following hyper-fast
     * insertion sort. For each remaining element, min, from [1] to [n-1],
     * set hi to the index of the element AFTER which this one goes.
     * Then, do the standard insertion sort shift on a character at a time
     * basis for each element in the frob.
     */
    for (min = base; (hi = min += qsz) < max;) {
        while ((*qcmp) (hi -= qsz, min) > 0)
            /* void */ ;
        if ((hi += qsz) != min) {
            for (lo = min + qsz; --lo >= min;) {
                c = *lo;
                for (i = j = lo; (j -= qsz) >= hi; i = j) {
                    *i = *j;
                }
                *i = c;
            }
        }
    }
}

/*
 * qst:
 * Do a quicksort
 * First, find the median element, and put that one in the first place as the
 * discriminator.  (This "median" is just the median of the first, last and
 * middle elements).  (Using this median instead of the first element is a big
 * win).  Then, the usual partitioning/swapping, followed by moving the
 * discriminator into the right place.  Then, figure out the sizes of the two
 * partions, do the smaller one recursively and the larger one via a repeat of
 * this code.  Stopping when there are less than THRESH elements in a partition
 * and cleaning up with an insertion sort (in our caller) is a huge win.
 * All data swaps are done in-line, which is space-losing but time-saving.
 * (And there are only three places where this is done).
 */

static void			/* added Kaivalya 8/18/03 */
qst (char *base, char *max)
{   /* changed Kaivalya/RRH 8/19/03 */
    register char c, *i, *j, *jj;
    register int ii;
    char *mid, *tmp;
    int lo, hi;

    /*
     * At the top here, lo is the number of characters of elements in the
     * current partition.  (Which should be max - base).
     * Find the median of the first, last, and middle element and make
     * that the middle element.  Set j to largest of first and middle.
     * If max is larger than that guy, then it's that guy, else compare
     * max with loser of first and take larger.  Things are set up to
     * prefer the middle, then the first in case of ties.
     */
    lo = max - base;		/* number of elements as chars */
    do {
        mid = i = base + qsz * ((lo / qsz) >> 1);
        if (lo >= mthresh) {
            j = ((*qcmp) ((jj = base), i) > 0 ? jj : i);
            if ((*qcmp) (j, (tmp = max - qsz)) > 0) {
                /* switch to first loser */
                j = (j == jj ? i : jj);
                if ((*qcmp) (j, tmp) < 0) {
                    j = tmp;
                }
            }
            if (j != i) {
                ii = qsz;
                do {
                    c = *i;
                    *i++ = *j;
                    *j++ = c;
                } while (--ii);
            }
        }
        /*
         * Semi-standard quicksort partitioning/swapping
         */
        for (i = base, j = max - qsz;;) {
            while (i < mid && (*qcmp) (i, mid) <= 0) {
                i += qsz;
            }
            while (j > mid) {
                if ((*qcmp) (mid, j) <= 0) {
                    j -= qsz;
                    continue;
                }
                tmp = i + qsz;		/* value of i after swap */
                if (i == mid) {
                    /* j <-> mid, new mid is j */
                    mid = jj = j;
                }
                else {
                    /* i <-> j */
                    jj = j;
                    j -= qsz;
                }
                goto swap;
            }
            if (i == mid) {
                break;
            }
            else {
                /* i <-> mid, new mid is i */
                jj = mid;
                tmp = mid = i;		/* value of i after swap */
                j -= qsz;
            }
swap:
            ii = qsz;
            do {
                c = *i;
                *i++ = *jj;
                *jj++ = c;
            } while (--ii);
            i = tmp;
        }
        /*
         * Look at sizes of the two partitions, do the smaller
         * one first by recursion, then do the larger one by
         * making sure lo is its size, base and max are update
         * correctly, and branching back.  But only repeat
         * (recursively or by branching) if the partition is
         * of at least size THRESH.
         */
        i = (j = mid) + qsz;
        if ((lo = j - base) <= (hi = max - i)) {
            if (lo >= thresh) {
                qst (base, j);
            }
            base = i;
            lo = hi;
        }
        else {
            if (hi >= thresh) {
                qst (i, max);
            }
            max = j;
        }
    } while (lo >= thresh);
}

/* a2m.c
 *
 * reading/writing A2M (aligned FASTA) files.
 *
 * CVS $Id: a2m.c,v 1.2 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "squid.h"
#include "msa.h"

/* Function: ReadA2M()
 * Date:     SRE, Sun Jun  6 17:11:29 1999 [bus from Madison 1999 worm mtg]
 *
 * Purpose:  Parse an alignment read from an open A2M format
 *           alignment file. A2M is a single alignment format.
 *           Return the alignment, or NULL if we've already
 *           read the alignment.
 *
 * Args:     afp - open alignment file
 *
 * Returns:  MSA *  - an alignment object.
 *                    Caller responsible for an MSAFree()
 */
MSA *
ReadA2M (MSAFILE * afp)
{
    MSA *msa;
    char *buf;
    char *name;
    char *desc;
    char *seq;
    int idx;
    int len1, len2;

    if (feof (afp->f)) {
        return NULL;
    }

    name = NULL;
    msa = MSAAlloc (10, 0);
    idx = 0;
    while ((buf = MSAFileGetLine (afp)) != NULL) {
        if (*buf == '>') {
            buf++;			/* skip the '>' */
            if ((name = sre_strtok (&buf, WHITESPACE, &len1)) == NULL)
                Die ("Blank name in A2M file %s (line %d)\n", afp->fname,
                     afp->linenumber);
            desc = sre_strtok (&buf, "\n", &len2);

            idx = GKIStoreKey (msa->index, name);
            if (idx >= msa->nseqalloc) {
                MSAExpand (msa);
            }

            msa->sqname[idx] = sre_strdup (name, len1);
            if (desc != NULL) {
                MSASetSeqDescription (msa, idx, desc);
            }
            msa->nseq++;
        }
        else if (name != NULL) {
            if ((seq = sre_strtok (&buf, WHITESPACE, &len1)) == NULL) {
                continue;
            }
            msa->sqlen[idx] =
                sre_strcat (&(msa->aseq[idx]), msa->sqlen[idx], seq, len1);
        }
    }
    if (name == NULL) {
        MSAFree (msa);
        return NULL;
    }

    MSAVerifyParse (msa);
    return msa;
}


/* Function: WriteA2M()
 * Date:     SRE, Sun Jun  6 17:40:35 1999 [bus from Madison, 1999 worm mtg]
 *
 * Purpose:  Write an "aligned FASTA" (aka a2m, to UCSC) formatted
 *           alignment.
 *
 * Args:     fp    - open FILE to write to.
 *           msa   - alignment to write
 *
 * Returns:  void
 */
void
WriteA2M (FILE * fp, MSA * msa)
{
    int idx;			/* sequence index */
    int pos;			/* position in sequence */
    char buf[64];			/* buffer for individual lines */
    int cpl = 60;			/* char per line; must be < 64 unless buf is bigger */

    buf[cpl] = '\0';
    for (idx = 0; idx < msa->nseq; idx++) {
        fprintf (fp, ">%s %s\n",
                 msa->sqname[idx],
                 (msa->sqdesc != NULL
                  && msa->sqdesc[idx] != NULL) ? msa->sqdesc[idx] : "");
        for (pos = 0; pos < msa->alen; pos += cpl) {
            strncpy (buf, &(msa->aseq[idx][pos]), cpl);
            fprintf (fp, "%s\n", buf);
        }
    }
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* aligneval.c
 *
 * Comparison of multiple alignments. Three functions are
 * provided, using subtly different scoring schemes:
 *    CompareMultAlignments()    - basic scoring scheme
 *    CompareRefMultAlignments() - only certain "canonical" columns
 *                                 are scored
 *
 * The similarity measure is a fractional alignment identity averaged
 * over all sequence pairs. The score for all pairs is:
 *      (identically aligned symbols) / (total aligned columns in
 *      known alignment)
 *
 * A column c is identically aligned for sequences i, j if:
 *    1) both i,j have a symbol aligned in column c, and the
 *       same pair of symbols is aligned somewhere in the test
 *       alignment
 *    2) S[i][c] is aligned to a gap in sequence j, and that symbol
 *       is aligned to a gap in the test alignment
 *    3) converse of 2)
 *
 *
 * The algorithm is as follows:
 *    1) For each known/test aligned pair of sequences (k1,k2 and t1,t2)
 *        construct a list for each sequence, in which for every
 *        counted symbol we record the raw index of the symbol in
 *        the other sequence that it aligns to, or -1 if it aligns
 *        to a gap or uncounted symbol.
 *
 *    2)  Compare the list for k1 to the list for t1 and count an identity
 *        for each correct alignment.
 *
 *    3) Repeat 2) for comparing k2 to t2. Note that this means correct sym/sym
 *       alignments count for 2; correct sym/gap alignments count for 1.
 *
 *    4) The score is (identities from 2 + identities from 3) /
 *       (totals from 2 + totals from 3).
 *
 * Written originally for koala's ss2 pairwise alignment package.
 *
 * Sean Eddy, Sun Nov  1 12:45:11 1992
 * SRE, Thu Jul 29 16:47:18 1993: major revision: all functions replaced by new algorithm
 * CVS $Id: aligneval.c,v 1.9 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"
#include "sre_random.h"

static int make_alilist (char *s1, char *s2, int **ret_s1_list,
                         int *ret_listlen);
static int make_ref_alilist (int *refcoords, char *k1, char *k2, char *s1,
                             char *s2, int **ret_s1_list, int *ret_listlen);
static int compare_lists (int *k1, int *k2, int *t1, int *t2, int len1,
                          int len2, float *ret_sc);


/* Function: ComparePairAlignments
 *
 * Purpose:  Calculate and return a number representing how well two different alignments
 *           of a pair of sequences compare. The number is, roughly speaking,
 *           the fraction of columns which are identically aligned.
 *
 *           For all columns c in which either known1[c] or known2[c]
 *           is a non-gap, count an identity if those same symbols are
 *           aligned somewhere in calc1/calc2. The score is identities/total
 *           columns examined. (i.e. fully gapped columns don't count)
 *
 *           more explicitly, identities come from:
 *             both known and test aligned pairs have the same symbol in the first sequence aligned to
 *               a gap in the second sequence;
 *             both known and test aligned pairs have the same symbol in the second sequence
 *               aligned to a gap in the first sequence;
 *             the known alignment has symbols aligned at this column, and the test
 *               alignment aligns the same two symbols.
 *
 * Args:     known1, known2: trusted alignment of two sequences
 *           calc1, calc2:   test alignment of two sequences
 *
 * Return:   Returns -1.0 on internal failure.
 */
float
ComparePairAlignments (char *known1, char *known2, char *calc1, char *calc2)
{
    int *klist1;
    int *klist2;
    int *tlist1;
    int *tlist2;
    int len1, len2;
    float score;

    if (!make_alilist (calc1, calc2, &tlist1, &len1)) {
        return -1.0;
    }
    if (!make_alilist (calc2, calc1, &tlist2, &len2)) {
        return -1.0;
    }
    if (!make_alilist (known1, known2, &klist1, &len1)) {
        return -1.0;
    }
    if (!make_alilist (known2, known1, &klist2, &len2)) {
        return -1.0;
    }
    if (!compare_lists (klist1, klist2, tlist1, tlist2, len1, len2, &score)) {
        return -1.0;
    }

    free (klist1);
    free (klist2);
    free (tlist1);
    free (tlist2);
    return score;
}



/* Function: CompareRefPairAlignments()
 *
 * Same as above, but the only columns that count are the ones
 * with indices in *refcoord. *refcoord and the known1, known2
 * pair must be in sync with each other (come from the same
 * multiple sequence alignment)
 *
 * Args:     ref           - 0..alen-1 array of 1 or 0
 *           known1,known2 - trusted alignment
 *           calc1, calc2  - test alignment
 *
 * Return:  the fractional alignment identity on success, -1.0 on failure.
 */
float
CompareRefPairAlignments (int *ref, char *known1, char *known2, char *calc1,
                          char *calc2)
{
    int *klist1;
    int *klist2;
    int *tlist1;
    int *tlist2;
    int len1, len2;
    float score;

    if (!make_ref_alilist (ref, known1, known2, calc1, calc2, &tlist1, &len1)) {
        return -1.0;
    }
    if (!make_ref_alilist (ref, known2, known1, calc2, calc1, &tlist2, &len2)) {
        return -1.0;
    }
    if (!make_ref_alilist (ref, known1, known2, known1, known2, &klist1, &len1)) {
        return -1.0;
    }
    if (!make_ref_alilist (ref, known2, known1, known2, known1, &klist2, &len2)) {
        return -1.0;
    }
    if (!compare_lists (klist1, klist2, tlist1, tlist2, len1, len2, &score)) {
        return -1.0;
    }

    free (klist1);
    free (klist2);
    free (tlist1);
    free (tlist2);
    return score;
}

/* Function: make_alilist()
 *
 * Purpose:  Construct a list (array) mapping the raw symbols of s1
 *           onto the indexes of the aligned symbols in s2 (or -1
 *           for gaps in s2). The list (s1_list) will be of the
 *           length of s1's raw sequence.
 *
 * Args:     s1          - sequence to construct the list for
 *           s2          - sequence s1 is aligned to
 *           ret_s1_list - RETURN: the constructed list (caller must free)
 *           ret_listlen - RETURN: length of the list
 *
 * Returns:  1 on success, 0 on failure
 */
static int
make_alilist (char *s1, char *s2, int **ret_s1_list, int *ret_listlen)
{
    int *s1_list;
    int col;			/* column position in alignment */
    int r1, r2;			/* raw symbol index at current col in s1, s2 */

    /* Malloc for s1_list. It can't be longer than s1 itself; we just malloc
     * for that (and waste a wee bit of space)
     */
    s1_list = (int *) MallocOrDie (sizeof (int) * strlen (s1));
    r1 = r2 = 0;
    for (col = 0; s1[col] != '\0'; col++) {
        /* symbol in s1? Record what it's aligned to, and bump
         * the r1 counter.
         */
        if (!isgap (s1[col])) {
            s1_list[r1] = isgap (s2[col]) ? -1 : r2;
            r1++;
        }

        /* symbol in s2? bump the r2 counter
         */
        if (!isgap (s2[col])) {
            r2++;
        }
    }

    *ret_listlen = r1;
    *ret_s1_list = s1_list;
    return 1;
}



/* Function: make_ref_alilist()
 *
 * Purpose:  Construct a list (array) mapping the raw symbols of s1
 *           which are under canonical columns of the ref alignment
 *           onto the indexes of the aligned symbols in s2 (or -1
 *           for gaps in s2 or noncanonical symbols in s2).
 *
 * Args:     ref:        - array of indices of canonical coords (1 canonical, 0 non)
 *           k1          - s1's known alignment (w/ respect to refcoords)
 *           k2          - s2's known alignment (w/ respect to refcoords)
 *           s1          - sequence to construct the list for
 *           s2          - sequence s1 is aligned to
 *           ret_s1_list - RETURN: the constructed list (caller must free)
 *           ret_listlen - RETURN: length of the list
 *
 * Returns:  1 on success, 0 on failure
 */
/*ARGSUSED*/ static int
make_ref_alilist (int *ref, char *k1, char *k2,
                  char *s1, char *s2, int **ret_s1_list, int *ret_listlen)
{
    int *s1_list;
    int col;			/* column position in alignment */
    int r1, r2;			/* raw symbol index at current col in s1, s2 */
    int *canons1;			/* flag array, 1 if position i in s1 raw seq is canonical */
    int lpos;			/* position in list */

    /* Allocations. No arrays can exceed the length of their
     * appropriate parent (s1 or s2)
     */
    s1_list = (int *) MallocOrDie (sizeof (int) * strlen (s1));
    canons1 = (int *) MallocOrDie (sizeof (int) * strlen (s1));

    /* First we use refcoords and k1,k2 to construct an array of 1's
     * and 0's, telling us whether s1's raw symbol number i is countable.
     * It's countable simply if it's under a canonical column.
     */
    r1 = 0;
    for (col = 0; k1[col] != '\0'; col++) {
        if (!isgap (k1[col])) {
            canons1[r1] = ref[col] ? 1 : 0;
            r1++;
        }
    }

    /* Now we can construct the list. We don't count pairs if the sym in s1
     * is non-canonical.
     * We have to keep separate track of our position in the list (lpos)
     * from our positions in the raw sequences (r1,r2)
     */
    r1 = r2 = lpos = 0;
    for (col = 0; s1[col] != '\0'; col++) {
        if (!isgap (s1[col]) && canons1[r1]) {
            s1_list[lpos] = isgap (s2[col]) ? -1 : r2;
            lpos++;
        }

        if (!isgap (s1[col])) {
            r1++;
        }
        if (!isgap (s2[col])) {
            r2++;
        }
    }

    free (canons1);
    *ret_listlen = lpos;
    *ret_s1_list = s1_list;
    return 1;
}

/* Function: compare_lists()
 *
 * Purpose:  Given four alignment lists (k1,k2, t1,t2), calculate the
 *           alignment score.
 *
 * Args:     k1   - list of k1's alignment to k2
 *           k2   - list of k2's alignment to k1
 *           t1   - list of t1's alignment to t2
 *           t2   - list of t2's alignment to t2
 *           len1 - length of k1, t1 lists (same by definition)
 *           len2 - length of k2, t2 lists (same by definition)
 *           ret_sc - RETURN: identity score of alignment
 *
 * Return:   1 on success, 0 on failure.
 */
static int
compare_lists (int *k1, int *k2, int *t1, int *t2, int len1, int len2,
               float *ret_sc)
{
    float id;
    float tot;
    int i;

    id = tot = 0.0;
    for (i = 0; i < len1; i++) {
        tot += 1.0;
        if (t1[i] == k1[i]) {
            id += 1.0;
        }
    }

    for (i = 0; i < len2; i++) {
        tot += 1.0;
        if (k2[i] == t2[i]) {
            id += 1.0;
        }
    }

    *ret_sc = id / tot;
    return 1;
}


/* Function: CompareMultAlignments
 *
 * Purpose:  Invokes pairwise alignment comparison for every possible pair,
 *           and returns the average score over all N(N-1) of them or -1.0
 *           on an internal failure.
 *
 *           Can be slow for large N, since it's quadratic.
 *
 * Args:     kseqs  - trusted multiple alignment
 *           tseqs  - test multiple alignment
 *           N      - number of sequences
 *
 * Return:   average identity score, or -1.0 on failure.
 */
float
CompareMultAlignments (char **kseqs, char **tseqs, int N)
{
    int i, j;			/* counters for sequences */
    float score;
    float tot_score = 0.0;
    /* do all pairwise comparisons */
    for (i = 0; i < N; i++) {
        for (j = i + 1; j < N; j++) {
            score = ComparePairAlignments (kseqs[i], kseqs[j], tseqs[i], tseqs[j]);
            if (score < 0.0) {
                return -1.0;
            }
            tot_score += score;
        }
    }
    return ((tot_score * 2.0) / ((float) N * ((float) N - 1.0)));
}



/* Function: CompareRefMultAlignments()
 *
 * Purpose:  Same as above, except an array of reference coords for
 *           the canonical positions of the known alignment is also
 *           provided.
 *
 * Args:     ref      : 0..alen-1 array of 1/0 flags, 1 if canon
 *           kseqs    : trusted alignment
 *           tseqs    : test alignment
 *           N        : number of sequences
 *
 * Return:   average identity score, or -1.0 on failure
 */
float
CompareRefMultAlignments (int *ref, char **kseqs, char **tseqs, int N)
{
    int i, j;			/* counters for sequences */
    float score;
    float tot_score = 0.0;

    /* do all pairwise comparisons */
    for (i = 0; i < N; i++) {
        for (j = i + 1; j < N; j++) {
            score =
                CompareRefPairAlignments (ref, kseqs[i], kseqs[j], tseqs[i],
                                          tseqs[j]);
            if (score < 0.0) {
                return -1.0;
            }
            tot_score += score;
        }
    }
    return ((tot_score * 2.0) / ((float) N * ((float) N - 1.0)));
}

/* Function: PairwiseIdentity()
 *
 * Purpose:  Calculate the pairwise fractional identity between
 *           two aligned sequences s1 and s2. This is simply
 *           (idents / MIN(len1, len2)).
 *
 *           Note how many ways there are to calculate pairwise identity,
 *           because of the variety of choices for the denominator:
 *           idents/(idents+mismat) has the disadvantage that artifactual
 *             gappy alignments would have high "identities".
 *           idents/(AVG|MAX)(len1,len2) both have the disadvantage that
 *             alignments of fragments to longer sequences would have
 *             artifactually low "identities".
 *
 *           Case sensitive; also, watch out in nucleic acid alignments;
 *           U/T RNA/DNA alignments will be counted as mismatches!
 */
float
PairwiseIdentity (char *s1, char *s2)
{
    int idents;			/* total identical positions  */
    int len1, len2;		/* lengths of seqs            */
    int x;			/* position in aligned seqs   */

    idents = len1 = len2 = 0;
    for (x = 0; s1[x] != '\0' && s2[x] != '\0'; x++) {
        if (!isgap (s1[x])) {
            len1++;
            if (s1[x] == s2[x]) {
                idents++;
            }
        }
        if (!isgap (s2[x])) {
            len2++;
        }
    }
    if (len2 < len1) {
        len1 = len2;
    }
    return (len1 == 0 ? 0.0 : (float) idents / (float) len1);
}



/* Function: AlignmentIdentityBySampling()
 * Date:     SRE, Mon Oct 19 14:29:01 1998 [St. Louis]
 *
 * Purpose:  Estimate and return the average pairwise
 *           fractional identity of an alignment,
 *           using sampling.
 *
 *           For use when there's so many sequences that
 *           an all vs. all rigorous calculation will
 *           take too long.
 *
 *           Case sensitive!
 *
 * Args:     aseq       - aligned sequences
 *           L          - length of alignment
 *           N          - number of seqs in alignment
 *           nsample    - number of samples
 *
 * Returns:  average fractional identity, 0..1.
 */
float
AlignmentIdentityBySampling (char **aseq, int L, int N, int nsample)
{
    int x, i, j;			/* counters */
    float sum;

    if (N < 2) {
        return 1.0;
    }

    sum = 0.;
    for (x = 0; x < nsample; x++) {
        i = CHOOSE (N);
        do {
            j = CHOOSE (N);
        } while (j == i);		/* make sure j != i */
        sum += PairwiseIdentity (aseq[i], aseq[j]);
    }
    return sum / (float) nsample;
}

/* Function: MajorityRuleConsensus()
 * Date:     SRE, Tue Mar  7 15:30:30 2000 [St. Louis]
 *
 * Purpose:  Given a set of aligned sequences, produce a
 *           majority rule consensus sequence. If >50% nonalphabetic
 *           (usually meaning gaps) in the column, ignore the column.
 *
 * Args:     aseq  - aligned sequences, [0..nseq-1][0..alen-1]
 *           nseq  - number of sequences
 *           alen  - length of alignment
 *
 * Returns:  ptr to allocated consensus sequence.
 *           Caller is responsible for free'ing this.
 */
char *
MajorityRuleConsensus (char **aseq, int nseq, int alen)
{
    char *cs;			/* RETURN: consensus sequence */
    int count[27];		/* counts for a..z and gaps in a column */
    int idx, apos;		/* counters for seq, column */
    int spos;			/* position in cs */
    int x;			/* counter for characters */
    int sym;
    int max, bestx;

    cs = MallocOrDie (sizeof (char) * (alen + 1));

    for (spos = 0, apos = 0; apos < alen; apos++) {
        for (x = 0; x < 27; x++) {
            count[x] = 0;
        }

        for (idx = 0; idx < nseq; idx++) {
            if (isalpha ((int) aseq[idx][apos])) {
                sym = toupper ((int) aseq[idx][apos]);
                count[sym - 'A']++;
            }
            else {
                count[26]++;
            }
        }

        if ((float) count[26] / (float) nseq <= 0.5) {
            max = bestx = -1;
            for (x = 0; x < 26; x++) {
                if (count[x] > max) {
                    max = count[x];
                    bestx = x;
                }
						}		
            cs[spos++] = (char) ('A' + bestx);
        }
    }
    cs[spos] = '\0';
    return cs;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* alignio.c
 * SRE, Mon Jul 12 11:57:37 1993
 * CVS $Id: alignio.c,v 1.12 2003/04/14 16:00:16 eddy Exp $
 *
 * Input/output of sequence alignments.
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"
#include "sre_random.h"

/* Function: AllocAlignment()
 *
 * Purpose:  Allocate space for an alignment, given the number
 *           of sequences and the alignment length in columns.
 *
 * Args:     nseq     - number of sequences
 *           alen     - width of alignment
 *           ret_aseq - RETURN: alignment itself
 *           ainfo    - RETURN: other info associated with alignment
 *
 * Return:   (void)
 *           aseq, ainfo free'd by caller: FreeAlignment(aseq, &ainfo).
 *           note that ainfo itself is alloc'ed in caller, usually
 *           just by a "AINFO ainfo" definition.
 */
void
AllocAlignment (int nseq, int alen, char ***ret_aseq, AINFO * ainfo)
{
    char **aseq;
    int idx;

    InitAinfo (ainfo);

    aseq = (char **) MallocOrDie (sizeof (char *) * nseq);
    for (idx = 0; idx < nseq; idx++) {
        aseq[idx] = (char *) MallocOrDie (sizeof (char) * (alen + 1));
    }

    ainfo->alen = alen;
    ainfo->nseq = nseq;

    ainfo->wgt = (float *) MallocOrDie (sizeof (float) * nseq);
    FSet (ainfo->wgt, nseq, 1.0);

    ainfo->sqinfo = (SQINFO *) MallocOrDie (sizeof (SQINFO) * nseq);
    for (idx = 0; idx < nseq; idx++) {
        ainfo->sqinfo[idx].flags = 0;
    }

    *ret_aseq = aseq;
}


/* Function: InitAinfo()
 * Date:     SRE, Tue Jan 19 10:16:02 1999 [St. Louis]
 *
 * Purpose:  Initialize the fields in ainfo structure to
 *           default (null) values. Does nothing with
 *           fields that are dependent on nseq or alen.
 *
 * Args:     ainfo  - optional info structure for an alignment
 *
 * Returns:  (void). ainfo is modified.
 */
void
InitAinfo (AINFO * ainfo)
{
    ainfo->name = NULL;
    ainfo->desc = NULL;
    ainfo->cs = NULL;
    ainfo->rf = NULL;
    ainfo->acc = NULL;
    ainfo->au = NULL;
    ainfo->flags = 0;

    ainfo->tc1 = ainfo->tc2 = 0.0;
    ainfo->nc1 = ainfo->nc2 = 0.0;
    ainfo->ga1 = ainfo->ga2 = 0.0;
}


/* Function: FreeAlignment()
 *
 * Purpose:  Free the space allocated to alignment, names, and optional
 *           information.
 *
 * Args:     aseqs - sequence alignment
 *           ainfo - associated alignment data.
 */
void
FreeAlignment (char **aseqs, AINFO * ainfo)
{
    int i;

    for (i = 0; i < ainfo->nseq; i++) {
        if (ainfo->sqinfo[i].flags & SQINFO_SS) {
            free (ainfo->sqinfo[i].ss);
        }
        if (ainfo->sqinfo[i].flags & SQINFO_SA) {
            free (ainfo->sqinfo[i].sa);
        }
    }
    if (ainfo->cs != NULL) {
        free (ainfo->cs);
    }
    if (ainfo->rf != NULL) {
        free (ainfo->rf);
    }
    if (ainfo->name != NULL) {
        free (ainfo->name);
    }
    if (ainfo->desc != NULL) {
        free (ainfo->desc);
    }
    if (ainfo->acc != NULL) {
        free (ainfo->acc);
    }
    if (ainfo->au != NULL) {
        free (ainfo->au);
    }

    free (ainfo->sqinfo);
    free (ainfo->wgt);
    Free2DArray ((void **) aseqs, ainfo->nseq);
}



/* Function: SAMizeAlignment()
 * Date:     SRE, Tue Jun 30 09:49:40 1998 [St. Louis]
 *
 * Purpose:  Make a "best effort" attempt to convert an alignment
 *           to SAM gap format: - in delete col, . in insert col.
 *           Only works if alignment adheres to SAM's upper/lower
 *           case convention, which is true for instance of old
 *           HMMER alignments.
 *
 * Args:     aseq  - alignment to convert
 *           nseq  - number of seqs in alignment
 *           alen  - length of alignment
 *
 * Returns:  (void)
 */
void
SAMizeAlignment (char **aseq, int nseq, int alen)
{
    int col;			/* counter for aligned columns */
    int i;			/* counter for seqs */
    int sawlower, sawupper, sawgap;
    char gapchar;

    for (col = 0; col < alen; col++) {
        sawlower = sawupper = sawgap = 0;
        /* pass 1: do we see only upper or lower? */
        for (i = 0; i < nseq; i++) {
            if (isgap (aseq[i][col])) {
                sawgap = 1;
                continue;
            }
            if (isupper ((int) aseq[i][col])) {
                sawupper = 1;
                continue;
            }
            if (islower ((int) aseq[i][col])) {
                sawlower = 1;
            }
        }
        /* select gap character for column */
        gapchar = '-';		/* default */
        if (sawlower && !sawupper) {
            gapchar = '.';
        }

        /* pass 2: set gap char */
        for (i = 0; i < nseq; i++) {
            if (isgap (aseq[i][col])) {
                aseq[i][col] = gapchar;
            }
				}		
    }
}


/* Function: SAMizeAlignmentByGapFrac()
 * Date:     SRE, Tue Jun 30 10:58:38 1998 [St. Louis]
 *
 * Purpose:  Convert an alignment to SAM's gap and case
 *           conventions, using gap fraction in a column
 *           to choose match versus insert columns. In match columns,
 *           residues are upper case and gaps are '-'.
 *           In insert columns, residues are lower case and
 *           gaps are '.'
 *
 * Args:     aseq   - aligned sequences
 *           nseq   - number of sequences
 *           alen   - length of alignment
 *           maxgap - if more gaps than this fraction, column is insert.
 *
 * Returns:  (void) Characters in aseq may be altered.
 */
void
SAMizeAlignmentByGapFrac (char **aseq, int nseq, int alen, float maxgap)
{
    int apos;			/* counter over columns */
    int idx;			/* counter over sequences */
    int ngap;			/* number of gaps seen */

    for (apos = 0; apos < alen; apos++) {
        /* count gaps */
        ngap = 0;
        for (idx = 0; idx < nseq; idx++) {
            if (isgap (aseq[idx][apos])) {
                ngap++;
            }
				}
        /* convert to SAM conventions */
        if ((float) ngap / (float) nseq > maxgap) {	/* insert column */
            for (idx = 0; idx < nseq; idx++) {
                if (isgap (aseq[idx][apos])) {
                    aseq[idx][apos] = '.';
                }
                else {
                    aseq[idx][apos] = (char) tolower ((int) aseq[idx][apos]);
                }
						}		
        }
        else {			/* match column */
            for (idx = 0; idx < nseq; idx++) {
                if (isgap (aseq[idx][apos])) {
                    aseq[idx][apos] = '-';
                }
                else {
                    aseq[idx][apos] = (char) toupper ((int) aseq[idx][apos]);
                }
						}		
        }
    }
}




/* Function: MakeAlignedString()
 *
 * Purpose:  Given a raw string of some type (secondary structure, say),
 *           align it to a given aseq by putting gaps wherever the
 *           aseq has gaps.
 *
 * Args:     aseq:  template for alignment
 *           alen:  length of aseq
 *           ss:    raw string to align to aseq
 *           ret_s: RETURN: aligned ss
 *
 * Return:   1 on success, 0 on failure (and squid_errno is set.)
 *           ret_ss is malloc'ed here and must be free'd by caller.
 */
int
MakeAlignedString (char *aseq, int alen, char *ss, char **ret_s)
{
    char *new;
    int apos, rpos;

    new = (char *) MallocOrDie ((alen + 1) * sizeof (char));
    for (apos = rpos = 0; apos < alen; apos++) {
        if (!isgap (aseq[apos])) {
            new[apos] = ss[rpos];
            rpos++;
        }
        else {
            new[apos] = '.';
        }
		}		
    new[apos] = '\0';

    if (rpos != strlen (ss)) {
        squid_errno = SQERR_PARAMETER;
        free (new);
        return 0;
    }
    *ret_s = new;
    return 1;
}


/* Function: MakeDealignedString()
 *
 * Purpose:  Given an aligned string of some type (either sequence or
 *           secondary structure, for instance), dealign it relative
 *           to a given aseq. Return a ptr to the new string.
 *
 * Args:     aseq  : template alignment
 *           alen  : length of aseq
 *           ss:   : string to make dealigned copy of; same length as aseq
 *           ret_s : RETURN: dealigned copy of ss
 *
 * Return:   1 on success, 0 on failure (and squid_errno is set)
 *           ret_s is alloc'ed here and must be freed by caller
 */
int
MakeDealignedString (char *aseq, int alen, char *ss, char **ret_s)
{
    char *new;
    int apos, rpos;

    new = (char *) MallocOrDie ((alen + 1) * sizeof (char));
    for (apos = rpos = 0; apos < alen; apos++) {
        if (!isgap (aseq[apos])) {
            new[rpos] = ss[apos];
            rpos++;
        }
		}		
    new[rpos] = '\0';
    if (alen != strlen (ss)) {
        squid_errno = SQERR_PARAMETER;
        free (new);
        return 0;
    }
    *ret_s = new;
    return 1;
}


/* Function: DealignedLength()
 *
 * Purpose:  Count the number of non-gap symbols in seq.
 *           (i.e. find the length of the unaligned sequence)
 *
 * Args:     aseq - aligned sequence to count symbols in, \0 terminated
 *
 * Return:   raw length of seq.
 */
int
DealignedLength (char *aseq)
{
    int rlen;
    for (rlen = 0; *aseq; aseq++) {
        if (!isgap (*aseq)) {
            rlen++;
        }
		}		
    return rlen;
}


/* Function: WritePairwiseAlignment()
 *
 * Purpose:  Write a nice formatted pairwise alignment out,
 *           with a BLAST-style middle line showing identities
 *           as themselves (single letter) and conservative
 *           changes as '+'.
 *
 * Args:     ofp          - open fp to write to (stdout, perhaps)
 *           aseq1, aseq2 - alignments to write (not necessarily
 *                          flushed right with gaps)
 *           name1, name2 - names of sequences
 *           spos1, spos2 - starting position in each (raw) sequence
 *           pam          - PAM matrix; positive values define
 *                          conservative changes
 *           indent       - how many extra spaces to print on left
 *
 * Return:  1 on success, 0 on failure
 */
int
WritePairwiseAlignment (FILE * ofp,
                        char *aseq1, char *name1, int spos1,
                        char *aseq2, char *name2, int spos2,
                        int **pam, int indent)
{
    char sname1[11];		/* shortened name               */
    char sname2[11];
    int still_going;		/* True if writing another block */
    char buf1[61];		/* buffer for writing seq1; CPL+1 */
    char bufmid[61];		/* buffer for writing consensus  */
    char buf2[61];
    char *s1, *s2;		/* ptrs into each sequence          */
    int count1, count2;		/* number of symbols we're writing  */
    int rpos1, rpos2;		/* position in raw seqs             */
    int rawcount1, rawcount2;	/* number of nongap symbols written */
    int apos;

    strncpy (sname1, name1, 10);
    sname1[10] = '\0';
    strtok (sname1, WHITESPACE);

    strncpy (sname2, name2, 10);
    sname2[10] = '\0';
    strtok (sname2, WHITESPACE);

    s1 = aseq1;
    s2 = aseq2;
    rpos1 = spos1;
    rpos2 = spos2;

    still_going = TRUE;
    while (still_going) {
        still_going = FALSE;

        /* get next line's worth from both */
        strncpy (buf1, s1, 60);
        buf1[60] = '\0';
        strncpy (buf2, s2, 60);
        buf2[60] = '\0';
        count1 = strlen (buf1);
        count2 = strlen (buf2);

        /* is there still more to go? */
        if ((count1 == 60 && s1[60] != '\0') || (count2 == 60 && s2[60] != '\0')) {
            still_going = TRUE;
        }

        /* shift seq ptrs by a line */
        s1 += count1;
        s2 += count2;

        /* assemble the consensus line */
        for (apos = 0; apos < count1 && apos < count2; apos++) {
            if (!isgap (buf1[apos]) && !isgap (buf2[apos])) {
                if (buf1[apos] == buf2[apos]) {
                    bufmid[apos] = buf1[apos];
                }
                else if (pam[buf1[apos] - 'A'][buf2[apos] - 'A'] > 0) {
                    bufmid[apos] = '+';
                }
                else {
                    bufmid[apos] = ' ';
                }
            }
            else {
                bufmid[apos] = ' ';
            }
        }
        bufmid[apos] = '\0';

        rawcount1 = 0;
        for (apos = 0; apos < count1; apos++) {
            if (!isgap (buf1[apos])) {
                rawcount1++;
            }
				}
        rawcount2 = 0;
        for (apos = 0; apos < count2; apos++) {
            if (!isgap (buf2[apos])) {
                rawcount2++;
            }
				}
        (void) fprintf (ofp, "%*s%-10.10s %5d %s %5d\n", indent, "",
                        sname1, rpos1, buf1, rpos1 + rawcount1 - 1);
        (void) fprintf (ofp, "%*s                 %s\n", indent, "", bufmid);
        (void) fprintf (ofp, "%*s%-10.10s %5d %s %5d\n", indent, "",
                        sname2, rpos2, buf2, rpos2 + rawcount2 - 1);
        (void) fprintf (ofp, "\n");

        rpos1 += rawcount1;
        rpos2 += rawcount2;
    }

    return 1;
}


/* Function: MingapAlignment()
 *
 * Purpose:  Remove all-gap columns from a multiple sequence alignment
 *           and its associated data. The alignment is assumed to be
 *           flushed (all aseqs the same length).
 */
int
MingapAlignment (char **aseqs, AINFO * ainfo)
{
    int apos;			/* position in original alignment */
    int mpos;			/* position in new alignment      */
    int idx;

    /* We overwrite aseqs, using its allocated memory.
     */
    for (apos = 0, mpos = 0; aseqs[0][apos] != '\0'; apos++) {
        /* check for all-gap in column */
        for (idx = 0; idx < ainfo->nseq; idx++) {
            if (!isgap (aseqs[idx][apos])) {
                break;
            }
				}		
        if (idx == ainfo->nseq) {
            continue;
        }

        /* shift alignment and ainfo */
        if (mpos != apos) {
            for (idx = 0; idx < ainfo->nseq; idx++) {
                aseqs[idx][mpos] = aseqs[idx][apos];
            }

            if (ainfo->cs != NULL) {
                ainfo->cs[mpos] = ainfo->cs[apos];
            }
            if (ainfo->rf != NULL) {
                ainfo->rf[mpos] = ainfo->rf[apos];
            }
        }
        mpos++;
    }
    /* null terminate everything */
    for (idx = 0; idx < ainfo->nseq; idx++) {
        aseqs[idx][mpos] = '\0';
    }
    ainfo->alen = mpos;		/* set new length */
    if (ainfo->cs != NULL) {
        ainfo->cs[mpos] = '\0';
    }
    if (ainfo->rf != NULL) {
        ainfo->rf[mpos] = '\0';
    }
    return 1;
}



/* Function: RandomAlignment()
 *
 * Purpose:  Create a random alignment from raw sequences.
 *
 *           Ideally, we would like to sample an alignment from the
 *           space of possible alignments according to its probability,
 *           given a prior probability distribution for alignments.
 *           I don't see how to describe such a distribution, let alone
 *           sample it.
 *
 *           This is a rough approximation that tries to capture some
 *           desired properties. We assume the alignment is generated
 *           by a simple HMM composed of match and insert states.
 *           Given parameters (pop, pex) for the probability of opening
 *           and extending an insertion, we can find the expected number
 *           of match states, M, in the underlying model for each sequence.
 *           We use an average M taken over all the sequences (this is
 *           an approximation. The expectation of M given all the sequence
 *           lengths is a nasty-looking summation.)
 *
 *           M = len / ( 1 + pop ( 1 + 1/ (1-pex) ) )
 *
 *           Then, we assign positions in each raw sequence onto the M match
 *           states and M+1 insert states of this "HMM", by rolling random
 *           numbers and inserting the (rlen-M) inserted positions randomly
 *           into the insert slots, taking into account the relative probability
 *           of open vs. extend.
 *
 *           The resulting alignment has two desired properties: insertions
 *           tend to follow the HMM-like exponential distribution, and
 *           the "sparseness" of the alignment is controllable through
 *           pop and pex.
 *
 * Args:     rseqs     - raw sequences to "align", 0..nseq-1
 *           sqinfo    - array of 0..nseq-1 info structures for the sequences
 *           nseq      - number of sequences
 *           pop       - probability to open insertion (0<pop<1)
 *           pex       - probability to extend insertion (0<pex<1)
 *           ret_aseqs - RETURN: alignment (flushed)
 *           ainfo     - fill in: alignment info
 *
 * Return:   1 on success, 0 on failure. Sets squid_errno to indicate cause
 *           of failure.
 */
int
RandomAlignment (char **rseqs, SQINFO * sqinfo, int nseq, float pop,
                 float pex, char ***ret_aseqs, AINFO * ainfo)
{
    char **aseqs;			/* RETURN: alignment   */
    int alen;			/* length of alignment */
    int *rlen;			/* lengths of each raw sequence */
    int M;			/* length of "model"   */
    int **ins;			/* insertion counts, 0..nseq-1 by 0..M */
    int *master_ins;		/* max insertion counts, 0..M */
    int apos, rpos, idx;
    int statepos;
    int count;
    int minlen;

    /* calculate expected length of model, M
     */
    rlen = (int *) MallocOrDie (sizeof (int) * nseq);
    M = 0;
    minlen = 9999999;
    for (idx = 0; idx < nseq; idx++) {
        rlen[idx] = strlen (rseqs[idx]);
        M += rlen[idx];
        minlen = (rlen[idx] < minlen) ? rlen[idx] : minlen;
    }
    M = (int) ((float) M / (1.0 + pop * (1.0 + 1.0 / (1.0 - pex))));
    M /= nseq;
    if (M > minlen) {
        M = minlen;
    }

    /* make arrays that count insertions in M+1 possible insert states
     */
    ins = (int **) MallocOrDie (sizeof (int *) * nseq);
    master_ins = (int *) MallocOrDie (sizeof (int) * (M + 1));
    for (idx = 0; idx < nseq; idx++) {
        ins[idx] = (int *) MallocOrDie (sizeof (int) * (M + 1));
        for (rpos = 0; rpos <= M; rpos++) {
            ins[idx][rpos] = 0;
        }
    }
    /* normalize */
    pop = pop / (pop + pex);
    pex = 1.0 - pop;
    /* make insertions for individual sequences */
    for (idx = 0; idx < nseq; idx++) {
        apos = -1;
        for (rpos = 0; rpos < rlen[idx] - M; rpos++) {
            if (sre_random () < pop || apos == -1) {	/* open insertion */
                apos = CHOOSE (M + 1);    /* choose 0..M */
            }
            ins[idx][apos]++;
        }
    }
    /* calculate master_ins, max inserts */
    alen = M;
    for (apos = 0; apos <= M; apos++) {
        master_ins[apos] = 0;
        for (idx = 0; idx < nseq; idx++) {
            if (ins[idx][apos] > master_ins[apos]) {
                master_ins[apos] = ins[idx][apos];
            }
				}		
        alen += master_ins[apos];
    }


    /* Now, construct alignment
     */
    aseqs = (char **) MallocOrDie (sizeof (char *) * nseq);
    for (idx = 0; idx < nseq; idx++) {
        aseqs[idx] = (char *) MallocOrDie (sizeof (char) * (alen + 1));
    }
    for (idx = 0; idx < nseq; idx++) {
        apos = rpos = 0;

        for (statepos = 0; statepos <= M; statepos++) {
            for (count = 0; count < ins[idx][statepos]; count++) {
                aseqs[idx][apos++] = rseqs[idx][rpos++];
            }
            for (; count < master_ins[statepos]; count++) {
                aseqs[idx][apos++] = ' ';
            }

            if (statepos != M) {
                aseqs[idx][apos++] = rseqs[idx][rpos++];
            }
        }
        aseqs[idx][alen] = '\0';
    }
    ainfo->flags = 0;
    ainfo->alen = alen;
    ainfo->nseq = nseq;
    ainfo->sqinfo = (SQINFO *) MallocOrDie (sizeof (SQINFO) * nseq);
    for (idx = 0; idx < nseq; idx++) {
        SeqinfoCopy (&(ainfo->sqinfo[idx]), &(sqinfo[idx]));
    }

    free (rlen);
    free (master_ins);
    Free2DArray ((void **) ins, nseq);
    *ret_aseqs = aseqs;
    return 1;
}

/* Function: AlignmentHomogenousGapsym()
 * Date:     SRE, Sun Mar 19 19:37:12 2000 [wren, St. Louis]
 *
 * Purpose:  Sometimes we've got to convert alignments to
 *           a lowest common denominator, and we need
 *           a single specific gap character -- for example,
 *           PSI-BLAST blastpgp -B takes a very simplistic
 *           alignment input format which appears to only
 *           allow '-' as a gap symbol.
 *
 *           Anything matching the isgap() macro is
 *           converted.
 *
 * Args:     aseq   - aligned character strings, [0..nseq-1][0..alen-1]
 *           nseq   - number of aligned strings
 *           alen   - length of alignment
 *           gapsym - character to use for gaps.
 *
 * Returns:  void ("never fails")
 */
void
AlignmentHomogenousGapsym (char **aseq, int nseq, int alen, char gapsym)
{
    int i, apos;

    for (i = 0; i < nseq; i++) {
        for (apos = 0; apos < alen; apos++) {
            if (isgap (aseq[i][apos])) {
                aseq[i][apos] = gapsym;
            }
				}
		}		
}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* alphabet.c
 * Configuration of the global symbol alphabet information.
 * CVS $Id: alphabet.c,v 1.11 2003/05/23 15:24:08 eddy Exp $
 */

#include "config.h"
#include "squidconf.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef HMMER_THREADS
#include <pthread.h>
#endif /* HMMER_THREADS */

#include "structs.h"
#include "funcs.h"
#include "squid.h"

static void set_degenerate (char iupac, char *syms);


/* Function: DetermineAlphabet()
 *
 * Purpose:  From a set of sequences (raw or aligned), make a good
 *           guess whether they're Nucleic, Amino, or something
 *           else, and set alphabet accordingly.
 *
 *           If Alphabet_type is already set, that means our
 *           autodetection was overridden from the command line,
 *           and we just set the other globals accordingly.
 */
void
DetermineAlphabet (char **rseqs, int nseq)
{
    int idx;
    int other, nucleic, amino;
    int type;

    /* Autodetection of alphabet type.
     */
    type = hmmNOTSETYET;
    other = nucleic = amino = 0;
    for (idx = 0; idx < nseq; idx++) {
        switch (Seqtype (rseqs[idx])) {
        case kRNA:
            nucleic++;
            break;
        case kDNA:
            nucleic++;
            break;
        case kAmino:
            amino++;
            break;
        case kOtherSeq:
            other++;
            break;
        default:
            Die ("No such alphabet type");
        }
    }

    if (nucleic == nseq) {
        type = hmmNUCLEIC;
    }
    else if (amino == nseq) {
        type = hmmAMINO;
    }
    else if (nucleic > amino && nucleic > other) {
        Warn ("Looks like nucleic acid sequence, hope that's right");
        type = hmmNUCLEIC;
    }
    else if (amino > nucleic && amino > other) {
        Warn ("Looks like amino acid sequence, hope that's right");
        type = hmmAMINO;
    }
    else {
        Die ("Sorry, I can't tell if that's protein or DNA");
    }

    /* Now set up the alphabet.
     */
    SetAlphabet (type);
}


/* Function: SetAlphabet()
 *
 * Purpose:  Set the alphabet globals, given an alphabet type
 *           of either hmmAMINO or hmmNUCLEIC.
 */
void
SetAlphabet (int type)
{
    int x;
#ifdef HMMER_THREADS
    pthread_mutex_t alphabet_lock;	/* alphabet is global; must protect to be threadsafe */
    int rtn;			/* return code from pthreads */

    if ((rtn = pthread_mutex_init (&alphabet_lock, NULL)) != 0) {
        Die ("pthread_mutex_init FAILED; %s\n", strerror (rtn));
    }
    if ((rtn = pthread_mutex_lock (&alphabet_lock)) != 0) {
        Die ("pthread_mutex_lock FAILED: %s\n", strerror (rtn));
    }
#endif

    /* Because the alphabet information is global, we must
     * be careful to make this a thread-safe function. The mutex
     * (above) takes care of that. But, indeed, it's also
     * just good sense (and more efficient) to simply never
     * allow resetting the alphabet. If type is Alphabet_type,
     * silently return; else die with an alphabet mismatch
     * warning.
     */
    if (Alphabet_type != hmmNOTSETYET) {
        if (type != Alphabet_type)
            Die
            ("An alphabet type conflict occurred.\nYou probably mixed a DNA seq file with a protein model, or vice versa.");

#ifdef HMMER_THREADS
        if ((rtn = pthread_mutex_unlock (&alphabet_lock)) != 0) {
            Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
        }
#endif
        return;
    }

    switch (type) {		/* Alphabet is not a string - careful! */
    case hmmAMINO:
        Alphabet_type = type;
        /*  strncpy(Alphabet, "ACDEFGHIKLMNPQRSTVWYUBZX", 24); 10/10 kmd  */
        strcpy (Alphabet, "ACDEFGHIKLMNPQRSTVWYUBZX");
        Alphabet_size = 20;
        Alphabet_iupac = 24;
        for (x = 0; x < Alphabet_iupac; x++) {
            memset (Degenerate[x], 0, Alphabet_size);
        }
        for (x = 0; x < Alphabet_size; x++) {
            Degenerate[x][x] = 1;
            DegenCount[x] = 1;
        }
        set_degenerate ('U', "S");	/* selenocysteine is treated as serine */
        set_degenerate ('B', "ND");
        set_degenerate ('Z', "QE");
        set_degenerate ('X', "ACDEFGHIKLMNPQRSTVWY");
        break;
    case hmmNUCLEIC:
        Alphabet_type = type;
        /*  strncpy(Alphabet, "ACGTUNRYMKSWHBVDX", 17); 10/1003 kmd */
        strcpy (Alphabet, "ACGTUNRYMKSWHBVDX");
        Alphabet_size = 4;
        Alphabet_iupac = 17;
        for (x = 0; x < Alphabet_iupac; x++) {
            memset (Degenerate[x], 0, Alphabet_size);
        }
        for (x = 0; x < Alphabet_size; x++) {
            Degenerate[x][x] = 1;
            DegenCount[x] = 1;
        }
        set_degenerate ('U', "T");
        set_degenerate ('N', "ACGT");
        set_degenerate ('X', "ACGT");
        set_degenerate ('R', "AG");
        set_degenerate ('Y', "CT");
        set_degenerate ('M', "AC");
        set_degenerate ('K', "GT");
        set_degenerate ('S', "CG");
        set_degenerate ('W', "AT");
        set_degenerate ('H', "ACT");
        set_degenerate ('B', "CGT");
        set_degenerate ('V', "ACG");
        set_degenerate ('D', "AGT");
        break;
    default:
        Die ("No support for non-nucleic or protein alphabets");
    }

#ifdef HMMER_THREADS
    if ((rtn = pthread_mutex_unlock (&alphabet_lock)) != 0) {
        Die ("pthread_mutex_unlock failure: %s\n", strerror (rtn));
    }
#endif
}

/* Function: SymbolIndex()
 *
 * Purpose:  Convert a symbol to its index in Alphabet[].
 *           Bogus characters are converted to 'X'.
 *           More robust than the SYMIDX() macro but
 *           presumably slower.
 */
int
SymbolIndex (char sym)
{
    char *s;
    return ((s = strchr (Alphabet, (char) toupper ((int) sym))) == NULL) ?
           Alphabet_iupac - 1 : s - Alphabet;
}


/* Function: DigitizeSequence()
 *
 * Purpose:  Internal representation of a sequence in HMMER is
 *           as a char array. 1..L are the indices
 *           of seq symbols in Alphabet[]. 0,L+1 are sentinel
 *           bytes, set to be Alphabet_iupac -- i.e. one more
 *           than the maximum allowed index.
 *
 *           Assumes that 'X', the fully degenerate character,
 *           is the last character in the allowed alphabet.
 *
 * Args:     seq - sequence to be digitized (0..L-1)
 *           L   - length of sequence
 *
 * Return:   digitized sequence, dsq.
 *           dsq is allocated here and must be free'd by caller.
 */
char *
DigitizeSequence (char *seq, int L)
{
    char *dsq;
    int i;

    dsq = MallocOrDie (sizeof (char) * (L + 2));
    dsq[0] = dsq[L + 1] = (char) Alphabet_iupac;
    for (i = 1; i <= L; i++) {
        dsq[i] = SymbolIndex (seq[i - 1]);
    }
    return dsq;
}


/* Function: DedigitizeSequence()
 * Date:     SRE, Tue Dec 16 10:39:19 1997 [StL]
 *
 * Purpose:  Returns a 0..L-1 character string, converting the
 *           dsq back to the real alphabet.
 */
char *
DedigitizeSequence (char *dsq, int L)
{
    char *seq;
    int i;

    seq = MallocOrDie (sizeof (char) * (L + 1));
    for (i = 0; i < L; i++) {
        seq[i] = Alphabet[(int) dsq[i + 1]];
    }
    seq[L] = '\0';
    return seq;
}


/* Function: DigitizeAlignment()
 *
 * Purpose:  Given an alignment, return digitized unaligned
 *           sequence array. (Tracebacks are always relative
 *           to digitized unaligned seqs, even if they are
 *           faked from an existing alignment in modelmakers.c.)
 *
 * Args:     msa      - alignment to digitize
 *           ret_dsqs - RETURN: array of digitized unaligned sequences
 *
 * Return:   (void)
 *           dsqs is alloced here. Free2DArray(dseqs, nseq).
 */
void
DigitizeAlignment (MSA * msa, char ***ret_dsqs)
{
    char **dsq;
    int idx;			/* counter for sequences     */
    int dpos;			/* position in digitized seq */
    int apos;			/* position in aligned seq   */

    dsq = (char **) MallocOrDie (sizeof (char *) * msa->nseq);
    for (idx = 0; idx < msa->nseq; idx++) {
        dsq[idx] = (char *) MallocOrDie (sizeof (char) * (msa->alen + 2));

        dsq[idx][0] = (char) Alphabet_iupac;	/* sentinel byte at start */

        for (apos = 0, dpos = 1; apos < msa->alen; apos++) {
            if (!isgap (msa->aseq[idx][apos])) {	/* skip gaps */
                dsq[idx][dpos++] = SymbolIndex (msa->aseq[idx][apos]);
            }
        }
        dsq[idx][dpos] = (char) Alphabet_iupac;	/* sentinel byte at end */
    }
    *ret_dsqs = dsq;
}


/* Function: P7CountSymbol()
 *
 * Purpose:  Given a possibly degenerate symbol code, increment
 *           a symbol counter array (generally an emission
 *           probability vector in counts form) appropriately.
 *
 * Args:     counters:  vector to count into. [0..Alphabet_size-1]
 *           symidx:    symbol index to count: [0..Alphabet_iupac-1]
 *           wt:        weight to use for the count; often 1.0
 *
 * Return:   (void)
 */
void
P7CountSymbol (float *counters, char symidx, float wt)
{
    int x;

    if (symidx < Alphabet_size) {
        counters[(int) symidx] += wt;
    }
    else
        for (x = 0; x < Alphabet_size; x++) {
            if (Degenerate[(int) symidx][x]) {
                counters[x] += wt / (float) DegenCount[(int) symidx];
            }
        }
}


/* Function: DefaultGeneticCode()
 *
 * Purpose:  Configure aacode, mapping triplets to amino acids.
 *           Triplet index: AAA = 0, AAC = 1, ... UUU = 63.
 *           AA index: alphabetical: A=0,C=1... Y=19
 *           Stop codon: -1.
 *           Uses the stdcode1[] global translation table from SQUID.
 *
 * Args:     aacode  - preallocated 0.63 array for genetic code
 *
 * Return:   (void)
 */
void
DefaultGeneticCode (int *aacode)
{
    int x;

    for (x = 0; x < 64; x++) {
        if (*(stdcode1[x]) == '*') {
            aacode[x] = -1;
        }
        else {
            aacode[x] = SYMIDX (*(stdcode1[x]));
        }
    }
}


/* Function: DefaultCodonBias()
 *
 * Purpose:  Configure a codonbias table, mapping triplets to
 *           probability of using the triplet for the amino acid
 *           it represents: P(triplet | aa).
 *           The default is to assume codons are used equiprobably.
 *
 * Args:     codebias:  0..63 array of P(triplet|aa), preallocated.
 *
 * Return:   (void)
 */
void
DefaultCodonBias (float *codebias)
{
    codebias[0] = 1. / 2.;	/* AAA Lys 2 */
    codebias[1] = 1. / 2.;	/* AAC Asn 2 */
    codebias[2] = 1. / 2.;	/* AAG Lys 2 */
    codebias[3] = 1. / 2.;	/* AAU Asn 2 */
    codebias[4] = 1. / 4.;	/* ACA Thr 4 */
    codebias[5] = 1. / 4.;	/* ACC Thr 4 */
    codebias[6] = 1. / 4.;	/* ACG Thr 4 */
    codebias[7] = 1. / 4.;	/* ACU Thr 4 */
    codebias[8] = 1. / 6.;	/* AGA Ser 6 */
    codebias[9] = 1. / 6.;	/* AGC Arg 6 */
    codebias[10] = 1. / 6.;	/* AGG Ser 6 */
    codebias[11] = 1. / 6.;	/* AGU Arg 6 */
    codebias[12] = 1. / 3.;	/* AUA Ile 3 */
    codebias[13] = 1. / 3.;	/* AUC Ile 3 */
    codebias[14] = 1.;		/* AUG Met 1 */
    codebias[15] = 1. / 3.;	/* AUU Ile 3 */
    codebias[16] = 1. / 2.;	/* CAA Gln 2 */
    codebias[17] = 1. / 2.;	/* CAC His 2 */
    codebias[18] = 1. / 2.;	/* CAG Gln 2 */
    codebias[19] = 1. / 2.;	/* CAU His 2 */
    codebias[20] = 1. / 4.;	/* CCA Pro 4 */
    codebias[21] = 1. / 4.;	/* CCC Pro 4 */
    codebias[22] = 1. / 4.;	/* CCG Pro 4 */
    codebias[23] = 1. / 4.;	/* CCU Pro 4 */
    codebias[24] = 1. / 6.;	/* CGA Arg 6 */
    codebias[25] = 1. / 6.;	/* CGC Arg 6 */
    codebias[26] = 1. / 6.;	/* CGG Arg 6 */
    codebias[27] = 1. / 6.;	/* CGU Arg 6 */
    codebias[28] = 1. / 6.;	/* CUA Leu 6 */
    codebias[29] = 1. / 6.;	/* CUC Leu 6 */
    codebias[30] = 1. / 6.;	/* CUG Leu 6 */
    codebias[31] = 1. / 6.;	/* CUU Leu 6 */
    codebias[32] = 1. / 2.;	/* GAA Glu 2 */
    codebias[33] = 1. / 2.;	/* GAC Asp 2 */
    codebias[34] = 1. / 2.;	/* GAG Glu 2 */
    codebias[35] = 1. / 2.;	/* GAU Asp 2 */
    codebias[36] = 1. / 4.;	/* GCA Ala 4 */
    codebias[37] = 1. / 4.;	/* GCC Ala 4 */
    codebias[38] = 1. / 4.;	/* GCG Ala 4 */
    codebias[39] = 1. / 4.;	/* GCU Ala 4 */
    codebias[40] = 1. / 4.;	/* GGA Gly 4 */
    codebias[41] = 1. / 4.;	/* GGC Gly 4 */
    codebias[42] = 1. / 4.;	/* GGG Gly 4 */
    codebias[43] = 1. / 4.;	/* GGU Gly 4 */
    codebias[44] = 1. / 4.;	/* GUA Val 4 */
    codebias[45] = 1. / 4.;	/* GUC Val 4 */
    codebias[46] = 1. / 4.;	/* GUG Val 4 */
    codebias[47] = 1. / 4.;	/* GUU Val 4 */
    codebias[48] = 0.;		/* UAA och - */
    codebias[49] = 1. / 2.;	/* UAC Tyr 2 */
    codebias[50] = 0.;		/* UAG amb - */
    codebias[51] = 1. / 2.;	/* UAU Tyr 2 */
    codebias[52] = 1. / 6.;	/* UCA Ser 6 */
    codebias[53] = 1. / 6.;	/* UCC Ser 6 */
    codebias[54] = 1. / 6.;	/* UCG Ser 6 */
    codebias[55] = 1. / 6.;	/* UCU Ser 6 */
    codebias[56] = 0.;		/* UGA opa - */
    codebias[57] = 1. / 2.;	/* UGC Cys 2 */
    codebias[58] = 1.;		/* UGG Trp 1 */
    codebias[59] = 1. / 2.;	/* UGU Cys 2 */
    codebias[60] = 1. / 6.;	/* UUA Leu 6 */
    codebias[61] = 1. / 2.;	/* UUC Phe 2 */
    codebias[62] = 1. / 6.;	/* UUG Leu 6 */
    codebias[63] = 1. / 2.;	/* UUU Phe 2 */
}



/* Function: set_degenerate()
 *
 * Purpose:  convenience function for setting up
 *           Degenerate[][] global for the alphabet.
 */
static void
set_degenerate (char iupac, char *syms)
{
    DegenCount[strchr (Alphabet, iupac) - Alphabet] = strlen (syms);
    while (*syms) {
        Degenerate[strchr (Alphabet, iupac) - Alphabet]
        [strchr (Alphabet, *syms) - Alphabet] = 1;
        syms++;
    }
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* clustal.c
 * SRE, Sun Jun  6 17:50:45 1999 [bus from Madison, 1999 worm mtg]
 *
 * Import/export of ClustalV/W multiple sequence alignment
 * formatted files. Derivative of msf.c; MSF is a pretty
 * generic interleaved format.
 *
 * CVS $Id: clustal.c,v 1.2 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"
#include "msa.h"

#ifdef TESTDRIVE_CLUSTAL
/*****************************************************************
 * msf.c test driver:
 * cc -DTESTDRIVE_CLUSTAL -g -O2 -Wall -o test clustal.c msa.c gki.c sqerror.c sre_string.c file.c hsregex.c sre_math.c sre_ctype.c -lm
 *
 */
int
main (int argc, char **argv)
{
    MSAFILE *afp;
    MSA *msa;
    char *file;

    file = argv[1];

    if ((afp = MSAFileOpen (file, MSAFILE_CLUSTAL, NULL)) == NULL) {
        Die ("Couldn't open %s\n", file);
    }

    while ((msa = ReadClustal (afp)) != NULL) {
        WriteClustal (stdout, msa);
        MSAFree (msa);
    }

    MSAFileClose (afp);
    exit (0);
}

/******************************************************************/
#endif /* testdrive_clustal */


/* Function: ReadClustal()
 * Date:     SRE, Sun Jun  6 17:53:49 1999 [bus from Madison, 1999 worm mtg]
 *
 * Purpose:  Parse an alignment read from an open Clustal format
 *           alignment file. Clustal is a single-alignment format.
 *           Return the alignment, or NULL if we have no data.
 *
 * Args:     afp  - open alignment file
 *
 * Returns:  MSA * - an alignment object
 *                   caller responsible for an MSAFree()
 *           NULL if no more alignments
 *
 * Diagnostics:
 *           Will Die() here with a (potentially) useful message
 *           if a parsing error occurs.
 */
MSA *
ReadClustal (MSAFILE * afp)
{
    MSA *msa;
    char *s;
    int slen;
    int sqidx;
    char *name;
    char *seq;
    char *s2;

    if (feof (afp->f)) {
        return NULL;
    }

    /* Skip until we see the CLUSTAL header
     */
    while ((s = MSAFileGetLine (afp)) != NULL) {
        if (strncmp (s, "CLUSTAL", 7) == 0 &&
                strstr (s, "multiple sequence alignment") != NULL) {
            break;
        }
    }
    if (s == NULL) {
        return NULL;
    }

    msa = MSAAlloc (10, 0);

    /* Now we're in the sequence section.
     * As discussed above, if we haven't seen a sequence name, then we
     * don't include the sequence in the alignment.
     * Watch out for conservation markup lines that contain *.: chars
     */
    while ((s = MSAFileGetLine (afp)) != NULL) {
        if ((name = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
            continue;
        }
        if ((seq = sre_strtok (&s, WHITESPACE, &slen)) == NULL) {
            continue;
        }
        s2 = sre_strtok (&s, "\n", NULL);

        /* The test for a conservation markup line
         */
        if (strpbrk (name, ".*:") != NULL && strpbrk (seq, ".*:") != NULL) {
            continue;
        }
        if (s2 != NULL)
            Die ("Parse failed at line %d, file %s: possibly using spaces as gaps",
                 afp->linenumber, afp->fname);

        /* It's not blank, and it's not a coord line: must be sequence
         */
        sqidx = MSAGetSeqidx (msa, name, msa->lastidx + 1);
        msa->lastidx = sqidx;
        msa->sqlen[sqidx] =
            sre_strcat (&(msa->aseq[sqidx]), msa->sqlen[sqidx], seq, slen);
    }

    MSAVerifyParse (msa);		/* verifies, and also sets alen and wgt. */
    return msa;
}


/* Function: WriteClustal()
 * Date:     SRE, Sun Jun  6 18:12:47 1999 [bus from Madison, worm mtg 1999]
 *
 * Purpose:  Write an alignment in Clustal format to an open file.
 *
 * Args:     fp    - file that's open for writing.
 *           msa   - alignment to write.
 *
 * Returns:  (void)
 */
void
WriteClustal (FILE * fp, MSA * msa)
{
    int idx;			/* counter for sequences         */
    int len;			/* tmp variable for name lengths */
    int namelen;			/* maximum name length used      */
    int pos;			/* position counter              */
    char buf[64];			/* buffer for writing seq        */
    int cpl = 50;			/* char per line (< 64)          */

    /* calculate max namelen used */
    namelen = 0;
    for (idx = 0; idx < msa->nseq; idx++) {
        if ((len = strlen (msa->sqname[idx])) > namelen) {
            namelen = len;
        }
		}
    fprintf (fp, "CLUSTAL W(1.5) multiple sequence alignment\n");

    /*****************************************************
     * Write the sequences
     *****************************************************/

    for (pos = 0; pos < msa->alen; pos += cpl) {
        fprintf (fp, "\n");		/* Blank line between sequence blocks */
        for (idx = 0; idx < msa->nseq; idx++) {
            strncpy (buf, msa->aseq[idx] + pos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "%*s %s\n", namelen, msa->sqname[idx], buf);
        }
    }

    return;
}



/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* core_algorithms.c
 * SRE, Mon Nov 11 15:58:52 1996
 * CVS $Id: core_algorithms.c,v 1.31 2003/05/26 19:33:09 eddy Exp $
 *
 * Simple and robust "research" implementations of Forward, Backward,
 * and Viterbi for Plan7. For optimized replacements for some of these functions,
 * see fast_algorithms.c.
 */
#include "config.h"
#include "squidconf.h"

#include "structs.h"
#include "funcs.h"
#include "squid.h"

#include <string.h>
#include <assert.h>

static float get_wee_midpt (struct plan7_s *hmm, char *dsq, int L,
                            int k1, char t1, int s1,
                            int k3, char t3, int s3,
                            int *ret_k2, char *ret_t2, int *ret_s2);


#ifndef ALTIVEC
/* Function: CreatePlan7Matrix()
 *
 * Purpose:  Create a dynamic programming matrix for standard Forward,
 *           Backward, or Viterbi, with scores kept as scaled log-odds
 *           integers. Keeps 2D arrays compact in RAM in an attempt
 *           to maximize cache hits.
 *
 *           The mx structure can be dynamically grown, if a new
 *           HMM or seq exceeds the currently allocated size. Dynamic
 *           growing is more efficient than an alloc/free of a whole
 *           matrix for every new target. The ResizePlan7Matrix()
 *           call does this reallocation, if needed. Here, in the
 *           creation step, we set up some pads - to inform the resizing
 *           call how much to overallocate when it realloc's. If a pad
 *           is zero, we will not resize in that dimension.
 *
 * Args:     N     - N+1 rows are allocated, for sequence.
 *           M     - size of model in nodes
 *           padN  - over-realloc in seq/row dimension, or 0
 *           padM  - over-realloc in HMM/column dimension, or 0
 *
 * Return:   mx
 *           mx is allocated here. Caller frees with FreePlan7Matrix(mx).
 */
struct dpmatrix_s *
CreatePlan7Matrix (int N, int M, int padN, int padM)
{
    struct dpmatrix_s *mx;
    int i;

    mx = (struct dpmatrix_s *) MallocOrDie (sizeof (struct dpmatrix_s));
    mx->xmx = (int **) MallocOrDie (sizeof (int *) * (N + 1));
    mx->mmx = (int **) MallocOrDie (sizeof (int *) * (N + 1));
    mx->imx = (int **) MallocOrDie (sizeof (int *) * (N + 1));
    mx->dmx = (int **) MallocOrDie (sizeof (int *) * (N + 1));
    mx->xmx_mem = (void *) MallocOrDie (sizeof (int) * ((N + 1) * 5));
    mx->mmx_mem = (void *) MallocOrDie (sizeof (int) * ((N + 1) * (M + 2)));
    mx->imx_mem = (void *) MallocOrDie (sizeof (int) * ((N + 1) * (M + 2)));
    mx->dmx_mem = (void *) MallocOrDie (sizeof (int) * ((N + 1) * (M + 2)));

    /* The indirect assignment below looks wasteful; it's actually
     * used for aligning data on 16-byte boundaries as a cache
     * optimization in the fast altivec implementation
     */
    mx->xmx[0] = (int *) mx->xmx_mem;
    mx->mmx[0] = (int *) mx->mmx_mem;
    mx->imx[0] = (int *) mx->imx_mem;
    mx->dmx[0] = (int *) mx->dmx_mem;
    for (i = 1; i <= N; i++) {
        mx->xmx[i] = mx->xmx[0] + (i * 5);
        mx->mmx[i] = mx->mmx[0] + (i * (M + 2));
        mx->imx[i] = mx->imx[0] + (i * (M + 2));
        mx->dmx[i] = mx->dmx[0] + (i * (M + 2));
    }

    if (padM > 0 && padN > 0) {
        Die ("there's trouble with RAMLIMIT if you grow in both M and N.");
    }
    mx->maxN = N;
    mx->maxM = M;
    mx->padN = padN;
    mx->padM = padM;

    return mx;
}
#endif /*ALTIVEC*/
#ifndef ALTIVEC
/* Function: ResizePlan7Matrix()
 *
 * Purpose:  Reallocate a dynamic programming matrix, if necessary,
 *           for a problem of NxM: sequence length N, model size M.
 *           (N=1 for small memory score-only variants; we allocate
 *           N+1 rows in the DP matrix.)
 *
 *           Returns individual ptrs to the four matrix components
 *           as a convenience.
 *
 * Args:     mx    - an already allocated model to grow.
 *           N     - seq length to allocate for; N+1 rows
 *           M     - size of model
 *           xmx, mmx, imx, dmx
 *                 - RETURN: ptrs to four mx components as a convenience
 *
 * Return:   (void)
 *           mx is (re)allocated here.
 */
void
ResizePlan7Matrix (struct dpmatrix_s *mx, int N, int M,
                   int ***xmx, int ***mmx, int ***imx, int ***dmx)
{
    int i;

    if (N <= mx->maxN && M <= mx->maxM) {
        goto DONE_1;
    }

    if (N > mx->maxN) {
        N += mx->padN;
        mx->maxN = N;
        mx->xmx = (int **) ReallocOrDie (mx->xmx, sizeof (int *) * (N + 1));
        mx->mmx = (int **) ReallocOrDie (mx->mmx, sizeof (int *) * (N + 1));
        mx->imx = (int **) ReallocOrDie (mx->imx, sizeof (int *) * (N + 1));
        mx->dmx = (int **) ReallocOrDie (mx->dmx, sizeof (int *) * (N + 1));
    }

    if (M > mx->maxM) {
        M += mx->padM;
        mx->maxM = M;
    }

    mx->xmx_mem =
        (void *) ReallocOrDie (mx->xmx_mem, sizeof (int) * ((N + 1) * 5));
    mx->mmx_mem =
        (void *) ReallocOrDie (mx->mmx_mem, sizeof (int) * ((N + 1) * (M + 2)));
    mx->imx_mem =
        (void *) ReallocOrDie (mx->imx_mem, sizeof (int) * ((N + 1) * (M + 2)));
    mx->dmx_mem =
        (void *) ReallocOrDie (mx->dmx_mem, sizeof (int) * ((N + 1) * (M + 2)));

    mx->xmx[0] = (int *) mx->xmx_mem;
    mx->mmx[0] = (int *) mx->mmx_mem;
    mx->imx[0] = (int *) mx->imx_mem;
    mx->dmx[0] = (int *) mx->dmx_mem;

    for (i = 1; i <= N; i++) {
        mx->xmx[i] = mx->xmx[0] + (i * 5);
        mx->mmx[i] = mx->mmx[0] + (i * (M + 2));
        mx->imx[i] = mx->imx[0] + (i * (M + 2));
        mx->dmx[i] = mx->dmx[0] + (i * (M + 2));
    }

DONE_1:
    if (xmx != NULL) {
        *xmx = mx->xmx;
    }
    if (mmx != NULL) {
        *mmx = mx->mmx;
    }
    if (imx != NULL) {
        *imx = mx->imx;
    }
    if (dmx != NULL) {
        *dmx = mx->dmx;
    }
}
#endif /*ALTIVEC*/
/* Function: AllocPlan7Matrix()
 * Date:     SRE, Tue Nov 19 07:14:47 2002 [St. Louis]
 *
 * Purpose:  Used to be the main allocator for dp matrices; we used to
 *           allocate, calculate, free. But this spent a lot of time
 *           in malloc(). Replaced with Create..() and Resize..() to
 *           allow matrix reuse in P7Viterbi(), the main alignment
 *           engine. But matrices are alloc'ed by other alignment engines
 *           too, ones that are less frequently called and less
 *           important to optimization of cpu performance. Instead of
 *           tracking changes through them, for now, provide
 *           an Alloc...() call with the same API that's just a wrapper.
 *
 * Args:     rows  - generally L+1, or 2; # of DP rows in seq dimension to alloc
 *           M     - size of model, in nodes
 *           xmx, mmx, imx, dmx
 *                 - RETURN: ptrs to four mx components as a convenience
 *
 * Returns:  mx
 *           Caller free's w/ FreePlan7Matrix()
 */
struct dpmatrix_s *
AllocPlan7Matrix (int rows, int M, int ***xmx, int ***mmx, int ***imx,
                  int ***dmx)
{
    struct dpmatrix_s *mx;
    mx = CreatePlan7Matrix (rows - 1, M, 0, 0);
    if (xmx != NULL) {
        *xmx = mx->xmx;
    }
    if (mmx != NULL) {
        *mmx = mx->mmx;
    }
    if (imx != NULL) {
        *imx = mx->imx;
    }
    if (dmx != NULL) {
        *dmx = mx->dmx;
    }
    return mx;
}


/* Function: FreePlan7Matrix()
 *
 * Purpose:  Free a dynamic programming matrix allocated by CreatePlan7Matrix().
 *
 * Return:   (void)
 */
void
FreePlan7Matrix (struct dpmatrix_s *mx)
{
    free (mx->xmx_mem);
    free (mx->mmx_mem);
    free (mx->imx_mem);
    free (mx->dmx_mem);
    free (mx->xmx);
    free (mx->mmx);
    free (mx->imx);
    free (mx->dmx);
    free (mx);
}

/* Function: AllocShadowMatrix()
 *
 * Purpose:  Allocate a dynamic programming traceback pointer matrix for
 *           a Viterbi algorithm.
 *
 * Args:     rows  - number of rows to allocate; typically L+1
 *           M     - size of model
 *           xtb, mtb, itb, dtb
 *                 - RETURN: ptrs to four mx components as a convenience
 *
 * Return:   mx
 *           mx is allocated here. Caller frees with FreeDPMatrix(mx).
 */

struct dpshadow_s *
AllocShadowMatrix (int rows, int M, char ***xtb, char ***mtb, char ***itb,
                   char ***dtb)
{
    struct dpshadow_s *tb;
    int i;

    tb = (struct dpshadow_s *) MallocOrDie (sizeof (struct dpshadow_s));
    tb->xtb = (char **) MallocOrDie (sizeof (char *) * rows);
    tb->mtb = (char **) MallocOrDie (sizeof (char *) * rows);
    tb->itb = (char **) MallocOrDie (sizeof (char *) * rows);
    tb->dtb = (char **) MallocOrDie (sizeof (char *) * rows);
    tb->esrc = (int *) MallocOrDie (sizeof (int) * rows);
    tb->xtb[0] = (char *) MallocOrDie (sizeof (char) * (rows * 5));
    tb->mtb[0] = (char *) MallocOrDie (sizeof (char) * (rows * (M + 2)));
    tb->itb[0] = (char *) MallocOrDie (sizeof (char) * (rows * (M + 2)));
    tb->dtb[0] = (char *) MallocOrDie (sizeof (char) * (rows * (M + 2)));
    for (i = 1; i < rows; i++) {
        tb->xtb[i] = tb->xtb[0] + (i * 5);
        tb->mtb[i] = tb->mtb[0] + (i * (M + 2));
        tb->itb[i] = tb->itb[0] + (i * (M + 2));
        tb->dtb[i] = tb->dtb[0] + (i * (M + 2));
    }

    if (xtb != NULL) {
        *xtb = tb->xtb;
    }
    if (mtb != NULL) {
        *mtb = tb->mtb;
    }
    if (itb != NULL) {
        *itb = tb->itb;
    }
    if (dtb != NULL) {
        *dtb = tb->dtb;
    }
    return tb;
}

/* Function: FreeShadowMatrix()
 *
 * Purpose:  Free a dynamic programming matrix allocated by AllocShadowMatrix().
 *
 * Return:   (void)
 */
void
FreeShadowMatrix (struct dpshadow_s *tb)
{
    free (tb->xtb[0]);
    free (tb->mtb[0]);
    free (tb->itb[0]);
    free (tb->dtb[0]);
    free (tb->esrc);
    free (tb->xtb);
    free (tb->mtb);
    free (tb->itb);
    free (tb->dtb);
    free (tb);
}

/* Function: P7ViterbiSize()
 * Date:     SRE, Fri Mar  6 15:13:20 1998 [St. Louis]
 *
 * Purpose:  Returns the ballpark predicted memory requirement for a
 *           P7Viterbi() alignment, in MB.
 *
 *           Currently L must fit in an int (< 2 GB), but we have
 *           to deal with LM > 2 GB - e.g. watch out for overflow, do
 *           the whole calculation in floating point. Bug here detected
 *           in 2.1.1 by David Harper, Sanger Centre.
 *
 * Args:     L  - length of sequence
 *           M  - length of HMM
 *
 * Returns:  # of MB
 */
int
P7ViterbiSize (int L, int M)
{
    float Mbytes;

    /* We're excessively precise here, but it doesn't cost
     * us anything to be pedantic. The four terms are:
     *   1. the matrix structure itself;
     *   2. the O(NM) main matrix (this dominates!)
     *   3. ptrs into the rows of the matrix
     *   4. storage for 5 special states. (xmx)
     */
    Mbytes = (float) sizeof (struct dpmatrix_s);
    Mbytes += 3. * (float) (L + 1) * (float) (M + 2) * (float) sizeof (int);
    Mbytes += 4. * (float) (L + 1) * (float) sizeof (int *);
    Mbytes += 5. * (float) (L + 1) * (float) sizeof (int);
    Mbytes /= 1048576.;
    return (int) Mbytes;
}

/* Function: P7SmallViterbiSize()
 * Date:     SRE, Fri Mar  6 15:20:04 1998 [St. Louis]
 *
 * Purpose:  Returns the ballpark predicted memory requirement for
 *           a P7SmallViterbi() alignment, in MB.
 *
 *           P7SmallViterbi() is a wrapper, calling both P7ParsingViterbi()
 *           and P7WeeViterbi(). P7ParsingViterbi() typically dominates
 *           the memory requirement, so the value returned
 *           is the P7ParsingViterbi() number.
 *
 *           We don't (yet) worry about overflow issues like we did with
 *           P7ViterbiSize(). We'll have many other 32-bit int issues in the
 *           code if we overflow here.
 *
 * Args:     L - length of sequence
 *           M - length of HMM
 *
 * Returns:  # of MB
 */
int
P7SmallViterbiSize (int L, int M)
{
    return ((2 * sizeof (struct dpmatrix_s) + 12 * (M + 2) * sizeof (int) +	/* 2 matrices w/ 2 rows */
             16 * sizeof (int *) +	/* ptrs into rows of matrix */
             20 * sizeof (int) +	/* 5 special states     */
             2 * (L + 1) * sizeof (int))	/* traceback indices    */
            /1000000);
}


/* Function: P7WeeViterbiSize()
 * Date:     SRE, Fri Mar  6 15:40:42 1998 [St. Louis]
 *
 * Purpose:  Returns the ballpark predicted memory requirement for
 *           a P7WeeViterbi() alignment, in MB.
 *
 * Args:     L - length of sequence
 *           M - length of HMM
 *
 * Returns:  # of MB
 */
int
P7WeeViterbiSize (int L, int M)
{
    return ((2 * sizeof (struct dpmatrix_s) + 12 * (M + 2) * sizeof (int) +	/* 2 matrices w/ 2 rows */
             16 * sizeof (int *) +	/* ptrs into rows of matrix */
             20 * sizeof (int) +	/* 5 special states      */
             2 * (L + 2) * sizeof (int) +	/* stacks for starts/ends (overkill) */
             (L + 2) * sizeof (int) +	/* k assignments to seq positions */
             (L + 2) * sizeof (char))	/* state assignments to seq pos */
            /1000000);
}


/* Function: P7Forward()
 *
 * Purpose:  The Forward dynamic programming algorithm.
 *           The scaling issue is dealt with by working in log space
 *           and calling ILogsum(); this is a slow but robust approach.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           ret_mx - RETURN: dp matrix; pass NULL if it's not wanted
 *
 * Return:   log P(S|M)/P(S|R), as a bit score.
 */
float
P7Forward (char *dsq, int L, struct plan7_s *hmm, struct dpmatrix_s **ret_mx)
{
    struct dpmatrix_s *mx;
    int **xmx;
    int **mmx;
    int **imx;
    int **dmx;
    int i, k;
    int sc;

    /* Allocate a DP matrix with 0..L rows, 0..M-1 columns.
     */
    mx = AllocPlan7Matrix (L + 1, hmm->M, &xmx, &mmx, &imx, &dmx);

    /* Initialization of the zero row.
     * Note that xmx[i][stN] = 0 by definition for all i,
     *    and xmx[i][stT] = xmx[i][stC], so neither stN nor stT need
     *    to be calculated in DP matrices.
     */
    xmx[0][XMN] = 0;		/* S->N, p=1            */
    xmx[0][XMB] = hmm->xsc[XTN][MOVE];	/* S->N->B, no N-tail   */
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need seq to get here */
    for (k = 0; k <= hmm->M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;    /* need seq to get here */
    }

    /* Recursion. Done as a pull.
     * Note some slightly wasteful boundary conditions:
     *    tsc[0] = -INFTY for all eight transitions (no node 0)
     *    D_M and I_M are wastefully calculated (they don't exist)
     */
    for (i = 1; i <= L; i++) {
        mmx[i][0] = imx[i][0] = dmx[i][0] = -INFTY;
        for (k = 1; k < hmm->M; k++) {
            mmx[i][k] = ILogsum (ILogsum (mmx[i - 1][k - 1] + hmm->tsc[TMM][k - 1],
                                          imx[i - 1][k - 1] + hmm->tsc[TIM][k - 1]),
                                 ILogsum (xmx[i - 1][XMB] + hmm->bsc[k],
                                          dmx[i - 1][k - 1] + hmm->tsc[TDM][k -
                                                  1]));
            mmx[i][k] += hmm->msc[(int) dsq[i]][k];

            dmx[i][k] = ILogsum (mmx[i][k - 1] + hmm->tsc[TMD][k - 1],
                                 dmx[i][k - 1] + hmm->tsc[TDD][k - 1]);
            imx[i][k] = ILogsum (mmx[i - 1][k] + hmm->tsc[TMI][k],
                                 imx[i - 1][k] + hmm->tsc[TII][k]);
            imx[i][k] += hmm->isc[(int) dsq[i]][k];
        }
        mmx[i][hmm->M] =
            ILogsum (ILogsum
                     (mmx[i - 1][hmm->M - 1] + hmm->tsc[TMM][hmm->M - 1],
                      imx[i - 1][hmm->M - 1] + hmm->tsc[TIM][hmm->M - 1]),
                     ILogsum (xmx[i - 1][XMB] + hmm->bsc[hmm->M - 1],
                              dmx[i - 1][hmm->M - 1] + hmm->tsc[TDM][hmm->M - 1]));
        mmx[i][hmm->M] += hmm->msc[(int) dsq[i]][hmm->M];

        /* Now the special states.
         * remember, C and J emissions are zero score by definition
         */
        xmx[i][XMN] = xmx[i - 1][XMN] + hmm->xsc[XTN][LOOP];

        xmx[i][XME] = -INFTY;
        for (k = 1; k <= hmm->M; k++) {
            xmx[i][XME] = ILogsum (xmx[i][XME], mmx[i][k] + hmm->esc[k]);
        }

        xmx[i][XMJ] = ILogsum (xmx[i - 1][XMJ] + hmm->xsc[XTJ][LOOP],
                               xmx[i][XME] + hmm->xsc[XTE][LOOP]);

        xmx[i][XMB] = ILogsum (xmx[i][XMN] + hmm->xsc[XTN][MOVE],
                               xmx[i][XMJ] + hmm->xsc[XTJ][MOVE]);

        xmx[i][XMC] = ILogsum (xmx[i - 1][XMC] + hmm->xsc[XTC][LOOP],
                               xmx[i][XME] + hmm->xsc[XTE][MOVE]);
    }

    sc = xmx[L][XMC] + hmm->xsc[XTC][MOVE];

    if (ret_mx != NULL) {
        *ret_mx = mx;
    }
    else {
        FreePlan7Matrix (mx);
    }

    return Scorify (sc);		/* the total Forward score. */
}


#ifdef SLOW
/* Function: P7Viterbi()
 *
 * Purpose:  The Viterbi dynamic programming algorithm.
 *           Identical to Forward() except that max's
 *           replace sum's.
 *
 *           This is the slower, more understandable version
 *           of P7Viterbi(). The default version in fast_algorithms.c
 *           is portably optimized and more difficult to understand;
 *           the ALTIVEC version in fast_algorithms.c is vectorized
 *           with Altivec-specific code, and is pretty opaque.
 *
 *           This function is not enabled by default; it is only
 *           activated by -DSLOW at compile time.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           mx     - reused DP matrix
 *           ret_tr - RETURN: traceback; pass NULL if it's not wanted
 *
 * Return:   log P(S|M)/P(S|R), as a bit score
 */
float
P7Viterbi (char *dsq, int L, struct plan7_s *hmm, struct dpmatrix_s *mx,
           struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;
    int **xmx;
    int **mmx;
    int **imx;
    int **dmx;
    int i, k;
    int sc;

    /* Allocate a DP matrix with 0..L rows, 0..M-1 columns.
     */
    ResizePlan7Matrix (mx, L, hmm->M, &xmx, &mmx, &imx, &dmx);

    /* Initialization of the zero row.
     */
    xmx[0][XMN] = 0;		/* S->N, p=1            */
    xmx[0][XMB] = hmm->xsc[XTN][MOVE];	/* S->N->B, no N-tail   */
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need seq to get here */
    for (k = 0; k <= hmm->M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;    /* need seq to get here */
    }

    /* Recursion. Done as a pull.
     * Note some slightly wasteful boundary conditions:
     *    tsc[0] = -INFTY for all eight transitions (no node 0)
     *    D_M and I_M are wastefully calculated (they don't exist)
     */
    for (i = 1; i <= L; i++) {
        mmx[i][0] = imx[i][0] = dmx[i][0] = -INFTY;

        for (k = 1; k <= hmm->M; k++) {
            /* match state */
            mmx[i][k] = -INFTY;
            if ((sc = mmx[i - 1][k - 1] + hmm->tsc[TMM][k - 1]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc = imx[i - 1][k - 1] + hmm->tsc[TIM][k - 1]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc = xmx[i - 1][XMB] + hmm->bsc[k]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc = dmx[i - 1][k - 1] + hmm->tsc[TDM][k - 1]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if (hmm->msc[(int) dsq[i]][k] != -INFTY) {
                mmx[i][k] += hmm->msc[(int) dsq[i]][k];
            }
            else {
                mmx[i][k] = -INFTY;
            }

            /* delete state */
            dmx[i][k] = -INFTY;
            if ((sc = mmx[i][k - 1] + hmm->tsc[TMD][k - 1]) > dmx[i][k]) {
                dmx[i][k] = sc;
            }
            if ((sc = dmx[i][k - 1] + hmm->tsc[TDD][k - 1]) > dmx[i][k]) {
                dmx[i][k] = sc;
            }

            /* insert state */
            if (k < hmm->M) {
                imx[i][k] = -INFTY;
                if ((sc = mmx[i - 1][k] + hmm->tsc[TMI][k]) > imx[i][k]) {
                    imx[i][k] = sc;
                }
                if ((sc = imx[i - 1][k] + hmm->tsc[TII][k]) > imx[i][k]) {
                    imx[i][k] = sc;
                }
                if (hmm->isc[(int) dsq[i]][k] != -INFTY) {
                    imx[i][k] += hmm->isc[(int) dsq[i]][k];
                }
                else {
                    imx[i][k] = -INFTY;
                }
            }
        }

        /* Now the special states. Order is important here.
         * remember, C and J emissions are zero score by definition,
         */
        /* N state */
        xmx[i][XMN] = -INFTY;
        if ((sc = xmx[i - 1][XMN] + hmm->xsc[XTN][LOOP]) > -INFTY) {
            xmx[i][XMN] = sc;
        }

        /* E state */
        xmx[i][XME] = -INFTY;
        for (k = 1; k <= hmm->M; k++) {
            if ((sc = mmx[i][k] + hmm->esc[k]) > xmx[i][XME]) {
                xmx[i][XME] = sc;
            }
				}		
        /* J state */
        xmx[i][XMJ] = -INFTY;
        if ((sc = xmx[i - 1][XMJ] + hmm->xsc[XTJ][LOOP]) > -INFTY) {
            xmx[i][XMJ] = sc;
        }
        if ((sc = xmx[i][XME] + hmm->xsc[XTE][LOOP]) > xmx[i][XMJ]) {
            xmx[i][XMJ] = sc;
        }

        /* B state */
        xmx[i][XMB] = -INFTY;
        if ((sc = xmx[i][XMN] + hmm->xsc[XTN][MOVE]) > -INFTY) {
            xmx[i][XMB] = sc;
        }
        if ((sc = xmx[i][XMJ] + hmm->xsc[XTJ][MOVE]) > xmx[i][XMB]) {
            xmx[i][XMB] = sc;
        }

        /* C state */
        xmx[i][XMC] = -INFTY;
        if ((sc = xmx[i - 1][XMC] + hmm->xsc[XTC][LOOP]) > -INFTY) {
            xmx[i][XMC] = sc;
        }
        if ((sc = xmx[i][XME] + hmm->xsc[XTE][MOVE]) > xmx[i][XMC]) {
            xmx[i][XMC] = sc;
        }
    }
    /* T state (not stored) */
    sc = xmx[L][XMC] + hmm->xsc[XTC][MOVE];

    if (ret_tr != NULL) {
        P7ViterbiTrace (hmm, dsq, L, mx, &tr);
        *ret_tr = tr;
    }

    return Scorify (sc);		/* the total Viterbi score. */
}
#endif /*SLOW*/
/* Function: P7ViterbiTrace()
 * Date:     SRE, Sat Aug 23 10:30:11 1997 (St. Louis Lambert Field)
 *
 * Purpose:  Traceback of a Viterbi matrix: i.e. retrieval
 *           of optimum alignment.
 *
 * Args:     hmm    - hmm, log odds form, used to make mx
 *           dsq    - sequence aligned to (digital form) 1..N
 *           N      - length of seq
 *           mx     - the matrix to trace back in, N x hmm->M
 *           ret_tr - RETURN: traceback.
 *
 * Return:   (void)
 *           ret_tr is allocated here. Free using P7FreeTrace().
 */
void
P7ViterbiTrace (struct plan7_s *hmm, char *dsq, int N,
                struct dpmatrix_s *mx, struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;
    int curralloc;		/* current allocated length of trace */
    int tpos;			/* position in trace */
    int i;			/* position in seq (1..N) */
    int k;			/* position in model (1..M) */
    int **xmx, **mmx, **imx, **dmx;
    int sc;			/* temp var for pre-emission score */

    /* Overallocate for the trace.
     * S-N-B- ... - E-C-T  : 6 states + N is minimum trace;
     * add N more as buffer.
     */
    curralloc = N * 2 + 6;
    P7AllocTrace (curralloc, &tr);

    xmx = mx->xmx;
    mmx = mx->mmx;
    imx = mx->imx;
    dmx = mx->dmx;

    /* Initialization of trace
     * We do it back to front; ReverseTrace() is called later.
     */
    tr->statetype[0] = STT;
    tr->nodeidx[0] = 0;
    tr->pos[0] = 0;
    tr->statetype[1] = STC;
    tr->nodeidx[1] = 0;
    tr->pos[1] = 0;
    tpos = 2;
    i = N;			/* current i (seq pos) we're trying to assign */

    /* Traceback
     */
    while (tr->statetype[tpos - 1] != STS) {
        switch (tr->statetype[tpos - 1]) {
        case STM:			/* M connects from i-1,k-1, or B */
            sc = mmx[i + 1][k + 1] - hmm->msc[(int) dsq[i + 1]][k + 1];
            if (sc <= -INFTY) {
                P7FreeTrace (tr);
                *ret_tr = NULL;
                return;
            }
            else if (sc == xmx[i][XMB] + hmm->bsc[k + 1]) {
                /* Check for wing unfolding */
                if (Prob2Score (hmm->begin[k + 1], hmm->p1) + 1 * INTSCALE <=
                        hmm->bsc[k + 1])
                    while (k > 0) {
                        tr->statetype[tpos] = STD;
                        tr->nodeidx[tpos] = k--;
                        tr->pos[tpos] = 0;
                        tpos++;
                        if (tpos == curralloc) {	/* grow trace if necessary  */
                            curralloc += N;
                            P7ReallocTrace (tr, curralloc);
                        }
                    }

                tr->statetype[tpos] = STB;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }
            else if (sc == mmx[i][k] + hmm->tsc[TMM][k]) {
                tr->statetype[tpos] = STM;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = i--;
            }
            else if (sc == imx[i][k] + hmm->tsc[TIM][k]) {
                tr->statetype[tpos] = STI;
                tr->nodeidx[tpos] = k;
                tr->pos[tpos] = i--;
            }
            else if (sc == dmx[i][k] + hmm->tsc[TDM][k]) {
                tr->statetype[tpos] = STD;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = 0;
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STD:			/* D connects from M,D */
            if (dmx[i][k + 1] <= -INFTY) {
                P7FreeTrace (tr);
                *ret_tr = NULL;
                return;
            }
            else if (dmx[i][k + 1] == mmx[i][k] + hmm->tsc[TMD][k]) {
                tr->statetype[tpos] = STM;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = i--;
            }
            else if (dmx[i][k + 1] == dmx[i][k] + hmm->tsc[TDD][k]) {
                tr->statetype[tpos] = STD;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = 0;
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STI:			/* I connects from M,I */
            sc = imx[i + 1][k] - hmm->isc[(int) dsq[i + 1]][k];
            if (sc <= -INFTY) {
                P7FreeTrace (tr);
                *ret_tr = NULL;
                return;
            }
            else if (sc == mmx[i][k] + hmm->tsc[TMI][k]) {
                tr->statetype[tpos] = STM;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = i--;
            }
            else if (sc == imx[i][k] + hmm->tsc[TII][k]) {
                tr->statetype[tpos] = STI;
                tr->nodeidx[tpos] = k;
                tr->pos[tpos] = i--;
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STN:			/* N connects from S, N */
            if (i == 0 && xmx[i][XMN] == 0) {
                tr->statetype[tpos] = STS;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }
            else if (i > 0 && xmx[i + 1][XMN] == xmx[i][XMN] + hmm->xsc[XTN][LOOP]) {
                tr->statetype[tpos] = STN;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* note convention adherence:  */
                tr->pos[tpos - 1] = i--;	/* first N doesn't emit        */
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STB:			/* B connects from N, J */
            if (xmx[i][XMB] <= -INFTY) {
                P7FreeTrace (tr);
                *ret_tr = NULL;
                return;
            }
            else if (xmx[i][XMB] == xmx[i][XMN] + hmm->xsc[XTN][MOVE]) {
                tr->statetype[tpos] = STN;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }
            else if (xmx[i][XMB] == xmx[i][XMJ] + hmm->xsc[XTJ][MOVE]) {
                tr->statetype[tpos] = STJ;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }

            else {
                Die ("traceback failed");
            }
            break;

        case STE:			/* E connects from any M state. k set here */
            if (xmx[i][XME] <= -INFTY) {
                P7FreeTrace (tr);
                *ret_tr = NULL;
                return;
            }
            for (k = hmm->M; k >= 1; k--) {
                if (xmx[i][XME] == mmx[i][k] + hmm->esc[k]) {
                    /* check for wing unfolding */
                    if (Prob2Score (hmm->end[k], 1.) + 1 * INTSCALE <= hmm->esc[k]) {
                        int dk;		/* need a tmp k while moving thru delete wing */
                        for (dk = hmm->M; dk > k; dk--) {
                            tr->statetype[tpos] = STD;
                            tr->nodeidx[tpos] = dk;
                            tr->pos[tpos] = 0;
                            tpos++;
                            if (tpos == curralloc) {	/* grow trace if necessary  */
                                curralloc += N;
                                P7ReallocTrace (tr, curralloc);
                            }
                        }
                    }

                    tr->statetype[tpos] = STM;
                    tr->nodeidx[tpos] = k--;
                    tr->pos[tpos] = i--;
                    break;
                }
						}		
            if (k < 0) {
                Die ("traceback failed");
            }
            break;

        case STC:			/* C comes from C, E */
            if (xmx[i][XMC] <= -INFTY) {
                P7FreeTrace (tr);
                *ret_tr = NULL;
                return;
            }
            else if (xmx[i][XMC] == xmx[i - 1][XMC] + hmm->xsc[XTC][LOOP]) {
                tr->statetype[tpos] = STC;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* note convention adherence: */
                tr->pos[tpos - 1] = i--;	/* first C doesn't emit       */
            }
            else if (xmx[i][XMC] == xmx[i][XME] + hmm->xsc[XTE][MOVE]) {
                tr->statetype[tpos] = STE;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* E is a nonemitter */
            }

            else {
                Die ("Traceback failed.");
            }
            break;

        case STJ:			/* J connects from E, J */
            if (xmx[i][XMJ] <= -INFTY) {
                P7FreeTrace (tr);
                *ret_tr = NULL;
                return;
            }
            else if (xmx[i][XMJ] == xmx[i - 1][XMJ] + hmm->xsc[XTJ][LOOP]) {
                tr->statetype[tpos] = STJ;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* note convention adherence: */
                tr->pos[tpos - 1] = i--;	/* first J doesn't emit       */
            }
            else if (xmx[i][XMJ] == xmx[i][XME] + hmm->xsc[XTE][LOOP]) {
                tr->statetype[tpos] = STE;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* E is a nonemitter */
            }

            else {
                Die ("Traceback failed.");
            }
            break;

        default:
            Die ("traceback failed");

        }				/* end switch over statetype[tpos-1] */

        tpos++;
        if (tpos == curralloc) {	/* grow trace if necessary  */
            curralloc += N;
            P7ReallocTrace (tr, curralloc);
        }

    }				/* end traceback, at S state; tpos == tlen now */
    tr->tlen = tpos;
    P7ReverseTrace (tr);
    *ret_tr = tr;
}


/* Function: P7SmallViterbi()
 * Date:     SRE, Fri Mar  6 15:29:41 1998 [St. Louis]
 *
 * Purpose:  Wrapper function, for linear memory alignment
 *           with same arguments as P7Viterbi().
 *
 *           Calls P7ParsingViterbi to break the sequence
 *           into fragments. Then, based on size of fragments,
 *           calls either P7Viterbi() or P7WeeViterbi() to
 *           get traces for them. Finally, assembles all these
 *           traces together to produce an overall optimal
 *           trace for the sequence.
 *
 *           If the trace isn't needed for some reason,
 *           all we do is call P7ParsingViterbi.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           mx     - DP matrix, growable
 *           ret_tr - RETURN: traceback; pass NULL if it's not wanted
 *
 * Returns:  Score of optimal alignment in bits.
 */
float
P7SmallViterbi (char *dsq, int L, struct plan7_s *hmm, struct dpmatrix_s *mx,
                struct p7trace_s **ret_tr)
{
    struct p7trace_s *ctr;	/* collapsed trace of optimal parse */
    struct p7trace_s *tr;		/* full trace of optimal alignment */
    struct p7trace_s **tarr;	/* trace array */
    int ndom;			/* number of subsequences */
    int i;			/* counter over domains   */
    int pos;			/* position in sequence */
    int tpos;			/* position in trace */
    int tlen;			/* length of full trace   */
    int sqlen;			/* length of a subsequence */
    int totlen;			/* length of L matched by model (as opposed to N/C/J) */
    float sc;			/* score of optimal alignment */
    int t2;			/* position in a subtrace */

    /* Step 1. Call P7ParsingViterbi to calculate an optimal parse
     *         of the sequence into single-hit subsequences; this parse
     *         is returned in a "collapsed" trace
     */
    sc = P7ParsingViterbi (dsq, L, hmm, &ctr);

    /* If we don't want full trace, we're done;
     * also, if parsing viterbi returned a NULL trace we're done. */
    if (ctr == NULL || ret_tr == NULL) {
        P7FreeTrace (ctr);
        return sc;
    }

    /* Step 2. Call either P7Viterbi or P7WeeViterbi on each subsequence
     *         to recover a full traceback of each, collecting them in
     *         an array.
     */
    ndom = ctr->tlen / 2 - 1;
    tarr = MallocOrDie (sizeof (struct p7trace_s *) * ndom);
    tlen = totlen = 0;
    for (i = 0; i < ndom; i++) {
        sqlen = ctr->pos[i * 2 + 2] - ctr->pos[i * 2 + 1];	/* length of subseq */

        if (P7ViterbiSize (sqlen, hmm->M) > RAMLIMIT) {
            P7WeeViterbi (dsq + ctr->pos[i * 2 + 1], sqlen, hmm, &(tarr[i]));
        }
        else {
            P7Viterbi (dsq + ctr->pos[i * 2 + 1], sqlen, hmm, mx, &(tarr[i]));
        }

        tlen += tarr[i]->tlen - 4;	/* not counting S->N,...,C->T */
        totlen += sqlen;
    }

    /* Step 3. Compose the subtraces into one big final trace.
     *         This is wasteful because we're going to TraceDecompose()
     *         it again in both hmmsearch and hmmpfam to look at
     *         individual domains; but we do it anyway so the P7SmallViterbi
     *         interface looks exactly like the P7Viterbi interface. Maybe
     *         long traces shouldn't include all the N/J/C states anyway,
     *         since they're unambiguously implied.
     */

    /* Calculate total trace len and alloc;
     * nonemitting SNCT + nonemitting J's + emitting NJC
     */
    tlen += 4 + (ndom - 1) + (L - totlen);
    P7AllocTrace (tlen, &tr);
    tr->tlen = tlen;

    /* Add N-terminal trace framework
     */
    tr->statetype[0] = STS;
    tr->nodeidx[0] = 0;
    tr->pos[0] = 0;
    tr->statetype[1] = STN;
    tr->nodeidx[1] = 0;
    tr->pos[1] = 0;
    tpos = 2;
    /* add implied N's */
    for (pos = 1; pos <= ctr->pos[1]; pos++) {
        tr->statetype[tpos] = STN;
        tr->nodeidx[tpos] = 0;
        tr->pos[tpos] = pos;
        tpos++;
    }

    /* Add each subseq trace in, with its appropriate
     * sequence offset derived from the collapsed trace
     */
    for (i = 0; i < ndom; i++) {	
				/* skip SN, CT framework at ends */
        for (t2 = 2; t2 < tarr[i]->tlen - 2; t2++) {
            tr->statetype[tpos] = tarr[i]->statetype[t2];
            tr->nodeidx[tpos] = tarr[i]->nodeidx[t2];
            if (tarr[i]->pos[t2] > 0) {
                tr->pos[tpos] = tarr[i]->pos[t2] + ctr->pos[i * 2 + 1];
            }
            else {
                tr->pos[tpos] = 0;
            }
            tpos++;
        }
        /* add nonemitting J or C */
        tr->statetype[tpos] = (i == ndom - 1) ? STC : STJ;
        tr->nodeidx[tpos] = 0;
        tr->pos[tpos] = 0;
        tpos++;
        /* add implied emitting J's */
        if (i != ndom - 1)
            for (pos = ctr->pos[i * 2 + 2] + 1; pos <= ctr->pos[(i + 1) * 2 + 1]; pos++) {
                tr->statetype[tpos] = STJ;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = pos;
                tpos++;
            }
    }

    /* add implied C's */
    for (pos = ctr->pos[ndom * 2] + 1; pos <= L; pos++) {
        tr->statetype[tpos] = STC;
        tr->nodeidx[tpos] = 0;
        tr->pos[tpos] = pos;
        tpos++;
    }
    /* add terminal T */
    tr->statetype[tpos] = STT;
    tr->nodeidx[tpos] = 0;
    tr->pos[tpos] = 0;
    tpos++;

    for (i = 0; i < ndom; i++) {
        P7FreeTrace (tarr[i]);
    }
    free (tarr);
    P7FreeTrace (ctr);

    *ret_tr = tr;
    return sc;
}




/* Function: P7ParsingViterbi()
 * Date:     SRE, Wed Mar  4 14:07:31 1998 [St. Louis]
 *
 * Purpose:  The "hmmfs" linear-memory algorithm for finding
 *           the optimal alignment of a very long sequence to
 *           a looping, multihit (e.g. Plan7) model, parsing it into
 *           a series of nonoverlapping subsequences that match
 *           the model once. Other algorithms (e.g. P7Viterbi()
 *           or P7WeeViterbi()) are applied subsequently to
 *           these subsequences to recover complete alignments.
 *
 *           The hmmfs algorithm appears briefly in [Durbin98],
 *           but is otherwise unpublished.
 *
 *           The traceback structure returned is special: a
 *           "collapsed" trace S->B->E->...->B->E->T, where
 *           stateidx is unused and pos is used to indicate the
 *           position of B and E in the sequence. The matched
 *           subsequence is B_pos+1...E_pos. The number of
 *           matches in the trace is (tlen/2)-1.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model (log odds scores ready)
 *           ret_tr - RETURN: a collapsed traceback.
 *
 * Returns:  Score of the optimal Viterbi alignment, in bits.
 */
float
P7ParsingViterbi (char *dsq, int L, struct plan7_s *hmm,
                  struct p7trace_s **ret_tr)
{
    struct dpmatrix_s *mx;	/* two rows of score matrix */
    struct dpmatrix_s *tmx;	/* two rows of misused score matrix: traceback ptrs */
    struct p7trace_s *tr;		/* RETURN: collapsed traceback */
    int **xmx, **mmx, **dmx, **imx;	/* convenience ptrs to score matrix */
    int **xtr, **mtr, **dtr, **itr;	/* convenience ptrs to traceback pointers */
    int *btr, *etr;		/* O(L) trace ptrs for B, E state pts in seq */
    int sc;			/* integer score of optimal alignment  */
    int i, k, tpos;		/* index for seq, model, trace position */
    int cur, prv;			/* indices for rolling dp matrix */
    int curralloc;		/* size of allocation for tr */


    /* Alloc a DP matrix and traceback pointers, two rows each, O(M).
     * Alloc two O(L) arrays to trace back through the sequence thru B and E.
     */
    mx = AllocPlan7Matrix (2, hmm->M, &xmx, &mmx, &imx, &dmx);
    tmx = AllocPlan7Matrix (2, hmm->M, &xtr, &mtr, &itr, &dtr);
    btr = MallocOrDie (sizeof (int) * (L + 1));
    etr = MallocOrDie (sizeof (int) * (L + 1));

    /* Initialization of the zero row.
     */
    xmx[0][XMN] = 0;		/* S->N, p=1            */
    xmx[0][XMB] = hmm->xsc[XTN][MOVE];	/* S->N->B, no N-tail   */
    btr[0] = 0;
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need seq to get here */
    etr[0] = -1;
    for (k = 0; k <= hmm->M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;    /* need seq to get here */
    }

    /* Recursion. Done as a pull. Rolling index trick. Trace ptr propagation trick.
     * Note some slightly wasteful boundary conditions:
     *    tsc[0] = -INFTY for all eight transitions (no node 0)
     *    D_M and I_M are wastefully calculated (they don't exist)
     *
     * Notes on traceback pointer propagation.
     *  - In the path B->E, we propagate the i that B was aligned to in the optimal
     *    alignment, via mtr, dtr, and itr.
     *  - When we reach an E, we record the i of the B it started from in etr.
     *  - In a looping path E->J...->B or terminal path E->C...->T, we propagate
     *    the i that E was aligned to in the optimal alignment via xtr[][XMC]
     *    and xtr[][XMJ].
     *  - When we enter B, we record the i of the best previous E, or 0 if there
     *    isn't one, in btr.
     */
    for (i = 1; i <= L; i++) {
        cur = i % 2;
        prv = !cur;

        mmx[cur][0] = imx[cur][0] = dmx[cur][0] = -INFTY;

        for (k = 1; k <= hmm->M; k++) {
            /* match state */
            mmx[cur][k] = -INFTY;
            if ((sc = mmx[prv][k - 1] + hmm->tsc[TMM][k - 1]) > -INFTY) {
                mmx[cur][k] = sc;
                mtr[cur][k] = mtr[prv][k - 1];
            }
            if ((sc = imx[prv][k - 1] + hmm->tsc[TIM][k - 1]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
                mtr[cur][k] = itr[prv][k - 1];
            }
            if ((sc = xmx[prv][XMB] + hmm->bsc[k]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
                mtr[cur][k] = i - 1;
            }
            if ((sc = dmx[prv][k - 1] + hmm->tsc[TDM][k - 1]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
                mtr[cur][k] = dtr[prv][k - 1];
            }
            if (hmm->msc[(int) dsq[i]][k] != -INFTY) {
                mmx[cur][k] += hmm->msc[(int) dsq[i]][k];
            }
            else {
                mmx[cur][k] = -INFTY;
            }

            /* delete state */
            dmx[cur][k] = -INFTY;
            if ((sc = mmx[cur][k - 1] + hmm->tsc[TMD][k - 1]) > -INFTY) {
                dmx[cur][k] = sc;
                dtr[cur][k] = mtr[cur][k - 1];
            }
            if ((sc = dmx[cur][k - 1] + hmm->tsc[TDD][k - 1]) > dmx[cur][k]) {
                dmx[cur][k] = sc;
                dtr[cur][k] = dtr[cur][k - 1];
            }

            /* insert state */
            if (k < hmm->M) {
                imx[cur][k] = -INFTY;
                if ((sc = mmx[prv][k] + hmm->tsc[TMI][k]) > -INFTY) {
                    imx[cur][k] = sc;
                    itr[cur][k] = mtr[prv][k];
                }
                if ((sc = imx[prv][k] + hmm->tsc[TII][k]) > imx[cur][k]) {
                    imx[cur][k] = sc;
                    itr[cur][k] = itr[prv][k];
                }
                if (hmm->isc[(int) dsq[i]][k] != -INFTY) {
                    imx[cur][k] += hmm->isc[(int) dsq[i]][k];
                }
                else {
                    imx[cur][k] = -INFTY;
                }
            }
        }

        /* Now the special states. Order is important here.
         * remember, C and J emissions are zero score by definition,
         */
        /* N state */
        xmx[cur][XMN] = -INFTY;
        if ((sc = xmx[prv][XMN] + hmm->xsc[XTN][LOOP]) > -INFTY) {
            xmx[cur][XMN] = sc;
        }
        /* E state */
        xmx[cur][XME] = -INFTY;
        for (k = 1; k <= hmm->M; k++) {
            if ((sc = mmx[cur][k] + hmm->esc[k]) > xmx[cur][XME]) {
                xmx[cur][XME] = sc;
                etr[i] = mtr[cur][k];
            }
				}		
        /* J state */
        xmx[cur][XMJ] = -INFTY;
        if ((sc = xmx[prv][XMJ] + hmm->xsc[XTJ][LOOP]) > -INFTY) {
            xmx[cur][XMJ] = sc;
            xtr[cur][XMJ] = xtr[prv][XMJ];
        }
        if ((sc = xmx[cur][XME] + hmm->xsc[XTE][LOOP]) > xmx[cur][XMJ]) {
            xmx[cur][XMJ] = sc;
            xtr[cur][XMJ] = i;
        }
        /* B state */
        xmx[cur][XMB] = -INFTY;
        if ((sc = xmx[cur][XMN] + hmm->xsc[XTN][MOVE]) > -INFTY) {
            xmx[cur][XMB] = sc;
            btr[i] = 0;
        }
        if ((sc = xmx[cur][XMJ] + hmm->xsc[XTJ][MOVE]) > xmx[cur][XMB]) {
            xmx[cur][XMB] = sc;
            btr[i] = xtr[cur][XMJ];
        }
        /* C state */
        xmx[cur][XMC] = -INFTY;
        if ((sc = xmx[prv][XMC] + hmm->xsc[XTC][LOOP]) > -INFTY) {
            xmx[cur][XMC] = sc;
            xtr[cur][XMC] = xtr[prv][XMC];
        }
        if ((sc = xmx[cur][XME] + hmm->xsc[XTE][MOVE]) > xmx[cur][XMC]) {
            xmx[cur][XMC] = sc;
            xtr[cur][XMC] = i;
        }
    }
    /* T state (not stored) */
    sc = xmx[cur][XMC] + hmm->xsc[XTC][MOVE];

    /*****************************************************************
     * Collapsed traceback stage.
     * xtr[L%2][XMC] contains the position j of the previous E
     * etr[j]        contains the position i of the previous B
     * btr[i]        contains the position j of the previous E, or 0
     * continue until btr[i] = 0.
     *****************************************************************/

    curralloc = 2;		/* minimum: no hits */
    P7AllocTrace (curralloc, &tr);

    /* Init of collapsed trace. Back to front; we ReverseTrace() later.
     */
    tpos = 0;
    tr->statetype[tpos] = STT;
    tr->pos[tpos] = 0;
    i = xtr[L % 2][XMC];
    while (i > 0) {
        curralloc += 2;
        P7ReallocTrace (tr, curralloc);

        tpos++;
        tr->statetype[tpos] = STE;
        tr->pos[tpos] = i;
        i = etr[i];

        tpos++;
        tr->statetype[tpos] = STB;
        tr->pos[tpos] = i;
        i = btr[i];
    }

    tpos++;
    tr->statetype[tpos] = STS;
    tr->pos[tpos] = 0;
    tr->tlen = tpos + 1;
    P7ReverseTrace (tr);

    FreePlan7Matrix (mx);
    FreePlan7Matrix (tmx);
    free (btr);
    free (etr);

    *ret_tr = tr;
    return Scorify (sc);
}

/* Function: P7WeeViterbi()
 * Date:     SRE, Wed Mar  4 08:24:04 1998 [St. Louis]
 *
 * Purpose:  Hirschberg/Myers/Miller linear memory alignment.
 *           See [Hirschberg75,MyM-88a] for the idea of the algorithm.
 *           Adapted to HMM implementation.
 *
 *           Requires that you /know/ that there's only
 *           one hit to the model in the sequence: either
 *           because you're forcing single-hit, or you've
 *           previously called P7ParsingViterbi to parse
 *           the sequence into single-hit segments. The reason
 *           for this is that a cyclic model (a la Plan7)
 *           defeats the nice divide and conquer trick.
 *           (I think some trickery with propagated trace pointers
 *           could get around this but haven't explored it.)
 *           This is implemented by ignoring transitions
 *           to/from J state.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           ret_tr - RETURN: traceback.
 *
 * Returns:  Score of the optimal Viterbi alignment.
 */
float
P7WeeViterbi (char *dsq, int L, struct plan7_s *hmm,
              struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;		/* RETURN: traceback */
    int *kassign;			/* 0..L+1, alignment of seq positions to model nodes */
    char *tassign;		/* 0..L+1, alignment of seq positions to state types */
    int *endlist;			/* stack of end points on sequence to work on */
    int *startlist;		/* stack of start points on sequence to work on */
    int lpos;			/* position in endlist, startlist */
    int k1, k2, k3;		/* start, mid, end in model      */
    char t1, t2, t3;		/* start, mid, end in state type */
    int s1, s2, s3;		/* start, mid, end in sequence   */
    float sc;			/* score of segment optimal alignment */
    float ret_sc;			/* optimal score over complete seq */
    int tlen;			/* length needed for trace */
    int i, k, tpos;		/* index in sequence, model, trace */


    /* Initialize.
     */
    kassign = MallocOrDie (sizeof (int) * (L + 1));
    tassign = MallocOrDie (sizeof (char) * (L + 1));
    endlist = MallocOrDie (sizeof (int) * (L + 1));
    startlist = MallocOrDie (sizeof (int) * (L + 1));

    lpos = 0;
    startlist[lpos] = 1;
    endlist[lpos] = L;
    kassign[1] = 1;
    kassign[L] = hmm->M;
    tassign[1] = STS;		/* temporary boundary condition! will become N or M */
    tassign[L] = STT;		/* temporary boundary condition! will become M or C */

    /* Recursive divide-and-conquer alignment.
     */
    while (lpos >= 0) {
        /* Pop a segment off the stack */
        s1 = startlist[lpos];
        k1 = kassign[s1];
        t1 = tassign[s1];
        s3 = endlist[lpos];
        k3 = kassign[s3];
        t3 = tassign[s3];
        lpos--;
        /* find optimal midpoint of segment */
        sc = get_wee_midpt (hmm, dsq, L, k1, t1, s1, k3, t3, s3, &k2, &t2, &s2);
        kassign[s2] = k2;
        tassign[s2] = t2;
        /* score is valid on first pass */
        if (t1 == STS && t3 == STT) {
            ret_sc = sc;
        }

        /* push N-terminal segment on stack */
        if (t2 != STN && (s2 - s1 > 1 || (s2 - s1 == 1 && t1 == STS))) {
            lpos++;
            startlist[lpos] = s1;
            endlist[lpos] = s2;
        }
        /* push C-terminal segment on stack */
        if (t2 != STC && (s3 - s2 > 1 || (s3 - s2 == 1 && t3 == STT))) {
            lpos++;
            startlist[lpos] = s2;
            endlist[lpos] = s3;
        }

        if (t2 == STN) {		/* if we see STN midpoint, we know the whole N-term is STN */
            for (; s2 >= s1; s2--) {
                kassign[s2] = 1;
                tassign[s2] = STN;
            }
        }
        if (t2 == STC) {		/* if we see STC midpoint, we know whole C-term is STC */
            for (; s2 <= s3; s2++) {
                kassign[s2] = hmm->M;
                tassign[s2] = STC;
            }
        }
    }

    /*****************************************************************
     * Construct a traceback structure from kassign/tassign by interpolating
     * necessary states.
     * Trace allocation is as follows. We clearly need L emitting states.
     * We also need nonemitting states as follows:
     * STS,STN,STB,STE,STC,STT = 6
     * STD: count k2-k1-1 in kassign M->M's
     * Also, count N->M's and M->C's (potential wing unfoldings)...
     *   ...and be careful to check wing unfoldings when there aren't
     *      any emitting N or C flanks! (bugfix, 2.1.1b)
     *****************************************************************/

    tlen = L + 6;
    for (i = 1; i < L; i++) {
        if (tassign[i] == STM && tassign[i + 1] == STM) {
            tlen += kassign[i + 1] - kassign[i] - 1;
        }
        if (tassign[i] == STN && tassign[i + 1] == STM) {
            tlen += kassign[i + 1] - 1;
        }
        if (tassign[i] == STM && tassign[i + 1] == STC) {
            tlen += hmm->M - kassign[i];
        }
    }
    if (tassign[1] == STM) {
        tlen += kassign[1] - 1;
    }
    if (tassign[L] == STM) {
        tlen += hmm->M - kassign[L];
    }
    P7AllocTrace (tlen, &tr);

    tr->statetype[0] = STS;
    tr->nodeidx[0] = 0;
    tr->pos[0] = 0;
    tr->statetype[1] = STN;
    tr->nodeidx[1] = 0;
    tr->pos[1] = 0;
    tpos = 2;

    for (i = 1; i <= L; i++) {
        switch (tassign[i]) {
        case STM:
            /* check for first match state */
            if (tr->statetype[tpos - 1] == STN) {
                tr->statetype[tpos] = STB;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
                tpos++;
                /* check for wing unfolding */
                if (Prob2Score (hmm->begin[kassign[i]], hmm->p1) + INTSCALE <=
                        hmm->bsc[kassign[i]])
                    for (k = 1; k < kassign[i]; k++) {
                        tr->statetype[tpos] = STD;
                        tr->nodeidx[tpos] = k;
                        tr->pos[tpos] = 0;
                        tpos++;
                    }
            }
            /* do the match state itself */
            tr->statetype[tpos] = STM;
            tr->nodeidx[tpos] = kassign[i];
            tr->pos[tpos] = i;
            tpos++;
            /* do any deletes necessary 'til next match */
            if (i < L && tassign[i + 1] == STM && kassign[i + 1] - kassign[i] > 1)
                for (k = kassign[i] + 1; k < kassign[i + 1]; k++) {
                    tr->statetype[tpos] = STD;
                    tr->nodeidx[tpos] = k;
                    tr->pos[tpos] = 0;
                    tpos++;
                }
            /* check for last match state */
            if (i == L || tassign[i + 1] == STC) {
                /* check for wing unfolding */
                if (Prob2Score (hmm->end[kassign[i - 1]], 1.) + INTSCALE <=
                        hmm->esc[kassign[i - 1]])
                    for (k = kassign[i] + 1; k <= hmm->M; k++) {
                        tr->statetype[tpos] = STD;
                        tr->nodeidx[tpos] = k;
                        tr->pos[tpos] = 0;
                        tpos++;
                    }
                /* add on the end state */
                tr->statetype[tpos] = STE;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
                tpos++;
                /* and a nonemitting C state */
                tr->statetype[tpos] = STC;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
                tpos++;
            }
            break;

        case STI:
            tr->statetype[tpos] = STI;
            tr->nodeidx[tpos] = kassign[i];
            tr->pos[tpos] = i;
            tpos++;
            break;

        case STN:
            tr->statetype[tpos] = STN;
            tr->nodeidx[tpos] = 0;
            tr->pos[tpos] = i;
            tpos++;
            break;

        case STC:
            tr->statetype[tpos] = STC;
            tr->nodeidx[tpos] = 0;
            tr->pos[tpos] = i;
            tpos++;
            break;

        default:
            Die ("Bogus state %s", Statetype (tassign[i]));
        }
    }
    /* terminate the trace */
    tr->statetype[tpos] = STT;
    tr->nodeidx[tpos] = 0;
    tr->pos[tpos] = 0;
    tr->tlen = tpos + 1;

    *ret_tr = tr;

    free (kassign);
    free (tassign);
    free (startlist);
    free (endlist);
    return ret_sc;
}


/* Function: Plan7ESTViterbi()
 *
 * Purpose:  Frameshift-tolerant alignment of protein model to cDNA EST.
 *
 *
 */
float
Plan7ESTViterbi (char *dsq, int L, struct plan7_s *hmm,
                 struct dpmatrix_s **ret_mx)
{
    struct dpmatrix_s *mx;
    int **xmx;
    int **mmx;
    int **imx;
    int **dmx;
    int i, k;
    int sc;
    int codon;

    /* Allocate a DP matrix with 0..L rows, 0..M+1 columns.
     */
    mx = AllocPlan7Matrix (L + 1, hmm->M, &xmx, &mmx, &imx, &dmx);

    /* Initialization of the zero row (DNA sequence of length 0)
     * Note that xmx[i][stN] = 0 by definition for all i,
     *    and xmx[i][stT] = xmx[i][stC], so neither stN nor stT need
     *    to be calculated in DP matrices.
     */
    xmx[0][XMN] = 0;		/* S->N, p=1            */
    xmx[0][XMB] = hmm->xsc[XTN][MOVE];	/* S->N->B, no N-tail   */
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need seq to get here */
    for (k = 0; k <= hmm->M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;    /* need seq to get here */
    }

    /* Initialization of the first row (DNA sequence of length 1);
     * only N state can make this nucleotide.
     */
    xmx[1][XMN] = xmx[0][XMN] + hmm->xsc[XTN][LOOP];
    xmx[1][XMB] = xmx[1][XMN] + hmm->xsc[XTN][MOVE];
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need 2 nt to get here */
    for (k = 0; k <= hmm->M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;    /* need 2 nt to get into model */
    }

    /* Recursion. Done as a pull.
     * Note some slightly wasteful boundary conditions:
     *    tsc[0] = -INFTY for all eight transitions (no node 0)
     *    D_M and I_M are wastefully calculated (they don't exist)
     */
    for (i = 2; i <= L; i++) {
        mmx[i][0] = imx[i][0] = dmx[i][0] = -INFTY;

        /* crude calculation of lookup value for codon */
        if (i > 2) {
            if (dsq[i - 2] < 4 && dsq[i - 1] < 4 && dsq[i] < 4) {
                codon = dsq[i - 2] * 16 + dsq[i - 1] * 4 + dsq[i];
            }
            else {
                codon = 64;    /* ambiguous codon; punt */
            }
        }

        for (k = 1; k <= hmm->M; k++) {
            /* match state */
            if (i > 2) {
                mmx[i][k] = mmx[i - 3][k - 1] + hmm->tsc[TMM][k - 1];
                if ((sc = imx[i - 3][k - 1] + hmm->tsc[TIM][k - 1]) > mmx[i][k]) {
                    mmx[i][k] = sc;
                }
                if ((sc = xmx[i - 3][XMB] + hmm->bsc[k]) > mmx[i][k]) {
                    mmx[i][k] = sc;
                }
                if ((sc = dmx[i - 3][k - 1] + hmm->tsc[TDM][k - 1]) > mmx[i][k]) {
                    mmx[i][k] = sc;
                }
                mmx[i][k] += hmm->dnam[codon][k];
            }
            /* -1 frameshifts into match state */
            if ((sc =
                        mmx[i - 2][k - 1] + hmm->tsc[TMM][k - 1] + hmm->dna2) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc =
                        imx[i - 2][k - 1] + hmm->tsc[TIM][k - 1] + hmm->dna2) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc = xmx[i - 2][XMB] + hmm->bsc[k] + hmm->dna2) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc =
                        dmx[i - 2][k - 1] + hmm->tsc[TDM][k - 1] + hmm->dna2) > mmx[i][k]) {
                mmx[i][k] = sc;
            }

            /* +1 frameshifts into match state */
            if (i > 3) {
                if ((sc =
                            mmx[i - 4][k - 1] + hmm->tsc[TMM][k - 1] + hmm->dna4) >
                        mmx[i][k]) {
                    mmx[i][k] = sc;
                }
                if ((sc =
                            imx[i - 4][k - 1] + hmm->tsc[TIM][k - 1] + hmm->dna4) >
                        mmx[i][k]) {
                    mmx[i][k] = sc;
                }
                if ((sc = xmx[i - 4][XMB] + hmm->bsc[k] + hmm->dna4) > mmx[i][k]) {
                    mmx[i][k] = sc;
                }
                if ((sc =
                            dmx[i - 4][k - 1] + hmm->tsc[TDM][k - 1] + hmm->dna4) >
                        mmx[i][k]) {
                    mmx[i][k] = sc;
                }
            }
            /* delete state */
            dmx[i][k] = mmx[i][k - 1] + hmm->tsc[TMD][k - 1];
            if ((sc = dmx[i][k - 1] + hmm->tsc[TDD][k - 1]) > dmx[i][k]) {
                dmx[i][k] = sc;
            }

            /* insert state */
            if (i > 2) {
                imx[i][k] = mmx[i - 3][k] + hmm->tsc[TMI][k];
                if ((sc = imx[i - 3][k] + hmm->tsc[TII][k]) > imx[i][k]) {
                    imx[i][k] = sc;
                }
                imx[i][k] += hmm->dnai[codon][k];
            }

            /* -1 frameshifts into insert state */
            if ((sc = mmx[i - 2][k] + hmm->tsc[TMI][k] + hmm->dna2) > imx[i][k]) {
                imx[i][k] = sc;
            }
            if ((sc = imx[i - 2][k] + hmm->tsc[TII][k] + hmm->dna2) > imx[i][k]) {
                imx[i][k] = sc;
            }

            /* +1 frameshifts into insert state */
            if (i > 4) {
                if ((sc = mmx[i - 4][k] + hmm->tsc[TMI][k] + hmm->dna4) > imx[i][k]) {
                    imx[i][k] = sc;
                }
                if ((sc = imx[i - 4][k] + hmm->tsc[TII][k] + hmm->dna4) > imx[i][k]) {
                    imx[i][k] = sc;
                }
            }
        }
        /* Now the special states. Order is important here.
         * remember, C and J emissions are zero score by definition,
         */
        /* N state: +1 nucleotide */
        xmx[i][XMN] = xmx[i - 1][XMN] + hmm->xsc[XTN][LOOP];
        /* E state: collect from M's, and last D  */
        xmx[i][XME] = dmx[i][hmm->M];	/* transition prob from last D = 1.0 */
        for (k = 1; k <= hmm->M; k++) {
            if ((sc = mmx[i][k] + hmm->esc[k]) > xmx[i][XME]) {
                xmx[i][XME] = sc;
            }
				}		
        /* J state: +1 nucleotide */
        xmx[i][XMJ] = xmx[i - 1][XMJ] + hmm->xsc[XTJ][LOOP];
        if ((sc = xmx[i][XME] + hmm->xsc[XTE][LOOP]) > xmx[i][XMJ]) {
            xmx[i][XMJ] = sc;
        }
        /* B state: collect from N,J */
        xmx[i][XMB] = xmx[i][XMN] + hmm->xsc[XTN][MOVE];
        if ((sc = xmx[i][XMJ] + hmm->xsc[XTJ][MOVE]) > xmx[i][XMB]) {
            xmx[i][XMB] = sc;
        }
        /* C state: +1 nucleotide */
        xmx[i][XMC] = xmx[i - 1][XMC] + hmm->xsc[XTC][LOOP];
        if ((sc = xmx[i][XME] + hmm->xsc[XTE][MOVE]) > xmx[i][XMC]) {
            xmx[i][XMC] = sc;
        }
    }

    sc = xmx[L][XMC] + hmm->xsc[XTC][MOVE];

    if (ret_mx != NULL) {
        *ret_mx = mx;
    }
    else {
        FreePlan7Matrix (mx);
    }

    return Scorify (sc);		/* the total Viterbi score. */
}



/* Function: get_wee_midpt()
 * Date:     SRE, Wed Mar  4 08:27:11 1998 [St. Louis]
 *
 * Purpose:  The heart of the divide and conquer algorithm
 *           for P7WeeViterbi(). This function is called
 *           recursively to find successive optimal midpoints
 *           in the alignment matrix. See P7WeeViterbi() for
 *           further comments on the assumptions of this algorithm.
 *
 * Args:     hmm   - the model, set up for integer scores
 *           dsq   - the sequence, digitized
 *           L     - length of the sequence
 *           k1    - model node to start with, 1..M
 *           t1    - state type to start with, STM | STI | STN | STC; STS to start
 *           s1    - sequence position to start with, 1..L; 1 to start
 *           k3    - model node to end with, 1..M
 *           t3    - state type to end with, STM | STI | STN | STC; STT to start
 *           s3    - sequence position to end with, 1..L; L to start
 *          ret_k2 - RETURN: optimal midpoint, node position in model
 *          ret_t2 - RETURN: optimal midpoint, state type
 *          ret_s2 - RETURN: optimal midpoint, sequence position
 *
 * Returns: score of optimal alignment, in bits.
 */
static float
get_wee_midpt (struct plan7_s *hmm, char *dsq, int L,
               int k1, char t1, int s1,
               int k3, char t3, int s3,
               int *ret_k2, char *ret_t2, int *ret_s2)
{
    struct dpmatrix_s *fwd;
    struct dpmatrix_s *bck;
    int **xmx;			/* convenience ptr into special states */
    int **mmx;			/* convenience ptr into match states   */
    int **imx;			/* convenience ptr into insert states  */
    int **dmx;			/* convenience ptr into delete states  */
    int k2;
    char t2;
    int s2;
    int cur, prv, nxt;		/* current, previous, next row index (0 or 1) */
    int i, k;			/* indices for seq, model */
    int sc;			/* integer score */
    int max;			/* maximum integer score */
    int start;			/* s1 to start at (need, for STS special case) */


    /* Choose our midpoint.
     * Special cases: s1, s3 adjacent and t1 == STS: s2 = s1
     *                s1, s3 adjacent and t3 == STT: s2 = s3
     *                (where we must replace STS, STT eventually)
     */
    s2 = s1 + (s3 - s1) / 2;
    if (s3 - s1 == 1 && t1 == STS) {
        s2 = s1;
    }
    if (s3 - s1 == 1 && t3 == STT) {
        s2 = s3;
    }

    /* STS is a special case. STS aligns to row zero by convention,
     * but we'll be passed s1=1, t1=STS. We have to init on row
     * zero then start DP on row 1.
     */
    start = (t1 == STS) ? 0 : s1;

    /* Allocate our forward two rows.
     * Initialize row zero.
     */
    fwd = AllocPlan7Matrix (2, hmm->M, &xmx, &mmx, &imx, &dmx);
    cur = start % 2;
    xmx[cur][XMN] = xmx[cur][XMB] = -INFTY;
    xmx[cur][XME] = xmx[cur][XMC] = -INFTY;
    for (k = k1; k <= k3; k++) {
        mmx[cur][k] = imx[cur][k] = dmx[cur][k] = -INFTY;
    }

    /* Where to put our zero for our start point...
     * (only possible to start on an emitting state; J disallowed)
     */
    switch (t1) {
    case STM:
        mmx[cur][k1] = 0;
        break;
    case STI:
        imx[cur][k1] = 0;
        break;
    case STN:
        xmx[cur][XMN] = 0;
        break;
    case STC:
        xmx[cur][XMC] = 0;
        break;
    case STS:
        xmx[cur][XMN] = 0;
        break;
    default:
        Die ("you can't init get_wee_midpt with a %s\n", Statetype (t1));
    }

    /* Still initializing.
     * Deal with pulling horizontal matrix moves in initial row.
     * These are any transitions to nonemitters:
     *    STM-> E, D
     *    STI-> none
     *    STN-> B
     *    STC-> (T, but we never observe this in the forward pass of a d&c)
     *    STE-> C
     *    STS-> (N, already implied by setting xmx[cur][XMN] = 0)
     *    STB-> M
     */
    if (t1 == STM) {
        for (k = k1 + 1; k <= k3; k++) {	
						/* transits into STD */
            dmx[cur][k] = -INFTY;
            if ((sc = mmx[cur][k - 1] + hmm->tsc[TMD][k - 1]) > -INFTY) {
                dmx[cur][k] = sc;
            }
            if ((sc = dmx[cur][k - 1] + hmm->tsc[TDD][k - 1]) > dmx[cur][k]) {
                dmx[cur][k] = sc;
            }
        }
        /* transit into STE */
        xmx[cur][XME] = -INFTY;
        if ((sc = mmx[cur][k1] + hmm->esc[k1]) > -INFTY) {
            xmx[cur][XME] = sc;
        }
    }
    /* transit into STB from STN */
    xmx[cur][XMB] = -INFTY;
    if ((sc = xmx[cur][XMN] + hmm->xsc[XTN][MOVE]) > -INFTY) {
        xmx[cur][XMB] = sc;
    }
    /* transit into STC from STE */
    xmx[cur][XMC] = -INFTY;
    if ((sc = xmx[cur][XME] + hmm->xsc[XTE][MOVE]) > -INFTY) {
        xmx[cur][XMC] = sc;
    }

    /* Done initializing.
     * Start recursive DP; sweep forward to chosen s2 midpoint. Done as a pull.
     */
    for (i = start + 1; i <= s2; i++) {
        cur = i % 2;
        prv = !cur;

        mmx[cur][k1] = imx[cur][k1] = dmx[cur][k1] = -INFTY;

        /* Insert state in column k1, and B->M transition in k1.
         */
        if (k1 < hmm->M) {
            imx[cur][k1] = -INFTY;
            if ((sc = mmx[prv][k1] + hmm->tsc[TMI][k1]) > -INFTY) {
                imx[cur][k1] = sc;
            }
            if ((sc = imx[prv][k1] + hmm->tsc[TII][k1]) > imx[cur][k1]) {
                imx[cur][k1] = sc;
            }
            if (hmm->isc[(int) dsq[i]][k1] != -INFTY) {
                imx[cur][k1] += hmm->isc[(int) dsq[i]][k1];
            }
            else {
                imx[cur][k1] = -INFTY;
            }
        }
        if ((sc = xmx[prv][XMB] + hmm->bsc[k1]) > -INFTY) {
            mmx[cur][k1] = sc;
        }
        if (hmm->msc[(int) dsq[i]][k1] != -INFTY) {
            mmx[cur][k1] += hmm->msc[(int) dsq[i]][k1];
        }
        else {
            mmx[cur][k1] = -INFTY;
        }

        /* Main chunk of recursion across model positions
         */
        for (k = k1 + 1; k <= k3; k++) {
            /* match state */
            mmx[cur][k] = -INFTY;
            if ((sc = mmx[prv][k - 1] + hmm->tsc[TMM][k - 1]) > -INFTY) {
                mmx[cur][k] = sc;
            }
            if ((sc = imx[prv][k - 1] + hmm->tsc[TIM][k - 1]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
            }
            if ((sc = xmx[prv][XMB] + hmm->bsc[k]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
            }
            if ((sc = dmx[prv][k - 1] + hmm->tsc[TDM][k - 1]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
            }
            if (hmm->msc[(int) dsq[i]][k] != -INFTY) {
                mmx[cur][k] += hmm->msc[(int) dsq[i]][k];
            }
            else {
                mmx[cur][k] = -INFTY;
            }

            /* delete state */
            dmx[cur][k] = -INFTY;
            if (k < hmm->M) {
                if ((sc = mmx[cur][k - 1] + hmm->tsc[TMD][k - 1]) > -INFTY) {
                    dmx[cur][k] = sc;
                }
                if ((sc = dmx[cur][k - 1] + hmm->tsc[TDD][k - 1]) > dmx[cur][k]) {
                    dmx[cur][k] = sc;
                }
            }

            /* insert state */
            imx[cur][k] = -INFTY;
            if (k < hmm->M) {
                if ((sc = mmx[prv][k] + hmm->tsc[TMI][k]) > -INFTY) {
                    imx[cur][k] = sc;
                }
                if ((sc = imx[prv][k] + hmm->tsc[TII][k]) > imx[cur][k]) {
                    imx[cur][k] = sc;
                }
                if (hmm->isc[(int) dsq[i]][k] != -INFTY) {
                    imx[cur][k] += hmm->isc[(int) dsq[i]][k];
                }
                else {
                    imx[cur][k] = -INFTY;
                }
            }
        }
        /* N state */
        xmx[cur][XMN] = -INFTY;
        if ((sc = xmx[prv][XMN] + hmm->xsc[XTN][LOOP]) > -INFTY) {
            xmx[cur][XMN] = sc;
        }
        /* E state */
        xmx[cur][XME] = -INFTY;
        for (k = k1; k <= k3 && k <= hmm->M; k++) {
            if ((sc = mmx[cur][k] + hmm->esc[k]) > xmx[cur][XME]) {
                xmx[cur][XME] = sc;
            }
				}		
        /* B state */
        xmx[cur][XMB] = -INFTY;
        if ((sc = xmx[cur][XMN] + hmm->xsc[XTN][MOVE]) > -INFTY) {
            xmx[cur][XMB] = sc;
        }
        /* C state */
        xmx[cur][XMC] = -INFTY;
        if ((sc = xmx[prv][XMC] + hmm->xsc[XTC][LOOP]) > -INFTY) {
            xmx[cur][XMC] = sc;
        }
        if ((sc = xmx[cur][XME] + hmm->xsc[XTE][MOVE]) > xmx[cur][XMC]) {
            xmx[cur][XMC] = sc;
        }
    }

    /* Row s2%2 in fwd matrix now contains valid scores from s1 (start) to s2,
     * with J transitions disallowed (no cycles through model).
     */

    /*****************************************************************
     * Backwards pass.
     *****************************************************************/

    /* Allocate our backwards two rows. Init last row.
     */
    bck = AllocPlan7Matrix (2, hmm->M, &xmx, &mmx, &imx, &dmx);
    nxt = s3 % 2;
    xmx[nxt][XMN] = xmx[nxt][XMB] = -INFTY;
    xmx[nxt][XME] = xmx[nxt][XMC] = -INFTY;
    for (k = k1; k <= k3 + 1; k++) {
        mmx[nxt][k] = imx[nxt][k] = dmx[nxt][k] = -INFTY;
    }
    cur = !nxt;
    mmx[cur][k3 + 1] = imx[cur][k3 + 1] = dmx[cur][k3 + 1] = -INFTY;

    /* Where to put the zero for our end point on last row.
     */
    switch (t3) {
    case STM:
        mmx[nxt][k3] = 0;
        break;
    case STI:
        imx[nxt][k3] = 0;
        break;
    case STN:
        xmx[nxt][XMN] = 0;
        break;
    case STC:
        xmx[nxt][XMC] = 0;
        break;			/* must be an emitting C */
    case STT:
        xmx[nxt][XMC] = hmm->xsc[XTC][MOVE];
        break;			/* C->T implied */
    default:
        Die ("you can't init get_wee_midpt with a %s\n", Statetype (t3));
    }

    /* Still initializing.
     * In the case t3==STT, there are a few horizontal moves possible
     * on row s3, because STT isn't an emitter. All other states are
     * emitters, so their connections have to be to the previous row s3-1.
     */
    if (t3 == STT) {		/* E->C */
        xmx[nxt][XME] = xmx[nxt][XMC] + hmm->xsc[XTE][MOVE];
        /* M->E */
        for (k = k3; k >= k1; k--) {
            mmx[nxt][k] = xmx[nxt][XME] + hmm->esc[k];
            if (s3 != s2) {
                mmx[nxt][k] += hmm->msc[(int) dsq[s3]][k];
            }
        }
    }

    /* Start recursive DP; sweep backwards to chosen s2 midpoint.
     * Done as a pull. M, I scores at current row do /not/ include
     * emission scores. Be careful of integer underflow.
     */
    for (i = s3 - 1; i >= s2; i--) {
        /* note i < L, so i+1 is always a legal index */
        cur = i % 2;
        nxt = !cur;
        /* C pulls from C (T is special cased) */
        xmx[cur][XMC] = -INFTY;
        if ((sc = xmx[nxt][XMC] + hmm->xsc[XTC][LOOP]) > -INFTY) {
            xmx[cur][XMC] = sc;
        }
        /* B pulls from M's */
        xmx[cur][XMB] = -INFTY;
        for (k = k1; k <= k3; k++) {
            if ((sc = mmx[nxt][k] + hmm->bsc[k]) > xmx[cur][XMB]) {
                xmx[cur][XMB] = sc;
            }
				}		
        /* E pulls from C (J disallowed) */
        xmx[cur][XME] = -INFTY;
        if ((sc = xmx[cur][XMC] + hmm->xsc[XTE][MOVE]) > -INFTY) {
            xmx[cur][XME] = sc;
        }
        /* N pulls from B, N */
        xmx[cur][XMN] = -INFTY;
        if ((sc = xmx[cur][XMB] + hmm->xsc[XTN][MOVE]) > -INFTY) {
            xmx[cur][XMN] = sc;
        }
        if ((sc = xmx[nxt][XMN] + hmm->xsc[XTN][LOOP]) > xmx[cur][XMN]) {
            xmx[cur][XMN] = sc;
        }

        /* Main recursion across model
         */
        for (k = k3; k >= k1; k--) {
            /* special case k == M */
            if (k == hmm->M) {
                mmx[cur][k] = xmx[cur][XME];	/* p=1 transition to E by definition */
                dmx[cur][k] = -INFTY;	/* doesn't exist */
                imx[cur][k] = -INFTY;	/* doesn't exist */
                if (i != s2) {
                    mmx[cur][k] += hmm->msc[(int) dsq[i]][k];
                }
                continue;
            }				/* below this k < M, so k+1 is a legal index */

            /* pull into match state */
            mmx[cur][k] = -INFTY;
            if ((sc = xmx[cur][XME] + hmm->esc[k]) > -INFTY) {
                mmx[cur][k] = sc;
            }
            if ((sc = mmx[nxt][k + 1] + hmm->tsc[TMM][k]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
            }
            if ((sc = imx[nxt][k] + hmm->tsc[TMI][k]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
            }
            if ((sc = dmx[cur][k + 1] + hmm->tsc[TMD][k]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
            }
            if (i != s2) {
                mmx[cur][k] += hmm->msc[(int) dsq[i]][k];
            }

            /* pull into delete state */
            dmx[cur][k] = -INFTY;
            if ((sc = mmx[nxt][k + 1] + hmm->tsc[TDM][k]) > -INFTY) {
                dmx[cur][k] = sc;
            }
            if ((sc = dmx[cur][k + 1] + hmm->tsc[TDD][k]) > dmx[cur][k]) {
                dmx[cur][k] = sc;
            }
            /* pull into insert state */
            imx[cur][k] = -INFTY;
            if ((sc = mmx[nxt][k + 1] + hmm->tsc[TIM][k]) > -INFTY) {
                imx[cur][k] = sc;
            }
            if ((sc = imx[nxt][k] + hmm->tsc[TII][k]) > imx[cur][k]) {
                imx[cur][k] = sc;
            }
            if (i != s2) {
                imx[cur][k] += hmm->isc[(int) dsq[i]][k];
            }

        }
    }

    /*****************************************************************
     * DP complete; we have both forward and backward passes. Now we
     * look across the s2 row and find the optimal emitting state.
     *****************************************************************/

    cur = s2 % 2;
    max = -INFTY;
    for (k = k1; k <= k3; k++) {
        if ((sc = fwd->mmx[cur][k] + bck->mmx[cur][k]) > max) {
            k2 = k;
            t2 = STM;
            max = sc;
        }
        if ((sc = fwd->imx[cur][k] + bck->imx[cur][k]) > max) {
            k2 = k;
            t2 = STI;
            max = sc;
        }
    }
    if ((sc = fwd->xmx[cur][XMN] + bck->xmx[cur][XMN]) > max) {
        k2 = 1;
        t2 = STN;
        max = sc;
    }
    if ((sc = fwd->xmx[cur][XMC] + bck->xmx[cur][XMC]) > max) {
        k2 = hmm->M;
        t2 = STC;
        max = sc;
    }

    /*****************************************************************
     * Garbage collection, return.
     *****************************************************************/

    FreePlan7Matrix (fwd);
    FreePlan7Matrix (bck);
    *ret_k2 = k2;
    *ret_t2 = t2;
    *ret_s2 = s2;
    return Scorify (max);
}


/* Function: P7ViterbiAlignAlignment()
 * Date:     SRE, Sat Jul  4 13:39:00 1998 [St. Louis]
 *
 * Purpose:  Align a multiple alignment to an HMM without
 *           changing the multiple alignment itself.
 *           Adapted from P7Viterbi().
 *
 *           Heuristic; not a guaranteed optimal alignment.
 *           Guaranteeing an optimal alignment appears difficult.
 *           [cryptic note to myself:] In paths connecting to I* metastates,
 *           recursion breaks down; if there is a gap in the
 *           previous column for a given seq, we can't determine what state the
 *           I* metastate corresponds to for this sequence, unless we
 *           look back in the DP matrix. The lookback would either involve
 *           recursing back to the previous M* metastate (giving a
 *           O(MN^2) algorithm instead of O(MN)) or expanding the I*
 *           metastate into 3^nseq separate I* metastates to keep track
 *           of which of three states each seq is in. Since the second
 *           option blows up exponentially w/ nseq, it is not attractive.
 *           If the first option were used, the correct algorithm would be related to
 *           modelmakers.c:Maxmodelmaker(), but somewhat more difficult.
 *
 *           The heuristic approach here is to calculate a "consensus"
 *           sequence from the alignment, and align the consensus to the HMM.
 *           Some hackery is employed, weighting transitions and emissions
 *           to make things work (re: con and mocc arrays).
 *
 * Args:     aseq  - aligned sequences
 *           ainfo - info for aseqs (includes alen, nseq, wgt)
 *           hmm   - model to align to
 *
 * Returns:  Traceback. Caller must free with P7FreeTrace().
 *           pos[] contains alignment columns, indexed 1..alen.
 *           statetype[] contains metastates M*, etc. as STM, etc.
 */
struct p7trace_s *
P7ViterbiAlignAlignment (MSA * msa, struct plan7_s *hmm)
{
    struct dpmatrix_s *mx;	/* Viterbi calculation lattice (two rows) */
    struct dpshadow_s *tb;	/* shadow matrix of traceback pointers */
    struct p7trace_s *tr;		/* RETURN: traceback */
    int **xmx, **mmx, **imx, **dmx;
    char **xtb, **mtb, **itb, **dtb;
    float **con;			/* [1..alen][0..Alphabet_size-1], consensus counts */
    float *mocc;			/* fractional occupancy of a column; used to weight transitions */
    int i;			/* counter for columns */
    int k;			/* counter for model positions */
    int idx;			/* counter for seqs */
    int sym;			/* counter for alphabet symbols */
    int sc;			/* temp variable for holding score */
    float denom;			/* total weight of seqs; used to "normalize" counts */
    int cur, prv;

    /* The "consensus" is a counts matrix, [1..alen][0..Alphabet_size-1].
     * Gaps are not counted explicitly, but columns with lots of gaps get
     * less total weight because they have fewer counts.
     */
    /* allocation */
    con = MallocOrDie (sizeof (float *) * (msa->alen + 1));
    mocc = MallocOrDie (sizeof (float) * (msa->alen + 1));
    for (i = 1; i <= msa->alen; i++) {
        con[i] = MallocOrDie (sizeof (float) * Alphabet_size);
        FSet (con[i], Alphabet_size, 0.0);
    }
    mocc[0] = -9999.;
    /* initialization */
    /* note: aseq is off by one, 0..alen-1 */
    /* "normalized" to have a max total count of 1 per col */
    denom = FSum (msa->wgt, msa->nseq);
    for (i = 1; i <= msa->alen; i++) {
        for (idx = 0; idx < msa->nseq; idx++) {
            if (!isgap (msa->aseq[idx][i - 1])) {
                P7CountSymbol (con[i], SYMIDX (msa->aseq[idx][i - 1]), msa->wgt[idx]);
            }
				}		
        FScale (con[i], Alphabet_size, 1. / denom);
        mocc[i] = FSum (con[i], Alphabet_size);
    }

    /* Allocate a DP matrix with 2 rows, 0..M columns,
     * and a shadow matrix with 0,1..alen rows, 0..M columns.
     */
    mx = AllocPlan7Matrix (2, hmm->M, &xmx, &mmx, &imx, &dmx);
    tb = AllocShadowMatrix (msa->alen + 1, hmm->M, &xtb, &mtb, &itb, &dtb);

    /* Initialization of the zero row.
     */
    xmx[0][XMN] = 0;		/* S->N, p=1            */
    xtb[0][XMN] = STS;
    xmx[0][XMB] = hmm->xsc[XTN][MOVE];	/* S->N->B, no N-tail   */
    xtb[0][XMB] = STN;
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need seq to get here */
    tb->esrc[0] = 0;
    xtb[0][XMC] = xtb[0][XMJ] = STBOGUS;
    for (k = 0; k <= hmm->M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;	/* need seq to get here */
        mtb[0][k] = itb[0][k] = dtb[0][k] = STBOGUS;
    }

    /* Recursion. Done as a pull.
     * Note some slightly wasteful boundary conditions:
     *    tsc[0] = -INFTY for all eight transitions (no node 0)
     *    D_M and I_M are wastefully calculated (they don't exist)
     */
    for (i = 1; i <= msa->alen; i++) {
        cur = i % 2;
        prv = !cur;

        mmx[cur][0] = imx[cur][0] = dmx[cur][0] = -INFTY;
        mtb[i][0] = itb[i][0] = dtb[i][0] = STBOGUS;

        for (k = 1; k <= hmm->M; k++) {
            /* match state */
            mmx[cur][k] = -INFTY;
            mtb[i][k] = STBOGUS;
            if (mmx[prv][k - 1] > -INFTY && hmm->tsc[TMM][k - 1] > -INFTY &&
                    (sc = mmx[prv][k - 1] + hmm->tsc[TMM][k - 1]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
                mtb[i][k] = STM;
            }
            if (imx[prv][k - 1] > -INFTY && hmm->tsc[TIM][k - 1] > -INFTY &&
                    (sc =
                         imx[prv][k - 1] + hmm->tsc[TIM][k - 1] * mocc[i - 1]) >
                    mmx[cur][k]) {
                mmx[cur][k] = sc;
                mtb[i][k] = STI;
            }
            if ((sc = xmx[prv][XMB] + hmm->bsc[k]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
                mtb[i][k] = STB;
            }
            if (dmx[prv][k - 1] > -INFTY && hmm->tsc[TDM][k - 1] > -INFTY &&
                    (sc = dmx[prv][k - 1] + hmm->tsc[TDM][k - 1]) > mmx[cur][k]) {
                mmx[cur][k] = sc;
                mtb[i][k] = STD;
            }
            /* average over "consensus" sequence */
            for (sym = 0; sym < Alphabet_size; sym++) {
                if (con[i][sym] > 0 && hmm->msc[sym][k] == -INFTY) {
                    mmx[cur][k] = -INFTY;
                    break;
                }
                mmx[cur][k] += hmm->msc[sym][k] * con[i][sym];
            }

            /* delete state */
            dmx[cur][k] = -INFTY;
            dtb[i][k] = STBOGUS;
            if (mmx[cur][k - 1] > -INFTY && hmm->tsc[TMD][k - 1] > -INFTY &&
                    (sc = mmx[cur][k - 1] + hmm->tsc[TMD][k - 1]) > dmx[cur][k]) {
                dmx[cur][k] = sc;
                dtb[i][k] = STM;
            }
            if (dmx[cur][k - 1] > -INFTY && hmm->tsc[TDD][k - 1] > -INFTY &&
                    (sc = dmx[cur][k - 1] + hmm->tsc[TDD][k - 1]) > dmx[cur][k]) {
                dmx[cur][k] = sc;
                dtb[i][k] = STD;
            }

            /* insert state */
            if (k < hmm->M) {
                imx[cur][k] = -INFTY;
                itb[i][k] = STBOGUS;
                if (mmx[prv][k] > -INFTY && hmm->tsc[TMI][k] > -INFTY &&
                        (sc = mmx[prv][k] + hmm->tsc[TMI][k] * mocc[i]) > imx[cur][k]) {
                    imx[cur][k] = sc;
                    itb[i][k] = STM;
                }
                if (imx[prv][k] > -INFTY && hmm->tsc[TII][k] > -INFTY &&
                        (sc =
                             imx[prv][k] + hmm->tsc[TII][k] * mocc[i - 1] * mocc[i]) >
                        imx[cur][k]) {
                    imx[cur][k] = sc;
                    itb[i][k] = STI;
                }
                /* average over "consensus" sequence */
                for (sym = 0; sym < Alphabet_size; sym++) {
                    if (con[i][sym] > 0 && hmm->isc[sym][k] == -INFTY) {
                        imx[cur][k] = -INFTY;
                        break;
                    }
                    imx[cur][k] += hmm->isc[sym][k] * con[i][sym];
                }
            }
        }

        /* Now the special states. Order is important here.
         * remember, N, C, and J emissions are zero score by definition.
         */
        /* N state */
        xmx[cur][XMN] = -INFTY;
        xtb[i][XMN] = STBOGUS;
        if (xmx[prv][XMN] > -INFTY && hmm->xsc[XTN][LOOP] > -INFTY &&
                (sc = xmx[prv][XMN] + hmm->xsc[XTN][LOOP] * mocc[i]) > -INFTY) {
            xmx[cur][XMN] = sc;
            xtb[i][XMN] = STN;
        }
        /* E state */
        xmx[cur][XME] = -INFTY;
        xtb[i][XME] = STBOGUS;
        for (k = 1; k <= hmm->M; k++) {
            if (mmx[cur][k] > -INFTY && hmm->esc[k] > -INFTY &&
                    (sc = mmx[cur][k] + hmm->esc[k]) > xmx[cur][XME]) {
                xmx[cur][XME] = sc;
                tb->esrc[i] = k;
            }
				}
        /* we don't check J state */
        /* B state; don't connect from J */
        xmx[cur][XMB] = -INFTY;
        xtb[i][XMB] = STBOGUS;
        if (xmx[cur][XMN] > -INFTY && hmm->xsc[XTN][MOVE] > -INFTY &&
                (sc = xmx[cur][XMN] + hmm->xsc[XTN][MOVE]) > xmx[cur][XMB]) {
            xmx[cur][XMB] = sc;
            xtb[i][XMB] = STN;
        }

        /* C state */
        xmx[cur][XMC] = -INFTY;
        xtb[i][XMC] = STBOGUS;
        if (xmx[prv][XMC] > -INFTY && hmm->xsc[XTC][LOOP] > -INFTY &&
                (sc = xmx[prv][XMC] + hmm->xsc[XTC][LOOP] * mocc[i]) > -INFTY) {
            xmx[cur][XMC] = sc;
            xtb[i][XMC] = STC;
        }
        if (xmx[cur][XME] > -INFTY && hmm->xsc[XTE][MOVE] > -INFTY &&
                (sc = xmx[cur][XME] + hmm->xsc[XTE][MOVE]) > xmx[cur][XMC]) {
            xmx[cur][XMC] = sc;
            xtb[i][XMC] = STE;
        }
    }
    /* T state (not stored in mx) */
    sc = xmx[msa->alen % 2][XMC] + hmm->xsc[XTC][MOVE];

    /* do the traceback */
    tr = ShadowTrace (tb, hmm, msa->alen);
    /* cleanup and return */
    FreePlan7Matrix (mx);
    FreeShadowMatrix (tb);
    for (i = 1; i <= msa->alen; i++) {
        free (con[i]);
    }
    free (con);
    free (mocc);

    return tr;
}



/* Function: ShadowTrace()
 * Date:     SRE, Sun Jul  5 11:38:24 1998 [St. Louis]
 *
 * Purpose:  Given a shadow matrix, trace it back, and return
 *           the trace.
 *
 * Args:     tb  - shadow matrix of traceback pointers
 *           hmm - the model (needed for figuring out wing unfolding)
 *           L   - sequence length
 *
 * Returns:  traceback. Caller must free w/ P7FreeTrace().
 */
struct p7trace_s *
ShadowTrace (struct dpshadow_s *tb, struct plan7_s *hmm, int L)
{
    struct p7trace_s *tr;
    int curralloc;		/* current allocated length of trace */
    int tpos;			/* position in trace */
    int i;			/* position in seq (1..N) */
    int k;			/* position in model (1..M) */
    char nxtstate;		/* next state to assign in traceback */

    /* Overallocate for the trace.
     * S-N-B- ... - E-C-T  : 6 states + L is minimum trace;
     * add L more as buffer.
     */
    curralloc = L * 2 + 6;
    P7AllocTrace (curralloc, &tr);

    /* Initialization of trace
     * We do it back to front; ReverseTrace() is called later.
     */
    tr->statetype[0] = STT;
    tr->nodeidx[0] = 0;
    tr->pos[0] = 0;
    tpos = 1;
    i = L;			/* current i (seq pos) we're trying to assign   */
    k = 0;			/* current k (model pos) we're trying to assign */
    nxtstate = STC;		/* assign the C state first, for C->T */

    /* Traceback
     */
    while (nxtstate != STS) {
        switch (nxtstate) {
        case STM:
            tr->statetype[tpos] = STM;
            nxtstate = tb->mtb[i][k];
            tr->nodeidx[tpos] = k--;
            tr->pos[tpos] = i--;
            tpos++;
            break;

        case STI:
            tr->statetype[tpos] = STI;
            nxtstate = tb->itb[i][k];
            tr->nodeidx[tpos] = k;
            tr->pos[tpos] = i--;
            tpos++;
            break;

        case STD:
            tr->statetype[tpos] = STD;
            nxtstate = tb->dtb[i][k];
            tr->nodeidx[tpos] = k--;
            tr->pos[tpos] = 0;
            tpos++;
            break;

        case STN:
            tr->statetype[tpos] = STN;
            nxtstate = tb->xtb[i][XMN];
            tr->nodeidx[tpos] = 0;
            tr->pos[tpos] = (nxtstate == STN) ? i-- : 0;	/* N->N; 2nd one emits. */
            tpos++;
            break;

        case STB:
            /* Check for wing unfolding */
            if (Prob2Score (hmm->begin[k + 1], hmm->p1) + 1 * INTSCALE <=
                    hmm->bsc[k + 1])
                while (k > 0) {
                    tr->statetype[tpos] = STD;
                    tr->nodeidx[tpos] = k--;
                    tr->pos[tpos] = 0;
                    tpos++;
                    if (tpos == curralloc) {	/* grow trace if necessary  */
                        curralloc += L;
                        P7ReallocTrace (tr, curralloc);
                    }
                }

            tr->statetype[tpos] = STB;
            nxtstate = tb->xtb[i][XMB];
            tr->nodeidx[tpos] = 0;
            tr->pos[tpos] = 0;
            tpos++;
            break;

        case STJ:
            tr->statetype[tpos] = STJ;
            nxtstate = tb->xtb[i][XMJ];
            tr->nodeidx[tpos] = 0;
            tr->pos[tpos] = (nxtstate == STJ) ? i-- : 0;	/* J->J; 2nd one emits. */
            tpos++;
            break;

        case STE:
            tr->statetype[tpos] = STE;
            tr->nodeidx[tpos] = 0;
            tr->pos[tpos] = 0;
            k = tb->esrc[i];
            nxtstate = STM;
            tpos++;
            /* check for wing unfolding */
            if (Prob2Score (hmm->end[k], 1.) + 1 * INTSCALE <= hmm->esc[k]) {
                int dk;			/* need a tmp k while moving thru delete wing */
                for (dk = hmm->M; dk > k; dk--) {
                    tr->statetype[tpos] = STD;
                    tr->nodeidx[tpos] = dk;
                    tr->pos[tpos] = 0;
                    tpos++;
                    if (tpos == curralloc) {	/* grow trace if necessary  */
                        curralloc += L;
                        P7ReallocTrace (tr, curralloc);
                    }
                }
            }
            break;

        case STC:
            tr->statetype[tpos] = STC;
            nxtstate = tb->xtb[i][XMC];
            tr->nodeidx[tpos] = 0;
            tr->pos[tpos] = (nxtstate == STC) ? i-- : 0;	/* C->C; 2nd one emits. */
            tpos++;
            break;

        default:
            Die ("HMMER: Bad state (%s) in ShadowTrace()\n", Statetype (nxtstate));

        }				/* end switch over nxtstate */

        if (tpos == curralloc) {	/* grow trace if necessary  */
            curralloc += L;
            P7ReallocTrace (tr, curralloc);
        }

    }				/* end traceback, just before assigning S state */

    tr->statetype[tpos] = STS;
    tr->nodeidx[tpos] = 0;
    tr->pos[tpos] = 0;
    tr->tlen = tpos + 1;

    P7ReverseTrace (tr);
    return tr;
}



/* Function: PostprocessSignificantHit()
 * Date:     SRE, Wed Dec 20 12:11:01 2000 [StL]
 *
 * Purpose:  Add a significant hit to per-seq and per-domain hit
 *           lists, after postprocessing the scores appropriately,
 *           and making sure per-domain scores add up to the per-seq
 *           score.
 *
 *          [doesn't really belong in core_algorithms.c, because
 *           it's more of a hack than an algorithm, but on the other
 *           hand it's now part of the core of how HMMER scores
 *           things. Maybe there should be a core_hacks.c.]
 *
 *           Given: active hit lists for per-seq and per-domain
 *           scores (e.g. hmmpfam and hmmsearch, collating their
 *           results), and a new hit that's significant enough
 *           that it may need to be reported in final output.
 *
 *           Breaks the traceback into individual domain traces;
 *           scores each one of them, then applies null2 correction
 *           for biased composition. Recalculates the per-seq score
 *           as the sum of the per-domain scores. Stores the hits
 *           in the lists, for eventual sorting and output by the
 *           caller.
 *
 * Notes:    In principle we've got the score, and a pvalue, and a traceback
 *           by doing the Viterbi algorithm, right? What else is left
 *           to do? Well, in practice, life is more complicated, because
 *           of the trace-dependent null2 score correction.
 *
 *           After a null2 score correction is carried out on
 *           each domain (the default) the number of detected domains
 *           with scores > 0 may have decreased. We want the
 *           global (per-seq) hit list to have the recalculated number of
 *           domains, not necessarily what Viterbi gave us.
 *
 *           Also, since we want the global score to be the sum of
 *           the individual domains, but the null2 correction is
 *           applied to each domain individually, we have to calculate
 *           an adjusted global score. (To do otherwise invites
 *           subtle inconsistencies; xref bug 2.)
 *
 *           We don't have final evalues, so we may put a few
 *           more hits into the hit lists than we end up reporting.
 *           The main output routine is responsible for final
 *           enforcement of the thresholds.
 *
 *           This routine is NOT THREADSAFE. When multithreaded,
 *           with using shared ghit/dhit output buffers, calls to
 *           PostprocessSignificantHit() need to be protected.
 *
 * Args:     ghit     - an active list of per-seq (global) hits
 *           dhit     - an active list of per-domain hits
 *           tr       - the significant HMM/seq traceback we'll report on
 *           hmm      - ptr to the HMM
 *           dsq      - digitized sequence (1..L)
 *           L        - length of dsq
 *           seqname  - name of sequence (same as targname, in hmmsearch)
 *           seqacc   - seq's accession (or NULL)
 *           seqdesc  - seq's description (or NULL)
 *           do_forward  - TRUE if we've already calculated final per-seq score
 *           sc_override - per-seq score to use if do_forward is TRUE
 *           do_null2    - TRUE to apply the null2 scoring correction
 *           thresh      - contains the threshold/cutoff information.
 *           hmmpfam_mode - TRUE if called by hmmpfam, else assumes hmmsearch;
 *                          affects how the lists' sort keys are set.
 *
 * Returns:  the recalculated per-seq score (or sc_override),
 *           as appropriate, for subsequent storage in the histogram
 */
float
PostprocessSignificantHit (struct tophit_s *ghit,
                           struct tophit_s *dhit,
                           struct p7trace_s *tr,
                           struct plan7_s *hmm,
                           char *dsq,
                           int L,
                           char *seqname,
                           char *seqacc,
                           char *seqdesc,
                           int do_forward,
                           float sc_override,
                           int do_null2,
                           struct threshold_s *thresh, int hmmpfam_mode)
{
    struct p7trace_s **tarr;	/* array of per-domain traces */
    struct fancyali_s *ali;	/* alignment of a domain      */
    int ntr;			/* number of domain traces from Viterbi */
    int tidx;			/* index for traces (0..ntr-1) */
    int ndom;			/* # of domains accepted in sequence */
    int didx;			/* index for domains (1..ndom) */
    int k1, k2;			/* start, stop coord in model */
    int i1, i2;			/* start, stop in sequence    */
    float whole_sc;		/* whole sequence score = \sum domain scores */
    float *score;			/* array of raw scores for each domain */
    int *usedomain;		/* TRUE if this domain is accepted */
    double whole_pval;
    double pvalue;
    double sortkey;

    /* Special case: rarely, the alignment was totally impossible
     * and tr is NULL.
     */
    if (tr == NULL) {
        return sc_override;
    }

    /* Break the trace into one or more individual domains.
     */
    TraceDecompose (tr, &tarr, &ntr);
    if (ntr == 0) {
        Die ("TraceDecompose() screwup");    /* "can't happen" (!) */
    }

    /* Rescore each domain, apply null2 correction if asked.
     * Mark positive-scoring ones (we'll definitely report those),
     * and include their score in the whole sequence score.
     */
    score = MallocOrDie (sizeof (float) * ntr);
    usedomain = MallocOrDie (sizeof (int) * ntr);
    ndom = 0;
    whole_sc = 0.;
    for (tidx = 0; tidx < ntr; tidx++) {
        score[tidx] = P7TraceScore (hmm, dsq, tarr[tidx]);
        if (do_null2) {
            score[tidx] -= TraceScoreCorrection (hmm, tarr[tidx], dsq);
        }
        if (score[tidx] > 0.0) {
            usedomain[tidx] = TRUE;
            ndom++;
            whole_sc += score[tidx];
        }
        else {
            usedomain[tidx] = FALSE;
        }
    }

    /* Make sure at least one positive scoring domain is in
     * the trace. If not, invoke "weak single domain" rules:
     * we will always report at least one domain per sequence, even
     * if it has a negative score. (HMMER's Plan7 architecture can report
     * one negative scoring domain but not more.)
     */
    if (ndom == 0) {
        tidx = FArgMax (score, ntr);
        usedomain[tidx] = TRUE;
        whole_sc = score[tidx];
        ndom = 1;
    }

    /* Implement --do_forward: override the trace-dependent sum-of-domain
     * whole score, use the P7Forward() score that the called passed
     * us instead. This is a hack; null2 is trace-dependent and
     * thus undefined for P7Forward() scoring; see commentary in hmmpfam.c.
     */
    if (do_forward) {
        whole_sc = sc_override;
    }

    /* Go through and put all the accepted domains into the hit list.
     */
    whole_pval = PValue (hmm, whole_sc);
    for (tidx = 0, didx = 1; tidx < ntr; tidx++) {
        if (!usedomain[tidx]) {
            continue;
        }

        TraceSimpleBounds (tarr[tidx], &i1, &i2, &k1, &k2);
        pvalue = PValue (hmm, score[tidx]);

        if (pvalue <= thresh->domE && score[tidx] >= thresh->domT) {
            ali = CreateFancyAli (tarr[tidx], hmm, dsq, seqname);

            if (hmmpfam_mode) {
                sortkey = -1. * (double) i1;    /* hmmpfam: sort on position in seq    */
            }
            else {
                sortkey = score[tidx];    /* hmmsearch: sort on E (monotonic w/ sc) */
            }

            RegisterHit (dhit, sortkey,
                         pvalue, score[tidx],
                         whole_pval, whole_sc,
                         hmmpfam_mode ? hmm->name : seqname,
                         hmmpfam_mode ? hmm->acc : seqacc,
                         hmmpfam_mode ? hmm->desc : seqdesc,
                         i1, i2, L, k1, k2, hmm->M, didx, ndom, ali);
        }
        didx++;
    }

    /* Now register the global hit, with the domain-derived score.
     */

    /* sorting:
     * hmmpfam has to worry that score and E-value are not monotonic
     * when multiple HMMs (with different EVD parameters) are potential
     * targets. Therefore in hmmpfam_mode we apply a weird hack
     * to sort primarily on E-value, but on score
     * for really good hits with E=0.0... works because we can
     * assume 100000. > -log(DBL_MIN).
     * hmmsearch simply sorts on score (which for a single HMM, we
     * know is monotonic with E-value).
     */
    if (hmmpfam_mode)
        sortkey =
            (whole_pval > 0.0) ? -1. * log (whole_pval) : 100000. + whole_sc;
    else {
        sortkey = whole_sc;
    }

    /* Note: we've recalculated whole_sc and it may have decreased
     *       after the null2 correction was applied. For Pfam GA, TC,
     *       or NC cutoffs, we have to be sure that everything on the
     *       hitlist is correct (the hmmpfam output routine assumes it,
     *       otherwise it would have to reload each HMM to get its
     *       cutoffs). In all other cases, though, we don't care if
     *       the hit list has a bit too many things on it, because the
     *       output routine in hmmsearch or hmmpfam will check against
     *       the cutoffs. Hence we only need to check against globT
     *       (it may be set by GA, TC, or NC) but not globE.
     *                 - SRE, CSHL genome mtg May 2001
     */
    if (whole_sc >= thresh->globT) {
        RegisterHit (ghit, sortkey, whole_pval, whole_sc, 0., 0.,	/* no mother seq */
                     hmmpfam_mode ? hmm->name : seqname, hmmpfam_mode ? hmm->acc : seqacc, hmmpfam_mode ? hmm->desc : seqdesc, 0, 0, 0,	/* seq positions  */
                     0, 0, 0,	/* HMM positions  */
                     0, ndom,	/* # domains info    */
                     NULL);		/* alignment info */
    }

    /* Clean up and return.
     */
    for (tidx = 0; tidx < ntr; tidx++) {
        P7FreeTrace (tarr[tidx]);
    }
    free (tarr);
    free (score);
    free (usedomain);
    return whole_sc;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* cluster.c
 * SRE, Sun Jul 18 09:49:47 1993
 * moved to squid Thu Mar  3 08:42:57 1994
 * CVS $Id: cluster.c,v 1.4 2003/04/14 16:00:16 eddy Exp $
 *
 * almost identical to bord.c, from fd
 * also now contains routines for constructing difference matrices
 * from alignments
 *
 * "branch ordering": Input a symmetric or upper-right-diagonal
 * NxN difference matrix (usually constructed by pairwise alignment
 * and similarity calculations for N sequences). Use the simple
 * cluster analysis part of the Fitch/Margoliash tree-building algorithm
 * (as described by Fitch and Margoliash 1967 as well as Feng
 * and Doolittle 1987) to calculate the topology of an "evolutionary
 * tree" consistent with the difference matrix. Returns an array
 * which represents the tree.
 *
 * The input difference matrix is just an NxN matrix of floats.
 * A good match is a small difference score (the algorithm is going
 * to search for minima among the difference scores). The original difference
 * matrix remains unchanged by the calculations.
 *
 * The output requires some explanation. A phylogenetic
 * tree is a binary tree, with N "leaves" and N-1 "nodes". The
 * topology of the tree may be completely described by N-1 structures
 * containing two pointers; each pointer points to either a leaf
 * or another node. Here, this is implemented with integer indices
 * rather than pointers. An array of N-1 pairs of ints is returned.
 * If the index is in the range (0..N-1), it is a "leaf" -- the
 * number of one of the sequences. If the index is in the range
 * (N..2N-2), it is another "node" -- (index-N) is the index
 * of the node in the returned array.
 *
 * If both indices of a member of the returned array point to
 * nodes, the tree is "compound": composed of more than one
 * cluster of related sequences.
 *
 * The higher-numbered elements of the returned array were the
 * first constructed, and hence represent the distal tips
 * of the tree -- the most similar sequences. The root
 * is node 0.
 ******************************************************************
 *
 * Algorithm
 *
 * INITIALIZATIONS:
 *  - copy the difference matrix (otherwise the caller's copy would
 *       get destroyed by the operations of this algorithm). If
 *       it's asymmetric, make it symmetric.
 *  - make a (0..N-1) array of ints to keep track of the indices in
 *       the difference matrix as they get swapped around. Initialize
 *       this matrix to 0..N-1.
 *  - make a (0..N-2) array of int[2] to store the results (the tree
 *       topology). Doesn't need to be initialized.
 *  - keep track of a "N'", the current size of the difference
 *       matrix being operated on.
 *
 * PROCESSING THE DIFFERENCE MATRIX:
 *  - for N' = N down to N' = 2  (N-1 steps):
 *    - in the half-diagonal N'xN' matrix, find the indices i,j at which
 *      there's the minimum difference score
 *
 *     Store the results:
 *    - at position N'-2 of the result array, store coords[i] and
 *         coords[j].
 *
 *     Move i,j rows, cols to the outside edges of the matrix:
 *    - swap row i and row N'-2
 *    - swap row j and row N'-1
 *    - swap column i and column N'-2
 *    - swap column j and column N'-1
 *    - swap indices i, N'-2 in the index array
 *    - swap indices j, N'-1 in the index array
 *
 *     Build a average difference score for differences to i,j:
 *    - for all columns, find avg difference between rows i and j and store in row i:
 *       row[i][col] = (row[i][col] + row[j][col]) / 2.0
 *    - copy the contents of row i to column i (it's a symmetric
 *       matrix, no need to recalculate)
 *    - store an index N'+N-2 at position N'-2 of the index array: means
 *       that this row/column is now a node rather than a leaf, and
 *       contains minimum values
 *
 *     Continue:
 *    - go to the next N'
 *
 * GARBAGE COLLECTION & RETURN.
 *
 **********************************************************************
 *
 * References:
 *
 * Feng D-F and R.F. Doolittle. "Progressive sequence alignment as a
 *    prerequisite to correct phylogenetic trees." J. Mol. Evol.
 *    25:351-360, 1987.
 *
 * Fitch W.M. and Margoliash E. "Construction of phylogenetic trees."
 *    Science 155:279-284, 1967.
 *
 **********************************************************************
 *
 * SRE, 18 March 1992 (bord.c)
 * SRE, Sun Jul 18 09:52:14 1993 (cluster.c)
 * added to squid Thu Mar  3 09:13:56 1994
 **********************************************************************
 * Mon May  4 09:47:02 1992: keep track of difference scores at each node
 */


#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "squid.h"
#include "sqfuncs.h"

/* Function: Cluster()
 *
 * Purpose:  Cluster analysis on a distance matrix. Constructs a
 *           phylogenetic tree which contains the topology
 *           and info for each node: branch lengths, how many
 *           sequences are included under the node, and which
 *           sequences are included under the node.
 *
 * Args:     dmx     - the NxN distance matrix ( >= 0.0, larger means more diverged)
 *           N       - size of mx (number of sequences)
 *           mode    - CLUSTER_MEAN, CLUSTER_MAX, or CLUSTER_MIN
 *           ret_tree- RETURN: the tree
 *
 * Return:   1 on success, 0 on failure.
 *           The caller is responsible for freeing the tree's memory,
 *           by calling FreePhylo(tree, N).
 */
int
Cluster (float **dmx, int N, enum clust_strategy mode,
         struct phylo_s **ret_tree)
{
    struct phylo_s *tree;		/* (0..N-2) phylogenetic tree          */
    float **mx;			/* copy of difference matrix           */
    int *coord;			/* (0..N-1), indices for matrix coords */
    int i, j;			/* coords of minimum difference        */
    int idx;			/* counter over seqs                   */
    int Np;			/* N', a working copy of N             */
    int row, col;			/* loop variables                      */
    float min;			/* best minimum score found            */
    float *trow;			/* tmp pointer for swapping rows       */
    float tcol;			/* tmp storage for swapping cols       */
    float *diff;			/* (0..N-2) difference scores at nodes */
    int swapfoo;			/* for SWAP() macro                    */

    /**************************
     * Initializations.
     **************************/
    /* We destroy the matrix we work on, so make a copy of dmx.
     */
    mx = MallocOrDie (sizeof (float *) * N);
    for (i = 0; i < N; i++) {
        mx[i] = MallocOrDie (sizeof (float) * N);
        for (j = 0; j < N; j++) {
            mx[i][j] = dmx[i][j];
        }
    }
    /* coord array alloc, (0..N-1) */
    coord = MallocOrDie (N * sizeof (int));
    diff = MallocOrDie ((N - 1) * sizeof (float));
    /* init the coord array to 0..N-1 */
    for (col = 0; col < N; col++) {
        coord[col] = col;
    }
    for (i = 0; i < N - 1; i++) {
        diff[i] = 0.0;
    }

    /* tree array alloc, (0..N-2) */
    if ((tree = AllocPhylo (N)) == NULL) {
        Die ("AllocPhylo() failed");
    }

    /*********************************
     * Process the difference matrix
     *********************************/

    /* N-prime, for an NxN down to a 2x2 diffmx */
    j = 0;			/* just to silence gcc uninit warnings */
    for (Np = N; Np >= 2; Np--) {
        /* find a minimum on the N'xN' matrix */
        min = 999999.;
        for (row = 0; row < Np; row++) {
            for (col = row + 1; col < Np; col++) {
                if (mx[row][col] < min) {
                    min = mx[row][col];
                    i = row;
                    j = col;
                }
						}		
				}
        /* We're clustering row i with col j. write necessary
         * data into a node on the tree
         */
        /* topology info */
        tree[Np - 2].left = coord[i];
        tree[Np - 2].right = coord[j];
        if (coord[i] >= N) {
            tree[coord[i] - N].parent = N + Np - 2;
        }
        if (coord[j] >= N) {
            tree[coord[j] - N].parent = N + Np - 2;
        }

        /* keep score info */
        diff[Np - 2] = tree[Np - 2].diff = min;

        /* way-simple branch length estimation */
        tree[Np - 2].lblen = tree[Np - 2].rblen = min;
        if (coord[i] >= N) {
            tree[Np - 2].lblen -= diff[coord[i] - N];
        }
        if (coord[j] >= N) {
            tree[Np - 2].rblen -= diff[coord[j] - N];
        }

        /* number seqs included at node */
        if (coord[i] < N) {
            tree[Np - 2].incnum++;
            tree[Np - 2].is_in[coord[i]] = 1;
        }
        else {
            tree[Np - 2].incnum += tree[coord[i] - N].incnum;
            for (idx = 0; idx < N; idx++) {
                tree[Np - 2].is_in[idx] |= tree[coord[i] - N].is_in[idx];
            }
        }

        if (coord[j] < N) {
            tree[Np - 2].incnum++;
            tree[Np - 2].is_in[coord[j]] = 1;
        }
        else {
            tree[Np - 2].incnum += tree[coord[j] - N].incnum;
            for (idx = 0; idx < N; idx++) {
                tree[Np - 2].is_in[idx] |= tree[coord[j] - N].is_in[idx];
            }
        }


        /* Now build a new matrix, by merging row i with row j and
         * column i with column j; see Fitch and Margoliash
         */
        /* Row and column swapping. */
        /* watch out for swapping i, j away: */
        if (i == Np - 1 || j == Np - 2) {
            SWAP (i, j);
        }

        if (i != Np - 2) {
            /* swap row i, row N'-2 */
            trow = mx[Np - 2];
            mx[Np - 2] = mx[i];
            mx[i] = trow;
            /* swap col i, col N'-2 */
            for (row = 0; row < Np; row++) {
                tcol = mx[row][Np - 2];
                mx[row][Np - 2] = mx[row][i];
                mx[row][i] = tcol;
            }
            /* swap coord i, coord N'-2 */
            SWAP (coord[i], coord[Np - 2]);
        }

        if (j != Np - 1) {
            /* swap row j, row N'-1 */
            trow = mx[Np - 1];
            mx[Np - 1] = mx[j];
            mx[j] = trow;
            /* swap col j, col N'-1 */
            for (row = 0; row < Np; row++) {
                tcol = mx[row][Np - 1];
                mx[row][Np - 1] = mx[row][j];
                mx[row][j] = tcol;
            }
            /* swap coord j, coord N'-1 */
            SWAP (coord[j], coord[Np - 1]);
        }

        /* average i and j together; they're now
           at Np-2 and Np-1 though */
        i = Np - 2;
        j = Np - 1;
        /* merge by saving avg of cols of row i and row j */
        for (col = 0; col < Np; col++) {
            switch (mode) {
            case CLUSTER_MEAN:
                mx[i][col] = (mx[i][col] + mx[j][col]) / 2.0;
                break;
            case CLUSTER_MIN:
                mx[i][col] = MIN (mx[i][col], mx[j][col]);
                break;
            case CLUSTER_MAX:
                mx[i][col] = MAX (mx[i][col], mx[j][col]);
                break;
            default:
                mx[i][col] = (mx[i][col] + mx[j][col]) / 2.0;
                break;
            }
        }
        /* copy those rows to columns */
        for (col = 0; col < Np; col++) {
            mx[col][i] = mx[i][col];
        }
        /* store the node index in coords */
        coord[Np - 2] = Np + N - 2;
    }

    /**************************
     * Garbage collection and return
     **************************/
    Free2DArray ((void **) mx, N);
    free (coord);
    free (diff);
    *ret_tree = tree;
    return 1;
}

/* Function: AllocPhylo()
 *
 * Purpose:  Allocate space for a phylo_s array. N-1 structures
 *           are allocated, one for each node; in each node, a 0..N
 *           is_in flag array is also allocated and initialized to
 *           all zeros.
 *
 * Args:     N  - size; number of sequences being clustered
 *
 * Return:   pointer to the allocated array
 *
 */
struct phylo_s *
AllocPhylo (int N)
{
    struct phylo_s *tree;
    int i;

    if ((tree =
                (struct phylo_s *) malloc ((N - 1) * sizeof (struct phylo_s))) == NULL) {
        return NULL;
    }

    for (i = 0; i < N - 1; i++) {
        tree[i].diff = 0.0;
        tree[i].lblen = tree[i].rblen = 0.0;
        tree[i].left = tree[i].right = tree[i].parent = -1;
        tree[i].incnum = 0;
        if ((tree[i].is_in = (char *) calloc (N, sizeof (char))) == NULL) {
            return NULL;
        }
    }
    return tree;
}


/* Function: FreePhylo()
 *
 * Purpose:  Free a clustree array that was built to cluster N sequences.
 *
 * Args:     tree - phylogenetic tree to free
 *           N    - size of clustree; number of sequences it clustered
 *
 * Return:   (void)
 */
void
FreePhylo (struct phylo_s *tree, int N)
{
    int idx;

    for (idx = 0; idx < N - 1; idx++) {
        free (tree[idx].is_in);
    }
    free (tree);
}


/* Function: MakeDiffMx()
 *
 * Purpose:  Given a set of aligned sequences, construct
 *           an NxN fractional difference matrix. (i.e. 1.0 is
 *           completely different, 0.0 is exactly identical).
 *
 * Args:     aseqs        - flushed, aligned sequences
 *           num          - number of aseqs
 *           ret_dmx      - RETURN: difference matrix
 *
 * Return:   1 on success, 0 on failure.
 *           Caller must free diff matrix with FMX2Free(dmx)
 */
void
MakeDiffMx (char **aseqs, int num, float ***ret_dmx)
{
    float **dmx;			/* RETURN: distance matrix           */
    int i, j;			/* counters over sequences           */

    /* Allocate 2D float matrix
     */
    dmx = FMX2Alloc (num, num);

    /* Calculate distances; symmetric matrix
     * record difference, not identity (1 - identity)
     */
    for (i = 0; i < num; i++) {
        for (j = i; j < num; j++) {
            dmx[i][j] = dmx[j][i] = 1.0 - PairwiseIdentity (aseqs[i], aseqs[j]);
        }
		}
    *ret_dmx = dmx;
    return;
}

/* Function: MakeIdentityMx()
 *
 * Purpose:  Given a set of aligned sequences, construct
 *           an NxN fractional identity matrix. (i.e. 1.0 is
 *           completely identical, 0.0 is completely different).
 *           Virtually identical to MakeDiffMx(). It's
 *           less confusing to have two distinct functions, I find.
 *
 * Args:     aseqs        - flushed, aligned sequences
 *           num          - number of aseqs
 *           ret_imx      - RETURN: identity matrix (caller must free)
 *
 * Return:   1 on success, 0 on failure.
 *           Caller must free imx using FMX2Free(imx)
 */
void
MakeIdentityMx (char **aseqs, int num, float ***ret_imx)
{
    float **imx;			/* RETURN: identity matrix           */
    int i, j;			/* counters over sequences           */

    /* Allocate 2D float matrix
     */
    imx = FMX2Alloc (num, num);

    /* Calculate distances, symmetric matrix
     */
    for (i = 0; i < num; i++) {
        for (j = i; j < num; j++) {
            imx[i][j] = imx[j][i] = PairwiseIdentity (aseqs[i], aseqs[j]);
        }
		}
    *ret_imx = imx;
    return;
}



/* Function: PrintNewHampshireTree()
 *
 * Purpose:  Print out a tree in the "New Hampshire" standard
 *           format. See PHYLIP's draw.doc for a definition of
 *           the New Hampshire format.
 *
 *           Like a CFG, we generate the format string left to
 *           right by a preorder tree traversal.
 *
 * Args:     fp   - file to print to
 *           ainfo- alignment info, including sequence names
 *           tree - tree to print
 *           N    - number of leaves
 *
 */
void
PrintNewHampshireTree (FILE * fp, AINFO * ainfo, struct phylo_s *tree, int N)
{
    struct intstack_s *stack;
    int code;
    float *blen;
    int docomma;

    blen = (float *) MallocOrDie (sizeof (float) * (2 * N - 1));
    stack = InitIntStack ();
    PushIntStack (stack, N);	/* push root on stack */
    docomma = FALSE;

    /* node index code:
     *     0..N-1   = leaves; indexes of sequences.
     *     N..2N-2  = interior nodes; node-N = index of node in tree structure.
     *                code N is the root.
     *     2N..3N-2 = special flags for closing interior nodes; node-2N = index in tree
     */
    while (PopIntStack (stack, &code)) {
        if (code < N) {		/* we're a leaf. */
            /* 1) print name:branchlength */
            if (docomma) {
                fputs (",", fp);
            }
            fprintf (fp, "%s:%.5f", ainfo->sqinfo[code].name, blen[code]);
            docomma = TRUE;
        }

        else if (code < 2 * N) {	/* we're an interior node */
            /* 1) print a '(' */
            if (docomma) {
                fputs (",\n", fp);
            }
            fputs ("(", fp);
            /* 2) push on stack: ), rchild, lchild */
            PushIntStack (stack, code + N);
            PushIntStack (stack, tree[code - N].right);
            PushIntStack (stack, tree[code - N].left);
            /* 3) record branch lengths */
            blen[tree[code - N].right] = tree[code - N].rblen;
            blen[tree[code - N].left] = tree[code - N].lblen;
            docomma = FALSE;
        }

        else {			/* we're closing an interior node */

            /* print a ):branchlength */
            if (code == 2 * N) {
                fprintf (fp, ");\n");
            }
            else {
                fprintf (fp, "):%.5f", blen[code - N]);
            }
            docomma = TRUE;
        }
    }

    FreeIntStack (stack);
    free (blen);
    return;
}


/* Function: PrintPhylo()
 *
 * Purpose:  Debugging output of a phylogenetic tree structure.
 */
void
PrintPhylo (FILE * fp, AINFO * ainfo, struct phylo_s *tree, int N)
{
    int idx;

    for (idx = 0; idx < N - 1; idx++) {
        fprintf (fp, "Interior node %d (code %d)\n", idx, idx + N);
        fprintf (fp, "\tParent: %d (code %d)\n", tree[idx].parent - N,
                 tree[idx].parent);
        fprintf (fp, "\tLeft:   %d (%s) %f\n",
                 tree[idx].left < N ? tree[idx].left - N : tree[idx].left,
                 tree[idx].left <
                 N ? ainfo->sqinfo[tree[idx].left].name : "interior",
                 tree[idx].lblen);
        fprintf (fp, "\tRight:   %d (%s) %f\n",
                 tree[idx].right < N ? tree[idx].right - N : tree[idx].right,
                 tree[idx].right <
                 N ? ainfo->sqinfo[tree[idx].right].name : "interior",
                 tree[idx].rblen);
        fprintf (fp, "\tHeight:  %f\n", tree[idx].diff);
        fprintf (fp, "\tIncludes:%d seqs\n", tree[idx].incnum);
    }
}



/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* dayhoff.c
 *
 * Routines for dealing with PAM matrices.
 *
 * Includes:
 *    ParsePAMFile()  -- read a PAM matrix from disk.
 *
 *
 * SRE - Fri Apr  2 11:23:45 1993
 * CVS $Id: dayhoff.c,v 1.7 2003/05/26 16:21:50 eddy Exp $
 */


#include "squidconf.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "squid.h"

/* Function: ParsePAMFile()
 *
 * Purpose:  Given a pointer to an open file containing a PAM matrix,
 *           parse the file and allocate and fill a 2D array of
 *           floats containing the matrix. The PAM file is
 *           assumed to be in the format that NCBI distributes
 *           with BLAST. BLOSUM matrices also work fine, as
 *           produced by Henikoff's program "MATBLAS".
 *
 *           Parses both old format and new format BLAST matrices.
 *           Old format just had rows of integers.
 *           New format includes a leading character on each row.
 *
 *           The PAM matrix is a 27x27 matrix, 0=A..25=Z,26=*.
 *           Note that it's not a 20x20 matrix as you might expect;
 *           this is for speed of indexing as well as the ability
 *           to deal with ambiguous characters.
 *
 * Args:     fp        - open PAM file
 *           ret_pam   - RETURN: pam matrix, integers
 *           ret_scale - RETURN: scale factor for converting
 *                       to real Sij. For instance, PAM120 is
 *                       given in units of ln(2)/2. This may
 *                       be passed as NULL if the caller
 *                       doesn't care.
 *
 * Returns:  1 on success; 0 on failure and sets squid_errno to
 *           indicate the cause. ret_pam is allocated here and
 *           must be freed by the caller (use FreePAM).
 */
int
ParsePAMFile (FILE * fp, int ***ret_pam, float *ret_scale)
{
    int **pam;
    char buffer[512];		/* input buffer from fp                  */
    int order[27];		/* order of fields, obtained from header */
    int nsymbols;			/* total number of symbols in matrix     */
    char *sptr;
    int idx;
    int row, col;
    float scale;
    int gotscale = FALSE;

    scale = 0.0;			/* just to silence gcc uninit warnings */
    if (fp == NULL) {
        squid_errno = SQERR_NODATA;
        return 0;
    }

    /* Look at the first non-blank, non-comment line in the file.
     * It gives single-letter codes in the order the PAM matrix
     * is arrayed in the file.
     */
    do {
        if (fgets (buffer, 512, fp) == NULL) {
            squid_errno = SQERR_NODATA;
            return 0;
        }

        /* Get the scale factor from the header.
         * For BLOSUM files, we assume the line looks like:
         *     BLOSUM Clustered Scoring Matrix in 1/2 Bit Units
         * and we assume that the fraction is always 1/x;
         *
         * For PAM files, we assume the line looks like:
         *     PAM 120 substitution matrix, scale = ln(2)/2 = 0.346574
         * and we assume that the number following the final '=' is our scale
         */
        if (strstr (buffer, "BLOSUM Clustered Scoring Matrix") != NULL &&
                (sptr = strchr (buffer, '/')) != NULL) {
            sptr++;
            if (!isdigit ((int) (*sptr))) {
                squid_errno = SQERR_FORMAT;
                return 0;
            }
            scale = (float) ((log (2.0)) / (atof (sptr)));
            gotscale = TRUE;
        }
        else if (strstr (buffer, "substitution matrix,") != NULL) {
            while ((sptr = strrchr (buffer, '=')) != NULL) {
                sptr += 2;
                if (IsReal (sptr)) {
                    scale = atof (sptr);
                    gotscale = TRUE;
                    break;
                }
            }
        }
    } while ((sptr = strtok (buffer, " \t\n")) == NULL || *sptr == '#');

    idx = 0;
    do {
        order[idx] = (int) *sptr - (int) 'A';
        if (order[idx] < 0 || order[idx] > 25) {
            order[idx] = 26;
        }
        idx++;
    } while ((sptr = strtok (NULL, " \t\n")) != NULL);
    nsymbols = idx;

    /* Allocate a pam matrix. For speed of indexing, we use
     * a 27x27 matrix so we can do lookups using the ASCII codes
     * of amino acid single-letter representations, plus one
     * extra field to deal with the "*" (terminators).
     */
    if ((pam = (int **) calloc (27, sizeof (int *))) == NULL) {
        Die ("calloc failed");
    }
    for (idx = 0; idx < 27; idx++) {
        if ((pam[idx] = (int *) calloc (27, sizeof (int))) == NULL) {
            Die ("calloc failed");
        }
		}
    /* Parse the rest of the file.
     */
    for (row = 0; row < nsymbols; row++) {
        if (fgets (buffer, 512, fp) == NULL) {
            squid_errno = SQERR_NODATA;
            return 0;
        }

        if ((sptr = strtok (buffer, " \t\n")) == NULL) {
            squid_errno = SQERR_NODATA;
            return 0;
        }
        for (col = 0; col < nsymbols; col++) {
            if (sptr == NULL) {
                squid_errno = SQERR_NODATA;
                return 0;
            }

            /* Watch out for new BLAST format, with leading characters
             */
            if (*sptr == '*' || isalpha ((int) *sptr)) {
                col--;    /* hack hack */
            }
            else {
                pam[order[row]][order[col]] = atoi (sptr);
            }

            sptr = strtok (NULL, " \t\n");
        }
    }

    /* Return
     */
    if (ret_scale != NULL) {
        if (gotscale) {
            *ret_scale = scale;
        }
        else {
            Warn
            ("Failed to parse PAM matrix scale factor. Defaulting to ln(2)/2!");
            *ret_scale = log (2.0) / 2.0;
        }
    }
    *ret_pam = pam;
    return 1;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* debug.c
 * Thu Nov 21 09:58:05 1996
 *
 * Printing out or naming various useful things from HMMER
 * innards.
 *
 * CVS $Id: debug.c,v 1.13 2003/04/14 16:00:16 eddy Exp $
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#include "structs.h"
#include "funcs.h"
#include "squid.h"

/* Function: Statetype()
 *
 * Purpose:  Returns the state type in text.
 * Example:  Statetype(S) = "S"
 */
char *
Statetype (char st)
{
    switch (st) {
    case STS:
        return "S";
    case STN:
        return "N";
    case STB:
        return "B";
    case STM:
        return "M";
    case STD:
        return "D";
    case STI:
        return "I";
    case STE:
        return "E";
    case STJ:
        return "J";
    case STC:
        return "C";
    case STT:
        return "T";
    default:
        return "BOGUS";
    }
}

/* Function: AlphabetType2String()
 * Date:     SRE, Sun Dec 24 11:33:40 2000 [St. Louis]
 *
 * Purpose:  Returns a string "protein" for hmmAMINO,
 *           "nucleic acid" for hmmNUCLEIC, etc... used
 *           for formatting diagnostics.
 *
 * Args:     type - Alphabet type, e.g. hmmAMINO
 *
 * Returns:  char *
 */
char *
AlphabetType2String (int type)
{
    switch (type) {
    case hmmAMINO:
        return "protein";
    case hmmNUCLEIC:
        return "nucleic acid";
    case hmmNOTSETYET:
        return "unknown";
    default:
        return "BOGUS";
    }
}


/* Function: P7PrintTrace()
 *
 * Purpose:  Print out a traceback structure.
 *           If hmm is non-NULL, also print transition and emission scores.
 *
 * Args:     fp  - stderr or stdout, often
 *           tr  - trace structure to print
 *           hmm - NULL or hmm containing scores to print
 *           dsq - NULL or digitized sequence trace refers to.
 */
void
P7PrintTrace (FILE * fp, struct p7trace_s *tr, struct plan7_s *hmm, char *dsq)
{
    int tpos;			/* counter for trace position */
    int sym;
    int sc;

    if (tr == NULL) {
        fprintf (fp, " [ trace is NULL ]\n");
        return;
    }

    if (hmm == NULL) {
        fprintf (fp, "st  node   rpos  - traceback len %d\n", tr->tlen);
        fprintf (fp, "--  ---- ------\n");
        for (tpos = 0; tpos < tr->tlen; tpos++) {
            fprintf (fp, "%1s  %4d %6d\n",
                     Statetype (tr->statetype[tpos]),
                     tr->nodeidx[tpos], tr->pos[tpos]);
        }
    }
    else {
        if (!(hmm->flags & PLAN7_HASBITS)) {
            Die ("oi, you can't print scores from that hmm, it's not ready.");
        }

        sc = 0;
        fprintf (fp, "st  node   rpos  transit emission - traceback len %d\n",
                 tr->tlen);
        fprintf (fp, "--  ---- ------  ------- --------\n");
        for (tpos = 0; tpos < tr->tlen; tpos++) {
            if (dsq != NULL) {
                sym = (int) dsq[tr->pos[tpos]];
            }

            fprintf (fp, "%1s  %4d %6d  %7d",
                     Statetype (tr->statetype[tpos]),
                     tr->nodeidx[tpos],
                     tr->pos[tpos],
                     (tpos < tr->tlen - 1) ?
                     TransitionScoreLookup (hmm, tr->statetype[tpos],
                                            tr->nodeidx[tpos],
                                            tr->statetype[tpos + 1],
                                            tr->nodeidx[tpos + 1]) : 0);

            if (tpos < tr->tlen - 1)
                sc +=
                    TransitionScoreLookup (hmm, tr->statetype[tpos], tr->nodeidx[tpos],
                                           tr->statetype[tpos + 1],
                                           tr->nodeidx[tpos + 1]);

            if (dsq != NULL) {
                if (tr->statetype[tpos] == STM) {
                    fprintf (fp, " %8d %c", hmm->msc[sym][tr->nodeidx[tpos]],
                             Alphabet[sym]);
                    sc += hmm->msc[sym][tr->nodeidx[tpos]];
                }
                else if (tr->statetype[tpos] == STI) {
                    fprintf (fp, " %8d %c", hmm->isc[sym][tr->nodeidx[tpos]],
                             (char) tolower ((int) Alphabet[sym]));
                    sc += hmm->isc[sym][tr->nodeidx[tpos]];
                }
                else if ((tr->statetype[tpos] == STN && tr->statetype[tpos - 1] == STN)
                         || (tr->statetype[tpos] == STC
                             && tr->statetype[tpos - 1] == STC)
                         || (tr->statetype[tpos] == STJ
                             && tr->statetype[tpos - 1] == STJ)) {
                    fprintf (fp, " %8d %c", 0, (char) tolower ((int) Alphabet[sym]));
                }
            }
            else {
                fprintf (fp, " %8s %c", "-", '-');
            }


            fputs ("\n", fp);
        }
        fprintf (fp, "                 ------- --------\n");
        fprintf (fp, "           total: %6d\n\n", sc);
    }
}

/* Function: P7PrintPrior()
 *
 * Purpose:  Print out a Plan 7 prior structure.
 */
void
P7PrintPrior (FILE * fp, struct p7prior_s *pri)
{
    int q, x;			/* counters for mixture component, element */

    if (pri->strategy == PRI_DCHLET) {
        fputs ("Dirichlet\n", fp);
    }
    else if (pri->strategy == PRI_PAM) {
        fputs ("PAM\n", fp);
    }
    else {
        Die ("No such strategy.");
    }

    if (Alphabet_type == hmmAMINO) {
        fputs ("Amino\n", fp);
    }
    else if (Alphabet_type == hmmNUCLEIC) {
        fputs ("Nucleic\n", fp);
    }

    /* Transitions
     */
    fprintf (fp, "\n%d\n", pri->tnum);
    for (q = 0; q < pri->tnum; q++) {
        fprintf (fp, "%.4f\n", pri->tq[q]);
        for (x = 0; x < 7; x++) {
            fprintf (fp, "%.4f ", pri->t[q][x]);
        }
        fputs ("\n", fp);
    }

    /* Match emissions
     */
    fprintf (fp, "\n%d\n", pri->mnum);
    for (q = 0; q < pri->mnum; q++) {
        fprintf (fp, "%.4f\n", pri->mq[q]);
        for (x = 0; x < Alphabet_size; x++) {
            fprintf (fp, "%.4f ", pri->m[q][x]);
        }
        fputs ("\n", fp);
    }

    /* Insert emissions
     */
    fprintf (fp, "\n%d\n", pri->inum);
    for (q = 0; q < pri->inum; q++) {
        fprintf (fp, "%.4f\n", pri->iq[q]);
        for (x = 0; x < Alphabet_size; x++) {
            fprintf (fp, "%.4f ", pri->i[q][x]);
        }
        fputs ("\n", fp);
    }
}

/* Function: TraceVerify()
 * Date:     SRE, Mon Feb  2 07:48:52 1998 [St. Louis]
 *
 * Purpose:  Check a traceback structure for internal consistency.
 *           Used in Shiva testsuite, for example.
 *
 * Args:     tr  - traceback to verify
 *           M   - length of HMM
 *           N   - length of sequence
 *
 * Returns:  1 if OK. 0 if not.
 */
int
TraceVerify (struct p7trace_s *tr, int M, int N)
{
    int tpos;			/* position in trace                  */
    int k;			/* current position in HMM nodes 1..M */
    int i;			/* current position in seq 1..N       */
    int nn, nc, nj;		/* number of STN's, STC's, STJ's seen */
    int nm;			/* number of STM's seen */

    /* Basic checks on ends.
     */
    if (tr->statetype[0] != STS) {
        return 0;
    }
    if (tr->statetype[1] != STN) {
        return 0;
    }
    if (tr->statetype[tr->tlen - 2] != STC) {
        return 0;
    }
    if (tr->statetype[tr->tlen - 1] != STT) {
        return 0;
    }
    if (tr->pos[1] != 0) {
        return 0;
    }

    /* Check for consistency throughout trace
     */
    k = i = nn = nc = nj = nm = 0;
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        switch (tr->statetype[tpos]) {
        case STS:
            if (tr->nodeidx[tpos] != 0) {
                return 0;
            }
            if (tr->pos[tpos] != 0) {
                return 0;
            }
            if (k != 0) {
                return 0;
            }
            if (i != 0) {
                return 0;
            }
            if (tpos != 0) {
                return 0;
            }
            break;

        case STN:			/* first N doesn't emit. */
            if (tr->nodeidx[tpos] != 0) {
                return 0;
            }
            if (k != 0) {
                return 0;
            }
            if (nn > 0) {
                if (tr->pos[tpos] != i + 1) {
                    return 0;
                }
                i++;
            }
            else {
                if (tr->pos[tpos] != 0) {
                    return 0;
                }
                if (i != 0) {
                    return 0;
                }
            }
            nn++;
            break;

        case STB:
            if (tr->nodeidx[tpos] != 0) {
                return 0;
            }
            if (tr->pos[tpos] != 0) {
                return 0;
            }
            nm = 0;
            break;

        case STM:			/* can enter anywhere on first M */
            if (tr->pos[tpos] != i + 1) {
                return 0;
            }
            if (tr->nodeidx[tpos] < 1 || tr->nodeidx[tpos] > M) {
                return 0;
            }
            i++;
            if (nm == 0) {
                k = tr->nodeidx[tpos];
            }
            else {
                if (tr->nodeidx[tpos] != k + 1) {
                    return 0;
                }
                k++;
            }
            nm++;
            break;

        case STI:
            if (tr->pos[tpos] != i + 1) {
                return 0;
            }
            if (tr->nodeidx[tpos] != k) {
                return 0;
            }
            if (tr->nodeidx[tpos] < 1 || tr->nodeidx[tpos] > M - 1) {
                return 0;
            }
            if (k >= M) {
                return 0;
            }
            i++;
            break;

        case STD:
            if (tr->pos[tpos] != 0) {
                return 0;
            }
            if (tr->nodeidx[tpos] != k + 1) {
                return 0;
            }
            if (tr->nodeidx[tpos] < 1 || tr->nodeidx[tpos] > M) {
                return 0;
            }
            k++;
            break;

        case STE:
            if (tr->nodeidx[tpos] != 0) {
                return 0;
            }
            if (tr->pos[tpos] != 0) {
                return 0;
            }
            nj = 0;
            break;

        case STJ:
            if (tr->nodeidx[tpos] != 0) {
                return 0;
            }
            if (nj > 0) {
                if (tr->pos[tpos] != i + 1) {
                    return 0;
                }
                i++;
            }
            else if (tr->pos[tpos] != 0) {
                return 0;
            }
            nj++;
            break;

        case STC:
            if (tr->nodeidx[tpos] != 0) {
                return 0;
            }
            if (nc > 0) {
                if (tr->pos[tpos] != i + 1) {
                    return 0;
                }
                i++;
            }
            else if (tr->pos[tpos] != 0) {
                return 0;
            }
            nc++;
            break;

        case STT:
            if (tpos != tr->tlen - 1) {
                return 0;
            }
            if (tr->nodeidx[tpos] != 0) {
                return 0;
            }
            if (tr->pos[tpos] != 0) {
                return 0;
            }
            if (i != N) {
                return 0;
            }
            break;

        case STBOGUS:
        default:
            return 0;
        }				/* end switch over statetypes */
    }				/* end loop over trace positions */

    return 1;
}


/* Function: TraceCompare()
 * Date:     SRE, Wed Mar  4 17:26:49 1998 [St. Louis]
 *
 * Purpose:  Compare two tracebacks; return 1 if they're
 *           identical, else 0. Written for Shiva testsuite.
 *
 * Args:     t1 - first trace
 *           t2 - second trace
 *
 * Returns:  1 if identical; 0 elsewise
 */
int
TraceCompare (struct p7trace_s *t1, struct p7trace_s *t2)
{
    int tpos;

    if (t1->tlen != t2->tlen) {
        return 0;
    }

    for (tpos = 0; tpos < t1->tlen; tpos++) {
        if (t1->statetype[tpos] != t2->statetype[tpos]) {
            return 0;
        }
        if (t1->nodeidx[tpos] != t2->nodeidx[tpos]) {
            return 0;
        }
        if (t1->pos[tpos] != t2->pos[tpos]) {
            return 0;
        }
    }
    return 1;
}

/************************************************************
 * Copyright (C) 1998 Ian Holmes
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* display.c
 * Author: Ian Holmes (ihh@sanger.ac.uk, Jun 5 1998)
 * Derived from core_algorithms.c (SRE, Nov 11 1996)
 * Incorporated SRE, Sat Nov  6 10:09:41 1999
 *
 * Functions for displaying HMMer2.0 structures.
 *
 * CVS $Id: display.c,v 1.4 2003/04/14 16:00:16 eddy Exp $
 */

#include "config.h"
#include "squidconf.h"

#include "structs.h"
#include "funcs.h"
#include "squid.h"

void PrintIscore (int sc);

void PrintTransition (char src,
                      int isrc,
                      int ksrc,
                      char dest,
                      int idest,
                      int kdest,
                      int sc,
                      struct p7trace_s **alignment,
                      int *min, int *max, int *on, int A);


/* Function: DisplayPlan7Posteriors()
 *
 * Purpose:  Print out posterior transition probabilities
 *           in modelpost format.
 *           NB only prints out transitions that touch
 *           either the Viterbi or the optimal accuracy path.
 *
 * Args:     L        - the length of the sequence
 *           hmm      - the model
 *           forward  - forward matrix
 *           backward - backward matrix
 *           viterbi  - Viterbi trace
 *           optacc   - optimal accuracy trace
 *
 * Return:   void
 *
 */
void
DisplayPlan7Posteriors (int L, struct plan7_s *hmm,
                        struct dpmatrix_s *forward,
                        struct dpmatrix_s *backward,
                        struct p7trace_s *viterbi, struct p7trace_s *optacc)
{
    struct p7trace_s *alignment[2];
    alignment[0] = viterbi;
    alignment[1] = optacc;
    DisplayPlan7PostAlign (L, hmm, forward, backward, alignment, 2);
}


/* Function: DisplayPlan7PostAlign()
 *
 * Purpose:  Print out posterior transition probabilities
 *           in modelpost format, for any set of alignments.
 *
 * Args:     L         - the length of the sequence
 *           hmm       - the model
 *           forward   - forward matrix
 *           backward  - backward matrix
 *           alignment - array of traces
 *           A         - size of alignment array
 *
 * Return:   void
 *
 */
void
DisplayPlan7PostAlign (int L, struct plan7_s *hmm,
                       struct dpmatrix_s *forward,
                       struct dpmatrix_s *backward,
                       struct p7trace_s **alignment, int A)
{
    int sc;
    int i;
    int j;
    int k;
    int kmin;
    int kmax;
    int *min;
    int *max;
    int *on;
    char state;

    sc = forward->xmx[L][XMC] + hmm->xsc[XTC][MOVE];	/* total Forward score */

    min = (int *) calloc (A, sizeof (int));
    max = (int *) calloc (A, sizeof (int));
    on = (int *) calloc (A, sizeof (int));

    for (i = 0; i <= L; i++) {
        for (j = 0; j < A; j++) {
            while (alignment[j]->pos[min[j]] < i - 1
                    && min[j] < alignment[j]->tlen - 1) {
                min[j]++;
            }

            while (alignment[j]->pos[max[j]] <= i + 1
                    && max[j] < alignment[j]->tlen - 1) {
                max[j]++;
            }
        }

        for (state = STM; state <= STJ; state++) {
            if (state == STM || state == STB) {
                kmin = 1;
                kmax = hmm->M;
            }
            else if (state == STD) {
                kmin = 2;
                kmax = hmm->M - 1;
            }
            else if (state == STI) {
                kmin = 1;
                kmax = hmm->M - 1;
            }
            else {
                kmin = kmax = 0;
            }

            for (k = kmin; k <= kmax; k++) {
                switch (state) {
                case STM:
                    if (i < L && k < hmm->M)
                        PrintTransition (STM, i, k, STM, i + 1, k + 1,
                                         forward->mmx[i][k] + hmm->tsc[TMM][k] +
                                         backward->mmx[i + 1][k + 1] - sc, alignment, min,
                                         max, on, A);

                    if (i < L && k < hmm->M)
                        PrintTransition (STM, i, k, STI, i + 1, k,
                                         forward->mmx[i][k] + hmm->tsc[TMI][k] +
                                         backward->imx[i + 1][k] - sc, alignment, min,
                                         max, on, A);

                    if (k < hmm->M - 1)
                        PrintTransition (STM, i, k, STD, i, k + 1,
                                         forward->mmx[i][k] + hmm->tsc[TMD][k] +
                                         backward->dmx[i][k + 1] - sc, alignment, min,
                                         max, on, A);

                    PrintTransition (STM, i, k, STE, i, 0,
                                     forward->mmx[i][k] + hmm->esc[k] +
                                     backward->xmx[i][XME] - sc, alignment, min, max,
                                     on, A);
                    break;

                case STD:
                    if (i < L)
                        PrintTransition (STD, i, k, STM, i + 1, k + 1,
                                         forward->dmx[i][k] + hmm->tsc[TDM][k] +
                                         backward->mmx[i + 1][k + 1] - sc, alignment, min,
                                         max, on, A);

                    PrintTransition (STD, i, k, STD, i, k + 1,
                                     forward->dmx[i][k] + hmm->tsc[TDD][k] +
                                     backward->dmx[i][k + 1] - sc, alignment, min, max,
                                     on, A);

                    break;

                case STI:
                    if (i < L)
                        PrintTransition (STI, i, k, STM, i + 1, k + 1,
                                         forward->imx[i][k] + hmm->tsc[TIM][k] +
                                         backward->mmx[i + 1][k + 1] - sc, alignment, min,
                                         max, on, A);

                    if (i < L)
                        PrintTransition (STI, i, k, STI, i + 1, k,
                                         forward->imx[i][k] + hmm->tsc[TII][k] +
                                         backward->imx[i + 1][k] - sc, alignment, min,
                                         max, on, A);

                    break;

                case STB:
                    if (i < L)
                        PrintTransition (STB, i, 0, STM, i + 1, k,
                                         forward->xmx[i][XMB] + hmm->bsc[k] +
                                         backward->mmx[i + 1][k] - sc, alignment, min,
                                         max, on, A);
                    break;

                default:
                    break;

                }
            }

            switch (state) {
            case STN:
                PrintTransition (STN, i, 0, STB, i, 0,
                                 forward->xmx[i][XMN] + hmm->xsc[XTN][MOVE] +
                                 backward->xmx[i][XMB] - sc, alignment, min, max, on,
                                 A);

                if (i < L)
                    PrintTransition (STN, i, 0, STN, i + 1, 0,
                                     forward->xmx[i][XMN] + hmm->xsc[XTN][LOOP] +
                                     backward->xmx[i + 1][XMN] - sc, alignment, min,
                                     max, on, A);
                break;

            case STJ:
                PrintTransition (STJ, i, 0, STB, i, 0,
                                 forward->xmx[i][XMJ] + hmm->xsc[XTJ][MOVE] +
                                 backward->xmx[i][XMB] - sc, alignment, min, max, on,
                                 A);

                if (i < L)
                    PrintTransition (STJ, i, 0, STJ, i + 1, 0,
                                     forward->xmx[i][XMJ] + hmm->xsc[XTJ][LOOP] +
                                     backward->xmx[i + 1][XMJ] - sc, alignment, min,
                                     max, on, A);
                break;

            case STC:
                PrintTransition (STC, i, 0, STT, i, 0, forward->xmx[i][XMC] + hmm->xsc[XTC][MOVE] - sc,	/* should be 1 */
                                 alignment, min, max, on, A);

                if (i < L)
                    PrintTransition (STC, i, 0, STC, i + 1, 0,
                                     forward->xmx[i][XMC] + hmm->xsc[XTC][LOOP] +
                                     backward->xmx[i + 1][XMC] - sc, alignment, min,
                                     max, on, A);
                break;

            case STE:
                PrintTransition (STE, i, 0, STC, i, 0,
                                 forward->xmx[i][XME] + hmm->xsc[XTE][MOVE] +
                                 backward->xmx[i][XMC] - sc, alignment, min, max, on,
                                 A);

                PrintTransition (STE, i, 0, STJ, i, 0,
                                 forward->xmx[i][XME] + hmm->xsc[XTE][LOOP] +
                                 backward->xmx[i][XMJ] - sc, alignment, min, max, on,
                                 A);
                break;

            case STS:
                if (i == 0)
                    PrintTransition (STS, i, 0, STN, i, 0, backward->xmx[i][XMN] - sc,	/* should be 1 */
                                     alignment, min, max, on, A);
                break;

            case STM:
            case STD:
            case STI:
            case STB:
            case STT:
                break;

            default:
                Die ("unknown state");

            }
        }
    }

    free (min);
    free (max);
    free (on);

}



/* Function: DisplayPlan7Matrix()
 *
 * Purpose:  Print out a dynamic programming matrix.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           mx     - dp matrix
 *
 * Return:   void
 *
 * The output of this function inverts HMMer's concept of rows and columns
 * (i.e. each row represents a state, and each column, a residue);
 * also, probabilities are displayed as natural logs, not bit scores.
 * It should probably only be used by ihh...
 *
 */
void
DisplayPlan7Matrix (char *dsq, int L, struct plan7_s *hmm,
                    struct dpmatrix_s *mx)
{
    int i;
    int k;

    printf ("         *      ");
    for (i = 1; i <= L; i++) {
        printf ("    %c      ", Alphabet[(int) dsq[i]]);
    }
    printf ("\nN    ");
    for (i = 0; i <= L; i++) {
        PrintIscore (mx->xmx[i][XMN]);
    }
    for (k = 1; k <= hmm->M; k++) {
        printf ("\nM%-3d ", k);
        for (i = 0; i <= L; i++) {
            PrintIscore (mx->mmx[i][k]);
        }
    }
    for (k = 1; k < hmm->M; k++) {
        printf ("\nI%-3d ", k);
        for (i = 0; i <= L; i++) {
            PrintIscore (mx->imx[i][k]);
        }
    }
    printf ("\nE    ");
    for (i = 0; i <= L; i++) {
        PrintIscore (mx->xmx[i][XME]);
    }
    printf ("\nC    ");
    for (i = 0; i <= L; i++) {
        PrintIscore (mx->xmx[i][XMC]);
    }
    printf ("\nJ    ");
    for (i = 0; i <= L; i++) {
        PrintIscore (mx->xmx[i][XMJ]);
    }
    printf ("\nB    ");
    for (i = 0; i <= L; i++) {
        PrintIscore (mx->xmx[i][XMB]);
    }
    for (k = 2; k < hmm->M; k++) {
        printf ("\nD%-3d ", k);
        for (i = 0; i <= L; i++) {
            PrintIscore (mx->dmx[i][k]);
        }
    }
    printf ("\n\n");
}


void
PrintIscore (int sc)
{
    double dsc;
    double div;
    dsc = (double) sc;
    div = INTSCALE / 0.693147180559945;	/* == INTSCALE / log(2) */
    dsc = dsc / div;
    printf ("%- #11.3e", dsc);
}


void
PrintTransition (char src,
                 int isrc,
                 int ksrc,
                 char dest,
                 int idest,
                 int kdest,
                 int sc,
                 struct p7trace_s **alignment,
                 int *min, int *max, int *on, int A)
{
    char src_str[6];		/* buffer for source state label        */
    char dest_str[6];		/* buffer for destination state label   */
    int j;
    int tpos;
    int tnext;
    int pos;
    int next;
    int near;

    near = 0;

    for (j = 0; j < A; j++) {
        on[j] = 0;
        for (pos = 0, tpos = min[j]; tpos <= max[j]; tpos++) {

            if (alignment[j]->pos[tpos] != 0) {
                pos = alignment[j]->pos[tpos];
            }

            if (src == alignment[j]->statetype[tpos]
                    && ksrc == alignment[j]->nodeidx[tpos]
                    && isrc == pos) {
                near = TRUE;
            }

            if (dest == alignment[j]->statetype[tpos]
                    && kdest == alignment[j]->nodeidx[tpos]
                    && idest == pos) {
                near = TRUE;
            }

            if (tpos < alignment[j]->tlen - 1) {
                tnext = tpos + 1;

                /* fold up B->D->M transitions into pseudo- B->M transitions */

                if (alignment[j]->statetype[tpos] == STB)
                    while (alignment[j]->statetype[tnext] == STD
                            && tnext < alignment[j]->tlen - 1) {
                        tnext++;
                    }

                next = alignment[j]->pos[tnext];
                if (next == 0) {
                    next = pos;
                }

                if (src == alignment[j]->statetype[tpos]
                        && ksrc == alignment[j]->nodeidx[tpos]
                        && isrc == pos && dest == alignment[j]->statetype[tnext]
                        && kdest == alignment[j]->nodeidx[tnext]
                        && idest == next) {
                    on[j] = TRUE;
                }
            }
        }
    }

    if (!near) {
        return;
    }

    switch (src) {
    case STM:
        sprintf (src_str, "M%d", ksrc);
        break;
    case STD:
        sprintf (src_str, "D%d", ksrc);
        break;
    case STI:
        sprintf (src_str, "I%d", ksrc);
        break;
    case STS:
        sprintf (src_str, "S");
        break;
    case STN:
        sprintf (src_str, "N");
        break;
    case STB:
        sprintf (src_str, "B");
        break;
    case STE:
        sprintf (src_str, "E");
        break;
    case STC:
        sprintf (src_str, "C");
        break;
    case STJ:
        sprintf (src_str, "J");
        break;
    case STT:
        sprintf (src_str, "T");
        break;
    default:
        Die ("bad transition");
    }

    switch (dest) {
    case STM:
        sprintf (dest_str, "M%d", kdest);
        break;
    case STD:
        sprintf (dest_str, "D%d", kdest);
        break;
    case STI:
        sprintf (dest_str, "I%d", kdest);
        break;
    case STS:
        sprintf (dest_str, "S");
        break;
    case STN:
        sprintf (dest_str, "N");
        break;
    case STB:
        sprintf (dest_str, "B");
        break;
    case STE:
        sprintf (dest_str, "E");
        break;
    case STC:
        sprintf (dest_str, "C");
        break;
    case STJ:
        sprintf (dest_str, "J");
        break;
    case STT:
        sprintf (dest_str, "T");
        break;
    default:
        Die ("bad transition");
    }

    printf ("%d\t%s\t%d\t%s\t%-14.7g\t", isrc, src_str, idest, dest_str,
            (double) Score2Prob (sc, 1.));

    for (j = 0; j < A; j++) {
        if (on[j]) {
            printf ("*");
        }
        if (j < A - 1) {
            printf ("\t");
        }
    }

    printf ("\n");

}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* emit.c
 * SRE, Sun Mar  8 12:26:58 1998
 * CVS $Id: emit.c,v 1.7 2003/04/14 16:00:16 eddy Exp $
 *
 * Generation of sequences/traces from an HMM.
 */

#include "config.h"
#include "squidconf.h"

#include <ctype.h>

#include "structs.h"
#include "funcs.h"
#include "squid.h"



/* Function: EmitSequence()
 * Date:     SRE, Sun Mar  8 12:28:03 1998 [St. Louis]
 *
 * Purpose:  Given a model, sample a sequence and/or traceback.
 *
 * Args:     hmm     - the model
 *           ret_dsq - RETURN: generated digitized sequence (pass NULL if unwanted)
 *           ret_L   - RETURN: length of generated sequence
 *           ret_tr  - RETURN: generated trace (pass NULL if unwanted)
 *
 * Returns:  void
 */
void
EmitSequence (struct plan7_s *hmm, char **ret_dsq, int *ret_L,
              struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;
    char type;			/* current state type */
    int k;			/* current node index */
    char *dsq;			/* generated sequence, digitized */
    int L;			/* length of sequence */
    int alloc_tlen;		/* allocated space for traceback */
    int alloc_L;			/* allocated space for sequence  */
    int tpos;			/* position in traceback */
    int sym;			/* a generated symbol index */
    float t[4];			/* little array for choosing M transition from */

    /* Initialize; allocations
     */
    P7AllocTrace (64, &tr);
    alloc_tlen = 64;
    dsq = MallocOrDie (sizeof (char) * 64);
    alloc_L = 64;

    TraceSet (tr, 0, STS, 0, 0);
    TraceSet (tr, 1, STN, 0, 0);
    dsq[0] = (char) Alphabet_iupac;
    L = 1;
    k = 0;
    type = STN;
    tpos = 2;

    while (type != STT) {
        /* Deal with state transition
         */
        switch (type) {
        case STB:
            hmm->begin[0] = hmm->tbd1;	/* begin[0] hack (documented in structs.h) */
            k = FChoose (hmm->begin, hmm->M + 1);
            if (k == 0) {
                type = STD;
                k = 1;
            }
            else {
                type = STM;
            }
            break;

        case STI:
            type = (FChoose (hmm->t[k] + TIM, 2) == 0) ? STM : STI;
            if (type == STM) {
                k++;
            }
            break;
        case STN:
            type = (FChoose (hmm->xt[XTN], 2) == LOOP) ? STN : STB;
            k = 0;
            break;
        case STE:
            type = (FChoose (hmm->xt[XTE], 2) == LOOP) ? STJ : STC;
            k = 0;
            break;
        case STC:
            type = (FChoose (hmm->xt[XTC], 2) == LOOP) ? STC : STT;
            k = 0;
            break;
        case STJ:
            type = (FChoose (hmm->xt[XTJ], 2) == LOOP) ? STJ : STB;
            k = 0;
            break;

        case STD:
            if (k < hmm->M) {
                type = (FChoose (hmm->t[k] + TDM, 2) == 0) ? STM : STD;
                k++;
            }
            else {
                type = STE;
                k = 0;
            }
            break;

        case STM:
            if (k < hmm->M) {
                FCopy (t, hmm->t[k], 3);
                t[3] = hmm->end[k];
                switch (FChoose (t, 4)) {
                case 0:
                    k++;
                    type = STM;
                    break;
                case 1:
                    type = STI;
                    break;
                case 2:
                    k++;
                    type = STD;
                    break;
                case 3:
                    k = 0;
                    type = STE;
                    break;
                default:
                    Die ("never happens");
                }
            }
            else {
                k = 0;
                type = STE;
            }
            break;

        case STT:
        case STBOGUS:
        default:
            Die ("can't happen.");
        }

        /* Choose a symbol emission, if necessary
         */
        sym = -1;
        if (type == STM) {
            sym = FChoose (hmm->mat[k], Alphabet_size);
        }
        else if (type == STI) {
            sym = FChoose (hmm->ins[k], Alphabet_size);
        }
        else if ((type == STN && tr->statetype[tpos - 1] == STN) ||
                 (type == STC && tr->statetype[tpos - 1] == STC) ||
                 (type == STJ && tr->statetype[tpos - 1] == STJ)) {
            sym = FChoose (hmm->null, Alphabet_size);
        }

        /* Add to the traceback; deal with realloc if necessary
         */
        TraceSet (tr, tpos, type, k, (sym != -1) ? L : 0);
        tpos++;
        if (tpos == alloc_tlen) {
            alloc_tlen += 64;
            P7ReallocTrace (tr, alloc_tlen);
        }

        /* Add to the digitized seq; deal with realloc, if necessary
         */
        if (sym != -1) {
            dsq[L] = (char) sym;
            L++;
            if (L + 1 == alloc_L) {	/* L+1 leaves room for sentinel byte + \0 */
                alloc_L += 64;
                dsq = ReallocOrDie (dsq, sizeof (char) * alloc_L);
            }
        }
    }

    /* Finish off the trace
     */
    tr->tlen = tpos;

    /* Finish off the dsq with sentinel byte and null terminator.
     * Emitted Sequence length is L-1.
     */
    dsq[L] = (char) Alphabet_iupac;
    dsq[L + 1] = '\0';
    L--;

    /* Return
     */
    if (ret_dsq != NULL) {
        *ret_dsq = dsq;
    }
    else {
        free (dsq);
    }
    if (ret_L != NULL) {
        *ret_L = L;
    }
    if (ret_tr != NULL) {
        *ret_tr = tr;
    }
    else {
        P7FreeTrace (tr);
    }
    return;
}

#ifdef SRE_REMOVED
/* Function: EmitBestSequence()
 * Date:     SRE, Tue Nov 10 16:21:59 1998 [St. Louis]
 *
 * Purpose:  Given a model, emit the maximum probability sequence
 *           from it: argmax_{seq} P(seq | model).
 *           This is a sensible HMM equivalent to a "consensus"
 *           sequence.
 *           The model should be Plan7NakedConfig()'ed;
 *           in particular, if we allowed B->M and M->E,
 *           the highest probability sequence would be
 *           artifactually short. (We could do the highest
 *           scoring sequence instead, to get around this problem,
 *           but the highest scoring sequence is prone to
 *           other artifacts -- any looping state N,C,J, or I
 *           with a positively scoring residue leads to
 *           an infinitely long "best scoring" sequence.)
 *
 * Args:     hmm     - the model
 *           ret_seq - RETURN: best sequence
 *           ret_L   - RETURN: length of sequence
 *           ret_tr  - RETURN: traceback of the model/seq alignment; or NULL.
 *
 * Returns:  void
 */
void
EmitBestSequence (struct plan7_s *hmm, char **ret_dsq, int *ret_L,
                  struct p7trace_s **ret_tr)
{
    char *seq;			/* RETURN: best seq */
    struct p7trace_s *tr;		/* RETURN: traceback */
    float *mmx, *imx, *dmx;	/* log P forward scores for M,D,I */
    char *mtb, *itb, *dtb;	/* traceback ptrs for M,D,I */
    int x;			/* counter for symbols */
    int k;			/* counter for nodes   */
    float sc;			/* tmp var for a log P */
    int bestsym;
    int rpos;			/* position in a sequence */
    int tpos;			/* position in a trace */
    int tlen;			/* length of the traceback */

    /* Initial allocations. We only need a 1D matrix and its shadow;
     * it's overkill to use the Plan7Matrix structures, so don't.
     */
    mmx = MallocOrDie (sizeof (float) * (hmm->M + 1));
    imx = MallocOrDie (sizeof (float) * (hmm->M));
    dmx = MallocOrDie (sizeof (float) * (hmm->M));
    mtb = MallocOrDie (sizeof (char) * (hmm->M + 1));
    itb = MallocOrDie (sizeof (char) * (hmm->M));
    dtb = MallocOrDie (sizeof (char) * (hmm->M));

    /* Initialization.
     * We can safely assume a max probability path of S->N->B->(M1 or D1),
     * so just init M1 and D1.
     */
    mmx[1] = log (hmm->xt[XTN][MOVE]) + log (1.F - hmm->tbd1);
    dmx[1] =
        /* Main recursion, done as a push.
         * The model is used in probability form; no wing folding needed.
         */
    for (k = 1; k < hmm->M; k++) {
        /* Transits out of match state (init with these)
         */
        mmx[k + 1] = mmx[k] + log (hmm->t[k][TMM]);
        mtb[k + 1] = STM;
        dmx[k + 1] = mmx[k] + log (hmm->t[k][TMD]);
        dtb[k + 1] = STM;
        if (k < hmm->M - 1) {
            imx[k] = mmx[k] + log (hmm->t[k][TMI]);
        }
        itb[k] = STM;

        /* Transits out of delete state
         */
        if ((sc = dmx[k] + log (hmm->t[k][TDM])) > mmx[k + 1]) {
            mmx[k + 1] = sc;
            mtb[k + 1] = STD;
        }
        if ((sc = dmx[k] + log (hmm->t[k][TDD])) > dmx[k + 1]) {
            dmx[k + 1] = sc;
            dtb[k + 1] = STD;
        }

        /* Transits out of insert state (self-loops are never good)
         */
        if ((sc = imx[k] + log (hmm->t[k][TIM])) > mmx[k + 1]) {
            mmx[k + 1] = sc;
            mtb[k + 1] = STI;
        }

        /* Best emissions
         */
        x = FArgMax (hmm->mat[k + 1], Alphabet_size);
        mmx[k + 1] += log (hmm->mat[k + 1][x]);

        if (k < hmm->M - 1) {
            x = FArgMax (hmm->ins[k + 1], Alphabet_size);
            imx[k + 1] += log (hmm->ins[k + 1][x]);
        }
    }
}
#endif /* SRE_REMOVED */


/* Function: EmitConsensusSequence()
 * Date:     SRE, Wed Nov 11 11:08:59 1998 [St. Louis]
 *
 * Purpose:  Generate a "consensus sequence". For the purposes
 *           of a profile HMM, this is defined as:
 *              - for each node:
 *                 - if StateOccupancy() says that M is used
 *                     with probability >= 0.5, this M is "consensus".
 *                     Then, choose maximally likely residue.
 *                     if P>0.5 (protein) or P>0.9 (DNA), make
 *                     it upper case; else make it lower case.
 *                 - if StateOccupancy() says that I
 *                     is used with P >= 0.5, this I is "consensus";
 *                     use it 1/(1-TII) times (its expectation value).
 *                     Generate an "x" from each I.
 *
 *           The function expects that the model is config'ed
 *           by Plan7NakedConfig(): that is, for a single global pass
 *           with no N,C,J involvement.
 *
 *
 * Args:     hmm     - the model
 *           ret_seq - RETURN: consensus sequence (pass NULL if unwanted)
 *           ret_dsq - RETURN: digitized consensus sequence (pass NULL if unwanted)
 *           ret_L   - RETURN: length of generated sequence
 *           ret_tr  - RETURN: generated trace (pass NULL if unwanted)
 *
 * Returns:  void
 */
void
EmitConsensusSequence (struct plan7_s *hmm, char **ret_seq, char **ret_dsq,
                       int *ret_L, struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;		/* RETURN: traceback */
    char *dsq, *seq;		/* sequence in digitized and undigitized form */
    float *mp, *ip, *dp;		/* state occupancies from StateOccupancy() */
    int nmat, ndel, nins;		/* number of matches, deletes, inserts used */
    int k;			/* counter for nodes */
    int tpos;			/* position in trace */
    int i;			/* position in seq (equiv pos in dsq is i+1 */
    int x;			/* symbol choice (M) or # symbols (I) */
    float mthresh;		/* >= this, show symbol as upper case */

    if (Alphabet_type == hmmAMINO) {
        mthresh = 0.5;
    }
    else {
        mthresh = 0.9;
    }

    StateOccupancy (hmm, &mp, &ip, &dp);

    /* First pass: how many states do we need in the trace?
     *             how long will the sequence be?
     */
    nmat = ndel = nins = 0;
    for (k = 1; k <= hmm->M; k++) {
        if (mp[k] >= 0.5) {
            nmat++;
        }
        else {
            ndel++;
        }
        if (k < hmm->M && ip[k] >= 0.5) {
            nins += (int) (1.f / (1.f - hmm->t[k][TII]));
        }
    }

    /* Allocations
     */
    P7AllocTrace (6 + nmat + ndel + nins, &tr);
    dsq = MallocOrDie (sizeof (char) * (nmat + nins + 3));
    seq = MallocOrDie (sizeof (char) * (nmat + nins + 1));

    /* Main pass.
     * Construct consensus trace, seq, and dsq.
     */
    TraceSet (tr, 0, STS, 0, 0);
    TraceSet (tr, 1, STN, 0, 0);
    TraceSet (tr, 2, STB, 0, 0);
    dsq[0] = Alphabet_iupac;	/* guard byte */
    tpos = 3;
    i = 0;
    for (k = 1; k <= hmm->M; k++) {
        if (mp[k] >= 0.5) {
            x = FArgMax (hmm->mat[k], Alphabet_size);
            TraceSet (tr, tpos, STM, k, i + 1);
            seq[i] = Alphabet[x];
            dsq[i + 1] = x;
            if (hmm->mat[k][x] < mthresh) {
                seq[i] = tolower ((int) seq[i]);
            }
            i++;
            tpos++;
        }
        else {
            TraceSet (tr, tpos, STD, k, 0);
            tpos++;
        }

        if (k < hmm->M && ip[k] >= 0.5) {
            x = (int) (1.f / (1.f - hmm->t[k][TII]));
            while (x--) {
                TraceSet (tr, tpos, STI, k, i + 1);
                seq[i] = 'x';
                dsq[i + 1] = Alphabet_iupac - 1;
                i++;
                tpos++;
            }
        }
    }
    TraceSet (tr, tpos, STE, 0, 0);
    tpos++;
    TraceSet (tr, tpos, STC, 0, 0);
    tpos++;
    TraceSet (tr, tpos, STT, 0, 0);
    tpos++;
    dsq[i + 1] = Alphabet_iupac;

    free (mp);
    free (ip);
    free (dp);
    if (ret_seq != NULL) {
        *ret_seq = seq;
    }
    else {
        free (seq);
    }
    if (ret_dsq != NULL) {
        *ret_dsq = dsq;
    }
    else {
        free (dsq);
    }
    if (ret_L != NULL) {
        *ret_L = i;
    }
    if (ret_tr != NULL) {
        *ret_tr = tr;
    }
    else {
        P7FreeTrace (tr);
    }
}



/* Function: StateOccupancy()
 * Date:     SRE, Wed Nov 11 09:46:15 1998 [St. Louis]
 *
 * Purpose:  Calculate the expected state occupancy for
 *           a given HMM in generated traces.
 *
 *           Note that expected prob of getting into
 *           any special state in a trace is trivial:
 *              S,N,B,E,C,T = 1.0
 *              J = E->J transition prob
 *
 * Args:     hmm    - the model
 *           ret_mp - RETURN: [1..M] prob's of occupying M
 *           ret_ip - RETURN: [1..M-1] prob's of occupying I
 *           ret_dp - RETURN: [1..M] prob's of occupying D
 *
 * Returns:  void
 *           mp, ip, dp are malloc'ed here. Caller must free().
 */
void
StateOccupancy (struct plan7_s *hmm, float **ret_mp, float **ret_ip,
                float **ret_dp)
{
    float *fmp, *fip, *fdp;	/* forward probabilities  */
    int k;			/* counter for nodes      */

    /* Initial allocations
     */
    fmp = MallocOrDie (sizeof (float) * (hmm->M + 1));
    fip = MallocOrDie (sizeof (float) * (hmm->M));
    fdp = MallocOrDie (sizeof (float) * (hmm->M + 1));

    /* Forward pass.
     */
    fdp[1] = hmm->tbd1;
    fmp[1] = hmm->begin[1];
    fip[1] = fmp[1] * hmm->t[1][TMI];
    for (k = 2; k <= hmm->M; k++) {
        /* M: from M,D,I at k-1, or B; count t_II as 1.0 */
        fmp[k] = fmp[k - 1] * hmm->t[k - 1][TMM] +
                 fip[k - 1] + fdp[k - 1] * hmm->t[k - 1][TDM] + hmm->begin[k];
        /* D: from M,D at k-1 */
        fdp[k] = fmp[k - 1] * hmm->t[k - 1][TMD] +
                 fdp[k - 1] * hmm->t[k - 1][TDD];
        /* I: from M at k; don't count II */
        if (k < hmm->M) {
            fip[k] = fmp[k] * hmm->t[k][TMI];
        }

        SQD_DASSERT2 ((fabs (1.0f - fmp[k] - fdp[k]) < 1e-6f));
        fmp[k] /= fmp[k] + fdp[k];	/* prevent propagating fp errors */
        fdp[k] /= fmp[k] + fdp[k];
    }
    /* We don't need a backward pass; all backwards P's are 1.0
     * by definition (you can always get out of a state with P=1).
     * The only situation where this might not be true is for
     * a TII of 1.0, when TIM = 0 -- but in that case, if there's
     * a finite chance of getting into that insert state, the model
     * generates infinitely long sequences, so we can consider this
     * situation "perverse" and disallow it elsewhere in building
     * profile HMMs.
     */

    /* Return.
     */
    *ret_mp = fmp;
    *ret_dp = fdp;
    *ret_ip = fip;
}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* emulation.c
 * SRE, Wed Jan 21 07:50:01 1998
 *
 * Interfaces between HMMER and other software packages.
 *
 * CVS $Id: emulation.c,v 1.9 2003/04/14 16:00:16 eddy Exp $
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <string.h>

#include "squid.h"
#include "structs.h"
#include "funcs.h"


/* Function: WriteProfile()
 * Date:     SRE, Wed Jan 21 07:58:09 1998 [St. Louis]
 *
 * Purpose:  Given an HMM, write a GCG profile .prf file as
 *           output. Based on examination of Michael Gribskov's Fortran
 *           source in GCG 9.1; on reverse engineering
 *           by examination of GCG 9.1 output from "profilemake"
 *           and how the .prf file is used by "profilesearch";
 *           and on the GCG 9.0 documentation.
 *
 *           See notes 28 Jan 98 for detail; in brief, the conversion goes like:
 *
 *           PROF(i,k) = match score         =  msc(i,k) + TMM(k-1)
 *
 *           GAP(k)    = cost per insertion  =  TMI(k-1) + TIM(k-1) - TMM(k-1) - TII(k-1)
 *           LEN(k)    = cost per inserted x =  TII(k-1)
 *
 *           QGAP(k)   = cost per deletion   =  TDM(k-1) + TMD(unknown) - TMM(k-1) - TDD(k-1)
 *           QLEN(k)   = cost per deleted k  =  TDD(k-1)
 *
 *           Note that GCG affine gaps are GAP + n * LEN;
 *           HMMER affine gaps count (n-1) * gap-extend, thus an
 *           extra TII gets taken away from GAP (and TDD from QGAP),
 *           since GCG will charge it.
 *
 *           Also note how the TMM transitions, which have no equivalent
 *           in a profile, get smuggled in OK.
 *
 *           Also note that GCG charges gaps using the profile position
 *           /after/ the gap, not preceding the gap as HMMER does.
 *
 *           Also note the TMD(unknown) in the QGAP calculation. HMMER
 *           distinguishes between gap-open and gap-close, but GCG does not,
 *           so there is a fundamental incompatibility here. Here
 *           we use an upper (best-scoring, minimum-cost) bound.
 *
 *           And finally note that GCG's implementation forces GAP=QGAP and
 *           LEN=QLEN. Here, we upper bound again. Compugen's implementation
 *           allows an "extended profile" format which distinguishes between
 *           the two.
 *
 *           The upper bound approach to these scores means that a
 *           score given by an emulated profile is an upper bound: the HMMER
 *           score (for a single Smith/Waterman style local alignment)
 *           cannot be better than this. This is intentional, so that
 *           the Compugen BIC can be used for rapid prefiltering of
 *           the database.
 *
 *           To get a close approximation of hmmsw scores, call
 *           profilesearch as
 *                profilesearch -noave -nonor -gap 10 -len 1
 *           On the Compugen BIC, using extended profiles, you want:
 *                om -model=xsw.model -gapop=10 -gapext=1 -qgapop=10 -qgapext=1 -noave -nonor
 *
 * Args:     fp      - open FILE to write to (or stdout, possibly)
 *           hmm     - the HMM to write
 *           do_xsw  - TRUE to write Compugen's experimental extended profile format
 *
 * Returns:  (void)
 */
void
WriteProfile (FILE * fp, struct plan7_s *hmm, int do_xsw)
{
    int k;			/* position in model      */
    int x;			/* symbol index           */
    int sc;			/* a score to print       */
    float nx;			/* expected # of symbol x */
    int gap, len, qgap, qlen;	/* penalties to charge    */

    P7Logoddsify (hmm, TRUE);

    /* GCG can't deal with long profiles. Their limit is 1000
     * positions. However, Compugen can. Therefore we warn,
     * but don't die.
     */
    if (hmm->M > 1000 && !do_xsw)
        Warn
        ("Profile %s will have more than 1000 positions. GCG won't read it; Compugen will.",
         hmm->name);

    /* Header information.
     * GCG will look for sequence type and length of model.
     * Other than this, nothing is parsed until we get to the
     * Cons line that has a ".." on it.
     * Lines that begin with "!" are comments.
     */
    if (Alphabet_type == hmmAMINO) {
        fprintf (fp, "!!AA_PROFILE 1.0\n");
    }
    else if (Alphabet_type == hmmNUCLEIC) {
        fprintf (fp, "!!NA_PROFILE 1.0\n");
    }
    else {
        Die ("No support for profiles with non-biological alphabets");
    }

    if (Alphabet_type == hmmAMINO) {
        fprintf (fp, "(Peptide) ");
    }
    else if (Alphabet_type == hmmNUCLEIC) {
        fprintf (fp, "(Nucleotide) ");
    }
    fprintf (fp, "HMMCONVERT v%s Length: %d %s|%s|%s\n",
             PACKAGE_VERSION, hmm->M, hmm->name,
             hmm->flags & PLAN7_ACC ? hmm->acc : "",
             hmm->flags & PLAN7_DESC ? hmm->desc : "");

    /* Insert some HMMER-specific commentary
     */
    if (do_xsw) {
        fprintf (fp,
                 "   Profile converted from a profile HMM using HMMER v%s emulation.\n",
                 PACKAGE_VERSION);
        fprintf (fp, "   Compugen XSW extended profile format.\n");
        fprintf (fp,
                 "   Use -model=xsw.model -nonor -noave -gapop=10 -gapext=1 -qgapop=10 -qgapext=1\n");
        fprintf (fp,
                 "      with om on the Compugen BIC to get the closest approximation to HMMER bit scores.\n");
        fprintf (fp,
                 "   WARNING: There is a loss of information in this conversion.\n");
        fprintf (fp,
                 "      Neither the scores nor even the rank order of hits will be precisely\n");
        fprintf (fp,
                 "      preserved in a comparison of HMMER hmmsearch to GCG profilesearch.\n");
        fprintf (fp,
                 "      The profile score is an approximation of the (single-hit) HMMER score.\n\n");
    }
    else {
        fprintf (fp,
                 "   Profile converted from a profile HMM using HMMER v%s emulation.\n",
                 PACKAGE_VERSION);
        fprintf (fp,
                 "   Use -nonor -noave -gap=10 -len=1 with profilesearch and friends\n");
        fprintf (fp,
                 "      to get the closest approximation to HMMER bit scores.\n");
        fprintf (fp,
                 "   WARNING: There is a loss of information in this conversion.\n");
        fprintf (fp,
                 "      Neither the scores nor even the rank order of hits will be precisely\n");
        fprintf (fp,
                 "      preserved in a comparison of HMMER hmmsearch to GCG profilesearch.\n");
        fprintf (fp,
                 "      The profile score is an approximation of the (single-hit) HMMER score.\n\n");
    }


    /* Do the CONS line, which gives the valid IUPAC symbols and their order
     */
    fprintf (fp, "Cons");
    for (x = 0; x < Alphabet_iupac; x++) {
        fprintf (fp, "    %c ", Alphabet[x]);
    }
    if (do_xsw) {
        fprintf (fp, "  Gap   Len  QGap  Qlen ..\n");
    }
    else {
        fprintf (fp, "  Gap   Len ..\n");
    }

    /* Now, the profile; for each position in the HMM, write a line of profile.
     */
    for (k = 1; k <= hmm->M; k++) {
        /* GCG adds some indexing as comments */
        if ((k - 1) % 10 == 0 && k > 10) {
            fprintf (fp, "! %d\n", k);
        }

        /* find consensus residue by max prob */
        x = FArgMax (hmm->mat[k], Alphabet_size);
        fprintf (fp, " %c  ", Alphabet[x]);
        /* generate emission score profile;
         * Profiles are scaled by a factor of 100
         */
        for (x = 0; x < Alphabet_iupac; x++) {
            sc = hmm->msc[x][k];
            if (k < hmm->M) {
                sc += hmm->tsc[TMM][k];
            }
            sc = sc * 100 / INTSCALE;
            fprintf (fp, "%5d ", sc);
        }
        /* Generate gap open, gap extend penalties;
           note we will force profilesearch to weights of 10, 1,
           and that GCG profile values are percentages
           of these base penalties, 0..100. */
        /* gap open (insertion) */
        if (k > 1) {
            gap =
                -1 * (hmm->tsc[TMI][k - 1] + hmm->tsc[TIM][k - 1] -
                      hmm->tsc[TMM][k - 1] - hmm->tsc[TII][k - 1]);
            gap = gap * 100 / (10.0 * INTSCALE);
        }
        else {
            gap = 100;    /* doesn't matter because GAP_1 is never used */
        }

        /* gap extend (insertion) */
        if (k > 1) {
            len = -1 * hmm->tsc[TII][k - 1];
            len = len * 100 / (1.0 * INTSCALE);
        }
        else {
            len = 100;    /* again, doesn't matter because LEN_1 is never used */
        }

        /* gap open (deletion) */
        if (k > 1) {
            qgap =
                -1 * (hmm->tsc[TDM][k - 1] + hmm->tsc[TMD][k - 1] -
                      hmm->tsc[TMM][k - 1] - hmm->tsc[TDD][k - 1]);
            qgap = qgap * 100 / (10.0 * INTSCALE);
        }
        else {
            qgap = 100;
        }
        /* gap extend (deletion) */
        if (k > 1) {
            qlen = -1 * hmm->tsc[TDD][k - 1];
            qlen = qlen * 100 / (1.0 * INTSCALE);
        }
        else {
            qlen = 100;
        }


        if (do_xsw) {
            fprintf (fp, "%5d %5d %5d %5d\n", gap, len, qgap, qlen);
        }
        else {
            fprintf (fp, "%5d %5d\n", gap, len);    /* assume insertions >= deletions */
        }
    }

    /* The final line of the profile is a count of the observed
     * residues in the training sequences. This information is not
     * available in an HMM, and I'm not sure that GCG ever uses it.
     * Approximate it by calculating a /very/ rough expectation.
     */
    fprintf (fp, " *  ");
    for (x = 0; x < Alphabet_size; x++) {
        nx = 0.0;
        for (k = 1; k <= hmm->M; k++) {
            nx += hmm->mat[k][x];
        }
        nx *= hmm->nseq;
        fprintf (fp, "%5d ", (int) nx);
    }
    for (; x < Alphabet_iupac; x++) {
        fprintf (fp, "%5d ", 0);
    }
    fprintf (fp, "\n");
    return;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* eps.c
 * SRE, Thu Jun 21 18:02:31 2001 [St. Louis]
 *
 * Some crude support for Encapsulated PostScript (EPS) output,
 * DSC compliant.
 *
 * CVS $Id: eps.c,v 1.5 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "squid.h"
#include "msa.h"

/* Function: EPSWriteSmallMSA()
 * Date:     SRE, Thu Jun 21 18:15:21 2001 [St. Louis]
 *
 * Purpose:  Write an alignment in singleblock, Stockholm/SELEX like
 *           format to an open file. Very crude.
 *           Currently fails if the alignment is >50 columns long, because
 *           it doesn't think it will fit on a single page.
 *
 * Args:     fp  - open file for writing
 *           msa - alignment to write
 *
 * Returns:  (void)
 */
void
EPSWriteSmallMSA (FILE * fp, MSA * msa)
{
    int namewidth;		/* namewidth in PostScript units */
    int fontwidth;		/* width of a character in this font */
    int hspace;			/* horizontal space between aligned chars */
    int vspace;			/* vertical space between sequences */
    char *font;			/* font name, e.g. "Courier" */
    int fontsize;			/* font size in pts */
    int i, j;			/* counter over sequences, columns */
    int len;			/* tmp var holding length of something */
    int width, height;		/* width and height of bounding box */
    int xpos, ypos;		/* x,y position */

    /* Set some font characteristics; done here, so it'll
     * be easy to change. Magic numbers for Courier 12 determined
     * by trial and error.
     */
    fontwidth = 8;
    hspace = 9;
    vspace = 15;
    font = sre_strdup ("Courier", -1);
    fontsize = 12;

    /* Find the width of the longest sequence name in characters.
     */
    namewidth = 0;
    for (i = 0; i < msa->nseq; i++) {
        if ((len = (int) strlen (msa->sqname[i])) > namewidth) {
            namewidth = len;
        }
		}		
    namewidth += 1;		/* add a space to separate name & aligned seq */
    namewidth *= fontwidth;

    /* Determine bounding box
     */
    if (msa->alen > 50) {
        Die ("No EPS fmt if alignment is >50 columns");
    }
    width = namewidth + hspace * msa->alen;
    if (width > 612) {
        Die ("Alignment too wide to write in EPS");
    }
    height = vspace * msa->nseq;
    if (height > 792) {
        Die ("Too many seqs to write in EPS");
    }

    /* Magic EPS header, bare-bones DSC-compliant.
     */
    fprintf (fp, "%%!PS-Adobe-3.0 EPSF-3.0\n");
    fprintf (fp, "%%%%BoundingBox: %d %d %d %d\n", 0, 0, width, height);
    fprintf (fp, "%%%%Pages: 1\n");
    fprintf (fp, "%%%%EndComments\n");

    /* More postscript magic before we start the alignment
     */
    fprintf (fp, "/%s findfont\n", font);
    fprintf (fp, "%d scalefont\n", fontsize);
    fprintf (fp, "setfont\n");
    fprintf (fp, "newpath\n");

    /* Write the alignment in PostScript in a single block
     */
    for (i = 0; i < msa->nseq; i++) {
        ypos = (msa->nseq - i - 1) * vspace;
        /* name first */
        fprintf (fp, "%d %d moveto\n", 0, ypos);
        fprintf (fp, "(%s) show\n", msa->sqname[i]);
        /* now seq */
        xpos = namewidth;
        for (j = 0; j < msa->alen; j++) {
            fprintf (fp, "%d %d moveto\n", xpos, ypos);
            fprintf (fp, "(%c) show\n", msa->aseq[i][j]);
            xpos += hspace;
        }
    }

    free (font);
}


/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/
/* Removed ALTIVEC stuff as suggested by John Henning and Cloyce
   Kaivalya Tue Dec  2 16:33:43 CST 2003
*/

/* fast_algorithms.c
 * SRE, Sun Nov 10 08:54:48 2002 [AA 3080, Denver to StL]
 * CVS $Id: fast_algorithms.c,v 1.7 2003/05/26 19:33:09 eddy Exp $
 *
 * Optimized routines to replace slower implementations in core_algorithms.c.
 *
 * The routines in core_algorithms.c are designed for clarity
 * and maintainability, not for speed. Implementations here
 * are designed for speed, not clarity. If you're trying to
 * understand the code, or optimize for a specific platform,
 * you are probably better off looking at core_algorithms.c.
 *
 * P7Viterbi() is the key function to target optimization to.
 * The implementation in core_algorithms.c is currently ifdef'ed
 * out of the code. The implementation that is used by default
 * is here, in fast_algorithms.c. A third implementation, from
 * Erik Lindahl at Stanford, is Mac/Altivec specific.
 *
 * Which implementation is used is controlled by ifdef's. The
 * default implementation uses a fast implementation of
 * P7Viterbi() from here. Other options (mutually exclusive):
 *
 * -DSLOW
 *   enable original core_algorithms.c code: slower than default,
 *   but might be easier to follow, for someone trying
 *   to understand the DP code.
 * -DALTIVEC
 *   enable Erik Lindahl's Altivec code for Macintosh OSX
 */

#include "config.h"
#include "squidconf.h"

#include "structs.h"
#include "funcs.h"
#include "squid.h"

/* the DEFAULT P7Viterbi() is portably optimized; code follows:
 */
#if !defined ALTIVEC && !defined SLOW
/* Function: P7Viterbi() - portably optimized version
 * Incept:   SRE, Fri Nov 15 13:14:33 2002 [St. Louis]
 *
 * Purpose:  The Viterbi dynamic programming algorithm.
 *           Derived from core_algorithms.c:P7Viterbi().
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           mx     - re-used DP matrix
 *           ret_tr - RETURN: traceback; pass NULL if it's not wanted
 *
 * Return:   log P(S|M)/P(S|R), as a bit score
 */
float
P7Viterbi (char *dsq, int L, struct plan7_s *hmm, struct dpmatrix_s *mx,
           struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;
    int **xmx;
    int **mmx;
    int **imx;
    int **dmx;
    int i, k;
    int sc;
    int *mc, *dc, *ic;		/* pointers to rows of mmx, dmx, imx */
    int *ms, *is;			/* pointers to msc[i], isc[i] */
    int *mpp, *mpc, *ip;		/* ptrs to mmx[i-1], mmx[i], imx[i-1] */
    int *bp;			/* ptr into bsc[] */
    int *ep;			/* ptr into esc[] */
    int xmb;			/* value of xmx[i-1][XMB] */
    int xme;			/* max for xmx[i][XME] */
    int *dpp;			/* ptr into dmx[i-1] (previous row) */
    int *tpmm, *tpmi, *tpmd, *tpim, *tpii, *tpdm, *tpdd;	/* ptrs into tsc */
    int M;

    /* Make sure we have space for a DP matrix with 0..L rows, 0..M-1 columns.
     */
    ResizePlan7Matrix (mx, L, hmm->M, &xmx, &mmx, &imx, &dmx);

    /* Initialization of the zero row.
     */
    xmx[0][XMN] = 0;		/* S->N, p=1            */
    xmx[0][XMB] = hmm->xsc[XTN][MOVE];	/* S->N->B, no N-tail   */
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need seq to get here */
    for (k = 0; k <= hmm->M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;    /* need seq to get here */
    }

    /* Initializations that help icc vectorize.
     */
    M = hmm->M;

    /* Recursion. Done as a pull.
     * Note some slightly wasteful boundary conditions:
     *    tsc[0] = -INFTY for all eight transitions (no node 0)
     *    D_M and I_M are wastefully calculated (they don't exist)
     */

    tpmm = hmm->tsc[TMM];
    tpim = hmm->tsc[TIM];
    tpdm = hmm->tsc[TDM];
    tpmd = hmm->tsc[TMD];
    tpdd = hmm->tsc[TDD];
    tpmi = hmm->tsc[TMI];
    tpii = hmm->tsc[TII];
    bp = hmm->bsc;

    //***********************************************************************
    // speccode: Initializing structures
    //specinit();
    //initializeCounters();
    //***********************************************************************
    // 
    // En TRAIN se realizan 85.000 veces estas operaciones
    //***********************************************************************
    // speccode: OMP threading directive
    omp_set_num_threads(threads);
    // speccode: initialize speculation structures
    //specstart(L);
    specbegin(L);
    // speccode: variable upper limit inicialization
    upper_limit = L;
    /*fprintf(stderr,"L=%d\n",L);
      fflush(stderr);*/
    //***********************************************************************
    //***********************************************************************
    // speccode: Start profiling
    // collector_resume();
    //***********************************************************************

    int iteration, ini, value;
    double spec_start, spec_stop;
    int aux;
    // speccode: Start profiling
    spec_start = get_time ();

    //***********************************************************************
    // speccode: Speculative loop
    #pragma omp parallel default(none) \
          private(mc, dc, ic, mpp, dpp, ip, xmb, ms, is, \
                  k, xme, mpc, ep, sc, \
                  iteration, ini, \
                  value, current, tid, retflag,aux)    \
          shared(xmx, mmx, dmx, imx, dsq, L, M, \
                 tpmm, tpim, tpdm, tpdd, tpmi, tpii, tpmd, bp, hmm, \
                 wheel_ns, wheel_ms, wheel,upper_limit, varblock)
    {

    #pragma omp for \
          schedule(static)

      // Format: initLoopSpecEngine(vector, iterator, ini,IDLOOP)
      //      Beginning of process
      // for (i = 1; i <= L; i++) 
      // initLoopSpecEngine(xmx, i, 1, 1)
      for (tid = 0; tid <= threads - 1; tid++) 
      {
        ini = 1;
        current = tid;
        iteration = varblock[0][current] + ini;

        //if (iteration > upper_limit - 1) ALVARO: este bucle -> i=1;i<=L
        if (iteration > upper_limit)
          goto labelSquash_1;

        varblock[2][current] = varblock[2][current] + 1;
        //printf("\t\tvarblock[0][current]:%d,current:%d,iteration:%d\n",varblock[0][current],current,iteration);
labelStartIteration_1: //printf("ENDLOOP,iteration:%d,current:%d\n",iteration,current);
        //  Codigo del bucle

        /* Now the special states. Order is important here.
         * remember, C and J emissions are zero score by definition,
         */
        /* N state */
        //***********************************************************************
        // speccode: speculative store. Original line:
        // xmx[iteration][XMN] = -INFTY;
        value = -INFTY;
        specstore_pointer((void *) &(xmx[iteration][XMN]),
            sizeof (xmx[iteration][XMN]),
            current,
            (void *) &value);
        // speccode: end of speculative store
        //***********************************************************************


        //***********************************************************************
        // speccode: speculative load. Original line:
        // if ((sc = xmx[iteration - 1][XMN] + hmm->xsc[XTN][LOOP]) > -INFTY) 
        if (specload_pointer((void *) &(xmx[iteration - 1][XMN]),
              sizeof (xmx[iteration - 1][XMN]),
              current,
              (void *) &value) == -1)
          earlySquash(1);
        //***********************************************************************
        if ((sc = value + hmm->xsc[XTN][LOOP]) > -INFTY) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          // xmx[iteration][XMN] = sc;
          specstore_pointer((void *) &(xmx[iteration][XMN]),
              sizeof (xmx[iteration][XMN]),
              current,
              (void *) &sc);
          // speccode: end of speculative store
          //***********************************************************************
        }

        /* E state */
        xme = -INFTY;
        mpc = mmx[iteration];
        ep = hmm->esc;
        for (k = 1; k <= hmm->M; k++) {
          if ((sc = mpc[k] + ep[k]) > xme) {
            xme = sc;
          }
        }		
        //***********************************************************************
        // speccode: speculative store. Original line:
        // xmx[iteration][XME] = xme;
        specstore_pointer((void *) &(xmx[iteration][XME]),
            sizeof (xmx[iteration][XME]),
            current,
            (void *) &xme);
        // speccode: end of speculative store
        //***********************************************************************

        /* J state */
        //***********************************************************************
        // speccode: speculative store. Original line:
        // xmx[iteration][XMJ] = -INFTY;
        value = -INFTY;
        specstore_pointer((void *) &(xmx[iteration][XMJ]),
            sizeof (xmx[iteration][XMJ]),
            current,
            (void *) &value);
        // speccode: end of speculative store
        //***********************************************************************

        //***********************************************************************
        // speccode: speculative load. Original line:
        // if ((sc = xmx[iteration - 1][XMJ] + hmm->xsc[XTJ][LOOP]) > -INFTY) 
        if (specload_pointer((void *) &(xmx[iteration - 1][XMJ]),
              sizeof (xmx[iteration - 1][XMJ]),
              current,
              (void *) &value) == -1)
          earlySquash(1);
        //***********************************************************************
        if ((sc = value + hmm->xsc[XTJ][LOOP]) > -INFTY) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          // xmx[iteration][XMJ] = sc;
          specstore_pointer((void *) &(xmx[iteration][XMJ]),
              sizeof (xmx[iteration][XMJ]),
              current,
              (void *) &sc);
          // speccode: end of speculative store
          //***********************************************************************
        }

        //***********************************************************************
        // speccode: speculative load. Original line:
        // if ((sc = xmx[iteration][XME] + hmm->xsc[XTE][LOOP]) > xmx[iteration][XMJ]) 
        if (specload_pointer((void *) &(xmx[iteration][XME]),
              sizeof (xmx[iteration][XME]),
              current,
              (void *) &sc) == -1)
          earlySquash(1);
        if (specload_pointer((void *) &(xmx[iteration][XMJ]),
              sizeof (xmx[iteration][XMJ]),
              current,
              (void *) &value) == -1)
          earlySquash(1);
        //***********************************************************************
        if ((sc = sc + hmm->xsc[XTE][LOOP]) > value) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          // xmx[iteration][XMJ] = sc;
          specstore_pointer((void *) &(xmx[iteration][XMJ]),
              sizeof (xmx[iteration][XMJ]),
              current,
              (void *) &sc);
          // speccode: end of speculative store
          //***********************************************************************
        }

        /* B state */
        //***********************************************************************
        // speccode: speculative store. Original line:
        // xmx[iteration][XMB] = -INFTY;
        value = -INFTY;
        specstore_pointer((void *) &(xmx[iteration][XMB]),
            sizeof (xmx[iteration][XMB]),
            current,
            (void *) &value);
        // speccode: end of speculative store
        //***********************************************************************

        //***********************************************************************
        // speccode: speculative load. Original line:
        // if ((sc = xmx[iteration][XMN] + hmm->xsc[XTN][MOVE]) > -INFTY) 
        if (specload_pointer((void *) &(xmx[iteration][XMN]),
              sizeof (xmx[iteration][XMN]),
              current,
              (void *) &value) == -1)
          earlySquash(1);
        //***********************************************************************
        if ((sc = value + hmm->xsc[XTN][MOVE]) > -INFTY) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          // xmx[iteration][XMB] = sc;
          specstore_pointer((void *) &(xmx[iteration][XMB]),
              sizeof (xmx[iteration][XMB]),
              current,
              (void *) &sc);
          // speccode: end of speculative store
          //***********************************************************************
        }

        //***********************************************************************
        // speccode: speculative load. Original line:
        // if ((sc = xmx[iteration][XMJ] + hmm->xsc[XTJ][MOVE]) > xmx[iteration][XMB]) 
        if (specload_pointer((void *) &(xmx[iteration][XMJ]),
              sizeof (xmx[iteration][XMJ]),
              current,
              (void *) &sc) == -1)
          earlySquash(1);
        if (specload_pointer((void *) &(xmx[iteration][XMB]),
              sizeof (xmx[iteration][XMB]),
              current,
              (void *) &value) == -1)
          earlySquash(1);
        //***********************************************************************
        if ((sc = sc + hmm->xsc[XTJ][MOVE]) > value) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          // xmx[iteration][XMB] = sc;
          specstore_pointer((void *) &(xmx[iteration][XMB]),
              sizeof (xmx[iteration][XMB]),
              current,
              (void *) &sc);
          // speccode: end of speculative store
          //***********************************************************************
        }

        /* C state */
        //***********************************************************************
        // speccode: speculative store. Original line:
        // xmx[iteration][XMC] = -INFTY;
        value = -INFTY;
        specstore_pointer((void *) &(xmx[iteration][XMC]),
            sizeof (xmx[iteration][XMC]),
            current,
            (void *) &value);
        // speccode: end of speculative store
        //***********************************************************************

        //***********************************************************************
        // speccode: speculative load. Original line:
        // if ((sc = xmx[iteration - 1][XMC] + hmm->xsc[XTC][LOOP]) > -INFTY) 
        if (specload_pointer((void *) &(xmx[iteration - 1][XMC]),
              sizeof (xmx[iteration - 1][XMC]),
              current,
              (void *) &value) == -1)
          earlySquash(1);
        //***********************************************************************
        if ((sc = value + hmm->xsc[XTC][LOOP]) > -INFTY) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          // xmx[iteration][XMC] = sc;
          specstore_pointer((void *) &(xmx[iteration][XMC]),
              sizeof (xmx[iteration][XMC]),
              current,
              (void *) &sc);
          // speccode: end of speculative store
          //***********************************************************************
        }

        // speccode: speculative load. Original line:
        // if ((sc = xmx[iteration][XME] + hmm->xsc[XTE][MOVE]) > xmx[iteration][XMC]) 
        if (specload_pointer((void *) &(xmx[iteration][XME]),
              sizeof (xmx[iteration][XME]),
              current,
              (void *) &sc) == -1)
          earlySquash(1);
        if (specload_pointer((void *) &(xmx[iteration][XMC]),
              sizeof (xmx[iteration][XMC]),
              current,
              (void *) &value) == -1)
          earlySquash(1);
        //***********************************************************************
        if ((sc = sc + hmm->xsc[XTE][MOVE]) > value) {
          //***********************************************************************
          // speccode: speculative store. Original line:
          // xmx[iteration][XMC] = sc;
          specstore_pointer((void *) &(xmx[iteration][XMC]),
              sizeof (xmx[iteration][XMC]),
              current,
              (void *) &sc);
          // speccode: end of speculative store
          //***********************************************************************
        }

        mc = mmx[iteration];
        dc = dmx[iteration];
        ic = imx[iteration];
        mpp = mmx[iteration - 1];
        dpp = dmx[iteration - 1];
        ip = imx[iteration - 1];
        ms = hmm->msc[(int) dsq[iteration]];
        is = hmm->isc[(int) dsq[iteration]];
        mc[0] = -INFTY;
        dc[0] = -INFTY;
        ic[0] = -INFTY;
        aux = 0;
        for (k = 1; k <= M; k++) {
          mc[k] = mpp[k - 1] + tpmm[k - 1];
          if ((sc = ip[k - 1] + tpim[k - 1]) > mc[k]) {
            mc[k] = sc;
          }
          if ((sc = dpp[k - 1] + tpdm[k - 1]) > mc[k]) {
            mc[k] = sc;
          }
          if (aux ==0) {
            //**********************************************************************
            // speccode: speculative load. Original line:
            // xmb = xmx[iteration - 1][XMB];
            if (specload_pointer((void *) &(xmx[iteration - 1][XMB]), 
                  sizeof (xmx[iteration - 1][XMB]), 
                  current, 
                  (void *) &xmb) == -1)
              earlySquash(1);
            //**********************************************************************
            aux=1;
          }
          if ((sc = xmb + bp[k]) > mc[k]) {
            mc[k] = sc;
          }
          mc[k] += ms[k];
          if (mc[k] < -INFTY) {
            mc[k] = -INFTY;
          }

          dc[k] = dc[k - 1] + tpdd[k - 1];
          if ((sc = mc[k - 1] + tpmd[k - 1]) > dc[k]) {
            dc[k] = sc;
          }
          if (dc[k] < -INFTY) {
            dc[k] = -INFTY;
          }

          if (k < M) {
            ic[k] = mpp[k] + tpmi[k];
            if ((sc = ip[k] + tpii[k]) > ic[k]) {
              ic[k] = sc;
            }
            ic[k] += is[k];
            if (ic[k] < -INFTY) {
              ic[k] = -INFTY;
            }
          }
        }

        //        nIterations = L; // From 1 while <=L

labelEndIteration_1:
        //if ((iteration != varblock[1][current] + ini) && (iteration < nIterations - 1))
        if ((iteration != varblock[1][current] + ini) && (iteration < L - 1))
        {
          iteration = iteration + 1;
          goto labelStartIteration_1;
        }
        /* Thread done, perform the commit */
labelSquash_1:
        retflag = threadend_pointer(&current);
        /* End if job done */
        if (retflag == JOBDONE) goto labelEndLoop_1;
        if (retflag == JOBTODO)
        {
          /*  Set loop variable */
          iteration = varblock[0][current] + ini;
          varblock[2][current] = varblock[2][current] + 1;
          goto labelStartIteration_1;
        }
labelEndLoop_1:
        ; /*freeWheel(current);*/
      } /*End for initLoop()*/
    } // #pragma omp parallel

    /************************************************************************/
    // speccode: Stop measuring time
    spec_stop = get_time ();
    spec_time += spec_stop - spec_start;
    /************************************************************************/

    // speccode: free speculation structures
    //specend();

    /* T state (not stored) */
    sc = xmx[L][XMC] + hmm->xsc[XTC][MOVE];

    if (ret_tr != NULL) {
      P7ViterbiTrace (hmm, dsq, L, mx, &tr);
      *ret_tr = tr;
    }

    return Scorify (sc);		/* the total Viterbi score. */
}
#endif /*default P7Viterbi, used when ALTIVEC and SLOW are not defined */

#if !defined(SPEC_CPU)
#ifdef ALTIVEC
/*################################################################
 * The Altivec port, for Macintosh PowerPC.
 * Erik Lindahl, Stanford University, 2002.
 *
 * Replaces the following functions:
 *    AllocPlan7Body()      plan7.c                (data alignment on 16-byte boundaries)
 *    CreatePlan7Matrix()   core_algorithms.c      (data alignment on 16-byte boundaries)
 *    ResizePlan7Matrix()   core_algorithms.c      (data alignment on 16-byte boundaries)
 *    P7Viterbi()           core_algorithms.c      (total recode, w/ Altivec instructions)
 ################################################################*/
void
AllocPlan7Body (struct plan7_s *hmm, int M)
{
    int k, x;

    hmm->M = M;

    hmm->rf = MallocOrDie ((M + 2) * sizeof (char));
    hmm->cs = MallocOrDie ((M + 2) * sizeof (char));
    hmm->ca = MallocOrDie ((M + 2) * sizeof (char));
    hmm->map = MallocOrDie ((M + 1) * sizeof (int));

    hmm->t = MallocOrDie (M * sizeof (float *));
    hmm->tsc = MallocOrDie (7 * sizeof (int *));
    hmm->mat = MallocOrDie ((M + 1) * sizeof (float *));
    hmm->ins = MallocOrDie (M * sizeof (float *));
    hmm->msc = MallocOrDie (MAXCODE * sizeof (int *));
    hmm->isc = MallocOrDie (MAXCODE * sizeof (int *));

    hmm->t[0] = MallocOrDie ((7 * M) * sizeof (float));
    /* Allocate extra memory so tsc[TMM,TIM,TDM,TMD,TDD] start on the
     * 16-byte cache boundary, and tsc[TMI,TII] start
     * 12 bytes offset from the boundary.
     */
    hmm->tsc_mem = MallocOrDie (((7 * (M + 16))) * sizeof (int));
    hmm->mat[0] = MallocOrDie ((MAXABET * (M + 1)) * sizeof (float));
    hmm->ins[0] = MallocOrDie ((MAXABET * M) * sizeof (float));
    /* Allocate extra mem. to make sure all members of msc,isc start
     * on 12-byte offsets from cache boundary.
     */
    hmm->msc_mem = MallocOrDie ((MAXCODE * (M + 1 + 16)) * sizeof (int));
    hmm->isc_mem = MallocOrDie ((MAXCODE * (M + 16)) * sizeof (int));

    /* note allocation strategy for important 2D arrays -- trying
     * to keep locality as much as possible, cache efficiency etc.
     */
    for (k = 1; k <= M; k++) {
        hmm->mat[k] = hmm->mat[0] + k * MAXABET;
        if (k < M) {
            hmm->ins[k] = hmm->ins[0] + k * MAXABET;
            hmm->t[k] = hmm->t[0] + k * 7;
        }
    }

    /* align tsc pointers */
    hmm->tsc[TMM] =
        (int *) (((((unsigned long int) hmm->tsc_mem) + 15) & (~0xf)));
    hmm->tsc[TMI] =
        (int
         *) (((((unsigned long int) hmm->tsc_mem) + (M + 12) * sizeof (int) +
               15) & (~0xf)) + 12);
    hmm->tsc[TMD] =
        (int
         *) (((((unsigned long int) hmm->tsc_mem) + 2 * (M + 12) * sizeof (int) +
               15) & (~0xf)));
    hmm->tsc[TIM] =
        (int
         *) (((((unsigned long int) hmm->tsc_mem) + 3 * (M + 12) * sizeof (int) +
               15) & (~0xf)));
    hmm->tsc[TII] =
        (int
         *) (((((unsigned long int) hmm->tsc_mem) + 4 * (M + 12) * sizeof (int) +
               15) & (~0xf)) + 12);
    hmm->tsc[TDM] =
        (int
         *) (((((unsigned long int) hmm->tsc_mem) + 5 * (M + 12) * sizeof (int) +
               15) & (~0xf)));
    hmm->tsc[TDD] =
        (int
         *) (((((unsigned long int) hmm->tsc_mem) + 6 * (M + 12) * sizeof (int) +
               15) & (~0xf)));

    for (x = 0; x < MAXCODE; x++) {
        hmm->msc[x] =
            (int
             *) (((((unsigned long int) hmm->msc_mem) +
                   x * (M + 1 + 12) * sizeof (int) + 15) & (~0xf)) + 12);
        hmm->isc[x] =
            (int
             *) (((((unsigned long int) hmm->isc_mem) +
                   x * (M + 12) * sizeof (int) + 15) & (~0xf)) + 12);
    }
    /* tsc[0] is used as a boundary condition sometimes [Viterbi()],
     * so set to -inf always.
     */
    for (x = 0; x < 7; x++) {
        hmm->tsc[x][0] = -INFTY;
    }

    hmm->begin = MallocOrDie ((M + 1) * sizeof (float));
    hmm->bsc_mem = MallocOrDie ((M + 1 + 12) * sizeof (int));
    hmm->end = MallocOrDie ((M + 1) * sizeof (float));
    hmm->esc_mem = MallocOrDie ((M + 1 + 12) * sizeof (int));

    hmm->bsc =
        (int *) (((((unsigned long int) hmm->bsc_mem) + 15) & (~0xf)) + 12);
    hmm->esc =
        (int *) (((((unsigned long int) hmm->esc_mem) + 15) & (~0xf)) + 12);

    return;
}


/* Function: CreatePlan7Matrix()
 *
 * Purpose:  Create a dynamic programming matrix for standard Forward,
 *           Backward, or Viterbi, with scores kept as scaled log-odds
 *           integers. Keeps 2D arrays compact in RAM in an attempt
 *           to maximize cache hits.
 *
 *           The mx structure can be dynamically grown, if a new
 *           HMM or seq exceeds the currently allocated size. Dynamic
 *           growing is more efficient than an alloc/free of a whole
 *           matrix for every new target. The ResizePlan7Matrix()
 *           call does this reallocation, if needed. Here, in the
 *           creation step, we set up some pads - to inform the resizing
 *           call how much to overallocate when it realloc's. If a pad
 *           is zero, we will not resize in that dimension.
 *
 * Args:     N     - N+1 rows are allocated, for sequence.
 *           M     - size of model in nodes
 *           padN  - over-realloc in seq/row dimension, or 0
 *           padM  - over-realloc in HMM/column dimension, or 0
 *
 * Return:   mx
 *           mx is allocated here. Caller frees with FreePlan7Matrix(mx).
 */
struct dpmatrix_s *
CreatePlan7Matrix (int N, int M, int padN, int padM)
{
    struct dpmatrix_s *mx;
    int i, n;

    mx = (struct dpmatrix_s *) MallocOrDie (sizeof (struct dpmatrix_s));
    mx->xmx = (int **) MallocOrDie (sizeof (int *) * (N + 1));
    mx->mmx = (int **) MallocOrDie (sizeof (int *) * (N + 1));
    mx->imx = (int **) MallocOrDie (sizeof (int *) * (N + 1));
    mx->dmx = (int **) MallocOrDie (sizeof (int *) * (N + 1));

    /* For the memory accessed by the altivec routines, we want to have
     * accesses aligned to 16-byte boundaries as far as possible.
     * To accomplish this, we align extra memory and then set the first
     * pointer on each row to point to 4 bytes before a boundary.
     * This means element 1, which is the first one we work on, will be
     * on a 16-byte boundary. We still make sure we 'own' the three bytes
     * before, though, so we can load the vector with element 0 cache-aligned too.
     * The real pointers to memory are kept in xmx_mem,mmx_mem,imx_mem,dmx_mem.
     */
    mx->xmx_mem = (void *) MallocOrDie (sizeof (int) * (N + 1) * (5 + 16));
    mx->mmx_mem = (void *) MallocOrDie (sizeof (int) * (N + 1) * (M + 2 + 16));
    mx->imx_mem = (void *) MallocOrDie (sizeof (int) * (N + 1) * (M + 2 + 16));
    mx->dmx_mem = (void *) MallocOrDie (sizeof (int) * (N + 1) * (M + 2 + 16));

    mx->xmx[0] =
        (int *) (((((unsigned long int) mx->xmx_mem) + 15) & (~0xf)) + 12);
    mx->mmx[0] =
        (int *) (((((unsigned long int) mx->mmx_mem) + 15) & (~0xf)) + 12);
    mx->imx[0] =
        (int *) (((((unsigned long int) mx->imx_mem) + 15) & (~0xf)) + 12);
    mx->dmx[0] =
        (int *) (((((unsigned long int) mx->dmx_mem) + 15) & (~0xf)) + 12);

    /* And make sure the beginning of each row is aligned the same way */
    for (i = 1; i <= N; i++) {
        mx->xmx[i] = mx->xmx[0] + i * (5 + 11);	/* add 11 bytes per row, making it divisible by 4 */
        n = 12 - (M + 2) % 4;
        mx->mmx[i] = mx->mmx[0] + i * (M + 2 + n);
        mx->imx[i] = mx->imx[0] + i * (M + 2 + n);
        mx->dmx[i] = mx->dmx[0] + i * (M + 2 + n);
    }

    if (padM > 0 && padN > 0) {
        Die ("there's trouble with RAMLIMIT if you grow in both M and N.");
    }
    mx->maxN = N;
    mx->maxM = M;
    mx->padN = padN;
    mx->padM = padM;

    return mx;
}

/* Function: ResizePlan7Matrix()
 *
 * Purpose:  Reallocate a dynamic programming matrix, if necessary,
 *           for a problem of NxM: sequence length N, model size M.
 *           (N=1 for small memory score-only variants; we allocate
 *           N+1 rows in the DP matrix.)
 *
 *           Returns individual ptrs to the four matrix components
 *           as a convenience.
 *
 * Args:     mx    - an already allocated model to grow.
 *           N     - seq length to allocate for; N+1 rows
 *           M     - size of model
 *           xmx, mmx, imx, dmx
 *                 - RETURN: ptrs to four mx components as a convenience
 *
 * Return:   (void)
 *           mx is (re)allocated here.
 */
void
ResizePlan7Matrix (struct dpmatrix_s *mx, int N, int M,
                   int ***xmx, int ***mmx, int ***imx, int ***dmx)
{
    int i, n;

    if (N <= mx->maxN && M <= mx->maxM) {
        if (xmx != NULL) {
            *xmx = mx->xmx;
        }
        if (mmx != NULL) {
            *mmx = mx->mmx;
        }
        if (imx != NULL) {
            *imx = mx->imx;
        }
        if (dmx != NULL) {
            *dmx = mx->dmx;
        }
        return;
    }

    if (N > mx->maxN) {
        N += mx->padN;
        mx->maxN = N;
        mx->xmx = (int **) ReallocOrDie (mx->xmx, sizeof (int *) * (N + 1));
        mx->mmx = (int **) ReallocOrDie (mx->mmx, sizeof (int *) * (N + 1));
        mx->imx = (int **) ReallocOrDie (mx->imx, sizeof (int *) * (N + 1));
        mx->dmx = (int **) ReallocOrDie (mx->dmx, sizeof (int *) * (N + 1));
    }

    if (M > mx->maxM) {
        M += mx->padM;
        mx->maxM = M;
    }

    mx->xmx_mem = ReallocOrDie (mx->xmx_mem, sizeof (int) * (N + 1) * (5 + 16));
    mx->mmx_mem =
        ReallocOrDie (mx->mmx_mem, sizeof (int) * (N + 1) * (M + 2 + 16));
    mx->imx_mem =
        ReallocOrDie (mx->imx_mem, sizeof (int) * (N + 1) * (M + 2 + 16));
    mx->dmx_mem =
        ReallocOrDie (mx->dmx_mem, sizeof (int) * (N + 1) * (M + 2 + 16));

    mx->xmx[0] =
        (int *) (((((unsigned long int) mx->xmx_mem) + 15) & (~0xf)) + 12);
    mx->mmx[0] =
        (int *) (((((unsigned long int) mx->mmx_mem) + 15) & (~0xf)) + 12);
    mx->imx[0] =
        (int *) (((((unsigned long int) mx->imx_mem) + 15) & (~0xf)) + 12);
    mx->dmx[0] =
        (int *) (((((unsigned long int) mx->dmx_mem) + 15) & (~0xf)) + 12);

    /* And make sure the beginning of each row is aligned the same way */
    for (i = 1; i <= N; i++) {
        mx->xmx[i] = mx->xmx[0] + i * (5 + 11);	/* add 11 bytes per row, making it divisible by 4 */
        n = 12 - (M + 2) % 4;
        mx->mmx[i] = mx->mmx[0] + i * (M + 2 + n);
        mx->imx[i] = mx->imx[0] + i * (M + 2 + n);
        mx->dmx[i] = mx->dmx[0] + i * (M + 2 + n);
    }

    if (xmx != NULL) {
        *xmx = mx->xmx;
    }
    if (mmx != NULL) {
        *mmx = mx->mmx;
    }
    if (imx != NULL) {
        *imx = mx->imx;
    }
    if (dmx != NULL) {
        *dmx = mx->dmx;
    }
}


/* Function: P7Viterbi()
 *
 * Purpose:  The Viterbi dynamic programming algorithm; Altivec implementation
 *           by Erik Lindahl, Stanford University, 2002.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           mx     - DP matrix (may get grown here)
 *           ret_tr - RETURN: traceback; pass NULL if it's not wanted
 *
 * Return:   log P(S|M)/P(S|R), as a bit score
 */
/* This first version of P7Viterbi has been accelerated with Altivec vectorization.
 * On Apple hardware, it is up to a factor 10 faster than the old non-altivec version.
 */
typedef union
{
    vector signed int v;
    int i[4];
} ivector;
void
printivec (vector signed int z)
{
    ivector q;
    q.v = z;
    printf ("%d  %d  %d  %d\n", q.i[0], q.i[1], q.i[2], q.i[3]);
}

float
P7Viterbi (char *dsq, int L, struct plan7_s *hmm, struct dpmatrix_s *mx,
           struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;
    int **xmx;
    int **mmx;
    int **imx;
    int **dmx;
    int *mmxi, *xmxi, *imxi, *dmxi;
    int *lmmxi, *lxmxi, *limxi, *ldmxi;
    int *p_tmm, *p_tim, *p_tdm, *p_bsc, *p_msc, *p_tmd, *p_tdd, *p_tmi, *p_tii,
        *p_isc, *p_esc;
    int i, n, k;
    int sc;
    /* gcc and motorola use different syntax for initializing vectors,
     * so we use a dummy variable instead to get an adress to load from...
     */
    int t_lowscore = -INFTY;

    /* vector variables. We avoid the stupid gcc register spill/fill code by
     * limiting ourselves to 32 generic variables and making the all registers.
     * (This reuse is the reason for the generic variable names).
     */
    register vector signed int v_lowscore;
    register vector signed int max_mmxesc;
    register vector signed int v_xmb;
    register vector unsigned int mask1;
    register vector unsigned int mask2;
    register vector unsigned int mask3;
    register vector unsigned int mask4;
    register vector signed int v_lmmx1;
    register vector signed int v_lmmx2;
    register vector signed int v_limx1;
    register vector signed int v_limx2;
    register vector signed int v_save_lmmx;
    register vector signed int v_save_ldmx;
    register vector signed int v_save_limx;
    register vector signed int v_save_mmx;
    register vector signed int v_save_dmx;
    register vector signed int v1;
    register vector signed int v2;
    register vector signed int v3;
    register vector signed int v4;
    register vector signed int v5;
    register vector signed int v6;
    register vector signed int v7;
    register vector signed int v8;
    register vector signed int v9;
    register vector signed int v10;
    register vector signed int v11;
    register vector signed int v12;
    register vector signed int v13;
    register vector signed int v14;
    register vector signed int v15;

    /* load (-infinity) to all four elements in v_lowscore */
    v_lowscore = vec_lde (0, &t_lowscore);
    mask1 = (vector unsigned int) vec_lvsl (0, &t_lowscore);
    v_lowscore =
        vec_perm (v_lowscore, v_lowscore, (vector unsigned char) mask1);
    v_lowscore = vec_splat (v_lowscore, 0);

    v1 = vec_splat_s32 (-1);
    v2 = vec_splat_s32 (0);
    mask1 = (vector unsigned int) vec_sld (v1, v2, 12);	/* FF in first pos, rest. are 00 */
    mask2 = vec_sld (mask1, mask1, 12);
    mask3 = vec_sld (mask1, mask1, 8);
    mask4 = vec_sld (mask1, mask1, 4);

    /* Make sure our DP matrix has 0..L rows, 0..M columns; grow it if needed.
     */
    ResizePlan7Matrix (mx, L, hmm->M, &xmx, &mmx, &imx, &dmx);

    /* Initialization of the zero row.
     */
    xmx[0][XMN] = 0;		/* S->N, p=1            */
    xmx[0][XMB] = hmm->xsc[XTN][MOVE];	/* S->N->B, no N-tail   */
    xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] = -INFTY;	/* need seq to get here */

    mmxi = mmx[0];
    imxi = imx[0];
    dmxi = dmx[0];
    xmxi = xmx[0];

    for (n = 0; n < 5 + hmm->M; n += 4) {
        vec_st (v_lowscore, n * 4, mmxi);
        vec_st (v_lowscore, n * 4, imxi);
        vec_st (v_lowscore, n * 4, dmxi);
    }

    /* Fill data beyound M with -INFTY, so we can take the maximum including
     * elements with k>M.
     */
    for (k = 1 + hmm->M; k <= (3 + hmm->M); k++) {
        hmm->esc[k] =-INFTY;
        hmm->bsc[k] =-INFTY;
        for (i = 0; i < 7; i++) {
            hmm->tsc[i][k] =-INFTY;
        }
        for (i = 0; i < MAXCODE; i++) {
            hmm->msc[i][k] =-INFTY;
            hmm->isc[i][k] =-INFTY;
        }
    }

    /* Recursion. Done as a pull
     * Note some slightly wasteful boundary conditions:
     *    tsc[0] = -INFTY for all eight transitions (no node 0)
     *    D_M and I_M are wastefully calculated (they don't exist)
     */

    for (i = 1; i <= L; i++) {
        /* pointers to last (i-1) row */
        lmmxi = mmxi;
        limxi = imxi;
        ldmxi = dmxi;
        lxmxi = xmxi;

        /* get pointers to this row */
        mmxi = mmx[i];
        imxi = imx[i];
        dmxi = dmx[i];
        xmxi = xmx[i];

        /* Set everything that doesnt depend on k here */

        /* load and splat (spread to all elements) XMX[i-1][XMB] */
        v13 = vec_lde (0, &(xmx[i - 1][XMB]));
        v14 = (vector signed int) vec_lvsl (0, &(xmx[i - 1][XMB]));
        v13 = vec_perm (v13, v13, (vector unsigned char) v14);
        v_xmb = vec_splat (v13, 0);
        p_tmm = hmm->tsc[TMM];
        p_tim = hmm->tsc[TIM];
        p_tdm = hmm->tsc[TDM];
        p_bsc = hmm->bsc;
        k = (int) dsq[i];
        p_msc = hmm->msc[k];
        p_isc = hmm->isc[k];
        p_tmd = hmm->tsc[TMD];
        p_tdd = hmm->tsc[TDD];
        p_tmi = hmm->tsc[TMI];
        p_tii = hmm->tsc[TII];
        p_esc = hmm->esc;
        max_mmxesc = v_lowscore;

        /* the 0 element of vectors are aligned 12 bytes up from the 16-byte boundary,
         * so we simply write the entire 16 bytes before the 16 byte boundary.
         */
        vec_st (v_lowscore, 0, mmxi);
        vec_st (v_lowscore, 0, imxi);
        vec_st (v_lowscore, 0, dmxi);

        /* Load the first (i.e. 'previous') vector on last row for mmx,imx,dmx,
         * and load the first vector of dmx and mmx on this row.
         */
        v_save_lmmx = vec_ld (-12, lmmxi);
        v_save_limx = vec_ld (-12, limxi);
        v_save_ldmx = vec_ld (-12, ldmxi);
        v_save_mmx = vec_ld (-12, mmxi);
        v_save_dmx = vec_ld (-12, dmxi);

        /* we have allocated extra memory, so it is perfectly OK
         * to do the calculations for a couple of extra cells where
         * k>hmm->M. These cells just wont be used.
         */
        for (n = 4, k = 1; k < (hmm->M - 3); n += 32, k += 8) {
            /* match state */

            /* 1: check which of mmx[i-1][k-1]+TMM and imx[i-1][k-1]+TIM is better,
             * but do it for 8 elements in parallel.
             * Since we are comparing with data on the previous row but one position
             * earlier, we have to shift stuff. Load two new vectors each round,
             * and use the saved register from last round.
             */
            /* load mmx data */
            v_lmmx1 = vec_ld (n, lmmxi);
            v_lmmx2 = vec_ld (n + 16, lmmxi);

            /* load imx data */
            v_limx1 = vec_ld (n, limxi);
            v_limx2 = vec_ld (n + 16, limxi);

            v5 = vec_ld (n, ldmxi);	/* Load dmx data */
            v10 = vec_ld (n + 16, ldmxi);

            /* shift mmx, imx & dmx data */
            v1 = vec_sld (v_save_lmmx, v_lmmx1, 12);
            v3 = vec_sld (v_save_limx, v_limx1, 12);
            v9 = vec_sld (v_save_ldmx, v5, 12);

            /* shift mmx, imx & dmx data */
            v2 = vec_sld (v_lmmx1, v_lmmx2, 12);
            v4 = vec_sld (v_limx1, v_limx2, 12);
            v_save_ldmx = v10;
            v10 = vec_sld (v5, v10, 12);

            v_save_lmmx = v_lmmx2;
            v_save_limx = v_limx2;

            /* v1,v2 now contains 8 element with mmx[i-1][k-1],
             * v3,v4 contain 8 elements with imx[i-1][k-1],
             * and v9,v10 contain 8 elements with dmx[i-1][k-1].
             */
            /* load TMM, TIM & TDM entries from the HMM - these are aligned in memory */
            v5 = vec_ld (n - 4, p_tmm);
            v6 = vec_ld (n + 12, p_tmm);
            v7 = vec_ld (n - 4, p_tim);
            v8 = vec_ld (n + 12, p_tim);
            v11 = vec_ld (n - 4, p_tdm);
            v12 = vec_ld (n + 12, p_tdm);
            /* load bsc[k] */
            v14 = vec_ld (n, p_bsc);
            v15 = vec_ld (n + 16, p_bsc);

            /* calc mmx+TMM, imx+TIM, dmx+TDM and XMX+bsc with saturated arithmetic, so
             * we don't loop if we add the large negative numbers used for -infinity.
             */
            v1 = vec_adds (v1, v5);
            v2 = vec_adds (v2, v6);
            v3 = vec_adds (v3, v7);
            v4 = vec_adds (v4, v8);
            v9 = vec_adds (v9, v11);
            v10 = vec_adds (v10, v12);
            v14 = vec_adds (v14, v_xmb);
            v15 = vec_adds (v15, v_xmb);
            /* Select max of mmx+TMM and imx+TIM in each element */
            v1 = vec_max (v1, v3);
            v2 = vec_max (v2, v4);
            /* select max of dmx+TDM and XMX+bsc */
            v9 = vec_max (v9, v14);
            v10 = vec_max (v10, v15);
            /* select max of the four alternatives */
            v1 = vec_max (v1, v9);
            v2 = vec_max (v2, v10);
            /* v1,v2 now contain the max values for the new mmx;
             * check if we should add msc.
             */

            v3 = vec_ld (n, p_msc);
            v4 = vec_ld (n + 16, p_msc);

            v5 = (vector signed int) vec_cmpgt (v3, v_lowscore);
            v6 = (vector signed int) vec_cmpgt (v4, v_lowscore);

            /* load esc[k] */
            v9 = vec_ld (n, p_esc);
            v10 = vec_ld (n + 16, p_esc);

            v1 = vec_adds (v1, v3);
            v2 = vec_adds (v2, v4);
            v1 = vec_sel (v3, v1, (vector unsigned int) v5);
            v2 = vec_sel (v4, v2, (vector unsigned int) v6);

            /* have final values for mmx on this row in v1,v2 - store it */
            vec_st (v1, n, mmxi);
            vec_st (v2, n + 16, mmxi);

            v9 = vec_adds (v9, v1);
            v10 = vec_adds (v10, v2);
            v9 = vec_max (v9, v10);
            max_mmxesc = vec_max (v9, max_mmxesc);

            /* OK. We finished the match state. Normally we could now first
             * do the delete and then the insertion. The problem is that the
             * delete is a pain in the ass to vectorize, since each element
             * depends on the previous one in the vector. This means I have
             * to write this relatively simple operation as eight independent
             * ones, just as we would have done in a non-vectorized code.
             * Since this is independent of the insertion state changes, I
             * try to hide the latencies by doing the delete and insert
             * calculations in parallel.
             * To make things easier I add 'del' in a comment on each
             * line for calculations that are on the delete state, and 'ins'
             * for the calculations for the insertion. Hang on...
             */

            /* We already have the match data on this row from the previous
             * iteration in v_save_mmx. Rotate it so the element that used to be
             * is pos 4 last iteration is in position 1, and pos 2-4 contain
             * the the first three elements of mmx this iteration.
             * And do the same type of rotation for v1/v2...
             */

            v4 = vec_sld (v1, v2, 12);	/* del */
            v3 = vec_sld (v_save_mmx, v1, 12);	/* del */
            v_save_mmx = v2;		/* Save for next iteration */

            /* Rotate last dmx data so we have the fourth element in pos 1. */
            v_save_dmx = vec_sld (v_save_dmx, v_save_dmx, 12);	/* del */

            /* load TMD & TDD data */
            v5 = vec_ld (n - 4, p_tmd);	/* del */
            v6 = vec_ld (n + 12, p_tmd);	/* del */
            v7 = vec_ld (n - 4, p_tdd);	/* del */
            v8 = vec_ld (n + 12, p_tdd);	/* del */

            /* calculate mmx+TMD */
            v3 = vec_adds (v3, v5);	/* del */
            v4 = vec_adds (v4, v6);	/* del */

            /* Start the ugly stuff. We have rotated last dmx data. Add TDD to
             * it and compare with v3/v4 (data from mmx+TMD alternative), but
             * we only compare & replace the first position, so we use a mask!
             */

            /* First position: Add TDD to v_save_dmx */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            /* Select max of this and mmx+TMD and save it temporary to v_save_dmx */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            /* use a mask to select only the first element from v_save_dmx, rest from v3. */
            v3 = vec_sel (v3, v_save_dmx, mask1);	/* del */

            /* Start doing the insertion calculations in parallel.
             * Load the TMI data.
             */
            v9 = vec_ld (n, p_tmi);	/* ins */
            v10 = vec_ld (n + 16, p_tmi);	/* ins */
            /* Deletion:
             * Now we have an accurate pos 1 in v3. continue with pos2 the same
             * way. Rotate to a temporary array, add to TDD, compare and use a mask
             * to write back to only position 2.
             */
            v_save_dmx = vec_sld (v3, v3, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            v3 = vec_sel (v3, v_save_dmx, mask2);	/* del */

            /* More insertion stuff - load TII data */
            v11 = vec_ld (n, p_tii);	/* ins */
            v12 = vec_ld (n + 16, p_tii);	/* ins */

            /* Deletion, position 3... */
            v_save_dmx = vec_sld (v3, v3, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            v3 = vec_sel (v3, v_save_dmx, mask3);	/* del */

            /* insertion stuff: calculate mmx+TMI */
            v9 = vec_adds (v_lmmx1, v9);	/* ins */
            v10 = vec_adds (v_lmmx2, v10);	/* ins */

            /* Deletion, position 4 */
            v_save_dmx = vec_sld (v3, v3, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            v3 = vec_sel (v3, v_save_dmx, mask4);	/* del */

            /* insertion stuff: calculate imx+TII */
            v11 = vec_adds (v_limx1, v11);	/* ins */
            v12 = vec_adds (v_limx2, v12);	/* ins */

            /* That was the first deletion vector, but we are unrolling.
             * The next step is position '5', i.e. the first in vector 2.
             * This one depends on the last position of vector 1, which we just finished.
             */
            v_save_dmx = vec_sld (v3, v3, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v8);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v4);	/* del */
            v4 = vec_sel (v4, v_save_dmx, mask1);	/* del */

            /* insertion stuff: select max of mmx+TMI and imx+TII */
            v9 = vec_max (v9, v11);	/* ins */
            v10 = vec_max (v10, v12);	/* ins */
            /* insertion stuff: load data from hmm->isc[tmpidx] */
            v11 = vec_ld (n, p_isc);	/* ins */
            v12 = vec_ld (n + 16, p_isc);	/* ins */

            /* position 6 (2 in vector 2) */
            v_save_dmx = vec_sld (v4, v4, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v8);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v4);	/* del */
            v4 = vec_sel (v4, v_save_dmx, mask2);	/* del */

            /* insertion: compare max of mmx+TMI, imx+TII with hmm->isc */
            v13 = (vector signed int) vec_cmpgt (v11, v_lowscore);
            v14 = (vector signed int) vec_cmpgt (v12, v_lowscore);

            /* position 7 (3 in vector 2) */
            v_save_dmx = vec_sld (v4, v4, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v8);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v4);	/* del */
            v4 = vec_sel (v4, v_save_dmx, mask3);	/* del */

            v9 = vec_adds (v9, v11);
            v10 = vec_adds (v10, v12);
            v9 = vec_sel (v11, v9, (vector unsigned int) v13);
            v10 = vec_sel (v12, v10, (vector unsigned int) v14);

            /* position 8 (4 in vector 2) */
            v_save_dmx = vec_sld (v4, v4, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v8);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v4);	/* del */
            v_save_dmx = vec_sel (v4, v_save_dmx, mask4);	/* del */

            /* Puh! that was the deletion... v3/v_save_dmx now contain the updated dmx data;
             * save it to memory. (v_save_dmx will be used next iteration)
             */
            vec_st (v3, n, dmxi);	/* del */
            vec_st (v_save_dmx, n + 16, dmxi);	/* del */

            /* save insertion too */
            vec_st (v9, n, imxi);
            vec_st (v10, n + 16, imxi);

        }
        /* odd loop */
        if (k < (1 + hmm->M)) {
            /* match state */
            /* 1: check which of mmx[i-1][k-1]+TMM and imx[i-1][k-1]+TIM is better,
             * but do it for 4 elements in parallel.
             * Since we are comparing with data on the previous row but one position
             * earlier, we have to shift stuff. Load two new vectors each round,
             * and use the saved register from last round.
             */
            /* load mmx data */
            v_lmmx1 = vec_ld (n, lmmxi);

            /* load imx data */
            v_limx1 = vec_ld (n, limxi);

            v5 = vec_ld (n, ldmxi);	/* Load dmx data */

            /* shift mmx, imx & dmx data */
            v1 = vec_sld (v_save_lmmx, v_lmmx1, 12);
            v3 = vec_sld (v_save_limx, v_limx1, 12);
            v9 = vec_sld (v_save_ldmx, v5, 12);

            /* v1,v2 now contains 8 element with mmx[i-1][k-1],
             * v3,v4 contain 8 elements with imx[i-1][k-1],
             * and v9,v10 contain 8 elements with dmx[i-1][k-1].
             */
            /* load TMM, TIM & TDM entries from the HMM - these are aligned in memory */
            v5 = vec_ld (n - 4, p_tmm);
            v7 = vec_ld (n - 4, p_tim);
            v11 = vec_ld (n - 4, p_tdm);
            /* load bsc[k] */
            v14 = vec_ld (n, p_bsc);

            /* calc mmx+TMM, imx+TIM, dmx+TDM and XMX+bsc with saturated arithmetic, so
             * we don't loop if we add the large negative numbers used for -infinity.
             */
            v1 = vec_adds (v1, v5);
            v3 = vec_adds (v3, v7);
            v9 = vec_adds (v9, v11);
            v14 = vec_adds (v14, v_xmb);
            /* Select max of mmx+TMM and imx+TIM in each element */
            v1 = vec_max (v1, v3);
            /* select max of dmx+TDM and XMX+bsc */
            v9 = vec_max (v9, v14);
            /* select max of the four alternatives */
            v1 = vec_max (v1, v9);
            /* v1,v2 now contain the max values for the new mmx;
             * check if we should add msc.
             */

            v3 = vec_ld (n, p_msc);

            v5 = (vector signed int) vec_cmpgt (v3, v_lowscore);

            /* load esc[k] */
            v9 = vec_ld (n, p_esc);

            v1 = vec_adds (v1, v3);
            v1 = vec_sel (v3, v1, (vector unsigned int) v5);

            /* have final values for mmx on this row in v1,v2 - store it */
            vec_st (v1, n, mmxi);

            v9 = vec_adds (v9, v1);
            max_mmxesc = vec_max (v9, max_mmxesc);

            /* OK. We finished the match state. Normally we could now first
             * do the delete and then the insertion. The problem is that the
             * delete is a pain in the ass to vectorize, since each element
             * depends on the previous one in the vector. This means I have
             * to write this relatively simple operation as eight independent
             * ones, just as we would have done in a non-vectorized code.
             * Since this is independent of the insertion state changes, I
             * try to hide the latencies by doing the delete and insert
             * calculations in parallel.
             * To make things easier I add 'del' in a comment on each
             * line for calculations that are on the delete state, and 'ins'
             * for the calculations for the insertion. Hang on...
             */

            /* We already have the match data on this row from the previous
             * iteration in v_save_mmx. Rotate it so the element that used to be
             * is pos 4 last iteration is in position 1, and pos 2-4 contain
             * the the first three elements of mmx this iteration.
             * And do the same type of rotation for v1/v2...
             */

            v3 = vec_sld (v_save_mmx, v1, 12);	/* del */

            /* Rotate last dmx data so we have the fourth element in pos 1. */
            v_save_dmx = vec_sld (v_save_dmx, v_save_dmx, 12);	/* del */

            /* load TMD & TDD data */
            v5 = vec_ld (n - 4, p_tmd);	/* del */
            v7 = vec_ld (n - 4, p_tdd);	/* del */

            /* calculate mmx+TMD */
            v3 = vec_adds (v3, v5);	/* del */

            /* Start the ugly stuff. We have rotated last dmx data. Add TDD to
             * it and compare with v3/v4 (data from mmx+TMD alternative), but
             * we only compare & replace the first position, so we use a mask!
             */

            /* First position: Add TDD to v_save_dmx */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            /* Select max of this and mmx+TMD and save it temporary to v_save_dmx */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            /* use a mask to select only the first element from v_save_dmx, rest from v3. */
            v3 = vec_sel (v3, v_save_dmx, mask1);	/* del */

            /* Start doing the insertion calculations in parallel.
             * Load the TMI data.
             */
            v9 = vec_ld (n, p_tmi);	/* ins */
            /* Deletion:
             * Now we have an accurate pos 1 in v3. continue with pos2 the same
             * way. Rotate to a temporary array, add to TDD, compare and use a mask
             * to write back to only position 2.
             */
            v_save_dmx = vec_sld (v3, v3, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            v3 = vec_sel (v3, v_save_dmx, mask2);	/* del */

            /* More insertion stuff - load TII data */
            v11 = vec_ld (n, p_tii);	/* ins */

            /* Deletion, position 3... */
            v_save_dmx = vec_sld (v3, v3, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            v3 = vec_sel (v3, v_save_dmx, mask3);	/* del */

            /* insertion stuff: calculate mmx+TMI */
            v9 = vec_adds (v_lmmx1, v9);	/* ins */

            /* Deletion, position 4 */
            v_save_dmx = vec_sld (v3, v3, 12);	/* del */
            v_save_dmx = vec_adds (v_save_dmx, v7);	/* del */
            v_save_dmx = vec_max (v_save_dmx, v3);	/* del */
            v3 = vec_sel (v3, v_save_dmx, mask4);	/* del */

            /* insertion stuff: calculate imx+TII */
            v11 = vec_adds (v_limx1, v11);	/* ins */

            /* insertion stuff: select max of mmx+TMI and imx+TII */
            v9 = vec_max (v9, v11);	/* ins */
            /* insertion stuff: load data from hmm->isc[tmpidx] */
            v11 = vec_ld (n, p_isc);	/* ins */

            /* insertion: compare max of mmx+TMI, imx+TII with hmm->isc */
            v13 = (vector signed int) vec_cmpgt (v11, v_lowscore);

            v9 = vec_adds (v9, v11);
            v9 = vec_sel (v11, v9, (vector unsigned int) v13);

            /* Puh! that was the deletion... v3/v_save_dmx now contain the updated dmx data;
             * save it to memory. (v_save_dmx will be used next iteration)
             */
            vec_st (v3, n, dmxi);	/* del */

            /* save insertion too */
            vec_st (v9, n, imxi);
        }
        /* end of k loops */

        /* Now the special states. Order is important here.
         * remember, C and J emissions are zero score by definition,
         */
        /* N state */
        xmx[i][XMN] = -INFTY;
        if ((sc = xmx[i - 1][XMN] + hmm->xsc[XTN][LOOP]) > -INFTY) {
            xmx[i][XMN] = sc;
        }

        /* E state */
        v2 = vec_sld (max_mmxesc, max_mmxesc, 8);
        v2 = vec_max (v2, max_mmxesc);
        v1 = vec_sld (v2, v2, 4);
        v1 = vec_max (v1, v2);
        vec_ste (v1, XME * 4, xmxi);

        /* J state */
        xmxi[XMJ] = -INFTY;
        if ((sc = lxmxi[XMJ] + hmm->xsc[XTJ][LOOP]) > -INFTY) {
            xmxi[XMJ] = sc;
        }
        if ((sc = xmxi[XME] + hmm->xsc[XTE][LOOP]) > xmxi[XMJ]) {
            xmxi[XMJ] = sc;
        }

        /* B state */
        xmxi[XMB] = -INFTY;
        if ((sc = xmxi[XMN] + hmm->xsc[XTN][MOVE]) > -INFTY) {
            xmxi[XMB] = sc;
        }
        if ((sc = xmxi[XMJ] + hmm->xsc[XTJ][MOVE]) > xmxi[XMB]) {
            xmxi[XMB] = sc;
        }

        /* C state */
        xmxi[XMC] = -INFTY;
        if ((sc = lxmxi[XMC] + hmm->xsc[XTC][LOOP]) > -INFTY) {
            xmxi[XMC] = sc;
        }
        if ((sc = xmxi[XME] + hmm->xsc[XTE][MOVE]) > xmxi[XMC]) {
            xmxi[XMC] = sc;
        }
    }
    /* T state (not stored) */
    sc = xmx[L][XMC] + hmm->xsc[XTC][MOVE];

    if (ret_tr != NULL) {
        P7ViterbiTrace (hmm, dsq, L, mx, &tr);
        *ret_tr = tr;
    }

    /* Note (Lindahl): we do NOT free the dpmatrix here anymore - the code was
     * spending 30% of the runtime allocating/freeing memory.
     * Provide a pointer to a dpmatrix_s structure to this routine,
     * and we try to reuse it. After the final call to P7Viterbi,
     * free it with FreePlan7Matrix.
     */
    return Scorify (sc);		/* the total Viterbi score. */
}
#endif /*the ALTIVEC port */
#endif /* do not allow ALTIVEC stuff for SPEC port Kaivalya 12/02/03 */
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/


/* file.c
 * SRE, Wed Jun 19 11:19:22 1996
 *
 * File operation utilities, dealing with pathnames, directories,
 * and environment variables.
 *
 * The goal is to have these be platform-independent but they
 * currently are UNIX-specific: i.e. this file is currently POSIX compliant
 * but it is NOT ANSI C compliant. (The sole offender is getenv().)
 *
 * CVS $Id: file.c,v 1.9 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "squid.h"
#include "sqfuncs.h"

/*
 * VMS:    #define DIRSLASH ']'
 * MacOS:  #define DIRSLASH ':'
 * DOS:    #define DIRSLASH '\\'
 *
 * The code assumes that '.' is used for file name extensions,
 * such as "foo.bar".
 */
#define DIRSLASH '/'		/* UNIX directory paths have /foo/bar */



/* Function: FileDirname()
 *
 * Purpose:  Returns the path from a filename:
 *             "/foo/bar/baz"  -> "/foo/bar"
 *             "foo/bar"       -> "foo"
 *             "foo"           -> "."
 *             "/"             -> "/"
 *           i.e. the string will be non-NULL; it will
 *           contain the string up to but not including the
 *           last '/' character; returns "." if
 *           there are no '/' characters, and returns "/"
 *           if the last slash is the first character.
 *           Modeled on Tcl's "file dirname" command.
 *
 * Args:     file - name of file "/foo/bar/baz".
 *
 * Return:   ptr to malloc'ed string "/foo/bar".
 */
char *
FileDirname (char *file)
{
    char *dirname;
    char *lastslash;
    int len;

    lastslash = strrchr (file, DIRSLASH);
    len = (lastslash == NULL) ? 0 : (int) (lastslash - file);
    dirname = (char *) MallocOrDie (sizeof (char) * (len + 2));
    if (len > 0) {
        strncpy (dirname, file, len);
    }
    else if (*file != DIRSLASH) {
        *dirname = '.';
        len = 1;
    }
    else {
        *dirname = DIRSLASH;
        len = 1;
    }
    dirname[len] = '\0';
    return dirname;
}


/* Function: FileTail()
 *
 * Purpose:  Return everything after the DIRSLASH:
 *             "/foo/bar/baz.1"  -> "baz.1"
 *             "foo/bar"         -> "bar"
 *             "foo"             -> "foo"
 *             "/"               -> ""
 *           If noextension is TRUE, removes a trailing ".foo" extension
 *           too.
 *
 * Args:     file        - name of file "/foo/bar/baz.1"
 *           noextension - TRUE to also remove extensions
 *
 * Return:   ptr to malloc'ed string "baz.1"
 */
char *
FileTail (char *file, int noextension)
{
    char *tail;
    char *lastslash;
    char *lastdot;
    /* remove directory prefix */
    lastslash = strrchr (file, DIRSLASH);
    tail = (char *) MallocOrDie (sizeof (char) * (strlen (file) + 1));
    if (lastslash == NULL) {
        strcpy (tail, file);
    }
    else {
        strcpy (tail, lastslash + 1);
    }
    /* remove trailing suffix */
    if (noextension) {
        if ((lastdot = strrchr (tail, '.')) != NULL) {
            *lastdot = '\0';
        }
    }

    return tail;
}


/* Function: FileSameDirectory()
 * Date:     SRE, Wed Mar  6 20:03:23 2002 [St. Louis]
 *
 * Purpose:  Given a path to one file, and the
 *           name of another file in the same directory,
 *           concat the path from file1 onto file2, and
 *           return the result. Caller must free the ptr
 *           that's returned.
 *
 *           Written for SSI - SSI indices contain filenames
 *           without paths, and we will need to convert that
 *           to a full path.
 *
 * Args:     file1 - a path to a file, e.g. "/foo/bar/baz.1"
 *           file2 - a simple filename, e.g. "quux.2"
 *
 * Returns:  path to file2: e.g. "/foo/bar/quux.2"
 *           Returns NULL if file2 already has a path, and the result
 *           would be a different place.
 */
char *
FileSameDirectory (char *file1, char *file2)
{
    char *path;
    char *tail;
    char *result;
    int seems_ok = 1;

    path = FileDirname (file1);
    tail = FileTail (file2, FALSE);
    if (strcmp (file2, tail) != 0) {
        seems_ok = 0;    /* ut-oh, file2 *had* a path */
    }
    result = FileConcat (path, tail);
    if (!seems_ok && strcmp (result, file2) != 0) {
        free (result);
        result = NULL;
    }
    free (path);
    free (tail);
    return result;
}

/* Function: FileConcat()
 *
 * Purpose:  Concatenate a directory path and a file name,
 *           returning a pointer to a malloc'ed string with the
 *           full filename. This isn't just a string concat,
 *           because we're careful about the dir slash.
 */
char *
FileConcat (char *dir, char *file)
{
    char *full;

    full =
        (char *) MallocOrDie (sizeof (char) * (strlen (dir) + strlen (file) + 2));
    if (*file == DIRSLASH) {
        strcpy (full, file);    /* file = "/foo", ignore directory. */
    }
    else {
        sprintf (full, "%s%c%s", dir, DIRSLASH, file);
    }
    return full;
}


/* Function: FileAddSuffix()
 * Date:     SRE, Wed Aug  1 11:19:33 2001 [Pasadena]
 *
 * Purpose:  Add a suffix to a filename, return a malloc'ed
 *           string containing the new filename.sfx name.
 *           Example:
 *             FileAddSuffix("genbank", "ssi")
 *           returns "genbank.ssi".
 */
char *
FileAddSuffix (char *filename, char *sfx)
{
    char *new;
    new = MallocOrDie (strlen (filename) + strlen (sfx) + 2);
    sprintf (new, "%s.%s", filename, sfx);
    return new;
}

/* Function: EnvFileOpen()
 * Date:     Sun Feb 12 10:55:29 1995
 *
 * Purpose:  Open a file, given a file name and an environment
 *           variable that contains a directory path. Files
 *           are opened read-only. Does not look at current directory
 *           unless "." is explicitly in the path specified by env.
 *
 *           For instance:
 *             fp = EnvFileOpen("BLOSUM45", "BLASTMAT", NULL);
 *           or:
 *             fp = EnvFileOpen("swiss", "BLASTDB", NULL);
 *
 *           Environment variables may contain a colon-delimited
 *           list of more than one path; e.g.
 *             setenv BLASTDB /nfs/databases/foo:/nfs/databases/bar
 *
 *           Sometimes a group of files may be found in
 *           one directory; for instance, an index file with a
 *           database. The caller can EnvFileOpen() the main
 *           file, and ask to get the name of the
 *           directory back in ret_dir, so it can construct
 *           the other auxiliary file names and fopen() them. (If it called
 *           EnvFileOpen(), it might get confused by
 *           file name clashes and open files in different
 *           directories.
 *
 * Args:     fname   - name of file to open
 *           env     - name of environment variable containing path
 *           ret_dir - if non-NULL, RETURN: name of dir that was used.
 *
 * Return:   FILE * to open file, or NULL on failure -- same as fopen()
 *           Caller must free ret_dir if it passed a non-NULL address.
 */
FILE *
EnvFileOpen (char *fname, char *env, char **ret_dir)
{
    FILE *fp;
    char *path;
    char *s;			/* ptr to indiv element in env list */
    char full[1024];		/* constructed file name */

    if (env == NULL) {
        return NULL;
    }
    if ((path = Strdup (getenv (env))) == NULL) {
        return NULL;
    }

    fp = NULL;
    s = strtok (path, ":");
    while (s != NULL) {
        if (((int) strlen (fname) + (int) strlen (s) + 2) > 1024) {
            free (path);
            return NULL;
        }
        sprintf (full, "%s%c%s", s, DIRSLASH, fname);
        if ((fp = fopen (full, "r")) != NULL) {
            break;
        }
        s = strtok (NULL, ":");
    }

    /* Return the path we used, if caller wants it
     */
    if (ret_dir != NULL) {
        *ret_dir = Strdup (s);
    }
    free (path);

    return fp;
}


/* Function: FileExists()
 *
 * Purpose:  Return TRUE if filename exists.
 *           Testing fopen() is the only possible platform-independent test
 *           I'm aware of.
 */
int
FileExists (char *filename)
{
    FILE *fp;
    if ((fp = fopen (filename, "r"))) {
        fclose (fp);
        return TRUE;
    }
    return FALSE;
}


/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* CVS $Id: getopt.c,v 1.8 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "squid.h"

/* Function: Getopt()
 *
 * Purpose:  Portable command line option parsing with abbreviated
 *           option switches. Replaces UNIX getopt(). Using UNIX getopt()
 *           hinders portability to non-UNIX platforms, and getopt()
 *           is also limited to single letter options.
 *
 *           Getopt() implements a superset of UNIX getopt().
 *           All of getopt()'s single-character switch behavior
 *           is emulated, and "--" by itself terminates the options.
 *           Additionally, Getopt() provides extended switches
 *           like "--youroptionhere", and Getopt() type checks
 *           arguments.
 *
 *           Extended options must start with "--", as in "--option1".
 *           Normal options must start with "-", as in "-o".
 *           Normal options may be concatenated, as in "-a -b" == "-ab".
 *
 *           See bottom of this .c file after #fdef GETOPT_TESTDRIVER
 *           for an example of calling Getopt().
 *
 * Args:     argc  - from main(). number of elems in argv.
 *           argv  - from main(). argv[0] is the name of the command.
 *           opt   - array of opt_s structures, defining option switches
 *           nopts - number of switches in opt
 *           usage - a (possibly long) string to print if usage error.
 *           ret_optind - RETURN: the index in argv[] of the next
 *                        valid command-line token.
 *           ret_optname- RETURN: ptr to the name of option switch
 *                        seen, or NULL if no option was seen.
 *           ret_optarg - RETURN: ptr to the optional argument, if any;
 *                        NULL if option takes no argument.
 *
 * Return:   1 if a valid option was parsed.
 *           0 if no option was found, and command-line parsing is complete.
 *           Die()'s here if an error is detected.
 */
int
Getopt (int argc, char **argv, struct opt_s *opt, int nopts, char *usage,
        int *ret_optind, char **ret_optname, char **ret_optarg)
{
    int i;
    int arglen;
    int nmatch;
    static int optind = 1;	/* init to 1 on first call  */
    static char *optptr = NULL;	/* ptr to next valid switch */
    int opti = 0;			/* init only to silence gcc uninit warnings */

    /* Check to see if we've run out of options.
     * A '-' by itself is an argument (e.g. "read from stdin")
     * not an option.
     */
    if (optind >= argc || argv[optind][0] != '-'
            || strcmp (argv[optind], "-") == 0) {
        *ret_optind = optind;
        *ret_optarg = NULL;
        *ret_optname = NULL;
        return 0;
    }

    /* Check to see if we're being told that this is the end
     * of the options with the special "--" flag.
     */
    if (strcmp (argv[optind], "--") == 0) {
        optind++;
        *ret_optind = optind;
        *ret_optname = NULL;
        *ret_optarg = NULL;
        return 0;
    }

    /* We have a real option. Find which one it is.
     * We handle single letter switches "-o" separately
     * from full switches "--option", based on the "-" vs. "--"
     * prefix -- single letter switches can be concatenated
     * as long as they don't have arguments.
     */
    /* full option */
    if (optptr == NULL && strncmp (argv[optind], "--", 2) == 0) {
        /* Use optptr to parse argument in options of form "--foo=666"
         */
        if ((optptr = strchr (argv[optind], '=')) != NULL) {
            *optptr = '\0';
            optptr++;
        }

        arglen = strlen (argv[optind]);
        nmatch = 0;
        for (i = 0; i < nopts; i++) {
            if (opt[i].single == FALSE &&
                    strncmp (opt[i].name, argv[optind], arglen) == 0) {
                nmatch++;
                opti = i;
                if (arglen == strlen (opt[i].name)) {
                    break;    /* exact match, stop now */
                }
            }
				}		
        if (nmatch > 1 && arglen != strlen (opt[i].name))
            Die ("Option \"%s\" is ambiguous; please be more specific.\n%s",
                 argv[optind], usage);
        if (nmatch == 0) {
            Die ("No such option \"%s\".\n%s", argv[optind], usage);
        }

        *ret_optname = opt[opti].name;

        /* Set the argument, if there is one
         */
        if (opt[opti].argtype != sqdARG_NONE) {
            if (optptr != NULL) {	/* --foo=666 style */
                *ret_optarg = optptr;
                optptr = NULL;
                optind++;
            }
            else if (optind + 1 >= argc) {
                Die ("Option %s requires an argument\n%s", opt[opti].name, usage);
            }
            else {			/* "--foo 666" style */

                *ret_optarg = argv[optind + 1];
                optind += 2;
            }
        }
        else {			/* sqdARG_NONE */

            if (optptr != NULL)
                Die ("Option %s does not take an argument\n%s", opt[opti].name,
                     usage);
            *ret_optarg = NULL;
            optind++;
        }
    }
    else {			/* else, a single letter option "-o" */

        /* find the option */
        if (optptr == NULL) {
            optptr = argv[optind] + 1;
        }
        for (opti = -1, i = 0; i < nopts; i++) {
            if (opt[i].single == TRUE && *optptr == opt[i].name[1]) {
                opti = i;
                break;
            }
				}		
        if (opti == -1) {
            Die ("No such option \"%c\".\n%s", *optptr, usage);
        }
        *ret_optname = opt[opti].name;

        /* set the argument, if there is one */
        if (opt[opti].argtype != sqdARG_NONE) {
            if (*(optptr + 1) != '\0') {	/* attached argument */
                *ret_optarg = optptr + 1;
                optind++;
            }
            else if (optind + 1 < argc) {	/* unattached argument */
                *ret_optarg = argv[optind + 1];
                optind += 2;
            }
            else {
                Die ("Option %s requires an argument\n%s", opt[opti].name, usage);
            }

            optptr = NULL;		/* can't concatenate after an argument */
        }
        else {			/* sqdARG_NONE */

            *ret_optarg = NULL;
            if (*(optptr + 1) != '\0') {	/* concatenation */
                optptr++;
            }
            else {
                optind++;		/* move to next field */
                optptr = NULL;
            }
        }

    }

    /* Type check the argument, if there is one
     */
    if (opt[opti].argtype != sqdARG_NONE) {
        if (opt[opti].argtype == sqdARG_INT && !IsInt (*ret_optarg))
            Die ("Option %s requires an integer argument\n%s",
                 opt[opti].name, usage);
        else if (opt[opti].argtype == sqdARG_FLOAT && !IsReal (*ret_optarg))
            Die ("Option %s requires a numerical argument\n%s",
                 opt[opti].name, usage);
        else if (opt[opti].argtype == sqdARG_CHAR && strlen (*ret_optarg) != 1)
            Die ("Option %s requires a single-character argument\n%s",
                 opt[opti].name, usage);
        /* sqdARG_STRING is always ok, no type check necessary */
    }

    *ret_optind = optind;
    return 1;
}



#ifdef GETOPT_TESTDRIVER
/* cc -DGETOPT_TESTDRIVER -L ~/lib/squid.linux/ getopt.c -lsquid
 */
struct opt_s OPTIONS[] = {
    {"--test1", FALSE, sqdARG_INT},
    {"--test2", FALSE, sqdARG_FLOAT},
    {"--test3", FALSE, sqdARG_STRING},
    {"--test4", FALSE, sqdARG_CHAR},
    {"-a", TRUE, sqdARG_NONE},
    {"-b", TRUE, sqdARG_INT},
};

#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))

int
main (int argc, char **argv)
{
    int optind;
    char *optarg;
    char *optname;

    while (Getopt (argc, argv, OPTIONS, NOPTIONS, "Usage/help here",
                   &optind, &optname, &optarg)) {
        printf ("Option:   index: %d name: %s argument: %s\n",
                optind, optname, optarg);
    }
    while (optind < argc) {
        printf ("Argument: index: %d name: %s\n", optind, argv[optind]);
        optind++;
    }


}


#endif /*GETOPT_TESTDRIVER */
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* gki.c
 * SRE, Sat May  1 14:49:08 1999
 *
 * "generic key index" module: emulation of Perl hashes.
 * Maps keys (ASCII char strings) to array index. Dynamically
 * resizes the hash table.
 *
 * Limitations:
 *     - hash table can only grow; no provision for deleting keys
 *       or downsizing the hash table.
 *     - Maximum hash table size set at 100003. Performance
 *       will degrade for key sets much larger than this.
 *     - Assumes that integers are 32 bits (or greater).
 *
 * Defines a typedef'd structure:
 *     gki           - a key index hash table.
 * Provides functions:
 *     GKIInit()     - start a hash table.
 *     GKIStoreKey() - store a new key, get a unique index.
 *     GKIKeyIndex() - retrieve an existing key's index.
 *     GKIFree()     - free a hash table.
 *     GKIStatus()   - Debugging: prints internal status of a hash struct
 *
 *
 * Note that there are no dependencies on squid; the gki.c/gki.h
 * pair are base ANSI C and can be reused anywhere.
 *****************************************************************
 *
 * API for storing/reading stuff:
 * moral equivalent of Perl's $foo{$key} = whatever, $bar{$key} = whatever:
 *       #include "gki.h"
 *
 *       gki  *hash;
 *       int   idx;
 *       char *key;
 *
 *       hash = GKIInit();
 * (Storing:)
 *       (foreach key) {
 *          idx = GKIStoreKey(hash, key);
 *          (reallocate foo, bar as needed)
 *          foo[idx] = whatever;
 *          bar[idx] = whatever;
 *       }
 * (Reading:)
 *       (foreach key) {
 *          idx = GKIKeyIndex(hash, key);
 *          if (idx == -1) {no_such_key; }
 *          (do something with) foo[idx];
 *          (do something with) bar[idx];
 *       }
 *       GKIFree();
 *
 *****************************************************************
 *
 * Timings on wrasse for 45402 keys in /usr/dict/words using
 * Tests/test_gki:
 *      250 msec store      (6 usec/store)
 *      140 msec retrieve   (3 usec/retrieve)
 * and using the 13408 names of Pfam's GP120.full alignment:
 *       70 msec store      (5 usec/store)
 *       50 msec retrieve   (4 usec/retrieve)
 *
 * CVS $Id: gki.c,v 1.4 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "squid.h"
#include "gki.h"

/*
 *   Best hash table sizes are prime numbers (see Knuth vol 3, Sorting
 * and Searching).
 *   gki_primes[] defines the ascending order of hash table sizes
 * that we use in upsizing the hash table dynamically.
 *   useful site for testing primes:
 * http://www.idbsu.edu/people/jbrennan/algebra/numbers/sieve.html
 *   Because of the way gki_hashvalue works, the largest number
 * must be < INT_MAX / 128 / 128 : 131072 on a 32 bit machine.
 */
static int gki_primes[] = { 101, 1009, 10007, 100003 };

#define GKI_NPRIMES      4
#define GKI_ALPHABETSIZE 128

static GKI *gki_alloc (int primelevel);
static int gki_hashvalue (GKI * hash, char *key);
static int gki_upsize (GKI * old);


/* Function: GKIInit()
 * Date:     SRE, Sat May  1 11:12:24 1999 [May Day geek-out]
 *
 * Purpose:  Initialize a hash table for key indexing.
 *           Simply a wrapper around a level 0 gki_alloc().
 *
 * Args:     (void)
 *
 * Returns:  An allocated hash table structure.
 *           Caller frees with GKIFree().
 */
GKI *
GKIInit (void)
{
    GKI *hash;
    hash = gki_alloc (0);
    return hash;
}

/* Function: GKIFree()
 * Date:     SRE, Sat May  1 11:13:26 1999 [May Day geek-out]
 *
 * Purpose:  Free a key index hash table.
 *
 * Args:     hash  - the gki structure
 *
 * Returns:  (void).
 *           hash table is destroyed.
 */
void
GKIFree (GKI * hash)
{
    struct gki_elem *ptr;
    int i;

    if (hash == NULL) {
        return;    /* tolerate a NULL */
    }

    for (i = 0; i < hash->nhash; i++) {
        while (hash->table[i] != NULL) {
            ptr = hash->table[i]->nxt;
            /* NULL keys can occur after we've gki_upsize'd */
            if (hash->table[i]->key != NULL) {
                free (hash->table[i]->key);
            }
            free (hash->table[i]);
            hash->table[i] = ptr;
        }
		}		
    free (hash->table);
    free (hash);
}

/* Function: GKIStoreKey()
 * Date:     SRE, Sat May  1 11:16:48 1999 [May Day geek-out]
 *
 * Purpose:  Store a key in the key index hash table.
 *           Associate it with a unique "key index", counting
 *           from 0. (It's this index that lets us map
 *           the hashed keys to indexed C arrays, (clumsily)
 *           emulating Perl's hashes.)
 *
 *           Does *not* check to see if the key's already
 *           in the table, so it's possible to store multiple
 *           copies of a key with different indices; probably
 *           not what you want, so if you're not sure the
 *           key is unique, check the table first with
 *           GKIKeyIndex().
 *
 * Args:     hash  - GKI structure to store the key in
 *           key   - string to store
 *
 * Returns:  the new key's index. Since it's always the
 *           last one in the current array, this index is
 *           just hash->nkeys-1.
 *           On a malloc failure, returns -1.
 *           hash table is modified.
 */
int
GKIStoreKey (GKI * hash, char *key)
{
    int val;
    struct gki_elem *ptr;

    val = gki_hashvalue (hash, key);

    ptr = hash->table[val];
    hash->table[val] = MallocOrDie (sizeof (struct gki_elem));
    hash->table[val]->key = MallocOrDie (sizeof (char) * (strlen (key) + 1));
    strcpy (hash->table[val]->key, key);

    hash->table[val]->idx = hash->nkeys;
    hash->table[val]->nxt = ptr;

    hash->nkeys++;
    /* time to upsize? */
    if (hash->nkeys > 3 * hash->nhash && hash->primelevel < GKI_NPRIMES - 1) {
        gki_upsize (hash);
    }

    return hash->nkeys - 1;
}

/* Function: GKIKeyIndex()
 * Date:     SRE, Sat May  1 11:20:42 1999 [May Day geek-out]
 *
 * Purpose:  Look up a key in the hash table. Return
 *           its index (0..nkeys-1), else -1 if the key
 *           isn't in the hash (yet).
 *
 * Args:     hash  - the GKI hash table to search in
 *           key   - the key to look up
 *
 * Returns:  -1 if key is not found;
 *           index of key if it is found (range 0..nkeys-1).
 *           hash table is unchanged.
 */
int
GKIKeyIndex (GKI * hash, char *key)
{
    struct gki_elem *ptr;
    int val;

    val = gki_hashvalue (hash, key);
    for (ptr = hash->table[val]; ptr != NULL; ptr = ptr->nxt) {
        if (strcmp (key, ptr->key) == 0) {
            return ptr->idx;
        }
		}		
    return -1;
}

/* Function: GKIStatus()
 * Date:     SRE, Sat May  1 11:11:13 1999 [St. Louis]
 *
 * Purpose:  (DEBUGGING) How are we doing? Calculate some
 *           simple statistics for the hash table.
 *
 * Args:     hash - the GKI hash table to look at
 *
 * Returns:  (void)
 *           Prints diagnostics on stdout.
 *           hash table is unchanged.
 */
void
GKIStatus (GKI * hash)
{
    struct gki_elem *ptr;
    int i;
    int nkeys;
    int nempty = 0;
    int maxkeys = -1;
    int minkeys = INT_MAX;

    for (i = 0; i < hash->nhash; i++) {
        nkeys = 0;
        for (ptr = hash->table[i]; ptr != NULL; ptr = ptr->nxt) {
            nkeys++;
        }

        if (nkeys == 0) {
            nempty++;
        }
        if (nkeys > maxkeys) {
            maxkeys = nkeys;
        }
        if (nkeys < minkeys) {
            minkeys = nkeys;
        }
    }

    printf ("Total keys:        %d\n", hash->nkeys);
    printf ("Hash table size:   %d\n", hash->nhash);
    printf ("Average occupancy: %.1f\n",
            (float) hash->nkeys / (float) hash->nhash);
    printf ("Unoccupied slots:  %d\n", nempty);
    printf ("Most in one slot:  %d\n", maxkeys);
    printf ("Least in one slot: %d\n", minkeys);

}


/* Function: gki_alloc()
 * Date:     SRE, Sat May  1 11:55:47 1999 [May Day geek-out]
 *
 * Purpose:  Allocate a hash table structure with the
 *           size given by primelevel.
 *
 * Args:     primelevel - level 0..GKI_NPRIMES-1, specifying
 *                        the size of the table; see gki_primes[]
 *                        array.
 *
 * Returns:  An allocated hash table structure.
 *           Caller frees with GKIFree().
 */
static GKI *
gki_alloc (int primelevel)
{
    GKI *hash;
    int i;

    if (primelevel < 0 || primelevel >= GKI_NPRIMES) {
        Die ("bad primelevel in gki_alloc()");
    }
    hash = MallocOrDie (sizeof (GKI));

    hash->primelevel = primelevel;
    hash->nhash = gki_primes[hash->primelevel];
    hash->table = MallocOrDie (sizeof (struct gki_elem) * hash->nhash);
    for (i = 0; i < hash->nhash; i++) {
        hash->table[i] = NULL;
    }
    hash->nkeys = 0;
    return hash;
}


/* Function: gki_hashvalue()
 * Date:     SRE, Sat May  1 11:14:10 1999 [May Day geek-out]
 *
 * Purpose:  Calculate the hash value for a key. Usually
 *           we expect a one-word key, but the function will
 *           hash any ASCII string effectively. The hash function
 *           is a simple one (see p. 233 of Sedgewick,
 *           Algorithms in C).
 *           Slightly optimized: does two characters at a time
 *           before doing the modulo; this gives us a significant
 *           speedup.
 *
 * Args:     hash - the gki structure (we need to know the hash table size)
 *           key  - a string to calculate the hash value for
 *
 * Returns:  a hash value, in the range 0..hash->nhash-1.
 *           hash table is unmodified.
 */
static int
gki_hashvalue (GKI * hash, char *key)
{
    int val = 0;

    for (; *key != '\0'; key++) {
        val = GKI_ALPHABETSIZE * val + *key;
        if (*(++key) == '\0') {
            val = val % hash->nhash;
            break;
        }
        val = (GKI_ALPHABETSIZE * val + *key) % hash->nhash;
    }
    return val;
}

/* Function: gki_upsize()
 * Date:     SRE, Sat May  1 11:46:07 1999 [May Day geek-out]
 *
 * Purpose:  Grow the hash table to the next available size.
 *
 * Args:     old - the GKI hash table to reallocate.
 *
 * Returns:  1 on success (the hash table is changed);
 *           0 on failure; the table is already at its maximum size,
 *              and the hash table is returned unchanged.
 */
static int
gki_upsize (GKI * old)
{
    GKI *new;
    int i;
    struct gki_elem *optr;
    struct gki_elem *nptr;
    int val;

    if (old->primelevel >= GKI_NPRIMES - 1) {
        return 0;
    }
    new = gki_alloc (old->primelevel + 1);

    /* Read the old, store in the new, while *not changing*
     * any key indices. Because of the way the lists are
     * treated as LIFO stacks, all the lists are reversed
     * in the new structure.
     */
    for (i = 0; i < old->nhash; i++) {
        optr = old->table[i];
        while (optr != NULL) {
            val = gki_hashvalue (new, optr->key);

            nptr = new->table[val];
            new->table[val] = optr;
            optr = optr->nxt;
            new->table[val]->nxt = nptr;
        }
    }
    free (old->table);

    /* Now swap within the interior of the structures, so the old
     * structure is updated to the new structure.
     * (nkeys is identical, so we don't need to swap that element.)
     */
    old->primelevel = new->primelevel;
    old->nhash = new->nhash;
    old->table = new->table;
    free (new);
    return 1;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* gsi.c
 * Interfaces for GSI "generic sequence index" files.
 * broken away from sqio.c and extended: SRE, Wed Aug  5 10:32:53 1998
 *
 *
 * GSI definition:
 *    1 + <nfiles> + <nkeys> total records.
 *    Each record = 38 bytes.
 *
 *  one header record     :  <"GSI"    (32)> <nfiles (2)> <nkeys (4)>
 *  <nfiles> file records :  <filename (32)> <fileno (2)> <fmt   (4)>
 *  <nkeys>  key records  :  <key      (32)> <fileno (2)> <offset(4)>
 *
 * Matches up with my Perl scripts that create GSI files.
 *
 * CVS $Id: gsi.c,v 1.6 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef SEEK_SET
#include <unistd.h>		/* needed for poor crippled SunOS */
#endif

#include "squid.h"
#include "gsi.h"


/*****************************************************************
 * GSI index file access routines
 *****************************************************************/

/* Function: GSIOpen()
 *
 * Purpose:  Open a GSI file. Returns the number of records in
 *           the file and a file pointer. Returns NULL on failure.
 *           The file pointer should be fclose()'d normally.
 */
GSIFILE *
GSIOpen (char *gsifile)
{
    GSIFILE *gsi;
    char magic[GSI_KEYSIZE];

    gsi = (GSIFILE *) MallocOrDie (sizeof (GSIFILE));
    if ((gsi->gsifp = fopen (gsifile, "r")) == NULL) {
        free (gsi);
        squid_errno = SQERR_NOFILE;
        return NULL;
    }

    if (!fread (magic, sizeof (char), GSI_KEYSIZE, gsi->gsifp)) {
        free (gsi);
        squid_errno = SQERR_NODATA;
        return NULL;
    }
    if (strcmp (magic, "GSI") != 0) {
        free (gsi);
        squid_errno = SQERR_FORMAT;
        return NULL;
    }

    if (!fread (&(gsi->nfiles), sizeof (sqd_uint16), 1, gsi->gsifp)) {
        free (gsi);
        squid_errno = SQERR_NODATA;
        return NULL;
    }
    if (!fread (&(gsi->recnum), sizeof (sqd_uint32), 1, gsi->gsifp)) {
        free (gsi);
        squid_errno = SQERR_NODATA;
        return NULL;
    }

    gsi->nfiles = sre_ntoh16 (gsi->nfiles);	/* convert from network short */
    gsi->recnum = sre_ntoh32 (gsi->recnum);	/* convert from network long  */

    return gsi;
}

/* Function: GSIGetRecord()
 *
 * Purpose:  Each non-header record of a GSI index files consists
 *           of 38 bytes: 32 bytes of character string, a 2 byte
 *           short, and a 4 byte long. This function returns the
 *           three values.
 *
 * Args:     gsi  - open GSI index file, correctly positioned at a record
 *           f1   - char[32], allocated by caller (or NULL if unwanted)
 *           f2   - pointer to short (or NULL if unwanted)
 *           f3   - pointer to long  (or NULL if unwanted)
 *
 * Return:   0 on failure and sets squid_errno.
 */
int
GSIGetRecord (GSIFILE * gsi, char *f1, sqd_uint16 * f2, sqd_uint32 * f3)
{
    if (f1 == NULL) {
        fseek (gsi->gsifp, GSI_KEYSIZE, SEEK_CUR);
    }
    else if (!fread (f1, GSI_KEYSIZE, 1, gsi->gsifp)) {
        squid_errno = SQERR_NODATA;
        return 0;
    }

    if (f2 == NULL) {
        fseek (gsi->gsifp, sizeof (sqd_uint16), SEEK_CUR);
    }
    else if (!fread (f2, sizeof (sqd_uint16), 1, gsi->gsifp)) {
        squid_errno = SQERR_NODATA;
        return 0;
    }

    if (f3 == NULL) {
        fseek (gsi->gsifp, sizeof (sqd_uint32), SEEK_CUR);
    }
    else if (!fread (f3, sizeof (sqd_uint32), 1, gsi->gsifp)) {
        squid_errno = SQERR_NODATA;
        return 0;
    }

    if (f2 != NULL) {
        *f2 = sre_ntoh16 (*f2);
    }
    if (f3 != NULL) {
        *f3 = sre_ntoh32 (*f3);
    }

    return 1;
}


/* Function: GSIGetOffset()
 *
 * Purpose:  From a key (sequence name), find a disk offset
 *           in an open general sequence index file by binary
 *           search. Presumably GSI indexing could be even faster
 *           if we used hashing.
 *
 * Args:     gsi         - GSI index file, opened by GSIOpen()
 *           key         - name of key to retrieve indices for
 *           ret_seqfile - pre-alloced char[32] array for seqfile name
 *           ret_fmt     - format of seqfile
 *           ret_offset  - return: disk offset in seqfile.
 */
int
GSIGetOffset (GSIFILE * gsi, char *key, char *ret_seqfile,
              int *ret_format, long *ret_offset)
{
    sqd_uint32 left, right, mid;
    int cmp;
    char name[GSI_KEYSIZE + 1];
    sqd_uint32 offset;
    sqd_uint16 filenum;
    sqd_uint32 fmt;

    name[GSI_KEYSIZE] = '\0';

    left = gsi->nfiles + 1;
    right = gsi->nfiles + gsi->recnum;
    mid = (left + right) / 2;
    fseek (gsi->gsifp, mid * GSI_RECSIZE, SEEK_SET);

    while (GSIGetRecord (gsi, name, &filenum, &offset)) {
        cmp = strcmp (name, key);
        if (cmp == 0) {
            break;    /* found it!              */
        }
        else if (left >= right) {
            return 0;    /* oops, missed it; fail. */
        }
        else if (cmp < 0) {
            left = mid + 1;    /* it's right of mid      */
        }
        else if (cmp > 0) {
            right = mid - 1;    /* it's left of mid       */
        }
        mid = (left + right) / 2;
        fseek (gsi->gsifp, mid * GSI_RECSIZE, SEEK_SET);
    }

    /* Using file number, look up the sequence file and format.
     */
    fseek (gsi->gsifp, filenum * GSI_RECSIZE, SEEK_SET);
    GSIGetRecord (gsi, ret_seqfile, NULL, &fmt);
    *ret_format = (int) fmt;
    *ret_offset = (long) offset;

    return 1;
}

/* Function: GSIClose()
 *
 * Purpose:  Close an open GSI sequence index file.
 */
void
GSIClose (GSIFILE * gsi)
{
    fclose (gsi->gsifp);
    free (gsi);
}


/*****************************************************************
 * GSI index construction routines
 * SRE, Wed Nov 10 11:49:14 1999 [St. Louis]
 *
 * API:
 *       g = GSIAllocIndex();
 *
 *       [foreach filename, <32 char, no directory path]
 *          GSIAddFileToIndex(g, filename);
 *          filenum++;
 *          [foreach key, <32 char, w/ filenum 1..nfiles, w/ 32bit offset]
 *             GSIAddKeyToIndex(g, key, filenum, offset);
 *
 *       GSISortIndex(g);
 *       GSIWriteIndex(fp, g);
 *       GSIFreeIndex(g);
 *****************************************************************/
struct gsiindex_s *
GSIAllocIndex (void)
{
    struct gsiindex_s *g;

    g = MallocOrDie (sizeof (struct gsiindex_s));
    g->filenames = MallocOrDie (sizeof (char *) * 10);
    g->fmt = MallocOrDie (sizeof (int) * 10);
    g->elems = MallocOrDie (sizeof (struct gsikey_s) * 100);
    g->nfiles = 0;
    g->nkeys = 0;
    return g;
}

void
GSIFreeIndex (struct gsiindex_s *g)
{
    int i;
    for (i = 0; i < g->nfiles; i++) {
        free (g->filenames[i]);
    }
    free (g->filenames);
    free (g->fmt);
    free (g->elems);
    free (g);
}

void
GSIAddFileToIndex (struct gsiindex_s *g, char *filename, int fmt)
{
    int len;

    len = strlen (filename);
    if (len >= GSI_KEYSIZE) {
        Die ("File name too long to be indexed.");
    }
    g->filenames[g->nfiles] = sre_strdup (filename, len);
    g->fmt[g->nfiles] = fmt;
    g->nfiles++;
    if (g->nfiles % 10 == 0) {
        g->filenames =
            ReallocOrDie (g->filenames, sizeof (char *) * (g->nfiles + 10));
        g->fmt = ReallocOrDie (g->fmt, sizeof (int) * (g->nfiles + 10));
    }
}

void
GSIAddKeyToIndex (struct gsiindex_s *g, char *key, int filenum, long offset)
{
    if (strlen (key) >= GSI_KEYSIZE) {
        Die ("key too long in GSI index");
    }
    if (filenum > SQD_UINT16_MAX) {
        Die ("too many files in GSI index");
    }
    if (offset > SQD_UINT32_MAX) {
        Die ("offset too big in GSI index");
    }

    strncpy (g->elems[g->nkeys].key, key, GSI_KEYSIZE - 1);
    g->elems[g->nkeys].key[GSI_KEYSIZE - 1] = '\0';
    g->elems[g->nkeys].filenum = (sqd_uint16) filenum;
    g->elems[g->nkeys].offset = (sqd_uint32) offset;
    g->nkeys++;

    if (g->nkeys % 100 == 0)
        g->elems =
            ReallocOrDie (g->elems, sizeof (struct gsikey_s) * (g->nkeys + 100));
}

static int
gsi_keysorter (const void *k1, const void *k2)
{
    struct gsikey_s *key1;
    struct gsikey_s *key2;
    key1 = (struct gsikey_s *) k1;
    key2 = (struct gsikey_s *) k2;
    return strcmp (key1->key, key2->key);
}

void
GSISortIndex (struct gsiindex_s *g)
{
    specqsort ((void *) g->elems, g->nkeys, sizeof (struct gsikey_s),
               gsi_keysorter);
}

void
GSIWriteIndex (FILE * fp, struct gsiindex_s *g)
{
    sqd_uint32 i;

    /* Range checking.
     */
    if (g->nfiles > SQD_UINT16_MAX) {
        Die ("Too many files in GSI index.");
    }
    if (g->nkeys > SQD_UINT32_MAX) {
        Die ("Too many keys in GSI index.");
    }

    GSIWriteHeader (fp, g->nfiles, g->nkeys);
    for (i = 0; i < g->nfiles; i++) {
        GSIWriteFileRecord (fp, g->filenames[i], i + 1, g->fmt[i]);
    }
    for (i = 0; i < g->nkeys; i++) {
        GSIWriteKeyRecord (fp, g->elems[i].key, g->elems[i].filenum, g->elems[i].offset);
		}		
}





/* Function: GSIWriteHeader()
 * Date:     SRE, Wed Aug  5 10:36:02 1998 [St. Louis]
 *
 * Purpose:  Write the first record to an open GSI file:
 *           "GSI" <nfiles> <nkeys>
 *
 * Args:     fp      - open file to write to.
 *           nfiles  - number of files indexed
 *           nkeys   - number of keys indexed
 *
 * Returns:  void
 */
void
GSIWriteHeader (FILE * fp, int nfiles, long nkeys)
{
    char key[GSI_KEYSIZE];
    sqd_uint16 f1;
    sqd_uint32 f2;

    /* beware potential range errors!
     */
    if (nfiles > SQD_UINT16_MAX) {
        Die ("GSI: nfiles out of range");
    }
    if (nkeys > SQD_UINT32_MAX) {
        Die ("GSI: nkeys out of range");
    }

    f1 = (sqd_uint16) nfiles;
    f2 = (sqd_uint32) nkeys;
    f1 = sre_hton16 (f1);
    f2 = sre_hton32 (f2);
    strcpy (key, "GSI");

    if (fwrite (key, 1, GSI_KEYSIZE, fp) < GSI_KEYSIZE) {
        PANIC;
    }
    if (fwrite (&f1, 2, 1, fp) < 1) {
        PANIC;
    }
    if (fwrite (&f2, 4, 1, fp) < 1) {
        PANIC;
    }
}


/* Function: GSIWriteFileRecord()
 * Date:     SRE, Wed Aug  5 10:45:51 1998 [St. Louis]
 *
 * Purpose:  Write a file record to an open GSI file.
 *
 * Args:     fp    - open GSI file
 *           fname - file name (max 31 characters)
 *           idx   - file number
 *           fmt   - file format (e.g. kPearson, etc.)
 *
 * Returns:  0 on failure. 1 on success.
 */
int
GSIWriteFileRecord (FILE * fp, char *fname, int idx, int fmt)
{
    sqd_uint16 f1;
    sqd_uint32 f2;

    if (strlen (fname) >= GSI_KEYSIZE) {
        return 0;
    }
    if (idx > SQD_UINT16_MAX) {
        Die ("GSI: file index out of range");
    }
    if (fmt > SQD_UINT32_MAX) {
        Die ("GSI: format index out of range");
    }

    f1 = (sqd_uint16) idx;
    f2 = (sqd_uint32) fmt;
    f1 = sre_hton16 (f1);
    f2 = sre_hton32 (f2);

    if (fwrite (fname, 1, GSI_KEYSIZE, fp) < GSI_KEYSIZE) {
        PANIC;
    }
    if (fwrite (&f1, 2, 1, fp) < 1) {
        PANIC;
    }
    if (fwrite (&f2, 4, 1, fp) < 1) {
        PANIC;
    }
    return 1;
}


/* Function: GSIWriteKeyRecord()
 * Date:     SRE, Wed Aug  5 10:52:30 1998 [St. Louis]
 *
 * Purpose:  Write a key record to a GSI file.
 *
 * Args:     fp      - open GSI file for writing
 *           key     - key (max 31 char + \0)
 *           fileidx - which file number to find this key in
 *           offset  - offset for this key
 *
 * Returns:  1 on success, else 0.
 *           will fail if key >= 32 chars, for instance.
 */
int
GSIWriteKeyRecord (FILE * fp, char *key, int fileidx, long offset)
{
    sqd_uint16 f1;
    sqd_uint32 f2;

    if (strlen (key) >= GSI_KEYSIZE) {
        return 0;
    }
    if (fileidx > SQD_UINT16_MAX) {
        Die ("GSI: file index out of range");
    }
    if (offset > SQD_UINT32_MAX) {
        Die ("GSI: offset out of range");
    }

    f1 = (sqd_uint16) fileidx;
    f2 = (sqd_uint32) offset;
    f1 = sre_hton16 (f1);
    f2 = sre_hton32 (f2);

    if (fwrite (key, 1, GSI_KEYSIZE, fp) < GSI_KEYSIZE) {
        PANIC;
    }
    if (fwrite (&f1, 2, 1, fp) < 1) {
        PANIC;
    }
    if (fwrite (&f2, 4, 1, fp) < 1) {
        PANIC;
    }
    return 1;
}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* histogram.c
 * SRE, Sat Jan 20 16:16:17 1996
 *
 * Accumulation, printing, and fitting of score histograms
 * from database searches.
 *
 * CVS $Id: histogram.c,v 1.18 2003/04/14 16:00:16 eddy Exp $
 ************************************************************
 * Basic API:
 *
 * struct histogram_s *h;
 *
 * h = AllocHistogram(min_hint, max_hint, lumpsize);
 *
 * while (getting scores x) AddToHistogram(h, x);
 *
 * ExtremeValueFitHistogram(h, high_hint);
 * PrintASCIIHistogram(fp, h);
 * FreeHistogram(h);
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>

#include "squid.h"
#include "structs.h"
#include "funcs.h"

/* Function: AllocHistogram()
 *
 * Purpose:  Allocate and return a histogram structure.
 *           min and max are your best guess. They need
 *           not be absolutely correct; the histogram
 *           will expand dynamically to accomodate scores
 *           that exceed these suggested bounds. The amount
 *           that the histogram grows by is set by "lumpsize".
 *
 * Args:     min:      minimum score (integer)
 *           max:      maximum score (integer)
 *           lumpsize: when reallocating histogram, pad the reallocation
 *                     by this much (saves excessive reallocation)
 */
struct histogram_s *
AllocHistogram (int min, int max, int lumpsize)
{
    struct histogram_s *h;
    int newsize;
    int i;

    newsize = max - min + 1;

    h = (struct histogram_s *) MallocOrDie (sizeof (struct histogram_s));
    h->min = min;
    h->max = max;
    h->total = 0;
    h->lowscore = INT_MAX;
    h->highscore = INT_MIN;
    h->lumpsize = lumpsize;
    h->histogram = (int *) MallocOrDie (sizeof (int) * newsize);
    for (i = 0; i < newsize; i++) {
        h->histogram[i] = 0;
    }

    h->expect = NULL;
    h->fit_type = HISTFIT_NONE;

    return h;
}


/* Function: FreeHistogram()
 *
 * Purpose:  free a histogram structure.
 */
void
FreeHistogram (struct histogram_s *h)
{
    free (h->histogram);
    if (h->expect != NULL) {
        free (h->expect);
    }
    free (h);
}

/* Function: UnfitHistogram()
 *
 * Purpose:  Free only the theoretical fit part of a histogram.
 */
void
UnfitHistogram (struct histogram_s *h)
{
    if (h->expect != NULL) {
        free (h->expect);
    }
    h->expect = NULL;
    h->fit_type = HISTFIT_NONE;
}


/* Function: AddToHistogram()
 *
 * Purpose:  Bump the appropriate counter in a histogram
 *           structure, given a score. The score is
 *           rounded off from float precision to the
 *           next lower integer.
 */
void
AddToHistogram (struct histogram_s *h, float sc)
{
    int score;
    int moveby;
    int prevsize;
    int newsize;
    int i;

    /* Adding to a histogram conflicts with existing fit:
     * prohibit this.
     */
    if (h->fit_type != HISTFIT_NONE) {
        Die ("AddToHistogram(): Can't add to a fitted histogram\n");
    }


    /* histogram bins are defined as:  score >= bin value, < bin+1
     * -1.9 -> -2    -0.4 -> -1    1.9 -> 1
     * -2.1 -> -3     0.4 -> 0     2.1 -> 2
     */
    score = (int) floor (sc);

    /* Check to see if we must reallocate the histogram.
     */
    if (score < h->min) {
        prevsize = h->max - h->min + 1;
        moveby = (h->min - score) + h->lumpsize;
        newsize = prevsize + moveby;
        h->min -= moveby;

        h->histogram =
            (int *) ReallocOrDie (h->histogram, sizeof (int) * newsize);
        memmove (h->histogram + moveby, h->histogram, sizeof (int) * prevsize);
        for (i = 0; i < moveby; i++) {
            h->histogram[i] = 0;
        }
    }
    else if (score > h->max) {
        prevsize = h->max - h->min + 1;
        h->max = h->lumpsize + score;
        newsize = h->max - h->min + 1;

        h->histogram =
            (int *) ReallocOrDie (h->histogram, sizeof (int) * newsize);
        for (i = prevsize; i < newsize; i++) {
            h->histogram[i] = 0;
        }
    }

    /* Bump the correct bin.
     * The bin number is score - h->min
     */
    h->histogram[score - h->min]++;
    h->total++;
    if (score < h->lowscore) {
        h->lowscore = score;
    }
    if (score > h->highscore) {
        h->highscore = score;
    }

    SQD_DPRINTF3 (("AddToHistogram(): added %.1f; rounded to %d; in bin %d (%d-%d)\n", sc, score, score - h->min, h->min, h->max));
    return;
}



/* Function: PrintASCIIHistogram()
 *
 * Purpose:  Print a "prettified" histogram to a file pointer.
 *           Deliberately a look-and-feel clone of Bill Pearson's
 *           excellent FASTA output.
 *
 * Args:     fp     - open file to print to (stdout works)
 *           h      - histogram to print
 */
void
PrintASCIIHistogram (FILE * fp, struct histogram_s *h)
{
    int units;
    int maxbar;
    int num;
    int i, idx;
    char buffer[81];		/* output line buffer */
    int pos;			/* position in output line buffer */
    int lowbound, lowcount;	/* cutoffs on the low side  */
    int highbound, highcount;	/* cutoffs on the high side */
    int emptybins = 3;

    /* Find out how we'll scale the histogram.
     * We have 59 characters to play with on a
     * standard 80-column terminal display:
     * leading "%5d %6d %6d|" occupies 20 chars.
     * Save the peak position, we'll use it later.
     */
    maxbar = 0;
    for (i = h->lowscore - h->min; i <= h->highscore - h->min; i++) {
        if (h->histogram[i] > maxbar) {
            maxbar = h->histogram[i];	/* max height    */
            lowbound = i + h->min;	/* peak position */
        }
		}
    /* Truncate histogram display on both sides, ad hoc fashion.
     * Start from the peak; then move out until we see <emptybins> empty bins,
     * and stop.
     */
    highbound = lowbound;		/* start at peak position */
    for (num = 0; lowbound > h->lowscore; lowbound--) {
        i = lowbound - h->min;
        if (h->histogram[i] > 0) {
            num = 0;
            continue;
        }				/* reset */
        if (++num == emptybins) {
            lowbound += emptybins;
            break;
        }				/* stop  */
    }
    for (num = 0; highbound < h->highscore; highbound++) {
        i = highbound - h->min;
        if (h->histogram[i] > 0) {
            num = 0;
            continue;
        }				/* reset */
        if (++num == emptybins) {
            highbound -= emptybins;
            break;
        }				/* stop  */
    }
    /* collect counts outside of bounds */
    for (lowcount = 0, i = h->lowscore - h->min; i <= lowbound - h->min; i++) {
        lowcount += h->histogram[i];
    }
    for (highcount = 0, i = h->highscore - h->min; i >= highbound - h->min; i--) {
        highcount += h->histogram[i];
    }

    /* maxbar might need raised now; then set our units  */
    if (lowcount > maxbar) {
        maxbar = lowcount;
    }
    if (highcount > maxbar) {
        maxbar = highcount;
    }
    units = ((maxbar - 1) / 59) + 1;


    /* Print the histogram
     */
    fprintf (fp, "%5s %6s %6s  (one = represents %d sequences)\n",
             "score", "obs", "exp", units);
    fprintf (fp, "%5s %6s %6s\n", "-----", "---", "---");
    buffer[80] = '\0';
    buffer[79] = '\n';
    for (i = h->lowscore; i <= h->highscore; i++) {
        memset (buffer, ' ', 79 * sizeof (char));
        idx = i - h->min;

        /* Deal with special cases at edges
         */
        if (i < lowbound) {
            continue;
        }
        else if (i > highbound) {
            continue;
        }
        else if (i == lowbound && i != h->lowscore) {
            sprintf (buffer, "<%4d %6d %6s|", i + 1, lowcount, "-");
            if (lowcount > 0) {
                num = 1 + (lowcount - 1) / units;
                if (num > 60) {
                    Die ("oops");
                }
                for (pos = 20; num > 0; num--) {
                    buffer[pos++] = '=';
                }
            }
            fputs (buffer, fp);
            continue;
        }
        else if (i == highbound && i != h->highscore) {
            sprintf (buffer, ">%4d %6d %6s|", i, highcount, "-");
            if (highcount > 0) {
                num = 1 + (highcount - 1) / units;
                for (pos = 20; num > 0; num--) {
                    buffer[pos++] = '=';
                }
            }
            fputs (buffer, fp);
            continue;
        }

        /* Deal with most cases
         */
        if (h->fit_type != HISTFIT_NONE)
            sprintf (buffer, "%5d %6d %6d|",
                     i, h->histogram[idx], (int) h->expect[idx]);
        else {
            sprintf (buffer, "%5d %6d %6s|", i, h->histogram[idx], "-");
        }
        buffer[20] = ' ';		/* sprintf writes a null char */

        /* Mark the histogram bar for observed hits
         */
        if (h->histogram[idx] > 0) {
            num = 1 + (h->histogram[idx] - 1) / units;
            for (pos = 20; num > 0; num--) {
                buffer[pos++] = '=';
            }
        }

        /* Mark the theoretically expected value
         */
        if (h->fit_type != HISTFIT_NONE && (int) h->expect[idx] > 0) {
            pos = 20 + (int) (h->expect[idx] - 1) / units;
            if (pos >= 78) {
                pos = 78;    /* be careful of buffer bounds */
            }
            buffer[pos] = '*';
        }

        /* Print the line
         */
        fputs (buffer, fp);
    }

    /* Print details about the statistics
     */
    switch (h->fit_type) {
    case HISTFIT_NONE:
        fprintf (fp, "\n\n%% No statistical fit available\n");
        break;

    case HISTFIT_EVD:
        fprintf (fp, "\n\n%% Statistical details of theoretical EVD fit:\n");
        fprintf (fp, "              mu = %10.4f\n", h->param[EVD_MU]);
        fprintf (fp, "          lambda = %10.4f\n", h->param[EVD_LAMBDA]);
        fprintf (fp, "chi-sq statistic = %10.4f\n", h->chisq);
        fprintf (fp, "  P(chi-square)  = %10.4g\n", h->chip);
        break;

    case HISTFIT_GAUSSIAN:
        fprintf (fp, "\n\n%% Statistical details of theoretical Gaussian fit:\n");
        fprintf (fp, "            mean = %10.4f\n", h->param[GAUSS_MEAN]);
        fprintf (fp, "              sd = %10.4f\n", h->param[GAUSS_SD]);
        fprintf (fp, "chi-sq statistic = %10.4f\n", h->chisq);
        fprintf (fp, "  P(chi-square)  = %10.4g\n", h->chip);
        break;
    }
    return;
}



/* Function: PrintXMGRHistogram()
 * Date:     SRE, Wed Nov 12 11:02:00 1997 [St. Louis]
 *
 * Purpose:  Print an XMGR data file that contains two data sets:
 *               - xy data for the observed histogram
 *               - xy data for the theoretical histogram
 */
void
PrintXMGRHistogram (FILE * fp, struct histogram_s *h)
{
    int sc;			/* integer score in histogram structure */
    double val;

    /* First data set is the observed histogram
     */
    for (sc = h->lowscore; sc <= h->highscore; sc++) {
        if (h->histogram[sc - h->min] > 0)
            fprintf (fp, "%-6d %f\n", sc,
                     (float) h->histogram[sc - h->min] / (float) h->total);
		}								 
    fprintf (fp, "&\n");

    /* Second data set is the theoretical histogram
     */
    if (h->fit_type != HISTFIT_NONE) {
        for (sc = h->lowscore; sc <= h->highscore; sc++) {
            val =
                (1. -
                 ExtremeValueP ((float) sc + 1, h->param[EVD_MU],
                                h->param[EVD_LAMBDA])) - (1. -
                                        ExtremeValueP ((float) sc,
                                                h->
                                                param
                                                [EVD_MU],
                                                h->
                                                param
                                                [EVD_LAMBDA]));
            fprintf (fp, "%-6d %f\n", sc, val);
        }
        fprintf (fp, "&\n");
    }
}

/* Function: PrintXMGRDistribution()
 * Date:     SRE, Wed Nov 12 11:02:09 1997 [St. Louis]
 *
 * Purpose:  Print an XMGR data file that contains two data sets:
 *               - xy data for the observed distribution P(S<x)
 *               - xy data for the theoretical distribution P(S<x)
 */
void
PrintXMGRDistribution (FILE * fp, struct histogram_s *h)
{
    int sc;			/* integer score in histogram structure */
    int cum;			/* cumulative count */
    double val;

    /* First data set is the observed distribution;
     * histogram bin x contains # of scores between x and x+1,
     * hence the sc+1 offset.
     */
    for (cum = 0, sc = h->lowscore; sc <= h->highscore; sc++) {
        cum += h->histogram[sc - h->min];
        fprintf (fp, "%-6d %f\n", sc + 1, (float) cum / (float) h->total);
    }
    fprintf (fp, "&\n");

    /* Second data set is the theoretical histogram
     */
    if (h->fit_type != HISTFIT_NONE) {
        for (sc = h->lowscore; sc <= h->highscore; sc++) {
            val = (1. - ExtremeValueP ((float) sc, h->param[EVD_MU],
                                       h->param[EVD_LAMBDA]));
            fprintf (fp, "%-6d %f\n", sc, val);
        }
        fprintf (fp, "&\n");
    }
}

/* Function: PrintXMGRRegressionLine()
 * Date:     SRE, Wed Nov 12 11:02:19 1997 [St. Louis]
 *
 * Purpose:  Print an XMGR data file that contains two data sets:
 *               - xy data for log log transform of observed distribution P(S<x)
 *               - xy data for log log transform of theoretical distribution P(S<x)
 */
void
PrintXMGRRegressionLine (FILE * fp, struct histogram_s *h)
{
    int sc;			/* integer score in histogram structure */
    int cum;
    double val;			/* log log transform */

    /* First data set is the observed distribution;
     * histogram bin x contains # of scores between x and x+1,
     * hence the sc+1 offset.
     */
    for (cum = 0, sc = h->lowscore; sc <= h->highscore; sc++) {
        cum += h->histogram[sc - h->min];
        val = log (-1. * log ((double) cum / (double) h->total));
        if (cum < h->total) {
            fprintf (fp, "%-6d %f\n", sc + 1, val);
        }
    }
    fprintf (fp, "&\n");

    /* Second data set is the theoretical histogram
     */
    if (h->fit_type != HISTFIT_NONE) {
        for (sc = h->lowscore; sc <= h->highscore; sc++) {
            val = log (-1. * log (1. - ExtremeValueP ((float) sc, h->param[EVD_MU],
                                  h->param[EVD_LAMBDA])));
            fprintf (fp, "%-6d %f\n", sc, val);
        }
        fprintf (fp, "&\n");
    }
}

/* Function: EVDBasicFit()
 * Date:     SRE, Wed Nov 12 11:02:27 1997 [St. Louis]
 *
 * Purpose:  Fit a score histogram to the extreme value
 *           distribution. Set the parameters lambda
 *           and mu in the histogram structure. Fill in the
 *           expected values in the histogram. Calculate
 *           a chi-square test as a measure of goodness of fit.
 *
 *           This is the basic version of ExtremeValueFitHistogram(),
 *           in a nonrobust form: simple linear regression with no
 *           outlier pruning.
 *
 * Methods:  Uses a linear regression fitting method [Collins88,Lawless82]
 *
 * Args:     h         - histogram to fit
 *
 * Return:   (void)
 */
void
EVDBasicFit (struct histogram_s *h)
{
    float *d;			/* distribution P(S < x)          */
    float *x;			/* x-axis of P(S<x) for Linefit() */
    int hsize;
    int sum;
    int sc, idx;			/* loop indices for score or score-h->min   */
    float slope, intercept;	/* m,b fit from Linefit()                   */
    float corr;			/* correlation coeff of line fit, not used  */
    float lambda, mu;		/* slope, intercept converted to EVD params */

    /* Allocations for x, y axes
     * distribution d runs from min..max with indices 0..max-min
     *     i.e. score - min = index into d, x, histogram, and expect
     */
    hsize = h->highscore - h->lowscore + 1;
    d = (float *) MallocOrDie (sizeof (float) * hsize);
    x = (float *) MallocOrDie (sizeof (float) * hsize);
    for (idx = 0; idx < hsize; idx++) {
        d[idx] = x[idx] = 0.;
    }

    /* Calculate P(S < x) distribution from histogram.
     * note off-by-one of sc, because histogram bin contains scores between
     * x and x+1.
     */
    sum = 0;
    for (sc = h->lowscore; sc <= h->highscore; sc++) {
        sum += h->histogram[sc - h->min];
        d[sc - h->lowscore] = (float) sum / (float) h->total;
        x[sc - h->lowscore] = (float) (sc + 1);
    }

    /* Do a linear regression fit to the log[-log(P(S<x))] "line".
     * we have log[-log(1-P(S>x))]  = -lambda * x + lambda * mu
     * so lambda = -m  and mu = b/lambda
     */
    /* convert y axis to log[-log(P(S<x))]  */
    for (sc = h->lowscore; sc < h->highscore; sc++) {
        d[sc - h->lowscore] = log (-1. * log (d[sc - h->lowscore]));
    }

    /* do the linear regression */
    Linefit (x, d, hsize - 1, &intercept, &slope, &corr);
    /* calc mu, lambda */
    lambda = -1. * slope;
    mu = intercept / lambda;

    /* Set the EVD parameters in the histogram;
     * pass 2 for additional lost degrees of freedom because we fit mu, lambda.
     */
    ExtremeValueSetHistogram (h, mu, lambda, h->lowscore, h->highscore, 2);

    free (x);
    free (d);
    return;
}


/* Function: ExtremeValueFitHistogram()
 * Date:     SRE, Sat Nov 15 17:16:15 1997 [St. Louis]
 *
 * Purpose:  Fit a score histogram to the extreme value
 *           distribution. Set the parameters lambda
 *           and mu in the histogram structure. Calculate
 *           a chi-square test as a measure of goodness of fit.
 *
 * Methods:  Uses a maximum likelihood method [Lawless82].
 *           Lower outliers are removed by censoring the data below the peak.
 *           Upper outliers are removed iteratively using method
 *           described by [Mott92].
 *
 * Args:     h         - histogram to fit
 *           censor    - TRUE to censor data left of the peak
 *           high_hint - score cutoff; above this are `real' hits that aren't fit
 *
 * Return:   1 if fit is judged to be valid.
 *           else 0 if fit is invalid (too few seqs.)
 */
int
ExtremeValueFitHistogram (struct histogram_s *h, int censor, float high_hint)
{
    float *x;			/* array of EVD samples to fit */
    int *y;			/* histogram counts            */
    int n;			/* number of observed samples  */
    int z;			/* number of censored samples  */
    int hsize;			/* size of histogram           */
    float lambda, mu;		/* new estimates of lambda, mu */
    int sc;			/* loop index for score        */
    int lowbound;			/* lower bound of fitted region */
    int highbound;		/* upper bound of fitted region */
    int new_highbound;
    int iteration;

    /* Determine lower bound on fitted region;
     * if we're censoring the data, choose the peak of the histogram.
     * if we're not, then we take the whole histogram.
     */
    lowbound = h->lowscore;
    if (censor) {
        int max = -1;
        for (sc = h->lowscore; sc <= h->highscore; sc++) {
            if (h->histogram[sc - h->min] > max) {
                max = h->histogram[sc - h->min];
                lowbound = sc;
            }
				}		
    }

    /* Determine initial upper bound on fitted region.
     */
    highbound = MIN (high_hint, h->highscore);

    /* Now, iteratively converge on our lambda, mu:
     */
    for (iteration = 0; iteration < 100; iteration++) {
        /* Construct x, y vectors.
         */
        x = NULL;
        y = NULL;
        hsize = highbound - lowbound + 1;
        if (hsize < 5) {
            goto FITFAILED;    /* require at least 5 bins or we don't fit */
        }

        x = MallocOrDie (sizeof (float) * hsize);
        y = MallocOrDie (sizeof (int) * hsize);
        n = 0;
        for (sc = lowbound; sc <= highbound; sc++) {
            x[sc - lowbound] = (float) sc + 0.5;	/* crude, but tests OK */
            y[sc - lowbound] = h->histogram[sc - h->min];
            n += h->histogram[sc - h->min];
        }

        if (n < 100) {
            goto FITFAILED;    /* require fitting to at least 100 points */
        }

        /* If we're censoring, estimate z, the number of censored guys
         * left of the bound. Our initial estimate is crudely that we're
         * missing e^-1 of the total distribution (which would be exact
         * if we censored exactly at mu; but we censored at the observed peak).
         * Subsequent estimates are more exact based on our current estimate of mu.
         */
        if (censor) {
            if (iteration == 0) {
                z = MIN (h->total - n, (int) (0.58198 * (float) n));
            }
            else {
                double psx;
                psx = EVDDistribution ((float) lowbound, mu, lambda);
                z = MIN (h->total - n, (int) ((double) n * psx / (1. - psx)));
            }
        }

        /* Do an ML fit
         */
        if (censor) {
            if (!EVDCensoredFit (x, y, hsize, z, (float) lowbound, &mu, &lambda)) {
                goto FITFAILED;
            }
        }
        else if (!EVDMaxLikelyFit (x, y, hsize, &mu, &lambda)) {
            goto FITFAILED;
        }

        /* Find the Eval = 1 point as a new highbound;
         * the total number of samples estimated to "belong" to the EVD is n+z
         */
        new_highbound = (int)
                        (mu -
                         (log (-1. * log ((double) (n + z - 1) / (double) (n + z))) / lambda));

        free (x);
        free (y);
        if (new_highbound >= highbound) {
            break;
        }
        highbound = new_highbound;
    }

    /* Set the histogram parameters;
     * - we fit from lowbound to highbound; thus we lose 2 degrees of freedom
     *   for fitting mu, lambda, but we get 1 back because we're unnormalized
     *   in this interval, hence we pass 2-1 = 1 as ndegrees.
     */
    ExtremeValueSetHistogram (h, mu, lambda, lowbound, highbound, 1);
    return 1;

FITFAILED:
    UnfitHistogram (h);
    if (x != NULL) {
        free (x);
    }
    if (y != NULL) {
        free (y);
    }
    return 0;
}


/* Function: ExtremeValueSetHistogram()
 *
 * Purpose:  Instead of fitting the histogram to an EVD,
 *           simply set the EVD parameters from an external source.
 *
 * Args:     h        - the histogram to set
 *           mu       - mu location parameter
 *           lambda   - lambda scale parameter
 *           lowbound - low bound of the histogram that was fit
 *           highbound- high bound of histogram that was fit
 *           ndegrees - extra degrees of freedom to subtract in X^2 test:
 *                        typically 0 if mu, lambda are parametric,
 *                        else 2 if mu, lambda are estimated from data
 */
void
ExtremeValueSetHistogram (struct histogram_s *h, float mu, float lambda,
                          float lowbound, float highbound, int ndegrees)
{
    int sc;
    int hsize, idx;
    int nbins;
    float delta;

    UnfitHistogram (h);
    h->fit_type = HISTFIT_EVD;
    h->param[EVD_LAMBDA] = lambda;
    h->param[EVD_MU] = mu;

    hsize = h->max - h->min + 1;
    h->expect = (float *) MallocOrDie (sizeof (float) * hsize);
    for (idx = 0; idx < hsize; idx++) {
        h->expect[idx] = 0.;
    }

    /* Calculate the expected values for the histogram.
     */
    for (sc = h->min; sc <= h->max; sc++) {
        h->expect[sc - h->min] =
            ExtremeValueE ((float) (sc), h->param[EVD_MU], h->param[EVD_LAMBDA],
                           h->total) -
            ExtremeValueE ((float) (sc + 1), h->param[EVD_MU], h->param[EVD_LAMBDA],
                           h->total);
		}
    /* Calculate the goodness-of-fit (within whole region)
     */
    h->chisq = 0.;
    nbins = 0;
    for (sc = lowbound; sc <= highbound; sc++) {
        if (h->expect[sc - h->min] >= 5. && h->histogram[sc - h->min] >= 5) {
            delta = (float) h->histogram[sc - h->min] - h->expect[sc - h->min];
            h->chisq += delta * delta / h->expect[sc - h->min];
            nbins++;
        }
		}
    /* Since we fit the whole histogram, there is at least
     * one constraint on chi-square: the normalization to h->total.
     */
    if (nbins > 1 + ndegrees)
        h->chip = (float) IncompleteGamma ((double) (nbins - 1 - ndegrees) / 2.,
                                           (double) h->chisq / 2.);
    else {
        h->chip = 0.;
    }
}



/* Function: GaussianFitHistogram()
 *
 * Purpose:  Fit a score histogram to a Gaussian distribution.
 *           Set the parameters mean and sd in the histogram
 *           structure, as well as a chi-squared test for
 *           goodness of fit.
 *
 * Args:     h         - histogram to fit
 *           high_hint - score cutoff; above this are `real' hits that aren't fit
 *
 * Return:   1 if fit is judged to be valid.
 *           else 0 if fit is invalid (too few seqs.)
 */
int
GaussianFitHistogram (struct histogram_s *h, float high_hint)
{
    float sum;
    float sqsum;
    float delta;
    int sc;
    int nbins;
    int hsize, idx;

    /* Clear any previous fitting from the histogram.
     */
    UnfitHistogram (h);

    /* Determine if we have enough hits to fit the histogram;
     * arbitrarily require 1000.
     */
    if (h->total < 1000) {
        h->fit_type = HISTFIT_NONE;
        return 0;
    }

    /* Simplest algorithm for mean and sd;
     * no outlier detection yet (not even using high_hint)
     *
     * Magic 0.5 correction is because our histogram is for
     * scores between x and x+1; we estimate the expectation
     * (roughly) as x + 0.5.
     */
    sum = sqsum = 0.;
    for (sc = h->lowscore; sc <= h->highscore; sc++) {
        delta = (float) sc + 0.5;
        sum += (float) h->histogram[sc - h->min] * delta;
        sqsum += (float) h->histogram[sc - h->min] * delta * delta;
    }
    h->fit_type = HISTFIT_GAUSSIAN;
    h->param[GAUSS_MEAN] = sum / (float) h->total;
    h->param[GAUSS_SD] = sqrt ((sqsum - (sum * sum / (float) h->total)) /
                               (float) (h->total - 1));

    /* Calculate the expected values for the histogram.
     * Note that the magic 0.5 correction appears again.
     * Calculating difference between distribution functions for Gaussian
     * would be correct but hard.
     */
    hsize = h->max - h->min + 1;
    h->expect = (float *) MallocOrDie (sizeof (float) * hsize);
    for (idx = 0; idx < hsize; idx++) {
        h->expect[idx] = 0.;
    }

    for (sc = h->min; sc <= h->max; sc++) {
        delta = (float) sc + 0.5 - h->param[GAUSS_MEAN];
        h->expect[sc - h->min] =
            (float) h->total * ((1. / (h->param[GAUSS_SD] * sqrt (2. * 3.14159))) *
                                (exp
                                 (-1. * delta * delta /
                                  (2. * h->param[GAUSS_SD] * h->param[GAUSS_SD]))));
    }

    /* Calculate the goodness-of-fit (within region that was fitted)
     */
    h->chisq = 0.;
    nbins = 0;
    for (sc = h->lowscore; sc <= h->highscore; sc++) {
        if (h->expect[sc - h->min] >= 5. && h->histogram[sc - h->min] >= 5) {
            delta = (float) h->histogram[sc - h->min] - h->expect[sc - h->min];
            h->chisq += delta * delta / h->expect[sc - h->min];
            nbins++;
        }
		}		
    /* -1 d.f. for normalization; -2 d.f. for two free parameters */
    if (nbins > 3)
        h->chip = (float) IncompleteGamma ((double) (nbins - 3) / 2.,
                                           (double) h->chisq / 2.);
    else {
        h->chip = 0.;
    }

    return 1;
}


/* Function: GaussianSetHistogram()
 *
 * Purpose:  Instead of fitting the histogram to a Gaussian,
 *           simply set the Gaussian parameters from an external source.
 */
void
GaussianSetHistogram (struct histogram_s *h, float mean, float sd)
{
    int sc;
    int hsize, idx;
    int nbins;
    float delta;

    UnfitHistogram (h);
    h->fit_type = HISTFIT_GAUSSIAN;
    h->param[GAUSS_MEAN] = mean;
    h->param[GAUSS_SD] = sd;

    /* Calculate the expected values for the histogram.
     */
    hsize = h->max - h->min + 1;
    h->expect = (float *) MallocOrDie (sizeof (float) * hsize);
    for (idx = 0; idx < hsize; idx++) {
        h->expect[idx] = 0.;
    }

    /* Note: ideally we'd use the Gaussian distribution function
     * to find the histogram occupancy in the window sc..sc+1.
     * However, the distribution function is hard to calculate.
     * Instead, estimate the histogram by taking the density at sc+0.5.
     */
    for (sc = h->min; sc <= h->max; sc++) {
        delta = ((float) sc + 0.5) - h->param[GAUSS_MEAN];
        h->expect[sc - h->min] =
            (float) h->total * ((1. / (h->param[GAUSS_SD] * sqrt (2. * 3.14159))) *
                                (exp
                                 (-1. * delta * delta /
                                  (2. * h->param[GAUSS_SD] * h->param[GAUSS_SD]))));
    }

    /* Calculate the goodness-of-fit (within whole region)
     */
    h->chisq = 0.;
    nbins = 0;
    for (sc = h->lowscore; sc <= h->highscore; sc++) {
        if (h->expect[sc - h->min] >= 5. && h->histogram[sc - h->min] >= 5) {
            delta = (float) h->histogram[sc - h->min] - h->expect[sc - h->min];
            h->chisq += delta * delta / h->expect[sc - h->min];
            nbins++;
        }
		}		
    /* -1 d.f. for normalization */
    if (nbins > 1)
        h->chip = (float) IncompleteGamma ((double) (nbins - 1) / 2.,
                                           (double) h->chisq / 2.);
    else {
        h->chip = 0.;
    }
}



/* Function: EVDDensity()
 * Date:     SRE, Sat Nov 15 19:37:52 1997 [St. Louis]
 *
 * Purpose:  Return the extreme value density P(S=x) at
 *           a given point x, for an EVD controlled by
 *           parameters mu and lambda.
 */
double
EVDDensity (float x, float mu, float lambda)
{
    return (lambda * exp (-1. * lambda * (x - mu)
                          - exp (-1. * lambda * (x - mu))));
}

/* Function: EVDDistribution()
 * Date:     SRE, Tue Nov 18 08:02:22 1997 [St. Louis]
 *
 * Purpose:  Returns the extreme value distribution P(S < x)
 *           evaluated at x, for an EVD controlled by parameters
 *           mu and lambda.
 */
double
EVDDistribution (float x, float mu, float lambda)
{
    return (exp (-1. * exp (-1. * lambda * (x - mu))));
}

/* Function: ExtremeValueP()
 *
 * Purpose:  Calculate P(S>x) according to an extreme
 *           value distribution, given x and the parameters
 *           of the distribution (characteristic
 *           value mu, decay constant lambda).
 *
 *           This function is exquisitely prone to
 *           floating point exceptions if it isn't coded
 *           carefully.
 *
 * Args:     x      = score
 *           mu     = characteristic value of extreme value distribution
 *           lambda = decay constant of extreme value distribution
 *
 * Return:   P(S>x)
 */
double
ExtremeValueP (float x, float mu, float lambda)
{
    double y;
    /* avoid exceptions near P=1.0 */
    /* typical 32-bit sys: if () < -3.6, return 1.0 */
    if ((lambda * (x - mu)) <= -1. * log (-1. * log (DBL_EPSILON))) {
        return 1.0;
    }
    /* avoid underflow fp exceptions near P=0.0 */
    if ((lambda * (x - mu)) >= 2.3 * (double) DBL_MAX_10_EXP) {
        return 0.0;
    }
    /* a roundoff issue arises; use 1 - e^-x --> x for small x */
    y = exp (-1. * lambda * (x - mu));
    if (y < 1e-7) {
        return y;
    }
    else {
        return (1.0 - exp (-1. * y));
    }
}


/* Function: ExtremeValueP2()
 *
 * Purpose:  Calculate P(S>x) in a database of size N,
 *           using P(S>x) for a single sequence, according
 *           to a Poisson distribution.
 *
 * Args:     x      = score
 *           mu     = characteristic value of extreme value distribution
 *           lambda = decay constant of extreme value distribution
 *           N      = number of trials (number of sequences)
 *
 * Return:   P(S>x) for database of size N
 */
double
ExtremeValueP2 (float x, float mu, float lambda, int N)
{
    double y;
    y = N * ExtremeValueP (x, mu, lambda);
    if (y < 1e-7) {
        return y;
    }
    else {
        return (1.0 - exp (-1. * y));
    }
}

/* Function: ExtremeValueE()
 *
 * Purpose:  Calculate E(S>x) in a database of size N,
 *           using P(S>x) for a single sequence: simply np.
 *
 * Args:     x      = score
 *           mu     = characteristic value of extreme value distribution
 *           lambda = decay constant of extreme value distribution
 *           N      = number of trials (number of sequences)
 *
 * Return:   E(S>x) for database of size N
 */
double
ExtremeValueE (float x, float mu, float lambda, int N)
{
    return (double) N *ExtremeValueP (x, mu, lambda);
}


/* Function: EVDrandom()
 *
 * Purpose:  Randomly sample an x from an EVD.
 *           Trivially done by the transformation method, since
 *           the distribution is analytical:
 *              x = \mu - \frac{\log \left[ -\log P(S<x) \right]}{\lambda}
 *           where P(S<x) is sampled uniformly on 0 < P(S<x) < 1.
 */
float
EVDrandom (float mu, float lambda)
{
    float p = 0.0;

    /* Very unlikely, but possible,
     * that sre_random() would give us exactly 0 or 1
     */
    while (p == 0. || p == 1.) {
        p = sre_random ();
    }
    return mu - log (-1. * log (p)) / lambda;
}


/* Function: Lawless416()
 * Date:     SRE, Thu Nov 13 11:48:50 1997 [St. Louis]
 *
 * Purpose:  Equation 4.1.6 from [Lawless82], pg. 143, and
 *           its first derivative with respect to lambda,
 *           for finding the ML fit to EVD lambda parameter.
 *           This equation gives a result of zero for the maximum
 *           likelihood lambda.
 *
 *           Can either deal with a histogram or an array.
 *
 *           Warning: beware overflow/underflow issues! not bulletproof.
 *
 * Args:     x      - array of sample values (or x-axis of a histogram)
 *           y      - NULL (or y-axis of a histogram)
 *           n      - number of samples (or number of histogram bins)
 *           lambda - a lambda to test
 *           ret_f  - RETURN: 4.1.6 evaluated at lambda
 *           ret_df - RETURN: first derivative of 4.1.6 evaluated at lambda
 *
 * Return:   (void)
 */
void
Lawless416 (float *x, int *y, int n, float lambda, float *ret_f,
            float *ret_df)
{

    double esum;			/* \sum e^(-lambda xi)      */
    double xesum;			/* \sum xi e^(-lambda xi)   */
    double xxesum;		/* \sum xi^2 e^(-lambda xi) */
    double xsum;			/* \sum xi                  */
    double mult;			/* histogram count multiplier */
    double total;			/* total samples            */
    int i;


    esum = xesum = xsum = xxesum = total = 0.;
    for (i = 0; i < n; i++) {
        mult = (y == NULL) ? 1. : (double) y[i];
        xsum += mult * x[i];
        xesum += mult * x[i] * exp (-1. * lambda * x[i]);
        xxesum += mult * x[i] * x[i] * exp (-1. * lambda * x[i]);
        esum += mult * exp (-1. * lambda * x[i]);
        total += mult;
    }
    *ret_f = 1. / lambda - xsum / total + xesum / esum;
    *ret_df = ((xesum / esum) * (xesum / esum))
              - (xxesum / esum)
              - (1. / (lambda * lambda));

    return;
}


/* Function: Lawless422()
 * Date:     SRE, Mon Nov 17 09:42:48 1997 [St. Louis]
 *
 * Purpose:  Equation 4.2.2 from [Lawless82], pg. 169, and
 *           its first derivative with respect to lambda,
 *           for finding the ML fit to EVD lambda parameter
 *           for Type I censored data.
 *           This equation gives a result of zero for the maximum
 *           likelihood lambda.
 *
 *           Can either deal with a histogram or an array.
 *
 *           Warning: beware overflow/underflow issues! not bulletproof.
 *
 * Args:     x      - array of sample values (or x-axis of a histogram)
 *           y      - NULL (or y-axis of a histogram)
 *           n      - number of observed samples (or number of histogram bins)
 *           z      - number of censored samples
 *           c      - censoring value; all observed x_i >= c
 *           lambda - a lambda to test
 *           ret_f  - RETURN: 4.2.2 evaluated at lambda
 *           ret_df - RETURN: first derivative of 4.2.2 evaluated at lambda
 *
 * Return:   (void)
 */
void
Lawless422 (float *x, int *y, int n, int z, float c,
            float lambda, float *ret_f, float *ret_df)
{
    double esum;			/* \sum e^(-lambda xi)      + z term    */
    double xesum;			/* \sum xi e^(-lambda xi)   + z term    */
    double xxesum;		/* \sum xi^2 e^(-lambda xi) + z term    */
    double xsum;			/* \sum xi                  (no z term) */
    double mult;			/* histogram count multiplier */
    double total;			/* total samples            */
    int i;

    esum = xesum = xsum = xxesum = total = 0.;
    for (i = 0; i < n; i++) {
        mult = (y == NULL) ? 1. : (double) y[i];
        xsum += mult * x[i];
        esum += mult * exp (-1. * lambda * x[i]);
        xesum += mult * x[i] * exp (-1. * lambda * x[i]);
        xxesum += mult * x[i] * x[i] * exp (-1. * lambda * x[i]);
        total += mult;
    }

    /* Add z terms for censored data
     */
    esum += (double) z *exp (-1. * lambda * c);
    xesum += (double) z *c * exp (-1. * lambda * c);
    xxesum += (double) z *c * c * exp (-1. * lambda * c);

    *ret_f = 1. / lambda - xsum / total + xesum / esum;
    *ret_df = ((xesum / esum) * (xesum / esum))
              - (xxesum / esum)
              - (1. / (lambda * lambda));

    return;
}



/* Function: EVDMaxLikelyFit()
 * Date:     SRE, Fri Nov 14 07:56:29 1997 [St. Louis]
 *
 * Purpose:  Given a list or a histogram of EVD-distributed samples,
 *           find maximum likelihood parameters lambda and
 *           mu.
 *
 * Algorithm: Uses approach described in [Lawless82]. Solves
 *           for lambda using Newton/Raphson iterations;
 *           then substitutes lambda into Lawless' equation 4.1.5
 *           to get mu.
 *
 *           Newton/Raphson algorithm developed from description in
 *           Numerical Recipes in C [Press88].
 *
 * Args:     x          - list of EVD distributed samples or x-axis of histogram
 *           c          - NULL, or y-axis of histogram
 *           n          - number of samples, or number of histogram bins
 *           ret_mu     : RETURN: ML estimate of mu
 *           ret_lambda : RETURN: ML estimate of lambda
 *
 * Return:   1 on success; 0 on any failure
 */
int
EVDMaxLikelyFit (float *x, int *c, int n, float *ret_mu, float *ret_lambda)
{
    float lambda, mu;
    float fx;			/* f(x)  */
    float dfx;			/* f'(x) */
    double esum;			/* \sum e^(-lambda xi) */
    double mult;
    double total;
    float tol = 1e-5;
    int i;

    /* 1. Find an initial guess at lambda: linear regression here?
     */
    lambda = 0.2;

    /* 2. Use Newton/Raphson to solve Lawless 4.1.6 and find ML lambda
     */
    for (i = 0; i < 100; i++) {
        Lawless416 (x, c, n, lambda, &fx, &dfx);
        if (fabs (fx) < tol) {
            break;    /* success */
        }
        lambda = lambda - fx / dfx;	/* Newton/Raphson is simple */
        if (lambda <= 0.) {
            lambda = 0.001;    /* but be a little careful  */
        }
    }

    /* 2.5: If we did 100 iterations but didn't converge, Newton/Raphson failed.
     *      Resort to a bisection search. Worse convergence speed
     *      but guaranteed to converge (unlike Newton/Raphson).
     *      We assume (!?) that fx is a monotonically decreasing function of x;
     *      i.e. fx > 0 if we are left of the root, fx < 0 if we
     *      are right of the root.
     */
    if (i == 100) {
        float left, right, mid;
        SQD_DPRINTF2 (("EVDMaxLikelyFit(): Newton/Raphson failed; switchover to bisection"));

        /* First we need to bracket the root */
        lambda = right = left = 0.2;
        Lawless416 (x, c, n, lambda, &fx, &dfx);
        if (fx < 0.) {		/* fix right; search left. */
            do {
                left -= 0.1;
                if (left < 0.) {
                    SQD_DPRINTF2 (("EVDMaxLikelyFit(): failed to bracket root"));
                    return 0;
                }
                Lawless416 (x, c, n, left, &fx, &dfx);
            } while (fx < 0.);
        }
        else {			/* fix left; search right. */
            do {
                right += 0.1;
                Lawless416 (x, c, n, right, &fx, &dfx);
                if (right > 100.) {
                    SQD_DPRINTF2 (("EVDMaxLikelyFit(): failed to bracket root"));
                    return 0;
                }
            } while (fx > 0.);
        }
        /* now we bisection search in left/right interval */
        for (i = 0; i < 100; i++) {
            mid = (left + right) / 2.;
            Lawless416 (x, c, n, mid, &fx, &dfx);
            if (fabs (fx) < tol) {
                break;    /* success */
            }
            if (fx > 0.) {
                left = mid;
            }
            else {
                right = mid;
            }
        }
        if (i == 100) {
            SQD_DPRINTF2 (("EVDMaxLikelyFit(): even the bisection search failed"));
            return 0;
        }
        lambda = mid;
    }

    /* 3. Substitute into Lawless 4.1.5 to find mu
     */
    esum = 0.;
    total = 0.;
    for (i = 0; i < n; i++) {
        mult = (c == NULL) ? 1. : (double) c[i];
        esum += mult * exp (-1 * lambda * x[i]);
        total += mult;
    }
    mu = -1. * log (esum / total) / lambda;

    *ret_lambda = lambda;
    *ret_mu = mu;
    return 1;
}


/* Function: EVDCensoredFit()
 * Date:     SRE, Mon Nov 17 10:01:05 1997 [St. Louis]
 *
 * Purpose:  Given a /left-censored/ list or histogram of EVD-distributed
 *           samples, as well as the number of censored samples z and the
 *           censoring value c,
 *           find maximum likelihood parameters lambda and
 *           mu.
 *
 * Algorithm: Uses approach described in [Lawless82]. Solves
 *           for lambda using Newton/Raphson iterations;
 *           then substitutes lambda into Lawless' equation 4.2.3
 *           to get mu.
 *
 *           Newton/Raphson algorithm developed from description in
 *           Numerical Recipes in C [Press88].
 *
 * Args:     x          - list of EVD distributed samples or x-axis of histogram
 *           y          - NULL, or y-axis of histogram
 *           n          - number of observed samples,or number of histogram bins
 *           z          - number of censored samples
 *           c          - censoring value (all x_i >= c)
 *           ret_mu     : RETURN: ML estimate of mu
 *           ret_lambda : RETURN: ML estimate of lambda
 *
 * Return:   (void)
 */
int
EVDCensoredFit (float *x, int *y, int n, int z, float c,
                float *ret_mu, float *ret_lambda)
{
    float lambda, mu;
    float fx;			/* f(x)  */
    float dfx;			/* f'(x) */
    double esum;			/* \sum e^(-lambda xi) */
    double mult;
    double total;
    float tol = 1e-5;
    int i;

    /* 1. Find an initial guess at lambda: linear regression here?
     */
    lambda = 0.2;

    /* 2. Use Newton/Raphson to solve Lawless 4.2.2 and find ML lambda
     */
    for (i = 0; i < 100; i++) {
        Lawless422 (x, y, n, z, c, lambda, &fx, &dfx);
        if (fabs (fx) < tol) {
            break;    /* success */
        }
        lambda = lambda - fx / dfx;	/* Newton/Raphson is simple */
        if (lambda <= 0.) {
            lambda = 0.001;    /* but be a little careful  */
        }
    }

    /* 2.5: If we did 100 iterations but didn't converge, Newton/Raphson failed.
     *      Resort to a bisection search. Worse convergence speed
     *      but guaranteed to converge (unlike Newton/Raphson).
     *      We assume (!?) that fx is a monotonically decreasing function of x;
     *      i.e. fx > 0 if we are left of the root, fx < 0 if we
     *      are right of the root.
     */
    if (i == 100) {
        float left, right, mid;
        /* First we need to bracket the root */
        SQD_DPRINTF2 (("EVDCensoredFit(): Newton/Raphson failed; switched to bisection"));
        lambda = right = left = 0.2;
        Lawless422 (x, y, n, z, c, lambda, &fx, &dfx);
        if (fx < 0.) {		/* fix right; search left. */
            do {
                left -= 0.03;
                if (left < 0.) {
                    SQD_DPRINTF2 (("EVDCensoredFit(): failed to bracket root"));
                    return 0;
                }
                Lawless422 (x, y, n, z, c, left, &fx, &dfx);
            } while (fx < 0.);
        }
        else {			/* fix left; search right. */
            do {
                right += 0.1;
                Lawless422 (x, y, n, z, c, left, &fx, &dfx);
                if (right > 100.) {
                    SQD_DPRINTF2 (("EVDCensoredFit(): failed to bracket root"));
                    return 0;
                }
            } while (fx > 0.);
        }
        /* now we bisection search in left/right interval */
        for (i = 0; i < 100; i++) {
            mid = (left + right) / 2.;
            Lawless422 (x, y, n, z, c, left, &fx, &dfx);
            if (fabs (fx) < tol) {
                break;    /* success */
            }
            if (fx > 0.) {
                left = mid;
            }
            else {
                right = mid;
            }
        }
        if (i == 100) {
            SQD_DPRINTF2 (("EVDCensoredFit(): even the bisection search failed"));
            return 0;
        }
        lambda = mid;
    }

    /* 3. Substitute into Lawless 4.2.3 to find mu
     */
    esum = total = 0.;
    for (i = 0; i < n; i++) {
        mult = (y == NULL) ? 1. : (double) y[i];
        esum += mult * exp (-1. * lambda * x[i]);
        total += mult;
    }
    esum += (double) z *exp (-1. * lambda * c);	/* term from censored data */
    mu = -1. * log (esum / total) / lambda;

    *ret_lambda = lambda;
    *ret_mu = mu;
    return 1;
}



/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/*****************************************************************
 * This code is an altered version of Henry Spencer's
 * regex library. Alterations are limited to minor streamlining,
 * and some name changes to protect the SQUID namespace.
 * Henry's copyright notice appears below.
 * You can obtain the original from
 *    ftp://ftp.zoo.toronto.edu/pub/bookregex.tar.Z
 * Thanks, Henry!
 *
 * The magic word for compiling a testdriver: NBA_TEAM_IN_STL
 * gcc -o test -g -DNBA_TEAM_IN_STL -L. hsregex.c -lsquid -lm
 *
 * Usage:
 *  test <pattern> <ntok> <string>
 *
 * SRE, Fri Aug 28 11:10:17 1998
 * CVS $Id: hsregex.c,v 1.8 2003/04/14 16:00:16 eddy Exp $
 *****************************************************************/

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"

/* global sqd_parse[] are managed by Strparse().
 * WARNING: TODO: this code is not threadsafe, and needs to be revised.
 */
char *sqd_parse[10];

/* Function: Strparse()
 *
 * Purpose:  Match a regexp to a string. Returns 1 if pattern matches,
 *           else 0.
 *
 *           Much like Perl, Strparse() makes copies of the matching
 *           substrings available via globals, sqd_parse[].
 *           sqd_parse[0] contains a copy of the complete matched
 *           text. sqd_parse[1-9] contain copies of up to nine
 *           different substrings matched within parentheses.
 *           The memory for these strings is internally managed and
 *           volatile; the next call to Strparse() may destroy them.
 *           If the caller needs the matched substrings to persist
 *           beyond a new Strparse() call, it must make its own
 *           copies.
 *
 *           A minor drawback of the memory management is that
 *           there will be a small amount of unfree'd memory being
 *           managed by Strparse() when a program exits; this may
 *           confuse memory debugging (Purify, dbmalloc). The
 *           general cleanup function SqdClean() is provided;
 *           you can call this before exiting.
 *
 *           Uses an extended POSIX regular expression interface.
 *           A copylefted GNU implementation is included in the squid
 *           implementation (gnuregex.c) for use on non-POSIX compliant
 *           systems. POSIX 1003.2-compliant systems (all UNIX,
 *           some WinNT, I believe) can omit the GNU code if necessary.
 *
 *           I built this for ease of use, not speed nor efficiency.
 *
 * Example:  Strparse("foo-...-baz", "foo-bar-baz")  returns 0
 *           Strparse("foo-(...)-baz", "foo-bar-baz")
 *              returns 0; sqd_parse[0] is "foo-bar-baz";
 *              sqd_parse[1] is "bar".
 *
 *           A real example:
 *            s   = ">gnl|ti|3 G10P69425RH2.T0 {SUB 81..737}  /len=657"
 *            pat = "SUB ([0-9]+)"
 *            Strparse(pat, s, 1)
 *               returns 1; sqd_parse[1] is "81".
 *
 * Args:     rexp  - regular expression, extended POSIX form
 *           s     - string to match against
 *           ntok  - number of () substrings we will save (maximum NSUBEXP-1)
 *
 * Return:   1 on match, 0 if no match
 */
int
Strparse (char *rexp, char *s, int ntok)
{
    sqd_regexp *pat;
    int code;
    int len;
    int i;
    /* sanity check */
    if (ntok >= NSUBEXP) {
        Die ("Strparse(): ntok must be <= %d", NSUBEXP - 1);
    }

    /* Free previous global substring buffers
     */
    for (i = 0; i <= ntok; i++) {
        if (sqd_parse[i] != NULL) {
            free (sqd_parse[i]);
            sqd_parse[i] = NULL;
        }
		}
    /* Compile and match the pattern, using our modified
     * copy of Henry Spencer's regexp library
     */
    if ((pat = sqd_regcomp (rexp)) == NULL) {
        Die ("regexp compilation failed.");
    }
    code = sqd_regexec (pat, s);

    /* Fill the global substring buffers
     */
    if (code == 1)
        for (i = 0; i <= ntok; i++) {
            if (pat->startp[i] != NULL && pat->endp[i] != NULL) {
                len = pat->endp[i] - pat->startp[i];
                sqd_parse[i] = (char *) MallocOrDie (sizeof (char) * (len + 1));
                strncpy (sqd_parse[i], pat->startp[i], len);
                sqd_parse[i][len] = '\0';
            }
				}
    free (pat);
    return code;
}

/* Function: SqdClean()
 * Date:     SRE, Wed Oct 29 12:52:08 1997 [TWA 721]
 *
 * Purpose:  Clean up any squid library allocations before exiting
 *           a program, so we don't leave unfree'd memory around
 *           and confuse a malloc debugger like Purify or dbmalloc.
 */
void
SqdClean (void)
{
    int i;

    /* Free global substring buffers that Strparse() uses
     */
    for (i = 0; i <= 9; i++) {
        if (sqd_parse[i] != NULL) {
            free (sqd_parse[i]);
            sqd_parse[i] = NULL;
        }
		}		
}



/* all code below is:
 * Copyright (c) 1986, 1993, 1995 by University of Toronto.
 * Written by Henry Spencer.  Not derived from licensed software.
 *
 * Permission is granted to anyone to use this software for any
 * purpose on any computer system, and to redistribute it in any way,
 * subject to the following restrictions:
 *
 * 1. The author is not responsible for the consequences of use of
 * 	this software, no matter how awful, even if they arise
 * 	from defects in it.
 *
 * 2. The origin of this software must not be misrepresented, either
 * 	by explicit claim or by omission.
 *
 * 3. Altered versions must be plainly marked as such, and must not
 * 	be misrepresented (by explicit claim or omission) as being
 *	the original software.
 *
 * 4. This notice must not be removed or altered.
 */

/*
 * sqd_regcomp and sqd_regexec -- sqd_regsub and sqd_regerror are elsewhere
 */

/*
 * The first byte of the regexp internal "program" is actually this magic
 * number; the start node begins in the second byte.
 */
#define	SQD_REGMAGIC	0234

/*
 * The "internal use only" fields in regexp.h are present to pass info from
 * compile to execute that permits the execute phase to run lots faster on
 * simple cases.  They are:
 *
 * regstart	char that must begin a match; '\0' if none obvious
 * reganch	is the match anchored (at beginning-of-line only)?
 * regmust	string (pointer into program) that match must include, or NULL
 * regmlen	length of regmust string
 *
 * Regstart and reganch permit very fast decisions on suitable starting points
 * for a match, cutting down the work a lot.  Regmust permits fast rejection
 * of lines that cannot possibly match.  The regmust tests are costly enough
 * that sqd_regcomp() supplies a regmust only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or +
 * at the start of the r.e., which can involve a lot of backup).  Regmlen is
 * supplied because the test in sqd_regexec() needs it and sqd_regcomp() is computing
 * it anyway.
 */

/*
 * Structure for regexp "program".  This is essentially a linear encoding
 * of a nondeterministic finite-state machine (aka syntax charts or
 * "railroad normal form" in parsing technology).  Each node is an opcode
 * plus a "next" pointer, possibly plus an operand.  "Next" pointers of
 * all nodes except BRANCH implement concatenation; a "next" pointer with
 * a BRANCH on both ends of it is connecting two alternatives.  (Here we
 * have one of the subtle syntax dependencies:  an individual BRANCH (as
 * opposed to a collection of them) is never concatenated with anything
 * because of operator precedence.)  The operand of some types of node is
 * a literal string; for others, it is a node leading into a sub-FSM.  In
 * particular, the operand of a BRANCH node is the first node of the branch.
 * (NB this is *not* a tree structure:  the tail of the branch connects
 * to the thing following the set of BRANCHes.)  The opcodes are:
 */

/* definition	number	opnd?	meaning */
#define	END	0		/* no   End of program. */
#define	BOL	1		/* no   Match beginning of line. */
#define	EOL	2		/* no   Match end of line. */
#define	ANY	3		/* no   Match any character. */
#define	ANYOF	4		/* str  Match any of these. */
#define	ANYBUT	5		/* str  Match any but one of these. */
#define	BRANCH	6		/* node Match this, or the next..\&. */
#define	BACK	7		/* no   "next" ptr points backward. */
#define	EXACTLY	8		/* str  Match this string. */
#define	NOTHING	9		/* no   Match empty string. */
#define	STAR	10		/* node Match this 0 or more times. */
#define	PLUS	11		/* node Match this 1 or more times. */
#define	OPEN	20		/* no   Sub-RE starts here. */
/*      OPEN+1 is number 1, etc. */
#define	CLOSE	30		/* no   Analogous to OPEN. */

/*
 * Opcode notes:
 *
 * BRANCH	The set of branches constituting a single choice are hooked
 *		together with their "next" pointers, since precedence prevents
 *		anything being concatenated to any individual branch.  The
 *		"next" pointer of the last BRANCH in a choice points to the
 *		thing following the whole choice.  This is also where the
 *		final "next" pointer of each individual branch points; each
 *		branch starts with the operand node of a BRANCH node.
 *
 * BACK		Normal "next" pointers all implicitly point forward; BACK
 *		exists to make loop structures possible.
 *
 * STAR,PLUS	'?', and complex '*' and '+', are implemented as circular
 *		BRANCH structures using BACK.  Simple cases (one character
 *		per match) are implemented with STAR and PLUS for speed
 *		and to minimize recursive plunges.
 *
 * OPEN,CLOSE	...are numbered at compile time.
 */

/*
 * A node is one char of opcode followed by two chars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define	OP(p)		(*(p))
#define	NEXT(p)		(((*((p)+1)&0177)<<8) + (*((p)+2)&0377))
#define	OPERAND(p)	((p) + 3)

/*
 * Utility definitions.
 */
#define	FAIL(m)		{ sqd_regerror(m); return(NULL); }
#define	ISREPN(c)	((c) == '*' || (c) == '+' || (c) == '?')
#define	META		"^$.[()|?+*\\"

/*
 * Flags to be passed up and down.
 */
#define	HASWIDTH	01	/* Known never to match null string. */
#define	SIMPLE		02	/* Simple enough to be STAR/PLUS operand. */
#define	SPSTART		04	/* Starts with * or +. */
#define	WORST		0	/* Worst case. */

/*
 * Work-variable struct for sqd_regcomp().
 */
struct comp
{
    char *regparse;		/* Input-scan pointer. */
    int regnpar;			/* () count. */
    char *regcode;		/* Code-emit pointer; &regdummy = don't. */
    char regdummy[3];		/* NOTHING, 0 next ptr */
    long regsize;			/* Code size. */
};
#define	EMITTING(cp)	((cp)->regcode != (cp)->regdummy)

/*
 * Forward declarations for sqd_regcomp()'s friends.
 */
static char *reg (struct comp *cp, int paren, int *flagp);
static char *regbranch (struct comp *cp, int *flagp);
static char *regpiece (struct comp *cp, int *flagp);
static char *regatom (struct comp *cp, int *flagp);
// sergio : Se cambia el prototipo de acuerdo a la definicion posterior que se hace
static char *regnode (struct comp *cp, int op);
//static char *regnode(register struct comp *cp, char op);
static char *regnext (char *node);
// sergio : Se cambia el prototipo de acuerdo a la definicion posterior que se hace
static void regc (struct comp *cp, char c);
// sergio : Se cambia el prototipo de acuerdo a la definicion posterior que se hace
static void reginsert (register struct comp *cp, char op, char *opnd);
static void regtail (struct comp *cp, char *p, char *val);
static void regoptail (struct comp *cp, char *p, char *val);

/*
 - sqd_regcomp - compile a regular expression into internal code
 *
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled regexp.
 */
sqd_regexp *
sqd_regcomp (exp)
const char *exp;
{
    register sqd_regexp *r;
    register char *scan;
    int flags;
    struct comp co;

    if (exp == NULL) {
        FAIL ("NULL argument to sqd_regcomp");
    }

    /* First pass: determine size, legality. */
    co.regparse = (char *) exp;
    co.regnpar = 1;
    co.regsize = 0L;
    co.regdummy[0] = NOTHING;
    co.regdummy[1] = co.regdummy[2] = 0;
    co.regcode = co.regdummy;
    regc (&co, SQD_REGMAGIC);
    if (reg (&co, 0, &flags) == NULL) {
        return (NULL);
    }

    /* Small enough for pointer-storage convention? */
    if (co.regsize >= 0x7fffL) {	/* Probably could be 0xffffL. */
        FAIL ("regexp too big");
    }

    /* Allocate space. */
    r = (sqd_regexp *) malloc (sizeof (sqd_regexp) + (size_t) co.regsize);
    if (r == NULL) {
        FAIL ("out of space");
    }

    /* Second pass: emit code. */
    co.regparse = (char *) exp;
    co.regnpar = 1;
    co.regcode = r->program;
    regc (&co, SQD_REGMAGIC);
    if (reg (&co, 0, &flags) == NULL) {
        return (NULL);
    }

    /* Dig out information for optimizations. */
    r->regstart = '\0';		/* Worst-case defaults. */
    r->reganch = 0;
    r->regmust = NULL;
    r->regmlen = 0;
    scan = r->program + 1;	/* First BRANCH. */
    if (OP (regnext (scan)) == END) {	/* Only one top-level choice. */
        scan = OPERAND (scan);

        /* Starting-point info. */
        if (OP (scan) == EXACTLY) {
            r->regstart = *OPERAND (scan);
        }
        else if (OP (scan) == BOL) {
            r->reganch = 1;
        }

        /*
         * If there's something expensive in the r.e., find the
         * longest literal string that must appear and make it the
         * regmust.  Resolve ties in favor of later strings, since
         * the regstart check works with the beginning of the r.e.
         * and avoiding duplication strengthens checking.  Not a
         * strong reason, but sufficient in the absence of others.
         */
        if (flags & SPSTART) {
            register char *longest = NULL;
            register size_t len = 0;

            for (; scan != NULL; scan = regnext (scan)) {
                if (OP (scan) == EXACTLY && strlen (OPERAND (scan)) >= len) {
                    longest = OPERAND (scan);
                    len = strlen (OPERAND (scan));
                }
						}		
            r->regmust = longest;
            r->regmlen = (int) len;
        }
    }

    return (r);
}

/*
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
 */
static char *
reg (cp, paren, flagp)
register struct comp *cp;
int paren;			/* Parenthesized? */
int *flagp;
{
    register char *ret = NULL;	/* SRE: NULL init added to silence gcc */
    register char *br;
    register char *ender;
    register int parno = 0;	/* SRE: init added to silence gcc */
    int flags;

    *flagp = HASWIDTH;		/* Tentatively. */

    if (paren) {
        /* Make an OPEN node. */
        if (cp->regnpar >= NSUBEXP) {
            FAIL ("too many ()");
        }
        parno = cp->regnpar;
        cp->regnpar++;
        ret = regnode (cp, OPEN + parno);
    }

    /* Pick up the branches, linking them together. */
    br = regbranch (cp, &flags);
    if (br == NULL) {
        return (NULL);
    }
    if (paren) {
        regtail (cp, ret, br);    /* OPEN -> first. */
    }
    else {
        ret = br;
    }
    *flagp &= ~(~flags & HASWIDTH);	/* Clear bit if bit 0. */
    *flagp |= flags & SPSTART;
    while (*cp->regparse == '|') {
        cp->regparse++;
        br = regbranch (cp, &flags);
        if (br == NULL) {
            return (NULL);
        }
        regtail (cp, ret, br);	/* BRANCH -> BRANCH. */
        *flagp &= ~(~flags & HASWIDTH);
        *flagp |= flags & SPSTART;
    }

    /* Make a closing node, and hook it on the end. */
    ender = regnode (cp, (paren) ? CLOSE + parno : END);
    regtail (cp, ret, ender);

    /* Hook the tails of the branches to the closing node. */
    for (br = ret; br != NULL; br = regnext (br)) {
        regoptail (cp, br, ender);
    }

    /* Check for proper termination. */
    if (paren && *cp->regparse++ != ')') {
        FAIL ("unterminated ()");
    }
    else if (!paren && *cp->regparse != '\0') {
        if (*cp->regparse == ')') {
            FAIL ("unmatched ()");
        }
        else {
            FAIL ("internal error: junk on end");
        }
        /* NOTREACHED */
    }

    return (ret);
}

/*
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
 */
static char *
regbranch (cp, flagp)
register struct comp *cp;
int *flagp;
{
    register char *ret;
    register char *chain;
    register char *latest;
    int flags;
    register int c;

    *flagp = WORST;		/* Tentatively. */

    ret = regnode (cp, BRANCH);
    chain = NULL;
    while ((c = *cp->regparse) != '\0' && c != '|' && c != ')') {
        latest = regpiece (cp, &flags);
        if (latest == NULL) {
            return (NULL);
        }
        *flagp |= flags & HASWIDTH;
        if (chain == NULL) {	/* First piece. */
            *flagp |= flags & SPSTART;
        }
        else {
            regtail (cp, chain, latest);
        }
        chain = latest;
    }
    if (chain == NULL) {	/* Loop ran zero times. */
        (void) regnode (cp, NOTHING);
    }

    return (ret);
}

/*
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
 */
static char *
regpiece (cp, flagp)
register struct comp *cp;
int *flagp;
{
    register char *ret;
    register char op;
    register char *next;
    int flags;

    ret = regatom (cp, &flags);
    if (ret == NULL) {
        return (NULL);
    }

    op = *cp->regparse;
    if (!ISREPN (op)) {
        *flagp = flags;
        return (ret);
    }

    if (!(flags & HASWIDTH) && op != '?') {
        FAIL ("*+ operand could be empty");
    }
    switch (op) {
    case '*':
        *flagp = WORST | SPSTART;
        break;
    case '+':
        *flagp = WORST | SPSTART | HASWIDTH;
        break;
    case '?':
        *flagp = WORST;
        break;
    }

    if (op == '*' && (flags & SIMPLE)) {
        reginsert (cp, STAR, ret);
    }
    else if (op == '*') {
        /* Emit x* as (x&|), where & means "self". */
        reginsert (cp, BRANCH, ret);	/* Either x */
        regoptail (cp, ret, regnode (cp, BACK));	/* and loop */
        regoptail (cp, ret, ret);	/* back */
        regtail (cp, ret, regnode (cp, BRANCH));	/* or */
        regtail (cp, ret, regnode (cp, NOTHING));	/* null. */
    }
    else if (op == '+' && (flags & SIMPLE)) {
        reginsert (cp, PLUS, ret);
    }
    else if (op == '+') {
        /* Emit x+ as x(&|), where & means "self". */
        next = regnode (cp, BRANCH);	/* Either */
        regtail (cp, ret, next);
        regtail (cp, regnode (cp, BACK), ret);	/* loop back */
        regtail (cp, next, regnode (cp, BRANCH));	/* or */
        regtail (cp, ret, regnode (cp, NOTHING));	/* null. */
    }
    else if (op == '?') {
        /* Emit x? as (x|) */
        reginsert (cp, BRANCH, ret);	/* Either x */
        regtail (cp, ret, regnode (cp, BRANCH));	/* or */
        next = regnode (cp, NOTHING);	/* null. */
        regtail (cp, ret, next);
        regoptail (cp, ret, next);
    }
    cp->regparse++;
    if (ISREPN (*cp->regparse)) {
        FAIL ("nested *?+");
    }

    return (ret);
}

/*
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary characters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed characters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
 */
static char *
regatom (cp, flagp)
register struct comp *cp;
int *flagp;
{
    register char *ret;
    int flags;

    *flagp = WORST;		/* Tentatively. */

    switch (*cp->regparse++) {
    case '^':
        ret = regnode (cp, BOL);
        break;
    case '$':
        ret = regnode (cp, EOL);
        break;
    case '.':
        ret = regnode (cp, ANY);
        *flagp |= HASWIDTH | SIMPLE;
        break;
    case '[': {
        register int range;
        register int rangeend;
        register int c;

        if (*cp->regparse == '^') {	/* Complement of range. */
            ret = regnode (cp, ANYBUT);
            cp->regparse++;
        }
        else {
            ret = regnode (cp, ANYOF);
        }
        if ((c = *cp->regparse) == ']' || c == '-') {
            regc (cp, c);
            cp->regparse++;
        }
        while ((c = *cp->regparse++) != '\0' && c != ']') {
            if (c != '-') {
                regc (cp, c);
            }
            else if ((c = *cp->regparse) == ']' || c == '\0') {
                regc (cp, '-');
            }
            else {
                range = (unsigned char) *(cp->regparse - 2);
                rangeend = (unsigned char) c;
                if (range > rangeend) {
                    FAIL ("invalid [] range");
                }
                for (range++; range <= rangeend; range++) {
                    regc (cp, range);
                }
                cp->regparse++;
            }
        }
        regc (cp, '\0');
        if (c != ']') {
            FAIL ("unmatched []");
        }
        *flagp |= HASWIDTH | SIMPLE;
        break;
    }
    case '(':
        ret = reg (cp, 1, &flags);
        if (ret == NULL) {
            return (NULL);
        }
        *flagp |= flags & (HASWIDTH | SPSTART);
        break;
    case '\0':
    case '|':
    case ')':
        /* supposed to be caught earlier */
        FAIL ("internal error: \\0|) unexpected");
        break;
    case '?':
    case '+':
    case '*':
        FAIL ("?+* follows nothing");
        break;
    case '\\':
        if (*cp->regparse == '\0') {
            FAIL ("trailing \\");
        }
        ret = regnode (cp, EXACTLY);
        regc (cp, *cp->regparse++);
        regc (cp, '\0');
        *flagp |= HASWIDTH | SIMPLE;
        break;
    default: {
        register size_t len;
        register char ender;

        cp->regparse--;
        len = strcspn (cp->regparse, META);
        if (len == 0) {
            FAIL ("internal error: strcspn 0");
        }
        ender = *(cp->regparse + len);
        if (len > 1 && ISREPN (ender)) {
            len--;    /* Back off clear of ?+* operand. */
        }
        *flagp |= HASWIDTH;
        if (len == 1) {
            *flagp |= SIMPLE;
        }
        ret = regnode (cp, EXACTLY);
        for (; len > 0; len--) {
            regc (cp, *cp->regparse++);
        }
        regc (cp, '\0');
        break;
    }
    }

    return (ret);
}

/*
 - regnode - emit a node
 */
// sergio: se pone de igual manera que el prototipo YA modificado
static char *			/* Location. */
regnode (cp, op)
register struct comp *cp;
char op;
//static char * regnode (register struct comp *cp, char op)
{
    register char *const ret = cp->regcode;
    register char *ptr;

    if (!EMITTING (cp)) {
        cp->regsize += 3;
        return (ret);
    }

    ptr = ret;
    *ptr++ = op;
    *ptr++ = '\0';		/* Null next pointer. */
    *ptr++ = '\0';
    cp->regcode = ptr;

    return (ret);
}

/*
 - regc - emit (if appropriate) a byte of code
 */
// sergio: se pone de igual manera que el prototipo YA modificado
static void
regc (cp, b)
register struct comp *cp;
char b;
//static void regc(register struct comp * cp, char b)
{
    if (EMITTING (cp)) {
        *cp->regcode++ = b;
    }
    else {
        cp->regsize++;
    }
}

/*
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
 */
// sergio: se pone de igual manera que el prototipo YA modificado
//static void
//reginsert(cp, op, opnd)
//register struct comp *cp;
//char op;
//char *opnd;
static void
reginsert (register struct comp *cp, char op, char *opnd)
{
    register char *place;

    if (!EMITTING (cp)) {
        cp->regsize += 3;
        return;
    }

    (void) memmove (opnd + 3, opnd, (size_t) (cp->regcode - opnd));
    cp->regcode += 3;

    place = opnd;			/* Op node, where operand used to be. */
    *place++ = op;
    *place++ = '\0';
    *place++ = '\0';
}

/*
 - regtail - set the next-pointer at the end of a node chain
 */
static void
regtail (cp, p, val)
register struct comp *cp;
char *p;
char *val;
{
    register char *scan;
    register char *temp;
    register int offset;

    if (!EMITTING (cp)) {
        return;
    }

    /* Find last node. */
    for (scan = p; (temp = regnext (scan)) != NULL; scan = temp) {
        continue;
    }

    offset = (OP (scan) == BACK) ? scan - val : val - scan;
    *(scan + 1) = (offset >> 8) & 0177;
    *(scan + 2) = offset & 0377;
}

/*
 - regoptail - regtail on operand of first argument; nop if operandless
 */
static void
regoptail (cp, p, val)
register struct comp *cp;
char *p;
char *val;
{
    /* "Operandless" and "op != BRANCH" are synonymous in practice. */
    if (!EMITTING (cp) || OP (p) != BRANCH) {
        return;
    }
    regtail (cp, OPERAND (p), val);
}

/*
 * sqd_regexec and friends
 */

/*
 * Work-variable struct for sqd_regexec().
 */
struct exec
{
    char *reginput;		/* String-input pointer. */
    char *regbol;			/* Beginning of input, for ^ check. */
    char **regstartp;		/* Pointer to startp array. */
    char **regendp;		/* Ditto for endp. */
};

/*
 * Forwards.
 */
static int regtry (struct exec *ep, sqd_regexp * rp, char *string);
static int regmatch (struct exec *ep, char *prog);
static size_t regrepeat (struct exec *ep, char *node);

#ifdef DEBUG
int regnarrate = 0;
void regdump ();
static char *regprop ();
#endif

/*
 - sqd_regexec - match a regexp against a string
 */
int
sqd_regexec (prog, str)
register sqd_regexp *prog;
const char *str;
{
    register char *string = (char *) str;	/* avert const poisoning */
    register char *s;
    struct exec ex;

    /* Be paranoid. */
    if (prog == NULL || string == NULL) {
        sqd_regerror ("NULL argument to sqd_regexec");
        return (0);
    }

    /* Check validity of program. */
    if ((unsigned char) *prog->program != SQD_REGMAGIC) {
        sqd_regerror ("corrupted regexp");
        return (0);
    }

    /* If there is a "must appear" string, look for it. */
    if (prog->regmust != NULL && strstr (string, prog->regmust) == NULL) {
        return (0);
    }

    /* Mark beginning of line for ^ . */
    ex.regbol = string;
    ex.regstartp = prog->startp;
    ex.regendp = prog->endp;

    /* Simplest case:  anchored match need be tried only once. */
    if (prog->reganch) {
        return (regtry (&ex, prog, string));
    }

    /* Messy cases:  unanchored match. */
    if (prog->regstart != '\0') {
        /* We know what char it must start with. */
        for (s = string; s != NULL; s = strchr (s + 1, prog->regstart)) {
            if (regtry (&ex, prog, s)) {
                return (1);
            }
				}		
        return (0);
    }
    else {
        /* We don't -- general case. */
        for (s = string; !regtry (&ex, prog, s); s++) {
            if (*s == '\0') {
                return (0);
            }
				}		
        return (1);
    }
    /* NOTREACHED */
}

/*
 - regtry - try match at specific point
 */
static int			/* 0 failure, 1 success */
regtry (ep, prog, string)
register struct exec *ep;
sqd_regexp *prog;
char *string;
{
    register int i;
    register char **stp;
    register char **enp;

    ep->reginput = string;

    stp = prog->startp;
    enp = prog->endp;
    for (i = NSUBEXP; i > 0; i--) {
        *stp++ = NULL;
        *enp++ = NULL;
    }
    if (regmatch (ep, prog->program + 1)) {
        prog->startp[0] = string;
        prog->endp[0] = ep->reginput;
        return (1);
    }
    else {
        return (0);
    }
}

/*
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
 */
static int			/* 0 failure, 1 success */
regmatch (ep, prog)
register struct exec *ep;
char *prog;
{
    register char *scan;		/* Current node. */
    char *next;			/* Next node. */

#ifdef DEBUG
    if (prog != NULL && regnarrate) {
        fprintf (stderr, "%s(\n", regprop (prog));
    }
#endif
    for (scan = prog; scan != NULL; scan = next) {
#ifdef DEBUG
        if (regnarrate) {
            fprintf (stderr, "%s...\n", regprop (scan));
        }
#endif
        next = regnext (scan);

        switch (OP (scan)) {
        case BOL:
            if (ep->reginput != ep->regbol) {
                return (0);
            }
            break;
        case EOL:
            if (*ep->reginput != '\0') {
                return (0);
            }
            break;
        case ANY:
            if (*ep->reginput == '\0') {
                return (0);
            }
            ep->reginput++;
            break;
        case EXACTLY: {
            register size_t len;
            register char *const opnd = OPERAND (scan);

            /* Inline the first character, for speed. */
            if (*opnd != *ep->reginput) {
                return (0);
            }
            len = strlen (opnd);
            if (len > 1 && strncmp (opnd, ep->reginput, len) != 0) {
                return (0);
            }
            ep->reginput += len;
            break;
        }
        case ANYOF:
            if (*ep->reginput == '\0' ||
                    strchr (OPERAND (scan), *ep->reginput) == NULL) {
                return (0);
            }
            ep->reginput++;
            break;
        case ANYBUT:
            if (*ep->reginput == '\0' ||
                    strchr (OPERAND (scan), *ep->reginput) != NULL) {
                return (0);
            }
            ep->reginput++;
            break;
        case NOTHING:
            break;
        case BACK:
            break;
        case OPEN + 1:
        case OPEN + 2:
        case OPEN + 3:
        case OPEN + 4:
        case OPEN + 5:
        case OPEN + 6:
        case OPEN + 7:
        case OPEN + 8:
        case OPEN + 9: {
            register const int no = OP (scan) - OPEN;
            register char *const input = ep->reginput;

            if (regmatch (ep, next)) {
                /*
                 * Don't set startp if some later
                 * invocation of the same parentheses
                 * already has.
                 */
                if (ep->regstartp[no] == NULL) {
                    ep->regstartp[no] = input;
                }
                return (1);
            }
            else {
                return (0);
            }
            break;
        }
        case CLOSE + 1:
        case CLOSE + 2:
        case CLOSE + 3:
        case CLOSE + 4:
        case CLOSE + 5:
        case CLOSE + 6:
        case CLOSE + 7:
        case CLOSE + 8:
        case CLOSE + 9: {
            register const int no = OP (scan) - CLOSE;
            register char *const input = ep->reginput;

            if (regmatch (ep, next)) {
                /*
                 * Don't set endp if some later
                 * invocation of the same parentheses
                 * already has.
                 */
                if (ep->regendp[no] == NULL) {
                    ep->regendp[no] = input;
                }
                return (1);
            }
            else {
                return (0);
            }
            break;
        }
        case BRANCH: {
            register char *const save = ep->reginput;

            if (OP (next) != BRANCH) {	/* No choice. */
                next = OPERAND (scan);    /* Avoid recursion. */
            }
            else {
                while (OP (scan) == BRANCH) {
                    if (regmatch (ep, OPERAND (scan))) {
                        return (1);
                    }
                    ep->reginput = save;
                    scan = regnext (scan);
                }
                return (0);
                /* NOTREACHED */
            }
            break;
        }
        case STAR:
        case PLUS: {
            register const char nextch =
                (OP (next) == EXACTLY) ? *OPERAND (next) : '\0';
            register size_t no;
            register char *const save = ep->reginput;
            register const size_t min = (OP (scan) == STAR) ? 0 : 1;

            for (no = regrepeat (ep, OPERAND (scan)) + 1; no > min; no--) {
                ep->reginput = save + no - 1;
                /* If it could work, try it. */
                if (nextch == '\0' || *ep->reginput == nextch)
                    if (regmatch (ep, next)) {
                        return (1);
                    }
            }
            return (0);
            break;
        }
        case END:
            return (1);		/* Success! */
            break;
        default:
            sqd_regerror ("regexp corruption");
            return (0);
            break;
        }
    }

    /*
     * We get here only if there's trouble -- normally "case END" is
     * the terminating point.
     */
    sqd_regerror ("corrupted pointers");
    return (0);
}

/*
 - regrepeat - report how many times something simple would match
 */
static size_t
regrepeat (ep, node)
register struct exec *ep;
char *node;
{
    register size_t count;
    register char *scan;
    register char ch;

    switch (OP (node)) {
    case ANY:
        return (strlen (ep->reginput));
        break;
    case EXACTLY:
        ch = *OPERAND (node);
        count = 0;
        for (scan = ep->reginput; *scan == ch; scan++) {
            count++;
        }
        return (count);
        break;
    case ANYOF:
        return (strspn (ep->reginput, OPERAND (node)));
        break;
    case ANYBUT:
        return (strcspn (ep->reginput, OPERAND (node)));
        break;
    default:			/* Oh dear.  Called inappropriately. */
        sqd_regerror ("internal error: bad call of regrepeat");
        return (0);			/* Best compromise. */
        break;
    }
    /* NOTREACHED */
}

/*
 - regnext - dig the "next" pointer out of a node
 */
static char *
regnext (p)
register char *p;
{
    register const int offset = NEXT (p);

    if (offset == 0) {
        return (NULL);
    }

    return ((OP (p) == BACK) ? p - offset : p + offset);
}

#ifdef DEBUG

static char *regprop ();

/*
 - regdump - dump a regexp onto stdout in vaguely comprehensible form
 */
void
regdump (r)
sqd_regexp *r;
{
    register char *s;
    register char op = EXACTLY;	/* Arbitrary non-END op. */
    register char *next;


    s = r->program + 1;
    while (op != END) {		/* While that wasn't END last time... */
        op = OP (s);
        printf ("%2d%s", s - r->program, regprop (s));	/* Where, what. */
        next = regnext (s);
        if (next == NULL) {	/* Next ptr. */
            printf ("(0)");
        }
        else {
            printf ("(%d)", (s - r->program) + (next - s));
        }
        s += 3;
        if (op == ANYOF || op == ANYBUT || op == EXACTLY) {
            /* Literal string, where present. */
            while (*s != '\0') {
                putchar (*s);
                s++;
            }
            s++;
        }
        putchar ('\n');
    }

    /* Header fields of interest. */
    if (r->regstart != '\0') {
        printf ("start `%c' ", r->regstart);
    }
    if (r->reganch) {
        printf ("anchored ");
    }
    if (r->regmust != NULL) {
        printf ("must have \"%s\"", r->regmust);
    }
    printf ("\n");
}

/*
 - regprop - printable representation of opcode
 */
static char *
regprop (op)
char *op;
{
    register char *p;
    static char buf[50];

    (void) strcpy (buf, ":");

    switch (OP (op)) {
    case BOL:
        p = "BOL";
        break;
    case EOL:
        p = "EOL";
        break;
    case ANY:
        p = "ANY";
        break;
    case ANYOF:
        p = "ANYOF";
        break;
    case ANYBUT:
        p = "ANYBUT";
        break;
    case BRANCH:
        p = "BRANCH";
        break;
    case EXACTLY:
        p = "EXACTLY";
        break;
    case NOTHING:
        p = "NOTHING";
        break;
    case BACK:
        p = "BACK";
        break;
    case END:
        p = "END";
        break;
    case OPEN + 1:
    case OPEN + 2:
    case OPEN + 3:
    case OPEN + 4:
    case OPEN + 5:
    case OPEN + 6:
    case OPEN + 7:
    case OPEN + 8:
    case OPEN + 9:
        sprintf (buf + strlen (buf), "OPEN%d", OP (op) - OPEN);
        p = NULL;
        break;
    case CLOSE + 1:
    case CLOSE + 2:
    case CLOSE + 3:
    case CLOSE + 4:
    case CLOSE + 5:
    case CLOSE + 6:
    case CLOSE + 7:
    case CLOSE + 8:
    case CLOSE + 9:
        sprintf (buf + strlen (buf), "CLOSE%d", OP (op) - CLOSE);
        p = NULL;
        break;
    case STAR:
        p = "STAR";
        break;
    case PLUS:
        p = "PLUS";
        break;
    default:
        sqd_regerror ("corrupted opcode");
        break;
    }
    if (p != NULL) {
        (void) strcat (buf, p);
    }
    return (buf);
}
#endif


/*
 - sqd_regsub - perform substitutions after a regexp match
 */
void
sqd_regsub (rp, source, dest)
const sqd_regexp *rp;
const char *source;
char *dest;
{
    register sqd_regexp *const prog = (sqd_regexp *) rp;
    register char *src = (char *) source;
    register char *dst = dest;
    register char c;
    register int no;
    register size_t len;

    if (prog == NULL || source == NULL || dest == NULL) {
        sqd_regerror ("NULL parameter to sqd_regsub");
        return;
    }
    if ((unsigned char) *(prog->program) != SQD_REGMAGIC) {
        sqd_regerror ("damaged regexp");
        return;
    }

    while ((c = *src++) != '\0') {
        if (c == '&') {
            no = 0;
        }
        else if (c == '\\' && isdigit ((int) (*src))) {
            no = *src++ - '0';
        }
        else {
            no = -1;
        }

        if (no < 0) {		/* Ordinary character. */
            if (c == '\\' && (*src == '\\' || *src == '&')) {
                c = *src++;
            }
            *dst++ = c;
        }
        else if (prog->startp[no] != NULL && prog->endp[no] != NULL &&
                 prog->endp[no] > prog->startp[no]) {
            len = prog->endp[no] - prog->startp[no];
            (void) strncpy (dst, prog->startp[no], len);
            dst += len;
            if (*(dst - 1) == '\0') {	/* strncpy hit NUL. */
                sqd_regerror ("damaged match string");
                return;
            }
        }
    }
    *dst++ = '\0';
}


void
sqd_regerror (s)
char *s;
{
    fprintf (stderr, "regexp(3): %s\n", s);
    exit (EXIT_FAILURE);
    /* NOTREACHED */
}

#ifdef NBA_TEAM_IN_STL
int
main (int argc, char **argv)
{
    char *pat;
    int ntok;
    char *s;
    int status;

    pat = argv[1];
    ntok = atoi (argv[2]);
    s = argv[3];

    status = Strparse (pat, s, ntok);
    if (status == 0) {
        printf ("no match\n");
    }
    else {
        int i;
        printf ("MATCH.\n");
        for (i = 1; i <= ntok; i++) {
            printf ("matched token %1d:  %s\n", i, sqd_parse[i]);
        }
    }
}
#endif /*NBA_TEAM_IN_STL */
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* iupac.c
 *
 * Globally defines the IUPAC symbols for nucleic acid sequence
 * Slowly evolving into a repository of globals. Tue Apr 20 1993
 *
 * CVS $Id: iupac.c,v 1.4 2003/04/14 16:00:16 eddy Exp $
 */
#include "squidconf.h"
#include "squid.h"

/* Default expected nucleotide occurrence frequencies, A/C/G/T.
 * Used (for instance) as the default distribution for
 * i.i.d. random nucleotide sequences.
 */
float dnafq[4] = { 0.25, 0.25, 0.25, 0.25 };

/* Dayhoff f(i) amino acid occurrence frequencies.
 * From SwissProt 34: 21,210,388 residues
 * In alphabetic order by single-letter code.
 * Used (for instance) as the default distribution for
 * i.i.d. random protein sequences.
 */
float aafq[20] = {
    0.075520,			/* A */
    0.016973,			/* C */
    0.053029,			/* D */
    0.063204,			/* E */
    0.040762,			/* F */
    0.068448,			/* G */
    0.022406,			/* H */
    0.057284,			/* I */
    0.059398,			/* K */
    0.093399,			/* L */
    0.023569,			/* M */
    0.045293,			/* N */
    0.049262,			/* P */
    0.040231,			/* Q */
    0.051573,			/* R */
    0.072214,			/* S */
    0.057454,			/* T */
    0.065252,			/* V */
    0.012513,			/* W */
    0.031985			/* Y */
};

char aa_alphabet[] = AMINO_ALPHABET;
/* aa_index converts to pam's 27x27 scheme */
int aa_index[20] = { 0, 2, 3, 4, 5, 6, 7, 8, 10, 11,
                     12, 13, 15, 16, 17, 18, 19, 21, 22, 24
                   };

/* IUPAC code translations */
/* note: sequence chars are UPPER CASE */
struct iupactype iupac[] = {
    {'A', 'T', NTA, NTT,},
    {'C', 'G', NTC, NTG,},
    {'G', 'C', NTG, NTC,},
    {'T', 'A', NTT, NTA,},
    {'U', 'A', NTU, NTA,},
    {'N', 'N', NTN, NTN,},
    {' ', ' ', NTGAP, NTGAP,},
    {'R', 'Y', NTR, NTY,},
    {'Y', 'R', NTY, NTR,},
    {'M', 'K', NTM, NTK,},
    {'K', 'M', NTK, NTM,},
    {'S', 'S', NTS, NTS,},
    {'W', 'W', NTW, NTW,},
    {'H', 'D', NTH, NTD,},
    {'B', 'V', NTB, NTV,},
    {'V', 'B', NTV, NTB,},
    {'D', 'H', NTD, NTH,},
};


char *stdcode1[65] = {
    "K",				/* AAA */
    "N",				/* AAC */
    "K",				/* AAG */
    "N",				/* AAU */
    "T",				/* ACA */
    "T",				/* ACC */
    "T",				/* ACG */
    "T",				/* ACU */
    "R",				/* AGA */
    "S",				/* AGC */
    "R",				/* AGG */
    "S",				/* AGU */
    "I",				/* AUA */
    "I",				/* AUC */
    "M",				/* AUG */
    "I",				/* AUU */
    "Q",				/* CAA */
    "H",				/* CAC */
    "Q",				/* CAG */
    "H",				/* CAU */
    "P",				/* CCA */
    "P",				/* CCC */
    "P",				/* CCG */
    "P",				/* CCU */
    "R",				/* CGA */
    "R",				/* CGC */
    "R",				/* CGG */
    "R",				/* CGU */
    "L",				/* CUA */
    "L",				/* CUC */
    "L",				/* CUG */
    "L",				/* CUU */
    "E",				/* GAA */
    "D",				/* GAC */
    "E",				/* GAG */
    "D",				/* GAU */
    "A",				/* GCA */
    "A",				/* GCC */
    "A",				/* GCG */
    "A",				/* GCU */
    "G",				/* GGA */
    "G",				/* GGC */
    "G",				/* GGG */
    "G",				/* GGU */
    "V",				/* GUA */
    "V",				/* GUC */
    "V",				/* GUG */
    "V",				/* GUU */
    "*",				/* UAA */
    "Y",				/* UAC */
    "*",				/* UAG */
    "Y",				/* UAU */
    "S",				/* UCA */
    "S",				/* UCC */
    "S",				/* UCG */
    "S",				/* UCU */
    "*",				/* UGA */
    "C",				/* UGC */
    "W",				/* UGG */
    "C",				/* UGU */
    "L",				/* UUA */
    "F",				/* UUC */
    "L",				/* UUG */
    "F",				/* UUU */
    "X",				/* unknown */
};




char *stdcode3[65] = {
    "Lys",			/* AAA */
    "Asn",			/* AAC */
    "Lys",			/* AAG */
    "Asn",			/* AAU */
    "Thr",			/* ACA */
    "Thr",			/* ACC */
    "Thr",			/* ACG */
    "Thr",			/* ACU */
    "Arg",			/* AGA */
    "Ser",			/* AGC */
    "Arg",			/* AGG */
    "Ser",			/* AGU */
    "Ile",			/* AUA */
    "Ile",			/* AUC */
    "Met",			/* AUG */
    "Ile",			/* AUU */
    "Gln",			/* CAA */
    "His",			/* CAC */
    "Gln",			/* CAG */
    "His",			/* CAU */
    "Pro",			/* CCA */
    "Pro",			/* CCC */
    "Pro",			/* CCG */
    "Pro",			/* CCU */
    "Arg",			/* CGA */
    "Arg",			/* CGC */
    "Arg",			/* CGG */
    "Arg",			/* CGU */
    "Leu",			/* CUA */
    "Leu",			/* CUC */
    "Leu",			/* CUG */
    "Leu",			/* CUU */
    "Glu",			/* GAA */
    "Asp",			/* GAC */
    "Glu",			/* GAG */
    "Asp",			/* GAU */
    "Ala",			/* GCA */
    "Ala",			/* GCC */
    "Ala",			/* GCG */
    "Ala",			/* GCU */
    "Gly",			/* GGA */
    "Gly",			/* GGC */
    "Gly",			/* GGG */
    "Gly",			/* GGU */
    "Val",			/* GUA */
    "Val",			/* GUC */
    "Val",			/* GUG */
    "Val",			/* GUU */
    "***",			/* UAA */
    "Tyr",			/* UAC */
    "***",			/* UAG */
    "Tyr",			/* UAU */
    "Ser",			/* UCA */
    "Ser",			/* UCC */
    "Ser",			/* UCG */
    "Ser",			/* UCU */
    "***",			/* UGA */
    "Cys",			/* UGC */
    "Trp",			/* UGG */
    "Cys",			/* UGU */
    "Leu",			/* UUA */
    "Phe",			/* UUC */
    "Leu",			/* UUG */
    "Trp",			/* UUU */
    "XXX",			/* unknown */
};

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* masks.c
 * SRE, Tue Nov 18 10:12:28 1997
 *
 * Sequence masking routines; corrections for biased composition
 * target sequences.
 *
 * The Claverie/States XNU code is not used by default because I
 * consider X'ing out sequence to be too black/white and too
 * aggressive, but it's available as an option.
 *
 * The Wooton/Federhen SEG code was studied, but deemed too
 * nonportable to include; it would've suffered the same drawback
 * as XNU.
 *
 * The TraceScoreCorrection() code is the default.
 *
 * CVS $Id: masks.c,v 1.15 2003/04/14 16:00:17 eddy Exp $
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <math.h>
#include <float.h>

#include "squid.h"
#include "structs.h"
#include "funcs.h"

/* The PAM120 score matrix, in HMMER's AMINO_ALPHABET alphabetic order
 */
static int xpam120[23][23] = {
    {   3, -3, 0, 0, -4, 1, -3, -1, -2, -3, -2, -1, 1, -1, -3, 1, 1, 0, -7, -4, 1,
        0, 0
    },
    {   -3, 9, -7, -7, -6, -4, -4, -3, -7, -7, -6, -5, -4, -7, -4, 0, -3, -3, -8,
        -1, -4, -6, 0
    },
    {   0, -7, 5, 3, -7, 0, 0, -3, -1, -5, -4, 2, -3, 1, -3, 0, -1, -3, -8, -5, 5,
        3, 0
    },
    {   0, -7, 3, 5, -7, -1, -1, -3, -1, -4, -3, 1, -2, 2, -3, -1, -2, -3, -8, -5,
        3, 5, 0
    },
    {   -4, -6, -7, -7, 8, -5, -3, 0, -7, 0, -1, -4, -5, -6, -5, -3, -4, -3, -1, 4,
        -4, -5, 0
    },
    {   1, -4, 0, -1, -5, 5, -4, -4, -3, -5, -4, 0, -2, -3, -4, 1, -1, -2, -8, -6,
        1, -1, 0
    },
    {   -3, -4, 0, -1, -3, -4, 7, -4, -2, -3, -4, 2, -1, 3, 1, -2, -3, -3, -3, -1,
        2, 2, 0
    },
    {   -1, -3, -3, -3, 0, -4, -4, 6, -3, 1, 1, -2, -3, -3, -2, -2, 0, 3, -6, -2,
        -2, -2, 0
    },
    {   -2, -7, -1, -1, -7, -3, -2, -3, 5, -4, 0, 1, -2, 0, 2, -1, -1, -4, -5, -5,
        1, 0, 0
    },
    {   -3, -7, -5, -4, 0, -5, -3, 1, -4, 5, 3, -4, -3, -2, -4, -4, -3, 1, -3, -2,
        -3, -2, 0
    },
    {   -2, -6, -4, -3, -1, -4, -4, 1, 0, 3, 8, -3, -3, -1, -1, -2, -1, 1, -6, -4,
        -3, -1, 0
    },
    {   -1, -5, 2, 1, -4, 0, 2, -2, 1, -4, -3, 4, -2, 0, -1, 1, 0, -3, -4, -2, 4,
        1, 0
    },
    {   1, -4, -3, -2, -5, -2, -1, -3, -2, -3, -3, -2, 6, 0, -1, 1, -1, -2, -7, -6,
        -1, 0, 0
    },
    {   -1, -7, 1, 2, -6, -3, 3, -3, 0, -2, -1, 0, 0, 6, 1, -2, -2, -3, -6, -5, 1,
        5, 0
    },
    {   -3, -4, -3, -3, -5, -4, 1, -2, 2, -4, -1, -1, -1, 1, 6, -1, -2, -3, 1, -5,
        -1, 0, 0
    },
    {   1, 0, 0, -1, -3, 1, -2, -2, -1, -4, -2, 1, 1, -2, -1, 3, 2, -2, -2, -3, 1,
        0, 0
    },
    {   1, -3, -1, -2, -4, -1, -3, 0, -1, -3, -1, 0, -1, -2, -2, 2, 4, 0, -6, -3,
        1, -1, 0
    },
    {   0, -3, -3, -3, -3, -2, -3, 3, -4, 1, 1, -3, -2, -3, -3, -2, 0, 5, -8, -3,
        -2, -2, 0
    },
    {   -7, -8, -8, -8, -1, -8, -3, -6, -5, -3, -6, -4, -7, -6, 1, -2, -6, -8, 12,
        -2, -5, -6, 0
    },
    {   -4, -1, -5, -5, 4, -6, -1, -2, -5, -2, -4, -2, -6, -5, -5, -3, -3, -3, -2,
        8, -2, -4, 0
    },
    {   1, -4, 5, 3, -4, 1, 2, -2, 1, -3, -3, 4, -1, 1, -1, 1, 1, -2, -5, -2, 6, 4,
        0
    },
    {   0, -6, 3, 5, -5, -1, 2, -2, 0, -2, -1, 1, 0, 5, 0, 0, -1, -2, -6, -4, 4, 6,
        0
    },
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
};


/* Function: XNU()
 * Date:     18 Nov 1997 [StL]
 *
 * Purpose:  x-out of repetitive sequence. XNU tends to be
 *           good at x'ing out short period tandem repeats.
 *
 * Note:     Apply /only/ to protein sequence.
 *
 * Args:     dsq: 1..len digitized sequence
 *           len: length of dsq
 *
 * Return:   number of characters x'ed out.
 */
int
XNU (char *dsq, int len)
{
    int i, k, off, sum, beg, end, top;
    int topcut, fallcut;
    double s0;
    int noff = 4;			/* maximum search offset */
    int mcut = 1;
    double pcut = 0.01;
    int *hit;
    double lambda = 0.346574;
    double K = 0.2;
    double Hconst = 0.664;
    int xnum = 0;

    if (len == 0) {
        return 0;
    }

    hit = MallocOrDie (sizeof (int) * (len + 1));
    for (i = 1; i <= len; i++) {
        hit[i] = 0;
    }

    /*
     ** Determine the score cutoff so that pcut will be the fraction
     ** of random sequence eliminated assuming lambda, K, and Hconst are
     ** characteristic of the database as a whole
     */
    s0 = -log (pcut * Hconst / (noff * K)) / lambda;
    if (s0 > 0) {
        topcut = floor (s0 + log (s0) / lambda + 0.5);
    }
    else {
        topcut = 0;
    }
    fallcut = (int) log (K / 0.001) / lambda;

    for (off = mcut; off <= noff; off++) {
        sum = top = 0;
        beg = off;
        end = 0;

        for (i = off + 1; i <= len; i++) {
            sum += xpam120[(int) dsq[i]][(int) dsq[i - off]];
            if (sum > top) {
                top = sum;
                end = i;
            }
            if (top >= topcut && top - sum > fallcut) {
                for (k = beg; k <= end; k++) {
                    hit[k] = hit[k - off] = 1;
                }
                sum = top = 0;
                beg = end = i + 1;
            }
            else if (top - sum > fallcut) {
                sum = top = 0;
                beg = end = i + 1;
            }
            if (sum < 0) {
                beg = end = i + 1;
                sum = top = 0;
            }
        }
        if (top >= topcut) {
            for (k = beg; k <= end; k++) {
                hit[k] = hit[k - off] = 1;
            }
        }
    }

    /* Now mask off detected repeats
     */
    for (i = 1; i <= len; i++) {
        if (hit[i]) {
            xnum++;
            dsq[i] = Alphabet_iupac - 1;
        }				/* e.g. 'X' */
		}
    free (hit);
    return xnum;
}


/* Function: TraceScoreCorrection()
 * Date:     Sun Dec 21 12:05:47 1997 [StL]
 *
 * Purpose:  Calculate a correction (in integer log_2 odds) to be
 *           applied to a sequence, using a second null model,
 *           based on a traceback. M/I emissions are corrected;
 *           C/N/J are not -- as if the nonmatching part and
 *           matching part were each generated by the best null model.
 *           The null model is constructed /post hoc/ as the
 *           average over all the M,I distributions used by the trace.
 *
 * Return:   the log_2-odds score correction.
 */
float
TraceScoreCorrection (struct plan7_s *hmm, struct p7trace_s *tr, char *dsq)
{
    float p[MAXABET];		/* null model distribution */
    int sc[MAXCODE];		/* null model scores       */
    int x;
    int tpos;
    int score;

    /* Rarely, the alignment was totally impossible, and tr is NULL.
     */
    if (tr == NULL) {
        return 0.0;
    }

    /* Set up model: average over the emission distributions of
     * all M, I states that appear in the trace. Ad hoc? Sure, you betcha.
     */
    FSet (p, Alphabet_size, 0.0);
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        if (tr->statetype[tpos] == STM) {
            FAdd (p, hmm->mat[tr->nodeidx[tpos]], Alphabet_size);
        }
        else if (tr->statetype[tpos] == STI) {
            FAdd (p, hmm->ins[tr->nodeidx[tpos]], Alphabet_size);
        }
		}		
    FNorm (p, Alphabet_size);

    for (x = 0; x < Alphabet_size; x++) {
        sc[x] = Prob2Score (p[x], hmm->null[x]);
    }
    /* could avoid this chunk if we knew
       we didn't need any degenerate char scores */
    for (x = Alphabet_size; x < Alphabet_iupac; x++) {
        sc[x] = DegenerateSymbolScore (p, hmm->null, x);
    }


    /* Score all the M,I state emissions that appear in the trace.
     */
    score = 0;
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        if (tr->statetype[tpos] == STM || tr->statetype[tpos] == STI) {
            score += sc[(int) dsq[tr->pos[tpos]]];
        }
		}
    /* Apply an ad hoc 8 bit fudge factor penalty;
     * interpreted as a prior, saying that the second null model is
     * 1/2^8 (1/256) as likely as the standard null model
     */
    score -= 8 * INTSCALE;

    /* Return the correction to the bit score.
     */
    return Scorify (ILogsum (0, score));
}


/* THE FOLLOWING CODE IS IN DEVELOPMENT.
 * it is commented out of the current release deliberately.
 * If you activate it, I'm not responsible for the consequences.
 */
#if MICHAEL_JORDAN_BUYS_THE_PACERS
/* Function: NewTraceScoreCorrection()
 * Date:     Wed Feb 17 14:32:45 1999 [StL]
 *
 * Purpose:  Calculate a correction (in integer log_2 odds) to be
 *           applied to a sequence, using a second null model,
 *           based on sequence endpoints. M/I emissions are corrected;
 *           C/N/J are not -- as if the nonmatching part and
 *           matching part were each generated by the best null model.
 *           Each null model is constructed /post hoc/ from the
 *           sequence composition of each matching domain (e.g.
 *           a null2 model is constructed for each domain in a
 *           multihit trace).
 *
 *           Constraints on the construction of this function include:
 *            1) Paracel hardware can't deal with trace-dependent
 *               null2 models. Original implementation of
 *               TraceScoreCorrection() was dependent on traceback
 *               and could not be reproduced on GeneMatcher.
 *               GeneMatcher may be able to deal w/ sequence endpoint
 *               dependent rescoring, though.
 *               Although this function looks like it's trace-
 *               dependent (because it's being passed a p7trace_s
 *               structure), it's really not; only the sequence
 *               endpoints are being used.
 *
 *            2) It is desirable that for multihit traces,
 *               per-domain scores sum to the per-sequence score.
 *               Otherwise people see this as a "bug" (cf.
 *               bug #2, David Kerk, NRC). HMMER calculates the
 *               per-domain scores by going through a separate
 *               TraceScore() call for each one and separately
 *               correcting them with TraceScoreCorrection(),
 *               so we have to do each domain in a full trace
 *               by a similar mechanism -- even if this means that
 *               we're adopting a very dubiously post hoc
 *               null model.
 *
 * Return:   the log_2-odds score correction.
 */
float
NewTraceScoreCorrection (struct plan7_s *hmm, struct p7trace_s *tr, char *dsq)
{
    float ct[MAXABET];		/* counts of observed residues            */
    float p[MAXABET];		/* null2 model distribution (also counts) */
    float sc[MAXCODE];		/* null2 model scores (as floats not int) */

    int x;
    int tpos;
    int score;			/* tmp score for real HMM, integer logodds */
    float hmmscore;		/* score for real HMM for this domain */
    float null2score;		/* score for null2 model for this domain */


    float totscore;		/* overall score for trace */
    float maxscore;		/* best score so far for single domain */
    int in_domain;		/* flag for whether we're counting this domain */
    int sym;			/* digitized symbol in dsq */
    int ndom;			/* number of domains counted towards score */

    int nsym;			/* number of symbols in this alignment */

    totscore = 0.;
    maxscore = -FLT_MAX;
    in_domain = FALSE;
    ndom = 0;
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        /* detect start of domain; start at N or J */
        if (tpos < tr->tlen - 1 && tr->statetype[tpos + 1] == STB) {
            FCopy (ct, hmm->null, Alphabet_size);	/* simple Dirichlet prior */
            score = 0;
            null2score = 0.;
            nsym = 0;
            in_domain = TRUE;
        }
        /* Count stuff in domain starting with N->B or J->B transition */
        if (in_domain) {
            sym = (int) dsq[tr->pos[tpos]];

            /* count emitted symbols in domain */
            if (tr->statetype[tpos] == STM || tr->statetype[tpos] == STI) {
                P7CountSymbol (ct, sym, 1.0);
                nsym++;
            }

            /* score emitted symbols in domain towards HMM */
            if (tr->statetype[tpos] == STM) {
                score += hmm->msc[sym][tr->nodeidx[tpos]];
            }
            else if (tr->statetype[tpos] == STI) {
                score += hmm->isc[sym][tr->nodeidx[tpos]];
            }
            /* score transitions in domain towards HMM */
            score += TransitionScoreLookup (hmm,
                                            tr->statetype[tpos], tr->nodeidx[tpos],
                                            tr->statetype[tpos + 1],
                                            tr->nodeidx[tpos + 1]);
        }


        if (tr->statetype[tpos] == STE) {	/* done w/ a domain; calc its score */
            /* convert counts to null2 prob distribution */
            FCopy (p, ct, Alphabet_size);
            FNorm (p, Alphabet_size);
            /* Convert probs to log-odds_e scores */
            /* p can't be zero, because of prior  */
            for (x = 0; x < Alphabet_size; x++) {
                sc[x] = log (p[x] / hmm->null[x]);
            }
            /* null2 score = counts \dot scores */
            null2score = FDot (ct, sc, Alphabet_size);

            printf ("NSYM = %d   NULL2 = %.1f\n", nsym, null2score);

            /* Apply an ad hoc 12 bit fudge factor penalty, per domain.
             * Interpreted probabilistically, saying that there's about
             * a 1/256 probability to transition into the second null model.
             */
            null2score -= 12.;

            /* Now correct score1 using the null2 score.
             * If it's still > 0, add it to accumulated score.
             */
            hmmscore = Scorify (score);
            hmmscore -= 1.44269504 * LogSum (0, null2score);
            if (hmmscore > 0.) {
                totscore += hmmscore;
                ndom++;
            }
            if (hmmscore > maxscore) {
                maxscore = hmmscore;
            }

            in_domain = FALSE;
        }
    }

    /* Single domain special case.
     */
    if (ndom == 0) {
        totscore = maxscore;
    }

    /* Return the correction to the bit score
     */
    return (P7TraceScore (hmm, dsq, tr) - totscore);
}
#endif /*0 */


float
SantaCruzCorrection (struct plan7_s *hmm, struct p7trace_s *tr, char *dsq)
{
    return 0.0;			/* UNFINISHED CODE */
}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/


/* mathsupport.c
 * SRE, Mon Nov 11 15:07:33 1996
 *
 * Miscellaneous mathematical functions.
 * General functions are in the SQUID library sre_math.c.
 * These functions are too HMM-specific to warrant being in the
 * SQUID library.
 *
 */


#include "config.h"
#include "squidconf.h"

#include <math.h>
#include <float.h>
#ifdef HMMER_THREADS
#include <pthread.h>
#endif

#include "squid.h"
#include "funcs.h"
#include "structs.h"


/* Function: Prob2Score()
 *
 * Purpose:  Convert a probability to a scaled integer log_2 odds score.
 *           Round to nearest integer (i.e. note use of +0.5 and floor())
 *           Return the score.
 */
int
Prob2Score (float p, float null)
{
    if (p == 0.0) {
        return -INFTY;
    }
    else {
        return (int) floor (0.5 + INTSCALE * sreLOG2 (p / null));
    }
}

/* Function: Score2Prob()
 *
 * Purpose:  Convert an integer log_2 odds score back to a probability;
 *           needs the null model probability, if any, to do the conversion.
 */
float
Score2Prob (int sc, float null)
{
    if (sc == -INFTY) {
        return 0.;
    }
    else {
        return (null * sreEXP2 ((float) sc / INTSCALE));
    }
}


/* Function: Scorify()
 *
 * Purpose:  Convert a scaled integer log-odds score to a floating
 *           point score for output. (could be a macro but who cares.)
 */
float
Scorify (int sc)
{
    return ((float) sc / INTSCALE);
}


/* Function: PValue()
 * Date:     SRE, Mon Oct 27 12:21:02 1997 [Sanger Centre, UK]
 *
 * Purpose:  Convert an HMM score to a P-value.
 *           We know P(S>x) is bounded by 1 / (1 + exp_2^x) for a bit score of x.
 *           We can also use EVD parameters for a tighter bound if we have
 *           them available.
 *
 * Args:     hmm - model structure, contains EVD parameters
 *           sc  - score in bits
 *
 * Returns:  P value for score significance.
 */
double
PValue (struct plan7_s *hmm, float sc)
{
    double pval;
    double pval2;

    /* the bound from Bayes */
    if (sc >= sreLOG2 (DBL_MAX)) {
        pval = 0.0;
    }
    else if (sc <= -1. * sreLOG2 (DBL_MAX)) {
        pval = 1.0;
    }
    else {
        pval = 1. / (1. + sreEXP2 (sc));
    }

    /* try for a better estimate from EVD fit */
    if (hmm != NULL && (hmm->flags & PLAN7_STATS)) {
        pval2 = ExtremeValueP (sc, hmm->mu, hmm->lambda);
        if (pval2 < pval) {
            pval = pval2;
        }
    }
    return pval;
}

/* Function: LogSum()
 *
 * Purpose:  Returns the log of the sum of two log probabilities.
 *           log(exp(p1)+exp(p2)) = p1 + log(1 + exp(p2-p1)) for p1 > p2
 *           Note that this is in natural log space, not log_2.
 */
float
LogSum (float p1, float p2)
{
    if (p1 > p2) {
        return (p1 - p2 > 50.) ? p1 : p1 + log (1. + exp (p2 - p1));
    }
    else {
        return (p2 - p1 > 50.) ? p2 : p2 + log (1. + exp (p1 - p2));
    }
}


/* Function: ILogsum()
 *
 * Purpose:  Return the scaled integer log probability of
 *           the sum of two probabilities p1 and p2, where
 *           p1 and p2 are also given as scaled log probabilities.
 *
 *           log(exp(p1)+exp(p2)) = p1 + log(1 + exp(p2-p1)) for p1 > p2
 *
 *           For speed, builds a lookup table the first time it's called.
 *           LOGSUM_TBL is set to 20000 by default, in config.h.
 *
 *           Because of the one-time initialization, we have to
 *           be careful in a multithreaded implementation... hence
 *           the use of pthread_once(), which forces us to put
 *           the initialization routine and the lookup table outside
 *           ILogsum(). (Thanks to Henry Gabb at Intel for pointing
 *           out this problem.)
 *
 * Args:     p1,p2 -- scaled integer log_2 probabilities to be summed
 *                    in probability space.
 *
 * Return:   scaled integer log_2 probability of the sum.
 */
static int ilogsum_lookup[LOGSUM_TBL];
static void
init_ilogsum (void)
{
    int i;
    for (i = 0; i < LOGSUM_TBL; i++) {
        ilogsum_lookup[i] = (int) (INTSCALE * 1.44269504 *
                                   (log
                                    (1. +
                                     exp (0.69314718 * (float) -i / INTSCALE))));
		}	 
}

int
ILogsum (int p1, int p2)
{
    int diff;
#ifdef HMMER_THREADS
    static pthread_once_t firsttime = PTHREAD_ONCE_INIT;
    pthread_once (&firsttime, init_ilogsum);
#else
    static int firsttime = 1;
    if (firsttime) {
        init_ilogsum ();
        firsttime = 0;
    }
#endif

    diff = p1 - p2;
    if (diff >= LOGSUM_TBL) {
        return p1;
    }
    else if (diff <= -LOGSUM_TBL) {
        return p2;
    }
    else if (diff > 0) {
        return p1 + ilogsum_lookup[diff];
    }
    else {
        return p2 + ilogsum_lookup[-diff];
    }
}

/* Function: LogNorm()
 *
 * Purpose:  Normalize a vector of log likelihoods, changing it
 *           to a probability vector. Be careful of overflowing exp().
 *           Implementation adapted from Graeme Mitchison.
 *
 * Args:     vec - vector destined to become log probabilities
 *           n   - length of vec
 */
void
LogNorm (float *vec, int n)
{
    int x;
    float max = -1.0e30;
    float denom = 0.;

    for (x = 0; x < n; x++) {
        if (vec[x] > max) {
            max = vec[x];
        }
		}		
    for (x = 0; x < n; x++) {
        if (vec[x] > max - 50.) {
            denom += exp (vec[x] - max);
        }
		}		
    for (x = 0; x < n; x++) {
        if (vec[x] > max - 50.) {
            vec[x] = exp (vec[x] - max) / denom;
        }
        else {
            vec[x] = 0.0;
        }
		}		
}


/* Function: Logp_cvec()
 *
 * Purpose:  Calculates ln P(cvec|dirichlet), the log probability of a
 *           count vector given a Dirichlet distribution. Adapted
 *           from an implementation by Graeme Mitchison.
 *
 * Args:     cvec  - count vector
 *           n     - length of cvec
 *           alpha - Dirichlet alpha terms
 *
 * Return:   log P(cvec|dirichlet)
 */
float
Logp_cvec (float *cvec, int n, float *alpha)
{
    float lnp;			/* log likelihood of P(cvec | Dirichlet) */
    float sum1, sum2, sum3;
    int x;

    sum1 = sum2 = sum3 = lnp = 0.0;
    for (x = 0; x < n; x++) {
        sum1 += cvec[x] + alpha[x];
        sum2 += alpha[x];
        sum3 += cvec[x];
        lnp += Gammln (alpha[x] + cvec[x]);
        lnp -= Gammln (cvec[x] + 1.);
        lnp -= Gammln (alpha[x]);
    }
    lnp -= Gammln (sum1);
    lnp += Gammln (sum2);
    lnp += Gammln (sum3 + 1.);
    return lnp;
}

/* Function: SampleDirichlet()
 *
 * Purpose:  Given a Dirichlet distribution defined by
 *           a vector of n alpha terms, sample of probability
 *           distribution of dimension n.
 *
 *           This code was derived from source provided
 *           by Betty Lazareva, from Gary Churchill's group.
 *
 * Args:     alpha - vector of Dirichlet alphas components
 *           n     - number of components
 *           ret_p - RETURN: sampled probability vector.
 *
 * Return:   (void)
 *           ret_p, an n-dimensional array alloced by the caller,
 *           is filled.
 */
void
SampleDirichlet (float *alpha, int n, float *p)
{
    int x;

    for (x = 0; x < n; x++) {
        p[x] = SampleGamma (alpha[x]);
    }
    FNorm (p, n);
}


/* Function: SampleGamma()
 *
 * Purpose:  Return a random deviate distributed as Gamma(alpha, 1.0).
 *           Uses two different accept/reject algorithms, one
 *           for 0<alpha<1, the other for 1<=alpha.
 *
 *           Code modified from source provided by Betty Lazareva
 *           and Gary Churchill.
 *
 * Args:     alpha - order of gamma function
 *
 * Return:   the gamma-distributed deviate.
 */
float
SampleGamma (float alpha)
{
    float U, V, X, W, lambda;

    if (alpha >= 1.0) {
        /*CONSTCOND*/ while (1) {
            lambda = sqrt (2.0 * alpha - 1.0);
            U = sre_random ();
            V = U / (1 - U);
            X = alpha * pow (V, 1 / lambda);
            W =
                .25 * exp (-X + alpha) * pow (V,
                                              1.0 + alpha / lambda) * pow (1.0 +
                                                      1.0 / V,
                                                      2.0);
            if (sre_random () <= W) {
                return X;
            }
        }
    }
    else if (alpha > 0.0) {
        /*CONSTCOND*/ while (1) {
            U = sre_random ();
            V = U * (1 + alpha / exp (1.0));
            if (V > 1.0) {
                X = -log ((1 - V + alpha / exp (1.0)) / alpha);
                if (sre_random () <= pow (X, alpha - 1.0)) {
                    return X;
                }
            }
            else {
                X = pow (V, 1.0 / alpha);
                if (sre_random () <= exp (-X)) {
                    return X;
                }
            }
        }
    }
    Die ("Invalid argument alpha < 0.0 to SampleGamma()");
    /*NOTREACHED*/
    return 0.0;
}

/* Function: SampleCountvector()
 *
 * Purpose:  Given a probability vector p of dimensionality
 *           n, sample c counts and store them in cvec.
 *           cvec is n-dimensional and is alloced by the caller.
 */
void
SampleCountvector (float *p, int n, int c, float *cvec)
{
    int i;

    FSet (cvec, n, 0.0);
    for (i = 0; i < c; i++) {
        cvec[FChoose (p, n)] += 1.0;
    }
}



/* Function: P_PvecGivenDirichlet()
 *
 * Purpose:  Calculate the log probability of a probability
 *           vector given a single Dirichlet component, alpha.
 *           Follows Sjolander (1996) appendix, lemma 2.
 *
 * Return:   log P(p | alpha)
 */
float
P_PvecGivenDirichlet (float *p, int n, float *alpha)
{
    float sum;			/* for Gammln(|alpha|) in Z     */
    float logp;			/* RETURN: log P(p|alpha)       */
    int x;

    sum = logp = 0.0;
    for (x = 0; x < n; x++) {
        if (p[x] > 0.0) {		/* any param that is == 0.0 doesn't exist */
            logp += (alpha[x] - 1.0) * log (p[x]);
            logp -= Gammln (alpha[x]);
            sum += alpha[x];
        }
		}		
    logp += Gammln (sum);
    return logp;
}


/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* misc.c
 * SRE, Thu Jul 15 18:49:19 1993
 *
 * Functions that I don't know quite where to put yet.
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <float.h>
#include <limits.h>

#include "squid.h"
#include "structs.h"


/* Function: HMMERBanner()
 * Date:     SRE, Sun Jun 20 17:19:41 1999 [Graeme's kitchen]
 *
 * Purpose:  Print a package version and copyright banner.
 *           Used by all the main()'s.
 *
 *    Expects to be able to pick up defined preprocessor variables:
 *    variable          example
 *    --------           --------------
 *    PACKAGE_NAME      "HMMER"
 *    PACKAGE_VERSION   "2.0.42"
 *    PACKAGE_DATE      "April 1999"
 *    PACKAGE_COPYRIGHT "Copyright (C) 1992-1999 Washington University School of Medicine"
 *    PACKAGE_LICENSE   "Freely distributed under the GNU General Public License (GPL)."
 *
 *    This gives us a general mechanism to update release information
 *    without changing multiple points in the code.
 *
 * Args:     fp     - where to print it
 *           banner - one-line program description, e.g.:
 *                    "foobar - make bars from foo with elan"
 * Returns:  (void)
 */
void
HMMERBanner (FILE * fp, char *banner)
{
    fprintf (fp, "%s\n", banner);
    fprintf (fp, "%s %s (%s)\n", PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_DATE);
    fprintf (fp, "%s\n", PACKAGE_COPYRIGHT);
    fprintf (fp, "%s\n", PACKAGE_LICENSE);
    fprintf (fp,
             "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
}



/* Function: Getword()
 *
 * Purpose:  little function used by ReadPrior() and ReadHMM() to parse
 *           next valid field out of an open file, ignoring
 *           comments. '#' marks the beginning of a comment.
 *
 * Arg:      fp   - open file for reading
 *           type - sqdARG_INT, sqdARG_FLOAT, or sqdARG_STRING from squid.h
 */
char *
Getword (FILE * fp, int type)
{
    static char buffer[512];
    static char *sptr = NULL;

    if (sptr != NULL) {
        sptr = strtok (NULL, " \t\n");
    }

    while (sptr == NULL) {
        if ((sptr = fgets (buffer, 512, fp)) == NULL) {
            return NULL;
        }
        if ((sptr = strchr (buffer, '#')) != NULL) {
            *sptr = '\0';
        }
        sptr = strtok (buffer, " \t\n");
    }

    switch (type) {
    case sqdARG_STRING:
        if (strlen (sptr) == 0) {
            Warn ("Parse failed: expected string, got nothing");
            sptr = NULL;
        }
        break;
    case sqdARG_INT:
        if (!IsInt (sptr)) {
            Warn ("Parse failed: expected integer, got %s", sptr);
            sptr = NULL;
        }
        break;
    case sqdARG_FLOAT:
        if (!IsReal (sptr)) {
            Warn ("Parse failed: expected real value, got %s", sptr);
            sptr = NULL;
        }
        break;
    }

    return sptr;
}


/* Function: Getline()
 *
 * Purpose:  Get the next non-blank, non-comment line from an open file.
 *           A comment line has '#' as the first non-whitespace character.
 *           Returns NULL if no line is found.
 *           Syntax is the same as fgets().
 *
 * Args:     s  - allocated storage for line
 *           n  - number of characters allocated for s
 *           fp - open FILE *
 *
 * Return:   Either s, or NULL if no new line is found.
 */
char *
Getline (char *s, int n, FILE * fp)
{
    char *first;

    do {
        if (fgets (s, n, fp) == NULL) {
            return NULL;
        }
        first = s;
        while (isspace ((int) (*first))) {
            first++;
        }
    } while (*first == '#' || *first == '\0');
    return s;
}


/* Function: SetAutocuts()
 * Date:     SRE, Thu Jun  8 08:19:46 2000 [TW721 over Ireland]
 *
 * Purpose:  Set score thresholds using the GA, TC, or NC information
 *           in an HMM.
 *
 * Args:     thresh - score threshold structure. autocut must be set
 *                    properly (CUT_GA, CUT_NC, or CUT_TC).
 *           hmm    - HMM containing appropriate score cutoff info
 *
 * Returns:  1 on success.
 *           0 if HMM does not have the score cutoffs available -- caller
 *             will have to decide on a fallback plan.
 *           Has no effect (and returns success) if autocut is
 *           CUT_NONE.
 */
int
SetAutocuts (struct threshold_s *thresh, struct plan7_s *hmm)
{
    if (thresh->autocut == CUT_GA) {
        if (!(hmm->flags & PLAN7_GA)) {
            return 0;
        }
        thresh->globT = hmm->ga1;
        thresh->domT = hmm->ga2;
        thresh->globE = thresh->domE = FLT_MAX;
    }
    else if (thresh->autocut == CUT_NC) {
        if (!(hmm->flags & PLAN7_NC)) {
            return 0;
        }
        thresh->globT = hmm->nc1;
        thresh->domT = hmm->nc2;
        thresh->globE = thresh->domE = FLT_MAX;
    }
    else if (thresh->autocut == CUT_TC) {
        if (!(hmm->flags & PLAN7_TC)) {
            return 0;
        }
        thresh->globT = hmm->tc1;
        thresh->domT = hmm->tc2;
        thresh->globE = thresh->domE = FLT_MAX;
    }
    return 1;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* msa.c
 * SRE, Mon May 17 10:48:47 1999
 *
 * SQUID's interface for multiple sequence alignment
 * manipulation: access to the MSA object.
 *
 * CVS $Id: msa.c,v 1.20 2003/05/26 16:21:50 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "squid.h"
#include "msa.h"		/* multiple sequence alignment object support */
#include "gki.h"		/* string indexing hashtable code  */
#include "ssi.h"		/* SSI sequence file indexing code */

/* Function: MSAAlloc()
 * Date:     SRE, Tue May 18 10:45:47 1999 [St. Louis]
 *
 * Purpose:  Allocate an MSA structure, return a pointer
 *           to it.
 *
 *           Designed to be used in three ways:
 *           1) We know exactly the dimensions of the alignment:
 *              both nseq and alen.
 *                    msa = MSAAlloc(nseq, alen);
 *
 *           2) We know the number of sequences but not alen.
 *              (We add sequences later.)
 *                    msa = MSAAlloc(nseq, 0);
 *
 *           3) We even don't know the number of sequences, so
 *              we'll have to dynamically expand allocations.
 *              We provide a blocksize for the allocation expansion,
 *              and expand when needed.
 *                    msa = MSAAlloc(10, 0);
 *                    if (msa->nseq == msa->nseqalloc) MSAExpand(msa);
 *
 * Args:     nseq - number of sequences, or nseq allocation blocksize
 *           alen - length of alignment in columns, or 0
 *
 * Returns:  pointer to new MSA object, w/ all values initialized.
 *           Note that msa->nseq is initialized to 0, though space
 *           is allocated.
 *
 * Diagnostics: "always works". Die()'s on memory allocation failure.
 *
 */
MSA *
MSAAlloc (int nseq, int alen)
{
    MSA *msa;
    int i;

    msa = MallocOrDie (sizeof (MSA));
    msa->aseq = MallocOrDie (sizeof (char *) * nseq);
    msa->sqname = MallocOrDie (sizeof (char *) * nseq);
    msa->sqlen = MallocOrDie (sizeof (int) * nseq);
    msa->wgt = MallocOrDie (sizeof (float) * nseq);

    for (i = 0; i < nseq; i++) {
        msa->sqname[i] = NULL;
        msa->sqlen[i] = 0;
        msa->wgt[i] = -1.0;

        if (alen != 0) {
            msa->aseq[i] = MallocOrDie (sizeof (char) * (alen + 1));
        }
        else {
            msa->aseq[i] = NULL;
        }
    }

    msa->alen = alen;
    msa->nseq = 0;
    msa->nseqalloc = nseq;
    msa->nseqlump = nseq;

    msa->flags = 0;
    msa->type = kOtherSeq;
    msa->name = NULL;
    msa->desc = NULL;
    msa->acc = NULL;
    msa->au = NULL;
    msa->ss_cons = NULL;
    msa->sa_cons = NULL;
    msa->rf = NULL;
    msa->sqacc = NULL;
    msa->sqdesc = NULL;
    msa->ss = NULL;
    msa->sslen = NULL;
    msa->sa = NULL;
    msa->salen = NULL;
    msa->index = GKIInit ();
    msa->lastidx = 0;

    for (i = 0; i < MSA_MAXCUTOFFS; i++) {
        msa->cutoff[i] = 0.;
        msa->cutoff_is_set[i] = FALSE;
    }

    /* Initialize unparsed optional markup
     */
    msa->comment = NULL;
    msa->ncomment = 0;
    msa->alloc_ncomment = 0;

    msa->gf_tag = NULL;
    msa->gf = NULL;
    msa->ngf = 0;

    msa->gs_tag = NULL;
    msa->gs = NULL;
    msa->gs_idx = NULL;
    msa->ngs = 0;

    msa->gc_tag = NULL;
    msa->gc = NULL;
    msa->gc_idx = NULL;
    msa->ngc = 0;

    msa->gr_tag = NULL;
    msa->gr = NULL;
    msa->gr_idx = NULL;
    msa->ngr = 0;

    /* Done. Return the alloced, initialized structure
     */
    return msa;
}

/* Function: MSAExpand()
 * Date:     SRE, Tue May 18 11:06:53 1999 [St. Louis]
 *
 * Purpose:  Increase the sequence allocation in an MSA
 *           by msa->nseqlump. (Typically used when we're reading
 *           in an alignment sequentially from a file,
 *           so we don't know nseq until we're done.)
 *
 * Args:     msa - the MSA object
 *
 * Returns:  (void)
 *
 */
void
MSAExpand (MSA * msa)
{
    int i, j;

    msa->nseqalloc += msa->nseqlump;

    msa->aseq = ReallocOrDie (msa->aseq, sizeof (char *) * msa->nseqalloc);
    msa->sqname = ReallocOrDie (msa->sqname, sizeof (char *) * msa->nseqalloc);
    msa->sqlen = ReallocOrDie (msa->sqlen, sizeof (char *) * msa->nseqalloc);
    msa->wgt = ReallocOrDie (msa->wgt, sizeof (float) * msa->nseqalloc);

    if (msa->ss != NULL) {
        msa->ss = ReallocOrDie (msa->ss, sizeof (char *) * msa->nseqalloc);
        msa->sslen = ReallocOrDie (msa->sslen, sizeof (int) * msa->nseqalloc);
    }
    if (msa->sa != NULL) {
        msa->sa = ReallocOrDie (msa->sa, sizeof (char *) * msa->nseqalloc);
        msa->salen = ReallocOrDie (msa->salen, sizeof (int) * msa->nseqalloc);
    }
    if (msa->sqacc != NULL) {
        msa->sqacc = ReallocOrDie (msa->sqacc, sizeof (char *) * msa->nseqalloc);
    }
    if (msa->sqdesc != NULL)
        msa->sqdesc =
            ReallocOrDie (msa->sqdesc, sizeof (char *) * msa->nseqalloc);

    for (i = msa->nseqalloc - msa->nseqlump; i < msa->nseqalloc; i++) {
        msa->sqname[i] = NULL;
        msa->wgt[i] = -1.0;

        if (msa->sqacc != NULL) {
            msa->sqacc[i] = NULL;
        }
        if (msa->sqdesc != NULL) {
            msa->sqdesc[i] = NULL;
        }

        if (msa->alen != 0)
            msa->aseq[i] =
                ReallocOrDie (msa->aseq[i], sizeof (char) * (msa->alen + 1));
        else {
            msa->aseq[i] = NULL;
        }
        msa->sqlen[i] = 0;

        if (msa->ss != NULL) {
            if (msa->alen != 0)
                msa->ss[i] =
                    ReallocOrDie (msa->ss[i], sizeof (char) * (msa->alen + 1));
            else {
                msa->ss[i] = NULL;
            }
            msa->sslen[i] = 0;
        }
        if (msa->sa != NULL) {
            if (msa->alen != 0)
                msa->sa[i] =
                    ReallocOrDie (msa->ss[i], sizeof (char) * (msa->alen + 1));
            else {
                msa->sa[i] = NULL;
            }
            msa->salen[i] = 0;
        }
    }

    /* Reallocate and re-init for unparsed #=GS tags, if we have some.
     * gs is [0..ngs-1][0..nseq-1][], so we're reallocing the middle
     * set of pointers.
     */
    if (msa->gs != NULL)
        for (i = 0; i < msa->ngs; i++) {
            if (msa->gs[i] != NULL) {
                msa->gs[i] =
                    ReallocOrDie (msa->gs[i], sizeof (char *) * msa->nseqalloc);
                for (j = msa->nseqalloc - msa->nseqlump; j < msa->nseqalloc; j++) {
                    msa->gs[i][j] = NULL;
                }
            }
        }

    /* Reallocate and re-init for unparsed #=GR tags, if we have some.
     * gr is [0..ngs-1][0..nseq-1][], so we're reallocing the middle
     * set of pointers.
     */
    if (msa->gr != NULL)
        for (i = 0; i < msa->ngr; i++) {
            if (msa->gr[i] != NULL) {
                msa->gr[i] =
                    ReallocOrDie (msa->gr[i], sizeof (char *) * msa->nseqalloc);
                for (j = msa->nseqalloc - msa->nseqlump; j < msa->nseqalloc; j++) {
                    msa->gr[i][j] = NULL;
                }
            }
        }

    return;
}

/* Function: MSAFree()
 * Date:     SRE, Tue May 18 11:20:16 1999 [St. Louis]
 *
 * Purpose:  Free a multiple sequence alignment structure.
 *
 * Args:     msa - the alignment
 *
 * Returns:  (void)
 */
void
MSAFree (MSA * msa)
{
    Free2DArray ((void **) msa->aseq, msa->nseq);
    Free2DArray ((void **) msa->sqname, msa->nseq);
    Free2DArray ((void **) msa->sqacc, msa->nseq);
    Free2DArray ((void **) msa->sqdesc, msa->nseq);
    Free2DArray ((void **) msa->ss, msa->nseq);
    Free2DArray ((void **) msa->sa, msa->nseq);

    if (msa->sqlen != NULL) {
        free (msa->sqlen);
    }
    if (msa->wgt != NULL) {
        free (msa->wgt);
    }

    if (msa->name != NULL) {
        free (msa->name);
    }
    if (msa->desc != NULL) {
        free (msa->desc);
    }
    if (msa->acc != NULL) {
        free (msa->acc);
    }
    if (msa->au != NULL) {
        free (msa->au);
    }
    if (msa->ss_cons != NULL) {
        free (msa->ss_cons);
    }
    if (msa->sa_cons != NULL) {
        free (msa->sa_cons);
    }
    if (msa->rf != NULL) {
        free (msa->rf);
    }
    if (msa->sslen != NULL) {
        free (msa->sslen);
    }
    if (msa->salen != NULL) {
        free (msa->salen);
    }

    Free2DArray ((void **) msa->comment, msa->ncomment);
    Free2DArray ((void **) msa->gf_tag, msa->ngf);
    Free2DArray ((void **) msa->gf, msa->ngf);
    Free2DArray ((void **) msa->gs_tag, msa->ngs);
    Free3DArray ((void ***) msa->gs, msa->ngs, msa->nseq);
    Free2DArray ((void **) msa->gc_tag, msa->ngc);
    Free2DArray ((void **) msa->gc, msa->ngc);
    Free2DArray ((void **) msa->gr_tag, msa->ngr);
    Free3DArray ((void ***) msa->gr, msa->ngr, msa->nseq);

    GKIFree (msa->index);
    GKIFree (msa->gs_idx);
    GKIFree (msa->gc_idx);
    GKIFree (msa->gr_idx);

    free (msa);
}


/* Function: MSASetSeqAccession()
 * Date:     SRE, Mon Jun 21 04:13:33 1999 [Sanger Centre]
 *
 * Purpose:  Set a sequence accession in an MSA structure.
 *           Handles some necessary allocation/initialization.
 *
 * Args:     msa      - multiple alignment to add accession to
 *           seqidx   - index of sequence to attach accession to
 *           acc      - accession
 *
 * Returns:  void
 */
void
MSASetSeqAccession (MSA * msa, int seqidx, char *acc)
{
    int x;

    if (msa->sqacc == NULL) {
        msa->sqacc = MallocOrDie (sizeof (char *) * msa->nseqalloc);
        for (x = 0; x < msa->nseqalloc; x++) {
            msa->sqacc[x] = NULL;
        }
    }
    msa->sqacc[seqidx] = sre_strdup (acc, -1);
}

/* Function: MSASetSeqDescription()
 * Date:     SRE, Mon Jun 21 04:21:09 1999 [Sanger Centre]
 *
 * Purpose:  Set a sequence description in an MSA structure.
 *           Handles some necessary allocation/initialization.
 *
 * Args:     msa      - multiple alignment to add accession to
 *           seqidx   - index of sequence to attach accession to
 *           desc     - description
 *
 * Returns:  void
 */
void
MSASetSeqDescription (MSA * msa, int seqidx, char *desc)
{
    int x;

    if (msa->sqdesc == NULL) {
        msa->sqdesc = MallocOrDie (sizeof (char *) * msa->nseqalloc);
        for (x = 0; x < msa->nseqalloc; x++) {
            msa->sqdesc[x] = NULL;
        }
    }
    msa->sqdesc[seqidx] = sre_strdup (desc, -1);
}


/* Function: MSAAddComment()
 * Date:     SRE, Tue Jun  1 17:37:21 1999 [St. Louis]
 *
 * Purpose:  Add an (unparsed) comment line to the MSA structure,
 *           allocating as necessary.
 *
 * Args:     msa - a multiple alignment
 *           s   - comment line to add
 *
 * Returns:  (void)
 */
void
MSAAddComment (MSA * msa, char *s)
{
    /* If this is our first recorded comment, we need to malloc();
     * and if we've filled available space, we need to realloc().
     * Note the arbitrary lumpsize of 10 lines per allocation...
     */
    if (msa->comment == NULL) {
        msa->comment = MallocOrDie (sizeof (char *) * 10);
        msa->alloc_ncomment = 10;
    }
    if (msa->ncomment == msa->alloc_ncomment) {
        msa->alloc_ncomment += 10;
        msa->comment =
            ReallocOrDie (msa->comment, sizeof (char *) * msa->alloc_ncomment);
    }

    msa->comment[msa->ncomment] = sre_strdup (s, -1);
    msa->ncomment++;
    return;
}

/* Function: MSAAddGF()
 * Date:     SRE, Wed Jun  2 06:53:54 1999 [bus to Madison]
 *
 * Purpose:  Add an unparsed #=GF markup line to the MSA
 *           structure, allocating as necessary.
 *
 * Args:     msa   - a multiple alignment
 *           tag   - markup tag (e.g. "AU")
 *           value - free text markup (e.g. "Alex Bateman")
 *
 * Returns:  (void)
 */
void
MSAAddGF (MSA * msa, char *tag, char *value)
{
    /* If this is our first recorded unparsed #=GF line, we need to malloc();
     * if we've filled availabl space If we already have a hash index, and the GF
     * Note the arbitrary lumpsize of 10 lines per allocation...
     */
    if (msa->gf_tag == NULL) {
        msa->gf_tag = MallocOrDie (sizeof (char *) * 10);
        msa->gf = MallocOrDie (sizeof (char *) * 10);
        msa->alloc_ngf = 10;
    }
    if (msa->ngf == msa->alloc_ngf) {
        msa->alloc_ngf += 10;
        msa->gf_tag =
            ReallocOrDie (msa->gf_tag, sizeof (char *) * msa->alloc_ngf);
        msa->gf = ReallocOrDie (msa->gf, sizeof (char *) * msa->alloc_ngf);
    }

    msa->gf_tag[msa->ngf] = sre_strdup (tag, -1);
    msa->gf[msa->ngf] = sre_strdup (value, -1);
    msa->ngf++;

    return;
}


/* Function: MSAAddGS()
 * Date:     SRE, Wed Jun  2 06:57:03 1999 [St. Louis]
 *
 * Purpose:  Add an unparsed #=GS markup line to the MSA
 *           structure, allocating as necessary.
 *
 *           It's possible that we could get more than one
 *           of the same type of GS tag per sequence; for
 *           example, "DR PDB;" structure links in Pfam.
 *           Hack: handle these by appending to the string,
 *           in a \n separated fashion.
 *
 * Args:     msa    - multiple alignment structure
 *           tag    - markup tag (e.g. "AC")
 *           sqidx  - index of sequence to assoc markup with (0..nseq-1)
 *           value  - markup (e.g. "P00666")
 *
 * Returns:  0 on success
 */
void
MSAAddGS (MSA * msa, char *tag, int sqidx, char *value)
{
    int tagidx;
    int i;

    /* Is this an unparsed tag name that we recognize?
     * If not, handle adding it to index, and reallocating
     * as needed.
     */
    if (msa->gs_tag == NULL) {	/* first tag? init w/ malloc  */
        msa->gs_idx = GKIInit ();
        tagidx = GKIStoreKey (msa->gs_idx, tag);
        SQD_DASSERT1 ((tagidx == 0));
        msa->gs_tag = MallocOrDie (sizeof (char *));
        msa->gs = MallocOrDie (sizeof (char **));
        msa->gs[0] = MallocOrDie (sizeof (char *) * msa->nseqalloc);
        for (i = 0; i < msa->nseqalloc; i++) {
            msa->gs[0][i] = NULL;
        }
    }
    else {
        /* new tag? */
        tagidx = GKIKeyIndex (msa->gs_idx, tag);
        if (tagidx < 0) {		/* it's a new tag name; realloc */
            tagidx = GKIStoreKey (msa->gs_idx, tag);
            /* since we alloc in blocks of 1,
               we always realloc upon seeing
               a new tag. */
            SQD_DASSERT1 ((tagidx == msa->ngs));
            msa->gs_tag =
                ReallocOrDie (msa->gs_tag, (msa->ngs + 1) * sizeof (char *));
            msa->gs = ReallocOrDie (msa->gs, (msa->ngs + 1) * sizeof (char **));
            msa->gs[msa->ngs] = MallocOrDie (sizeof (char *) * msa->nseqalloc);
            for (i = 0; i < msa->nseqalloc; i++) {
                msa->gs[msa->ngs][i] = NULL;
            }
        }
    }

    if (tagidx == msa->ngs) {
        msa->gs_tag[tagidx] = sre_strdup (tag, -1);
        msa->ngs++;
    }

    if (msa->gs[tagidx][sqidx] == NULL) {	/* first annotation of this seq with this tag? */
        msa->gs[tagidx][sqidx] = sre_strdup (value, -1);
    }
    else {
        /* >1 annotation of this seq with this tag; append */
        int len;
        if ((len = sre_strcat (&(msa->gs[tagidx][sqidx]), -1, "\n", 1)) < 0) {
            Die ("failed to sre_strcat()");
        }
        if (sre_strcat (&(msa->gs[tagidx][sqidx]), len, value, -1) < 0) {
            Die ("failed to sre_strcat()");
        }
    }
    return;
}

/* Function: MSAAppendGC()
 * Date:     SRE, Thu Jun  3 06:25:14 1999 [Madison]
 *
 * Purpose:  Add an unparsed #=GC markup line to the MSA
 *           structure, allocating as necessary.
 *
 *           When called multiple times for the same tag,
 *           appends value strings together -- used when
 *           parsing multiblock alignment files, for
 *           example.
 *
 * Args:     msa   - multiple alignment structure
 *           tag   - markup tag (e.g. "CS")
 *           value - markup, one char per aligned column
 *
 * Returns:  (void)
 */
void
MSAAppendGC (MSA * msa, char *tag, char *value)
{
    int tagidx;

    /* Is this an unparsed tag name that we recognize?
     * If not, handle adding it to index, and reallocating
     * as needed.
     */
    if (msa->gc_tag == NULL) {	/* first tag? init w/ malloc  */
        msa->gc_tag = MallocOrDie (sizeof (char *));
        msa->gc = MallocOrDie (sizeof (char *));
        msa->gc_idx = GKIInit ();
        tagidx = GKIStoreKey (msa->gc_idx, tag);
        SQD_DASSERT1 ((tagidx == 0));
        msa->gc[0] = NULL;
    }
    else {			/* new tag? */
        tagidx = GKIKeyIndex (msa->gc_idx, tag);
        if (tagidx < 0) {		/* it's a new tag name; realloc */
            tagidx = GKIStoreKey (msa->gc_idx, tag);
            /* since we alloc in blocks of 1,
               we always realloc upon seeing
               a new tag. */
            SQD_DASSERT1 ((tagidx == msa->ngc));
            msa->gc_tag =
                ReallocOrDie (msa->gc_tag, (msa->ngc + 1) * sizeof (char **));
            msa->gc = ReallocOrDie (msa->gc, (msa->ngc + 1) * sizeof (char **));
            msa->gc[tagidx] = NULL;
        }
    }

    if (tagidx == msa->ngc) {
        msa->gc_tag[tagidx] = sre_strdup (tag, -1);
        msa->ngc++;
    }
    sre_strcat (&(msa->gc[tagidx]), -1, value, -1);
    return;
}

/* Function: MSAGetGC()
 * Date:     SRE, Fri Aug 13 13:25:57 1999 [St. Louis]
 *
 * Purpose:  Given a tagname for a miscellaneous #=GC column
 *           annotation, return a pointer to the annotation
 *           string.
 *
 * Args:     msa  - alignment and its annotation
 *           tag  - name of the annotation
 *
 * Returns:  ptr to the annotation string. Caller does *not*
 *           free; is managed by msa object still.
 */
char *
MSAGetGC (MSA * msa, char *tag)
{
    int tagidx;

    if (msa->gc_idx == NULL) {
        return NULL;
    }
    if ((tagidx = GKIKeyIndex (msa->gc_idx, tag)) < 0) {
        return NULL;
    }
    return msa->gc[tagidx];
}


/* Function: MSAAppendGR()
 * Date:     SRE, Thu Jun  3 06:34:38 1999 [Madison]
 *
 * Purpose:  Add an unparsed #=GR markup line to the
 *           MSA structure, allocating as necessary.
 *
 *           When called multiple times for the same tag,
 *           appends value strings together -- used when
 *           parsing multiblock alignment files, for
 *           example.
 *
 * Args:     msa    - multiple alignment structure
 *           tag    - markup tag (e.g. "SS")
 *           sqidx  - index of seq to assoc markup with (0..nseq-1)
 *           value  - markup, one char per aligned column
 *
 * Returns:  (void)
 */
void
MSAAppendGR (MSA * msa, char *tag, int sqidx, char *value)
{
    int tagidx;
    int i;

    /* Is this an unparsed tag name that we recognize?
     * If not, handle adding it to index, and reallocating
     * as needed.
     */
    if (msa->gr_tag == NULL) {	/* first tag? init w/ malloc  */
        msa->gr_tag = MallocOrDie (sizeof (char *));
        msa->gr = MallocOrDie (sizeof (char **));
        msa->gr[0] = MallocOrDie (sizeof (char *) * msa->nseqalloc);
        for (i = 0; i < msa->nseqalloc; i++) {
            msa->gr[0][i] = NULL;
        }
        msa->gr_idx = GKIInit ();
        tagidx = GKIStoreKey (msa->gr_idx, tag);
        SQD_DASSERT1 ((tagidx == 0));
    }
    else {
        /* new tag? */
        tagidx = GKIKeyIndex (msa->gr_idx, tag);
        if (tagidx < 0) {		/* it's a new tag name; realloc */
            tagidx = GKIStoreKey (msa->gr_idx, tag);
            /* since we alloc in blocks of 1,
               we always realloc upon seeing
               a new tag. */
            SQD_DASSERT1 ((tagidx == msa->ngr));
            msa->gr_tag =
                ReallocOrDie (msa->gr_tag, (msa->ngr + 1) * sizeof (char *));
            msa->gr = ReallocOrDie (msa->gr, (msa->ngr + 1) * sizeof (char **));
            msa->gr[msa->ngr] = MallocOrDie (sizeof (char *) * msa->nseqalloc);
            for (i = 0; i < msa->nseqalloc; i++) {
                msa->gr[msa->ngr][i] = NULL;
            }
        }
    }

    if (tagidx == msa->ngr) {
        msa->gr_tag[tagidx] = sre_strdup (tag, -1);
        msa->ngr++;
    }
    sre_strcat (&(msa->gr[tagidx][sqidx]), -1, value, -1);
    return;
}


/* Function: MSAVerifyParse()
 * Date:     SRE, Sat Jun  5 14:24:24 1999 [Madison, 1999 worm mtg]
 *
 * Purpose:  Last function called after a multiple alignment is
 *           parsed. Checks that parse was successful; makes sure
 *           required information is present; makes sure required
 *           information is consistent. Some fields that are
 *           only use during parsing may be freed (sqlen, for
 *           example).
 *
 *           Some fields in msa may be modified (msa->alen is set,
 *           for example).
 *
 * Args:     msa - the multiple alignment
 *                 sqname, aseq must be set
 *                 nseq must be correct
 *                 alen need not be set; will be set here.
 *                 wgt will be set here if not already set
 *
 * Returns:  (void)
 *           Will Die() here with diagnostics on error.
 *
 * Example:
 */
void
MSAVerifyParse (MSA * msa)
{
    int idx;

    if (msa->nseq == 0)
        Die ("Parse error: no sequences were found for alignment %s",
             msa->name != NULL ? msa->name : "");

    msa->alen = msa->sqlen[0];

    /* We can rely on msa->sqname[] being valid for any index,
     * because of the way the line parsers always store any name
     * they add to the index.
     */
    for (idx = 0; idx < msa->nseq; idx++) {
        /* aseq is required. */
        if (msa->aseq[idx] == NULL)
            Die ("Parse error: No sequence for %s in alignment %s",
                 msa->sqname[idx], msa->name != NULL ? msa->name : "");
        /* either all weights must be set, or none of them */
        if ((msa->flags & MSA_SET_WGT) && msa->wgt[idx] == -1.0)
            Die
            ("Parse error: some weights are set, but %s doesn't have one in alignment %s",
             msa->sqname[idx], msa->name != NULL ? msa->name : "");
        /* all aseq must be same length. */
        if (msa->sqlen[idx] != msa->alen)
            Die ("Parse error: sequence %s: length %d, expected %d in alignment %s",
                 msa->sqname[idx], msa->sqlen[idx], msa->alen,
                 msa->name != NULL ? msa->name : "");
        /* if SS is present, must have length right */
        if (msa->ss != NULL && msa->ss[idx] != NULL
                && msa->sslen[idx] != msa->alen)
            Die
            ("Parse error: #=GR SS annotation for %s: length %d, expected %d in alignment %s",
             msa->sqname[idx], msa->sslen[idx], msa->alen,
             msa->name != NULL ? msa->name : "");
        /* if SA is present, must have length right */
        if (msa->sa != NULL && msa->sa[idx] != NULL
                && msa->salen[idx] != msa->alen)
            Die
            ("Parse error: #=GR SA annotation for %s: length %d, expected %d in alignment %s",
             msa->sqname[idx], msa->salen[idx], msa->alen,
             msa->name != NULL ? msa->name : "");
    }

    /* if cons SS is present, must have length right */
    if (msa->ss_cons != NULL && strlen (msa->ss_cons) != msa->alen)
        Die
        ("Parse error: #=GC SS_cons annotation: length %d, expected %d in alignment %s",
         strlen (msa->ss_cons), msa->alen, msa->name != NULL ? msa->name : "");

    /* if cons SA is present, must have length right */
    if (msa->sa_cons != NULL && strlen (msa->sa_cons) != msa->alen)
        Die
        ("Parse error: #=GC SA_cons annotation: length %d, expected %d in alignment %s",
         strlen (msa->sa_cons), msa->alen, msa->name != NULL ? msa->name : "");

    /* if RF is present, must have length right */
    if (msa->rf != NULL && strlen (msa->rf) != msa->alen)
        Die
        ("Parse error: #=GC RF annotation: length %d, expected %d in alignment %s",
         strlen (msa->rf), msa->alen, msa->name != NULL ? msa->name : "");

    /* Check that all or no weights are set */
    if (!(msa->flags & MSA_SET_WGT)) {
        FSet (msa->wgt, msa->nseq, 1.0);    /* default weights */
    }

    /* Clean up a little from the parser */
    if (msa->sqlen != NULL) {
        free (msa->sqlen);
        msa->sqlen = NULL;
    }
    if (msa->sslen != NULL) {
        free (msa->sslen);
        msa->sslen = NULL;
    }
    if (msa->salen != NULL) {
        free (msa->salen);
        msa->salen = NULL;
    }

    return;
}




/* Function: MSAFileOpen()
 * Date:     SRE, Tue May 18 13:22:01 1999 [St. Louis]
 *
 * Purpose:  Open an alignment database file and prepare
 *           for reading one alignment, or sequentially
 *           in the (rare) case of multiple MSA databases
 *           (e.g. Stockholm format).
 *
 * Args:     filename - name of file to open
 *                      if "-", read stdin
 *                      if it ends in ".gz", read from pipe to gunzip -dc
 *           format   - format of file (e.g. MSAFILE_STOCKHOLM)
 *           env      - environment variable for path (e.g. BLASTDB)
 *
 * Returns:  opened MSAFILE * on success.
 *           NULL on failure:
 *             usually, because the file doesn't exist;
 *             for gzip'ed files, may also mean that gzip isn't in the path.
 */
MSAFILE *
MSAFileOpen (char *filename, int format, char *env)
{
    MSAFILE *afp;

    afp = MallocOrDie (sizeof (MSAFILE));
    if (strcmp (filename, "-") == 0) {
        afp->f = stdin;
        afp->do_stdin = TRUE;
        afp->do_gzip = FALSE;
        afp->fname = sre_strdup ("[STDIN]", -1);
        afp->ssi = NULL;		/* can't index stdin because we can't seek */
    }
#ifndef SRE_STRICT_ANSI
    /* popen(), pclose() aren't portable to non-POSIX systems; disable */
    else if (Strparse ("^.*\\.gz$", filename, 0)) {
        char cmd[256];

        /* Note that popen() will return "successfully"
         * if file doesn't exist, because gzip works fine
         * and prints an error! So we have to check for
         * existence of file ourself.
         */
        if (!FileExists (filename)) {
            Die ("%s: file does not exist", filename);
        }
        if (strlen (filename) + strlen ("gzip -dc ") >= 256) {
            Die ("filename > 255 char in MSAFileOpen()");
        }
        sprintf (cmd, "gzip -dc %s", filename);
        if ((afp->f = popen (cmd, "r")) == NULL) {
            return NULL;
        }

        afp->do_stdin = FALSE;
        afp->do_gzip = TRUE;
        afp->fname = sre_strdup (filename, -1);
        /* we can't index a .gz file, because we can't seek in a pipe afaik */
        afp->ssi = NULL;
    }
#endif /*SRE_STRICT_ANSI */
    else {
        char *ssifile;
        char *dir;

        /* When we open a file, it may be either in the current
         * directory, or in the directory indicated by the env
         * argument - and we have to construct the SSI filename accordingly.
         */
        if ((afp->f = fopen (filename, "r")) != NULL) {
            ssifile = MallocOrDie (sizeof (char) * (strlen (filename) + 5));
            sprintf (ssifile, "%s.ssi", filename);
        }
        else if ((afp->f = EnvFileOpen (filename, env, &dir)) != NULL) {
            char *full;
            full = FileConcat (dir, filename);
            ssifile =
                MallocOrDie (sizeof (char) * (strlen (full) + strlen (filename) + 5));
            sprintf (ssifile, "%s.ssi", full);
            free (dir);
        }
        else {
            return NULL;
        }

        afp->do_stdin = FALSE;
        afp->do_gzip = FALSE;
        afp->fname = sre_strdup (filename, -1);
        afp->ssi = NULL;

        /* Open the SSI index file. If it doesn't exist, or
         * it's corrupt, or some error happens, afp->ssi stays NULL.
         */
        SSIOpen (ssifile, &(afp->ssi));
        free (ssifile);
    }

    /* Invoke autodetection if we haven't already been told what
     * to expect.
     */
    if (format == MSAFILE_UNKNOWN) {
        if (afp->do_stdin == TRUE || afp->do_gzip)
            Die
            ("Can't autodetect alignment file format from a stdin or gzip pipe");
        format = MSAFileFormat (afp);
        if (format == MSAFILE_UNKNOWN)
            Die ("Can't determine format of multiple alignment file %s",
                 afp->fname);
    }

    afp->format = format;
    afp->linenumber = 0;
    afp->buf = NULL;
    afp->buflen = 0;

    return afp;
}


/* Function: MSAFilePositionByKey()
 *           MSAFilePositionByIndex()
 *           MSAFileRewind()
 *
 * Date:     SRE, Tue Nov  9 19:02:54 1999 [St. Louis]
 *
 * Purpose:  Family of functions for repositioning in
 *           open MSA files; analogous to a similarly
 *           named function series in HMMER's hmmio.c.
 *
 * Args:     afp    - open alignment file
 *           offset - disk offset in bytes
 *           key    - key to look up in SSI indices
 *           idx    - index of alignment.
 *
 * Returns:  0 on failure.
 *           1 on success.
 *           If called on a non-fseek()'able file (e.g. a gzip'ed
 *           or pipe'd alignment), returns 0 as a failure flag.
 */
int
MSAFileRewind (MSAFILE * afp)
{
    if (afp->do_gzip || afp->do_stdin) {
        return 0;
    }
    rewind (afp->f);
    return 1;
}

int
MSAFilePositionByKey (MSAFILE * afp, char *key)
{
    int fh;			/* filehandle is ignored       */
    SSIOFFSET offset;		/* offset of the key alignment */

    if (afp->ssi == NULL) {
        return 0;
    }
    if (SSIGetOffsetByName (afp->ssi, key, &fh, &offset) != 0) {
        return 0;
    }
    if (SSISetFilePosition (afp->f, &offset) != 0) {
        return 0;
    }
    return 1;
}

int
MSAFilePositionByIndex (MSAFILE * afp, int idx)
{
    int fh;			/* filehandled is passed but ignored */
    SSIOFFSET offset;		/* disk offset of desired alignment  */

    if (afp->ssi == NULL) {
        return 0;
    }
    if (SSIGetOffsetByNumber (afp->ssi, idx, &fh, &offset) != 0) {
        return 0;
    }
    if (SSISetFilePosition (afp->f, &offset) != 0) {
        return 0;
    }
    return 1;
}


/* Function: MSAFileRead()
 * Date:     SRE, Fri May 28 16:01:43 1999 [St. Louis]
 *
 * Purpose:  Read the next msa from an open alignment file.
 *           This is a wrapper around format-specific calls.
 *
 * Args:     afp     - open alignment file
 *
 * Returns:  next alignment, or NULL if out of alignments
 */
MSA *
MSAFileRead (MSAFILE * afp)
{
    MSA *msa = NULL;

    switch (afp->format) {
    case MSAFILE_STOCKHOLM:
        msa = ReadStockholm (afp);
        break;
    case MSAFILE_MSF:
        msa = ReadMSF (afp);
        break;
    case MSAFILE_A2M:
        msa = ReadA2M (afp);
        break;
    case MSAFILE_CLUSTAL:
        msa = ReadClustal (afp);
        break;
    case MSAFILE_SELEX:
        msa = ReadSELEX (afp);
        break;
    case MSAFILE_PHYLIP:
        msa = ReadPhylip (afp);
        break;
    default:
        Die ("MSAFILE corrupted: bad format index");
    }
    return msa;
}

/* Function: MSAFileClose()
 * Date:     SRE, Tue May 18 14:05:28 1999 [St. Louis]
 *
 * Purpose:  Close an open MSAFILE.
 *
 * Args:     afp  - ptr to an open MSAFILE.
 *
 * Returns:  void
 */
void
MSAFileClose (MSAFILE * afp)
{
#ifndef SRE_STRICT_ANSI
    /* gzip functionality only on POSIX systems */
    if (afp->do_gzip) {
        pclose (afp->f);
    }
#endif
    if (!afp->do_stdin) {
        fclose (afp->f);
    }
    if (afp->buf != NULL) {
        free (afp->buf);
    }
    if (afp->ssi != NULL) {
        SSIClose (afp->ssi);
    }
    if (afp->fname != NULL) {
        free (afp->fname);
    }
    free (afp);
}

char *
MSAFileGetLine (MSAFILE * afp)
{
    char *s;
    if ((s = sre_fgets (&(afp->buf), &(afp->buflen), afp->f)) == NULL) {
        return NULL;
    }
    afp->linenumber++;
    return afp->buf;
}

void
MSAFileWrite (FILE * fp, MSA * msa, int outfmt, int do_oneline)
{
    switch (outfmt) {
    case MSAFILE_A2M:
        WriteA2M (fp, msa);
        break;
    case MSAFILE_CLUSTAL:
        WriteClustal (fp, msa);
        break;
    case MSAFILE_MSF:
        WriteMSF (fp, msa);
        break;
    case MSAFILE_PHYLIP:
        WritePhylip (fp, msa);
        break;
    case MSAFILE_SELEX:
        WriteSELEX (fp, msa);
        break;
    case MSAFILE_STOCKHOLM:
        if (do_oneline) {
            WriteStockholmOneBlock (fp, msa);
        }
        else {
            WriteStockholm (fp, msa);
        }
        break;
    default:
        Die ("can't write. no such alignment format %d\n", outfmt);
    }
}

/* Function: MSAGetSeqidx()
 * Date:     SRE, Wed May 19 15:08:25 1999 [St. Louis]
 *
 * Purpose:  From a sequence name, return seqidx appropriate
 *           for an MSA structure.
 *
 *           1) try to guess the index. (pass -1 if you can't guess)
 *           2) Look up name in msa's hashtable.
 *           3) If it's a new name, store in msa's hashtable;
 *                                  expand allocs as needed;
 *                                  save sqname.
 *
 * Args:     msa   - alignment object
 *           name  - a sequence name
 *           guess - a guess at the right index, or -1 if no guess.
 *
 * Returns:  seqidx
 */
int
MSAGetSeqidx (MSA * msa, char *name, int guess)
{
    int seqidx;
    /* can we guess? */
    if (guess >= 0 && guess < msa->nseq
            && strcmp (name, msa->sqname[guess]) == 0) {
        return guess;
    }
    /* else, a lookup in the index */
    if ((seqidx = GKIKeyIndex (msa->index, name)) >= 0) {
        return seqidx;
    }
    /* else, it's a new name */
    seqidx = GKIStoreKey (msa->index, name);
    if (seqidx >= msa->nseqalloc) {
        MSAExpand (msa);
    }

    msa->sqname[seqidx] = sre_strdup (name, -1);
    msa->nseq++;
    return seqidx;
}


/* Function: MSAFromAINFO()
 * Date:     SRE, Mon Jun 14 11:22:24 1999 [St. Louis]
 *
 * Purpose:  Convert the old aseq/ainfo alignment structure
 *           to new MSA structure. Enables more rapid conversion
 *           of codebase to the new world order.
 *
 * Args:     aseq  - [0..nseq-1][0..alen-1] alignment
 *           ainfo - old-style optional info
 *
 * Returns:  MSA *
 */
MSA *
MSAFromAINFO (char **aseq, AINFO * ainfo)
{
    MSA *msa;
    int i, j;

    msa = MSAAlloc (ainfo->nseq, ainfo->alen);
    for (i = 0; i < ainfo->nseq; i++) {
        strcpy (msa->aseq[i], aseq[i]);
        msa->wgt[i] = ainfo->wgt[i];
        msa->sqname[i] = sre_strdup (ainfo->sqinfo[i].name, -1);
        msa->sqlen[i] = msa->alen;
        GKIStoreKey (msa->index, msa->sqname[i]);

        if (ainfo->sqinfo[i].flags & SQINFO_ACC) {
            MSASetSeqAccession (msa, i, ainfo->sqinfo[i].acc);
        }

        if (ainfo->sqinfo[i].flags & SQINFO_DESC) {
            MSASetSeqDescription (msa, i, ainfo->sqinfo[i].desc);
        }

        if (ainfo->sqinfo[i].flags & SQINFO_SS) {
            if (msa->ss == NULL) {
                msa->ss = MallocOrDie (sizeof (char *) * msa->nseqalloc);
                msa->sslen = MallocOrDie (sizeof (int) * msa->nseqalloc);
                for (j = 0; j < msa->nseqalloc; j++) {
                    msa->ss[j] = NULL;
                    msa->sslen[j] = 0;
                }
            }
            MakeAlignedString (msa->aseq[i], msa->alen, ainfo->sqinfo[i].ss,
                               &(msa->ss[i]));
            msa->sslen[i] = msa->alen;
        }

        if (ainfo->sqinfo[i].flags & SQINFO_SA) {
            if (msa->sa == NULL) {
                msa->sa = MallocOrDie (sizeof (char *) * msa->nseqalloc);
                msa->salen = MallocOrDie (sizeof (int) * msa->nseqalloc);
                for (j = 0; j < msa->nseqalloc; j++) {
                    msa->sa[j] = NULL;
                    msa->salen[j] = 0;
                }
            }
            MakeAlignedString (msa->aseq[i], msa->alen, ainfo->sqinfo[i].sa,
                               &(msa->sa[i]));
            msa->salen[i] = msa->alen;
        }
    }
    /* note that sre_strdup() returns NULL when passed NULL */
    msa->name = sre_strdup (ainfo->name, -1);
    msa->desc = sre_strdup (ainfo->desc, -1);
    msa->acc = sre_strdup (ainfo->acc, -1);
    msa->au = sre_strdup (ainfo->au, -1);
    msa->ss_cons = sre_strdup (ainfo->cs, -1);
    msa->rf = sre_strdup (ainfo->rf, -1);
    if (ainfo->flags & AINFO_TC) {
        msa->cutoff[MSA_CUTOFF_TC1] = ainfo->tc1;
        msa->cutoff_is_set[MSA_CUTOFF_TC1] = TRUE;
        msa->cutoff[MSA_CUTOFF_TC2] = ainfo->tc2;
        msa->cutoff_is_set[MSA_CUTOFF_TC2] = TRUE;
    }
    if (ainfo->flags & AINFO_NC) {
        msa->cutoff[MSA_CUTOFF_NC1] = ainfo->nc1;
        msa->cutoff_is_set[MSA_CUTOFF_NC1] = TRUE;
        msa->cutoff[MSA_CUTOFF_NC2] = ainfo->nc2;
        msa->cutoff_is_set[MSA_CUTOFF_NC2] = TRUE;
    }
    if (ainfo->flags & AINFO_GA) {
        msa->cutoff[MSA_CUTOFF_GA1] = ainfo->ga1;
        msa->cutoff_is_set[MSA_CUTOFF_GA1] = TRUE;
        msa->cutoff[MSA_CUTOFF_GA2] = ainfo->ga2;
        msa->cutoff_is_set[MSA_CUTOFF_GA2] = TRUE;
    }
    msa->nseq = ainfo->nseq;
    msa->alen = ainfo->alen;
    return msa;
}




/* Function: MSAFileFormat()
 * Date:     SRE, Fri Jun 18 14:26:49 1999 [Sanger Centre]
 *
 * Purpose:  (Attempt to) determine the format of an alignment file.
 *           Since it rewinds the file pointer when it's done,
 *           cannot be used on a pipe or gzip'ed file. Works by
 *           calling SeqfileFormat() from sqio.c, then making sure
 *           that the format is indeed an alignment. If the format
 *           comes back as FASTA, it assumes that the format as A2M
 *           (e.g. aligned FASTA).
 *
 * Args:     fname   - file to evaluate
 *
 * Returns:  format code; e.g. MSAFILE_STOCKHOLM
 */
int
MSAFileFormat (MSAFILE * afp)
{
    int fmt;

    fmt = SeqfileFormat (afp->f);

    if (fmt == SQFILE_FASTA) {
        fmt = MSAFILE_A2M;
    }

    if (fmt != MSAFILE_UNKNOWN && !IsAlignmentFormat (fmt))
        Die ("File %s does not appear to be an alignment file;\n\
rather, it appears to be an unaligned file in %s format.\n\
I'm expecting an alignment file in this context.\n", afp->fname, SeqfileFormat2String (fmt));
    return fmt;
}


/* Function: MSAMingap()
 * Date:     SRE, Mon Jun 28 18:57:54 1999 [on jury duty, St. Louis Civil Court]
 *
 * Purpose:  Remove all-gap columns from a multiple sequence alignment
 *           and its associated per-residue data.
 *
 * Args:     msa - the alignment
 *
 * Returns:  (void)
 */
void
MSAMingap (MSA * msa)
{
    int *useme;			/* array of TRUE/FALSE flags for which columns to keep */
    int apos;			/* position in original alignment */
    int idx;			/* sequence index */

    useme = MallocOrDie (sizeof (int) * msa->alen);
    for (apos = 0; apos < msa->alen; apos++) {
        for (idx = 0; idx < msa->nseq; idx++) {
            if (!isgap (msa->aseq[idx][apos])) {
                break;
            }
				}		
        if (idx == msa->nseq) {
            useme[apos] = FALSE;
        }
        else {
            useme[apos] = TRUE;
        }
    }
    MSAShorterAlignment (msa, useme);
    free (useme);
    return;
}

/* Function: MSANogap()
 * Date:     SRE, Wed Nov 17 09:59:51 1999 [St. Louis]
 *
 * Purpose:  Remove all columns from a multiple sequence alignment that
 *           contain any gaps -- used for filtering before phylogenetic
 *           analysis.
 *
 * Args:     msa - the alignment
 *
 * Returns:  (void). The alignment is modified, so if you want to keep
 *           the original for something, make a copy.
 */
void
MSANogap (MSA * msa)
{
    int *useme;			/* array of TRUE/FALSE flags for which columns to keep */
    int apos;			/* position in original alignment */
    int idx;			/* sequence index */

    useme = MallocOrDie (sizeof (int) * msa->alen);
    for (apos = 0; apos < msa->alen; apos++) {
        for (idx = 0; idx < msa->nseq; idx++) {
            if (isgap (msa->aseq[idx][apos])) {
                break;
            }
				}		
        if (idx == msa->nseq) {
            useme[apos] = TRUE;
        }
        else {
            useme[apos] = FALSE;
        }
    }
    MSAShorterAlignment (msa, useme);
    free (useme);
    return;
}


/* Function: MSAShorterAlignment()
 * Date:     SRE, Wed Nov 17 09:49:32 1999 [St. Louis]
 *
 * Purpose:  Given an array "useme" (0..alen-1) of TRUE/FALSE flags,
 *           where TRUE means "keep this column in the new alignment":
 *           Remove all columns annotated as "FALSE" in the useme
 *           array.
 *
 * Args:     msa   - the alignment. The alignment is changed, so
 *                   if you don't want the original screwed up, make
 *                   a copy of it first.
 *           useme - TRUE/FALSE flags for columns to keep: 0..alen-1
 *
 * Returns:  (void)
 */
void
MSAShorterAlignment (MSA * msa, int *useme)
{
    int apos;			/* position in original alignment */
    int mpos;			/* position in new alignment      */
    int idx;			/* sequence index */
    int i;			/* markup index */

    /* Since we're minimizing, we can overwrite, using already allocated
     * memory.
     */
    for (apos = 0, mpos = 0; apos < msa->alen; apos++) {
        if (useme[apos] == FALSE) {
            continue;
        }

        /* shift alignment and associated per-column+per-residue markup */
        if (mpos != apos) {
            for (idx = 0; idx < msa->nseq; idx++) {
                msa->aseq[idx][mpos] = msa->aseq[idx][apos];
                if (msa->ss != NULL && msa->ss[idx] != NULL) {
                    msa->ss[idx][mpos] = msa->ss[idx][apos];
                }
                if (msa->sa != NULL && msa->sa[idx] != NULL) {
                    msa->sa[idx][mpos] = msa->sa[idx][apos];
                }

                for (i = 0; i < msa->ngr; i++) {
                    if (msa->gr[i][idx] != NULL) {
                        msa->gr[i][idx][mpos] = msa->gr[i][idx][apos];
                    }
								}		
            }

            if (msa->ss_cons != NULL) {
                msa->ss_cons[mpos] = msa->ss_cons[apos];
            }
            if (msa->sa_cons != NULL) {
                msa->sa_cons[mpos] = msa->sa_cons[apos];
            }
            if (msa->rf != NULL) {
                msa->rf[mpos] = msa->rf[apos];
            }

            for (i = 0; i < msa->ngc; i++) {
                msa->gc[i][mpos] = msa->gc[i][apos];
            }
        }
        mpos++;
    }

    msa->alen = mpos;		/* set new length */
    /* null terminate everything */
    for (idx = 0; idx < msa->nseq; idx++) {
        msa->aseq[idx][mpos] = '\0';
        if (msa->ss != NULL && msa->ss[idx] != NULL) {
            msa->ss[idx][mpos] = '\0';
        }
        if (msa->sa != NULL && msa->sa[idx] != NULL) {
            msa->sa[idx][mpos] = '\0';
        }

        for (i = 0; i < msa->ngr; i++) {
            if (msa->gr[i][idx] != NULL) {
                msa->gr[i][idx][mpos] = '\0';
            }
				}		
    }

    if (msa->ss_cons != NULL) {
        msa->ss_cons[mpos] = '\0';
    }
    if (msa->sa_cons != NULL) {
        msa->sa_cons[mpos] = '\0';
    }
    if (msa->rf != NULL) {
        msa->rf[mpos] = '\0';
    }

    for (i = 0; i < msa->ngc; i++) {
        msa->gc[i][mpos] = '\0';
    }

    return;
}


/* Function: MSASmallerAlignment()
 * Date:     SRE, Wed Jun 30 09:56:08 1999 [St. Louis]
 *
 * Purpose:  Given an array "useme" of TRUE/FALSE flags for
 *           each sequence in an alignment, construct
 *           and return a new alignment containing only
 *           those sequences that are flagged useme=TRUE.
 *
 *           Used by routines such as MSAFilterAlignment()
 *           and MSASampleAlignment().
 *
 * Limitations:
 *           Does not copy unparsed Stockholm markup.
 *
 *           Does not make assumptions about meaning of wgt;
 *           if you want the new wgt vector renormalized, do
 *           it yourself with FNorm(new->wgt, new->nseq).
 *
 * Args:     msa     -- the original (larger) alignment
 *           useme   -- [0..nseq-1] array of TRUE/FALSE flags; TRUE means include
 *                      this seq in new alignment
 *           ret_new -- RETURN: new alignment
 *
 * Returns:  void
 *           ret_new is allocated here; free with MSAFree()
 */
void
MSASmallerAlignment (MSA * msa, int *useme, MSA ** ret_new)
{
    MSA *new;			/* RETURN: new alignment */
    int nnew;			/* number of seqs in new msa (e.g. # of TRUEs) */
    int oidx, nidx;		/* old, new indices */
    int i;

    nnew = 0;
    for (oidx = 0; oidx < msa->nseq; oidx++) {
        if (useme[oidx]) {
            nnew++;
        }
		}		
    if (nnew == 0) {
        *ret_new = NULL;
        return;
    }

    new = MSAAlloc (nnew, 0);
    nidx = 0;
    for (oidx = 0; oidx < msa->nseq; oidx++) {
        if (useme[oidx]) {
            new->aseq[nidx] = sre_strdup (msa->aseq[oidx], msa->alen);
            new->sqname[nidx] = sre_strdup (msa->sqname[oidx], msa->alen);
            GKIStoreKey (new->index, msa->sqname[oidx]);
            new->wgt[nidx] = msa->wgt[oidx];
            if (msa->sqacc != NULL) {
                MSASetSeqAccession (new, nidx, msa->sqacc[oidx]);
            }
            if (msa->sqdesc != NULL) {
                MSASetSeqDescription (new, nidx, msa->sqdesc[oidx]);
            }
            if (msa->ss != NULL && msa->ss[oidx] != NULL) {
                if (new->ss == NULL) {
                    new->ss = MallocOrDie (sizeof (char *) * new->nseq);
                }
                new->ss[nidx] = sre_strdup (msa->ss[oidx], -1);
            }
            if (msa->sa != NULL && msa->sa[oidx] != NULL) {
                if (new->sa == NULL) {
                    new->sa = MallocOrDie (sizeof (char *) * new->nseq);
                }
                new->sa[nidx] = sre_strdup (msa->sa[oidx], -1);
            }
            nidx++;
        }
		}		

    new->nseq = nnew;
    new->alen = msa->alen;
    new->flags = msa->flags;
    new->type = msa->type;
    new->name = sre_strdup (msa->name, -1);
    new->desc = sre_strdup (msa->desc, -1);
    new->acc = sre_strdup (msa->acc, -1);
    new->au = sre_strdup (msa->au, -1);
    new->ss_cons = sre_strdup (msa->ss_cons, -1);
    new->sa_cons = sre_strdup (msa->sa_cons, -1);
    new->rf = sre_strdup (msa->rf, -1);
    for (i = 0; i < MSA_MAXCUTOFFS; i++) {
        new->cutoff[i] = msa->cutoff[i];
        new->cutoff_is_set[i] = msa->cutoff_is_set[i];
    }
    free (new->sqlen);

    MSAMingap (new);
    *ret_new = new;
    return;
}


/*****************************************************************
 * Retrieval routines
 *
 * Access to MSA structure data is possible through these routines.
 * I'm not doing this because of object oriented design, though
 * it might work in my favor someday.
 * I'm doing this because lots of MSA data is optional, and
 * checking through the chain of possible NULLs is a pain.
 *****************************************************************/

char *
MSAGetSeqAccession (MSA * msa, int idx)
{
    if (msa->sqacc != NULL && msa->sqacc[idx] != NULL) {
        return msa->sqacc[idx];
    }
    else {
        return NULL;
    }
}

char *
MSAGetSeqDescription (MSA * msa, int idx)
{
    if (msa->sqdesc != NULL && msa->sqdesc[idx] != NULL) {
        return msa->sqdesc[idx];
    }
    else {
        return NULL;
    }
}

char *
MSAGetSeqSS (MSA * msa, int idx)
{
    if (msa->ss != NULL && msa->ss[idx] != NULL) {
        return msa->ss[idx];
    }
    else {
        return NULL;
    }
}

char *
MSAGetSeqSA (MSA * msa, int idx)
{
    if (msa->sa != NULL && msa->sa[idx] != NULL) {
        return msa->sa[idx];
    }
    else {
        return NULL;
    }
}


/*****************************************************************
 * Information routines
 *
 * Access information about the MSA.
 *****************************************************************/

/* Function: MSAAverageSequenceLength()
 * Date:     SRE, Sat Apr  6 09:41:34 2002 [St. Louis]
 *
 * Purpose:  Return the average length of the (unaligned) sequences
 *           in the MSA.
 *
 * Args:     msa  - the alignment
 *
 * Returns:  average length
 */
float
MSAAverageSequenceLength (MSA * msa)
{
    int i;
    float avg;

    avg = 0.;
    for (i = 0; i < msa->nseq; i++) {
        avg += (float) DealignedLength (msa->aseq[i]);
    }

    if (msa->nseq == 0) {
        return 0.;
    }
    else {
        return (avg / msa->nseq);
    }
}


/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* msf.c
 * SRE, Sun Jul 11 16:17:32 1993
 *
 * Import/export of GCG MSF multiple sequence alignment
 * formatted files. Designed using format specifications
 * kindly provided by Steve Smith of Genetics Computer Group.
 *
 * CVS $Id: msf.c,v 1.6 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include  <time.h>
#include "squid.h"
#include "msa.h"

#ifdef TESTDRIVE_MSF
/*****************************************************************
 * msf.c test driver:
 * cc -DTESTDRIVE_MSF -g -O2 -Wall -o test msf.c msa.c gki.c sqerror.c sre_string.c file.c hsregex.c sre_math.c sre_ctype.c sqio.c alignio.c selex.c interleaved.c types.c -lm
 *
 */
int
main (int argc, char **argv)
{
    MSAFILE *afp;
    MSA *msa;
    char *file;

    file = argv[1];

    if ((afp = MSAFileOpen (file, MSAFILE_STOCKHOLM, NULL)) == NULL) {
        Die ("Couldn't open %s\n", file);
    }

    while ((msa = ReadMSF (afp)) != NULL) {
        WriteMSF (stdout, msa);
        MSAFree (msa);
    }

    MSAFileClose (afp);
    exit (0);
}

/******************************************************************/
#endif /* testdrive_msf */



/* Function: ReadMSF()
 * Date:     SRE, Tue Jun  1 08:07:22 1999 [St. Louis]
 *
 * Purpose:  Parse an alignment read from an open MSF format
 *           alignment file. (MSF is a single-alignment format.)
 *           Return the alignment, or NULL if we've already
 *           read the alignment.
 *
 * Args:     afp  - open alignment file
 *
 * Returns:  MSA * - an alignment object
 *                   caller responsible for an MSAFree()
 *           NULL if no more alignments
 *
 * Diagnostics:
 *           Will Die() here with a (potentially) useful message
 *           if a parsing error occurs.
 */
MSA *
ReadMSF (MSAFILE * afp)
{
    MSA *msa;
    char *s;
    int alleged_alen;
    int alleged_type;
    int alleged_checksum;
    char *tok;
    char *sp;
    int slen;
    int sqidx;
    char *name;
    char *seq;

    if (feof (afp->f)) {
        return NULL;
    }
    if ((s = MSAFileGetLine (afp)) == NULL) {
        return NULL;
    }

    /* The first line is the header.
     * This is a new-ish GCG feature. Don't count on it, so
     * we can be a bit more tolerant towards non-GCG software
     * generating "MSF" files.
     */
    msa = MSAAlloc (10, 0);
    if (strncmp (s, "!!AA_MULTIPLE_ALIGNMENT", 23) == 0) {
        msa->type = kAmino;
        if ((s = MSAFileGetLine (afp)) == NULL) {
            return NULL;
        }
    }
    else if (strncmp (s, "!!NA_MULTIPLE_ALIGNMENT", 23) == 0) {
        msa->type = kRNA;
        if ((s = MSAFileGetLine (afp)) == NULL) {
            return NULL;
        }
    }

    /* Now we're in the free text comment section of the MSF file.
     * It ends when we see the "MSF: Type: Check: .." line.
     * This line must be present.
     */
    do {
        if ((strstr (s, "..") != NULL && strstr (s, "MSF:") != NULL) &&
                Strparse
                ("^.+MSF: +([0-9]+) +Type: +([PNX]).+Check: +([0-9]+) +\\.\\.", s,
                 3)) {
            alleged_alen = atoi (sqd_parse[0]);
            switch (*(sqd_parse[1])) {
            case 'N':
                alleged_type = kRNA;
                break;
            case 'P':
                alleged_type = kAmino;
                break;
            case 'X':
                alleged_type = kOtherSeq;
                break;
            default:
                alleged_type = kOtherSeq;
            }
            alleged_checksum = atoi (sqd_parse[3]);
            if (msa->type == kOtherSeq) {
                msa->type = alleged_type;
            }
            break;			/* we're done with comment section. */
        }
        if (!IsBlankline (s)) {
            MSAAddComment (msa, s);
        }
    } while ((s = MSAFileGetLine (afp)) != NULL);

    /* Now we're in the name section.
     * GCG has a relatively poorly documented feature: only sequences that
     * appear in this list will be read from the alignment section. Commenting
     * out sequences in the name list (by preceding them with "!") is
     * allowed as a means of manually defining subsets of sequences in
     * the alignment section. We can support this feature reasonably
     * easily because of the hash table for names in the MSA: we
     * only add names to the hash table when we see 'em in the name section.
     */
    while ((s = MSAFileGetLine (afp)) != NULL) {
        while ((*s == ' ' || *s == '\t') && *s) {
            s++;    /* skip leading whitespace */
        }

        if (*s == '\n') {
            continue;    /* skip blank lines */
        }
        else if (*s == '!') {
            MSAAddComment (msa, s);
        }
        else if ((sp = strstr (s, "Name:")) != NULL) {
            /* We take the name and the weigh, and that's it */
            sp += 5;
            tok = sre_strtok (&sp, " \t", &slen);	/* <sequence name> */
            sqidx = GKIStoreKey (msa->index, tok);
            if (sqidx >= msa->nseqalloc) {
                MSAExpand (msa);
            }
            msa->sqname[sqidx] = sre_strdup (tok, slen);
            msa->nseq++;

            if ((sp = strstr (sp, "Weight:")) == NULL)
                Die ("No Weight: on line %d for %s in name section of MSF file %s\n",
                     afp->linenumber, msa->sqname[sqidx], afp->fname);
            sp += 7;
            tok = sre_strtok (&sp, " \t", &slen);
            msa->wgt[sqidx] = atof (tok);
            msa->flags |= MSA_SET_WGT;
        }
        else if (strncmp (s, "//", 2) == 0) {
            break;
        }
        else {
            Die ("Invalid line (probably %d) in name section of MSF file %s:\n%s\n",
                 afp->linenumber, afp->fname, s);
            squid_errno = SQERR_FORMAT;	/* NOT THREADSAFE */
            return NULL;
        }

    }

    /* And now we're in the sequence section.
     * As discussed above, if we haven't seen a sequence name, then we
     * don't include the sequence in the alignment.
     * Also, watch out for coordinate-only lines.
     */
    while ((s = MSAFileGetLine (afp)) != NULL) {
        sp = s;
        if ((name = sre_strtok (&sp, " \t", NULL)) == NULL) {
            continue;
        }
        if ((seq = sre_strtok (&sp, "\n", &slen)) == NULL) {
            continue;
        }

        /* The test for a coord line: digits starting both fields
         */
        if (isdigit ((int) *name) && isdigit ((int) *seq)) {
            continue;
        }

        /* It's not blank, and it's not a coord line: must be sequence
         */
        sqidx = GKIKeyIndex (msa->index, name);
        if (sqidx < 0) {
            continue;    /* not a sequence we recognize */
        }

        msa->sqlen[sqidx] =
            sre_strcat (&(msa->aseq[sqidx]), msa->sqlen[sqidx], seq, slen);
    }

    /* We've left blanks in the aseqs; take them back out.
     */
    for (sqidx = 0; sqidx < msa->nseq; sqidx++) {
        if (msa->aseq[sqidx] == NULL)
            Die ("Didn't find a sequence for %s in MSF file %s\n",
                 msa->sqname[sqidx], afp->fname);

        for (s = sp = msa->aseq[sqidx]; *s != '\0'; s++) {
            if (*s == ' ' || *s == '\t') {
                msa->sqlen[sqidx]--;
            }
            else {
                *sp = *s;
                sp++;
            }
        }
        *sp = '\0';
    }

    MSAVerifyParse (msa);		/* verifies, and also sets alen and wgt. */
    return msa;
}


/* Function: WriteMSF()
 * Date:     SRE, Mon May 31 11:25:18 1999 [St. Louis]
 *
 * Purpose:  Write an alignment in MSF format to an open file.
 *
 * Args:     fp    - file that's open for writing.
 *           msa   - alignment to write.
 *
 *                   Note that msa->type, usually optional, must be
 *                   set for WriteMSF to work. If it isn't, a fatal
 *                   error is generated.
 *
 * Returns:  (void)
 */
void
WriteMSF (FILE * fp, MSA * msa)
{
    time_t now;			/* current time as a time_t */
    char date[64];		/* today's date in GCG's format "October 3, 1996 15:57" */
    char **gcg_aseq;		/* aligned sequences with gaps converted to GCG format */
    char **gcg_sqname;		/* sequence names with GCG-valid character sets */
    int idx;			/* counter for sequences         */
    char *s;			/* pointer into sqname or seq    */
    int len;			/* tmp variable for name lengths */
    int namelen;			/* maximum name length used      */
    int pos;			/* position counter              */
    char buffer[51];		/* buffer for writing seq        */
    int i;			/* another position counter */

    /*****************************************************************
     * Make copies of sequence names and sequences.
     *   GCG recommends that name characters should only contain
     *   alphanumeric characters, -, or _
     *   Some GCG and GCG-compatible software is sensitive to this.
     *   We silently convert all other characters to '_'.
     *
     *   For sequences, GCG allows only ~ and . for gaps.
     *   Otherwise, everthing is interpreted as a residue;
     *   so squid's IUPAC-restricted chars are fine. ~ means
     *   an external gap. . means an internal gap.
     *****************************************************************/

    /* make copies that we can edit */
    gcg_aseq = MallocOrDie (sizeof (char *) * msa->nseq);
    gcg_sqname = MallocOrDie (sizeof (char *) * msa->nseq);
    for (idx = 0; idx < msa->nseq; idx++) {
        gcg_aseq[idx] = sre_strdup (msa->aseq[idx], msa->alen);
        gcg_sqname[idx] = sre_strdup (msa->sqname[idx], -1);
    }
    /* alter names as needed  */
    for (idx = 0; idx < msa->nseq; idx++) {
        for (s = gcg_sqname[idx]; *s != '\0'; s++) {
            if (!isalnum ((int) *s) && *s != '-' && *s != '_') {
                *s = '_';
            }
				}
		}		
    /* alter gap chars in seq  */
    for (idx = 0; idx < msa->nseq; idx++) {
        for (s = gcg_aseq[idx]; *s != '\0' && isgap (*s); s++) {
            *s = '~';
        }
        for (; *s != '\0'; s++) {
            if (isgap (*s)) {
                *s = '.';
            }
				}		
        for (pos = msa->alen - 1; pos > 0 && isgap (gcg_aseq[idx][pos]); pos--) {
            gcg_aseq[idx][pos] = '~';
        }
    }
    /* calculate max namelen used */
    namelen = 0;
    for (idx = 0; idx < msa->nseq; idx++) {
        if ((len = strlen (msa->sqname[idx])) > namelen) {
            namelen = len;
        }
		}
    /*****************************************************
     * Write the MSF header
     *****************************************************/
    /* required file type line */
    if (msa->type == kOtherSeq) {
        msa->type = GuessAlignmentSeqtype (msa->aseq, msa->nseq);
    }

    if (msa->type == kRNA) {
        fprintf (fp, "!!NA_MULTIPLE_ALIGNMENT 1.0\n");
    }
    else if (msa->type == kDNA) {
        fprintf (fp, "!!NA_MULTIPLE_ALIGNMENT 1.0\n");
    }
    else if (msa->type == kAmino) {
        fprintf (fp, "!!AA_MULTIPLE_ALIGNMENT 1.0\n");
    }
    else if (msa->type == kOtherSeq)
        Die
        ("WriteMSF(): couldn't guess whether that alignment is RNA or protein.\n");
    else {
        Die ("Invalid sequence type %d in WriteMSF()\n", msa->type);
    }

    /* free text comments */
    if (msa->ncomment > 0) {
        for (idx = 0; idx < msa->ncomment; idx++) {
            fprintf (fp, "%s\n", msa->comment[idx]);
        }
        fprintf (fp, "\n");
    }
    /* required checksum line */
    now = time (NULL);
    if (strftime (date, 64, "%B %d, %Y %H:%M", localtime (&now)) == 0) {
        Die ("What time is it on earth? strftime() failed in WriteMSF().\n");
    }
    fprintf (fp, " %s  MSF: %d  Type: %c  %s  Check: %d  ..\n",
             msa->name != NULL ? msa->name : "squid.msf",
             msa->alen,
             msa->type == kRNA ? 'N' : 'P',
             date, GCGMultchecksum (gcg_aseq, msa->nseq));
    fprintf (fp, "\n");

    /*****************************************************
     * Names/weights section
     *****************************************************/

    for (idx = 0; idx < msa->nseq; idx++) {
        fprintf (fp, " Name: %-*.*s  Len:  %5d  Check: %4d  Weight: %.2f\n",
                 namelen, namelen,
                 gcg_sqname[idx],
                 msa->alen,
                 GCGchecksum (gcg_aseq[idx], msa->alen), msa->wgt[idx]);
    }
    fprintf (fp, "\n");
    fprintf (fp, "//\n");

    /*****************************************************
     * Write the sequences
     *****************************************************/

    for (pos = 0; pos < msa->alen; pos += 50) {
        fprintf (fp, "\n");		/* Blank line between sequence blocks */

        /* Coordinate line */
        len = (pos + 50) > msa->alen ? msa->alen - pos : 50;
        if (len > 10)
            fprintf (fp, "%*s  %-6d%*s%6d\n", namelen, "",
                     pos + 1, len + ((len - 1) / 10) - 12, "", pos + len);
        else {
            fprintf (fp, "%*s  %-6d\n", namelen, "", pos + 1);
        }

        for (idx = 0; idx < msa->nseq; idx++) {
            fprintf (fp, "%-*s ", namelen, gcg_sqname[idx]);
            /* get next line's worth of 50 from seq */
            strncpy (buffer, gcg_aseq[idx] + pos, 50);
            buffer[50] = '\0';
            /* draw the sequence line */
            for (i = 0; i < len; i++) {
                if (!(i % 10)) {
                    fputc (' ', fp);
                }
                fputc (buffer[i], fp);
            }
            fputc ('\n', fp);
        }
    }

    Free2DArray ((void **) gcg_aseq, msa->nseq);
    Free2DArray ((void **) gcg_sqname, msa->nseq);
    return;
}



/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* modelmakers.c
 * SRE, Fri Nov 15 10:00:04 1996
 *
 * Construction of models from multiple alignments. Three versions:
 *    Handmodelmaker() -- use #=RF annotation to indicate match columns
 *    Fastmodelmaker() -- Krogh/Haussler heuristic
 *    Maxmodelmaker()  -- MAP model construction algorithm (Eddy,
 *                          unpublished)
 *
 * The meat of the model construction code is in matassign2hmm().
 * The three model construction strategies simply label which columns
 * are supposed to be match states, and then hand this info to
 * matassign2hmm().
 *
 * Two wrinkles to watch for:
 * 1) The alignment is assumed to contain sequence fragments. Look in
 *    fake_tracebacks() for how internal entry/exit points are handled.
 * 2) Plan7 disallows DI and ID transitions, but an alignment may
 *    imply these. Look in trace_doctor() for how DI and ID transitions
 *    are removed.
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <float.h>
#include <ctype.h>

#include "squid.h"
#include "msa.h"
#include "structs.h"
#include "funcs.h"



/* flags used for matassign[] arrays --
 *   assignment of aligned columns to match/insert states
 */
#define ASSIGN_MATCH      (1<<0)
#define FIRST_MATCH       (1<<1)
#define LAST_MATCH        (1<<2)
#define ASSIGN_INSERT     (1<<3)
#define EXTERNAL_INSERT_N (1<<4)
#define EXTERNAL_INSERT_C (1<<5)

static int build_cij (char **aseqs, int nseq, int *insopt, int i, int j,
                      float *wgt, float *cij);
static int estimate_model_length (MSA * msa);
static void matassign2hmm (MSA * msa, char **dsq,
                           int *matassign, struct plan7_s **ret_hmm,
                           struct p7trace_s ***ret_tr);
static void fake_tracebacks (char **aseq, int nseq, int alen, int *matassign,
                             struct p7trace_s ***ret_tr);
static void trace_doctor (struct p7trace_s *tr, int M, int *ret_ndi,
                          int *ret_nid);
static void annotate_model (struct plan7_s *hmm, int *matassign, MSA * msa);
static void print_matassign (int *matassign, int alen);



/* Function: P7Handmodelmaker()
 *
 * Purpose:  Manual model construction:
 *           Construct an HMM from an alignment, where the #=RF line
 *           of a HMMER alignment file is given to indicate
 *           the columns assigned to matches vs. inserts.
 *
 *           NOTE: Handmodelmaker() will slightly revise the alignment
 *           if necessary, if the assignment of columns implies
 *           DI and ID transitions.
 *
 *           Returns both the HMM in counts form (ready for applying
 *           Dirichlet priors as the next step), and fake tracebacks
 *           for each aligned sequence.
 *
 * Args:     msa  - multiple sequence alignment
 *           dsq  - digitized unaligned aseq's
 *           ret_hmm - RETURN: counts-form HMM
 *           ret_tr  - RETURN: array of tracebacks for aseq's
 *
 * Return:   (void)
 *           ret_hmm and ret_tr alloc'ed here; FreeTrace(tr[i]), free(tr),
 *           FreeHMM(hmm).
 */
void
P7Handmodelmaker (MSA * msa, char **dsq,
                  struct plan7_s **ret_hmm, struct p7trace_s ***ret_tr)
{
    int *matassign;		/* MAT state assignments if 1; 1..alen */
    int apos;			/* counter for aligned columns         */

    /* Make sure we have all the info about the alignment that we need */
    if (msa->rf == NULL) {
        Die ("Alignment must have RF annotation to hand-build an HMM");
    }

    /* Allocation */
    matassign = (int *) MallocOrDie (sizeof (int) * (msa->alen + 1));

    /* Determine match assignment from optional annotation
     */
    matassign[0] = 0;
    for (apos = 0; apos < msa->alen; apos++) {
        matassign[apos + 1] = 0;
        if (!isgap (msa->rf[apos])) {
            matassign[apos + 1] |= ASSIGN_MATCH;
        }
        else {
            matassign[apos + 1] |= ASSIGN_INSERT;
        }
    }

    /* Hand matassign off for remainder of model construction
     */
    /*   print_matassign(matassign, msa->alen); */
    matassign2hmm (msa, dsq, matassign, ret_hmm, ret_tr);

    free (matassign);
    return;
}


/* Function: P7Fastmodelmaker()
 *
 * Purpose:  Heuristic model construction:
 *           Construct an HMM from an alignment using the original
 *           Krogh/Haussler heuristic; any column with more
 *           symbols in it than a given fraction is assigned to
 *           match.
 *
 *           NOTE: Fastmodelmaker() will slightly revise the
 *           alignment if the assignment of columns implies
 *           DI and ID transitions.
 *
 *           Returns the HMM in counts form (ready for applying Dirichlet
 *           priors as the next step). Also returns fake traceback
 *           for each training sequence.
 *
 * Args:     msa     - multiple sequence alignment
 *           dsq     - digitized unaligned aseq's
 *           maxgap  - if more gaps than this, column becomes insert.
 *           ret_hmm - RETURN: counts-form HMM
 *           ret_tr  - RETURN: array of tracebacks for aseq's
 *
 * Return:   (void)
 *           ret_hmm and ret_tr alloc'ed here; FreeTrace(tr[i]), free(tr),
 *           FreeHMM(hmm).
 */
void
P7Fastmodelmaker (MSA * msa, char **dsq, float maxgap,
                  struct plan7_s **ret_hmm, struct p7trace_s ***ret_tr)
{
    int *matassign;		/* MAT state assignments if 1; 1..alen */
    int idx;			/* counter over sequences              */
    int apos;			/* counter for aligned columns         */
    int ngap;			/* number of gaps in a column          */

    /* Allocations: matassign is 1..alen array of bit flags
     */
    matassign = (int *) MallocOrDie (sizeof (int) * (msa->alen + 1));

    /* Determine match assignment by counting symbols in columns
     */
    matassign[0] = 0;
    for (apos = 0; apos < msa->alen; apos++) {
        matassign[apos + 1] = 0;

        ngap = 0;
        for (idx = 0; idx < msa->nseq; idx++) {
            if (isgap (msa->aseq[idx][apos])) {
                ngap++;
            }
				}
        if ((float) ngap / (float) msa->nseq > maxgap) {
            matassign[apos + 1] |= ASSIGN_INSERT;
        }
        else {
            matassign[apos + 1] |= ASSIGN_MATCH;
        }
    }

    /* Once we have matassign calculated, all modelmakers behave
     * the same; matassign2hmm() does this stuff (traceback construction,
     * trace counting) and sets up ret_hmm and ret_tr.
     */
    matassign2hmm (msa, dsq, matassign, ret_hmm, ret_tr);

    free (matassign);
    return;
}


/* Function: P7Maxmodelmaker()
 *
 * Purpose:  The Unholy Beast of HMM model construction algorithms --
 *           maximum a posteriori construction. A tour de force and
 *           probably overkill. MAP construction for Krogh
 *           HMM-profiles is fairly straightforward, but MAP construction of
 *           Plan 7 HMM-profiles is, er, intricate.
 *
 *           Given a multiple alignment, construct an optimal (MAP) model
 *           architecture. Return a counts-based HMM.
 *
 * Args:     msa     - multiple sequence alignment
 *           dsq     - digitized, unaligned seqs
 *           maxgap  - above this, trailing columns are assigned to C
 *           prior   - priors on parameters to use for model construction
 *           null    - random sequence model emissions
 *           null_p1 - random sequence model p1 transition
 *           mpri    - prior on architecture: probability of new match node
 *           ret_hmm - RETURN: new hmm (counts form)
 *           ret_tr  - RETURN: array of tracebacks for aseq's
 *
 * Return:   (void)
 *           ret_hmm and ret_tr (if !NULL) must be free'd by the caller.
 */
void
P7Maxmodelmaker (MSA * msa, char **dsq, float maxgap,
                 struct p7prior_s *prior,
                 float *null, float null_p1, float mpri,
                 struct plan7_s **ret_hmm, struct p7trace_s ***ret_tr)
{
    int idx;			/* counter for seqs                */
    int i, j;			/* positions in alignment          */
    int x;			/* counter for syms or transitions */
    float **matc;			/* count vectors: [1..alen][0..19] */
    float cij[8], tij[8];		/* count and score transit vectors */
    float matp[MAXABET];		/* match emission vector           */
    float insp[MAXABET];		/* insert score vector             */
    float insc[MAXABET];		/* insert count vector             */
    float *sc;			/* DP scores [0,1..alen,alen+1]    */
    int *tbck;			/* traceback ptrs for sc           */
    int *matassign;		/* match assignments [1..alen]     */
    int *insopt;			/* number of inserted chars [0..nseq-1] */
    int first, last;		/* positions of first and last cols [1..alen] */
    float bm1, bm2;		/* estimates for start,internal b->m t's  */
    int est_M;			/* estimate for the size of the model */
    float t_me;			/* estimate for an internal M->E transition */
    float new, bestsc;		/* new score, best score so far     */
    int code;			/* optimization: return code from build_cij() */
    int ngap;			/* gap count in a column                      */
    float wgtsum;			/* sum of weights; do not assume it is nseq */

    /* Allocations
     */
    matc = (float **) MallocOrDie (sizeof (float *) * (msa->alen + 1));
    sc = (float *) MallocOrDie (sizeof (float) * (msa->alen + 2));
    tbck = (int *) MallocOrDie (sizeof (int) * (msa->alen + 2));
    matassign = (int *) MallocOrDie (sizeof (int) * (msa->alen + 1));
    insopt = (int *) MallocOrDie (sizeof (int) * msa->nseq);
    for (i = 0; i < msa->alen; i++) {
        matc[i + 1] = (float *) MallocOrDie (Alphabet_size * sizeof (float));
        FSet (matc[i + 1], Alphabet_size, 0.);
    }

    /* Precalculations
     */
    for (i = 0; i < msa->alen; i++) {
        for (idx = 0; idx < msa->nseq; idx++) {
            if (!isgap (msa->aseq[idx][i]))
                P7CountSymbol (matc[i + 1], SymbolIndex (msa->aseq[idx][i]),
                               msa->wgt[idx]);
				}
		}		
    mpri = sreLOG2 (mpri);

    FCopy (insp, prior->i[0], Alphabet_size);
    FNorm (insp, Alphabet_size);
    wgtsum = FSum (msa->wgt, msa->nseq);
    for (x = 0; x < Alphabet_size; x++) {
        insp[x] = sreLOG2 (insp[x] / null[x]);
    }

    /* Estimate the relevant special transitions.
     */
    est_M = estimate_model_length (msa);
    t_me = 0.5 / (float) (est_M - 1);
    bm1 = 0.5;
    bm2 = 0.5 / (float) (est_M - 1);
    bm1 = sreLOG2 (bm1 / null_p1);
    bm2 = sreLOG2 (bm2 / null_p1);

    /* Estimate the position of the last match-assigned column
     * by counting gap frequencies.
     */
    maxgap = 0.5;
    for (last = msa->alen; last >= 1; last--) {
        ngap = 0;
        for (idx = 0; idx < msa->nseq; idx++) { 
            if (isgap (msa->aseq[idx][last - 1])) {
                ngap++;
            }
				}		
        if ((float) ngap / (float) msa->nseq <= maxgap) {
            break;
        }
    }

    /* Initialization
     */
    sc[last] = 0.;
    tbck[last] = 0;

    /* Set ME gaps to '_'
     */
    for (idx = 0; idx < msa->nseq; idx++) {
        for (i = last; i > 0 && isgap (msa->aseq[idx][i - 1]); i--) {
            msa->aseq[idx][i - 1] = '_';
        }
		}
    /* Main recursion moves from right to left.
     */
    for (i = last - 1; i > 0; i--) {
        /* Calculate match emission scores for i  */
        FCopy (matp, matc[i], Alphabet_size);
        P7PriorifyEmissionVector (matp, prior, prior->mnum, prior->mq, prior->m,
                                  NULL);
        for (x = 0; x < Alphabet_size; x++) {
            matp[x] = sreLOG2 (matp[x] / null[x]);
        }

        /* Initialize insert counters to zero */
        FSet (insc, Alphabet_size, 0.);
        for (idx = 0; idx < msa->nseq; idx++) {
            insopt[idx] = 0;
        }

        sc[i] = -FLT_MAX;
        for (j = i + 1; j <= last; j++) {
            /* build transition matrix for column pair i,j */
            code = build_cij (msa->aseq, msa->nseq, insopt, i, j, msa->wgt, cij);
            if (code == -1) {
                break;    /* no j to our right can work for us */
            }
            if (code == 1) {
                FCopy (tij, cij, 7);
                P7PriorifyTransitionVector (tij, prior, prior->tq);
                FNorm (tij, 3);
                tij[TMM] = sreLOG2 (tij[TMM] / null_p1);
                tij[TMI] = sreLOG2 (tij[TMI] / null_p1);
                tij[TMD] = sreLOG2 (tij[TMD]);
                tij[TIM] = sreLOG2 (tij[TIM] / null_p1);
                tij[TII] = sreLOG2 (tij[TII] / null_p1);
                tij[TDM] = sreLOG2 (tij[TDM] / null_p1);
                tij[TDD] = sreLOG2 (tij[TDD]);
                /* calculate the score of using this j. */
                new = sc[j] + FDot (tij, cij, 7) + FDot (insp, insc, Alphabet_size);

                SQD_DPRINTF2 (("%3d %3d new=%6.2f scj=%6.2f m=%6.2f i=%6.2f t=%6.2f\n", i, j, new, sc[j], FDot (matp, matc[i], Alphabet_size), FDot (insp, insc, Alphabet_size), FDot (tij, cij, 7)));

                /* keep it if it's better */
                if (new > sc[i]) {
                    sc[i] = new;
                    tbck[i] = j;
                }
            }
            /* bump insc, insopt insert symbol counters */
            FAdd (insc, matc[j], Alphabet_size);
            for (idx = 0; idx < msa->nseq; idx++) {
                if (!isgap (msa->aseq[idx][j - 1])) {
                    insopt[idx]++;
                }
						}		
        }
        /* add in constant contributions for col i */
        /* note ad hoc scaling of mpri by wgtsum (us. nseq) */
        sc[i] += FDot (matp, matc[i], Alphabet_size) + mpri * wgtsum;
    }				/* end loop over start positions i */

    /* Termination: place the begin state.
     * log odds score for S->N->B is all zero except for NB transition, which
     * is a constant. So we only have to evaluate BM transitions.
     */
    bestsc = -FLT_MAX;
    first = 0;
    for (i = 1; i <= last; i++) {
        new = sc[i];
        for (idx = 0; idx < msa->nseq; idx++) {
            if (isgap (msa->aseq[idx][i - 1])) {
                new += bm2;    /* internal B->M transition */
            }
            else {
                new += bm1;    /* B->M1 transition         */
            }
        }
        if (new > bestsc) {
            bestsc = new;
            first = i;
        }
    }

    /* Traceback
     */
    matassign[0] = 0;
    for (i = 1; i <= msa->alen; i++) {
        matassign[i] = ASSIGN_INSERT;
    }
    for (i = first; i != 0; i = tbck[i]) {
        matassign[i] &= ~ASSIGN_INSERT;
        matassign[i] |= ASSIGN_MATCH;
    }

    /* Hand matassign off for remainder of model construction
     */
    /*  print_matassign(matassign, ainfo->alen); */
    matassign2hmm (msa, dsq, matassign, ret_hmm, ret_tr);

    /* Clean up.
     */
    for (i = 1; i <= msa->alen; i++) {
        free (matc[i]);
    }
    free (matc);
    free (sc);
    free (tbck);
    free (matassign);
    free (insopt);
}


/* Function: build_cij()
 *
 * Purpose:  Construct a counts vector for transitions between
 *           column i and column j in a multiple alignment.
 *
 *           '_' gap characters indicate "external" gaps which
 *           are to be dealt with by B->M and M->E transitions.
 *           These characters must be placed by a preprocessor.
 *
 *           insopt is an "insert optimization" -- an incrementor
 *           which keeps track of the number of insert symbols
 *           between i and j.
 *
 * Args:     aseqs  - multiple alignment. [0.nseq-1][0.alen-1]
 *           nseq   - number of seqs in aseqs
 *           insopt - number of inserts per seq between i/j [0.nseq-1]
 *           i      - i column [1.alen], off by one from aseqs
 *           j      - j column [1.alen], off by one from aseqs
 *           wgt    - per-seq weights [0.nseq-1]
 *           cij    - transition count vectors [0..7]
 *
 * Return:   -1 if an illegal transition was seen for this i/j assignment *and*
 *              we are guaranteed that any j to the right will also
 *              have illegal transitions.
 *           0  if an illegal transition was seen, but a j further to the
 *              right may work.
 *           1 if all transitions were legal.
 */
static int
build_cij (char **aseqs, int nseq, int *insopt, int i, int j,
           float *wgt, float *cij)
{
    int idx;			/* counter for seqs */

    i--;				/* make i,j relative to aseqs [0..alen-1] */
    j--;
    FSet (cij, 8, 0.);		/* zero cij */
    for (idx = 0; idx < nseq; idx++) {
        if (insopt[idx] > 0) {
            if (isgap (aseqs[idx][i])) {
                return -1;    /* D->I prohibited. */
            }
            if (isgap (aseqs[idx][j])) {
                return 0;    /* I->D prohibited. */
            }
            cij[TMI] += wgt[idx];
            cij[TII] += (insopt[idx] - 1) * wgt[idx];
            cij[TIM] += wgt[idx];
        }
        else {
            if (!isgap (aseqs[idx][i])) {
                if (aseqs[idx][j] == '_');	/* YO! what to do with trailer? */
                else if (isgap (aseqs[idx][j])) {
                    cij[TMD] += wgt[idx];
                }
                else {
                    cij[TMM] += wgt[idx];
                }
            }
            else {			/* ignores B->E possibility */
                if (aseqs[idx][j] == '_') {
                    continue;
                }
                else if (isgap (aseqs[idx][j])) {
                    cij[TDD] += wgt[idx];
                }
                else {
                    cij[TDM] += wgt[idx];
                }
            }
        }
    }
    return 1;
}


/* Function: estimate_model_length()
 *
 * Purpose:  Return a decent guess about the length of the model,
 *           based on the lengths of the sequences.
 *
 *           Algorithm is dumb: use weighted average length.
 *
 *           Don't assume that weights sum to nseq!
 */
static int
estimate_model_length (MSA * msa)
{
    int idx;
    float total = 0.;
    float wgtsum = 0.;

    for (idx = 0; idx < msa->nseq; idx++) {
        total += msa->wgt[idx] * DealignedLength (msa->aseq[idx]);
        wgtsum += msa->wgt[idx];
    }

    return (int) (total / wgtsum);
}


/* Function: matassign2hmm()
 *
 * Purpose:  Given an assignment of alignment columns to match vs.
 *           insert, finish the final part of the model construction
 *           calculation that is constant between model construction
 *           algorithms.
 *
 * Args:     msa       - multiple sequence alignment
 *           dsq       - digitized unaligned aseq's
 *           matassign - 1..alen bit flags for column assignments
 *           ret_hmm   - RETURN: counts-form HMM
 *           ret_tr    - RETURN: array of tracebacks for aseq's
 *
 * Return:   (void)
 *           ret_hmm and ret_tr alloc'ed here for the calling
 *           modelmaker function.
 */
static void
matassign2hmm (MSA * msa, char **dsq, int *matassign,
               struct plan7_s **ret_hmm, struct p7trace_s ***ret_tr)
{
    struct plan7_s *hmm;		/* RETURN: new hmm                     */
    struct p7trace_s **tr;	/* fake tracebacks for each seq        */
    int M;			/* length of new model in match states */
    int idx;			/* counter over sequences              */
    int apos;			/* counter for aligned columns         */

    /* how many match states in the HMM? */
    M = 0;
    for (apos = 1; apos <= msa->alen; apos++) {
        if (matassign[apos] & ASSIGN_MATCH) {
            M++;
        }
    }

    if (M == 0)
        Die ("No conserved consensus columns found; aborting construction!\n\
This is an unusual situation. Reexamine your sequence alignment. It is\n\
probably unusually full of gaps, or lots of sequence fragments. You may be\n\
able to force HMMER to model it; see the --fast (and --gapmax), or --hand\n\
options to hmmbuild.");

    /* delimit N-terminal tail */
    for (apos = 1; matassign[apos] & ASSIGN_INSERT && apos <= msa->alen; apos++) {
        matassign[apos] |= EXTERNAL_INSERT_N;
    }
    if (apos <= msa->alen) {
        matassign[apos] |= FIRST_MATCH;
    }

    /* delimit C-terminal tail */
    for (apos = msa->alen; matassign[apos] & ASSIGN_INSERT && apos > 0; apos--) {
        matassign[apos] |= EXTERNAL_INSERT_C;
    }
    if (apos > 0) {
        matassign[apos] |= LAST_MATCH;
    }

    /* print_matassign(matassign, msa->alen);  */

    /* make fake tracebacks for each seq */
    fake_tracebacks (msa->aseq, msa->nseq, msa->alen, matassign, &tr);
    /* build model from tracebacks */
    hmm = AllocPlan7 (M);
    ZeroPlan7 (hmm);
    for (idx = 0; idx < msa->nseq; idx++) {
        /* P7PrintTrace(stdout, tr[idx], NULL, NULL);   */
        P7TraceCount (hmm, dsq[idx], msa->wgt[idx], tr[idx]);
    }
    /* annotate new model */
    annotate_model (hmm, matassign, msa);

    /* Set #=RF line of alignment to reflect our assignment
     * of match, delete. matassign is valid from 1..alen and is off
     * by one from msa->rf.
     */
    if (msa->rf != NULL) {
        free (msa->rf);
    }
    msa->rf = (char *) MallocOrDie (sizeof (char) * (msa->alen + 1));
    for (apos = 0; apos < msa->alen; apos++) {
        msa->rf[apos] = matassign[apos + 1] & ASSIGN_MATCH ? 'x' : '.';
    }
    msa->rf[msa->alen] = '\0';

    /* Cleanup and return. */
    if (ret_tr != NULL) {
        *ret_tr = tr;
    }
    else {
        for (idx = 0; idx < msa->nseq; idx++) {
            P7FreeTrace (tr[idx]);
        }
        free (tr);
    }
    if (ret_hmm != NULL) {
        *ret_hmm = hmm;
    }
    else {
        FreePlan7 (hmm);
    }
    return;
}



/* Function: fake_tracebacks()
 *
 * Purpose:  From a consensus assignment of columns to MAT/INS, construct fake
 *           tracebacks for each individual sequence.
 *
 * Note:     Fragment tolerant by default. Internal entries are
 *           B->M_x, instead of B->D1->D2->...->M_x; analogously
 *           for internal exits.
 *
 * Args:     aseqs     - alignment [0..nseq-1][0..alen-1]
 *           nseq      - number of seqs in alignment
 *           alen      - length of alignment in columns
 *           matassign - assignment of column; [1..alen] (off one from aseqs)
 *           ret_tr    - RETURN: array of tracebacks
 *
 * Return:   (void)
 *           ret_tr is alloc'ed here. Caller must free.
 */
static void
fake_tracebacks (char **aseq, int nseq, int alen, int *matassign,
                 struct p7trace_s ***ret_tr)
{
    struct p7trace_s **tr;
    int idx;			/* counter over sequences          */
    int i;			/* position in raw sequence (1..L) */
    int k;			/* position in HMM                 */
    int apos;			/* position in alignment columns   */
    int tpos;			/* position in traceback           */

    tr = (struct p7trace_s **) MallocOrDie (sizeof (struct p7trace_s *) * nseq);

    for (idx = 0; idx < nseq; idx++) {
        P7AllocTrace (alen + 6, &tr[idx]);	/* allow room for S,N,B,E,C,T */

        /* all traces start with S state... */
        tr[idx]->statetype[0] = STS;
        tr[idx]->nodeidx[0] = 0;
        tr[idx]->pos[0] = 0;
        /* ...and transit to N state; N-term tail
           is emitted on N->N transitions */
        tr[idx]->statetype[1] = STN;
        tr[idx]->nodeidx[1] = 0;
        tr[idx]->pos[1] = 0;

        i = 1;
        k = 0;
        tpos = 2;
        for (apos = 0; apos < alen; apos++) {
            tr[idx]->statetype[tpos] = STBOGUS;	/* bogus, deliberately, to debug */

            if (matassign[apos + 1] & FIRST_MATCH) {	/* BEGIN */
                tr[idx]->statetype[tpos] = STB;
                tr[idx]->nodeidx[tpos] = 0;
                tr[idx]->pos[tpos] = 0;
                tpos++;
            }

            if (matassign[apos + 1] & ASSIGN_MATCH && !isgap (aseq[idx][apos])) {	/* MATCH */
                k++;			/* move to next model pos */
                tr[idx]->statetype[tpos] = STM;
                tr[idx]->nodeidx[tpos] = k;
                tr[idx]->pos[tpos] = i;
                i++;
                tpos++;
            }
            else if (matassign[apos + 1] & ASSIGN_MATCH) {	/* DELETE */
                /* being careful about S/W transitions; no B->D transitions */
                k++;			/* *always* move on model when ASSIGN_MATCH */
                if (tr[idx]->statetype[tpos - 1] != STB) {
                    tr[idx]->statetype[tpos] = STD;
                    tr[idx]->nodeidx[tpos] = k;
                    tr[idx]->pos[tpos] = 0;
                    tpos++;
                }
            }
            else if (matassign[apos + 1] & EXTERNAL_INSERT_N && !isgap (aseq[idx][apos])) {	/* N-TERMINAL TAIL */
                tr[idx]->statetype[tpos] = STN;
                tr[idx]->nodeidx[tpos] = 0;
                tr[idx]->pos[tpos] = i;
                i++;
                tpos++;
            }
            else if (matassign[apos + 1] & EXTERNAL_INSERT_C && !isgap (aseq[idx][apos])) {	/* C-TERMINAL TAIL */
                tr[idx]->statetype[tpos] = STC;
                tr[idx]->nodeidx[tpos] = 0;
                tr[idx]->pos[tpos] = i;
                i++;
                tpos++;
            }
            else if (!isgap (aseq[idx][apos])) {	/* INSERT */
                tr[idx]->statetype[tpos] = STI;
                tr[idx]->nodeidx[tpos] = k;
                tr[idx]->pos[tpos] = i;
                i++;
                tpos++;
            }

            if (matassign[apos + 1] & LAST_MATCH) {	/* END */
                /* be careful about S/W transitions; may need to roll
                 * back over some D's because there's no D->E transition
                 */
                while (tr[idx]->statetype[tpos - 1] == STD) {
                    tpos--;
                }
                tr[idx]->statetype[tpos] = STE;
                tr[idx]->nodeidx[tpos] = 0;
                tr[idx]->pos[tpos] = 0;
                tpos++;
                /* and then transit E->C;
                   alignments that use J are undefined;
                   C-term tail is emitted on C->C transitions */
                tr[idx]->statetype[tpos] = STC;
                tr[idx]->nodeidx[tpos] = 0;
                tr[idx]->pos[tpos] = 0;
                tpos++;
            }
        }
        /* all traces end with T state */
        tr[idx]->statetype[tpos] = STT;
        tr[idx]->nodeidx[tpos] = 0;
        tr[idx]->pos[tpos] = 0;
        tr[idx]->tlen = ++tpos;
        /* deal with DI, ID transitions */
        /* k == M here */
        trace_doctor (tr[idx], k, NULL, NULL);

    }				/* end for sequence # idx */

    *ret_tr = tr;
    return;
}

/* Function: trace_doctor()
 *
 * Purpose:  Plan 7 disallows D->I and I->D "chatter" transitions.
 *           However, these transitions may be implied by many
 *           alignments for hand- or heuristic- built HMMs.
 *           trace_doctor() collapses I->D or D->I into a
 *           single M position in the trace.
 *           Similarly, B->I and I->E transitions may be implied
 *           by an alignment.
 *
 *           trace_doctor does not examine any scores when it does
 *           this. In ambiguous situations (D->I->D) the symbol
 *           will be pulled arbitrarily to the left, regardless
 *           of whether that's the best column to put it in or not.
 *
 * Args:     tr      - trace to doctor
 *           M       - length of model that traces are for
 *           ret_ndi - number of DI transitions doctored
 *           ret_nid - number of ID transitions doctored
 *
 * Return:   (void)
 *           tr is modified
 */
static void
trace_doctor (struct p7trace_s *tr, int mlen, int *ret_ndi, int *ret_nid)
{
    int opos;			/* position in old trace                 */
    int npos;			/* position in new trace (<= opos)       */
    int ndi, nid;			/* number of DI, ID transitions doctored */

    /* overwrite the trace from left to right */
    ndi = nid = 0;
    opos = npos = 0;
    while (opos < tr->tlen) {
        /* fix implied D->I transitions; D transforms to M, I pulled in */
        if (tr->statetype[opos] == STD && tr->statetype[opos + 1] == STI) {
            tr->statetype[npos] = STM;
            tr->nodeidx[npos] = tr->nodeidx[opos];	/* D transforms to M      */
            tr->pos[npos] = tr->pos[opos + 1];	/* insert char moves back */
            opos += 2;
            npos += 1;
            ndi++;
        }				/* fix implied I->D transitions; D transforms to M, I is pushed in */
        else if (tr->statetype[opos] == STI && tr->statetype[opos + 1] == STD) {
            tr->statetype[npos] = STM;
            tr->nodeidx[npos] = tr->nodeidx[opos + 1];	/* D transforms to M    */
            tr->pos[npos] = tr->pos[opos];	/* insert char moves up */
            opos += 2;
            npos += 1;
            nid++;
        }				/* fix implied B->I transitions; pull I back to its M */
        else if (tr->statetype[opos] == STI && tr->statetype[opos - 1] == STB) {
            tr->statetype[npos] = STM;
            tr->nodeidx[npos] = tr->nodeidx[opos];	/* offending I transforms to M */
            tr->pos[npos] = tr->pos[opos];
            opos++;
            npos++;
        }				/* fix implied I->E transitions; push I to next M */
        else if (tr->statetype[opos] == STI && tr->statetype[opos + 1] == STE) {
            tr->statetype[npos] = STM;
            tr->nodeidx[npos] = tr->nodeidx[opos] + 1;	/* offending I transforms to M */
            tr->pos[npos] = tr->pos[opos];
            opos++;
            npos++;
        }				/* rare: N-N-B-E becomes N-B-M_1-E (swap B,N) */
        else if (tr->statetype[opos] == STB && tr->statetype[opos + 1] == STE
                 && tr->statetype[opos - 1] == STN && tr->pos[opos - 1] > 0) {
            tr->statetype[npos] = STM;
            tr->nodeidx[npos] = 1;
            tr->pos[npos] = tr->pos[opos - 1];
            tr->statetype[npos - 1] = STB;
            tr->nodeidx[npos - 1] = 0;
            tr->pos[npos - 1] = 0;
            opos++;
            npos++;
        }				/* rare: B-E-C-C-x becomes B-M_M-E-C-x (swap E,C) */
        else if (tr->statetype[opos] == STE && tr->statetype[opos - 1] == STB
                 && tr->statetype[opos + 1] == STC
                 && tr->statetype[opos + 2] == STC) {
            tr->statetype[npos] = STM;
            tr->nodeidx[npos] = mlen;
            tr->pos[npos] = tr->pos[opos + 2];
            tr->statetype[npos + 1] = STE;
            tr->nodeidx[npos + 1] = 0;
            tr->pos[npos + 1] = 0;
            tr->statetype[npos + 2] = STC;	/* first C must be a nonemitter  */
            tr->nodeidx[npos + 2] = 0;
            tr->pos[npos + 2] = 0;
            opos += 3;
            npos += 3;
        }				/* everything else is just copied */
        else {
            tr->statetype[npos] = tr->statetype[opos];
            tr->nodeidx[npos] = tr->nodeidx[opos];
            tr->pos[npos] = tr->pos[opos];
            opos++;
            npos++;
        }
    }
    tr->tlen = npos;

    if (ret_ndi != NULL) {
        *ret_ndi = ndi;
    }
    if (ret_nid != NULL) {
        *ret_nid = nid;
    }
    return;
}


/* Function: annotate_model()
 *
 * Purpose:  Add rf, cs optional annotation to a new model.
 *
 * Args:     hmm       - new model
 *           matassign - which alignment columns are MAT; [1..alen]
 *           msa       - alignment, including annotation to transfer
 *
 * Return:   (void)
 */
static void
annotate_model (struct plan7_s *hmm, int *matassign, MSA * msa)
{
    int apos;			/* position in matassign, 1.alen  */
    int k;			/* position in model, 1.M         */
    char *pri;			/* X-PRM, X-PRI, X-PRT annotation */

    /* Transfer reference coord annotation from alignment,
     * if available
     */
    if (msa->rf != NULL) {
        hmm->rf[0] = ' ';
        for (apos = k = 1; apos <= msa->alen; apos++) {
            if (matassign[apos] & ASSIGN_MATCH) {	/* ainfo is off by one from HMM */
                hmm->rf[k++] = (msa->rf[apos - 1] == ' ') ? '.' : msa->rf[apos - 1];
            }
				}		
        hmm->rf[k] = '\0';
        hmm->flags |= PLAN7_RF;
    }

    /* Transfer consensus structure annotation from alignment,
     * if available
     */
    if (msa->ss_cons != NULL) {
        hmm->cs[0] = ' ';
        for (apos = k = 1; apos <= msa->alen; apos++) {
            if (matassign[apos] & ASSIGN_MATCH)
                hmm->cs[k++] =
                    (msa->ss_cons[apos - 1] == ' ') ? '.' : msa->ss_cons[apos - 1];
				}						
        hmm->cs[k] = '\0';
        hmm->flags |= PLAN7_CS;
    }

    /* Transfer surface accessibility annotation from alignment,
     * if available
     */
    if (msa->sa_cons != NULL) {
        hmm->ca[0] = ' ';
        for (apos = k = 1; apos <= msa->alen; apos++) {
            if (matassign[apos] & ASSIGN_MATCH)
                hmm->ca[k++] =
                    (msa->sa_cons[apos - 1] == ' ') ? '.' : msa->sa_cons[apos - 1];
				}						
        hmm->ca[k] = '\0';
        hmm->flags |= PLAN7_CA;
    }

    /* Store the alignment map
     */
    for (apos = k = 1; apos <= msa->alen; apos++) {
        if (matassign[apos] & ASSIGN_MATCH) {
            hmm->map[k++] = apos;
        }
		}		
    hmm->flags |= PLAN7_MAP;

    /* Translate and transfer X-PRM annotation.
     * 0-9,[a-zA-Z] are legal; translate as 0-9,10-35 into hmm->mpri.
     * Any other char is translated as -1, and this will be interpreted
     * as a flag that means "unknown", e.g. use the normal mixture Dirichlet
     * procedure for this column.
     */
    if ((pri = MSAGetGC (msa, "X-PRM")) != NULL) {
        hmm->mpri = MallocOrDie (sizeof (int) * (hmm->M + 1));
        for (apos = k = 1; apos <= msa->alen; apos++) {
            if (matassign[apos] & ASSIGN_MATCH) {
                if (isdigit ((int) pri[apos - 1])) {
                    hmm->mpri[k] = pri[apos - 1] - '0';
                }
                else if (islower ((int) pri[apos - 1])) {
                    hmm->mpri[k] = pri[apos - 1] - 'a' + 10;
                }
                else if (isupper ((int) pri[apos - 1])) {
                    hmm->mpri[k] = pri[apos - 1] - 'A' + 10;
                }
                else {
                    hmm->mpri[k] = -1;
                }
                k++;
            }
				}		
    }
    /* And again for X-PRI annotation on insert priors:
     */
    if ((pri = MSAGetGC (msa, "X-PRI")) != NULL) {
        hmm->ipri = MallocOrDie (sizeof (int) * (hmm->M + 1));
        for (apos = k = 1; apos <= msa->alen; apos++) {
            if (matassign[apos] & ASSIGN_MATCH) {
                if (isdigit ((int) pri[apos - 1])) {
                    hmm->ipri[k] = pri[apos - 1] - '0';
                }
                else if (islower ((int) pri[apos - 1])) {
                    hmm->ipri[k] = pri[apos - 1] - 'a' + 10;
                }
                else if (isupper ((int) pri[apos - 1])) {
                    hmm->ipri[k] = pri[apos - 1] - 'A' + 10;
                }
                else {
                    hmm->ipri[k] = -1;
                }
                k++;
            }
				}		
    }
    /* And one last time for X-PRT annotation on transition priors:
     */
    if ((pri = MSAGetGC (msa, "X-PRT")) != NULL) {
        hmm->tpri = MallocOrDie (sizeof (int) * (hmm->M + 1));
        for (apos = k = 1; apos <= msa->alen; apos++) {
            if (matassign[apos] & ASSIGN_MATCH) {
                if (isdigit ((int) pri[apos - 1])) {
                    hmm->tpri[k] = pri[apos - 1] - '0';
                }
                else if (islower ((int) pri[apos - 1])) {
                    hmm->tpri[k] = pri[apos - 1] - 'a' + 10;
                }
                else if (isupper ((int) pri[apos - 1])) {
                    hmm->tpri[k] = pri[apos - 1] - 'A' + 10;
                }
                else {
                    hmm->tpri[k] = -1;
                }
                k++;
            }
				}		
    }

}

static void
print_matassign (int *matassign, int alen)
{
    int apos;

    for (apos = 0; apos <= alen; apos++) {
        printf ("%3d %c %c %c\n",
                apos,
                (matassign[apos] & ASSIGN_MATCH) ? 'x' : ' ',
                (matassign[apos] & FIRST_MATCH
                 || matassign[apos] & LAST_MATCH) ? '<' : ' ',
                (matassign[apos] & EXTERNAL_INSERT_N
                 || matassign[apos] & EXTERNAL_INSERT_C) ? '|' : ' ');
    }
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* phylip.c
 * SRE, Mon Jun 14 14:08:33 1999 [St. Louis]
 *
 * Import/export of PHYLIP interleaved multiple sequence alignment
 * format files.
 *
 * CVS $Id: phylip.c,v 1.3 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"
#include "msa.h"

#ifdef TESTDRIVE_PHYLIP
/*****************************************************************
 * phylip.c test driver:
 *
 */
int
main (int argc, char **argv)
{
    MSAFILE *afp;
    MSA *msa;
    char *file;

    file = argv[1];

    if ((afp = MSAFileOpen (file, MSAFILE_UNKNOWN, NULL)) == NULL) {
        Die ("Couldn't open %s\n", file);
    }

    printf ("format %d\n", afp->format);

    while ((msa = ReadPhylip (afp)) != NULL) {
        WritePhylip (stdout, msa);
        MSAFree (msa);
    }

    MSAFileClose (afp);
    exit (0);
}

/******************************************************************/
#endif /* testdrive_phylip */



/* Function: ReadPhylip()
 * Date:     SRE, Fri Jun 18 12:59:37 1999 [Sanger Centre]
 *
 * Purpose:  Parse an alignment from an open Phylip format
 *           alignment file. Phylip is a single-alignment format.
 *           Return the alignment, or NULL if we have no data.
 *
 * Args:     afp - open alignment file
 *
 * Returns:  MSA * - an alignment object
 *                   Caller responsible for an MSAFree()
 *           NULL if no more alignments
 */
MSA *
ReadPhylip (MSAFILE * afp)
{
    MSA *msa;
    char *s, *s1, *s2;
    char name[11];		/* seq name max len = 10 char */
    int nseq, alen;
    int idx;			/* index of current sequence */
    int slen;
    int nblock;

    if (feof (afp->f)) {
        return NULL;
    }

    /* Skip until we see a nonblank line; it's the header,
     * containing nseq/alen
     */
    nseq = 0;
    alen = 0;
    while ((s = MSAFileGetLine (afp)) != NULL) {
        if ((s1 = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
            continue;
        }
        if ((s2 = sre_strtok (&s, WHITESPACE, NULL)) == NULL)
            Die ("Failed to parse nseq/alen from first line of PHYLIP file %s\n",
                 afp->fname);
        if (!IsInt (s1) || !IsInt (s2))
            Die
            ("nseq and/or alen not an integer in first line of PHYLIP file %s\n",
             afp->fname);
        nseq = atoi (s1);
        alen = atoi (s2);
        break;
    }

    msa = MSAAlloc (nseq, 0);
    idx = 0;
    nblock = 0;
    while ((s = MSAFileGetLine (afp)) != NULL) {
        /* ignore blank lines. nonblank lines start w/ nonblank char */
        if (isspace ((int) *s)) {
            continue;
        }
        /* First block has seq names */
        if (nblock == 0) {
            strncpy (name, s, 10);
            name[10] = '\0';
            GKIStoreKey (msa->index, name);
            msa->sqname[idx] = sre_strdup (name, -1);
            s += 10;
        }
        /* be careful of trailing whitespace on lines */
        if ((s1 = sre_strtok (&s, WHITESPACE, &slen)) == NULL)
            Die ("Failed to parse sequence at line %d of PHYLIP file %s\n",
                 afp->linenumber, afp->fname);
        msa->sqlen[idx] =
            sre_strcat (&(msa->aseq[idx]), msa->sqlen[idx], s1, slen);

        idx++;
        if (idx == nseq) {
            idx = 0;
            nblock++;
        }
    }
    msa->nseq = nseq;
    MSAVerifyParse (msa);		/* verifies; sets alen, wgt; frees sqlen[] */
    return msa;
}



/* Function: WritePhylip()
 * Date:     SRE, Fri Jun 18 12:07:41 1999 [Sanger Centre]
 *
 * Purpose:  Write an alignment in Phylip format to an open file.
 *
 * Args:     fp    - file that's open for writing.
 *           msa   - alignment to write.
 *
 * Returns:  (void)
 */
void
WritePhylip (FILE * fp, MSA * msa)
{
    int idx;			/* counter for sequences         */
    int cpl = 50;			/* 50 seq char per line          */
    char buf[51];			/* buffer for writing seq        */
    int pos;

    /* First line has nseq, alen
     */
    fprintf (fp, " %d  %d\n", msa->nseq, msa->alen);

    /* Alignment section.
     * PHYLIP is a multiblock format, blocks (optionally) separated
     * by blanks; names only attached to first block. Names are
     * restricted to ten char; we achieve this by simple truncation (!).
     * (Do we need to convert gap characters from our ./- convention?)
     */
    for (pos = 0; pos < msa->alen; pos += cpl) {
        if (pos > 0) {
            fprintf (fp, "\n");
        }

        for (idx = 0; idx < msa->nseq; idx++) {
            strncpy (buf, msa->aseq[idx] + pos, cpl);
            buf[cpl] = '\0';
            if (pos > 0) {
                fprintf (fp, "%s\n", buf);
            }
            else {
                fprintf (fp, "%-10.10s%s\n", msa->sqname[idx], buf);
            }
        }
    }
    return;
}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/


/* plan7.c
 * SRE, Sat Nov 16 14:19:56 1996
 *
 * Support for Plan 7 HMM data structure, plan7_s.
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "squid.h"
#include "funcs.h"
#include "structs.h"


/* Functions: AllocPlan7(), AllocPlan7Shell(), AllocPlan7Body(), FreePlan7()
 *
 * Purpose:   Allocate or free a Plan7 HMM structure.
 *            Can either allocate all at one (AllocPlan7()) or
 *            in two steps (AllocPlan7Shell(), AllocPlan7Body()).
 *            The two step method is used in hmmio.c where we start
 *            parsing the header of an HMM file but don't
 *            see the size of the model 'til partway thru the header.
 */
struct plan7_s *
AllocPlan7 (int M)
{
    struct plan7_s *hmm;

    hmm = AllocPlan7Shell ();
    AllocPlan7Body (hmm, M);
    return hmm;
}

struct plan7_s *
AllocPlan7Shell (void)
{
    struct plan7_s *hmm;

    hmm = (struct plan7_s *) MallocOrDie (sizeof (struct plan7_s));
    hmm->M = 0;

    hmm->name = NULL;
    hmm->acc = NULL;
    hmm->desc = NULL;
    hmm->rf = NULL;
    hmm->cs = NULL;
    hmm->ca = NULL;
    hmm->comlog = NULL;
    hmm->nseq = 0;
    hmm->ctime = NULL;
    hmm->map = NULL;
    hmm->checksum = 0;

    hmm->tpri = NULL;
    hmm->mpri = NULL;
    hmm->ipri = NULL;

    hmm->ga1 = hmm->ga2 = 0.0;
    hmm->tc1 = hmm->tc2 = 0.0;
    hmm->nc1 = hmm->nc2 = 0.0;

    hmm->t = NULL;
    hmm->mat = NULL;
    hmm->ins = NULL;

    hmm->tsc = hmm->msc = hmm->isc = NULL;
    hmm->tsc_mem = hmm->msc_mem = hmm->msc_mem = NULL;

    hmm->begin = NULL;
    hmm->end = NULL;

    hmm->bsc = hmm->bsc_mem = NULL;
    hmm->esc = hmm->esc_mem = NULL;

    /* DNA translation is not enabled by default */
    hmm->dnam = NULL;
    hmm->dnai = NULL;
    hmm->dna2 = -INFTY;
    hmm->dna4 = -INFTY;
    /* statistical parameters set to innocuous empty values */
    hmm->mu = 0.;
    hmm->lambda = 0.;

    hmm->flags = 0;
    return hmm;
}

#ifndef ALTIVEC
/* in Altivec port, replaced in fast_algorithms.c */
void
AllocPlan7Body (struct plan7_s *hmm, int M)
{
    int k, x;

    hmm->M = M;

    hmm->rf = MallocOrDie ((M + 2) * sizeof (char));
    hmm->cs = MallocOrDie ((M + 2) * sizeof (char));
    hmm->ca = MallocOrDie ((M + 2) * sizeof (char));
    hmm->map = MallocOrDie ((M + 1) * sizeof (int));

    hmm->t = MallocOrDie (M * sizeof (float *));
    hmm->mat = MallocOrDie ((M + 1) * sizeof (float *));
    hmm->ins = MallocOrDie (M * sizeof (float *));
    hmm->t[0] = MallocOrDie ((7 * M) * sizeof (float));
    hmm->mat[0] = MallocOrDie ((MAXABET * (M + 1)) * sizeof (float));
    hmm->ins[0] = MallocOrDie ((MAXABET * M) * sizeof (float));

    hmm->tsc = MallocOrDie (7 * sizeof (int *));
    hmm->msc = MallocOrDie (MAXCODE * sizeof (int *));
    hmm->isc = MallocOrDie (MAXCODE * sizeof (int *));
    hmm->tsc_mem = MallocOrDie ((7 * M) * sizeof (int));
    hmm->msc_mem = MallocOrDie ((MAXCODE * (M + 1)) * sizeof (int));
    hmm->isc_mem = MallocOrDie ((MAXCODE * M) * sizeof (int));

    hmm->tsc[0] = hmm->tsc_mem;
    hmm->msc[0] = hmm->msc_mem;
    hmm->isc[0] = hmm->isc_mem;

    /* note allocation strategy for important 2D arrays -- trying
     * to keep locality as much as possible, cache efficiency etc.
     */
    for (k = 1; k <= M; k++) {
        hmm->mat[k] = hmm->mat[0] + k * MAXABET;
        if (k < M) {
            hmm->ins[k] = hmm->ins[0] + k * MAXABET;
            hmm->t[k] = hmm->t[0] + k * 7;
        }
    }
    for (x = 1; x < MAXCODE; x++) {
        hmm->msc[x] = hmm->msc[0] + x * (M + 1);
        hmm->isc[x] = hmm->isc[0] + x * M;
    }
    for (x = 0; x < 7; x++) {
        hmm->tsc[x] = hmm->tsc[0] + x * M;
    }

    /* tsc[x][0] is used as a boundary condition sometimes [Viterbi()],
     * so set to -inf always.
     */
    for (x = 0; x < 7; x++) {
        hmm->tsc[x][0] = -INFTY;
    }

    hmm->begin = MallocOrDie ((M + 1) * sizeof (float));
    hmm->end = MallocOrDie ((M + 1) * sizeof (float));

    hmm->bsc_mem = MallocOrDie ((M + 1) * sizeof (int));
    hmm->esc_mem = MallocOrDie ((M + 1) * sizeof (int));

    hmm->bsc = hmm->bsc_mem;
    hmm->esc = hmm->esc_mem;

    return;
}
#endif /* ALTIVEC */

void
FreePlan7 (struct plan7_s *hmm)
{
    if (hmm->name != NULL) {
        free (hmm->name);
    }
    if (hmm->acc != NULL) {
        free (hmm->acc);
    }
    if (hmm->desc != NULL) {
        free (hmm->desc);
    }
    if (hmm->rf != NULL) {
        free (hmm->rf);
    }
    if (hmm->cs != NULL) {
        free (hmm->cs);
    }
    if (hmm->ca != NULL) {
        free (hmm->ca);
    }
    if (hmm->comlog != NULL) {
        free (hmm->comlog);
    }
    if (hmm->ctime != NULL) {
        free (hmm->ctime);
    }
    if (hmm->map != NULL) {
        free (hmm->map);
    }
    if (hmm->tpri != NULL) {
        free (hmm->tpri);
    }
    if (hmm->mpri != NULL) {
        free (hmm->mpri);
    }
    if (hmm->ipri != NULL) {
        free (hmm->ipri);
    }
    if (hmm->bsc_mem != NULL) {
        free (hmm->bsc_mem);
    }
    if (hmm->begin != NULL) {
        free (hmm->begin);
    }
    if (hmm->esc_mem != NULL) {
        free (hmm->esc_mem);
    }
    if (hmm->end != NULL) {
        free (hmm->end);
    }
    if (hmm->msc_mem != NULL) {
        free (hmm->msc_mem);
    }
    if (hmm->isc_mem != NULL) {
        free (hmm->isc_mem);
    }
    if (hmm->tsc_mem != NULL) {
        free (hmm->tsc_mem);
    }
    if (hmm->mat != NULL) {
        free (hmm->mat[0]);
    }
    if (hmm->ins != NULL) {
        free (hmm->ins[0]);
    }
    if (hmm->t != NULL) {
        free (hmm->t[0]);
    }
    if (hmm->msc != NULL) {
        free (hmm->msc);
    }
    if (hmm->isc != NULL) {
        free (hmm->isc);
    }
    if (hmm->tsc != NULL) {
        free (hmm->tsc);
    }
    if (hmm->mat != NULL) {
        free (hmm->mat);
    }
    if (hmm->ins != NULL) {
        free (hmm->ins);
    }
    if (hmm->t != NULL) {
        free (hmm->t);
    }
    if (hmm->dnam != NULL) {
        free (hmm->dnam);
    }
    if (hmm->dnai != NULL) {
        free (hmm->dnai);
    }
    free (hmm);
}

/* Function: ZeroPlan7()
 *
 * Purpose:  Zeros the counts/probabilities fields in a model.
 *           Leaves null model untouched.
 */
void
ZeroPlan7 (struct plan7_s *hmm)
{
    int k;
    for (k = 1; k < hmm->M; k++) {
        FSet (hmm->t[k], 7, 0.);
        FSet (hmm->mat[k], Alphabet_size, 0.);
        FSet (hmm->ins[k], Alphabet_size, 0.);
    }
    FSet (hmm->mat[hmm->M], Alphabet_size, 0.);
    hmm->tbd1 = 0.;
    FSet (hmm->begin + 1, hmm->M, 0.);
    FSet (hmm->end + 1, hmm->M, 0.);
    for (k = 0; k < 4; k++) {
        FSet (hmm->xt[k], 2, 0.);
    }
    hmm->flags &= ~PLAN7_HASBITS;	/* invalidates scores */
    hmm->flags &= ~PLAN7_HASPROB;	/* invalidates probabilities */
}


/* Function: Plan7SetName()
 *
 * Purpose:  Change the name of a Plan7 HMM. Convenience function.
 *
 * Note:     Trailing whitespace and \n's are chopped.
 */
void
Plan7SetName (struct plan7_s *hmm, char *name)
{
    if (hmm->name != NULL) {
        free (hmm->name);
    }
    hmm->name = Strdup (name);
    StringChop (hmm->name);
}

/* Function: Plan7SetAccession()
 *
 * Purpose:  Change the accession number of a Plan7 HMM. Convenience function.
 *
 * Note:     Trailing whitespace and \n's are chopped.
 */
void
Plan7SetAccession (struct plan7_s *hmm, char *acc)
{
    if (hmm->acc != NULL) {
        free (hmm->acc);
    }
    hmm->acc = Strdup (acc);
    StringChop (hmm->acc);
    hmm->flags |= PLAN7_ACC;
}

/* Function: Plan7SetDescription()
 *
 * Purpose:  Change the description line of a Plan7 HMM. Convenience function.
 *
 * Note:     Trailing whitespace and \n's are chopped.
 */
void
Plan7SetDescription (struct plan7_s *hmm, char *desc)
{
    if (hmm->desc != NULL) {
        free (hmm->desc);
    }
    hmm->desc = Strdup (desc);
    StringChop (hmm->desc);
    hmm->flags |= PLAN7_DESC;
}

/* Function: Plan7ComlogAppend()
 * Date:     SRE, Wed Oct 29 09:57:30 1997 [TWA 721 over Greenland]
 *
 * Purpose:  Concatenate command line options and append to the
 *           command line log.
 */
void
Plan7ComlogAppend (struct plan7_s *hmm, int argc, char **argv)
{
    int len;
    int i;

    /* figure out length of command line, w/ spaces and \n */
    len = argc;
    for (i = 0; i < argc; i++) {
        len += strlen (argv[i]);
    }

    /* allocate */
    if (hmm->comlog != NULL) {
        len += strlen (hmm->comlog);
        hmm->comlog = ReallocOrDie (hmm->comlog, sizeof (char) * (len + 1));
    }
    else {
        hmm->comlog = MallocOrDie (sizeof (char) * (len + 1));
        *(hmm->comlog) = '\0';	/* need this to make strcat work */
    }

    /* append */
    strcat (hmm->comlog, "\n");
    for (i = 0; i < argc; i++) {
        strcat (hmm->comlog, argv[i]);
        if (i < argc - 1) {
            strcat (hmm->comlog, " ");
        }
    }
}

/* Function: Plan7SetCtime()
 * Date:     SRE, Wed Oct 29 11:53:19 1997 [TWA 721 over the Atlantic]
 *
 * Purpose:  Set the ctime field in a new HMM to the current time.
 */
void
Plan7SetCtime (struct plan7_s *hmm)
{
    time_t date = time (NULL);
    if (hmm->ctime != NULL) {
        free (hmm->ctime);
    }
    hmm->ctime = Strdup (ctime (&date));
    StringChop (hmm->ctime);
}


/* Function: Plan7SetNullModel()
 *
 * Purpose:  Set the null model section of an HMM.
 *           Convenience function.
 */
void
Plan7SetNullModel (struct plan7_s *hmm, float null[MAXABET], float p1)
{
    int x;
    for (x = 0; x < Alphabet_size; x++) {
        hmm->null[x] = null[x];
    }
    hmm->p1 = p1;
}


/* Function: P7Logoddsify()
 *
 * Purpose:  Take an HMM with valid probabilities, and
 *           fill in the integer log-odds score section of the model.
 *
 *    Notes on log-odds scores:
 *         type of parameter       probability        score
 *         -----------------       -----------        ------
 *         any emission             p_x           log_2 p_x/null_x
 *             N,J,C /assume/ p_x = null_x so /always/ score zero.
 *         transition to emitters   t_x           log_2 t_x/p1
 *            (M,I; N,C; J)
 *             NN and CC loops are often equal to p1, so usu. score zero.
 *         C->T transition          t_x            log_2 t_x/p2
 *            often zero, usu. C->T = p2.
 *         all other transitions    t_x           log_2 t_x
 *             (no null model counterpart, so null prob is 1)
 *
 *    Notes on entry/exit scores, B->M and M->E:
 *         The probability form model includes delete states 1 and M.
 *         these states are removed from a search form model to
 *         prevent B->D...D->E->J->B mute cycles, which would complicate
 *         dynamic programming algorithms. The data-independent
 *         S/W B->M and M->E transitions are folded together with
 *         data-dependent B->D...D->M and M->D...D->E paths.
 *
 *         This process is referred to in the code as "wing folding"
 *         or "wing retraction"... the analogy is to a swept-wing
 *         fighter in landing vs. high speed flight configuration.
 *
 *    Note on Viterbi vs. forward flag:
 *         Wing retraction must take forward vs. Viterbi
 *         into account. If forward, sum two paths; if Viterbi, take
 *         max. I tried to slide this by as a sum, without
 *         the flag, but Alex detected it as a bug, because you can
 *         then find cases where the Viterbi score doesn't match
 *         the P7TraceScore().
 *
 * Args:      hmm          - the hmm to calculate scores in.
 *            viterbi_mode - TRUE to fold wings in Viterbi configuration.
 *
 * Return:    (void)
 *            hmm scores are filled in.
 */
void
P7Logoddsify (struct plan7_s *hmm, int viterbi_mode)
{
    int k;			/* counter for model position */
    int x;			/* counter for symbols        */
    float accum;
    float tbm, tme;

    if (hmm->flags & PLAN7_HASBITS) {
        return;
    }

    /* Symbol emission scores
     */
    for (k = 1; k <= hmm->M; k++) {
        /* match/insert emissions in main model */
        for (x = 0; x < Alphabet_size; x++) {
            hmm->msc[x][k] = Prob2Score (hmm->mat[k][x], hmm->null[x]);
            if (k < hmm->M) {
                hmm->isc[x][k] = Prob2Score (hmm->ins[k][x], hmm->null[x]);
            }
        }
        /* degenerate match/insert emissions */
        for (x = Alphabet_size; x < Alphabet_iupac; x++) {
            hmm->msc[x][k] = DegenerateSymbolScore (hmm->mat[k], hmm->null, x);
            if (k < hmm->M) {
                hmm->isc[x][k] = DegenerateSymbolScore (hmm->ins[k], hmm->null, x);
            }
        }
    }

    /* State transitions.
     *
     * A note on "folding" of D_1 and D_M.
     * These two delete states are folded out of search form models
     * in order to prevent null cycles in the dynamic programming
     * algorithms (see code below). However, we use their log transitions
     * when we save the model! So the following log transition probs
     * are used *only* in save files, *never* in search algorithms:
     *    log (tbd1), D1 -> M2, D1 -> D2
     *    Mm-1 -> Dm, Dm-1 -> Dm
     *
     * In a search algorithm, these have to be interpreted as -INFTY
     * because their contributions are folded into bsc[] and esc[]
     * entry/exit scores. They can't be set to -INFTY here because
     * we need them in save files.
     */
    for (k = 1; k < hmm->M; k++) {
        hmm->tsc[TMM][k] = Prob2Score (hmm->t[k][TMM], hmm->p1);
        hmm->tsc[TMI][k] = Prob2Score (hmm->t[k][TMI], hmm->p1);
        hmm->tsc[TMD][k] = Prob2Score (hmm->t[k][TMD], 1.0);
        hmm->tsc[TIM][k] = Prob2Score (hmm->t[k][TIM], hmm->p1);
        hmm->tsc[TII][k] = Prob2Score (hmm->t[k][TII], hmm->p1);
        hmm->tsc[TDM][k] = Prob2Score (hmm->t[k][TDM], hmm->p1);
        hmm->tsc[TDD][k] = Prob2Score (hmm->t[k][TDD], 1.0);
    }

    /* B->M entry transitions. Note how D_1 is folded out.
     * M1 is just B->M1
     * M2 is sum (or max) of B->M2 and B->D1->M2
     * M_k is sum (or max) of B->M_k and B->D1...D_k-1->M_k
     * These have to be done in log space, else you'll get
     * underflow errors; and we also have to watch for log(0).
     * A little sloppier than it probably has to be; historically,
     * doing in this in log space was in response to a bug report.
     */
    accum = hmm->tbd1 > 0.0 ? log (hmm->tbd1) : -9999.;
    for (k = 1; k <= hmm->M; k++) {
        tbm = hmm->begin[k] > 0. ? log (hmm->begin[k]) : -9999.;	/* B->M_k part */

        /* B->D1...D_k-1->M_k part we get from accum */
        if (k > 1 && accum > -9999.) {
            if (hmm->t[k - 1][TDM] > 0.0) {
                if (viterbi_mode) {
                    tbm = MAX (tbm, accum + log (hmm->t[k - 1][TDM]));
                }
                else {
                    tbm = LogSum (tbm, accum + log (hmm->t[k - 1][TDM]));
                }
            }

            accum =
                (hmm->t[k - 1][TDD] >
                 0.0) ? accum + log (hmm->t[k - 1][TDD]) : -9999.;
        }
        /* Convert from log_e to scaled integer log_2 odds. */
        if (tbm > -9999.)
            hmm->bsc[k] =
                (int) floor (0.5 + INTSCALE * 1.44269504 * (tbm - log (hmm->p1)));
        else {
            hmm->bsc[k] = -INFTY;
        }
    }

    /* M->E exit transitions. Note how D_M is folded out.
     * M_M is 1 by definition
     * M_M-1 is sum of M_M-1->E and M_M-1->D_M->E, where D_M->E is 1 by definition
     * M_k is sum of M_k->E and M_k->D_k+1...D_M->E
     * Must be done in log space to avoid underflow errors.
     * A little sloppier than it probably has to be; historically,
     * doing in this in log space was in response to a bug report.
     */
    hmm->esc[hmm->M] = 0;
    accum = 0.;
    for (k = hmm->M - 1; k >= 1; k--) {
        tme = hmm->end[k] > 0. ? log (hmm->end[k]) : -9999.;
        if (accum > -9999.) {
            if (hmm->t[k][TMD] > 0.0) {
                if (viterbi_mode) {
                    tme = MAX (tme, accum + log (hmm->t[k][TMD]));
                }
                else {
                    tme = LogSum (tme, accum + log (hmm->t[k][TMD]));
                }
            }
            accum = (hmm->t[k][TDD] > 0.0) ? accum + log (hmm->t[k][TDD]) : -9999.;
        }
        /* convert from log_e to scaled integer log odds. */
        hmm->esc[k] =
            (tme >
             -9999.) ? (int) floor (0.5 + INTSCALE * 1.44269504 * tme) : -INFTY;
    }

    /* special transitions */
    hmm->xsc[XTN][LOOP] = Prob2Score (hmm->xt[XTN][LOOP], hmm->p1);
    hmm->xsc[XTN][MOVE] = Prob2Score (hmm->xt[XTN][MOVE], 1.0);
    hmm->xsc[XTE][LOOP] = Prob2Score (hmm->xt[XTE][LOOP], 1.0);
    hmm->xsc[XTE][MOVE] = Prob2Score (hmm->xt[XTE][MOVE], 1.0);
    hmm->xsc[XTC][LOOP] = Prob2Score (hmm->xt[XTC][LOOP], hmm->p1);
    hmm->xsc[XTC][MOVE] = Prob2Score (hmm->xt[XTC][MOVE], 1. - hmm->p1);
    hmm->xsc[XTJ][LOOP] = Prob2Score (hmm->xt[XTJ][LOOP], hmm->p1);
    hmm->xsc[XTJ][MOVE] = Prob2Score (hmm->xt[XTJ][MOVE], 1.0);

    hmm->flags |= PLAN7_HASBITS;	/* raise the log-odds ready flag */
}



/* Function: Plan7Renormalize()
 *
 * Purpose:  Take an HMM in counts form, and renormalize
 *           all of its probability vectors. Also enforces
 *           Plan7 restrictions on nonexistent transitions.
 *
 * Args:     hmm - the model to renormalize.
 *
 * Return:   (void)
 *           hmm is changed.
 */
void
Plan7Renormalize (struct plan7_s *hmm)
{
    int k;			/* counter for model position */
    int st;			/* counter for special states */
    float d;			/* denominator */

    /* match emissions */
    for (k = 1; k <= hmm->M; k++) {
        FNorm (hmm->mat[k], Alphabet_size);
    }
    /* insert emissions */
    for (k = 1; k < hmm->M; k++) {
        FNorm (hmm->ins[k], Alphabet_size);
    }
    /* begin transitions */
    d = FSum (hmm->begin + 1, hmm->M) + hmm->tbd1;
    FScale (hmm->begin + 1, hmm->M, 1. / d);
    hmm->tbd1 /= d;
    /* main model transitions */
    for (k = 1; k < hmm->M; k++) {
        d = FSum (hmm->t[k], 3) + hmm->end[k];
        FScale (hmm->t[k], 3, 1. / d);
        hmm->end[k] /= d;

        FNorm (hmm->t[k] + 3, 2);	/* insert */
        FNorm (hmm->t[k] + 5, 2);	/* delete */
    }
    /* null model emissions */
    FNorm (hmm->null, Alphabet_size);
    /* special transitions  */
    for (st = 0; st < 4; st++) {
        FNorm (hmm->xt[st], 2);
    }
    /* enforce nonexistent transitions */
    /* (is this necessary?) */
    hmm->t[0][TDM] = hmm->t[0][TDD] = 0.0;

    hmm->flags &= ~PLAN7_HASBITS;	/* clear the log-odds ready flag */
    hmm->flags |= PLAN7_HASPROB;	/* set the probabilities OK flag */
}


/* Function: Plan7RenormalizeExits()
 * Date:     SRE, Fri Aug 14 11:22:19 1998 [St. Louis]
 *
 * Purpose:  Renormalize just the match state transitions;
 *           for instance, after a Config() function has
 *           modified the exit distribution.
 *
 * Args:     hmm - hmm to renormalize
 *
 * Returns:  void
 */
void
Plan7RenormalizeExits (struct plan7_s *hmm)
{
    int k;
    float d;

    for (k = 1; k < hmm->M; k++) {
        d = FSum (hmm->t[k], 3);
        FScale (hmm->t[k], 3, 1. / (d + d * hmm->end[k]));
    }
}


/*****************************************************************
 * Plan7 configuration functions
 * The following few functions are the Plan7 equivalent of choosing
 * different alignment styles (fully local, fully global, global/local,
 * multihit, etc.)
 *
 * There is (at least) one constraint worth noting.
 * If you want per-domain scores to sum up to per-sequence scores,
 * then one of the following two sets of conditions must be met:
 *
 *   1) t(E->J) = 0
 *      e.g. no multidomain hits
 *
 *   2) t(N->N) = t(C->C) = t(J->J) = hmm->p1
 *      e.g. unmatching sequence scores zero, and
 *      N->B first-model score is equal to J->B another-model score.
 *
 * These constraints are obeyed in the default Config() functions below,
 * but in the future (when HMM editing may be allowed) we'll have
 * to remember this. Non-equality of the summed domain scores and
 * the total sequence score is a really easy "red flag" for people to
 * notice and report as a bug, even if it may make probabilistic
 * sense not to meet either constraint for certain modeling problems.
 *****************************************************************
 */

/* Function: Plan7NakedConfig()
 *
 * Purpose:  Set the alignment-independent, algorithm-dependent parameters
 *           of a Plan7 model so that no special states (N,C,J) emit anything:
 *           one simple, full global pass through the model.
 *
 * Args:     hmm - the plan7 model
 *
 * Return:   (void)
 *           The HMM is modified; algorithm dependent parameters are set.
 *           Previous scores are invalidated if they existed.
 */
void
Plan7NakedConfig (struct plan7_s *hmm)
{
    hmm->xt[XTN][MOVE] = 1.;	/* disallow N-terminal tail */
    hmm->xt[XTN][LOOP] = 0.;
    hmm->xt[XTE][MOVE] = 1.;	/* only 1 domain/sequence ("global" alignment) */
    hmm->xt[XTE][LOOP] = 0.;
    hmm->xt[XTC][MOVE] = 1.;	/* disallow C-terminal tail */
    hmm->xt[XTC][LOOP] = 0.;
    hmm->xt[XTJ][MOVE] = 0.;	/* J state unused */
    hmm->xt[XTJ][LOOP] = 1.;
    FSet (hmm->begin + 2, hmm->M - 1, 0.);	/* disallow internal entries. */
    hmm->begin[1] = 1. - hmm->tbd1;
    FSet (hmm->end + 1, hmm->M - 1, 0.);	/* disallow internal exits. */
    hmm->end[hmm->M] = 1.;
    Plan7RenormalizeExits (hmm);
    hmm->flags &= ~PLAN7_HASBITS;	/* reconfig invalidates log-odds scores */
}

/* Function: Plan7GlobalConfig()
 *
 * Purpose:  Set the alignment-independent, algorithm-dependent parameters
 *           of a Plan7 model to global (Needleman/Wunsch) configuration.
 *
 *           Like a non-looping hmmls, since we actually allow flanking
 *           N and C terminal sequence.
 *
 * Args:     hmm - the plan7 model
 *
 * Return:   (void)
 *           The HMM is modified; algorithm dependent parameters are set.
 *           Previous scores are invalidated if they existed.
 */
void
Plan7GlobalConfig (struct plan7_s *hmm)
{
    hmm->xt[XTN][MOVE] = 1. - hmm->p1;	/* allow N-terminal tail */
    hmm->xt[XTN][LOOP] = hmm->p1;
    hmm->xt[XTE][MOVE] = 1.;	/* only 1 domain/sequence ("global" alignment) */
    hmm->xt[XTE][LOOP] = 0.;
    hmm->xt[XTC][MOVE] = 1. - hmm->p1;	/* allow C-terminal tail */
    hmm->xt[XTC][LOOP] = hmm->p1;
    hmm->xt[XTJ][MOVE] = 0.;	/* J state unused */
    hmm->xt[XTJ][LOOP] = 1.;
    FSet (hmm->begin + 2, hmm->M - 1, 0.);	/* disallow internal entries. */
    hmm->begin[1] = 1. - hmm->tbd1;
    FSet (hmm->end + 1, hmm->M - 1, 0.);	/* disallow internal exits. */
    hmm->end[hmm->M] = 1.;
    Plan7RenormalizeExits (hmm);
    hmm->flags &= ~PLAN7_HASBITS;	/* reconfig invalidates log-odds scores */
}

/* Function: Plan7LSConfig()
 *
 * Purpose:  Set the alignment independent parameters of a Plan7 model
 *           to hmmls (global in HMM, local in sequence) configuration.
 *
 * Args:     hmm  - the plan7 model
 *
 * Return:   (void);
 *           the HMM probabilities are modified.
 */
void
Plan7LSConfig (struct plan7_s *hmm)
{
    hmm->xt[XTN][MOVE] = 1. - hmm->p1;	/* allow N-terminal tail */
    hmm->xt[XTN][LOOP] = hmm->p1;
    hmm->xt[XTE][MOVE] = 0.5;	/* expectation 2 domains/seq */
    hmm->xt[XTE][LOOP] = 0.5;
    hmm->xt[XTC][MOVE] = 1. - hmm->p1;	/* allow C-terminal tail */
    hmm->xt[XTC][LOOP] = hmm->p1;
    hmm->xt[XTJ][MOVE] = 1. - hmm->p1;	/* allow J junction state */
    hmm->xt[XTJ][LOOP] = hmm->p1;
    FSet (hmm->begin + 2, hmm->M - 1, 0.);	/* start at M1/D1 */
    hmm->begin[1] = 1. - hmm->tbd1;
    FSet (hmm->end + 1, hmm->M - 1, 0.);	/* end at M_m/D_m */
    hmm->end[hmm->M] = 1.;
    Plan7RenormalizeExits (hmm);
    hmm->flags &= ~PLAN7_HASBITS;	/* reconfig invalidates log-odds scores */
}


/* Function: Plan7SWConfig()
 *
 * Purpose:  Set the alignment independent parameters of
 *           a Plan7 model to hmmsw (Smith/Waterman) configuration.
 *
 * Notes:    entry/exit is asymmetric because of the left/right
 *           nature of the HMM/profile. Entry probability is distributed
 *           simply by assigning p_x = pentry / (M-1) to M-1
 *           internal match states. However, the same approach doesn't
 *           lead to a flat distribution over exit points. Exit p's
 *           must be corrected for the probability of a previous exit
 *           from the model. Requiring a flat distribution over exit
 *           points leads to an easily solved piece of algebra, giving:
 *                      p_1 = pexit / (M-1)
 *                      p_x = p_1 / (1 - (x-1) p_1)
 *
 * Args:     hmm    - the Plan7 model w/ data-dep prob's valid
 *           pentry - probability of an internal entry somewhere;
 *                    will be evenly distributed over M-1 match states
 *           pexit  - probability of an internal exit somewhere;
 *                    will be distributed over M-1 match states.
 *
 * Return:   (void)
 *           HMM probabilities are modified.
 */
void
Plan7SWConfig (struct plan7_s *hmm, float pentry, float pexit)
{
    float basep;			/* p1 for exits: the base p */
    int k;			/* counter over states      */

    /* Configure special states.
     */
    hmm->xt[XTN][MOVE] = 1 - hmm->p1;	/* allow N-terminal tail */
    hmm->xt[XTN][LOOP] = hmm->p1;
    hmm->xt[XTE][MOVE] = 1.;	/* disallow jump state   */
    hmm->xt[XTE][LOOP] = 0.;
    hmm->xt[XTC][MOVE] = 1 - hmm->p1;	/* allow C-terminal tail */
    hmm->xt[XTC][LOOP] = hmm->p1;
    hmm->xt[XTJ][MOVE] = 1.;	/* J is unused */
    hmm->xt[XTJ][LOOP] = 0.;

    /* Configure entry.
     */
    hmm->begin[1] = (1. - pentry) * (1. - hmm->tbd1);
    FSet (hmm->begin + 2, hmm->M - 1,
          (pentry * (1. - hmm->tbd1)) / (float) (hmm->M - 1));

    /* Configure exit.
     */
    hmm->end[hmm->M] = 1.0;
    basep = pexit / (float) (hmm->M - 1);
    for (k = 1; k < hmm->M; k++) {
        hmm->end[k] = basep / (1. - basep * (float) (k - 1));
    }
    Plan7RenormalizeExits (hmm);
    hmm->flags &= ~PLAN7_HASBITS;	/* reconfig invalidates log-odds scores */
}

/* Function: Plan7FSConfig()
 * Date:     SRE, Fri Jan  2 15:34:40 1998 [StL]
 *
 * Purpose:  Set the alignment independent parameters of
 *           a Plan7 model to hmmfs (multihit Smith/Waterman) configuration.
 *
 *           See comments on Plan7SWConfig() for explanation of
 *           how pentry and pexit are used.
 *
 * Args:     hmm    - the Plan7 model w/ data-dep prob's valid
 *           pentry - probability of an internal entry somewhere;
 *                    will be evenly distributed over M-1 match states
 *           pexit  - probability of an internal exit somewhere;
 *                    will be distributed over M-1 match states.
 *
 * Return:   (void)
 *           HMM probabilities are modified.
 */
void
Plan7FSConfig (struct plan7_s *hmm, float pentry, float pexit)
{
    float basep;			/* p1 for exits: the base p */
    int k;			/* counter over states      */

    /* Configure special states.
     */
    hmm->xt[XTN][MOVE] = 1 - hmm->p1;	/* allow N-terminal tail     */
    hmm->xt[XTN][LOOP] = hmm->p1;
    hmm->xt[XTE][MOVE] = 0.5;	/* allow loops / multihits   */
    hmm->xt[XTE][LOOP] = 0.5;
    hmm->xt[XTC][MOVE] = 1 - hmm->p1;	/* allow C-terminal tail     */
    hmm->xt[XTC][LOOP] = hmm->p1;
    hmm->xt[XTJ][MOVE] = 1. - hmm->p1;	/* allow J junction between domains */
    hmm->xt[XTJ][LOOP] = hmm->p1;

    /* Configure entry.
     */
    hmm->begin[1] = (1. - pentry) * (1. - hmm->tbd1);
    FSet (hmm->begin + 2, hmm->M - 1,
          (pentry * (1. - hmm->tbd1)) / (float) (hmm->M - 1));

    /* Configure exit.
     */
    hmm->end[hmm->M] = 1.0;
    basep = pexit / (float) (hmm->M - 1);
    for (k = 1; k < hmm->M; k++) {
        hmm->end[k] = basep / (1. - basep * (float) (k - 1));
    }
    Plan7RenormalizeExits (hmm);
    hmm->flags &= ~PLAN7_HASBITS;	/* reconfig invalidates log-odds scores */
}




/* Function: Plan7ESTConfig()
 *
 * Purpose:  Configure a Plan7 model for EST Smith/Waterman
 *           analysis.
 *
 *           OUTDATED; DO NOT USE WITHOUT RECHECKING
 *
 * Args:     hmm        - hmm to configure.
 *           aacode     - 0..63 vector mapping genetic code to amino acids
 *           estmodel   - 20x64 translation matrix, w/ codon bias and substitution error
 *           dna2       - probability of a -1 frameshift in a triplet
 *           dna4       - probability of a +1 frameshift in a triplet
 */
void
Plan7ESTConfig (struct plan7_s *hmm, int *aacode, float **estmodel,
                float dna2, float dna4)
{
    int k;
    int x;
    float p;
    /* fixed kaivalya  8/1/8/03 UNFINISHED!!! */
    /*   float *tripnull;	*/
    float *tripnull = NULL;	/* Kaivalya 8/25/03 * UNFINISHED!!! */

    /* configure specials */
    hmm->xt[XTN][MOVE] = 1. / 351.;
    hmm->xt[XTN][LOOP] = 350. / 351.;
    hmm->xt[XTE][MOVE] = 1.;
    hmm->xt[XTE][LOOP] = 0.;
    hmm->xt[XTC][MOVE] = 1. / 351.;
    hmm->xt[XTC][LOOP] = 350. / 351.;
    hmm->xt[XTJ][MOVE] = 1.;
    hmm->xt[XTJ][LOOP] = 0.;
    /* configure entry/exit */
    hmm->begin[1] = 0.5;
    FSet (hmm->begin + 2, hmm->M - 1, 0.5 / ((float) hmm->M - 1.));
    hmm->end[hmm->M] = 1.;
    FSet (hmm->end, hmm->M - 1, 0.5 / ((float) hmm->M - 1.));

    /* configure dna triplet/frameshift emissions */
    for (k = 1; k <= hmm->M; k++) {
        /* translate aa to triplet probabilities */
        for (x = 0; x < 64; x++) {
            p =
                hmm->mat[k][aacode[x]] * estmodel[aacode[x]][x] * (1. - dna2 - dna4);
            hmm->dnam[x][k] = Prob2Score (p, tripnull[x]);

            p =
                hmm->ins[k][aacode[x]] * estmodel[aacode[x]][x] * (1. - dna2 - dna4);
            hmm->dnai[x][k] = Prob2Score (p, tripnull[x]);
        }
        hmm->dnam[64][k] = 0;	/* ambiguous codons score 0 (danger?) */
        hmm->dna2 = Prob2Score (dna2, 1.);
        hmm->dna4 = Prob2Score (dna4, 1.);
    }
}

/* Function: PrintPlan7Stats()
 *
 * Purpose:  Given a newly constructed HMM and the tracebacks
 *           of the sequences it was trained on, print out all
 *           the interesting information at the end of hmmb
 *           and hmmt runs that convinces the user we actually
 *           did something.
 *
 * Args:     fp   - where to send the output (stdout, usually)
 *           hmm  - the new HMM, probability form
 *           dsq  - digitized training seqs
 *           nseq - number of dsq's
 *           tr   - array of tracebacks for dsq
 *
 * Return:   (void)
 */
void
PrintPlan7Stats (FILE * fp, struct plan7_s *hmm, char **dsq, int nseq,
                 struct p7trace_s **tr)
{
    int idx;			/* counter for sequences                */
    float score;			/* an individual trace score            */
    float total, best, worst;	/* for the avg. and range of the scores */
    float sqsum, stddev;		/* for the std. deviation of the scores */

    P7Logoddsify (hmm, TRUE);	/* make sure model scores are ready */

    /* find individual trace scores */
    score = P7TraceScore (hmm, dsq[0], tr[0]);
    total = best = worst = score;
    sqsum = score * score;
    for (idx = 1; idx < nseq; idx++) {
        /* P7PrintTrace(stdout, tr[idx], hmm, dsq[idx]); */
        score = P7TraceScore (hmm, dsq[idx], tr[idx]);
        total += score;
        sqsum += score * score;
        if (score > best) {
            best = score;
        }
        if (score < worst) {
            worst = score;
        }
    }
    if (nseq > 1) {
        stddev = (sqsum - (total * total / (float) nseq)) / ((float) nseq - 1.);
        stddev = (stddev > 0) ? sqrt (stddev) : 0.0;
    }
    else {
        stddev = 0.0;
    }
    /* print out stuff. */
    fprintf (fp, "Average score:  %10.2f bits\n", total / (float) nseq);
    fprintf (fp, "Minimum score:  %10.2f bits\n", worst);
    fprintf (fp, "Maximum score:  %10.2f bits\n", best);
    fprintf (fp, "Std. deviation: %10.2f bits\n", stddev);
}

/* Function: DegenerateSymbolScore()
 *
 * Purpose:  Given a sequence character x and an hmm emission probability
 *           vector, calculate the log-odds (base 2) score of
 *           the symbol.
 *
 *           Easy if x is in the emission alphabet, but not so easy
 *           is x is a degenerate symbol. The "correct" Bayesian
 *           philosophy is to calculate score(X) by summing over
 *           p(x) for all x in the degenerate symbol X to get P(X),
 *           doing the same sum over the prior to get F(X), and
 *           doing log_2 (P(X)/F(X)). This gives an X a zero score,
 *           for instance.
 *
 *           Though this is correct in a formal Bayesian sense --
 *           we have no information on the sequence, so we can't
 *           say if it's random or model, so it scores zero --
 *           it sucks, big time, for scoring biological sequences.
 *           Sequences with lots of X's score near zero, while
 *           real sequences have average scores that are negative --
 *           so the X-laden sequences appear to be lifted out
 *           of the noise of a full histogram of a database search.
 *           Correct or not, this is highly undesirable.
 *
 *           So therefore we calculated the expected score of
 *           the degenerate symbol by summing over all x in X:
 *                 e_x log_2 (p(x)/f(x))
 *           where the expectation of x, e_x, is calculated from
 *           the random model.
 *
 *           Empirically, this works; it also has a wooly hand-waving
 *           probabilistic justification that I'm happy enough about.
 *
 * Args:     p      - probabilities of normal symbols
 *           null   - null emission model
 *           ambig  - index of the degenerate character in Alphabet[]
 *
 * Return:   the integer log odds score of x given the emission
 *           vector and the null model, scaled up by INTSCALE.
 */
int
DegenerateSymbolScore (float *p, float *null, int ambig)
{
    int x;
    float numer = 0.;
    float denom = 0.;

    for (x = 0; x < Alphabet_size; x++) {
        if (Degenerate[ambig][x]) {
            numer += null[x] * sreLOG2 (p[x] / null[x]);
            denom += null[x];
        }
    }
    return (int) (INTSCALE * numer / denom);
}

/*****************************************************************
 *
 * Plan9/Plan7 interface
 *
 * Very important code during the evolutionary takeover by Plan7 --
 * convert between Krogh/Haussler and Plan7 models.
 *****************************************************************/

/* Function: Plan9toPlan7()
 *
 * Purpose:  Convert an old HMM into Plan7. Configures it in
 *           ls mode.
 *
 * Args:     hmm       - old ugly plan9 style HMM
 *           ret_plan7 - new wonderful Plan7 HMM
 *
 * Return:   (void)
 *           Plan7 HMM is allocated here. Free w/ FreePlan7().
 */
void
Plan9toPlan7 (struct plan9_s *hmm, struct plan7_s **ret_plan7)
{
    struct plan7_s *plan7;
    int k, x;

    plan7 = AllocPlan7 (hmm->M);

    for (k = 1; k < hmm->M; k++) {
        plan7->t[k][TMM] = hmm->mat[k].t[MATCH];
        plan7->t[k][TMD] = hmm->mat[k].t[DELETE];
        plan7->t[k][TMI] = hmm->mat[k].t[INSERT];
        plan7->t[k][TDM] = hmm->del[k].t[MATCH];
        plan7->t[k][TDD] = hmm->del[k].t[DELETE];
        plan7->t[k][TIM] = hmm->ins[k].t[MATCH];
        plan7->t[k][TII] = hmm->ins[k].t[INSERT];
    }

    for (k = 1; k <= hmm->M; k++) {
        for (x = 0; x < Alphabet_size; x++) {
            plan7->mat[k][x] = hmm->mat[k].p[x];
        }
		}
    for (k = 1; k < hmm->M; k++) {
        for (x = 0; x < Alphabet_size; x++) {
            plan7->ins[k][x] = hmm->ins[k].p[x];
        }
		}
    plan7->tbd1 =
        hmm->mat[0].t[DELETE] / (hmm->mat[0].t[DELETE] + hmm->mat[0].t[MATCH]);

    /* We have to make up the null transition p1; use default */
    P7DefaultNullModel (plan7->null, &(plan7->p1));
    for (x = 0; x < Alphabet_size; x++) {
        plan7->null[x] = hmm->null[x];
    }

    if (hmm->name != NULL) {
        Plan7SetName (plan7, hmm->name);
    }
    if (hmm->flags & HMM_REF) {
        strcpy (plan7->rf, hmm->ref);
        plan7->flags |= PLAN7_RF;
    }
    if (hmm->flags & HMM_CS) {
        strcpy (plan7->cs, hmm->cs);
        plan7->flags |= PLAN7_CS;
    }

    Plan7LSConfig (plan7);	/* configure specials for ls-style alignment */
    Plan7Renormalize (plan7);	/* mainly to correct for missing ID and DI */
    plan7->flags |= PLAN7_HASPROB;	/* probabilities are valid */
    plan7->flags &= ~PLAN7_HASBITS;	/* scores are not valid    */
    *ret_plan7 = plan7;
}


/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* plan9.c
 * SRE, Wed Apr  8 07:35:30 1998
 *
 * alloc, free, and initialization of old Plan9 (HMMER 1.x) functions.
 * Rescued from the wreckage of HMMER 1.9m code.
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "squid.h"
#include "structs.h"
#include "funcs.h"

struct plan9_s *
P9AllocHMM (int M)
{   /* length of model to make */
    struct plan9_s *hmm;		/* RETURN: blank HMM */

    hmm = (struct plan9_s *) MallocOrDie (sizeof (struct plan9_s));
    hmm->ins =
        (struct basic_state *) MallocOrDie (sizeof (struct basic_state) *
                                            (M + 2));
    hmm->del =
        (struct basic_state *) MallocOrDie (sizeof (struct basic_state) *
                                            (M + 2));
    hmm->mat =
        (struct basic_state *) MallocOrDie (sizeof (struct basic_state) *
                                            (M + 2));
    hmm->ref = (char *) MallocOrDie ((M + 2) * sizeof (char));
    hmm->cs = (char *) MallocOrDie ((M + 2) * sizeof (char));
    hmm->xray = (float *) MallocOrDie ((M + 2) * sizeof (float) * NINPUTS);
    hmm->M = M;
    hmm->name = Strdup ("unnamed");	/* name is not optional. */

    hmm->flags = 0;
    P9ZeroHMM (hmm);
    return hmm;
}

int
P9FreeHMM (struct plan9_s *hmm)
{
    if (hmm == NULL) {
        return 0;
    }
    free (hmm->ref);
    free (hmm->cs);
    free (hmm->xray);
    free (hmm->name);
    if (hmm->mat != NULL) {
        free (hmm->mat);
    }
    if (hmm->ins != NULL) {
        free (hmm->ins);
    }
    if (hmm->del != NULL) {
        free (hmm->del);
    }
    free (hmm);
    return 1;
}


/* Function: P9ZeroHMM()
 *
 * Purpose:  Zero emission and transition counts in an HMM.
 */
void
P9ZeroHMM (struct plan9_s *hmm)
{
    int k, ts, idx;

    for (k = 0; k <= hmm->M + 1; k++) {
        for (ts = 0; ts < 3; ts++) {
            hmm->mat[k].t[ts] = 0.0;
            hmm->ins[k].t[ts] = 0.0;
            hmm->del[k].t[ts] = 0.0;
        }
        for (idx = 0; idx < Alphabet_size; idx++) {
            hmm->mat[k].p[idx] = 0.0;
            hmm->ins[k].p[idx] = 0.0;
            hmm->del[k].p[idx] = 0.0;
        }
    }
}





/* Function: P9Renormalize()
 *
 * Normalize all P distributions so they sum to 1.
 * P distributions that are all 0, or contain negative
 * probabilities, are left untouched.
 *
 * Returns 1 on success, or 0 on failure.
 */
void
P9Renormalize (struct plan9_s *hmm)
{
    int k;			/* counter for states                  */

    for (k = 0; k <= hmm->M; k++) {
        /* match state transition frequencies */
        FNorm (hmm->mat[k].t, 3);
        FNorm (hmm->ins[k].t, 3);
        if (k > 0) {
            FNorm (hmm->del[k].t, 3);
        }

        if (k > 0) {
            FNorm (hmm->mat[k].p, Alphabet_size);
        }
        FNorm (hmm->ins[k].p, Alphabet_size);
    }
}

/* Function: P9DefaultNullModel()
 *
 * Purpose:  Set up a default random sequence model, using
 *           global aafq[]'s for protein or 0.25 for nucleic
 *           acid. randomseq is alloc'ed in caller. Alphabet information
 *           must already be known.
 */
void
P9DefaultNullModel (float *null)
{
    int x;
    if (Alphabet_type == hmmAMINO)
        for (x = 0; x < Alphabet_size; x++) {
            null[x] = aafq[x];
        }
    else if (Alphabet_type == hmmNUCLEIC)
        for (x = 0; x < Alphabet_size; x++) {
            null[x] = 0.25;
        }
    else {
        Die ("No support for non-protein, non-nucleic acid alphabets.");
    }
}

/************************************************************
 * Copyright (C) 1998 Ian Holmes (ihh@sanger.ac.uk)
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* postprob.c
 * Author: Ian Holmes (ihh@sanger.ac.uk, Jun 5 1998)
 * Derived from core_algorithms.c (SRE, Nov 11 1996)
 * Incorporated SRE, Sat Nov  6 09:07:12 1999 [Cold Spring Harbor]
 *
 * RCS $Id: postprob.c,v 1.3 2003/04/14 16:00:17 eddy Exp $
 *****************************************************************
 * IHH's notes:
 *
 * Functions for working with posterior probabilities,
 * including unfussed "backwards" and "optimal accuracy"
 * implementations.
 *****************************************************************
 * SRE's notes:
 *
 * Simple API example:
 *     struct p7trace_s  *tr;
 *     struct dpmatrix_s *fwd;
 *     struct dpmatrix_s *bck;
 *     struct dpmatrix_s *posterior;
 *     char *postcode;
 *
 *     (get a traceback from somewhere: P7Viterbi() or a modelmaker)
 *     (get an HMM from somewhere: read file or construct it)
 *     P7Forward (dsq, len, hmm, &fwd);
 *     P7Backward(dsq, len, hmm, &bck);
 *     posterior = bck;              -- can alloc posterior, but also can re-use bck --
 *     P7EmitterPosterior(len, hmm, fwd, bck, posterior);
 *     postcode = PostalCode(len, posterior, tr);
 *
 *     MSAAppendGR(msa, "POST", seqidx, postcode);  -- or a similar annotation call --
 *
 *     free(postcode);
 *     FreePlan7Matrix(fwd);
 *     FreePlan7Matrix(bck);
 *
 * P7OptimalAccuracy() - the Durbin/Holmes optimal accuracy
 *                       alignment algorithm. Takes a sequence
 *                       and an HMM, returns an alignment as
 *                       a trace structure.
 *
 * P7Backward()        - The Backward() algorithm, counterpart
 *                       of P7Forward() in core_algorithms.c.
 *
 * P7EmitterPosterior()- The heart of postprob.c: given a Forward
 *                       and a Backward matrix, calculate a new matrix
 *                       that contains the posterior probabilities
 *                       for each symbol i being emitted by
 *                       state k (so, \sum_k p(k | x_i) = 1.0).
 *
 * P7FillOptimalAccuracy() - The core DP algorithm called by
 *                       P7OptimalAccuracy().
 *
 * P7OptimalAccuracyTrace() - the traceback algorithm called by
 *                       P7FillOptimalAccuracy().
 *
 * PostalCode()        - Create a character string for annotating
 *                       an alignment.
 *
 * No small memory variants of these algorithms are available
 * right now.
 */

#include "config.h"
#include "squidconf.h"

#include "squid.h"
#include "structs.h"
#include "funcs.h"

/* Function: P7OptimalAccuracy()
 *
 * Purpose:  The optimal accuracy dynamic programming algorithm.
 *           Identical to Viterbi() except that posterior residue
 *           label probabilities are used as scores.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           ret_tr - RETURN: traceback; pass NULL if it's not wanted
 *
 * Return:   log ( sum_{residues} P(label|M,D) ), as a bit score
 *           (i.e. log of expected accuracy)
 */
float
P7OptimalAccuracy (char *dsq, int L, struct plan7_s *hmm,
                   struct p7trace_s **ret_tr)
{
    double sc;
    struct dpmatrix_s *forward;
    struct dpmatrix_s *backward;

    (void) P7Forward (dsq, L, hmm, &forward);
    (void) P7Backward (dsq, L, hmm, &backward);

    P7EmitterPosterior (L, hmm, forward, backward, backward);	/* Re-use backward matrix for posterior scores */

    sc = P7FillOptimalAccuracy (L, hmm->M, backward, forward, ret_tr);	/* Re-use forward matrix for optimal accuracy scores */

    FreePlan7Matrix (forward);
    FreePlan7Matrix (backward);

    return sc;
}



/* Function: P7Backward()
 *
 * Purpose:  The Backward dynamic programming algorithm.
 *           The scaling issue is dealt with by working in log space
 *           and calling ILogsum(); this is a slow but robust approach.
 *
 * Args:     dsq    - sequence in digitized form
 *           L      - length of dsq
 *           hmm    - the model
 *           ret_mx - RETURN: dp matrix; pass NULL if it's not wanted
 *
 * Return:   log P(S|M)/P(S|R), as a bit score.
 */
float
P7Backward (char *dsq, int L, struct plan7_s *hmm, struct dpmatrix_s **ret_mx)
{
    struct dpmatrix_s *mx;
    int **xmx;
    int **mmx;
    int **imx;
    int **dmx;
    int i, k;
    int sc;

    /* Allocate a DP matrix with 0..L rows, 0..M-1 columns.
     */
    mx = AllocPlan7Matrix (L + 1, hmm->M, &xmx, &mmx, &imx, &dmx);

    /* Initialization of the L row.
     * Note that xmx[i][stS] = xmx[i][stN] by definition for all i,
     *    so stS need not be calculated in backward DP matrices.
     */
    xmx[L][XMC] = hmm->xsc[XTC][MOVE];	/* C<-T                          */
    xmx[L][XME] = xmx[L][XMC] + hmm->xsc[XTE][MOVE];	/* E<-C, no C-tail               */
    xmx[L][XMJ] = xmx[L][XMB] = xmx[L][XMN] = -INFTY;	/* need seq to get out from here */
    for (k = hmm->M; k >= 1; k--) {
        mmx[L][k] = xmx[L][XME] + hmm->esc[k];	/* M<-E ...                      */
        mmx[L][k] += hmm->msc[(int) dsq[L]][k];	/* ... + emitted match symbol    */
        imx[L][k] = dmx[L][k] = -INFTY;	/* need seq to get out from here */
    }

    /* Recursion. Done as a pull.
     * Note slightly wasteful boundary conditions:
     *    M_M precalculated, D_M set to -INFTY,
     *    D_1 wastefully calculated.
     * Scores for transitions to D_M also have to be hacked to -INFTY,
     * as Plan7Logoddsify does not do this for us (I think? - ihh).
     */
    hmm->tsc[TDD][hmm->M - 1] = hmm->tsc[TMD][hmm->M - 1] = -INFTY;	/* no D_M state -- HACK -- should be in Plan7Logoddsify */
    for (i = L - 1; i >= 0; i--) {
        /* Do the special states first.
         * remember, C, N and J emissions are zero score by definition
         */
        xmx[i][XMC] = xmx[i + 1][XMC] + hmm->xsc[XTC][LOOP];

        xmx[i][XMB] = -INFTY;
        /* The following section has been hacked to fit a bug in core_algorithms.c
         * The "correct" code is:
         * for (k = hmm->M; k >= 1; k--)
         * xmx[i][XMB] = ILogsum(xmx[i][XMB], mmx[i+1][k] + hmm->bsc[k];
         *
         * The following code gives the same results as core_algorithms.c:
         */
        xmx[i][XMB] =
            ILogsum (xmx[i][XMB], mmx[i + 1][hmm->M] + hmm->bsc[hmm->M - 1]);
        for (k = hmm->M - 1; k >= 1; k--) {
            xmx[i][XMB] = ILogsum (xmx[i][XMB], mmx[i + 1][k] + hmm->bsc[k]);
        }

        xmx[i][XMJ] = ILogsum (xmx[i][XMB] + hmm->xsc[XTJ][MOVE],
                               xmx[i + 1][XMJ] + hmm->xsc[XTJ][LOOP]);

        xmx[i][XME] = ILogsum (xmx[i][XMC] + hmm->xsc[XTE][MOVE],
                               xmx[i][XMJ] + hmm->xsc[XTE][LOOP]);

        xmx[i][XMN] = ILogsum (xmx[i][XMB] + hmm->xsc[XTN][MOVE],
                               xmx[i + 1][XMN] + hmm->xsc[XTN][LOOP]);

        /* Now the main states. Note the boundary conditions at M.
         */

        if (i > 0) {
            mmx[i][hmm->M] =
                xmx[i][XME] + hmm->esc[hmm->M] + hmm->msc[(int) dsq[i]][hmm->M];
            dmx[i][hmm->M] = -INFTY;
            for (k = hmm->M - 1; k >= 1; k--) {
                mmx[i][k] = ILogsum (ILogsum (xmx[i][XME] + hmm->esc[k],
                                              mmx[i + 1][k + 1] + hmm->tsc[TMM][k]),
                                     ILogsum (imx[i + 1][k] + hmm->tsc[TMI][k],
                                              dmx[i][k + 1] + hmm->tsc[TMD][k]));
                mmx[i][k] += hmm->msc[(int) dsq[i]][k];

                imx[i][k] = ILogsum (imx[i + 1][k] + hmm->tsc[TII][k],
                                     mmx[i + 1][k + 1] + hmm->tsc[TIM][k]);
                imx[i][k] += hmm->isc[(int) dsq[i]][k];

                dmx[i][k] = ILogsum (dmx[i][k + 1] + hmm->tsc[TDD][k],
                                     mmx[i + 1][k + 1] + hmm->tsc[TDM][k]);

            }
        }

    }

    sc = xmx[0][XMN];

    if (ret_mx != NULL) {
        *ret_mx = mx;
    }
    else {
        FreePlan7Matrix (mx);
    }

    return Scorify (sc);		/* the total Backward score. */
}


/* Function: P7EmitterPosterior()
 *
 * Purpose:  Combines Forward and Backward matrices into a posterior
 *           probability matrix.
 *           The entries in row i of this matrix are the logs of the
 *           posterior probabilities of each state emitting symbol i of
 *           the sequence, i.e. all entries for non-emitting states are -INFTY.
 *           The caller must allocate space for the matrix, although the
 *           backward matrix can be used instead (overwriting it will not
 *           compromise the algorithm).
 *
 * Args:     L        - length of sequence
 *           hmm      - the model
 *           forward  - pre-calculated forward matrix
 *           backward - pre-calculated backward matrix
 *           mx       - pre-allocated dynamic programming matrix
 *
 * Return:   void
 */
void
P7EmitterPosterior (int L,
                    struct plan7_s *hmm,
                    struct dpmatrix_s *forward,
                    struct dpmatrix_s *backward, struct dpmatrix_s *mx)
{
    int i;
    int k;
    int sc;

    sc = backward->xmx[0][XMN];

    for (i = L; i >= 1; i--) {
        mx->xmx[i][XMC] =
            forward->xmx[i - 1][XMC] + hmm->xsc[XTC][LOOP] + backward->xmx[i][XMC] -
            sc;

        mx->xmx[i][XMJ] =
            forward->xmx[i - 1][XMJ] + hmm->xsc[XTJ][LOOP] + backward->xmx[i][XMJ] -
            sc;

        mx->xmx[i][XMN] =
            forward->xmx[i - 1][XMN] + hmm->xsc[XTN][LOOP] + backward->xmx[i][XMN] -
            sc;

        mx->xmx[i][XMB] = mx->xmx[i][XME] = -INFTY;

        for (k = 1; k < hmm->M; k++) {
            mx->mmx[i][k] = backward->mmx[i][k];
            mx->mmx[i][k] +=
                ILogsum (ILogsum
                         (forward->mmx[i - 1][k - 1] + hmm->tsc[TMM][k - 1],
                          forward->imx[i - 1][k - 1] + hmm->tsc[TIM][k - 1]),
                         ILogsum (forward->xmx[i - 1][XMB] + hmm->bsc[k],
                                  forward->dmx[i - 1][k - 1] + hmm->tsc[TDM][k - 1]));
            mx->mmx[i][k] -= sc;

            mx->imx[i][k] = backward->imx[i][k];
            mx->imx[i][k] += ILogsum (forward->mmx[i - 1][k] + hmm->tsc[TMI][k],
                                      forward->imx[i - 1][k] + hmm->tsc[TII][k]);
            mx->imx[i][k] -= sc;

            mx->dmx[i][k] = -INFTY;
        }
        mx->mmx[i][hmm->M] = backward->mmx[i][hmm->M];
        mx->mmx[i][hmm->M] +=
            ILogsum (ILogsum
                     (forward->mmx[i - 1][hmm->M - 1] + hmm->tsc[TMM][hmm->M - 1],
                      forward->imx[i - 1][hmm->M - 1] + hmm->tsc[TIM][hmm->M - 1]),
                     ILogsum (forward->xmx[i - 1][XMB] + hmm->bsc[hmm->M],
                              forward->dmx[i - 1][hmm->M - 1] +
                              hmm->tsc[TDM][hmm->M - 1]));
        mx->mmx[i][hmm->M] -= sc;

        mx->imx[i][hmm->M] = mx->dmx[i][hmm->M] = mx->dmx[i][0] = -INFTY;

    }
}


/* Function: P7FillOptimalAccuracy()
 *
 * Purpose:  The core of the optimal accuracy dynamic programming algorithm.
 *           Identical to Viterbi() except that scores are given by a
 *           posterior matrix (that the caller must pre-calculate).
 *           Also, the caller must pre-allocate the optimal accuracy matrix
 *           (this allows the forward matrix to be re-used).
 *           P7OptimalAccuracy() does all this for you and cleans up.
 *
 *
 * Args:     L         - length of sequence
 *           M         - length of model
 *           posterior - pre-calculated emitter posterior matrix
 *           mx        - pre-allocated dynamic programming matrix
 *           ret_tr    - RETURN: traceback; pass NULL if it's not wanted
 *
 * Return:   log ( sum_{residues} P(label|M,D) ), as a bit score
 *           (i.e. log of expected accuracy)
 */
float
P7FillOptimalAccuracy (int L,
                       int M,
                       struct dpmatrix_s *posterior,
                       struct dpmatrix_s *mx, struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;
    int **xmx;
    int **mmx;
    int **imx;
    int **dmx;
    int i, k;
    int sc;

    xmx = mx->xmx;
    mmx = mx->mmx;
    imx = mx->imx;
    dmx = mx->dmx;

    /* Initialization of the zero row.
     * Each cell in the optimal accuracy matrix holds the log of the expected
     * of correctly assigned symbols up to that point.
     * To begin with, everything is log(0) = -INFTY.
     */
    xmx[0][XMN] = xmx[0][XMB] = xmx[0][XME] = xmx[0][XMC] = xmx[0][XMJ] =
                                    -INFTY;
    for (k = 0; k <= M; k++) {
        mmx[0][k] = imx[0][k] = dmx[0][k] = -INFTY;
    }

    /* Recursion. Done as a pull.
     * Note some slightly wasteful boundary conditions:
     *    D_M and I_M are wastefully calculated (they don't exist)
     */
    for (i = 1; i <= L; i++) {
        mmx[i][0] = imx[i][0] = dmx[i][0] = -INFTY;

        for (k = 1; k <= M; k++) {
            /* match state */
            mmx[i][k] = -INFTY;
            if ((sc = mmx[i - 1][k - 1]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc = imx[i - 1][k - 1]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc = dmx[i - 1][k - 1]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            if ((sc = xmx[i - 1][XMB]) > mmx[i][k]) {
                mmx[i][k] = sc;
            }
            mmx[i][k] = ILogsum (mmx[i][k], posterior->mmx[i][k]);

            /* delete state */
            dmx[i][k] = -INFTY;
            if ((sc = mmx[i][k - 1]) > dmx[i][k]) {
                dmx[i][k] = sc;
            }
            if ((sc = dmx[i][k - 1]) > dmx[i][k]) {
                dmx[i][k] = sc;
            }

            /* insert state */
            imx[i][k] = -INFTY;
            if ((sc = mmx[i - 1][k]) > imx[i][k]) {
                imx[i][k] = sc;
            }
            if ((sc = imx[i - 1][k]) > imx[i][k]) {
                imx[i][k] = sc;
            }
            imx[i][k] = ILogsum (imx[i][k], posterior->imx[i][k]);
        }

        /* Now the special states. Order is important here.
         * remember, C and J emissions are zero score by definition,
         */

        /* N state */
        xmx[i][XMN] = -INFTY;
        if ((sc = ILogsum (xmx[i - 1][XMN], posterior->xmx[i][XMN])) > -INFTY) {
            xmx[i][XMN] = sc;
        }

        /* E state */
        xmx[i][XME] = -INFTY;
        for (k = 1; k <= M; k++) {
            if ((sc = mmx[i][k]) > xmx[i][XME]) {
                xmx[i][XME] = sc;
            }
				}
        /* J state */
        xmx[i][XMJ] = -INFTY;
        if ((sc = ILogsum (xmx[i - 1][XMJ], posterior->xmx[i][XMJ])) > -INFTY) {
            xmx[i][XMJ] = sc;
        }
        if ((sc = xmx[i][XME]) > xmx[i][XMJ]) {	/* no E->J emission */
            xmx[i][XMJ] = sc;
        }

        /* B state */
        xmx[i][XMB] = -INFTY;
        if ((sc = xmx[i][XMN]) > -INFTY) {
            xmx[i][XMB] = sc;
        }
        if ((sc = xmx[i][XMJ]) > xmx[i][XMB]) {
            xmx[i][XMB] = sc;
        }

        /* C state */
        xmx[i][XMC] = -INFTY;
        if ((sc = ILogsum (xmx[i - 1][XMC], posterior->xmx[i][XMC])) > -INFTY) {
            xmx[i][XMC] = sc;
        }
        if ((sc = xmx[i][XME]) > xmx[i][XMC]) {	/* no E->C emission */
            xmx[i][XMC] = sc;
        }
    }

    /* T state (not stored) */
    sc = xmx[L][XMC];

    if (ret_tr != NULL) {
        P7OptimalAccuracyTrace (L, M, posterior, mx, &tr);
        *ret_tr = tr;
    }

    return Score2Prob (sc, 1);	/* the log of the expected accuracy. */
}


/* Function: P7OptimalAccuracyTrace()
 *
 * Purpose:  Traceback of an optimal accuracy matrix: i.e. retrieval
 *           of optimum alignment.
 *
 * Args:     L         - length of sequence
 *           M         - length of HMM
 *           posterior - the posterior matrix
 *           mx        - the matrix to trace back in, (L+1) x M
 *           ret_tr    - RETURN: traceback.
 *
 * Return:   (void)
 *           ret_tr is allocated here. Free using P7FreeTrace().
 */
void
P7OptimalAccuracyTrace (int L,
                        int M,
                        struct dpmatrix_s *posterior,
                        struct dpmatrix_s *mx, struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;
    int curralloc;		/* current allocated length of trace */
    int tpos;			/* position in trace */
    int i;			/* position in seq (1..L) */
    int k;			/* position in model (1..M) */
    int **xmx, **mmx, **imx, **dmx;
    int sc;			/* temp var for pre-emission score */

    /* Overallocate for the trace.
     * S-N-B- ... - E-C-T  : 6 states + L is minimum trace;
     * add L more as buffer.
     */
    curralloc = L * 2 + 6;
    P7AllocTrace (curralloc, &tr);

    xmx = mx->xmx;
    mmx = mx->mmx;
    imx = mx->imx;
    dmx = mx->dmx;

    /* Initialization of trace
     * We do it back to front; ReverseTrace() is called later.
     */
    tr->statetype[0] = STT;
    tr->nodeidx[0] = 0;
    tr->pos[0] = 0;
    tr->statetype[1] = STC;
    tr->nodeidx[1] = 0;
    tr->pos[1] = 0;
    tpos = 2;
    i = L;			/* current i (seq pos) we're trying to assign */

    /* Traceback
     */
    while (tr->statetype[tpos - 1] != STS) {
        switch (tr->statetype[tpos - 1]) {
        case STM:			/* M connects from i-1,k-1, or B */
            sc = mmx[i + 1][k + 1];
            if (sc == ILogsum (mmx[i][k], posterior->mmx[i + 1][k + 1]) && i > 0
                    && k > 0) {
                tr->statetype[tpos] = STM;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = i--;
            }
            else if (sc == ILogsum (imx[i][k], posterior->mmx[i + 1][k + 1])
                     && i > 0 && k > 0) {
                tr->statetype[tpos] = STI;
                tr->nodeidx[tpos] = k;
                tr->pos[tpos] = i--;
            }
            else if (sc == ILogsum (dmx[i][k], posterior->mmx[i + 1][k + 1])
                     && i > 0 && k > 1) {
                tr->statetype[tpos] = STD;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = 0;
            }
            else if (sc == ILogsum (xmx[i][XMB], posterior->mmx[i + 1][k + 1])) {
                tr->statetype[tpos] = STB;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STD:			/* D connects from M,D */
            if (dmx[i][k + 1] == mmx[i][k] && i > 0 && k > 0) {
                tr->statetype[tpos] = STM;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = i--;
            }
            else if (dmx[i][k + 1] == dmx[i][k] && k > 1) {
                tr->statetype[tpos] = STD;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = 0;
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STI:			/* I connects from M,I */
            sc = imx[i + 1][k];
            if (sc == ILogsum (mmx[i][k], posterior->imx[i + 1][k]) && i > 0
                    && k > 0) {
                tr->statetype[tpos] = STM;
                tr->nodeidx[tpos] = k--;
                tr->pos[tpos] = i--;
            }
            else if (sc == ILogsum (imx[i][k], posterior->imx[i + 1][k]) && i > 0
                     && k > 0) {
                tr->statetype[tpos] = STI;
                tr->nodeidx[tpos] = k;
                tr->pos[tpos] = i--;
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STN:			/* N connects from S, N */
            if (i == 0 && xmx[i][XMN] == -INFTY) {
                tr->statetype[tpos] = STS;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }
            else if (i > 0
                     && xmx[i + 1][XMN] == ILogsum (xmx[i][XMN],
                                                    posterior->xmx[i + 1][XMN])
                     && i > 0) {
                tr->statetype[tpos] = STN;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* note convention adherence:  */
                tr->pos[tpos - 1] = i--;	/* first N doesn't emit        */
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STB:			/* B connects from N, J */
            if (xmx[i][XMB] == xmx[i][XMN]) {
                tr->statetype[tpos] = STN;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }
            else if (xmx[i][XMB] == xmx[i][XMJ]) {
                tr->statetype[tpos] = STJ;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;
            }
            else {
                Die ("traceback failed");
            }
            break;

        case STE:			/* E connects from any M state. k set here */
            for (k = M; k >= 1; k--) {
                if (xmx[i][XME] == mmx[i][k] && i > 0) {
                    tr->statetype[tpos] = STM;
                    tr->nodeidx[tpos] = k--;
                    tr->pos[tpos] = i--;
                    break;
                }
						}		
            if (k <= 0) {
                Die ("traceback failed");
            }
            break;

        case STC:			/* C comes from C, E */
            if (xmx[i][XMC] == ILogsum (xmx[i - 1][XMC], posterior->xmx[i][XMC])
                    && i > 0) {
                tr->statetype[tpos] = STC;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* note convention adherence: */
                tr->pos[tpos - 1] = i--;	/* first C doesn't emit       */
            }
            else if (xmx[i][XMC] == xmx[i][XME]) {
                tr->statetype[tpos] = STE;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* E is a nonemitter */
            }
            else {
                Die ("Traceback failed.");
            }
            break;

        case STJ:			/* J connects from E, J */
            if (xmx[i][XMJ] == ILogsum (xmx[i - 1][XMJ], posterior->xmx[i][XMJ])
                    && i > 0) {
                tr->statetype[tpos] = STJ;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* note convention adherence: */
                tr->pos[tpos - 1] = i--;	/* first J doesn't emit       */
            }
            else if (xmx[i][XMJ] == xmx[i][XME]) {
                tr->statetype[tpos] = STE;
                tr->nodeidx[tpos] = 0;
                tr->pos[tpos] = 0;	/* E is a nonemitter */
            }
            else {
                Die ("Traceback failed.");
            }
            break;

        default:
            Die ("traceback failed");

        }				/* end switch over statetype[tpos-1] */

        tpos++;
        if (tpos == curralloc) {	/* grow trace if necessary  */
            curralloc += L;
            P7ReallocTrace (tr, curralloc);
        }

    }				/* end traceback, at S state; tpos == tlen now */
    tr->tlen = tpos;
    P7ReverseTrace (tr);
    *ret_tr = tr;

}


/* Function: PostalCode()
 * Date:     SRE, Sun Nov  7 15:31:35 1999 [Cold Spring Harbor]
 *
 * Purpose:  Given a traceback and one of Ian's posterior
 *           probability matrices, calculate a string that
 *           represents the confidence values on each
 *           residue in the sequence.
 *
 *           The code string is 0..L-1  (L = len of target seq),
 *           so it's in the coordinate system of the sequence string;
 *           off by one from dsq; and convertible to the coordinate
 *           system of aseq using MakeAlignedString().
 *
 *           Values are 0-9,*
 *           for example, 9 means with >=90% posterior probabiility,
 *           residue i is aligned to the state k that it
 *           is assigned to in the given trace.
 *
 * Args:     L  - length of seq
 *           mx - posterior prob matrix: see P7EmitterPosterior()
 *           tr - a traceback to get a Postal code string for.
 *
 * Returns:  char * array of codes, 0..L-1
 *           Caller is responsible for free'ing it.
 */
static char
score2postcode (int sc)
{
    char i;
    i = (char) (Score2Prob (sc, 1.) * 10.);
    return ((i > 9) ? '*' : '0' + i);
}

char *
PostalCode (int L, struct dpmatrix_s *mx, struct p7trace_s *tr)
{
    int tpos;
    int i;
    int k;
    char *postcode;

    postcode = MallocOrDie ((L + 1) * sizeof (char));
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        i = tr->pos[tpos];
        k = tr->nodeidx[tpos];
        if (i == 0) {
            continue;
        }

        switch (tr->statetype[tpos]) {
        case STM:
            postcode[i - 1] = score2postcode (mx->mmx[i][k]);
            break;
        case STI:
            postcode[i - 1] = score2postcode (mx->imx[i][k]);
            break;
        case STN:
            postcode[i - 1] = score2postcode (mx->xmx[i][XMN]);
            break;
        case STC:
            postcode[i - 1] = score2postcode (mx->xmx[i][XMC]);
            break;
        case STJ:
            postcode[i - 1] = score2postcode (mx->xmx[i][XMJ]);
            break;
        }
    }
    postcode[L] = '\0';

    return postcode;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* prior.c
 * SRE, Mon Nov 18 15:44:08 1996
 *
 * Support for Dirichlet prior data structure, p7prior_s.
 */

#include "config.h"
#include "squidconf.h"

#include <string.h>

#include "squid.h"
#include "structs.h"
#include "funcs.h"


static struct p7prior_s *default_amino_prior (void);
static struct p7prior_s *default_nucleic_prior (void);

/* Function: P7AllocPrior(), P7FreePrior()
 *
 * Purpose:  Allocation and free'ing of a prior structure.
 *           Very simple, but might get more complex someday.
 */
struct p7prior_s *
P7AllocPrior (void)
{
    return (struct p7prior_s *) MallocOrDie (sizeof (struct p7prior_s));
}

void
P7FreePrior (struct p7prior_s *pri)
{
    free (pri);
}


/* Function: P7LaplacePrior()
 *
 * Purpose:  Create a Laplace plus-one prior. (single component Dirichlets).
 *           Global alphabet info is assumed to have been set already.
 *
 * Args:     (void)
 *
 * Return:   prior. Allocated here; call FreePrior() to free it.
 */
struct p7prior_s *
P7LaplacePrior (void)
{
    struct p7prior_s *pri;

    pri = P7AllocPrior ();
    pri->strategy = PRI_DCHLET;

    pri->tnum = 1;
    pri->tq[0] = 1.;
    FSet (pri->t[0], 8, 1.);

    pri->mnum = 1;
    pri->mq[0] = 1.;
    FSet (pri->m[0], Alphabet_size, 1.);

    pri->inum = 1;
    pri->iq[0] = 1.;
    FSet (pri->i[0], Alphabet_size, 1.);

    return pri;
}

/* Function: P7DefaultPrior()
 *
 * Purpose:  Set up a somewhat more realistic single component
 *           Dirichlet prior than Laplace.
 */
struct p7prior_s *
P7DefaultPrior (void)
{
    switch (Alphabet_type) {
    case hmmAMINO:
        return default_amino_prior ();
    case hmmNUCLEIC:
        return default_nucleic_prior ();
    case hmmNOTSETYET:
        Die ("Can't set prior; alphabet type not set yet");
    }
    /*NOTREACHED*/ return NULL;
}

/* Function: P7ReadPrior()
 *
 * Purpose:  Input a prior from disk file.
 */
struct p7prior_s *
P7ReadPrior (char *prifile)
{
    FILE *fp;
    struct p7prior_s *pri;
    char *sptr;
    int q, x;

    if ((fp = fopen (prifile, "r")) == NULL) {
        Die ("Failed to open HMMER prior file %s\n", prifile);
    }
    pri = P7AllocPrior ();

    /* First entry is the strategy:
     * Only standard Dirichlet prior (simple or mixture) is supported in Plan7 so far
     */
    sptr = Getword (fp, sqdARG_STRING);
    s2upper (sptr);
    if (strcmp (sptr, "DIRICHLET") == 0) {
        pri->strategy = PRI_DCHLET;
    }
    else {
        Die ("No such prior strategy %s; failed to parse file %s", sptr, prifile);
    }

    /* Second entry is the alphabet type:
     * Amino or Nucleic
     */
    sptr = Getword (fp, sqdARG_STRING);
    s2upper (sptr);
    if (strcmp (sptr, "AMINO") == 0) {
        if (Alphabet_type != hmmAMINO)
            Die ("HMM and/or sequences are DNA/RNA; can't use protein prior %s",
                 prifile);
    }
    else if (strcmp (sptr, "NUCLEIC") == 0) {
        if (Alphabet_type != hmmNUCLEIC)
            Die ("HMM and/or sequences are protein; can't use DNA/RNA prior %s",
                 prifile);
    }
    else {
        Die ("Alphabet \"%s\" in prior file %s isn't valid.", sptr, prifile);
    }

    /* State transition priors:
     * # of mixtures.
     * then for each mixture:
     *    prior P(q)
     *    Dirichlet terms for Tmm, Tmi, Tmd, Tim, Tii, Tid, Tdm, Tdi, Tdd
     */
    pri->tnum = atoi (Getword (fp, sqdARG_INT));
    if (pri->tnum < 0)
        Die ("%d is bad; need at least one state transition mixture component",
             pri->tnum);
    if (pri->tnum > MAXDCHLET)
        Die ("%d is bad, too many transition components (MAXDCHLET = %d)\n",
             MAXDCHLET);
    for (q = 0; q < pri->tnum; q++) {
        pri->tq[q] = (float) atof (Getword (fp, sqdARG_FLOAT));
        for (x = 0; x < 7; x++) {
            pri->t[q][x] = (float) atof (Getword (fp, sqdARG_FLOAT));
        }
    }

    /* Match emission priors:
     * # of mixtures.
     * then for each mixture:
     *    prior P(q)
     *    Dirichlet terms for Alphabet_size symbols in Alphabet
     */
    pri->mnum = atoi (Getword (fp, sqdARG_INT));
    if (pri->mnum < 0)
        Die ("%d is bad; need at least one match emission mixture component",
             pri->mnum);
    if (pri->mnum > MAXDCHLET)
        Die ("%d is bad; too many match components (MAXDCHLET = %d)\n", pri->mnum,
             MAXDCHLET);

    for (q = 0; q < pri->mnum; q++) {
        pri->mq[q] = (float) atof (Getword (fp, sqdARG_FLOAT));
        for (x = 0; x < Alphabet_size; x++) {
            pri->m[q][x] = (float) atof (Getword (fp, sqdARG_FLOAT));
        }
    }

    /* Insert emission priors:
     * # of mixtures.
     * then for each mixture component:
     *    prior P(q)
     *    Dirichlet terms for Alphabet_size symbols in Alphabet
     */
    pri->inum = atoi (Getword (fp, sqdARG_INT));
    if (pri->inum < 0)
        Die ("%d is bad; need at least one insert emission mixture component",
             pri->inum);
    if (pri->inum > MAXDCHLET)
        Die ("%d is bad; too many insert components (MAXDCHLET = %d)\n",
             pri->inum, MAXDCHLET);
    for (q = 0; q < pri->inum; q++) {
        pri->iq[q] = (float) atof (Getword (fp, sqdARG_FLOAT));
        for (x = 0; x < Alphabet_size; x++) {
            pri->i[q][x] = (float) atof (Getword (fp, sqdARG_FLOAT));
        }
    }

    fclose (fp);
    return pri;
}


/* Function: PAMPrior()
 *
 * Purpose:  Produces an ad hoc "Dirichlet mixture" prior for
 *           match emissions, using a PAM matrix.
 *
 *           Side effect notice: PAMPrior() replaces the match
 *           emission section of an existing Dirichlet prior,
 *           which is /expected/ to be a simple one-component
 *           kind of prior. The insert emissions /must/ be a
 *           one-component prior (because of details in how
 *           PriorifyEmissionVector() is done). However,
 *           the transitions /could/ be a mixture Dirichlet prior
 *           without causing problems. In other words, the
 *           -p and -P options of hmmb can coexist, but there
 *           may be conflicts. PAMPrior() checks for these,
 *           so there's no serious problem, except that the
 *           error message from PAMPrior() might be confusing to
 *           a user.
 */
void
PAMPrior (char *pamfile, struct p7prior_s *pri, float wt)
{
    FILE *fp;
    char *blastpamfile;		/* BLAST looks in aa/ subdirectory of BLASTMAT */
    int **pam;
    float scale;
    int xi, xj;
    int idx1, idx2;

    if (Alphabet_type != hmmAMINO) {
        Die ("PAM prior is only valid for protein sequences");
    }
    if (pri->strategy != PRI_DCHLET) {
        Die ("PAM prior may only be applied over an existing Dirichlet prior");
    }
    if (pri->inum != 1)
        Die
        ("PAM prior requires that the insert emissions be a single Dirichlet");
    if (MAXDCHLET < 20)
        Die
        ("Whoa, code is misconfigured; MAXDCHLET must be >= 20 for PAM prior");

    blastpamfile = FileConcat ("aa", pamfile);

    if ((fp = fopen (pamfile, "r")) == NULL &&
            (fp = EnvFileOpen (pamfile, "BLASTMAT", NULL)) == NULL &&
            (fp = EnvFileOpen (blastpamfile, "BLASTMAT", NULL)) == NULL) {
        Die ("Failed to open PAM scoring matrix file %s", pamfile);
    }
    if (!ParsePAMFile (fp, &pam, &scale)) {
        Die ("Failed to parse PAM scoring matrix file %s", pamfile);
    }
    fclose (fp);
    free (blastpamfile);

    pri->strategy = PRI_PAM;
    pri->mnum = 20;

    /* Convert PAM entries back to conditional prob's P(xj | xi),
     * which we'll use as "pseudocounts" weighted by wt.
     */
    for (xi = 0; xi < Alphabet_size; xi++) {
        for (xj = 0; xj < Alphabet_size; xj++) {
            idx1 = Alphabet[xi] - 'A';
            idx2 = Alphabet[xj] - 'A';
            pri->m[xi][xj] = aafq[xj] * exp ((float) pam[idx1][idx2] * scale);
        }
		}
    /* Normalize so that rows add up to wt.
     * i.e. Sum(xj) mat[xi][xj] = wt for every row xi
     */
    for (xi = 0; xi < Alphabet_size; xi++) {
        pri->mq[xi] = 1. / Alphabet_size;
        FNorm (pri->m[xi], Alphabet_size);
        FScale (pri->m[xi], Alphabet_size, wt);
    }

    Free2DArray ((void **) pam, 27);
}


/* Function: P7DefaultNullModel()
 *
 * Purpose:  Set up a default random sequence model, using
 *           global aafq[]'s for protein or 1/Alphabet_size for anything
 *           else. randomseq is alloc'ed in caller. Alphabet information
 *           must already be known.
 */
void
P7DefaultNullModel (float *null, float *ret_p1)
{
    int x;
    if (Alphabet_type == hmmAMINO) {
        for (x = 0; x < Alphabet_size; x++) {
            null[x] = aafq[x];
        }
        *ret_p1 = 350. / 351.;	/* rationale: approx avg protein length. */
    }
    else {
        for (x = 0; x < Alphabet_size; x++) {
            null[x] = 1.0 / (float) Alphabet_size;
        }
        *ret_p1 = 1000. / 1001.;	/* rationale: approx inter-Alu distance. */
    }
}

void
P7ReadNullModel (char *rndfile, float *null, float *ret_p1)
{
    FILE *fp;
    char *s;
    int x;
    int type = 0;

    if ((fp = fopen (rndfile, "r")) == NULL) {
        Die ("Failed to open null model file %s\n", rndfile);
    }
    if ((s = Getword (fp, sqdARG_STRING)) == NULL) {
        goto FAILURE;
    }
    s2upper (s);
    if (strcmp (s, "NUCLEIC") == 0) {
        type = hmmNUCLEIC;
    }
    else if (strcmp (s, "AMINO") == 0) {
        type = hmmAMINO;
    }
    else {
        goto FAILURE;
    }
    /* check/set alphabet type */
    if (Alphabet_type == 0) {
        SetAlphabet (type);
    }
    else if (Alphabet_type != type)
        Die ("Alphabet type conflict; null model in %s is inappropriate\n",
             rndfile);
    /* parse the file */
    for (x = 0; x < Alphabet_size; x++) {
        if ((s = Getword (fp, sqdARG_FLOAT)) == NULL) {
            goto FAILURE;
        }
        null[x] = atof (s);
    }
    if ((s = Getword (fp, sqdARG_FLOAT)) == NULL) {
        goto FAILURE;
    }
    *ret_p1 = atof (s);

    fclose (fp);
    return;

FAILURE:
    fclose (fp);
    Die ("%s is not in HMMER null model file format", rndfile);
}


/* Function: P7PriorifyHMM()
 *
 * Purpose:  Add pseudocounts to an HMM using Dirichlet priors,
 *           and renormalize the HMM.
 *
 * Args:     hmm -- the HMM to add counts to (counts form)
 *           pri -- the Dirichlet prior to use
 *
 * Return:   (void)
 *           HMM returns in probability form.
 */
void
P7PriorifyHMM (struct plan7_s *hmm, struct p7prior_s *pri)
{
    int k;			/* counter for model position   */
    float d;			/* a denominator */
    float tq[MAXDCHLET];		/* prior distribution over mixtures */
    float mq[MAXDCHLET];		/* prior distribution over mixtures */
    float iq[MAXDCHLET];		/* prior distribution over mixtures */

    /* Model-dependent transitions are handled simply; Laplace.
     */
    FSet (hmm->begin + 2, hmm->M - 1, 0.);	/* wipe internal BM entries */
    FSet (hmm->end + 1, hmm->M - 1, 0.);	/* wipe internal ME exits   */
    d = hmm->tbd1 + hmm->begin[1] + 2.;
    hmm->tbd1 = (hmm->tbd1 + 1.) / d;
    hmm->begin[1] = (hmm->begin[1] + 1.) / d;
    hmm->end[hmm->M] = 1.0;

    /* Main model transitions and emissions
     */
    for (k = 1; k < hmm->M; k++) {
        /* The following code chunk is experimental.
         * Collaboration with Michael Asman, Erik Sonnhammer, CGR Stockholm.
         * Only activated if X-PR* annotation has been used, in which
         * priors are overridden and a single Dirichlet component is
         * specified for each column (using structural annotation).
         * If X-PR* annotation is not used, which is usually the case,
         * the following code has no effect (observe how the real prior
         * distributions are copied into tq, mq, iq).
         */
        if (hmm->tpri != NULL && hmm->tpri[k] >= 0) {
            if (hmm->tpri[k] >= pri->tnum) {
                Die ("X-PRT annotation out of range");
            }
            FSet (tq, pri->tnum, 0.0);
            tq[hmm->tpri[k]] = 1.0;
        }
        else {
            FCopy (tq, pri->tq, pri->tnum);
        }
        if (hmm->mpri != NULL && hmm->mpri[k] >= 0) {
            if (hmm->mpri[k] >= pri->mnum) {
                Die ("X-PRM annotation out of range");
            }
            FSet (mq, pri->mnum, 0.0);
            mq[hmm->mpri[k]] = 1.0;
        }
        else {
            FCopy (mq, pri->mq, pri->mnum);
        }
        if (hmm->ipri != NULL && hmm->ipri[k] >= 0) {
            if (hmm->ipri[k] >= pri->inum) {
                Die ("X-PRI annotation out of range");
            }
            FSet (iq, pri->inum, 0.0);
            iq[hmm->ipri[k]] = 1.0;
        }
        else {
            FCopy (iq, pri->iq, pri->inum);
        }

        /* This is the main line of the code:
         */
        P7PriorifyTransitionVector (hmm->t[k], pri, tq);
        P7PriorifyEmissionVector (hmm->mat[k], pri, pri->mnum, mq, pri->m, NULL);
        P7PriorifyEmissionVector (hmm->ins[k], pri, pri->inum, iq, pri->i, NULL);
    }

    /* We repeat the above steps just for the final match state, M.
     */
    if (hmm->mpri != NULL && hmm->mpri[hmm->M] >= 0) {
        if (hmm->mpri[hmm->M] >= pri->mnum) {
            Die ("X-PRM annotation out of range");
        }
        FSet (mq, pri->mnum, 0.0);
        mq[hmm->mpri[hmm->M]] = 1.0;
    }
    else {
        FCopy (mq, pri->mq, pri->mnum);
    }

    P7PriorifyEmissionVector (hmm->mat[hmm->M], pri, pri->mnum, mq, pri->m,
                              NULL);

    /* Now we're done. Convert the counts-based HMM to probabilities.
     */
    Plan7Renormalize (hmm);
}


/* Function: P7PriorifyEmissionVector()
 *
 * Purpose:  Add prior pseudocounts to an observed
 *           emission count vector and renormalize.
 *
 *           Can return the posterior mixture probabilities
 *           P(q | counts) if ret_mix[MAXDCHLET] is passed.
 *           Else, pass NULL.
 *
 * Args:     vec     - the 4 or 20-long vector of counts to modify
 *           pri     - prior data structure
 *           num     - pri->mnum or pri->inum; # of mixtures
 *           eq      - pri->mq or pri->iq; prior mixture probabilities
 *           e       - pri->i or pri->m; Dirichlet components
 *           ret_mix - filled with posterior mixture probabilities, or NULL
 *
 * Return:   (void)
 *           The counts in vec are changed and normalized to probabilities.
 */
void
P7PriorifyEmissionVector (float *vec, struct p7prior_s *pri,
                          int num, float eq[MAXDCHLET],
                          float e[MAXDCHLET][MAXABET], float *ret_mix)
{
    int x;			/* counter over vec                     */
    int q;			/* counter over mixtures                */
    float mix[MAXDCHLET];		/* posterior distribution over mixtures */
    float totc;			/* total counts                         */
    float tota;			/* total alpha terms                    */
    float xi;			/* X_i term, Sjolander eq. 41           */

    /* Calculate mix[], which is the posterior probability
     * P(q | n) of mixture component q given the count vector n
     *
     * (side effect note: note that an insert vector in a PAM prior
     * is passed with num = 1, bypassing pam prior code; this means
     * that inserts cannot be mixture Dirichlets...)
     * [SRE, 12/24/00: the above comment is cryptic! what the hell does that
     *  mean, inserts can't be mixtures? doesn't seem to be true. it
     *  may mean that in a PAM prior, you can't have a mixture for inserts,
     *  but I don't even understand that. The insert vectors aren't passed
     *  with num=1!!]
     */
    mix[0] = 1.0;
    if (pri->strategy == PRI_DCHLET && num > 1) {
        for (q = 0; q < num; q++) {
            mix[q] = eq[q] > 0.0 ? log (eq[q]) : -999.;
            mix[q] += Logp_cvec (vec, Alphabet_size, e[q]);
        }
        LogNorm (mix, num);		/* now mix[q] is P(component_q | n) */
    }
    else if (pri->strategy == PRI_PAM && num > 1) {	/* pam prior uses aa frequencies as `P(q|n)' */
        for (q = 0; q < Alphabet_size; q++) {
            mix[q] = vec[q];
        }
        FNorm (mix, Alphabet_size);
    }

    /* Convert the counts to probabilities, following Sjolander (1996)
     */
    totc = FSum (vec, Alphabet_size);
    for (x = 0; x < Alphabet_size; x++) {
        xi = 0.0;
        for (q = 0; q < num; q++) {
            tota = FSum (e[q], Alphabet_size);
            xi += mix[q] * (vec[x] + e[q][x]) / (totc + tota);
        }
        vec[x] = xi;
    }
    FNorm (vec, Alphabet_size);

    if (ret_mix != NULL)
        for (q = 0; q < num; q++) {
            ret_mix[q] = mix[q];
        }
}



/* Function: P7PriorifyTransitionVector()
 *
 * Purpose:  Add prior pseudocounts to transition vector,
 *           which contains three different probability vectors
 *           for m, d, and i.
 *
 * Args:     t     - state transitions, counts: 3 for M, 2 for I, 2 for D.
 *           prior - Dirichlet prior information
 *           tq    - prior distribution over Dirichlet components.
 *                   (overrides prior->iq[]; used for alternative
 *                   methods of conditioning prior on structural data)
 *
 * Return:   (void)
 *           t is changed, and renormalized -- comes back as
 *           probability vectors.
 */
void
P7PriorifyTransitionVector (float *t, struct p7prior_s *prior,
                            float tq[MAXDCHLET])
{
    int ts;
    int q;
    float mix[MAXDCHLET];
    float totm, totd, toti;	/* total counts in three transition vecs */
    float xi;			/* Sjolander's X_i term */

    mix[0] = 1.0;			/* default is simple one component */
    if ((prior->strategy == PRI_DCHLET || prior->strategy == PRI_PAM)
            && prior->mnum > 1) {
        for (q = 0; q < prior->tnum; q++) {
            mix[q] = tq[q] > 0.0 ? log (tq[q]) : -999.;
            mix[q] += Logp_cvec (t, 3, prior->t[q]);	/* 3 match  */
            mix[q] += Logp_cvec (t + 3, 2, prior->t[q] + 3);	/* 2 insert */
            mix[q] += Logp_cvec (t + 5, 2, prior->t[q] + 5);	/* 2 delete */
        }
        LogNorm (mix, prior->tnum);	/* mix[q] is now P(q | counts) */
    }
    /* precalc some denominators */
    totm = FSum (t, 3);
    toti = t[TIM] + t[TII];
    totd = t[TDM] + t[TDD];

    for (ts = 0; ts < 7; ts++) {
        xi = 0.0;
        for (q = 0; q < prior->tnum; q++) {
            switch (ts) {
            case TMM:
            case TMI:
            case TMD:
                xi += mix[q] * (t[ts] + prior->t[q][ts]) /
                      (totm + FSum (prior->t[q], 3));
                break;
            case TIM:
            case TII:
                xi += mix[q] * (t[ts] + prior->t[q][ts]) /
                      (toti + prior->t[q][TIM] + prior->t[q][TII]);
                break;
            case TDM:
            case TDD:
                xi += mix[q] * (t[ts] + prior->t[q][ts]) /
                      (totd + prior->t[q][TDM] + prior->t[q][TDD]);
                break;
            }
        }
        t[ts] = xi;
    }
    FNorm (t, 3);			/* match  */
    FNorm (t + 3, 2);		/* insert */
    FNorm (t + 5, 2);		/* delete */
}


/* Function: default_amino_prior()
 *
 * Purpose:  Set the default protein prior.
 */
static struct p7prior_s *
default_amino_prior (void)
{
    struct p7prior_s *pri;
    int q, x;
    /* default match mixture coefficients */
    static float defmq[9] = {
        0.178091, 0.056591, 0.0960191, 0.0781233, 0.0834977,
        0.0904123, 0.114468, 0.0682132, 0.234585
    };

    /* default match mixture Dirichlet components */
    static float defm[9][20] = {
        {   0.270671, 0.039848, 0.017576, 0.016415, 0.014268,
            0.131916, 0.012391, 0.022599, 0.020358, 0.030727,
            0.015315, 0.048298, 0.053803, 0.020662, 0.023612,
            0.216147, 0.147226, 0.065438, 0.003758, 0.009621
        },
        {   0.021465, 0.010300, 0.011741, 0.010883, 0.385651,
            0.016416, 0.076196, 0.035329, 0.013921, 0.093517,
            0.022034, 0.028593, 0.013086, 0.023011, 0.018866,
            0.029156, 0.018153, 0.036100, 0.071770, 0.419641
        },
        {   0.561459, 0.045448, 0.438366, 0.764167, 0.087364,
            0.259114, 0.214940, 0.145928, 0.762204, 0.247320,
            0.118662, 0.441564, 0.174822, 0.530840, 0.465529,
            0.583402, 0.445586, 0.227050, 0.029510, 0.121090
        },
        {   0.070143, 0.011140, 0.019479, 0.094657, 0.013162,
            0.048038, 0.077000, 0.032939, 0.576639, 0.072293,
            0.028240, 0.080372, 0.037661, 0.185037, 0.506783,
            0.073732, 0.071587, 0.042532, 0.011254, 0.028723
        },
        {   0.041103, 0.014794, 0.005610, 0.010216, 0.153602,
            0.007797, 0.007175, 0.299635, 0.010849, 0.999446,
            0.210189, 0.006127, 0.013021, 0.019798, 0.014509,
            0.012049, 0.035799, 0.180085, 0.012744, 0.026466
        },
        {   0.115607, 0.037381, 0.012414, 0.018179, 0.051778,
            0.017255, 0.004911, 0.796882, 0.017074, 0.285858,
            0.075811, 0.014548, 0.015092, 0.011382, 0.012696,
            0.027535, 0.088333, 0.944340, 0.004373, 0.016741
        },
        {   0.093461, 0.004737, 0.387252, 0.347841, 0.010822,
            0.105877, 0.049776, 0.014963, 0.094276, 0.027761,
            0.010040, 0.187869, 0.050018, 0.110039, 0.038668,
            0.119471, 0.065802, 0.025430, 0.003215, 0.018742
        },
        {   0.452171, 0.114613, 0.062460, 0.115702, 0.284246,
            0.140204, 0.100358, 0.550230, 0.143995, 0.700649,
            0.276580, 0.118569, 0.097470, 0.126673, 0.143634,
            0.278983, 0.358482, 0.661750, 0.061533, 0.199373
        },
        {   0.005193, 0.004039, 0.006722, 0.006121, 0.003468,
            0.016931, 0.003647, 0.002184, 0.005019, 0.005990,
            0.001473, 0.004158, 0.009055, 0.003630, 0.006583,
            0.003172, 0.003690, 0.002967, 0.002772, 0.002686
        },
    };

    pri = P7AllocPrior ();
    pri->strategy = PRI_DCHLET;

    /* Transition priors are subjective, but borrowed from GJM's estimations
     * on Pfam
     */
    pri->tnum = 1;
    pri->tq[0] = 1.0;
    pri->t[0][TMM] = 0.7939;
    pri->t[0][TMI] = 0.0278;	/* Markus suggests: ~10x MD: ~0.036: test!  */
    pri->t[0][TMD] = 0.0135;	/* Markus suggests: ~0.1x MI: ~0.004 */
    pri->t[0][TIM] = 0.1551;
    pri->t[0][TII] = 0.1331;
    pri->t[0][TDM] = 0.9002;
    pri->t[0][TDD] = 0.5630;

    /* Match emission priors are a mixture Dirichlet,
     * from Kimmen Sjolander (Blocks9)
     */
    pri->mnum = 9;
    for (q = 0; q < pri->mnum; q++) {
        pri->mq[q] = defmq[q];
        for (x = 0; x < 20; x++) {
            pri->m[q][x] = defm[q][x];
        }
    }

    /* These insert emission priors are subjective. Observed frequencies
     * were obtained from PFAM 1.0, 10 Nov 96;
     *      see ~/projects/plan7/InsertStatistics.
     * Inserts are slightly biased towards polar residues and away from
     * hydrophobic residues.
     */
    pri->inum = 1;
    pri->iq[0] = 1.;
    pri->i[0][0] = 681.;		/* A */
    pri->i[0][1] = 120.;		/* C */
    pri->i[0][2] = 623.;		/* D */
    pri->i[0][3] = 651.;		/* E */
    pri->i[0][4] = 313.;		/* F */
    pri->i[0][5] = 902.;		/* G */
    pri->i[0][6] = 241.;		/* H */
    pri->i[0][7] = 371.;		/* I */
    pri->i[0][8] = 687.;		/* K */
    pri->i[0][9] = 676.;		/* L */
    pri->i[0][10] = 143.;		/* M */
    pri->i[0][11] = 548.;		/* N */
    pri->i[0][12] = 647.;		/* P */
    pri->i[0][13] = 415.;		/* Q */
    pri->i[0][14] = 551.;		/* R */
    pri->i[0][15] = 926.;		/* S */
    pri->i[0][16] = 623.;		/* T */
    pri->i[0][17] = 505.;		/* V */
    pri->i[0][18] = 102.;		/* W */
    pri->i[0][19] = 269.;		/* Y */

    return pri;
}


/* Function: default_nucleic_prior()
 *
 * Purpose:  Set the default DNA prior. (for now, almost a Laplace)
 */
static struct p7prior_s *
default_nucleic_prior (void)
{
    struct p7prior_s *pri;

    pri = P7AllocPrior ();
    pri->strategy = PRI_DCHLET;

    /* The use of the Pfam-trained amino acid transition priors
     * here is TOTALLY bogus. But it works better than a straight
     * Laplace, esp. for Maxmodelmaker(). For example, a Laplace
     * prior builds M=1 models for a single sequence GAATTC (at
     * one time an open "bug").
     */
    pri->tnum = 1;
    pri->tq[0] = 1.;
    pri->t[0][TMM] = 0.7939;
    pri->t[0][TMI] = 0.0278;
    pri->t[0][TMD] = 0.0135;
    pri->t[0][TIM] = 0.1551;
    pri->t[0][TII] = 0.1331;
    pri->t[0][TDM] = 0.9002;
    pri->t[0][TDD] = 0.5630;

    pri->mnum = 1;
    pri->mq[0] = 1.;
    FSet (pri->m[0], Alphabet_size, 1.);

    pri->inum = 1;
    pri->iq[0] = 1.;
    FSet (pri->i[0], Alphabet_size, 1.);

    return pri;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* revcomp.c
 *
 * Reverse complement of a IUPAC character string
 * CVS $Id: revcomp.c,v 1.6 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"

/* Function: revcomp()
 *
 * Purpose:  Reverse complement seq; store in comp.
 *           Can revcomp "in place" (revcomp(seq, seq)).
 *
 * Args:     comp  - destination for reverse complement of seq
 *           seq   - sequence to reverse complement
 *
 * Returns:  NULL on failure; or a (useless) pointer to comp.
 */
char *
revcomp (char *comp, char *seq)
{
    char *s;
    char c;

    if (comp == NULL) {
        return NULL;
    }
    if (seq == NULL) {
        return NULL;
    }

    StrReverse (comp, seq);
    for (s = comp; *s != '\0'; s++) {
        c = *s;
        c = sre_toupper (c);
        switch (c) {
        case 'A':
            c = 'T';
            break;
        case 'C':
            c = 'G';
            break;
        case 'G':
            c = 'C';
            break;
        case 'T':
            c = 'A';
            break;
        case 'U':
            c = 'A';
            break;
        case 'R':
            c = 'Y';
            break;
        case 'Y':
            c = 'R';
            break;
        case 'M':
            c = 'K';
            break;
        case 'K':
            c = 'M';
            break;
        case 'S':
            c = 'S';
            break;
        case 'W':
            c = 'W';
            break;
        case 'H':
            c = 'D';
            break;
        case 'D':
            c = 'H';
            break;
        case 'B':
            c = 'V';
            break;
        case 'V':
            c = 'B';
            break;
        default:
            break;			/* anything else? leave it; it's prob a gap or an X */
        }
        if (islower ((int) *s)) {
            c = (char) sre_tolower ((int) c);
        }
        *s = c;
    }
    return comp;
}

#ifdef REVCOMP_TESTDRIVER
/* gcc -g -DREVCOMP_TESTDRIVER revcomp.c sre_string.c shuffle.c sre_math.c sre_ctype.c sqerror.c -lm
*/
int
main (void)
{
    float p[4] = { 0.25, 0.25, 0.25, 0.25 };
    char *alphabet = "ACGT";
    int len = 10;
    char *seq;

    seq = RandomSequence (alphabet, p, 4, len);
    printf ("%s\n", seq);
    revcomp (seq, seq);
    printf ("%s\n", seq);
    free (seq);
    exit (0);
}
#endif
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* rk.c (originally from rnabob's patsearch.c)
 *
 * Contains a compiler and a search engine for Rabin-Karp
 * based primary sequence pattern searching on encoded
 * sequences.
 *
 * See Sedgewick, _Algorithms_, for a general discussion of
 * the Rabin-Karp algorithm. See the rkcomp or rkexec man
 * pages for specific details.
 *
 * CVS $Id: rk.c,v 1.3 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"		/* seq encoding utilities and typedefs */
#include "rk.h"


Hashseq
rkcomp (char *probe)
{   /* A,C,G,T/U, N probe string, 0-8 nt long */
    Hashseq hashprobe = 0;
    char coded[RK_HASHSIZE + 1];
    int len;
    int i;
    /* check bounds violation on probe */
    if ((len = strlen (probe)) > RK_HASHSIZE) {
        return 0;
    }
    /* encode the probe */
    if (seqencode (coded, probe) == 0) {
        return 0;
    }
    /* pack the probe into a Hashseq */
    for (i = 0; i < len; i++) {
        hashprobe <<= 4;
        hashprobe |= (Hashseq) coded[i];
    }
    /* left adjust as needed */
    for (; i < RK_HASHSIZE; i++) {
        hashprobe <<= 4;
        hashprobe |= (Hashseq) NTN;
    }
    /* return the compiled probe */
    return hashprobe;
}

int
rkseq (Hashseq hashprobe,	/* up to 8 nt packed into the probe */
       char *sequence)
{   /* encoded sequence                 */
    long i;
    long pos = 0;
    Hashseq target = 0;

    /* initialize the target hashseq */
    for (i = 0; i < RK_HASHSIZE; i++) {
        if (*(sequence + i) == NTEND) {
            break;
        }
        target <<= 4;
        target |= (Hashseq) (*(sequence + i));
    }

    while (*(sequence + pos + RK_HASHSIZE - 1) != NTEND) {
#ifdef DEBUG
        printf ("hashprobe: ");
        writehash (hashprobe);
        printf ("\ttarget: ");
        writehash (target);
        printf ("\nhashprobe & target: ");
        writehash (hashprobe & target);
        printf ("\n");
#endif
        if ((hashprobe & target) == target) {
            return ((int) pos);
        }
        target <<= 4;
        target |= (Hashseq) (*(sequence + pos + RK_HASHSIZE));
        pos++;
    }
    /* now we deal with an end effect */
    for (i = 0; i < RK_HASHSIZE; i++) {
        target |= (Hashseq) NTN;
        if ((hashprobe & target) == target) {
            return ((int) pos);
        }
        target <<= 4;
        pos++;
    }

    return (-1);
}


/* Debugging aids */
#ifdef DEBUG

static void
writehash (Hashseq hashseq)
{
    int idx;
    int sym;

    if (hashseq / 16) {
        writehash (hashseq / 16);
    }

    sym = (int) (hashseq % 16);
    if (sym == 0) {
        putchar ('-');
    }
    else {
        for (idx = 0; sym != iupac[idx].code && idx < IUPACSYMNUM; idx++) {
				};
        if (idx > IUPACSYMNUM) {
            printf ("(%d)", sym);
        }
        else {
            putchar (iupac[idx].sym);
        }
    }
}

#endif
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* selex.c
 *
 * SRE, Mon Jun 14 11:08:38 1999
 * SELEX  obsolete as the preferred HMMER/SQUID format
 * replaced by Stockholm format
 * selex support retained for backwards compatibility
 * kludged to use the MSA interface
 *
 * SRE, Mon Jan 30 14:41:49 1995:
 * #=SA side chain % surface accessibility annotation supported
 *
 * SRE, Tue Nov  9 17:40:50 1993:
 * major revision. #= special comments and aliinfo_s optional
 * alignment info support added. Support for #=CS (consensus
 * secondary structure), #=SS (individual secondary structure),
 * #=RF (reference coordinate system), #=SQ (per-sequence header info),
 * and #=AU ("author") added.
 *
 * Fri Dec  4 17:43:24 1992, SRE:
 * Reading and writing aligned sequences to/from disk files.
 * Implements a new, broader specification of SELEX format
 * and supercedes alignio.c.
 *
 * SELEX format is documented in Docs/formats.tex.
 ****************************************************************************
 * CVS $Id: selex.c,v 1.12 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include "squid.h"
#include "msa.h"

static int copy_alignment_line (char *aseq, int apos, int name_rcol,
                                char *buffer, int lcol, int rcol,
                                char gapsym);
static void actually_write_selex (FILE * fp, MSA * msa, int cpl);

static char commentsyms[] = "%#";

/* Function: ReadSELEX()
 * Date:     SRE, Sun Jun  6 18:24:09 1999 [St. Louis]
 *
 * Purpose:  Parse an alignment read from an open SELEX format
 *           alignment file. (SELEX is a single alignment format).
 *           Return the alignment, or NULL if we've already read the
 *           alignment or there's no alignment data in the file.
 *
 * Limitations: SELEX is the only remaining multipass parser for
 *           alignment files. It cannot read from gzip or from stdin.
 *           It Die()'s here if you try. The reason for this
 *           that SELEX allows space characters as gaps, so we don't
 *           know the borders of an alignment block until we've seen
 *           the whole block. I could rewrite to allow single-pass
 *           parsing (by storing the whole block in memory) but
 *           since SELEX is now legacy, why bother.
 *
 *           Note that the interface is totally kludged: fastest
 *           possible adaptation of old ReadSELEX() to the new
 *           MSA interface.
 *
 * Args:     afp  - open alignment file
 *
 * Returns:  MSA *  - an alignment object
 *                    caller responsible for an MSAFree()
 *           NULL if no alignment data.
 */
MSA *
ReadSELEX (MSAFILE * afp)
{
    MSA *msa;			/* RETURN: mult seq alignment   */
    FILE *fp;			/* ptr to opened seqfile        */
    char **aseqs;			/* aligned seqs                 */
    int num = 0;			/* number of seqs read          */
    char buffer[LINEBUFLEN];	/* input buffer for lines       */
    char bufcpy[LINEBUFLEN];	/* strtok'able copy of buffer   */
    struct block_struc
    {   /** alignment data for a block: */
        int lcol;			/* furthest left aligned sym    */
        int rcol;			/* furthest right aligned sym   */
    } *blocks = NULL;
    int blocknum;			/* number of blocks in file     */
    char *nptr;			/* ptr to start of name on line */
    char *sptr;			/* ptr into sequence on line    */
    int currnum;			/* num. seqs in given block     */
    int currblock;		/* index for blocks             */
    int i;			/* loop counter                 */
    int seqidx;			/* counter for seqs             */
    int alen;			/* length of alignment          */
    int warn_names;		/* becomes TRUE if names don't match between blocks */
    int headnum;			/* seqidx in per-sequence header info */
    int currlen;
    int count;
    int have_cs = 0;
    int have_rf = 0;
    AINFO base_ainfo, *ainfo;	/* hack: used to be passed ptr to AINFO */


    /* Convert from MSA interface to what old ReadSELEX() did:
     *     - copy our open fp, rather than opening file
     *     - verify that we're not reading a gzip or stdin
     */
    if (feof (afp->f)) {
        return NULL;
    }
    if (afp->do_gzip || afp->do_stdin)
        Die
        ("Can't read a SELEX format alignment from a pipe, stdin, or gzip'ed file");
    fp = afp->f;
    ainfo = &base_ainfo;

    /***************************************************
     * First pass across file.
     * Count seqs, get names, determine column info
     * Determine what sorts of info are active in this file.
     ***************************************************/

    InitAinfo (ainfo);
    /* get first line of the block
     * (non-comment, non-blank) */
    do {
        if (fgets (buffer, LINEBUFLEN, fp) == NULL) {
            squid_errno = SQERR_NODATA;
            return 0;
        }
        strcpy (bufcpy, buffer);
        if (*buffer == '#') {
            if (strncmp (buffer, "#=CS", 4) == 0) {
                have_cs = 1;
            }
            else if (strncmp (buffer, "#=RF", 4) == 0) {
                have_rf = 1;
            }
        }
    }
    while ((nptr = strtok (bufcpy, WHITESPACE)) == NULL ||
            (strchr (commentsyms, *nptr) != NULL));

    blocknum = 0;
    warn_names = FALSE;
    while (!feof (fp)) {
        /* allocate for info about this block. */
        if (blocknum == 0)
            blocks =
                (struct block_struc *) MallocOrDie (sizeof (struct block_struc));
        else
            blocks =
                (struct block_struc *) ReallocOrDie (blocks,
                        (blocknum +
                         1) *
                        sizeof (struct block_struc));
        blocks[blocknum].lcol = LINEBUFLEN + 1;
        blocks[blocknum].rcol = -1;

        currnum = 0;
        while (nptr != NULL) {	/* becomes NULL when this block ends. */
            /* First block only: save names */
            if (blocknum == 0) {
                if (currnum == 0) {
                    ainfo->sqinfo = (SQINFO *) MallocOrDie (sizeof (SQINFO));
                }
                else
                    ainfo->sqinfo =
                        (SQINFO *) ReallocOrDie (ainfo->sqinfo,
                                                 (currnum + 1) * sizeof (SQINFO));

                ainfo->sqinfo[currnum].flags = 0;
                SetSeqinfoString (&(ainfo->sqinfo[currnum]), nptr, SQINFO_NAME);
            }
            else {			/* in each additional block: check names */

                if (strcmp (ainfo->sqinfo[currnum].name, nptr) != 0) {
                    warn_names = TRUE;
                }
            }
            currnum++;

            /* check rcol, lcol */
            if ((sptr = strtok (NULL, WHITESPACE)) != NULL) {
                /* is this the furthest left we've
                   seen word 2 in this block? */
                if (sptr - bufcpy < blocks[blocknum].lcol) {
                    blocks[blocknum].lcol = sptr - bufcpy;
                }
                /* look for right side in buffer */
                for (sptr = buffer + strlen (buffer) - 1; strchr (WHITESPACE, *sptr) != NULL; sptr--) {
                }    /* do nothing */ ;
                if (sptr - buffer > blocks[blocknum].rcol) {
                    blocks[blocknum].rcol = sptr - buffer;
                }
            }

            /* get the next line; blank line means end of block */
            do {
                if (fgets (buffer, LINEBUFLEN, fp) == NULL) {
                    nptr = NULL;
                    break;
                }
                strcpy (bufcpy, buffer);

                if (strncmp (buffer, "#=SS", 4) == 0) {
                    ainfo->sqinfo[currnum - 1].flags |= SQINFO_SS;
                }
                else if (strncmp (buffer, "#=SA", 4) == 0) {
                    ainfo->sqinfo[currnum - 1].flags |= SQINFO_SA;
                }
                else if (strncmp (buffer, "#=CS", 4) == 0) {
                    have_cs = 1;
                }
                else if (strncmp (buffer, "#=RF", 4) == 0) {
                    have_rf = 1;
                }

                if ((nptr = strtok (bufcpy, WHITESPACE)) == NULL) {
                    break;
                }
            } while (strchr (commentsyms, *nptr) != NULL);
        }


        /* check that number of sequences matches expected */
        if (blocknum == 0) {
            num = currnum;
        }
        else if (currnum != num) {
            Die ("Parse error in ReadSELEX()");
        }
        blocknum++;

        /* get first line of next block
         * (non-comment, non-blank) */
        do {
            if (fgets (buffer, LINEBUFLEN, fp) == NULL) {
                nptr = NULL;
                break;
            }
            strcpy (bufcpy, buffer);
        }
        while ((nptr = strtok (bufcpy, WHITESPACE)) == NULL ||
                (strchr (commentsyms, *nptr) != NULL));
    }


    /***************************************************
     * Get ready for second pass:
     *   figure out the length of the alignment
     *   malloc space
     *   rewind the file
     ***************************************************/

    alen = 0;
    for (currblock = 0; currblock < blocknum; currblock++) {
        alen += blocks[currblock].rcol - blocks[currblock].lcol + 1;
    }

    rewind (fp);

    /* allocations. we can't use AllocateAlignment because of
     * the way we already used ainfo->sqinfo.
     */
    aseqs = (char **) MallocOrDie (num * sizeof (char *));
    if (have_cs) {
        ainfo->cs = (char *) MallocOrDie ((alen + 1) * sizeof (char));
    }
    if (have_rf) {
        ainfo->rf = (char *) MallocOrDie ((alen + 1) * sizeof (char));
    }



    for (i = 0; i < num; i++) {
        aseqs[i] = (char *) MallocOrDie ((alen + 1) * sizeof (char));
        if (ainfo->sqinfo[i].flags & SQINFO_SS) {
            ainfo->sqinfo[i].ss = (char *) MallocOrDie ((alen + 1) * sizeof (char));
        }
        if (ainfo->sqinfo[i].flags & SQINFO_SA) {
            ainfo->sqinfo[i].sa = (char *) MallocOrDie ((alen + 1) * sizeof (char));
        }
    }

    ainfo->alen = alen;
    ainfo->nseq = num;
    ainfo->wgt = (float *) MallocOrDie (sizeof (float) * num);
    FSet (ainfo->wgt, num, 1.0);

    /***************************************************
     * Second pass across file. Parse header; assemble sequences
     ***************************************************/
    /* We've now made a complete first pass over the file. We know how
     * many blocks it contains, we know the number of seqs in the first
     * block, and we know every block has the same number of blocks;
     * so we can be a bit more cavalier about error-checking as we
     * make the second pass.
     */

    /* Look for header
     */
    headnum = 0;
    for (;;) {
        if (fgets (buffer, LINEBUFLEN, fp) == NULL) {
            Die ("Parse error in ReadSELEX()");
        }
        strcpy (bufcpy, buffer);
        if ((nptr = strtok (bufcpy, WHITESPACE)) == NULL) {
            continue;    /* skip blank lines */
        }

        if (strcmp (nptr, "#=AU") == 0 && (sptr = strtok (NULL, "\n")) != NULL) {
            ainfo->au = Strdup (sptr);
        }
        else if (strcmp (nptr, "#=ID") == 0
                 && (sptr = strtok (NULL, "\n")) != NULL) {
            ainfo->name = Strdup (sptr);
        }
        else if (strcmp (nptr, "#=AC") == 0
                 && (sptr = strtok (NULL, "\n")) != NULL) {
            ainfo->acc = Strdup (sptr);
        }
        else if (strcmp (nptr, "#=DE") == 0
                 && (sptr = strtok (NULL, "\n")) != NULL) {
            ainfo->desc = Strdup (sptr);
        }
        else if (strcmp (nptr, "#=GA") == 0) {
            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=GA line in ReadSELEX()");
            }
            ainfo->ga1 = atof (sptr);

            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=GA line in ReadSELEX()");
            }
            ainfo->ga2 = atof (sptr);

            ainfo->flags |= AINFO_GA;
        }
        else if (strcmp (nptr, "#=TC") == 0) {
            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=TC line in ReadSELEX()");
            }
            ainfo->tc1 = atof (sptr);

            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=TC line in ReadSELEX()");
            }
            ainfo->tc2 = atof (sptr);

            ainfo->flags |= AINFO_TC;
        }
        else if (strcmp (nptr, "#=NC") == 0) {
            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=NC line in ReadSELEX()");
            }
            ainfo->nc1 = atof (sptr);

            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=NC line in ReadSELEX()");
            }
            ainfo->nc2 = atof (sptr);

            ainfo->flags |= AINFO_NC;
        }
        else if (strcmp (nptr, "#=SQ") == 0) {	/* per-sequence header info */
            /* first field is the name */
            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=SQ line in ReadSELEX()");
            }
            if (strcmp (sptr, ainfo->sqinfo[headnum].name) != 0) {
                warn_names = TRUE;
            }

            /* second field is the weight */
            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=SQ line in ReadSELEX()");
            }
            if (!IsReal (sptr))
                Die
                ("Parse error in #=SQ line in ReadSELEX(): weight is not a number");
            ainfo->wgt[headnum] = atof (sptr);

            /* third field is database source id */
            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=SQ line in ReadSELEX(): incomplete line");
            }
            SetSeqinfoString (&(ainfo->sqinfo[headnum]), sptr, SQINFO_ID);

            /* fourth field is database accession number */
            if ((sptr = strtok (NULL, WHITESPACE)) == NULL) {
                Die ("Parse error in #=SQ line in ReadSELEX(): incomplete line");
            }
            SetSeqinfoString (&(ainfo->sqinfo[headnum]), sptr, SQINFO_ACC);

            /* fifth field is start..stop::olen */
            if ((sptr = strtok (NULL, ".:")) == NULL) {
                Die ("Parse error in #=SQ line in ReadSELEX(): incomplete line");
            }
            SetSeqinfoString (&(ainfo->sqinfo[headnum]), sptr, SQINFO_START);

            if ((sptr = strtok (NULL, ".:")) == NULL) {
                Die ("Parse error in #=SQ line in ReadSELEX(): incomplete line");
            }
            SetSeqinfoString (&(ainfo->sqinfo[headnum]), sptr, SQINFO_STOP);

            if ((sptr = strtok (NULL, ":\t ")) == NULL) {
                Die ("Parse error in #=SQ line in ReadSELEX(): incomplete line");
            }
            SetSeqinfoString (&(ainfo->sqinfo[headnum]), sptr, SQINFO_OLEN);

            /* rest of line is optional description */
            if ((sptr = strtok (NULL, "\n")) != NULL) {
                SetSeqinfoString (&(ainfo->sqinfo[headnum]), sptr, SQINFO_DESC);
            }

            headnum++;
        }
        else if (strcmp (nptr, "#=CS") == 0) {
            break;
        }
        else if (strcmp (nptr, "#=RF") == 0) {
            break;
        }
        else if (strchr (commentsyms, *nptr) == NULL) {
            break;    /* non-comment, non-header */
        }
    }


    currlen = 0;
    for (currblock = 0; currblock < blocknum; currblock++) {
        /* parse the block */
        seqidx = 0;
        while (nptr != NULL) {
            /* Consensus structure */
            if (strcmp (nptr, "#=CS") == 0) {
                if (!copy_alignment_line (ainfo->cs, currlen, strlen (nptr) - 1,
                                          buffer, blocks[currblock].lcol,
                                          blocks[currblock].rcol, (char) '.')) {
                    Die ("Parse error in #=CS line in ReadSELEX()");
                }
            }

            /* Reference coordinates */
            else if (strcmp (nptr, "#=RF") == 0) {
                if (!copy_alignment_line (ainfo->rf, currlen, strlen (nptr) - 1,
                                          buffer, blocks[currblock].lcol,
                                          blocks[currblock].rcol, (char) '.')) {
                    Die ("Parse error in #=RF line in ReadSELEX()");
                }
            }
            /* Individual secondary structure */
            else if (strcmp (nptr, "#=SS") == 0) {
                if (!copy_alignment_line
                        (ainfo->sqinfo[seqidx - 1].ss, currlen, strlen (nptr) - 1, buffer,
                         blocks[currblock].lcol, blocks[currblock].rcol, (char) '.')) {
                    Die ("Parse error in #=SS line in ReadSELEX()");
                }
            }

            /* Side chain % surface accessibility code */
            else if (strcmp (nptr, "#=SA") == 0) {
                if (!copy_alignment_line
                        (ainfo->sqinfo[seqidx - 1].sa, currlen, strlen (nptr) - 1, buffer,
                         blocks[currblock].lcol, blocks[currblock].rcol, (char) '.')) {
                    Die ("Parse error in #=SA line in ReadSELEX()");
                }
            }
            /* Aligned sequence; avoid unparsed machine comments */
            else if (strncmp (nptr, "#=", 2) != 0) {
                if (!copy_alignment_line (aseqs[seqidx], currlen, strlen (nptr) - 1,
                                          buffer, blocks[currblock].lcol,
                                          blocks[currblock].rcol, (char) '.')) {
                    Die ("Parse error in alignment line in ReadSELEX()");
                }
                seqidx++;
            }

            /* get next line */
            for (;;) {
                nptr = NULL;
                if (fgets (buffer, LINEBUFLEN, fp) == NULL) {
                    break;    /* EOF */
                }
                strcpy (bufcpy, buffer);
                if ((nptr = strtok (bufcpy, WHITESPACE)) == NULL) {
                    break;    /* blank */
                }
                if (strncmp (buffer, "#=", 2) == 0) {
                    break;    /* machine comment */
                }
                if (strchr (commentsyms, *nptr) == NULL) {
                    break;    /* data */
                }
            }
        }				/* end of a block */

        currlen += blocks[currblock].rcol - blocks[currblock].lcol + 1;

        /* get line 1 of next block */
        for (;;) {
            if (fgets (buffer, LINEBUFLEN, fp) == NULL) {
                break;    /* no data */
            }
            strcpy (bufcpy, buffer);
            if ((nptr = strtok (bufcpy, WHITESPACE)) == NULL) {
                continue;    /* blank */
            }
            if (strncmp (buffer, "#=", 2) == 0) {
                break;    /* machine comment */
            }
            if (strchr (commentsyms, *nptr) == NULL) {
                break;    /* non-comment */
            }
        }
    }				/* end of the file */

    /* Lengths in sqinfo are for raw sequence (ungapped),
     * and SS, SA are 0..rlen-1 not 0..alen-1.
     * Only the seqs with structures come out of here with lengths set.
     */
    for (seqidx = 0; seqidx < num; seqidx++) {
        int apos, rpos;
        /* secondary structures */
        if (ainfo->sqinfo[seqidx].flags & SQINFO_SS) {
            for (apos = rpos = 0; apos < alen; apos++) {
                if (!isgap (aseqs[seqidx][apos])) {
                    ainfo->sqinfo[seqidx].ss[rpos] = ainfo->sqinfo[seqidx].ss[apos];
                    rpos++;
                }
						}		
            ainfo->sqinfo[seqidx].ss[rpos] = '\0';
        }
        /* Surface accessibility */
        if (ainfo->sqinfo[seqidx].flags & SQINFO_SA) {
            for (apos = rpos = 0; apos < alen; apos++) {
                if (!isgap (aseqs[seqidx][apos])) {
                    ainfo->sqinfo[seqidx].sa[rpos] = ainfo->sqinfo[seqidx].sa[apos];
                    rpos++;
                }
						}		
            ainfo->sqinfo[seqidx].sa[rpos] = '\0';
        }
    }

    /* NULL-terminate all the strings */
    if (ainfo->rf != NULL) {
        ainfo->rf[alen] = '\0';
    }
    if (ainfo->cs != NULL) {
        ainfo->cs[alen] = '\0';
    }
    for (seqidx = 0; seqidx < num; seqidx++) {
        aseqs[seqidx][alen] = '\0';
    }

    /* find raw sequence lengths for sqinfo */
    for (seqidx = 0; seqidx < num; seqidx++) {
        count = 0;
        for (sptr = aseqs[seqidx]; *sptr != '\0'; sptr++) {
            if (!isgap (*sptr)) {
                count++;
            }
				}		
        ainfo->sqinfo[seqidx].len = count;
        ainfo->sqinfo[seqidx].flags |= SQINFO_LEN;
    }


    /***************************************************
     * Garbage collection and return
     ***************************************************/
    free (blocks);
    if (warn_names)
        Warn ("sequences may be in different orders in blocks of %s?",
              afp->fname);

    /* Convert back to MSA structure. (Wasteful kludge.)
     */
    msa = MSAFromAINFO (aseqs, ainfo);
    MSAVerifyParse (msa);
    FreeAlignment (aseqs, ainfo);
    return msa;
}


/* Function: WriteSELEX()
 * Date:     SRE, Mon Jun 14 13:13:14 1999 [St. Louis]
 *
 * Purpose:  Write a SELEX file in multiblock format.
 *
 * Args:     fp  - file that's open for writing
 *           msa - multiple sequence alignment object
 *
 * Returns:  (void)
 */
void
WriteSELEX (FILE * fp, MSA * msa)
{
    actually_write_selex (fp, msa, 50);	/* 50 char per block */
}

/* Function: WriteSELEXOneBlock()
 * Date:     SRE, Mon Jun 14 13:14:56 1999 [St. Louis]
 *
 * Purpose:  Write a SELEX alignment file in Pfam's single-block
 *           format style. A wrapper for actually_write_selex().
 *
 * Args:     fp - file that's open for writing
 *           msa- alignment to write
 *
 * Returns:  (void)
 */
void
WriteSELEXOneBlock (FILE * fp, MSA * msa)
{
    actually_write_selex (fp, msa, msa->alen);	/* one big block */
}


/* Function: actually_write_selex()
 * Date:     SRE, Mon Jun 14 12:54:46 1999 [St. Louis]
 *
 * Purpose:  Write an alignment in SELEX format to an open
 *           file. This is the function that actually does
 *           the work. The API's WriteSELEX() and
 *           WriteSELEXOneBlock() are wrappers.
 *
 * Args:     fp  - file that's open for writing
 *           msa - alignment to write
 *           cpl - characters to write per line in alignment block
 *
 * Returns:  (void)
 */
static void
actually_write_selex (FILE * fp, MSA * msa, int cpl)
{
    int i;
    int len = 0;
    int namewidth;
    char *buf;
    int currpos;

    buf = malloc (sizeof (char) * (cpl + 101));	/* 100 chars allowed for name, etc. */

    /* Figure out how much space we need for name + markup
     * to keep the alignment in register, for easier human viewing --
     * even though Stockholm format doesn't care about visual
     * alignment.
     */
    namewidth = 0;
    for (i = 0; i < msa->nseq; i++) {
        if ((len = strlen (msa->sqname[i])) > namewidth) {
            namewidth = len;
        }
		}		
    if (namewidth < 6) {
        namewidth = 6;    /* minimum space for markup tags */
    }

    /* Free text comments
     */
    for (i = 0; i < msa->ncomment; i++) {
        fprintf (fp, "# %s\n", msa->comment[i]);
    }
    if (msa->ncomment > 0) {
        fprintf (fp, "\n");
    }

    /* Per-file annotation
     */
    if (msa->name != NULL) {
        fprintf (fp, "#=ID %s\n", msa->name);
    }
    if (msa->acc != NULL) {
        fprintf (fp, "#=AC %s\n", msa->acc);
    }
    if (msa->desc != NULL) {
        fprintf (fp, "#=DE %s\n", msa->desc);
    }
    if (msa->au != NULL) {
        fprintf (fp, "#=AU %s\n", msa->au);
    }

    /* Thresholds are hacky. Pfam has two. Rfam has one.
     */
    if (msa->cutoff_is_set[MSA_CUTOFF_GA1]
            && msa->cutoff_is_set[MSA_CUTOFF_GA2])
        fprintf (fp, "#=GA %.1f %.1f\n", msa->cutoff[MSA_CUTOFF_GA1],
                 msa->cutoff[MSA_CUTOFF_GA2]);
    else if (msa->cutoff_is_set[MSA_CUTOFF_GA1]) {
        fprintf (fp, "#=GA %.1f\n", msa->cutoff[MSA_CUTOFF_GA1]);
    }
    if (msa->cutoff_is_set[MSA_CUTOFF_NC1]
            && msa->cutoff_is_set[MSA_CUTOFF_NC2])
        fprintf (fp, "#=NC %.1f %.1f\n", msa->cutoff[MSA_CUTOFF_NC1],
                 msa->cutoff[MSA_CUTOFF_NC2]);
    else if (msa->cutoff_is_set[MSA_CUTOFF_NC1]) {
        fprintf (fp, "#=NC %.1f\n", msa->cutoff[MSA_CUTOFF_NC1]);
    }
    if (msa->cutoff_is_set[MSA_CUTOFF_TC1]
            && msa->cutoff_is_set[MSA_CUTOFF_TC2])
        fprintf (fp, "#=TC %.1f %.1f\n", msa->cutoff[MSA_CUTOFF_TC1],
                 msa->cutoff[MSA_CUTOFF_TC2]);
    else if (msa->cutoff_is_set[MSA_CUTOFF_TC1]) {
        fprintf (fp, "#=TC %.1f\n", msa->cutoff[MSA_CUTOFF_TC1]);
    }

    /* Per-sequence annotation
     */
    for (i = 0; i < msa->nseq; i++) {
        fprintf (fp, "#=SQ %-*.*s %6.4f %s %s %d..%d::%d %s\n", namewidth, namewidth, msa->sqname[i], msa->wgt[i], "-",	/* MSA has no ID field */
                 (msa->sqacc != NULL && msa->sqacc[i] != NULL) ? msa->sqacc[i] : "-", 0, 0, 0,	/* MSA has no start, stop, olen field */
                 (msa->sqdesc != NULL
                  && msa->sqdesc[i] != NULL) ? msa->sqdesc[i] : "-");
		}							
    fprintf (fp, "\n");

    /* Alignment section:
     */
    for (currpos = 0; currpos < msa->alen; currpos += cpl) {
        if (currpos > 0) {
            fprintf (fp, "\n");
        }

        if (msa->ss_cons != NULL) {
            strncpy (buf, msa->ss_cons + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "%-*.*s %s\n", namewidth, namewidth, "#=CS", buf);
        }
        if (msa->rf != NULL) {
            strncpy (buf, msa->rf + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "%-*.*s %s\n", namewidth, namewidth, "#=RF", buf);
        }
        for (i = 0; i < msa->nseq; i++) {
            strncpy (buf, msa->aseq[i] + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "%-*.*s %s\n", namewidth, namewidth, msa->sqname[i], buf);

            if (msa->ss != NULL && msa->ss[i] != NULL) {
                strncpy (buf, msa->ss[i] + currpos, cpl);
                buf[cpl] = '\0';
                fprintf (fp, "%-*.*s %s\n", namewidth, namewidth, "#=SS", buf);
            }
            if (msa->sa != NULL && msa->sa[i] != NULL) {
                strncpy (buf, msa->sa[i] + currpos, cpl);
                buf[cpl] = '\0';
                fprintf (fp, "%-*.*s %s\n", namewidth, namewidth, "#=SA", buf);
            }
        }
    }
    free (buf);
}


/* Function: copy_alignment_line()
 *
 * Purpose:  Given a line from an alignment file, and bounds lcol,rcol
 *           on what part of it may be sequence, save the alignment into
 *           aseq starting at position apos.
 *
 *           name_rcol is set to the rightmost column this aseqs's name
 *           occupies; if name_rcol >= lcol, we have a special case in
 *           which the name intrudes into the sequence zone.
 */
static int
copy_alignment_line (char *aseq, int apos, int name_rcol,
                     char *buffer, int lcol, int rcol, char gapsym)
{
    char *s1, *s2;
    int i;

    s1 = aseq + apos;
    s2 = buffer;			/* be careful that buffer doesn't end before lcol! */
    for (i = 0; i < lcol; i++) {
        if (*s2) {
            s2++;
        }
		}
    for (i = lcol; i <= rcol; i++) {
        if (*s2 == '\t') {
            Warn
            ("TAB characters will corrupt a SELEX alignment! Please remove them first.");
            return 0;
        }
        if (name_rcol >= i) {	/* name intrusion special case: pad left w/ gaps */
            *s1 = gapsym;
        }
        /* short buffer special case: pad right w/ gaps  */
        else if (*s2 == '\0' || *s2 == '\n') {
            *s1 = gapsym;
        }

        else if (*s2 == ' ') {	/* new: disallow spaces as gap symbols */
            *s1 = gapsym;
        }

        else {		/* normal case: copy buffer into aseq */
            *s1 = *s2;
        }

        s1++;
        if (*s2) {
            s2++;
        }
    }
    return 1;
}





/* Function: DealignAseqs()
 *
 * Given an array of (num) aligned sequences aseqs,
 * strip the gaps. Store the raw sequences in a new allocated array.
 *
 * Caller is responsible for free'ing the memory allocated to
 * rseqs.
 *
 * Returns 1 on success. Returns 0 and sets squid_errno on
 * failure.
 */
int
DealignAseqs (char **aseqs, int num, char ***ret_rseqs)
{
    char **rseqs;			/* de-aligned sequence array   */
    int idx;			/* counter for sequences       */
    int depos;			/* position counter for dealigned seq */
    int apos;			/* position counter for aligned seq */
    int seqlen;			/* length of aligned seq */

    /* alloc space */
    rseqs = (char **) MallocOrDie (num * sizeof (char *));
    /* main loop */
    for (idx = 0; idx < num; idx++) {
        seqlen = strlen (aseqs[idx]);
        /* alloc space */
        rseqs[idx] = (char *) MallocOrDie ((seqlen + 1) * sizeof (char));

        /* strip gaps */
        depos = 0;
        for (apos = 0; aseqs[idx][apos] != '\0'; apos++) {
            if (!isgap (aseqs[idx][apos])) {
                rseqs[idx][depos] = aseqs[idx][apos];
                depos++;
            }
				}		
        rseqs[idx][depos] = '\0';
    }
    *ret_rseqs = rseqs;
    return 1;
}


/* Function: IsSELEXFormat()
 *
 * Return TRUE if filename may be in SELEX format.
 *
 * Accuracy is sacrificed for speed; a TRUE return does
 * *not* guarantee that the file will pass the stricter
 * error-checking of ReadSELEX(). All it checks is that
 * the first 500 non-comment lines of a file are
 * blank, or if there's a second "word" on the line
 * it looks like sequence (i.e., it's not kOtherSeq).
 *
 * Returns TRUE or FALSE.
 */
int
IsSELEXFormat (char *filename)
{
    FILE *fp;			/* ptr to open sequence file */
    char buffer[LINEBUFLEN];
    char *sptr;			/* ptr to first word          */
    int linenum;


    if ((fp = fopen (filename, "r")) == NULL) {
        squid_errno = SQERR_NOFILE;
        return 0;
    }

    linenum = 0;
    while (linenum < 500 && fgets (buffer, LINEBUFLEN, fp) != NULL) {
        linenum++;
        /* dead giveaways for extended SELEX */
        if (strncmp (buffer, "#=AU", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=ID", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=AC", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=DE", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=GA", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=TC", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=NC", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=SQ", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=SS", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=CS", 4) == 0) {
            goto DONE_2;
        }
        else if (strncmp (buffer, "#=RF", 4) == 0) {
            goto DONE_2;
        }

        /* a comment? */
        if (strchr (commentsyms, *buffer) != NULL) {
            continue;
        }

        /* a blank line? */
        if ((sptr = strtok (buffer, WHITESPACE)) == NULL) {
            continue;
        }

        /* a one-word line (name only)
           is possible, though rare */
        if ((sptr = strtok (NULL, "\n")) == NULL) {
            continue;
        }

        if (Seqtype (sptr) == kOtherSeq) {
            fclose (fp);
            return 0;
        }
    }

DONE_2:
    fclose (fp);
    return 1;
}








/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* seqencode.c
 *
 * Routines for creating and manipulating encoded sequence strings.
 * CVS $Id: seqencode.c,v 1.4 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "squid.h"

/* seqcmp()
   returns 0 if s1 == s2
   mismatch number otherwise */
int
seqcmp (char *s1, char *s2, int allow)
{
    int mmat = 0;

    while ((*s1 != NTEND) && (*s2 != NTEND) && (mmat <= allow)) {
        if (!(ntmatch (*s1, *s2))) {
            mmat++;
        };
        s1++;
        s2++;
    }
    while ((*s1++ != NTEND) && (mmat <= allow)) {
        mmat++;
    }
    return (mmat);
}

/* seqncmp()
   same as seqcmp but it looks at,
   at most, n positions */
int
seqncmp (char *s1, char *s2, int n, int allow)
{
    int mmat = 0;

    while ((*s2 != NTEND) && (n-- != 0)) {
        if ((!(ntmatch (*s1, *s2))) && (++mmat > allow)) {
            return (mmat);
        }
        s1++;
        s2++;
    }
    while ((n-- != 0) && (*s1++ != NTEND) && (mmat <= allow)) {
        mmat++;
    }
    return (mmat);
}

/* seqencode()
   given a character text string str (A,C,G,T),
   convert to an encoded seq string;
   return 1 for success, 0 if fail */
int
seqencode (char *codeseq,	/* pre-allocated space for answer */
           char *str)
{   /* character string to convert    */
    char *ptr;
    int idx;

    ptr = codeseq;
    while (*str != '\0') {
        if (islower ((int) (*str))) {
            *str = (char) toupper ((int) (*str));
        }
        for (idx = 0; *str != iupac[idx].sym && idx <= IUPACSYMNUM; idx++) {
				};
        if (idx > IUPACSYMNUM) {
            *ptr = (char) NTEND;
            return 0;
        }
        else {
            *ptr = iupac[idx].code;
        }
        ptr++;
        str++;
    }
    *ptr = NTEND;
    return 1;
}


int
coded_revcomp (char *comp, char *seq)
{
    long bases;
    char *bckp, *fwdp;
    int idx;
    long pos;

    bases = strlen (seq);

    fwdp = comp;
    bckp = seq + bases - 1;
    for (pos = 0; pos < bases; pos++) {
        for (idx = 0; *bckp != iupac[idx].code && idx < IUPACSYMNUM; idx++) {
				};
        if (idx > IUPACSYMNUM) {
            *fwdp = NTEND;
            return 0;
        }
        else {
            *fwdp = iupac[idx].comp;
        }
        fwdp++;
        bckp--;
    }
    *fwdp = NTEND;
    return (1);
}

int
seqdecode (char *str, char *codeseq)
{
    int idx;
    int pos;

    pos = 0;
    while (*codeseq != NTEND) {
        for (idx = 0; *codeseq != iupac[idx].code && idx < IUPACSYMNUM; idx++) {
				};
        if (idx > IUPACSYMNUM) {
            str[pos] = 'X';
            return 0;
        }
        else {
            str[pos] = iupac[idx].sym;
        }
        codeseq++;
        pos++;
    }
    str[pos] = '\0';
    return 1;
}

int
seqndecode (char *str,		/* pre-allocated string to write into */
            char *codeseq,	/* sequence to decode */
            int n)
{   /* how many bases to decode */
    int idx;
    int pos = 0;

    while (--n >= 0) {
        for (idx = 0; *codeseq != iupac[idx].code && idx < IUPACSYMNUM; idx++) {
				};
        if (idx > IUPACSYMNUM) {
            str[pos] = 'X';
            return 0;
        }
        else {
            str[pos] = iupac[idx].sym;
        }
        codeseq++;
        pos++;
    }
    str[pos] = '\0';
    return 1;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* shuffle.c
 *
 * Routines for randomizing sequences.
 *
 * All routines are alphabet-independent (DNA, protein, RNA, whatever);
 * they assume that input strings are purely alphabetical [a-zA-Z], and
 * will return strings in all upper case [A-Z].
 *
 * All return 1 on success, and 0 on failure; 0 status invariably
 * means the input string was not alphabetical.
 *
 * StrShuffle()   - shuffled string, preserve mono-symbol composition.
 * StrDPShuffle() - shuffled string, preserve mono- and di-symbol composition.
 *
 * StrMarkov0()   - random string, same zeroth order Markov properties.
 * StrMarkov1()   - random string, same first order Markov properties.
 *
 * StrReverse()   - simple reversal of string
 * StrRegionalShuffle() -  mono-symbol shuffled string in regional windows
 *
 * There are also similar routines for shuffling alignments:
 *
 * AlignmentShuffle()   - alignment version of StrShuffle().
 * AlignmentBootstrap() - sample with replacement; a bootstrap dataset.
 * QRNAShuffle()        - shuffle a pairwise alignment, preserving all gap positions.
 *
 * CVS $Id: shuffle.c,v 1.8 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <string.h>
#include <ctype.h>

#include "squid.h"
#include "sre_random.h"

/* Function: StrShuffle()
 *
 * Purpose:  Returns a shuffled version of s2, in s1.
 *           (s1 and s2 can be identical, to shuffle in place.)
 *
 * Args:     s1 - allocated space for shuffled string.
 *           s2 - string to shuffle.
 *
 * Return:   1 on success.
 */
int
StrShuffle (char *s1, char *s2)
{
    int len;
    int pos;
    char c;

    if (s1 != s2) {
        strcpy (s1, s2);
    }
    for (len = strlen (s1); len > 1; len--) {
        pos = CHOOSE (len);
        c = s1[pos];
        s1[pos] = s1[len - 1];
        s1[len - 1] = c;
    }
    return 1;
}

/* Function: StrDPShuffle()
 * Date:     SRE, Fri Oct 29 09:15:17 1999 [St. Louis]
 *
 * Purpose:  Returns a shuffled version of s2, in s1.
 *           (s1 and s2 may be identical; i.e. a string
 *           may be shuffled in place.) The shuffle is a
 *           "doublet-preserving" (DP) shuffle. Both
 *           mono- and di-symbol composition are preserved.
 *
 *           Done by searching for a random Eulerian
 *           walk on a directed multigraph.
 *           Reference: S.F. Altschul and B.W. Erickson, Mol. Biol.
 *           Evol. 2:526-538, 1985. Quoted bits in my comments
 *           are from Altschul's outline of the algorithm.
 *
 * Args:     s1   - RETURN: the string after it's been shuffled
 *                    (space for s1 allocated by caller)
 *           s2   - the string to be shuffled
 *
 * Returns:  0 if string can't be shuffled (it's not all [a-zA-z]
 *             alphabetic.
 *           1 on success.
 */
int
StrDPShuffle (char *s1, char *s2)
{
    int len;
    int pos;			/* a position in s1 or s2 */
    int x, y;			/* indices of two characters */
    char **E;			/* edge lists: E[0] is the edge list from vertex A */
    int *nE;			/* lengths of edge lists */
    int *iE;			/* positions in edge lists */
    int n;			/* tmp: remaining length of an edge list to be shuffled */
    char sf;			/* last character in s2 */
    char Z[26];			/* connectivity in last edge graph Z */
    int keep_connecting;		/* flag used in Z connectivity algorithm */
    int is_eulerian;		/* flag used for when we've got a good Z */

    /* First, verify that the string is entirely alphabetic.
     */
    len = strlen (s2);
    for (pos = 0; pos < len; pos++) {
        if (!isalpha ((int) s2[pos])) {
            return 0;
        }
		}		

    /* "(1) Construct the doublet graph G and edge ordering E
     *      corresponding to S."
     *
     * Note that these also imply the graph G; and note,
     * for any list x with nE[x] = 0, vertex x is not part
     * of G.
     */
    E = MallocOrDie (sizeof (char *) * 26);
    nE = MallocOrDie (sizeof (int) * 26);
    for (x = 0; x < 26; x++) {
        E[x] = MallocOrDie (sizeof (char) * (len - 1));
        nE[x] = 0;
    }

    x = toupper ((int) s2[0]) - 'A';
    for (pos = 1; pos < len; pos++) {
        y = toupper ((int) s2[pos]) - 'A';
        E[x][nE[x]] = y;
        nE[x]++;
        x = y;
    }

    /* Now we have to find a random Eulerian edge ordering.
     */
    sf = toupper ((int) s2[len - 1]) - 'A';
    is_eulerian = 0;
    while (!is_eulerian) {
        /* "(2) For each vertex s in G except s_f, randomly select
         *      one edge from the s edge list of E(S) to be the
         *      last edge of the s list in a new edge ordering."
         *
         * select random edges and move them to the end of each
         * edge list.
         */
        for (x = 0; x < 26; x++) {
            if (nE[x] == 0 || x == sf) {
                continue;
            }

            pos = CHOOSE (nE[x]);
            y = E[x][pos];
            E[x][pos] = E[x][nE[x] - 1];
            E[x][nE[x] - 1] = y;
        }

        /* "(3) From this last set of edges, construct the last-edge
         *      graph Z and determine whether or not all of its
         *      vertices are connected to s_f."
         *
         * a probably stupid algorithm for looking at the
         * connectivity in Z: iteratively sweep through the
         * edges in Z, and build up an array (confusing called Z[x])
         * whose elements are 1 if x is connected to sf, else 0.
         */
        for (x = 0; x < 26; x++) {
            Z[x] = 0;
        }
        Z[(int) sf] = keep_connecting = 1;

        while (keep_connecting) {
            keep_connecting = 0;
            for (x = 0; x < 26; x++) {
                y = E[x][nE[x] - 1];	/* xy is an edge in Z */
                if (Z[x] == 0 && Z[y] == 1) {	/* x is connected to sf in Z */
                    Z[x] = 1;
                    keep_connecting = 1;
                }
            }
        }

        /* if any vertex in Z is tagged with a 0, it's
         * not connected to sf, and we won't have a Eulerian
         * walk.
         */
        is_eulerian = 1;
        for (x = 0; x < 26; x++) {
            if (nE[x] == 0 || x == sf) {
                continue;
            }
            if (Z[x] == 0) {
                is_eulerian = 0;
                break;
            }
        }

        /* "(4) If any vertex is not connected in Z to s_f, the
         *      new edge ordering will not be Eulerian, so return to
         *      (2). If all vertices are connected in Z to s_f,
         *      the new edge ordering will be Eulerian, so
         *      continue to (5)."
         *
         * e.g. note infinite loop while is_eulerian is FALSE.
         */
    }

    /* "(5) For each vertex s in G, randomly permute the remaining
     *      edges of the s edge list of E(S) to generate the s
     *      edge list of the new edge ordering E(S')."
     *
     * Essentially a StrShuffle() on the remaining nE[x]-1 elements
     * of each edge list; unfortunately our edge lists are arrays,
     * not strings, so we can't just call out to StrShuffle().
     */
    for (x = 0; x < 26; x++) {
        for (n = nE[x] - 1; n > 1; n--) {
            pos = CHOOSE (n);
            y = E[x][pos];
            E[x][pos] = E[x][n - 1];
            E[x][n - 1] = y;
        }
		}
    /* "(6) Construct sequence S', a random DP permutation of
     *      S, from E(S') as follows. Start at the s_1 edge list.
     *      At each s_i edge list, add s_i to S', delete the
     *      first edge s_i,s_j of the edge list, and move to
     *      the s_j edge list. Continue this process until
     *      all edge lists are exhausted."
     */
    iE = MallocOrDie (sizeof (int) * 26);
    for (x = 0; x < 26; x++) {
        iE[x] = 0;
    }

    pos = 0;
    x = toupper ((int) s2[0]) - 'A';
    while (1) {
        s1[pos++] = 'A' + x;	/* add s_i to S' */

        y = E[x][iE[x]];
        iE[x]++;			/* "delete" s_i,s_j from edge list */

        x = y;			/* move to s_j edge list. */

        if (iE[x] == nE[x]) {
            break;    /* the edge list is exhausted. */
        }
    }
    s1[pos++] = 'A' + sf;
    s1[pos] = '\0';

    /* Reality checks.
     */
    if (x != sf) {
        Die ("hey, you didn't end on s_f.");
    }
    if (pos != len) {
        Die ("hey, pos (%d) != len (%d).", pos, len);
    }

    /* Free and return.
     */
    Free2DArray ((void **) E, 26);
    free (nE);
    free (iE);
    return 1;
}


/* Function: StrMarkov0()
 * Date:     SRE, Fri Oct 29 11:08:31 1999 [St. Louis]
 *
 * Purpose:  Returns a random string s1 with the same
 *           length and zero-th order Markov properties
 *           as s2.
 *
 *           s1 and s2 may be identical, to randomize s2
 *           in place.
 *
 * Args:     s1 - allocated space for random string
 *           s2 - string to base s1's properties on.
 *
 * Returns:  1 on success; 0 if s2 doesn't look alphabetical.
 */
int
StrMarkov0 (char *s1, char *s2)
{
    int len;
    int pos;
    float p[26];			/* symbol probabilities */

    /* First, verify that the string is entirely alphabetic.
     */
    len = strlen (s2);
    for (pos = 0; pos < len; pos++) {
        if (!isalpha ((int) s2[pos])) {
            return 0;
        }
		}
    /* Collect zeroth order counts and convert to frequencies.
     */
    FSet (p, 26, 0.);
    for (pos = 0; pos < len; pos++) {
        p[(int) (toupper ((int) s2[pos]) - 'A')] += 1.0;
    }
    FNorm (p, 26);

    /* Generate a random string using those p's.
     */
    for (pos = 0; pos < len; pos++) {
        s1[pos] = FChoose (p, 26) + 'A';
    }
    s1[pos] = '\0';

    return 1;
}


/* Function: StrMarkov1()
 * Date:     SRE, Fri Oct 29 11:22:20 1999 [St. Louis]
 *
 * Purpose:  Returns a random string s1 with the same
 *           length and first order Markov properties
 *           as s2.
 *
 *           s1 and s2 may be identical, to randomize s2
 *           in place.
 *
 * Args:     s1 - allocated space for random string
 *           s2 - string to base s1's properties on.
 *
 * Returns:  1 on success; 0 if s2 doesn't look alphabetical.
 */
int
StrMarkov1 (char *s1, char *s2)
{
    int len;
    int pos;
    int x, y;
    int i;			/* initial symbol */
    float p[26][26];		/* symbol probabilities */

    /* First, verify that the string is entirely alphabetic.
     */
    len = strlen (s2);
    for (pos = 0; pos < len; pos++) {
        if (!isalpha ((int) s2[pos])) {
            return 0;
        }
		}
    /* Collect first order counts and convert to frequencies.
     */
    for (x = 0; x < 26; x++) {
        FSet (p[x], 26, 0.);
    }

    i = x = toupper ((int) s2[0]) - 'A';
    for (pos = 1; pos < len; pos++) {
        y = toupper ((int) s2[pos]) - 'A';
        p[x][y] += 1.0;
        x = y;
    }
    for (x = 0; x < 26; x++) {
        FNorm (p[x], 26);
    }

    /* Generate a random string using those p's.
     */
    x = i;
    s1[0] = x + 'A';
    for (pos = 1; pos < len; pos++) {
        y = FChoose (p[x], 26);
        s1[pos] = y + 'A';
        x = y;
    }
    s1[pos] = '\0';

    return 1;
}



/* Function: StrReverse()
 * Date:     SRE, Thu Nov 20 10:54:52 1997 [St. Louis]
 *
 * Purpose:  Returns a reversed version of s2, in s1.
 *           (s1 and s2 can be identical, to reverse in place)
 *
 * Args:     s1 - allocated space for reversed string.
 *           s2 - string to reverse.
 *
 * Return:   1.
 */
int
StrReverse (char *s1, char *s2)
{
    int len;
    int pos;
    char c;

    len = strlen (s2);
    for (pos = 0; pos < len / 2; pos++) {	
				/* swap ends */
        c = s2[len - pos - 1];
        s1[len - pos - 1] = s2[pos];
        s1[pos] = c;
    }
    if (len % 2) {
        s1[pos] = s2[pos];
    }				/* copy middle residue in odd-len s2 */
    s1[len] = '\0';
    return 1;
}

/* Function: StrRegionalShuffle()
 * Date:     SRE, Thu Nov 20 11:02:34 1997 [St. Louis]
 *
 * Purpose:  Returns a regionally shuffled version of s2, in s1.
 *           (s1 and s2 can be identical to regionally
 *           shuffle in place.) See [Pearson88].
 *
 * Args:     s1 - allocated space for regionally shuffled string.
 *           s2 - string to regionally shuffle
 *           w  - window size (typically 10 or 20)
 *
 * Return:   1.
 */
int
StrRegionalShuffle (char *s1, char *s2, int w)
{
    int len;
    char c;
    int pos;
    int i, j;

    if (s1 != s2) {
        strcpy (s1, s2);
    }
    len = strlen (s1);

    for (i = 0; i < len; i += w) {
        for (j = MIN (len - 1, i + w - 1); j > i; j--) {
            pos = i + CHOOSE (j - i);
            c = s1[pos];
            s1[pos] = s1[j];
            s1[j] = c;
        }
		}		
    return 1;
}


/* Function: AlignmentShuffle()
 * Date:     SRE, Sun Apr 22 18:37:15 2001 [St. Louis]
 *
 * Purpose:  Returns a shuffled version of ali2, in ali1.
 *           (ali1 and ali2 can be identical, to shuffle
 *           in place.) The alignment columns are shuffled,
 *           preserving % identity within the columns.
 *
 * Args:     ali1 - allocated space for shuffled alignment
 *                  [0..nseq-1][0..alen-1]
 *           ali2 - alignment to be shuffled
 *           nseq - number of sequences in the alignment
 *           alen - length of alignment, in columns.
 *
 * Returns:  int
 */
int
AlignmentShuffle (char **ali1, char **ali2, int nseq, int alen)
{
    int i;
    int pos;
    char c;

    if (ali1 != ali2) {
        for (i = 0; i < nseq; i++) {
            strcpy (ali1[i], ali2[i]);
        }
    }

    for (i = 0; i < nseq; i++) {
        ali1[i][alen] = '\0';
    }

    for (; alen > 1; alen--) {
        pos = CHOOSE (alen);
        for (i = 0; i < nseq; i++) {
            c = ali1[i][pos];
            ali1[i][pos] = ali1[i][alen - 1];
            ali1[i][alen - 1] = c;
        }
    }

    return 1;
}

/* Function: AlignmentBootstrap()
 * Date:     SRE, Sun Apr 22 18:49:14 2001 [St. Louis]
 *
 * Purpose:  Returns a bootstrapped alignment sample in ali1,
 *           constructed from ali2 by sampling columns with
 *           replacement.
 *
 *           Unlike the other shuffling routines, ali1 and
 *           ali2 cannot be the same. ali2 is left unchanged.
 *           ali1 must be a properly allocated space for an
 *           alignment the same size as ali2.
 *
 * Args:     ali1 - allocated space for bootstrapped alignment
 *                  [0..nseq-1][0..alen-1]
 *           ali2 - alignment to be bootstrapped
 *           nseq - number of sequences in the alignment
 *           alen - length of alignment, in columns.
 *
 * Returns:  1 on success.
 */
int
AlignmentBootstrap (char **ali1, char **ali2, int nseq, int alen)
{
    int pos;
    int col;
    int i;

    for (pos = 0; pos < alen; pos++) {
        col = CHOOSE (alen);
        for (i = 0; i < nseq; i++) {
            ali1[i][pos] = ali2[i][col];
        }
    }
    for (i = 0; i < nseq; i++) {
        ali1[i][alen] = '\0';
    }

    return 1;
}


/* Function: QRNAShuffle()
 * Date:     SRE, Mon Dec 10 10:14:12 2001 [St. Louis]
 *
 * Purpose:  Shuffle a pairwise alignment x,y while preserving the
 *           position of gaps; return the shuffled alignment in xs,
 *           ys.
 *
 *           Works by doing three separate
 *           shuffles, of (1) columns with residues in both
 *           x and y, (2) columns with residue in x and gap in y,
 *           and (3) columns with gap in x and residue in y.
 *
 *           xs,x and ys,y may be identical: that is, to shuffle
 *           an alignment "in place", destroying the original
 *           alignment, just call:
 *              QRNAShuffle(x,y,x,y);
 *
 * Args:     xs, ys: allocated space for shuffled pairwise ali of x,y [L+1]
 *           x, y: pairwise alignment to be shuffled [0..L-1]
 *
 * Returns:  1 on success, 0 on failure.
 *           The shuffled alignment is returned in xs, ys.
 */
int
QRNAShuffle (char *xs, char *ys, char *x, char *y)
{
    int L;
    int *xycol, *xcol, *ycol;
    int nxy, nx, ny;
    int i;
    int pos, c;
    char xsym, ysym;

    if (xs != x) {
        strcpy (xs, x);
    }
    if (ys != y) {
        strcpy (ys, y);
    }

    /* First, construct three arrays containing lists of the column positions
     * of the three types of columns. (If a column contains gaps in both x and y,
     * we've already simply copied it to the shuffled sequence.)
     */
    L = strlen (x);
    xycol = MallocOrDie (sizeof (int) * L);
    xcol = MallocOrDie (sizeof (int) * L);
    ycol = MallocOrDie (sizeof (int) * L);
    nxy = nx = ny = 0;

    for (i = 0; i < L; i++) {
        if (isgap (x[i]) && isgap (y[i])) {
            continue;
        }
        else if (!isgap (x[i]) && !isgap (y[i])) {
            xycol[nxy] = i;
            nxy++;
        }
        else if (isgap (x[i])) {
            ycol[ny] = i;
            ny++;
        }
        else if (isgap (y[i])) {
            xcol[nx] = i;
            nx++;
        }
    }

    /* Second, shuffle the sequences indirectly, via shuffling these arrays.
     * Yow, careful with those indices, and with order of the statements...
     */
    for (; nxy > 1; nxy--) {
        pos = CHOOSE (nxy);
        xsym = xs[xycol[pos]];
        ysym = ys[xycol[pos]];
        c = xycol[pos];
        xs[xycol[pos]] = xs[xycol[nxy - 1]];
        ys[xycol[pos]] = ys[xycol[nxy - 1]];
        xycol[pos] = xycol[nxy - 1];
        xs[xycol[nxy - 1]] = xsym;
        ys[xycol[nxy - 1]] = ysym;
        xycol[pos] = xycol[nxy - 1];
    }
    for (; nx > 1; nx--) {
        pos = CHOOSE (nx);
        xsym = xs[xcol[pos]];
        ysym = ys[xcol[pos]];
        c = xcol[pos];
        xs[xcol[pos]] = xs[xcol[nx - 1]];
        ys[xcol[pos]] = ys[xcol[nx - 1]];
        xcol[pos] = xcol[nx - 1];
        xs[xcol[nx - 1]] = xsym;
        ys[xcol[nx - 1]] = ysym;
        xcol[nx - 1] = c;
    }
    for (; ny > 1; ny--) {
        pos = CHOOSE (ny);
        xsym = xs[ycol[pos]];
        ysym = ys[ycol[pos]];
        c = ycol[pos];
        xs[ycol[pos]] = xs[ycol[ny - 1]];
        ys[ycol[pos]] = ys[ycol[ny - 1]];
        ycol[pos] = ycol[ny - 1];
        xs[ycol[ny - 1]] = xsym;
        ys[ycol[ny - 1]] = ysym;
        ycol[ny - 1] = c;
    }

    free (xycol);
    free (xcol);
    free (ycol);
    return 1;
}


#ifdef TESTDRIVER
/*
 * cc -g -o testdriver -DTESTDRIVER -L. shuffle.c -lsquid -lm
 */
int
main (int argc, char **argv)
{
    char s1[100];
    char s2[100];

    sre_srandom (42);
    strcpy (s2, "GGGGGGGGGGCCCCCCCCCC");
    /*  strcpy(s2, "AGACATAAAGTTCCGTACTGCCGGGAT");
     */
    StrDPShuffle (s1, s2);
    printf ("DPshuffle: %s\n", s1);
    StrMarkov0 (s1, s2);
    printf ("Markov 0 : %s\n", s1);
    StrMarkov1 (s1, s2);
    printf ("Markov 1 : %s\n", s1);

    strcpy (s1, "ACGTACGT--------ACGTACGT----ACGTACGT");
    strcpy (s2, "ACGTACGTACGTACGT------------ACGTACGT");
    QRNAShuffle (s1, s2, s1, s2);
    printf ("QRNA : %s\n", s1);
    printf ("     : %s\n", s2);

    return 0;
}
#endif
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* sqerror.c
 *
 * error handling for the squid library
 * CVS $Id: sqerror.c,v 1.6 2003/05/26 16:21:50 eddy Exp $
 */

#include "squidconf.h"
#include "squid.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

int squid_errno;		/* a global errno equivalent */


/* Function: Die()
 *
 * Purpose:  Print an error message and die. The arguments
 *           are formatted exactly like arguments to printf().
 *
 * Return:   None. Exits the program.
 */
/* VARARGS0 */
void
Die (char *format, ...)
{
    va_list argp;
    /* format the error mesg */
    fprintf (stderr, "\nFATAL: ");
    va_start (argp, format);
    vfprintf (stderr, format, argp);
    va_end (argp);
    fprintf (stderr, "\n");
    fflush (stderr);
    /* exit  */
    exit (1);
}



/* Function: Warn()
 *
 * Purpose:  Print an error message and return. The arguments
 *           are formatted exactly like arguments to printf().
 *
 * Return:   (void)
 */
/* VARARGS0 */
void
Warn (char *format, ...)
{
    va_list argp;
    /* format the error mesg */
    fprintf (stderr, "WARNING: ");
    va_start (argp, format);
    vfprintf (stderr, format, argp);
    va_end (argp);
    fprintf (stderr, "\n");
    fflush (stderr);
}

/* Function: Panic()
 *
 * Purpose:  Die from a lethal error that's not my problem,
 *           but instead a failure of a StdC/POSIX call that
 *           shouldn't fail. Call perror() to get the
 *           errno flag, then die.
 *
 *           Usually called by the PANIC macro which adds
 *           the __FILE__ and __LINE__ information; see
 *           structs.h.
 *
 *           Inspired by code in Donald Lewine's book, _POSIX
 *           Programmer's Guide_.
 */
void
Panic (char *file, int line)
{
    (void) fprintf (stderr, "\nPANIC [%s line %d] ", file, line);
    (void) perror ("Unusual error");
    exit (EXIT_FAILURE);
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* File: sqio.c
 * From: ureadseq.c in Don Gilbert's sequence i/o package
 *
 * Reads and writes nucleic/protein sequence in various
 * formats. Data files may have multiple sequences.
 *
 * Heavily modified from READSEQ package
 * Copyright (C) 1990 by D.G. Gilbert
 * Biology Dept., Indiana University, Bloomington, IN 47405
 * email: gilbertd@bio.indiana.edu
 * Thanks Don!
 *
 * SRE: Modifications as noted. Fri Jul  3 09:44:54 1992
 *      Packaged for squid, Thu Oct  1 10:07:11 1992
 *      ANSI conversion in full swing, Mon Jul 12 12:22:21 1993
 *
 * CVS $Id: sqio.c,v 1.30 2003/04/14 16:00:16 eddy Exp $
 *
 *****************************************************************
 * Basic API for single sequence reading:
 *
 * SQFILE *sqfp;
 * char   *seqfile;
 * int     format;        - see squid.h for formats; example: SQFILE_FASTA
 * char   *seq;
 * SQINFO  sqinfo;
 *
 * if ((sqfp = SeqfileOpen(seqfile, format, "BLASTDB")) == NULL)
 *   Die("Failed to open sequence database file %s\n%s\n", seqfile, usage);
 * while (ReadSeq(sqfp, sqfp->format, &seq, &sqinfo)) {
 *   do_stuff;
 *   FreeSequence(seq, &sqinfo);
 * }
 * SeqfileClose(sqfp);
 *
 *****************************************************************
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifndef SEEK_SET
#include <unistd.h>
#endif

#include "squid.h"
#include "msa.h"
#include "ssi.h"

static void SeqfileGetLine (SQFILE * V);

#define kStartLength  500

static char *aminos = "ABCDEFGHIKLMNPQRSTVWXYZ*";
static char *primenuc = "ACGTUN";
static char *protonly = "EFIPQZ";

static SQFILE *seqfile_open (char *filename, int format, char *env,
                             int ssimode);

/* Function: SeqfileOpen()
 *
 * Purpose : Open a sequence database file and prepare for reading
 *           sequentially.
 *
 * Args:     filename - name of file to open
 *           format   - format of file
 *           env      - environment variable for path (e.g. BLASTDB)
 *           ssimode  - -1, SSI_OFFSET_I32, or SSI_OFFSET_I64
 *
 *           Returns opened SQFILE ptr, or NULL on failure.
 */
SQFILE *
SeqfileOpen (char *filename, int format, char *env)
{
    return seqfile_open (filename, format, env, -1);
}

SQFILE *
SeqfileOpenForIndexing (char *filename, int format, char *env, int ssimode)
{
    return seqfile_open (filename, format, env, ssimode);
}

static SQFILE *
seqfile_open (char *filename, int format, char *env, int ssimode)
{
    SQFILE *dbfp;

    dbfp = (SQFILE *) MallocOrDie (sizeof (SQFILE));

    dbfp->ssimode = ssimode;
    dbfp->rpl = -1;		/* flag meaning "unset" */
    dbfp->lastrpl = 0;
    dbfp->maxrpl = 0;
    dbfp->bpl = -1;		/* flag meaning "unset" */
    dbfp->lastbpl = 0;
    dbfp->maxbpl = 0;

    /* Open our file handle.
     * Three possibilities:
     *    1. normal file open
     *    2. filename = "-";    read from stdin
     *    3. filename = "*.gz"; read thru pipe from gzip
     * If we're reading from stdin or a pipe, we can't reliably
     * back up, so we can't do two-pass parsers like the interleaved alignment
     * formats.
     */
    if (strcmp (filename, "-") == 0) {
        dbfp->f = stdin;
        dbfp->do_stdin = TRUE;
        dbfp->do_gzip = FALSE;
        dbfp->fname = sre_strdup ("[STDIN]", -1);
    }
#ifndef SRE_STRICT_ANSI
    /* popen(), pclose() aren't portable to non-POSIX systems; disable */
    else if (Strparse ("^.*\\.gz$", filename, 0)) {
        char cmd[256];

        /* Note that popen() will return "successfully"
         * if file doesn't exist, because gzip works fine
         * and prints an error! So we have to check for
         * existence of file ourself.
         */
        if (!FileExists (filename)) {
            Die ("%s: file does not exist", filename);
        }

        if (strlen (filename) + strlen ("gzip -dc ") >= 256) {
            Die ("filename > 255 char in SeqfileOpen()");
        }
        sprintf (cmd, "gzip -dc %s", filename);
        if ((dbfp->f = popen (cmd, "r")) == NULL) {
            return NULL;
        }

        dbfp->do_stdin = FALSE;
        dbfp->do_gzip = TRUE;
        dbfp->fname = sre_strdup (filename, -1);
    }
#endif /*SRE_STRICT_ANSI */
    else {
        if ((dbfp->f = fopen (filename, "r")) == NULL &&
                (dbfp->f = EnvFileOpen (filename, env, NULL)) == NULL) {
            return NULL;
        }

        dbfp->do_stdin = FALSE;
        dbfp->do_gzip = FALSE;
        dbfp->fname = sre_strdup (filename, -1);
    }


    /* Invoke autodetection if we haven't already been told what
     * to expect.
     */
    if (format == SQFILE_UNKNOWN) {
        if (dbfp->do_stdin == TRUE || dbfp->do_gzip) {
            Die ("Can't autodetect sequence file format from a stdin or gzip pipe");
        }
        format = SeqfileFormat (dbfp->f);
        if (format == SQFILE_UNKNOWN) {
            Die ("Can't determine format of sequence file %s", dbfp->fname);
        }
    }

    /* The hack for sequential access of an interleaved alignment file:
     * read the alignment in, we'll copy sequences out one at a time.
     */
    dbfp->msa = NULL;
    dbfp->afp = NULL;
    dbfp->format = format;
    dbfp->linenumber = 0;
    dbfp->buf = NULL;
    dbfp->buflen = 0;
    if (IsAlignmentFormat (format)) {
        /* We'll be reading from the MSA interface. Copy our data
         * to the MSA afp's structure.
         */
        dbfp->afp = MallocOrDie (sizeof (MSAFILE));
        dbfp->afp->f = dbfp->f;	/* just a ptr, don't close */
        dbfp->afp->do_stdin = dbfp->do_stdin;
        dbfp->afp->do_gzip = dbfp->do_gzip;
        dbfp->afp->fname = dbfp->fname;	/* just a ptr, don't free */
        dbfp->afp->format = dbfp->format;	/* e.g. format */
        dbfp->afp->linenumber = dbfp->linenumber;	/* e.g. 0 */
        dbfp->afp->buf = NULL;
        dbfp->afp->buflen = 0;

        if ((dbfp->msa = MSAFileRead (dbfp->afp)) == NULL) {
            Die ("Failed to read any alignment data from file %s", dbfp->fname);
        }
        /* hack: overload/reuse msa->lastidx; indicates
           next seq to return upon a ReadSeq() call */
        dbfp->msa->lastidx = 0;

        return dbfp;
    }

    /* Load the first line.
     */
    SeqfileGetLine (dbfp);
    return dbfp;
}

/* Function: SeqfilePosition()
 *
 * Purpose:  Move to a particular offset in a seqfile.
 *           Will not work on alignment files.
 */
void
SeqfilePosition (SQFILE * sqfp, SSIOFFSET * offset)
{
    if (sqfp->do_stdin || sqfp->do_gzip || IsAlignmentFormat (sqfp->format)) {
        Die ("SeqfilePosition() failed: in a nonrewindable data file or stream");
    }

    if (SSISetFilePosition (sqfp->f, offset) != 0) {
        Die ("SSISetFilePosition failed, but that shouldn't happen.");
    }
    SeqfileGetLine (sqfp);
}


/* Function: SeqfileRewind()
 *
 * Purpose:  Set a sequence file back to the first sequence.
 *
 *           Won't work on alignment files. Although it would
 *           seem that it could (just set msa->lastidx back to 0),
 *           that'll fail on "multiple multiple" alignment file formats
 *           (e.g. Stockholm).
 */
void
SeqfileRewind (SQFILE * sqfp)
{
    if (sqfp->do_stdin || sqfp->do_gzip) {
        Die ("SeqfileRewind() failed: in a nonrewindable data file or stream");
    }

    rewind (sqfp->f);
    SeqfileGetLine (sqfp);
}

/* Function: SeqfileLineParameters()
 * Date:     SRE, Thu Feb 15 17:00:41 2001 [St. Louis]
 *
 * Purpose:  After all the sequences have been read from the file,
 *           but before closing it, retrieve overall bytes-per-line and
 *           residues-per-line info. If non-zero, these mean that
 *           the file contains homogeneous sequence line lengths (except
 *           the last line in each record).
 *
 *           If either of bpl or rpl is determined to be inhomogeneous,
 *           both are returned as 0.
 *
 * Args:     *sqfp   - an open but fully read sequence file
 *           ret_bpl - RETURN: bytes per line, or 0 if inhomogeneous
 *           ret_rpl - RETURN: residues per line, or 0 if inhomogenous.
 *
 * Returns:  void
 */
void
SeqfileLineParameters (SQFILE * V, int *ret_bpl, int *ret_rpl)
{
    if (V->rpl > 0 && V->maxrpl == V->rpl && V->bpl > 0 && V->maxbpl == V->bpl) {
        *ret_bpl = V->bpl;
        *ret_rpl = V->rpl;
    }
    else {
        *ret_bpl = 0;
        *ret_rpl = 0;
    }
}


void
SeqfileClose (SQFILE * sqfp)
{
    /* note: don't test for sqfp->msa being NULL. Now that
     * we're holding afp open and allowing access to multi-MSA
     * databases (e.g. Stockholm format, Pfam), msa ends
     * up being NULL when we run out of alignments.
     */
    if (sqfp->afp != NULL) {
        if (sqfp->msa != NULL) {
            MSAFree (sqfp->msa);
        }
        if (sqfp->afp->buf != NULL) {
            free (sqfp->afp->buf);
        }
        free (sqfp->afp);
    }
#ifndef SRE_STRICT_ANSI
    /* gunzip functionality only on POSIX systems */
    if (sqfp->do_gzip) {
        pclose (sqfp->f);
    }
#endif
    else if (!sqfp->do_stdin) {
        fclose (sqfp->f);
    }
    if (sqfp->buf != NULL) {
        free (sqfp->buf);
    }
    if (sqfp->fname != NULL) {
        free (sqfp->fname);
    }
    free (sqfp);
}


/* Function: SeqfileGetLine()
 * Date:     SRE, Tue Jun 22 09:15:49 1999 [Sanger Centre]
 *
 * Purpose:  read a line from a sequence file into V->buf
 *           If the fgets() is NULL, sets V->buf[0] to '\0'.
 *
 * Args:     V
 *
 * Returns:  void
 */
static void
SeqfileGetLine (SQFILE * V)
{
    if (V->ssimode >= 0)
        if (0 != SSIGetFilePosition (V->f, V->ssimode, &(V->ssioffset))) {
            Die ("SSIGetFilePosition() failed");
        }
    if (sre_fgets (&(V->buf), &(V->buflen), V->f) == NULL) {
        *(V->buf) = '\0';
    }
    V->linenumber++;
}


void
FreeSequence (char *seq, SQINFO * sqinfo)
{
    if (seq != NULL) {
        free (seq);
    }
    if (sqinfo->flags & SQINFO_SS) {
        free (sqinfo->ss);
    }
    if (sqinfo->flags & SQINFO_SA) {
        free (sqinfo->sa);
    }
}

int
SetSeqinfoString (SQINFO * sqinfo, char *sptr, int flag)
{
    int len;
    int pos;

    /* silently ignore NULL. */
    if (sptr == NULL) {
        return 1;
    }

    while (*sptr == ' ') {
        sptr++;    /* ignore leading whitespace */
    }
    for (pos = strlen (sptr) - 1; pos >= 0; pos--) {
        if (!isspace ((int) sptr[pos])) {
            break;
        }
		}		
    sptr[pos + 1] = '\0';		/* ignore trailing whitespace */

    switch (flag) {
    case SQINFO_NAME:
        if (*sptr != '-') {
            strncpy (sqinfo->name, sptr, SQINFO_NAMELEN - 1);
            sqinfo->name[SQINFO_NAMELEN - 1] = '\0';
            sqinfo->flags |= SQINFO_NAME;
        }
        break;

    case SQINFO_ID:
        if (*sptr != '-') {
            strncpy (sqinfo->id, sptr, SQINFO_NAMELEN - 1);
            sqinfo->id[SQINFO_NAMELEN - 1] = '\0';
            sqinfo->flags |= SQINFO_ID;
        }
        break;

    case SQINFO_ACC:
        if (*sptr != '-') {
            strncpy (sqinfo->acc, sptr, SQINFO_NAMELEN - 1);
            sqinfo->acc[SQINFO_NAMELEN - 1] = '\0';
            sqinfo->flags |= SQINFO_ACC;
        }
        break;

    case SQINFO_DESC:
        if (*sptr != '-') {
            if (sqinfo->flags & SQINFO_DESC) {	/* append? */
                len = strlen (sqinfo->desc);
                if (len < SQINFO_DESCLEN - 2) {	/* is there room? */
                    strncat (sqinfo->desc, " ", SQINFO_DESCLEN - 1 - len);
                    len++;
                    strncat (sqinfo->desc, sptr, SQINFO_DESCLEN - 1 - len);
                }
            }
            else {		/* else copy */
                strncpy (sqinfo->desc, sptr, SQINFO_DESCLEN - 1);
            }
            sqinfo->desc[SQINFO_DESCLEN - 1] = '\0';
            sqinfo->flags |= SQINFO_DESC;
        }
        break;

    case SQINFO_START:
        if (!IsInt (sptr)) {
            squid_errno = SQERR_FORMAT;
            return 0;
        }
        sqinfo->start = atoi (sptr);
        if (sqinfo->start != 0) {
            sqinfo->flags |= SQINFO_START;
        }
        break;

    case SQINFO_STOP:
        if (!IsInt (sptr)) {
            squid_errno = SQERR_FORMAT;
            return 0;
        }
        sqinfo->stop = atoi (sptr);
        if (sqinfo->stop != 0) {
            sqinfo->flags |= SQINFO_STOP;
        }
        break;

    case SQINFO_OLEN:
        if (!IsInt (sptr)) {
            squid_errno = SQERR_FORMAT;
            return 0;
        }
        sqinfo->olen = atoi (sptr);
        if (sqinfo->olen != 0) {
            sqinfo->flags |= SQINFO_OLEN;
        }
        break;

    default:
        Die ("Invalid flag %d to SetSeqinfoString()", flag);
    }
    return 1;
}

void
SeqinfoCopy (SQINFO * sq1, SQINFO * sq2)
{
    sq1->flags = sq2->flags;
    if (sq2->flags & SQINFO_NAME) {
        strcpy (sq1->name, sq2->name);
    }
    if (sq2->flags & SQINFO_ID) {
        strcpy (sq1->id, sq2->id);
    }
    if (sq2->flags & SQINFO_ACC) {
        strcpy (sq1->acc, sq2->acc);
    }
    if (sq2->flags & SQINFO_DESC) {
        strcpy (sq1->desc, sq2->desc);
    }
    if (sq2->flags & SQINFO_LEN) {
        sq1->len = sq2->len;
    }
    if (sq2->flags & SQINFO_START) {
        sq1->start = sq2->start;
    }
    if (sq2->flags & SQINFO_STOP) {
        sq1->stop = sq2->stop;
    }
    if (sq2->flags & SQINFO_OLEN) {
        sq1->olen = sq2->olen;
    }
    if (sq2->flags & SQINFO_TYPE) {
        sq1->type = sq2->type;
    }
    if (sq2->flags & SQINFO_SS) {
        sq1->ss = Strdup (sq2->ss);
    }
    if (sq2->flags & SQINFO_SA) {
        sq1->sa = Strdup (sq2->sa);
    }
}

/* Function: ToDNA()
 *
 * Purpose:  Convert a sequence to DNA.
 *           U --> T
 */
void
ToDNA (char *seq)
{
    for (; *seq != '\0'; seq++) {
        if (*seq == 'U') {
            *seq = 'T';
        }
        else if (*seq == 'u') {
            *seq = 't';
        }
    }
}

/* Function: ToRNA()
 *
 * Purpose:  Convert a sequence to RNA.
 *           T --> U
 */
void
ToRNA (char *seq)
{
    for (; *seq != '\0'; seq++) {
        if (*seq == 'T') {
            *seq = 'U';
        }
        else if (*seq == 't') {
            *seq = 'u';
        }
    }
}


/* Function: ToIUPAC()
 *
 * Purpose:  Convert X's, o's, other junk in a nucleic acid sequence to N's,
 *           to comply with IUPAC code. If is_aseq is TRUE, will allow gap
 *           characters though, so we can call ToIUPAC() on aligned seqs.
 *
 *           NUCLEOTIDES is defined in squid.h as:
 *               "ACGTUNRYMKSWHBVDacgtunrymkswhbvd"
 *           gap chars allowed by isgap() are defined in squid.h as:
 *               " ._-~"
 *
 *           WU-BLAST's pressdb will
 *           choke on X's, for instance, necessitating conversion
 *           of certain genome centers' data.
 */
void
ToIUPAC (char *seq, int is_aseq)
{
    if (is_aseq) {
        for (; *seq != '\0'; seq++) {
            if (strchr (NUCLEOTIDES, *seq) == NULL && !isgap (*seq)) {
                *seq = 'N';
            }
				}		
    }
    else {
        for (; *seq != '\0'; seq++) {
            if (strchr (NUCLEOTIDES, *seq) == NULL) {
                *seq = 'N';
            }
				}		
    }
}


/* Function: addseq()
 *
 * Purpose:  Add a line of sequence to the growing string in V.
 *
 *           In the seven supported unaligned formats, all sequence
 *           lines may contain whitespace that must be filtered out;
 *           four formats (PIR, EMBL, Genbank, GCG) include coordinates
 *           that must be filtered out. Thus an (!isdigit && !isspace)
 *           test on each character before we accept it.
 */
static void
addseq (char *s, struct ReadSeqVars *V)
{
    char *s0;
    char *sq;
    int rpl;			/* valid residues per line */
    int bpl;			/* characters per line     */

    if (V->ssimode == -1) {	/* Normal mode: keeping the seq */
        /* Make sure we have enough room. We know that s is <= buflen,
         * so just make sure we've got room for a whole new buflen worth
         * of sequence.
         */
        if (V->seqlen + V->buflen > V->maxseq) {
            V->maxseq += MAX (V->buflen, kStartLength);
            V->seq = ReallocOrDie (V->seq, V->maxseq + 1);
            /*  following added because calloc clears memory and
               strings are null terminated. But resizing does not
               have null termination. Joydeep found this bug and
               has kindly provided the following fix -Kaivalya 9/15/03
               added jray :: null termination Kaivalya 9/15/03
               Author okayed the change.
             */
            V->seq[V->maxseq] = 0;	/* Kaivalya 09/15/03 */

        }

        sq = V->seq + V->seqlen;
        while (*s != 0) {
            if (!isdigit ((int) *s) && !isspace ((int) *s)) {
                *sq = *s;
                sq++;
            }
            s++;
        }
        V->seqlen = sq - V->seq;
    }
    else {			/* else: indexing mode, discard the seq */

        s0 = s;
        rpl = 0;
        while (*s != 0) {
            if (!isdigit ((int) *s) && !isspace ((int) *s)) {
                rpl++;
            }
            s++;
        }
        V->seqlen += rpl;
        bpl = s - s0;

        /* Keep track of the global rpl, bpl for the file.
         * This is overly complicated because we have to
         * allow the last line of each record (e.g. the last addseq() call
         * on each sequence) to have a different length - and sometimes
         * we'll have one-line sequence records, too.  Thus we only
         * do something with the global V->rpl when we have *passed over*
         * a line - we keep the last line's rpl in last_rpl. And because
         * a file might consist entirely of single-line records, we keep
         * a third guy, maxrpl, that tells us the maximum rpl of any line
         * in the file. If we reach the end of file and rpl is still unset,
         * we'll set it to maxrpl. If we reach eof and rpl is set, but is
         * less than maxrpl, that's a weird case where a last line in some
         * record is longer than every other line.
         */
        if (V->rpl != 0) {		/* 0 means we already know rpl is invalid       */
            if (V->lastrpl > 0) {	/* we're on something that's not the first line */
                if (V->rpl > 0 && V->lastrpl != V->rpl) {
                    V->rpl = 0;
                }
                else if (V->rpl == -1) {
                    V->rpl = V->lastrpl;
                }
            }
            V->lastrpl = rpl;
            if (rpl > V->maxrpl) {
                V->maxrpl = rpl;    /* make sure we check max length of final lines */
            }
        }
        if (V->bpl != 0) {		/* 0 means we already know bpl is invalid       */
            if (V->lastbpl > 0) {	/* we're on something that's not the first line */
                if (V->bpl > 0 && V->lastbpl != V->bpl) {
                    V->bpl = 0;
                }
                else if (V->bpl == -1) {
                    V->bpl = V->lastbpl;
                }
            }
            V->lastbpl = bpl;
            if (bpl > V->maxbpl) {
                V->maxbpl = bpl;    /* make sure we check max length of final lines */
            }
        }
    }				/* end of indexing mode of addseq(). */

}

static void
readLoop (int addfirst, int (*endTest) (char *, int *), struct ReadSeqVars *V)
{
    int addend = 0;
    int done = 0;

    V->seqlen = 0;
    V->lastrpl = V->lastbpl = 0;
    if (addfirst) {
        if (V->ssimode >= 0) {
            V->d_off = V->ssioffset;
        }
        addseq (V->buf, V);
    }
    else if (V->ssimode >= 0)
        if (0 != SSIGetFilePosition (V->f, V->ssimode, &(V->d_off))) {
            Die ("SSIGetFilePosition() failed");
        }

    do {
        SeqfileGetLine (V);
        /* feof() alone is a bug; files not necessarily \n terminated */
        if (*(V->buf) == '\0' && feof (V->f)) {
            done = TRUE;
        }
        done |= (*endTest) (V->buf, &addend);
        if (addend || !done) {
            addseq (V->buf, V);
        }
    } while (!done);
}


static int
endPIR (char *s, int *addend)
{
    *addend = 0;
    if ((strncmp (s, "///", 3) == 0) || (strncmp (s, "ENTRY", 5) == 0)) {
        return 1;
    }
    else {
        return 0;
    }
}

static void
readPIR (struct ReadSeqVars *V)
{
    char *sptr;
    /* load first line of entry  */
    while (!feof (V->f) && strncmp (V->buf, "ENTRY", 5) != 0) {
        SeqfileGetLine (V);
    }
    if (feof (V->f)) {
        return;
    }
    if (V->ssimode >= 0) {
        V->r_off = V->ssioffset;
    }

    if ((sptr = strtok (V->buf + 15, "\n\t ")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_NAME);
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_ID);
    }
    do {
        SeqfileGetLine (V);
        if (!feof (V->f) && strncmp (V->buf, "TITLE", 5) == 0) {
            SetSeqinfoString (V->sqinfo, V->buf + 15, SQINFO_DESC);
        }
        else if (!feof (V->f) && strncmp (V->buf, "ACCESSION", 9) == 0) {
            if ((sptr = strtok (V->buf + 15, " \t\n")) != NULL) {
                SetSeqinfoString (V->sqinfo, sptr, SQINFO_ACC);
            }
        }
    } while (!feof (V->f) && (strncmp (V->buf, "SEQUENCE", 8) != 0));
    SeqfileGetLine (V);		/* skip next line, coords */

    readLoop (0, endPIR, V);

    /* reading a real PIR-CODATA database file, we keep the source coords
     */
    V->sqinfo->start = 1;
    V->sqinfo->stop = V->seqlen;
    V->sqinfo->olen = V->seqlen;
    V->sqinfo->flags |= SQINFO_START | SQINFO_STOP | SQINFO_OLEN;

    /* get next line
     */
    while (!feof (V->f) && strncmp (V->buf, "ENTRY", 5) != 0) {
        SeqfileGetLine (V);
    }
}



static int
endIG (char *s, int *addend)
{
    *addend = 1;			/* 1 or 2 occur in line w/ bases */
    return ((strchr (s, '1') != NULL) || (strchr (s, '2') != NULL));
}

static void
readIG (struct ReadSeqVars *V)
{
    char *nm;
    /* position past ';' comments */
    do {
        SeqfileGetLine (V);
    } while (!(feof (V->f) || ((*V->buf != 0) && (*V->buf != ';'))));

    if (!feof (V->f)) {
        if ((nm = strtok (V->buf, "\n\t ")) != NULL) {
            SetSeqinfoString (V->sqinfo, nm, SQINFO_NAME);
        }

        readLoop (0, endIG, V);
    }

    while (!(feof (V->f) || ((*V->buf != '\0') && (*V->buf == ';')))) {
        SeqfileGetLine (V);
    }
}

static int
endStrider (char *s, int *addend)
{
    *addend = 0;
    return (strstr (s, "//") != NULL);
}

static void
readStrider (struct ReadSeqVars *V)
{
    char *nm;

    while ((!feof (V->f)) && (*V->buf == ';')) {
        if (strncmp (V->buf, "; DNA sequence", 14) == 0) {
            if ((nm = strtok (V->buf + 16, ",\n\t ")) != NULL) {
                SetSeqinfoString (V->sqinfo, nm, SQINFO_NAME);
            }
        }
        SeqfileGetLine (V);
    }

    if (!feof (V->f)) {
        readLoop (1, endStrider, V);
    }

    /* load next line
     */
    while ((!feof (V->f)) && (*V->buf != ';')) {
        SeqfileGetLine (V);
    }
}


static int
endGB (char *s, int *addend)
{
    *addend = 0;
    return ((strstr (s, "//") != NULL) || (strstr (s, "LOCUS") == s));
}

static void
readGenBank (struct ReadSeqVars *V)
{
    char *sptr;
    int in_definition;

    /* We'll map three genbank identifiers onto names:
     *     LOCUS     -> sqinfo.name
     *     ACCESSION -> sqinfo.acc   [primary accession only]
     *     VERSION   -> sqinfo.id
     * We don't currently store the GI number, or secondary accessions.
     */
    while (strncmp (V->buf, "LOCUS", 5) != 0) {
        SeqfileGetLine (V);
    }
    if (V->ssimode >= 0) {
        V->r_off = V->ssioffset;
    }

    if ((sptr = strtok (V->buf + 12, "\n\t ")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_NAME);
    }

    in_definition = FALSE;
    while (!feof (V->f)) {
        SeqfileGetLine (V);
        if (!feof (V->f) && strstr (V->buf, "DEFINITION") == V->buf) {
            if ((sptr = strtok (V->buf + 12, "\n")) != NULL) {
                SetSeqinfoString (V->sqinfo, sptr, SQINFO_DESC);
            }
            in_definition = TRUE;
        }
        else if (!feof (V->f) && strstr (V->buf, "ACCESSION") == V->buf) {
            if ((sptr = strtok (V->buf + 12, "\n\t ")) != NULL) {
                SetSeqinfoString (V->sqinfo, sptr, SQINFO_ACC);
            }
            in_definition = FALSE;
        }
        else if (!feof (V->f) && strstr (V->buf, "VERSION") == V->buf) {
            if ((sptr = strtok (V->buf + 12, "\n\t ")) != NULL) {
                SetSeqinfoString (V->sqinfo, sptr, SQINFO_ID);
            }
            in_definition = FALSE;
        }
        else if (strncmp (V->buf, "ORIGIN", 6) != 0) {
            if (in_definition) {
                SetSeqinfoString (V->sqinfo, V->buf, SQINFO_DESC);
            }
        }
        else {
            break;
        }
    }

    readLoop (0, endGB, V);

    /* reading a real GenBank database file, we keep the source coords
     */
    V->sqinfo->start = 1;
    V->sqinfo->stop = V->seqlen;
    V->sqinfo->olen = V->seqlen;
    V->sqinfo->flags |= SQINFO_START | SQINFO_STOP | SQINFO_OLEN;


    while (!
            (feof (V->f)
             || ((*V->buf != 0) && (strstr (V->buf, "LOCUS") == V->buf)))) {
        SeqfileGetLine (V);
    }
    /* SRE: V->s now holds "//", so sequential
       reads are wedged: fixed Tue Jul 13 1993 */
    while (!feof (V->f) && strstr (V->buf, "LOCUS  ") != V->buf) {
        SeqfileGetLine (V);
    }
}

static int
endGCGdata (char *s, int *addend)
{
    *addend = 0;
    return (*s == '>');
}

static void
readGCGdata (struct ReadSeqVars *V)
{
    int binary = FALSE;		/* whether data are binary or not */
    int blen = 0;			/* length of binary sequence */

    /* first line contains ">>>>" followed by name */
    if (Strparse (">>>>([^ ]+) .+2BIT +Len: ([0-9]+)", V->buf, 2)) {
        binary = TRUE;
        SetSeqinfoString (V->sqinfo, sqd_parse[1], SQINFO_NAME);
        blen = atoi (sqd_parse[2]);
    }
    else if (Strparse (">>>>([^ ]+) .+ASCII +Len: [0-9]+", V->buf, 1)) {
        SetSeqinfoString (V->sqinfo, sqd_parse[1], SQINFO_NAME);
    }
    else {
        Die ("bogus GCGdata format? %s", V->buf);
    }

    /* second line contains free text description */
    SeqfileGetLine (V);
    SetSeqinfoString (V->sqinfo, V->buf, SQINFO_DESC);

    if (binary) {
        /* allocate for blen characters +3... (allow for 3 bytes of slop) */
        if (blen >= V->maxseq) {
            V->maxseq = blen;
            if ((V->seq =
                        (char *) realloc (V->seq,
                                          sizeof (char) * (V->maxseq + 4))) == NULL) {
                Die ("malloc failed");
            }
        }
        /* read (blen+3)/4 bytes from file */
        if (fread (V->seq, sizeof (char), (blen + 3) / 4, V->f) <
                (size_t) ((blen + 3) / 4)) {
            Die ("fread failed");
        }
        V->seqlen = blen;
        /* convert binary code to seq */
        GCGBinaryToSequence (V->seq, blen);
    }
    else {
        readLoop (0, endGCGdata, V);
    }

    while (!(feof (V->f) || ((*V->buf != 0) && (*V->buf == '>')))) {
        SeqfileGetLine (V);
    }
}

static int
endPearson (char *s, int *addend)
{
    *addend = 0;
    return (*s == '>');
}

static void
readPearson (struct ReadSeqVars *V)
{
    char *sptr;

    if (V->ssimode >= 0) {
        V->r_off = V->ssioffset;
    }

    if (*V->buf != '>')
        Die ("\
File %s does not appear to be in FASTA format at line %d.\n\
You may want to specify the file format on the command line.\n\
Usually this is done with an option --informat <fmt>.\n", V->fname, V->linenumber);

    if ((sptr = strtok (V->buf + 1, "\n\t ")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_NAME);
    }
    if ((sptr = strtok (NULL, "\n")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_DESC);
    }

    readLoop (0, endPearson, V);

    while (!(feof (V->f) || ((*V->buf != 0) && (*V->buf == '>')))) {
        SeqfileGetLine (V);
    }
}


static int
endEMBL (char *s, int *addend)
{
    *addend = 0;
    /* Some people (Berlin 5S rRNA database, f'r instance) use
     * an extended EMBL format that attaches extra data after
     * the sequence -- watch out for that. We use the fact that
     * real EMBL sequence lines begin with five spaces.
     *
     * We can use this as the sole end test because readEMBL() will
     * advance to the next ID line before starting to read again.
     */
    return (strncmp (s, "     ", 5) != 0);
    /*  return ((strstr(s,"//") != NULL) || (strstr(s,"ID   ") == s)); */
}

static void
readEMBL (struct ReadSeqVars *V)
{
    char *sptr;

    /* make sure we have first line */
    while (!feof (V->f) && strncmp (V->buf, "ID  ", 4) != 0) {
        SeqfileGetLine (V);
    }
    if (V->ssimode >= 0) {
        V->r_off = V->ssioffset;
    }

    if ((sptr = strtok (V->buf + 5, "\n\t ")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_NAME);
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_ID);
    }

    do {
        SeqfileGetLine (V);
        if (!feof (V->f) && strstr (V->buf, "AC  ") == V->buf) {
            if ((sptr = strtok (V->buf + 5, ";  \t\n")) != NULL) {
                SetSeqinfoString (V->sqinfo, sptr, SQINFO_ACC);
            }
        }
        else if (!feof (V->f) && strstr (V->buf, "DE  ") == V->buf) {
            if ((sptr = strtok (V->buf + 5, "\n")) != NULL) {
                SetSeqinfoString (V->sqinfo, sptr, SQINFO_DESC);
            }
        }
    } while (!feof (V->f) && strncmp (V->buf, "SQ", 2) != 0);

    readLoop (0, endEMBL, V);

    /* Hack for Staden experiment files: convert - to N
     */
    if (V->ssimode == -1)		/* if we're in ssi mode, we're not keeping the seq */
        for (sptr = V->seq; *sptr != '\0'; sptr++) {
            if (*sptr == '-') {
                *sptr = 'N';
            }
				}		

    /* reading a real EMBL database file, we keep the source coords
     */
    V->sqinfo->start = 1;
    V->sqinfo->stop = V->seqlen;
    V->sqinfo->olen = V->seqlen;
    V->sqinfo->flags |= SQINFO_START | SQINFO_STOP | SQINFO_OLEN;

    /* load next record's ID line */
    while (!feof (V->f) && strncmp (V->buf, "ID  ", 4) != 0) {
        SeqfileGetLine (V);
    }

}


static int
endZuker (char *s, int *addend)
{
    *addend = 0;
    return (*s == '(');
}

static void
readZuker (struct ReadSeqVars *V)
{
    char *sptr;

    SeqfileGetLine (V);		/*s == "seqLen seqid string..." */

    if ((sptr = strtok (V->buf + 6, " \t\n")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_NAME);
    }

    if ((sptr = strtok (NULL, "\n")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_DESC);
    }

    readLoop (0, endZuker, V);

    while (!(feof (V->f) | ((*V->buf != '\0') & (*V->buf == '(')))) {
        SeqfileGetLine (V);
    }
}

static void
readUWGCG (struct ReadSeqVars *V)
{
    char *si;
    char *sptr;
    int done;

    V->seqlen = 0;

    /*writeseq: "    %s  Length: %d  (today)  Check: %d  ..\n" */
    /*drop above or ".." from id */
    if ((si = strstr (V->buf, "  Length: ")) != NULL) {
        *si = 0;
    }
    else if ((si = strstr (V->buf, "..")) != NULL) {
        *si = 0;
    }

    if ((sptr = strtok (V->buf, "\n\t ")) != NULL) {
        SetSeqinfoString (V->sqinfo, sptr, SQINFO_NAME);
    }

    do {
        done = feof (V->f);
        SeqfileGetLine (V);
        if (!done) {
            addseq (V->buf, V);
        }
    } while (!done);
}


/* Function: ReadSeq()
 *
 * Purpose:  Read next sequence from an open database file.
 *           Return the sequence and associated info.
 *
 * Args:     fp      - open sequence database file pointer
 *           format  - format of the file (previously determined
 *                      by call to SeqfileFormat()).
 *                      Currently unused, since we carry it in V.
 *           ret_seq - RETURN: sequence
 *           sqinfo  - RETURN: filled in w/ other information
 *
 * Limitations: uses squid_errno, so it's not threadsafe.
 *
 * Return:   1 on success, 0 on failure.
 *           ret_seq and some field of sqinfo are allocated here,
 *           The preferred call mechanism to properly free the memory is:
 *
 *           SQINFO sqinfo;
 *           char  *seq;
 *
 *           ReadSeq(fp, format, &seq, &sqinfo);
 *           ... do something...
 *           FreeSequence(seq, &sqinfo);
 */
int
ReadSeq (SQFILE * V, int format, char **ret_seq, SQINFO * sqinfo)
{
    int gotuw;

    squid_errno = SQERR_OK;

    /* Here's the hack for sequential access of sequences from
     * the multiple sequence alignment formats
     */
    if (IsAlignmentFormat (V->format)) {
        if (V->msa->lastidx >= V->msa->nseq) {	/* out of data. try to read another alignment */
            MSAFree (V->msa);
            if ((V->msa = MSAFileRead (V->afp)) == NULL) {
                return 0;
            }
            V->msa->lastidx = 0;
        }
        /* copy and dealign the appropriate aligned seq */
        MakeDealignedString (V->msa->aseq[V->msa->lastidx], V->msa->alen,
                             V->msa->aseq[V->msa->lastidx], &(V->seq));
        V->seqlen = strlen (V->seq);

        /* Extract sqinfo stuff for this sequence from the msa.
         * Tedious; code that should be cleaned.
         */
        sqinfo->flags = 0;
        if (V->msa->sqname[V->msa->lastidx] != NULL) {
            SetSeqinfoString (sqinfo, V->msa->sqname[V->msa->lastidx], SQINFO_NAME);
        }
        if (V->msa->sqacc != NULL && V->msa->sqacc[V->msa->lastidx] != NULL) {
            SetSeqinfoString (sqinfo, V->msa->sqacc[V->msa->lastidx], SQINFO_ACC);
        }
        if (V->msa->sqdesc != NULL && V->msa->sqdesc[V->msa->lastidx] != NULL) {
            SetSeqinfoString (sqinfo, V->msa->sqdesc[V->msa->lastidx], SQINFO_DESC);
        }
        if (V->msa->ss != NULL && V->msa->ss[V->msa->lastidx] != NULL) {
            MakeDealignedString (V->msa->aseq[V->msa->lastidx], V->msa->alen,
                                 V->msa->ss[V->msa->lastidx], &(sqinfo->ss));
            sqinfo->flags |= SQINFO_SS;
        }
        if (V->msa->sa != NULL && V->msa->sa[V->msa->lastidx] != NULL) {
            MakeDealignedString (V->msa->aseq[V->msa->lastidx], V->msa->alen,
                                 V->msa->sa[V->msa->lastidx], &(sqinfo->sa));
            sqinfo->flags |= SQINFO_SA;
        }
        V->msa->lastidx++;
    }
    else {
        if (feof (V->f)) {
            return 0;
        }

        if (V->ssimode == -1) {	/* normal mode */
            V->seq = (char *) calloc (kStartLength + 1, sizeof (char));
            V->maxseq = kStartLength;
        }
        else {			/* index mode: discarding seq */
            V->seq = NULL;
            V->maxseq = 0;
        }
        V->seqlen = 0;
        V->sqinfo = sqinfo;
        V->sqinfo->flags = 0;

        switch (V->format) {
        case SQFILE_IG:
            readIG (V);
            break;
        case SQFILE_STRIDER:
            readStrider (V);
            break;
        case SQFILE_GENBANK:
            readGenBank (V);
            break;
        case SQFILE_FASTA:
            readPearson (V);
            break;
        case SQFILE_EMBL:
            readEMBL (V);
            break;
        case SQFILE_ZUKER:
            readZuker (V);
            break;
        case SQFILE_PIR:
            readPIR (V);
            break;
        case SQFILE_GCGDATA:
            readGCGdata (V);
            break;

        case SQFILE_GCG:
            do {			/* skip leading comments on GCG file */
                gotuw = (strstr (V->buf, "..") != NULL);
                if (gotuw) {
                    readUWGCG (V);
                }
                SeqfileGetLine (V);
            } while (!feof (V->f));
            break;

        case SQFILE_IDRAW:		/* SRE: no attempt to read idraw postscript */
        default:
            squid_errno = SQERR_FORMAT;
            free (V->seq);
            return 0;
        }
        if (V->seq != NULL) {	/* (it can be NULL in indexing mode) */
            V->seq[V->seqlen] = 0;    /* stick a string terminator on it */
        }
    }

    /* Cleanup
     */
    sqinfo->len = V->seqlen;
    sqinfo->flags |= SQINFO_LEN;
    *ret_seq = V->seq;
    if (squid_errno == SQERR_OK) {
        return 1;
    }
    else {
        return 0;
    }
}

/* Function: SeqfileFormat()
 * Date:     SRE, Tue Jun 22 10:58:58 1999 [Sanger Centre]
 *
 * Purpose:  Determine format of an open file.
 *           Returns format code.
 *           Rewinds the file.
 *
 *           Autodetects the following unaligned formats:
 *              SQFILE_FASTA
 *              SQFILE_GENBANK
 *              SQFILE_EMBL
 *              SQFILE_GCG
 *              SQFILE_GCGDATA
 *              SQFILE_PIR
 *           Also autodetects the following alignment formats:
 *              MSAFILE_STOCKHOLM
 *              MSAFILE_MSF
 *              MSAFILE_CLUSTAL
 *              MSAFILE_SELEX
 *              MSAFILE_PHYLIP
 *
 *           Can't autodetect MSAFILE_A2M, calls it SQFILE_FASTA.
 *           MSAFileFormat() does the opposite.
 *
 * Args:     sfp -  open SQFILE
 *
 * Return:   format code, or SQFILE_UNKNOWN if unrecognized
 */
int
SeqfileFormat (FILE * fp)
{
    char *buf;
    int len;
    int fmt = SQFILE_UNKNOWN;
    int ndataline;
    char *bufcpy, *s, *s1, *s2;
    int has_junk;

    buf = NULL;
    len = 0;
    ndataline = 0;
    has_junk = FALSE;
    while (sre_fgets (&buf, &len, fp) != NULL) {
        if (IsBlankline (buf)) {
            continue;
        }

        /* Well-behaved formats identify themselves in first nonblank line.
         */
        if (ndataline == 0) {
            if (strncmp (buf, ">>>>", 4) == 0 && strstr (buf, "Len: ")) {
                fmt = SQFILE_GCGDATA;
                goto DONE_3;
            }

            if (buf[0] == '>') {
                fmt = SQFILE_FASTA;
                goto DONE_3;
            }

            if (strncmp (buf, "!!AA_SEQUENCE", 13) == 0 ||
                    strncmp (buf, "!!NA_SEQUENCE", 13) == 0) {
                fmt = SQFILE_GCG;
                goto DONE_3;
            }

            if (strncmp (buf, "# STOCKHOLM 1.", 14) == 0) {
                fmt = MSAFILE_STOCKHOLM;
                goto DONE_3;
            }

            if (strncmp (buf, "CLUSTAL", 7) == 0 &&
                    strstr (buf, "multiple sequence alignment") != NULL) {
                fmt = MSAFILE_CLUSTAL;
                goto DONE_3;
            }

            if (strncmp (buf, "!!AA_MULTIPLE_ALIGNMENT", 23) == 0 ||
                    strncmp (buf, "!!NA_MULTIPLE_ALIGNMENT", 23) == 0) {
                fmt = MSAFILE_MSF;
                goto DONE_3;
            }

            /* PHYLIP id: also just a good bet */
            bufcpy = sre_strdup (buf, -1);
            s = bufcpy;
            if ((s1 = sre_strtok (&s, WHITESPACE, NULL)) != NULL &&
                    (s2 = sre_strtok (&s, WHITESPACE, NULL)) != NULL &&
                    IsInt (s1) && IsInt (s2)) {
                free (bufcpy);
                fmt = MSAFILE_PHYLIP;
                goto DONE_3;
            }
            free (bufcpy);
        }

        /* We trust that other formats identify themselves soon.
         */
        /* dead giveaways for extended SELEX */
        if (strncmp (buf, "#=AU", 4) == 0 ||
                strncmp (buf, "#=ID", 4) == 0 ||
                strncmp (buf, "#=AC", 4) == 0 ||
                strncmp (buf, "#=DE", 4) == 0 ||
                strncmp (buf, "#=GA", 4) == 0 ||
                strncmp (buf, "#=TC", 4) == 0 ||
                strncmp (buf, "#=NC", 4) == 0 ||
                strncmp (buf, "#=SQ", 4) == 0 ||
                strncmp (buf, "#=SS", 4) == 0 ||
                strncmp (buf, "#=CS", 4) == 0 || strncmp (buf, "#=RF", 4) == 0) {
            fmt = MSAFILE_SELEX;
            goto DONE_3;
        }

        if (strncmp (buf, "///", 3) == 0 || strncmp (buf, "ENTRY ", 6) == 0) {
            fmt = SQFILE_PIR;
            goto DONE_3;
        }

        /* a ha, diagnostic of an (old) MSF file */
        if ((strstr (buf, "..") != NULL) &&
                (strstr (buf, "MSF:") != NULL) && (strstr (buf, "Check:") != NULL)) {
            fmt = MSAFILE_MSF;
            goto DONE_3;
        }

        /* unaligned GCG (must follow MSF test!) */
        if (strstr (buf, " Check: ") != NULL && strstr (buf, "..") != NULL) {
            fmt = SQFILE_GCG;
            goto DONE_3;
        }

        if (strncmp (buf, "LOCUS ", 6) == 0 || strncmp (buf, "ORIGIN ", 6) == 0) {
            fmt = SQFILE_GENBANK;
            goto DONE_3;
        }

        if (strncmp (buf, "ID   ", 5) == 0 || strncmp (buf, "SQ   ", 5) == 0) {
            fmt = SQFILE_EMBL;
            goto DONE_3;
        }

        /* But past here, we're being desperate. A simple SELEX file is
         * very difficult to detect; we can only try to disprove it.
         */
        s = buf;
        if ((s1 = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
            continue;    /* skip blank lines */
        }
        if (strchr ("#%", *s1) != NULL) {
            continue;    /* skip comment lines */
        }

        /* Disproof 1. Noncomment, nonblank lines in a SELEX file
         * must have at least two space-delimited fields (name/seq)
         */
        if ((s2 = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
            has_junk = TRUE;
        }

        /* Disproof 2.
         * The sequence field should look like a sequence.
         */
        if (s2 != NULL && Seqtype (s2) == kOtherSeq) {
            has_junk = TRUE;
        }

        ndataline++;
        if (ndataline == 300) {
            break;    /* only look at first 300 lines */
        }
    }

    if (ndataline == 0) {
        Die ("Sequence file contains no data");
    }

    /* If we've made it this far, we've run out of data, but there
     * was at least one line of it; check if we've
     * disproven SELEX. If not, cross our fingers, pray, and guess SELEX.
     */
    if (has_junk == TRUE) {
        fmt = SQFILE_UNKNOWN;
    }
    else {
        fmt = MSAFILE_SELEX;
    }

DONE_3:
    if (buf != NULL) {
        free (buf);
    }
    rewind (fp);
    return fmt;
}

/* Function: GCGBinaryToSequence()
 *
 * Purpose:  Convert a GCG 2BIT binary string to DNA sequence.
 *           0 = C  1 = T  2 = A  3 = G
 *           4 nts/byte
 *
 * Args:     seq  - binary sequence. Converted in place to DNA.
 *           len  - length of DNA. binary is (len+3)/4 bytes
 */
int
GCGBinaryToSequence (char *seq, int len)
{
    int bpos;			/* position in binary   */
    int spos;			/* position in sequence */
    char twobit;
    int i;

    for (bpos = (len - 1) / 4; bpos >= 0; bpos--) {
        twobit = seq[bpos];
        spos = bpos * 4;

        for (i = 3; i >= 0; i--) {
            switch (twobit & 0x3) {
            case 0:
                seq[spos + i] = 'C';
                break;
            case 1:
                seq[spos + i] = 'T';
                break;
            case 2:
                seq[spos + i] = 'A';
                break;
            case 3:
                seq[spos + i] = 'G';
                break;
            }
            twobit = twobit >> 2;
        }
    }
    seq[len] = '\0';
    return 1;
}


/* Function: GCGchecksum()
 * Date:     SRE, Mon May 31 11:13:21 1999 [St. Louis]
 *
 * Purpose:  Calculate a GCG checksum for a sequence.
 *           Code provided by Steve Smith of Genetics
 *           Computer Group.
 *
 * Args:     seq -  sequence to calculate checksum for.
 *                  may contain gap symbols.
 *           len -  length of sequence (usually known,
 *                  so save a strlen() call)
 *
 * Returns:  GCG checksum.
 */
int
GCGchecksum (char *seq, int len)
{
    int i;			/* position in sequence */
    int chk = 0;			/* calculated checksum  */

    for (i = 0; i < len; i++) {
        chk = (chk + (i % 57 + 1) * (sre_toupper ((int) seq[i]))) % 10000;
    }
    return chk;
}


/* Function: GCGMultchecksum()
 *
 * Purpose:  GCG checksum for a multiple alignment: sum of
 *           individual sequence checksums (including their
 *           gap characters) modulo 10000.
 *
 *           Implemented using spec provided by Steve Smith of
 *           Genetics Computer Group.
 *
 * Args:     seqs - sequences to be checksummed; aligned or not
 *           nseq - number of sequences
 *
 * Return:   the checksum, a number between 0 and 9999
 */
int
GCGMultchecksum (char **seqs, int nseq)
{
    int chk = 0;
    int idx;

    for (idx = 0; idx < nseq; idx++) {
        chk = (chk + GCGchecksum (seqs[idx], strlen (seqs[idx]))) % 10000;
    }
    return chk;
}




/* Function: Seqtype()
 *
 * Purpose:  Returns a (very good) guess about type of sequence:
 *           kDNA, kRNA, kAmino, or kOtherSeq.
 *
 *           Modified from, and replaces, Gilbert getseqtype().
 */
int
Seqtype (char *seq)
{
    int saw;			/* how many non-gap characters I saw */
    char c;
    int po = 0;			/* count of protein-only */
    int nt = 0;			/* count of t's */
    int nu = 0;			/* count of u's */
    int na = 0;			/* count of nucleotides */
    int aa = 0;			/* count of amino acids */
    int no = 0;			/* count of others */

    /* Look at the first 300 non-gap characters
     */
    for (saw = 0; *seq != '\0' && saw < 300; seq++) {
        c = sre_toupper ((int) *seq);
        if (!isgap (c)) {
            if (strchr (protonly, c)) {
                po++;
            }
            else if (strchr (primenuc, c)) {
                na++;
                if (c == 'T') {
                    nt++;
                }
                else if (c == 'U') {
                    nu++;
                }
            }
            else if (strchr (aminos, c)) {
                aa++;
            }
            else if (isalpha ((int) c)) {
                no++;
            }
            saw++;
        }
    }

    if (no > 0) {
        return kOtherSeq;
    }
    else if (po > 0) {
        return kAmino;
    }
    else if (na > aa) {
        if (nu > nt) {
            return kRNA;
        }
        else {
            return kDNA;
        }
    }
    else {
        return kAmino;    /* ooooh. risky. */
    }
}


/* Function: GuessAlignmentSeqtype()
 * Date:     SRE, Wed Jul  7 09:42:34 1999 [St. Louis]
 *
 * Purpose:  Try to guess whether an alignment is protein
 *           or nucleic acid; return a code for the
 *           type (kRNA, kDNA, or kAmino).
 *
 * Args:     aseq  - array of aligned sequences. (Could also
 *                   be an rseq unaligned sequence array)
 *           nseq  - number of aseqs
 *
 * Returns:  kRNA, kDNA, kAmino;
 *           kOtherSeq if inconsistency is detected.
 */
int
GuessAlignmentSeqtype (char **aseq, int nseq)
{
    int idx;
    int nrna = 0;
    int ndna = 0;
    int namino = 0;
    int nother = 0;

    for (idx = 0; idx < nseq; idx++) {
        switch (Seqtype (aseq[idx])) {
        case kRNA:
            nrna++;
            break;
        case kDNA:
            ndna++;
            break;
        case kAmino:
            namino++;
            break;
        default:
            nother++;
        }
		}		

    /* Unambiguous decisions:
     */
    if (nother) {
        return kOtherSeq;
    }
    if (namino == nseq) {
        return kAmino;
    }
    if (ndna == nseq) {
        return kDNA;
    }
    if (nrna == nseq) {
        return kRNA;
    }

    /* Ambiguous decisions:
     */
    if (namino == 0) {
        return kRNA;    /* it's nucleic acid, but seems mixed RNA/DNA */
    }
    return kAmino;		/* some amino acid seen; others probably short seqs, some
				   of which may be entirely ACGT (ala,cys,gly,thr). We
				   could be a little more sophisticated: U would be a giveaway
				   that we're not in protein seqs */
}

/* Function: WriteSimpleFASTA()
 * Date:     SRE, Tue Nov 16 18:06:00 1999 [St. Louis]
 *
 * Purpose:  Just write a FASTA format sequence to a file;
 *           minimal interface, mostly for quick and dirty programs.
 *
 * Args:     fp   - open file handle (stdout, possibly)
 *           seq  - sequence to output
 *           name - name for the sequence
 *           desc - optional description line, or NULL.
 *
 * Returns:  void
 */
void
WriteSimpleFASTA (FILE * fp, char *seq, char *name, char *desc)
{
    char buf[61];
    int len;
    int pos;

    len = strlen (seq);
    buf[60] = '\0';
    fprintf (fp, ">%s %s\n", name, desc != NULL ? desc : "");
    for (pos = 0; pos < len; pos += 60) {
        strncpy (buf, seq + pos, 60);
        fprintf (fp, "%s\n", buf);
    }
}

int
WriteSeq (FILE * outf, int outform, char *seq, SQINFO * sqinfo)
{
    int numline = 0;
    int lines = 0, spacer = 0, width = 50, tab = 0;
    int i, j, l, l1, ibase;
    char endstr[10];
    char s[100];			/* buffer for sequence  */
    char ss[100];			/* buffer for structure */
    int checksum = 0;
    int seqlen;
    int which_case;		/* 0 = do nothing. 1 = upper case. 2 = lower case */
    int dostruc;			/* TRUE to print structure lines */

    which_case = 0;
    dostruc = FALSE;
    seqlen = (sqinfo->flags & SQINFO_LEN) ? sqinfo->len : strlen (seq);

    if (IsAlignmentFormat (outform)) {
        Die ("Tried to write an aligned format with WriteSeq() -- bad, bad.");
    }


    strcpy (endstr, "");
    l1 = 0;
    checksum = GCGchecksum (seq, seqlen);

    switch (outform) {
    case SQFILE_UNKNOWN:		/* no header, just sequence */
        strcpy (endstr, "\n");	/* end w/ extra blank line */
        break;

    case SQFILE_GENBANK:
        fprintf (outf, "LOCUS       %s       %d bp\n", sqinfo->name, seqlen);
        fprintf (outf, "ACCESSION   %s\n",
                 (sqinfo->flags & SQINFO_ACC) ? sqinfo->acc : ".");
        fprintf (outf, "DEFINITION  %s\n",
                 (sqinfo->flags & SQINFO_DESC) ? sqinfo->desc : ".");
        fprintf (outf, "VERSION     %s\n",
                 (sqinfo->flags & SQINFO_ID) ? sqinfo->id : ".");
        fprintf (outf, "ORIGIN      \n");
        spacer = 11;
        numline = 1;
        strcpy (endstr, "\n//");
        break;

    case SQFILE_GCGDATA:
        fprintf (outf, ">>>>%s  9/95  ASCII  Len: %d\n", sqinfo->name, seqlen);
        fprintf (outf, "%s\n",
                 (sqinfo->flags & SQINFO_DESC) ? sqinfo->desc : "-");
        break;

    case SQFILE_PIR:
        fprintf (outf, "ENTRY          %s\n",
                 (sqinfo->flags & SQINFO_ID) ? sqinfo->id : sqinfo->name);
        fprintf (outf, "TITLE          %s\n",
                 (sqinfo->flags & SQINFO_DESC) ? sqinfo->desc : "-");
        fprintf (outf, "ACCESSION      %s\n",
                 (sqinfo->flags & SQINFO_ACC) ? sqinfo->acc : "-");
        fprintf (outf,
                 "SUMMARY                                #Length %d  #Checksum  %d\n",
                 sqinfo->len, checksum);
        fprintf (outf, "SEQUENCE\n");
        fprintf (outf,
                 "                  5        10        15        20        25        30\n");
        spacer = 2;			/* spaces after every residue */
        numline = 1;		/* number lines w/ coords     */
        width = 30;			/* 30 aa per line             */
        strcpy (endstr, "\n///");
        break;

    case SQFILE_SQUID:
        fprintf (outf, "NAM  %s\n", sqinfo->name);
        if (sqinfo->
                flags & (SQINFO_ID | SQINFO_ACC | SQINFO_START | SQINFO_STOP |
                         SQINFO_OLEN))
            fprintf (outf, "SRC  %s %s %d..%d::%d\n",
                     (sqinfo->flags & SQINFO_ID) ? sqinfo->id : "-",
                     (sqinfo->flags & SQINFO_ACC) ? sqinfo->acc : "-",
                     (sqinfo->flags & SQINFO_START) ? sqinfo->start : 0,
                     (sqinfo->flags & SQINFO_STOP) ? sqinfo->stop : 0,
                     (sqinfo->flags & SQINFO_OLEN) ? sqinfo->olen : 0);
        if (sqinfo->flags & SQINFO_DESC) {
            fprintf (outf, "DES  %s\n", sqinfo->desc);
        }
        if (sqinfo->flags & SQINFO_SS) {
            fprintf (outf, "SEQ  +SS\n");
            dostruc = TRUE;		/* print structure lines too */
        }
        else {
            fprintf (outf, "SEQ\n");
        }
        numline = 1;		/* number seq lines w/ coords  */
        strcpy (endstr, "\n++");
        break;

    case SQFILE_EMBL:
        fprintf (outf, "ID   %s\n",
                 (sqinfo->flags & SQINFO_ID) ? sqinfo->id : sqinfo->name);
        fprintf (outf, "AC   %s\n",
                 (sqinfo->flags & SQINFO_ACC) ? sqinfo->acc : "-");
        fprintf (outf, "DE   %s\n",
                 (sqinfo->flags & SQINFO_DESC) ? sqinfo->desc : "-");
        fprintf (outf, "SQ             %d BP\n", seqlen);
        strcpy (endstr, "\n//");	/* 11Oct90: bug fix */
        tab = 5;	 /** added 31jan91 */
        spacer = 11; /** added 31jan91 */
        break;

    case SQFILE_GCG:
        fprintf (outf, "%s\n", sqinfo->name);
        if (sqinfo->flags & SQINFO_ACC) {
            fprintf (outf, "ACCESSION   %s\n", sqinfo->acc);
        }
        if (sqinfo->flags & SQINFO_DESC) {
            fprintf (outf, "DEFINITION  %s\n", sqinfo->desc);
        }
        fprintf (outf, "    %s  Length: %d  (today)  Check: %d  ..\n",
                 sqinfo->name, seqlen, checksum);
        spacer = 11;
        numline = 1;
        strcpy (endstr, "\n");	/* this is insurance to help prevent misreads at eof */
        break;

    case SQFILE_STRIDER:		/* ?? map ? */
        fprintf (outf, "; ### from DNA Strider ;-)\n");
        fprintf (outf, "; DNA sequence  %s, %d bases, %d checksum.\n;\n",
                 sqinfo->name, seqlen, checksum);
        strcpy (endstr, "\n//");
        break;

        /* SRE: Don had Zuker default to Pearson, which is not
           intuitive or helpful, since Zuker's MFOLD can't read
           Pearson format. More useful to use kIG */
    case SQFILE_ZUKER:
        which_case = 1;		/* MFOLD requires upper case. */
        /*FALLTHRU*/
    case SQFILE_IG:
        fprintf (outf, ";%s %s\n",
                 sqinfo->name, (sqinfo->flags & SQINFO_DESC) ? sqinfo->desc : "");
        fprintf (outf, "%s\n", sqinfo->name);
        strcpy (endstr, "1");	/* == linear dna */
        break;

    case SQFILE_RAW:		/* Raw: no header at all. */
        break;

    default:
    case SQFILE_FASTA:
        fprintf (outf, ">%s %s\n", sqinfo->name,
                 (sqinfo->flags & SQINFO_DESC) ? sqinfo->desc : "");
        break;
    }

    if (which_case == 1) {
        s2upper (seq);
    }
    if (which_case == 2) {
        s2lower (seq);
    }


    width = MIN (width, 100);
    for (i = 0, l = 0, ibase = 1, lines = 0; i < seqlen;) {
        if (l1 < 0) {
            l1 = 0;
        }
        else if (l1 == 0) {
            if (numline) {
                fprintf (outf, "%8d ", ibase);
            }
            for (j = 0; j < tab; j++) {
                fputc (' ', outf);
            }
        }
        if ((spacer != 0) && ((l + 1) % spacer == 1)) {
            s[l] = ' ';
            ss[l] = ' ';
            l++;
        }
        s[l] = seq[i];
        ss[l] = (sqinfo->flags & SQINFO_SS) ? sqinfo->ss[i] : '.';
        l++;
        i++;
        l1++;			/* don't count spaces for width */
        if (l1 == width || i == seqlen) {
            s[l] = ss[l] = '\0';
            l = 0;
            l1 = 0;
            if (dostruc) {
                fprintf (outf, "%s\n", s);
                if (numline) {
                    fprintf (outf, "         ");
                }
                for (j = 0; j < tab; j++) {
                    fputc (' ', outf);
                }
                if (i == seqlen) {
                    fprintf (outf, "%s%s\n", ss, endstr);
                }
                else {
                    fprintf (outf, "%s\n", ss);
                }
            }
            else {
                if (i == seqlen) {
                    fprintf (outf, "%s%s\n", s, endstr);
                }
                else {
                    fprintf (outf, "%s\n", s);
                }
            }
            lines++;
            ibase = i + 1;
        }
    }
    return lines;
}


/* Function: ReadMultipleRseqs()
 *
 * Purpose:  Open a data file and
 *           parse it into an array of rseqs (raw, unaligned
 *           sequences).
 *
 *           Caller is responsible for free'ing memory allocated
 *           to ret_rseqs, ret_weights, and ret_names.
 *
 *           Weights are currently only supported for MSF format.
 *           Sequences read from all other formats will be assigned
 *           weights of 1.0. If the caller isn't interested in
 *           weights, it passes NULL as ret_weights.
 *
 * Returns 1 on success. Returns 0 on failure and sets
 * squid_errno to indicate the cause.
 */
int
ReadMultipleRseqs (char *seqfile,
                   int fformat,
                   char ***ret_rseqs, SQINFO ** ret_sqinfo, int *ret_num)
{
    SQINFO *sqinfo;		/* array of sequence optional info         */
    SQFILE *dbfp;			/* open ptr for sequential access of file  */
    char **rseqs;			/* sequence array                          */
    int numalloced;		/* num of seqs currently alloced for       */
    int num;


    num = 0;
    numalloced = 16;
    rseqs = (char **) MallocOrDie (numalloced * sizeof (char *));
    sqinfo = (SQINFO *) MallocOrDie (numalloced * sizeof (SQINFO));
    if ((dbfp = SeqfileOpen (seqfile, fformat, NULL)) == NULL) {
        return 0;
    }

    while (ReadSeq (dbfp, dbfp->format, &rseqs[num], &(sqinfo[num]))) {
        num++;
        if (num == numalloced) {	/* more seqs coming, alloc more room */
            numalloced += 16;
            rseqs = (char **) ReallocOrDie (rseqs, numalloced * sizeof (char *));
            sqinfo = (SQINFO *) ReallocOrDie (sqinfo, numalloced * sizeof (SQINFO));
        }
    }
    SeqfileClose (dbfp);

    *ret_rseqs = rseqs;
    *ret_sqinfo = sqinfo;
    *ret_num = num;
    return 1;
}


/* Function: String2SeqfileFormat()
 * Date:     SRE, Sun Jun 27 15:25:54 1999 [TW 723 over Canadian Shield]
 *
 * Purpose:  Convert a string (e.g. from command line option arg)
 *           to a format code. Case insensitive. Return
 *           MSAFILE_UNKNOWN/SQFILE_UNKNOWN if string is bad.
 *           Uses codes defined in squid.h (unaligned formats) and
 *           msa.h (aligned formats).
 *
 * Args:     s   - string to convert; e.g. "stockholm"
 *
 * Returns:  format code; e.g. MSAFILE_STOCKHOLM
 */
int
String2SeqfileFormat (char *s)
{
    char *s2;
    int code = SQFILE_UNKNOWN;

    if (s == NULL) {
        return SQFILE_UNKNOWN;
    }
    s2 = sre_strdup (s, -1);
    s2upper (s2);

    if (strcmp (s2, "FASTA") == 0) {
        code = SQFILE_FASTA;
    }
    else if (strcmp (s2, "GENBANK") == 0) {
        code = SQFILE_GENBANK;
    }
    else if (strcmp (s2, "EMBL") == 0) {
        code = SQFILE_EMBL;
    }
    else if (strcmp (s2, "GCG") == 0) {
        code = SQFILE_GCG;
    }
    else if (strcmp (s2, "GCGDATA") == 0) {
        code = SQFILE_GCGDATA;
    }
    else if (strcmp (s2, "RAW") == 0) {
        code = SQFILE_RAW;
    }
    else if (strcmp (s2, "IG") == 0) {
        code = SQFILE_IG;
    }
    else if (strcmp (s2, "STRIDER") == 0) {
        code = SQFILE_STRIDER;
    }
    else if (strcmp (s2, "IDRAW") == 0) {
        code = SQFILE_IDRAW;
    }
    else if (strcmp (s2, "ZUKER") == 0) {
        code = SQFILE_ZUKER;
    }
    else if (strcmp (s2, "PIR") == 0) {
        code = SQFILE_PIR;
    }
    else if (strcmp (s2, "SQUID") == 0) {
        code = SQFILE_SQUID;
    }
    else if (strcmp (s2, "STOCKHOLM") == 0) {
        code = MSAFILE_STOCKHOLM;
    }
    else if (strcmp (s2, "SELEX") == 0) {
        code = MSAFILE_SELEX;
    }
    else if (strcmp (s2, "MSF") == 0) {
        code = MSAFILE_MSF;
    }
    else if (strcmp (s2, "CLUSTAL") == 0) {
        code = MSAFILE_CLUSTAL;
    }
    else if (strcmp (s2, "A2M") == 0) {
        code = MSAFILE_A2M;
    }
    else if (strcmp (s2, "PHYLIP") == 0) {
        code = MSAFILE_PHYLIP;
    }
    else if (strcmp (s2, "EPS") == 0) {
        code = MSAFILE_EPS;
    }

    free (s2);
    return code;
}

char *
SeqfileFormat2String (int code)
{
    switch (code) {
    case SQFILE_UNKNOWN:
        return "unknown";
    case SQFILE_FASTA:
        return "FASTA";
    case SQFILE_GENBANK:
        return "Genbank";
    case SQFILE_EMBL:
        return "EMBL";
    case SQFILE_GCG:
        return "GCG";
    case SQFILE_GCGDATA:
        return "GCG data library";
    case SQFILE_RAW:
        return "raw";
    case SQFILE_IG:
        return "Intelligenetics";
    case SQFILE_STRIDER:
        return "MacStrider";
    case SQFILE_IDRAW:
        return "Idraw Postscript";
    case SQFILE_ZUKER:
        return "Zuker";
    case SQFILE_PIR:
        return "PIR";
    case SQFILE_SQUID:
        return "SQUID";
    case MSAFILE_STOCKHOLM:
        return "Stockholm";
    case MSAFILE_SELEX:
        return "SELEX";
    case MSAFILE_MSF:
        return "MSF";
    case MSAFILE_CLUSTAL:
        return "Clustal";
    case MSAFILE_A2M:
        return "a2m";
    case MSAFILE_PHYLIP:
        return "Phylip";
    case MSAFILE_EPS:
        return "EPS";
    default:
        Die ("Bad code passed to MSAFormat2String()");
    }
    /*NOTREACHED*/ return NULL;
}


/* Function: MSAToSqinfo()
 * Date:     SRE, Tue Jul 20 14:36:56 1999 [St. Louis]
 *
 * Purpose:  Take an MSA and generate a SQINFO array suitable
 *           for use in annotating the unaligned sequences.
 *           Return the array.
 *
 *           Permanent temporary code. sqinfo was poorly designed.
 *           it must eventually be replaced, but the odds
 *           of this happening soon are nil, so I have to deal.
 *
 * Args:     msa   - the alignment
 *
 * Returns:  ptr to allocated sqinfo array.
 *           Freeing is ghastly: free in each individual sqinfo[i]
 *           with FreeSequence(NULL, &(sqinfo[i])), then
 *           free(sqinfo).
 */
SQINFO *
MSAToSqinfo (MSA * msa)
{
    int idx;
    SQINFO *sqinfo;

    sqinfo = MallocOrDie (sizeof (SQINFO) * msa->nseq);

    for (idx = 0; idx < msa->nseq; idx++) {
        sqinfo[idx].flags = 0;
        SetSeqinfoString (&(sqinfo[idx]), msa->sqname[idx], SQINFO_NAME);
        SetSeqinfoString (&(sqinfo[idx]),
                          MSAGetSeqAccession (msa, idx), SQINFO_ACC);
        SetSeqinfoString (&(sqinfo[idx]),
                          MSAGetSeqDescription (msa, idx), SQINFO_DESC);

        if (msa->ss != NULL && msa->ss[idx] != NULL) {
            MakeDealignedString (msa->aseq[idx], msa->alen,
                                 msa->ss[idx], &(sqinfo[idx].ss));
            sqinfo[idx].flags |= SQINFO_SS;
        }

        if (msa->sa != NULL && msa->sa[idx] != NULL) {
            MakeDealignedString (msa->aseq[idx], msa->alen,
                                 msa->sa[idx], &(sqinfo[idx].sa));
            sqinfo[idx].flags |= SQINFO_SA;
        }

        sqinfo[idx].len = DealignedLength (msa->aseq[idx]);
        sqinfo[idx].flags |= SQINFO_LEN;
    }
    return sqinfo;
}



/* cc -o sqio_test -DA_QUIET_DAY -L. sqio.c -lsquid */
#ifdef A_QUIET_DAY
#include "ssi.h"
int
main (int argc, char **argv)
{
    FILE *fp;
    char *filename;
    char *buf;
    int len;
    int mode = 3;
    SSIOFFSET off;

    filename = argv[1];

    if (mode == 1) {
        buf = malloc (sizeof (char) * 256);
        if ((fp = fopen (filename, "r")) == NULL) {
            Die ("open of %s failed", filename);
        }
        while (fgets (buf, 255, fp) != NULL);
        fclose (fp);
        free (buf);
    }
    else if (mode == 2) {
        if ((fp = fopen (filename, "r")) == NULL) {
            Die ("open of %s failed", filename);
        }
        buf = NULL;
        len = 0;
        while (sre_fgets (&buf, &len, fp) != NULL) {
            SSIGetFilePosition (fp, SSI_OFFSET_I32, &off);
        }
        fclose (fp);
        free (buf);
    }
    else if (mode == 3) {
        SQFILE *dbfp;
        SQINFO info;

        if ((dbfp = SeqfileOpen (filename, SQFILE_FASTA, NULL)) == NULL) {
            Die ("open of %s failed", filename);
        }
        while (ReadSeq (dbfp, dbfp->format, &buf, &info)) {
            SSIGetFilePosition (dbfp->f, SSI_OFFSET_I32, &off);
            FreeSequence (buf, &info);
        }
        SeqfileClose (dbfp);
    }

}


#endif
/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* squidcore.c
 * SRE, Sun Jun 20 17:19:04 1999 [Graeme's kitchen]
 *
 * Core functions for SQUID library.
 * CVS $Id: squidcore.c,v 1.3 2003/05/26 16:21:50 eddy Exp $
 */

#include "squidconf.h"
#include "squid.h"

#include <stdio.h>

/* Function: SqdBanner()
 * Date:     SRE, Sun Jun 20 17:19:41 1999 [Graeme's kitchen]
 *
 * Purpose:  Print a package version and copyright banner.
 *           Used by all the main()'s.
 *
 *    Expects to be able to pick up preprocessor #define's from squidconf.h:
 *    symbol           example
 *    ------           --------------
 *    SQUID_VERSION    "2.0.42"
 *    SQUID_DATE       "April 1999"
 *    SQUID_COPYRIGHT  "Copyright (C) 1992-1999 Washington University School of Medicine"
 *    SQUID_LICENSE    "Freely distributed under the GNU General Public License (GPL)."
 *
 *           This gives us a general mechanism to update release information
 *           without changing multiple points in the code.
 *
 * Args:     fp     - where to print it
 *           banner - one-line program description, e.g.:
 *                    "foobar - make bars from foo with elan"
 * Returns:  (void)
 */
void
SqdBanner (FILE * fp, char *banner)
{
    fprintf (fp, "%s\n", banner);
    fprintf (fp, "SQUID %s (%s)\n", SQUID_VERSION, SQUID_DATE);
    fprintf (fp, "%s\n", SQUID_COPYRIGHT);
    fprintf (fp, "%s\n", SQUID_LICENSE);
    fprintf (fp,
             "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
}


/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* sre_ctype.c
 *
 * For portability. Some systems have functions tolower, toupper
 * as macros (for instance, MIPS M-2000 RISC/os!)
 *
 * CVS $Id: sre_ctype.c,v 1.4 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <ctype.h>
#include "squid.h"

int
sre_tolower (int c)
{
    if (isupper (c)) {
        return tolower (c);
    }
    else {
        return c;
    }
}

int
sre_toupper (int c)
{
    if (islower (c)) {
        return toupper (c);
    }
    else {
        return c;
    }
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* sre_math.c
 *
 * Portability for and extensions to C math library.
 * RCS $Id: sre_math.c,v 1.13 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "squid.h"


/* Function: Linefit()
 *
 * Purpose:  Given points x[0..N-1] and y[0..N-1], fit to
 *           a straight line y = a + bx.
 *           a, b, and the linear correlation coefficient r
 *           are filled in for return.
 *
 * Args:     x     - x values of data
 *           y     - y values of data
 *           N     - number of data points
 *           ret_a - RETURN: intercept
 *           ret_b - RETURN: slope
 *           ret_r - RETURN: correlation coefficient
 *
 * Return:   1 on success, 0 on failure.
 */
int
Linefit (float *x, float *y, int N, float *ret_a, float *ret_b, float *ret_r)
{
    float xavg, yavg;
    float sxx, syy, sxy;
    int i;

    /* Calculate averages, xavg and yavg
     */
    xavg = yavg = 0.0;
    for (i = 0; i < N; i++) {
        xavg += x[i];
        yavg += y[i];
    }
    xavg /= (float) N;
    yavg /= (float) N;

    sxx = syy = sxy = 0.0;
    for (i = 0; i < N; i++) {
        sxx += (x[i] - xavg) * (x[i] - xavg);
        syy += (y[i] - yavg) * (y[i] - xavg);
        sxy += (x[i] - xavg) * (y[i] - yavg);
    }
    *ret_b = sxy / sxx;
    *ret_a = yavg - xavg * (*ret_b);
    *ret_r = sxy / (sqrt (sxx) * sqrt (syy));
    return 1;
}


/* Function: WeightedLinefit()
 *
 * Purpose:  Given points x[0..N-1] and y[0..N-1] with
 *           variances (measurement errors) var[0..N-1],
 *           fit to a straight line y = mx + b.
 *
 * Method:   Algorithm from Numerical Recipes in C, [Press88].
 *
 * Return:   (void)
 *           ret_m contains slope; ret_b contains intercept
 */
void
WeightedLinefit (float *x, float *y, float *var, int N, float *ret_m,
                 float *ret_b)
{
    int i;
    double s;
    double sx, sy;
    double sxx, sxy;
    double delta;
    double m, b;

    s = sx = sy = sxx = sxy = 0.;
    for (i = 0; i < N; i++) {
        s += 1. / var[i];
        sx += x[i] / var[i];
        sy += y[i] / var[i];
        sxx += x[i] * x[i] / var[i];
        sxy += x[i] * y[i] / var[i];
    }

    delta = s * sxx - (sx * sx);
    b = (sxx * sy - sx * sxy) / delta;
    m = (s * sxy - sx * sy) / delta;

    *ret_m = m;
    *ret_b = b;
}


/* Function: Gammln()
 *
 * Returns the natural log of the gamma function of x.
 * x is > 0.0.
 *
 * Adapted from a public domain implementation in the
 * NCBI core math library. Thanks to John Spouge and
 * the NCBI. (According to the NCBI, that's Dr. John
 * "Gammas Galore" Spouge to you, pal.)
 */
double
Gammln (double x)
{
    int i;
    double xx, tx;
    double tmp, value;
    static double cof[11] = {
        4.694580336184385e+04,
        -1.560605207784446e+05,
        2.065049568014106e+05,
        -1.388934775095388e+05,
        5.031796415085709e+04,
        -9.601592329182778e+03,
        8.785855930895250e+02,
        -3.155153906098611e+01,
        2.908143421162229e-01,
        -2.319827630494973e-04,
        1.251639670050933e-10
    };

    /* Protect against x=0. We see this in Dirichlet code,
     * for terms alpha = 0. This is a severe hack but it is effective
     * and (we think?) safe. (due to GJM)
     */
    if (x <= 0.0) {
        return 999999.;
    }

    xx = x - 1.0;
    tx = tmp = xx + 11.0;
    value = 1.0;
    for (i = 10; i >= 0; i--) {	
				/* sum least significant terms first */
        value += cof[i] / tmp;
        tmp -= 1.0;
    }
    value = log (value);
    tx += 0.5;
    value += 0.918938533 + (xx + 0.5) * log (tx) - tx;
    return value;
}


/* 2D matrix operations
 */
float **
FMX2Alloc (int rows, int cols)
{
    float **mx;
    int r;

    mx = (float **) MallocOrDie (sizeof (float *) * rows);
    mx[0] = (float *) MallocOrDie (sizeof (float) * rows * cols);
    for (r = 1; r < rows; r++) {
        mx[r] = mx[0] + r * cols;
    }
    return mx;
}

void
FMX2Free (float **mx)
{
    free (mx[0]);
    free (mx);
}

double **
DMX2Alloc (int rows, int cols)
{
    double **mx;
    int r;

    mx = (double **) MallocOrDie (sizeof (double *) * rows);
    mx[0] = (double *) MallocOrDie (sizeof (double) * rows * cols);
    for (r = 1; r < rows; r++) {
        mx[r] = mx[0] + r * cols;
    }
    return mx;
}

void
DMX2Free (double **mx)
{
    free (mx[0]);
    free (mx);
}

/* Function: FMX2Multiply()
 *
 * Purpose:  Matrix multiplication.
 *           Multiply an m x p matrix A by a p x n matrix B,
 *           giving an m x n matrix C.
 *           Matrix C must be a preallocated matrix of the right
 *           size.
 */
void
FMX2Multiply (float **A, float **B, float **C, int m, int p, int n)
{
    int i, j, k;

    for (i = 0; i < m; i++) {
        for (j = 0; j < n; j++) {
            C[i][j] = 0.;
            for (k = 0; k < p; k++) {
                C[i][j] += A[i][p] * B[p][j];
            }
        }
		}		
}


/* Function: IncompleteGamma()
 *
 * Purpose:  Returns 1 - P(a,x) where:
 *           P(a,x) = \frac{1}{\Gamma(a)} \int_{0}^{x} t^{a-1} e^{-t} dt
 *                  = \frac{\gamma(a,x)}{\Gamma(a)}
 *                  = 1 - \frac{\Gamma(a,x)}{\Gamma(a)}
 *
 *           Used in a chi-squared test: for a X^2 statistic x
 *           with v degrees of freedom, call:
 *                  p = IncompleteGamma(v/2., x/2.)
 *           to get the probability p that a chi-squared value
 *           greater than x could be obtained by chance even for
 *           a correct model. (i.e. p should be large, say
 *           0.95 or more).
 *
 * Method:   Based on ideas from Numerical Recipes in C, Press et al.,
 *           Cambridge University Press, 1988.
 *
 * Args:     a  - for instance, degrees of freedom / 2     [a > 0]
 *           x  - for instance, chi-squared statistic / 2  [x >= 0]
 *
 * Return:   1 - P(a,x).
 */
double
IncompleteGamma (double a, double x)
{
    int iter;			/* iteration counter */

    if (a <= 0.) {
        Die ("IncompleteGamma(): a must be > 0");
    }
    if (x < 0.) {
        Die ("IncompleteGamma(): x must be >= 0");
    }

    /* For x > a + 1 the following gives rapid convergence;
     * calculate 1 - P(a,x) = \frac{\Gamma(a,x)}{\Gamma(a)}:
     *     use a continued fraction development for \Gamma(a,x).
     */
    if (x > a + 1) {
        double oldp;		/* previous value of p    */
        double nu0, nu1;		/* numerators for continued fraction calc   */
        double de0, de1;		/* denominators for continued fraction calc */

        nu0 = 0.;			/* A_0 = 0       */
        de0 = 1.;			/* B_0 = 1       */
        nu1 = 1.;			/* A_1 = 1       */
        de1 = x;			/* B_1 = x       */

        oldp = nu1;
        for (iter = 1; iter < 100; iter++) {
            /* Continued fraction development:
             * set A_j = b_j A_j-1 + a_j A_j-2
             *     B_j = b_j B_j-1 + a_j B_j-2
             * We start with A_2, B_2.
             */
            /* j = even: a_j = iter-a, b_j = 1 */
            /* A,B_j-2 are in nu0, de0; A,B_j-1 are in nu1,de1 */
            nu0 = nu1 + ((double) iter - a) * nu0;
            de0 = de1 + ((double) iter - a) * de0;

            /* j = odd: a_j = iter, b_j = x */
            /* A,B_j-2 are in nu1, de1; A,B_j-1 in nu0,de0 */
            nu1 = x * nu0 + (double) iter *nu1;
            de1 = x * de0 + (double) iter *de1;

            /* rescale */
            if (de1 != 0.) {
                nu0 /= de1;
                de0 /= de1;
                nu1 /= de1;
                de1 = 1.;
            }
            /* check for convergence */
            if (fabs ((nu1 - oldp) / nu1) < 1.e-7) {
                return nu1 * exp (a * log (x) - x - Gammln (a));
            }

            oldp = nu1;
        }
        Die
        ("IncompleteGamma(): failed to converge using continued fraction approx");
    }
    else {			/* x <= a+1 */

        double p;			/* current sum               */
        double val;			/* current value used in sum */

        /* For x <= a+1 we use a convergent series instead:
         *   P(a,x) = \frac{\gamma(a,x)}{\Gamma(a)},
         * where
         *   \gamma(a,x) = e^{-x}x^a \sum_{n=0}{\infty} \frac{\Gamma{a}}{\Gamma{a+1+n}} x^n
         * which looks appalling but the sum is in fact rearrangeable to
         * a simple series without the \Gamma functions:
         *   = \frac{1}{a} + \frac{x}{a(a+1)} + \frac{x^2}{a(a+1)(a+2)} ...
         * and it's obvious that this should converge nicely for x <= a+1.
         */

        p = val = 1. / a;
        for (iter = 1; iter < 10000; iter++) {
            val *= x / (a + (double) iter);
            p += val;

            if (fabs (val / p) < 1.e-7) {
                return 1. - p * exp (a * log (x) - x - Gammln (a));
            }
        }
        Die ("IncompleteGamma(): failed to converge using series approx");
    }
    /*NOTREACHED*/ return 0.;
}

/* sre_random.c
 *
 * Portable random number generator, and sampling routines.
 *
 * SRE, Tue Oct  1 15:24:11 2002 [St. Louis]
 * CVS $Id: sre_random.c,v 1.2 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "sre_random.h"

static int sre_randseed = 42;	/* default seed for sre_random()   */

/* Function: sre_random()
 *
 * Purpose:  Return a uniform deviate x, 0.0 <= x < 1.0.
 *
 *           sre_randseed is a static variable, set
 *           by sre_srandom(). When it is non-zero,
 *           we re-seed.
 *
 *           Implements L'Ecuyer's algorithm for combining output
 *           of two linear congruential generators, plus a Bays-Durham
 *           shuffle. This is essentially ran2() from Numerical Recipes,
 *           sans their nonhelpful Rand/McNally-esque code obfuscation.
 *
 *           Overflow errors are avoided by Schrage's algorithm:
 *               az % m = a(z%q) - r(z/q) (+m if <0)
 *           where q=m/a, r=m%a
 *
 *           Requires that long int's have at least 32 bits.
 *           This function uses statics and is NOT THREADSAFE.
 *
 * Reference: Press et al. Numerical Recipes in C, 1992.
 *
 * Reliable and portable, but slow. Benchmarks on wrasse,
 * using Linux gcc and Linux glibc rand() (see randspeed, in Testsuite):
 *     sre_random():    0.5 usec/call
 *     rand():          0.2 usec/call
 */
double
sre_random (void)
{
    static long rnd1;		/* random number from LCG1 */
    static long rnd2;		/* random number from LCG2 */
    static long rnd;		/* random number we return */
    static long tbl[64];		/* table for Bays/Durham shuffle */
    long x, y;
    int i;

    /* Magic numbers a1,m1, a2,m2 from L'Ecuyer, for 2 LCGs.
     * q,r derive from them (q=m/a, r=m%a) and are needed for Schrage's algorithm.
     */
    long a1 = 40014;
    long m1 = 2147483563;
    long q1 = 53668;
    long r1 = 12211;

    long a2 = 40692;
    long m2 = 2147483399;
    long q2 = 52774;
    long r2 = 3791;

    if (sre_randseed > 0) {
        rnd1 = sre_randseed;
        rnd2 = sre_randseed;
        /* Fill the table for Bays/Durham */
        for (i = 0; i < 64; i++) {
            x = a1 * (rnd1 % q1);	/* LCG1 in action... */
            y = r1 * (rnd1 / q1);
            rnd1 = x - y;
            if (rnd1 < 0) {
                rnd1 += m1;
            }

            x = a2 * (rnd2 % q2);	/* LCG2 in action... */
            y = r2 * (rnd2 / q2);
            rnd2 = x - y;
            if (rnd2 < 0) {
                rnd2 += m2;
            }

            tbl[i] = rnd1 - rnd2;
            if (tbl[i] < 0) {
                tbl[i] += m1;
            }
        }
        sre_randseed = 0;		/* drop the flag. */
    }				/* end of initialization */


    x = a1 * (rnd1 % q1);		/* LCG1 in action... */
    y = r1 * (rnd1 / q1);
    rnd1 = x - y;
    if (rnd1 < 0) {
        rnd1 += m1;
    }

    x = a2 * (rnd2 % q2);		/* LCG2 in action... */
    y = r2 * (rnd2 / q2);
    rnd2 = x - y;
    if (rnd2 < 0) {
        rnd2 += m2;
    }

    /* Choose our random number from the table... */
    i = (int) (((double) rnd / (double) m1) * 64.);
    rnd = tbl[i];
    /* and replace with a new number by L'Ecuyer. */
    tbl[i] = rnd1 - rnd2;
    if (tbl[i] < 0) {
        tbl[i] += m1;
    }

    return ((double) rnd / (double) m1);
}

/* Function: sre_srandom()
 *
 * Purpose:  Initialize with a random seed. Seed must be
 *           >= 0 to work; we silently enforce this.
 */
void
sre_srandom (int seed)
{
    if (seed < 0) {
        seed = -1 * seed;
    }
    if (seed == 0) {
        seed = 42;
    }
    sre_randseed = seed;
}

/* Function: sre_random_positive()
 * Date:     SRE, Wed Apr 17 13:34:32 2002 [St. Louis]
 *
 * Purpose:  Assure 0 < x < 1 (positive uniform deviate)
 */
double
sre_random_positive (void)
{
    double x;
    do {
        x = sre_random ();
    } while (x == 0.0);
    return x;
}

/* Function: ExponentialRandom()
 * Date:     SRE, Mon Sep  6 21:24:29 1999 [St. Louis]
 *
 * Purpose:  Pick an exponentially distributed random variable
 *           0 > x >= infinity
 *
 * Args:     (void)
 *
 * Returns:  x
 */
double
ExponentialRandom (void)
{
    double x;

    do {
        x = sre_random ();
    }
    while (x == 0.0);
    return -log (x);
}

/* Function: Gaussrandom()
 *
 * Pick a Gaussian-distributed random variable
 * with some mean and standard deviation, and
 * return it.
 *
 * Based on RANLIB.c public domain implementation.
 * Thanks to the authors, Barry W. Brown and James Lovato,
 * University of Texas, M.D. Anderson Cancer Center, Houston TX.
 * Their implementation is from Ahrens and Dieter, "Extensions
 * of Forsythe's method for random sampling from the normal
 * distribution", Math. Comput. 27:927-937 (1973).
 *
 * Impenetrability of the code is to be blamed on its FORTRAN/f2c lineage.
 *
 */
double
Gaussrandom (double mean, double stddev)
{
    static double a[32] = {
        0.0, 3.917609E-2, 7.841241E-2, 0.11777, 0.1573107, 0.1970991, 0.2372021,
        0.2776904, 0.3186394, 0.36013, 0.4022501, 0.4450965, 0.4887764,
        0.5334097, 0.5791322,
        0.626099, 0.6744898, 0.7245144, 0.7764218, 0.8305109, 0.8871466,
        0.9467818,
        1.00999, 1.077516, 1.150349, 1.229859, 1.318011, 1.417797, 1.534121,
        1.67594,
        1.862732, 2.153875
    };
    static double d[31] = {
        0.0, 0.0, 0.0, 0.0, 0.0, 0.2636843, 0.2425085, 0.2255674, 0.2116342,
        0.1999243,
        0.1899108, 0.1812252, 0.1736014, 0.1668419, 0.1607967, 0.1553497,
        0.1504094,
        0.1459026, 0.14177, 0.1379632, 0.1344418, 0.1311722, 0.128126, 0.1252791,
        0.1226109, 0.1201036, 0.1177417, 0.1155119, 0.1134023, 0.1114027,
        0.1095039
    };
    static double t[31] = {
        7.673828E-4, 2.30687E-3, 3.860618E-3, 5.438454E-3, 7.0507E-3, 8.708396E-3,
        1.042357E-2, 1.220953E-2, 1.408125E-2, 1.605579E-2, 1.81529E-2,
        2.039573E-2,
        2.281177E-2, 2.543407E-2, 2.830296E-2, 3.146822E-2, 3.499233E-2,
        3.895483E-2,
        4.345878E-2, 4.864035E-2, 5.468334E-2, 6.184222E-2, 7.047983E-2,
        8.113195E-2,
        9.462444E-2, 0.1123001, 0.136498, 0.1716886, 0.2276241, 0.330498,
        0.5847031
    };
    static double h[31] = {
        3.920617E-2, 3.932705E-2, 3.951E-2, 3.975703E-2, 4.007093E-2, 4.045533E-2,
        4.091481E-2, 4.145507E-2, 4.208311E-2, 4.280748E-2, 4.363863E-2,
        4.458932E-2,
        4.567523E-2, 4.691571E-2, 4.833487E-2, 4.996298E-2, 5.183859E-2,
        5.401138E-2,
        5.654656E-2, 5.95313E-2, 6.308489E-2, 6.737503E-2, 7.264544E-2,
        7.926471E-2,
        8.781922E-2, 9.930398E-2, 0.11556, 0.1404344, 0.1836142, 0.2790016,
        0.7010474
    };
    static long i;
    static double snorm, u, s, ustar, aa, w, y, tt;

    u = sre_random ();
    s = 0.0;
    if (u > 0.5) {
        s = 1.0;
    }
    u += (u - s);
    u = 32.0 * u;
    i = (long) (u);
    if (i == 32) {
        i = 31;
    }
    if (i == 0) {
        goto S100;
    }
    /*
     * START CENTER
     */
    ustar = u - (double) i;
    aa = *(a + i - 1);
S40:
    if (ustar <= *(t + i - 1)) {
        goto S60;
    }
    w = (ustar - *(t + i - 1)) ** (h + i - 1);
S50:
    /*
     * EXIT   (BOTH CASES)
     */
    y = aa + w;
    snorm = y;
    if (s == 1.0) {
        snorm = -y;
    }
    return (stddev * snorm + mean);
S60:
    /*
     * CENTER CONTINUED
     */
    u = sre_random ();
    w = u * (*(a + i) - aa);
    tt = (0.5 * w + aa) * w;
    goto S80;
S70:
    tt = u;
    ustar = sre_random ();
S80:
    if (ustar > tt) {
        goto S50;
    }
    u = sre_random ();
    if (ustar >= u) {
        goto S70;
    }
    ustar = sre_random ();
    goto S40;
S100:
    /*
     * START TAIL
     */
    i = 6;
    aa = *(a + 31);
    goto S120;
S110:
    aa += *(d + i - 1);
    i += 1;
S120:
    u += u;
    if (u < 1.0) {
        goto S110;
    }
    u -= 1.0;
S140:
    w = u ** (d + i - 1);
    tt = (0.5 * w + aa) * w;
    goto S160;
S150:
    tt = u;
S160:
    ustar = sre_random ();
    if (ustar > tt) {
        goto S50;
    }
    u = sre_random ();
    if (ustar >= u) {
        goto S150;
    }
    u = sre_random ();
    goto S140;
}


/* Functions: DChoose(), FChoose()
 *
 * Purpose:   Make a random choice from a normalized distribution.
 *            DChoose() is for double-precision vectors;
 *            FChoose() is for single-precision float vectors.
 *            Returns the number of the choice.
 */
int
DChoose (double *p, int N)
{
    double roll;			/* random fraction */
    double sum;			/* integrated prob */
    int i;			/* counter over the probs */

    roll = sre_random ();
    sum = 0.0;
    for (i = 0; i < N; i++) {
        sum += p[i];
        if (roll < sum) {
            return i;
        }
    }
    return (int) (sre_random () * N);	/* bulletproof */
}

int
FChoose (float *p, int N)
{
    float roll;			/* random fraction */
    float sum;			/* integrated prob */
    int i;			/* counter over the probs */

    roll = sre_random ();
    sum = 0.0;
    for (i = 0; i < N; i++) {
        sum += p[i];
        if (roll < sum) {
            return i;
        }
    }
    return (int) (sre_random () * N);	/* bulletproof */
}


/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* sre_string.c
 *
 * my library of extra string functions. Some for portability
 * across UNIXes
 *
 * CVS $Id: sre_string.c,v 1.14 2003/05/26 16:21:50 eddy Exp $
 */

#include "squidconf.h"
#include "squid.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>


/* Function: Strdup()
 *
 * Purpose:  Implementation of the common (but non-ANSI) function
 *           strdup(). Robust against being passed a NULL pointer.
 *
 */
char *
Strdup (char *s)
{
    char *new;
    if (s == NULL) {
        return NULL;
    }
    if ((new = (char *) malloc (strlen (s) + 1)) == NULL) {
        return NULL;
    }
    strcpy (new, s);
    return new;
}

/* Function: StringChop()
 * Date:     SRE, Wed Oct 29 12:10:02 1997 [TWA 721]
 *
 * Purpose:  Chop trailing whitespace off of a string.
 */
void
StringChop (char *s)
{
    int i;

    i = strlen (s) - 1;		/* set i at last char in string     */
    while (i >= 0 && isspace ((int) s[i])) {
        i--;    /* i now at last non-whitespace char, or -1 */
    }
    s[i + 1] = '\0';
}

int
Strinsert (char *s1,		/* string to insert a char into  */
           char c,		/* char to insert                */
           int pos)
{   /* position in s1 to insert c at */
    char oldc;
    char *s;

    for (s = s1 + pos; c; s++) {
        /* swap current char for inserted one */
        oldc = *s;			/* pick up current */
        *s = c;			/* put down inserted one    */
        c = oldc;			/* old becomes next to insert */
    }
    *s = '\0';

    return 1;
}


int
Strdelete (char *s1,		/* string to delete a char from       */
           int pos)
{   /* position of char to delete 0..n-1  */
    char *s;

    for (s = s1 + pos; *s; s++) {
        *s = *(s + 1);
    }

    return 1;
}

void
s2lower (char *s)
{
    for (; *s != '\0'; s++) {
        *s = sre_tolower ((int) *s);
    }
}

void
s2upper (char *s)
{
    for (; *s != '\0'; s++) {
        *s = sre_toupper ((int) *s);
    }
}


void *
sre_malloc (char *file, int line, size_t size)
{
    void *ptr;

    SQD_DPRINTF3 (("MALLOC: %d bytes (file %s line %d)\n", size, file, line));
    if ((ptr = malloc (size)) == NULL) {
        Die ("malloc of %ld bytes failed: file %s line %d", size, file, line);
    }
    return ptr;
}

void *
sre_realloc (char *file, int line, void *p, size_t size)
{
    void *ptr;

    if ((ptr = realloc (p, size)) == NULL) {
        Die ("realloc of %ld bytes failed: file %s line %d", size, file, line);
    }
    return ptr;
}



/* Function: Free2DArray(), Free3DArray()
 * Date:     SRE, Tue Jun  1 14:47:14 1999 [St. Louis]
 *
 * Purpose:  Convenience functions for free'ing 2D
 *           and 3D pointer arrays. Tolerates any of the
 *           pointers being NULL, to allow "sparse"
 *           arrays.
 *
 * Args:     p     - array to be freed
 *           dim1  - n for first dimension
 *           dim2  - n for second dimension
 *
 *           e.g. a 2d array is indexed p[0..dim1-1][]
 *                a 3D array is indexed p[0..dim1-1][0..dim2-1][]
 *
 * Returns:  void
 *
 * Diagnostics: (void)
 *              "never fails"
 */
void
Free2DArray (void **p, int dim1)
{
    int i;

    if (p != NULL) {
        for (i = 0; i < dim1; i++) {
            if (p[i] != NULL) {
                free (p[i]);
            }
				}		
        free (p);
    }
}

void
Free3DArray (void ***p, int dim1, int dim2)
{
    int i, j;

    if (p != NULL) {
        for (i = 0; i < dim1; i++) {
            if (p[i] != NULL) {
                for (j = 0; j < dim2; j++) {
                    if (p[i][j] != NULL) {
                        free (p[i][j]);
                    }
								}		
                free (p[i]);
            }
				}		
        free (p);
    }
}


/* Function: RandomSequence()
 *
 * Purpose:  Generate an iid symbol sequence according
 *           to some alphabet, alphabet_size, probability
 *           distribution, and length. Return the
 *           sequence.
 *
 * Args:     alphabet  - e.g. "ACGT"
 *           p         - probability distribution [0..n-1]
 *           n         - number of symbols in alphabet
 *           len       - length of generated sequence
 *
 * Return:   ptr to random sequence, or NULL on failure.
 */
char *
RandomSequence (char *alphabet, float *p, int n, int len)
{
    char *s;
    int x;

		/* FChoose() variables */
		float roll;			/* random fraction */
    float sum;			/* integrated prob */
    int i;			/* counter over the probs */
		int output; /* Output value of FChoose() inlining */


    s = (char *) MallocOrDie (sizeof (char) * (len + 1));
    for (x = 0; x < len; x++) {
        //s[x] = alphabet[FChoose (p, n)];
				
				/* Begin inlining FChoose() function */
				roll = sre_random();
				sum = 0.0;
				for (i = 0; i < n; i++) {
        	sum += p[i];
        	if (roll < sum) {
          	output = i; /* Exit point 1 */
						goto exit_FChoose;
					}
    		}
    		output = (int) (sre_random () * n);	/* Exit Point 2 */
				exit_FChoose:
				/* End inlining FChoose() function */
				
				s[x] = alphabet[output];


    }
    s[x] = '\0';
    return s;
}

/* Function: sre_fgets()
 * Date:     SRE, Thu May 13 10:56:28 1999 [St. Louis]
 *
 * Purpose:  Dynamic allocation version of fgets(),
 *           capable of reading unlimited line lengths.
 *
 * Args:     buf - ptr to a string (may be reallocated)
 *           n   - ptr to current allocated length of buf,
 *                 (may be changed)
 *           fp  - open file ptr for reading
 *
 *           Before the first call to sre_fgets(),
 *           buf should be initialized to NULL and n to 0.
 *           They're a linked pair, so don't muck with the
 *           allocation of buf or the value of n while
 *           you're still doing sre_fgets() calls with them.
 *
 * Returns:  ptr to the buffer on success.
 *           NULL on EOF (buf isn't to be used in this case)
 *           sre_fgets() *always* results in an allocation
 *           in buf.
 *
 *           The reason to have it return a ptr to buf
 *           is that it makes wrapper macros easy; see
 *           MSAFileGetLine() for an example.
 *
 * Example:  char *buf;
 *           int   n;
 *           FILE *fp;
 *
 *           fp  = fopen("my_file", "r");
 *           buf = NULL;
 *           n   = 0;
 *           while (sre_fgets(&buf, &n, fp) != NULL)
 *           {
 *             do stuff with buf;
 *           }
 */
char *
sre_fgets (char **buf, int *n, FILE * fp)
{
    char *s;
    int len;
    int pos;

    if (*n == 0) {
        *buf = MallocOrDie (sizeof (char) * 128);
        *n = 128;
    }

    /* Simple case 1. We're sitting at EOF, or there's an error.
     *                fgets() returns NULL, so we return NULL.
     */
    if (fgets (*buf, *n, fp) == NULL) {
        return NULL;
    }

    /* Simple case 2. fgets() got a string, and it reached EOF.
     *                return success status, so caller can use
     *                the last line; on the next call we'll
     *                return the 0 for the EOF.
     */
    if (feof (fp)) {
        return *buf;
    }

    /* Simple case 3. We got a complete string, with \n,
     *                and don't need to extend the buffer.
     */
    len = strlen (*buf);
    if ((*buf)[len - 1] == '\n') {
        return *buf;
    }

    /* The case we're waiting for. We have an incomplete string,
     * and we have to extend the buffer one or more times. Make
     * sure we overwrite the previous fgets's \0 (hence +(n-1)
     * in first step, rather than 128, and reads of 129, not 128).
     */
    pos = (*n) - 1;
    while (1) {
        *n += 128;
        *buf = ReallocOrDie (*buf, sizeof (char) * (*n));
        s = *buf + pos;
        if (fgets (s, 129, fp) == NULL) {
            return *buf;
        }
        len = strlen (s);
        if (s[len - 1] == '\n') {
            return *buf;
        }
        pos += 128;
    }
    /*NOTREACHED*/

}

/* Function: sre_strcat()
 * Date:     SRE, Thu May 13 09:36:32 1999 [St. Louis]
 *
 * Purpose:  Dynamic memory version of strcat().
 *           appends src to the string that dest points to,
 *           extending allocation for dest if necessary.
 *
 *           One timing experiment (100 successive appends of
 *           1-255 char) shows sre_strcat() has about a 20%
 *           overhead relative to strcat(). However, if optional
 *           length info is passed, sre_strcat() is about 30%
 *           faster than strcat().
 *
 * Args:     dest  - ptr to string (char **), '\0' terminated
 *           ldest - length of dest, if known; or -1 if length unknown.
 *           src   - string to append to dest, '\0' terminated
 *           lsrc  - length of src, if known; or -1 if length unknown.
 *
 *           dest may be NULL, in which case this is
 *           the equivalent of dest = Strdup(src).
 *
 *           src may also be NULL, in which case
 *           dest is unmodified (but why would you want to pass
 *           a NULL src?)
 *
 *           if both dest and src are NULL, dest is
 *           unmodified; it stays NULL.
 *
 *           the length parameters are optional. If a -1
 *           is passed, sre_strcat() will call strlen() to
 *           determine the length itself. Passing length
 *           info saves the strlen() calls and can speed things
 *           up if lots of successive appends need to be done.
 *
 * Returns:  new length of dest (>=0 on success);
 *           dest is (probably) reallocated, and modified
 *           to a longer string, '\0' terminated.
 */
int
sre_strcat (char **dest, int ldest, char *src, int lsrc)
{
    int len1, len2;

    if (ldest < 0) {
        len1 = ((*dest == NULL) ? 0 : strlen (*dest));
    }
    else {
        len1 = ldest;
    }

    if (lsrc < 0) {
        len2 = ((src == NULL) ? 0 : strlen (src));
    }
    else {
        len2 = lsrc;
    }

    if (len2 == 0) {
        return len1;
    }

    if (*dest == NULL) {
        *dest = MallocOrDie (sizeof (char) * (len2 + 1));
    }
    else {
        *dest = ReallocOrDie (*dest, sizeof (char) * (len1 + len2 + 1));
    }

    memcpy ((*dest) + len1, src, len2 + 1);
    return len1 + len2;
}

/* Function: sre_strtok()
 * Date:     SRE, Wed May 19 16:30:20 1999 [St. Louis]
 *
 * Purpose:  Thread-safe version of strtok().
 *
 *           Returns ptr to next token in a string: skips
 *            until it reaches a character that is not in the delim
 *            string, and sets beginning of token. Skips to
 *            next delim character (or '\0') to set the end; replaces that
 *            character with '\0'.
 *           If there's still more string left, sets s to point to next
 *            character after the '\0' that was written, so successive
 *            calls extract tokens in succession. If there was no string
 *            left, s points at the terminal '\0'.
 *
 *           If no token is found, returns NULL.
 *
 *           Also returns the length of the token, which
 *           may save us a strlen() call in some applications.
 *
 * Limitations:
 *           *s can't be a constant string, since we write to it.
 *
 * Example:
 *           char *tok;
 *           int   len;
 *           char *s;
 *           char  buf[50] = "This is  a sentence.";
 *
 *           s = buf;
 *           tok = sre_strtok(&s, " ", &len);
 *                tok is "This"; s is "is  a sentence."; len is 4.
 *           tok = sre_strtok(&s, " ", &len);
 *                tok is "is"; s is " a sentence."; len is 2.
 *           tok = sre_strtok(&s, " ", &len);
 *                tok is "a"; s is "sentence."; len is 1.
 *           tok = sre_strtok(&s, " ", &len);
 *                tok is "sentence."; s is "\0"; len is 9.
 *           tok = sre_strtok(&s, " ", &len);
 *                tok is NULL; s is "\0", len is undefined.
 *
 * Args:     s     - a tmp, modifiable ptr to string
 *           delim - characters that delimits tokens
 *           len   - RETURN: length of token; pass NULL if not wanted
 *
 * Returns:  ptr to next token, or NULL if there aren't any.
 */
char *
sre_strtok (char **s, char *delim, int *len)
{
    char *begin, *end;
    int n;

    begin = *s;
    begin += strspn (begin, delim);
    if (!*begin) {
        return NULL;
    }

    n = strcspn (begin, delim);
    end = begin + n;
    if (*end == '\0') {
        *s = end;
    }
    else {
        *end = '\0';
        *s = end + 1;
    }

    if (len != NULL) {
        *len = n;
    }
    return begin;
}



/* Function: sre_strdup()
 * Date:     SRE, Wed May 19 17:57:28 1999 [St. Louis]
 *
 * Purpose:  A version of the common but non-ANSI strdup()
 *           function. Can pass len, if known, to save a
 *           strlen() call.
 *
 * Args:     s  - string to duplicate
 *           n  - length of string, if known; -1 if unknown.
 *
 * Returns:  allocated copy of string.
 *           NULL on failure.
 */
char *
sre_strdup (char *s, int n)
{
    char *new;

    if (s == NULL) {
        return NULL;
    }
    if (n < 0) {
        n = strlen (s);
    }
    new = MallocOrDie (sizeof (char) * (n + 1));
    strcpy (new, s);
    return new;
}


/* Function: sre_strncpy()
 * Date:     SRE, Tue Jun 22 10:10:46 1999 [Sanger Centre]
 *
 * Purpose:  a strncpy() that makes sure it adds a trailing \0.
 *
 * Args:     s1   - string to copy to (allocated n+1 or larger)
 *           s2   - string to copy from
 *           n    - number of chars to copy
 *
 * Returns:  s1.
 *           Done only for consistency with strncpy(). Not clear
 *           why it's useful for a strncpy() to return s1.
 */
char *
sre_strncpy (char *s1, char *s2, int n)
{
    strncpy (s1, s2, n);
    s1[n] = '\0';
    return s1;
}

/* Function: IsBlankline()
 * Date:     SRE, Fri Jun 18 14:36:08 1999 [St. Louis]
 *
 * Purpose:  Returns TRUE if string consists solely of whitespace.
 *
 * Args:     s   - string to check
 */
int
IsBlankline (char *s)
{
    for (; *s != '\0'; s++) {
        if (!isspace ((int) *s)) {
            return FALSE;
        }
		}		
    return TRUE;
}



#ifdef CUBS_WIN
/* A timing test for sre_strcat()
 * cc -O2 -g sre_string.c sre_ctype.c sqerror.c sre_math.c hsregex.c -lm
 * 15.200u - 5.360u = 9.84u if sre_strcat() with no length info passed
 * 13.660u - 5.360u = 8.30u if strcat(), with a single malloc().
 * 11.370u - 5.360u = 6.01u if sre_strcat() with length info passed.
 */
int
main (void)
{
    float p[4] = { 0.25, 0.25, 0.25, 0.25 };
    int buflen;
    int len;
    int nappends;
    int nstrings;
    char *s1 = NULL;
    char *s2;
    int i;

    nappends = 100;
    nstrings = 1000;
    while (nstrings--) {
        /* s1 = malloc(sizeof(char) * (255*nappends+1));
           s1[0] = '\0';
         */

        s1 = NULL;
        len = 0;
        for (i = 0; i < nappends; i++) {
            buflen = CHOOSE (255) + 1;
            s2 = RandomSequence ("ACGT", p, 4, buflen);

            /* strcat(s1,s2); */
            if ((len = sre_strcat (&s1, len, s2, buflen)) < 0) {
                exit (1);
            }
            free (s2);
        }
        free (s1);
    }
    exit (0);
}
#endif /*CUBS_WIN */
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#if !defined(SPEC_CPU_WINDOWS)
#include <unistd.h>
#endif /* SPEC_CPU_WINDOWS */
#include "squid.h"
#include "ssi.h"

static sqd_uint32 v20magicssi = 0xf3f3e9b1;	/* SSI 1.0: "ssi1" + 0x80808080 */
static sqd_uint32 v20swapssi = 0xb1e9f3f3;	/* byteswapped */

static int read_i16 (FILE * fp, sqd_uint16 * ret_result);
static int read_i32 (FILE * fp, sqd_uint32 * ret_result);
static int read_i64 (FILE * fp, sqd_uint64 * ret_result);
static int read_offset (FILE * fp, char mode, SSIOFFSET * ret_offset);
static int write_i16 (FILE * fp, sqd_uint16 n);
static int write_i32 (FILE * fp, sqd_uint32 n);
static int write_i64 (FILE * fp, sqd_uint64 n);
static int write_offset (FILE * fp, SSIOFFSET * offset);
static int binary_search (SSIFILE * sfp, char *key, int klen,
                          SSIOFFSET * base, sqd_uint32 recsize,
                          sqd_uint32 maxidx);
static int indexfile_position (SSIFILE * sfp, SSIOFFSET * base,
                               sqd_uint32 len, sqd_uint32 n);
static void clear_ssifile (SSIFILE * sfp);
static sqd_uint64 current_index_size (SSIINDEX * g);
static int activate_external_sort (SSIINDEX * g);
static int load_indexfile (SSIFILE * sfp);
static int parse_pkey_info (char *buf, char mode, struct ssipkey_s *pkey);
static int parse_skey_info (char *buf, struct ssiskey_s *skey);

/* Function: SSIOpen()
 * Date:     SRE, Sun Dec 31 12:40:03 2000 [St. Louis]
 *
 * Purpose:  Opens the SSI index file {filename} and returns
 *           a SSIFILE * stream thru {ret_sfp}.
 *           The caller must eventually close this stream using
 *           SSIClose(). More than one index file can be open
 *           at once.
 *
 * Args:     filename - full path to a SSI index file
 *
 * Returns:  Returns 0 on success, nonzero on failure.
 */
int
SSIOpen (char *filename, SSIFILE ** ret_sfp)
{
    SSIFILE *sfp = NULL;
    int status;
    if ((sfp = malloc (sizeof (SSIFILE))) == NULL) {
        return SSI_ERR_MALLOC;
    }
    if ((sfp->fp = fopen (filename, "rb")) == NULL) {
        free (sfp);
        return SSI_ERR_NOFILE;
    }
    status = load_indexfile (sfp);
    *ret_sfp = sfp;
    return status;
}

/* load_indexfile(): given a SSIFILE structure with an open and positioned
 *    stream (fp) -- but no other data loaded -- read the next SSIFILE
 *    in from disk. We use this routine without its SSIOpen() wrapper
 *    as part of the external mergesort when creating large indices.
 */
static int
load_indexfile (SSIFILE * sfp)
{
    sqd_uint32 magic;
    sqd_uint16 i;			/* counter over files */
    int status;			/* overall return status if an error is thrown */

    status = SSI_ERR_BADFORMAT;	/* default: almost every kind of error is a bad format error */

    sfp->filename = NULL;
    sfp->fileformat = NULL;
    sfp->fileflags = NULL;
    sfp->bpl = NULL;
    sfp->rpl = NULL;
    sfp->nfiles = 0;
    if (!read_i32 (sfp->fp, &magic)) {
        status = SSI_ERR_BADMAGIC;
        goto FAILURE;
    }
    if (magic != v20magicssi && magic != v20swapssi) {
        status = SSI_ERR_BADMAGIC;
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->flags))) {
        goto FAILURE;
    }

    /* If we have 64-bit offsets, make sure we can deal with them.
     */
#ifndef HAS_64BIT_FILE_OFFSETS
    if ((sfp->flags & SSI_USE64_INDEX) || (sfp->flags & SSI_USE64)) {
        status = SSI_ERR_NO64BIT;
        goto FAILURE;
    }
#endif

    sfp->imode =
        (sfp->flags & SSI_USE64_INDEX) ? SSI_OFFSET_I64 : SSI_OFFSET_I32;
    sfp->smode = (sfp->flags & SSI_USE64) ? SSI_OFFSET_I64 : SSI_OFFSET_I32;

    if (!read_i16 (sfp->fp, &(sfp->nfiles))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->nprimary))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->nsecondary))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->flen))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->plen))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->slen))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->frecsize))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->precsize))) {
        goto FAILURE;
    }
    if (!read_i32 (sfp->fp, &(sfp->srecsize))) {
        goto FAILURE;
    }

    if (!read_offset (sfp->fp, sfp->imode, &(sfp->foffset))) {
        goto FAILURE;
    }
    if (!read_offset (sfp->fp, sfp->imode, &(sfp->poffset))) {
        goto FAILURE;
    }
    if (!read_offset (sfp->fp, sfp->imode, &(sfp->soffset))) {
        goto FAILURE;
    }

    /* Read the file information and keep it.
     * We expect the number of files to be small, so reading it
     * once should be advantageous overall. If SSI ever had to
     * deal with large numbers of files, you'd probably want to
     * read file information on demand.
     */
    if (sfp->nfiles == 0) {
        goto FAILURE;
    }
    if ((sfp->filename = malloc (sizeof (char *) * sfp->nfiles)) == NULL) {
        status = SSI_ERR_MALLOC;
        goto FAILURE;
    }
    for (i = 0; i < sfp->nfiles; i++) {
        sfp->filename[i] = NULL;
    }
    if ((sfp->fileformat = malloc (sizeof (sqd_uint32) * sfp->nfiles)) == NULL) {
        status = SSI_ERR_MALLOC;
        goto FAILURE;
    }
    if ((sfp->fileflags = malloc (sizeof (sqd_uint32) * sfp->nfiles)) == NULL) {
        status = SSI_ERR_MALLOC;
        goto FAILURE;
    }
    if ((sfp->bpl = malloc (sizeof (sqd_uint32) * sfp->nfiles)) == NULL) {
        status = SSI_ERR_MALLOC;
        goto FAILURE;
    }
    if ((sfp->rpl = malloc (sizeof (sqd_uint32) * sfp->nfiles)) == NULL) {
        status = SSI_ERR_MALLOC;
        goto FAILURE;
    }

    for (i = 0; i < sfp->nfiles; i++) {
        /* We have to explicitly position, because header and file
         * records may expand in the future; frecsize and foffset
         * give us forwards compatibility.
         */
        if (indexfile_position (sfp, &(sfp->foffset), sfp->frecsize, i) != 0) {
            goto FAILURE;
        }
        if ((sfp->filename[i] = malloc (sizeof (char) * sfp->flen)) == NULL) {
            status = SSI_ERR_MALLOC;
            goto FAILURE;
        }
        if (fread (sfp->filename[i], sizeof (char), sfp->flen, sfp->fp) !=
                sfp->flen) {
            goto FAILURE;
        }
        if (!read_i32 (sfp->fp, &(sfp->fileformat[i]))) {
            goto FAILURE;
        }
        if (!read_i32 (sfp->fp, &(sfp->fileflags[i]))) {
            goto FAILURE;
        }
        if (!read_i32 (sfp->fp, &(sfp->bpl[i]))) {
            goto FAILURE;
        }
        if (!read_i32 (sfp->fp, &(sfp->rpl[i]))) {
            goto FAILURE;
        }
    }

    /* Success. Return 0.
     */
    return 0;

FAILURE:
    /* Failure: free the damaged structure, return status code.
     */
    SSIClose (sfp);
    return status;
}



/* Function: SSIGetOffsetByName()
 * Date:     SRE, Sun Dec 31 13:55:31 2000 [St. Louis]
 *
 * Purpose:  Looks up the string {key} in the open index {sfp}.
 *           {key} can be either a primary or secondary key. If {key}
 *           is found, {*ret_fh} contains a unique handle on
 *           the file that contains {key} (suitable for an SSIFileInfo()
 *           call, or for comparison to the handle of the last file
 *           that was opened for retrieval), and {offset} is filled
 *           in with the offset in that file.
 *
 * Args:     sfp         - open index file
 *           key         - string to search for
 *           ret_fh      - RETURN: handle on file that key is in
 *           ret_offset  - RETURN: offset of the start of that key's record
 *
 * Returns:  0 on success.
 *           non-zero on error.
 */
int
SSIGetOffsetByName (SSIFILE * sfp, char *key, int *ret_fh,
                    SSIOFFSET * ret_offset)
{
    int status;
    sqd_uint16 fnum;

    /* Look in the primary keys.
     */
    status = binary_search (sfp, key, sfp->plen, &(sfp->poffset), sfp->precsize,
                            sfp->nprimary);
    if (status == 0) {
        /* We found it as a primary key; get our data & return.
         */
        if (!read_i16 (sfp->fp, &fnum)) {
            return SSI_ERR_NODATA;
        }
        *ret_fh = (int) fnum;
        if (!read_offset (sfp->fp, sfp->smode, ret_offset)) {
            return SSI_ERR_NODATA;
        }

        return 0;			/* success! (we don't need the other key data) */
    }
    else if (status == SSI_ERR_NO_SUCH_KEY) {
        /* Not in the primary keys? OK, try the secondary keys.
         */
        if (sfp->nsecondary > 0) {
            char *pkey;
            status =
                binary_search (sfp, key, sfp->slen, &(sfp->soffset), sfp->srecsize,
                               sfp->nsecondary);
            if (status != 0) {
                return status;
            }
            if ((pkey = malloc (sizeof (char) * sfp->plen)) == NULL) {
                return SSI_ERR_MALLOC;
            }
            if (fread (pkey, sizeof (char), sfp->plen, sfp->fp) != sfp->plen) {
                return SSI_ERR_NODATA;
            }

            status = SSIGetOffsetByName (sfp, pkey, ret_fh, ret_offset);
            free (pkey);
        }
        return status;

    }
    else {
        return status;
    }
    /*NOTREACHED*/

}

/* Function: SSIGetOffsetByNumber()
 * Date:     SRE, Mon Jan  1 19:42:42 2001 [St. Louis]
 *
 * Purpose:  Looks up primary key #{n} in the open index {sfp}.
 *           {n} ranges from 0..nprimary-1. When key #{n}
 *           is found, {*ret_fh} contains a unique
 *           handle on the file that contains {key} (suitable
 *           for an SSIFileInfo() call, or for comparison to
 *           the handle of the last file that was opened for retrieval),
 *           and {offset} is filled in with the offset in that file.
 *
 * Args:     sfp        - open index file
 *           n          - primary key number to retrieve.
 *           ret_fh     - RETURN: handle on file that key is in
 *           ret_offset - RETURN: offset of the start of that key's record
 *
 * Returns:  0 on success.
 *           non-zero on error.
 */
int
SSIGetOffsetByNumber (SSIFILE * sfp, int n, int *ret_fh,
                      SSIOFFSET * ret_offset)
{
    sqd_uint16 fnum;
    char *pkey;

    if (n >= sfp->nprimary) {
        return SSI_ERR_NO_SUCH_KEY;
    }
    if (indexfile_position (sfp, &(sfp->poffset), sfp->precsize, n) != 0) {
        return SSI_ERR_SEEK_FAILED;
    }

    if ((pkey = malloc (sizeof (char) * sfp->plen)) == NULL) {
        return SSI_ERR_MALLOC;
    }
    if (fread (pkey, sizeof (char), sfp->plen, sfp->fp) != sfp->plen) {
        return SSI_ERR_NODATA;
    }
    if (!read_i16 (sfp->fp, &fnum)) {
        return SSI_ERR_NODATA;
    }
    if (!read_offset (sfp->fp, sfp->smode, ret_offset)) {
        return SSI_ERR_NODATA;
    }
    *ret_fh = fnum;
    free (pkey);
    return 0;
}

/* Function: SSIGetSubseqOffset()
 * Date:     SRE, Mon Jan  1 19:49:31 2001 [St. Louis]
 *
 * Purpose:  Implements SSI_FAST_SUBSEQ.
 *
 *           Looks up a primary or secondary {key} in the open
 *           index {sfp}. Asks for the nearest offset to a
 *           subsequence starting at position {requested_start}
 *           in the sequence (numbering the sequence 1..L).
 *           If {key} is found, on return, {ret_fh}
 *           contains a unique handle on the file that contains
 *           {key} (suitable for an SSIFileInfo() call, or for
 *           comparison to the handle of the last file that was
 *           opened for retrieval); {record_offset} contains the
 *           disk offset to the start of the record; {data_offset}
 *           contains the disk offset either exactly at the requested
 *           residue, or at the start of the line containing the
 *           requested residue; {ret_actual_start} contains the
 *           coordinate (1..L) of the first valid residue at or
 *           after {data_offset}. {ret_actual_start} is <=
 *           {requested_start}.
 *
 * Args:     sfp             - open index file
 *           key             - primary or secondary key to find
 *           requested_start - residue we'd like to start at (1..L)
 *           ret_fh          - RETURN: handle for file the key is in
 *           record_offset   - RETURN: offset of entire record
 *           data_offset     - RETURN: offset of subseq (see above)
 *           ret_actual_start- RETURN: coord (1..L) of residue at data_offset
 *
 * Returns:  0 on success, non-zero on failure.
 */
int
SSIGetSubseqOffset (SSIFILE * sfp, char *key, int requested_start,
                    int *ret_fh, SSIOFFSET * record_offset,
                    SSIOFFSET * data_offset, int *ret_actual_start)
{
    int status;
    sqd_uint32 len;
    int r, b, i, l;		/* tmp variables for "clarity", to match docs */

    /* Look up the key. Rely on the fact that SSIGetOffsetByName()
     * leaves the index file positioned at the rest of the data for this key.
     */
    status = SSIGetOffsetByName (sfp, key, ret_fh, record_offset);
    if (status != 0) {
        return status;
    }

    /* Check that we're allowed to do subseq lookup on that file.
     */
    if (!(sfp->fileflags[*ret_fh] & SSI_FAST_SUBSEQ)) {
        return SSI_ERR_NO_SUBSEQS;
    }

    /* Read the data we need for subseq lookup
     */
    if (!read_offset (sfp->fp, sfp->smode, data_offset)) {
        return SSI_ERR_NODATA;
    }
    if (!read_i32 (sfp->fp, &len)) {
        return SSI_ERR_NODATA;
    }

    /* Set up tmp variables for clarity of equations below,
     * and to make them match documentation (ssi-format.tex).
     */
    r = sfp->rpl[*ret_fh];	/* residues per line */
    b = sfp->bpl[*ret_fh];	/* bytes per line    */
    i = requested_start;		/* start position 1..L */
    l = (i - 1) / r;		/* data line # (0..) that the residue is on */
    if (r == 0 || b == 0) {
        return SSI_ERR_NO_SUBSEQS;
    }
    if (i < 0 || i > len) {
        return SSI_ERR_RANGE;
    }

    /* When b = r+1, there's nothing but sequence on each data line (and the \0),
     * and we can find each residue precisely.
     */
    if (b == r + 1) {
        if (sfp->smode == SSI_OFFSET_I32) {
            data_offset->mode = SSI_OFFSET_I32;
            data_offset->off.i32 = data_offset->off.i32 + l * b + (i - 1) % r;
        }
        else if (sfp->smode == SSI_OFFSET_I64) {
            data_offset->mode = SSI_OFFSET_I64;
            data_offset->off.i64 = data_offset->off.i64 + l * b + (i - 1) % r;
        }
        *ret_actual_start = requested_start;
    }
    else {
        /* else, there's other stuff on seq lines, so the best
         * we can do easily is to position at start of relevant line.
         */
        if (sfp->smode == SSI_OFFSET_I32) {
            data_offset->mode = SSI_OFFSET_I32;
            data_offset->off.i32 = data_offset->off.i32 + l * b;
        }
        else if (sfp->smode == SSI_OFFSET_I64) {
            data_offset->mode = SSI_OFFSET_I64;
            data_offset->off.i64 = data_offset->off.i64 + l * b;
        }
        /* yes, the eq below is = 1 + (i-1)/r*r but it's not = i. that's an integer /. */
        *ret_actual_start = 1 + l * r;
    }
    return 0;
}

/* Function: SSISetFilePosition()
 * Date:     SRE, Tue Jan  2 09:13:46 2001 [St. Louis]
 *
 * Purpose:  Uses {offset} to sets the file position for {fp}, usually an
 *           open sequence file, relative to the start of the file.
 *           Hides the details of system-dependent shenanigans necessary for
 *           file positioning in large (>2 GB) files.
 *
 *           Behaves just like fseek(fp, offset, SEEK_SET) for 32 bit
 *           offsets and <2 GB files.
 *
 *           Warning: if all else fails, in desperation, it will try to
 *           use fsetpos(). This requires making assumptions about fpos_t
 *           that may be unwarranted... assumptions that ANSI C prohibits
 *           me from making... though I believe the ./configure
 *           script robustly tests whether I can play with fpos_t like this.
 *
 * Args:     fp      - file to position.
 *           offset  - SSI offset relative to file start.
 *
 * Returns:  0 on success, nonzero on error.
 */
int
SSISetFilePosition (FILE * fp, SSIOFFSET * offset)
{
    if (offset->mode == SSI_OFFSET_I32) {
        if (fseek (fp, offset->off.i32, SEEK_SET) != 0) {
            return SSI_ERR_SEEK_FAILED;
        }
    }
#ifndef HAS_64BIT_FILE_OFFSETS
    else {
        return SSI_ERR_NO64BIT;
    }
#elif defined HAVE_FSEEKO && SIZEOF_OFF_T == 8
    else if (fseeko (fp, offset->off.i64, SEEK_SET) != 0) {
        return SSI_ERR_SEEK_FAILED;
    }
#elif defined HAVE_FSEEKO64 && SIZEOF_OFF64_T == 8
    else if (fseeko64 (fp, offset->off.i64, SEEK_SET) != 0) {
        return SSI_ERR_SEEK_FAILED;
    }
#elif defined HAVE_FSEEK64
    else if (fseek64 (fp, offset->off.i64, SEEK_SET) != 0) {
        return SSI_ERR_SEEK_FAILED;
    }
#elif defined ARITHMETIC_FPOS_T && SIZEOF_FPOS_T == 8
    else if (fsetpos (fp, &(offset->off.i64)) != 0) {
        return SSI_ERR_SEEK_FAILED;
    }
#endif
    return 0;
}


/* Function: SSIFileInfo()
 * Date:     SRE, Tue Jan  2 10:31:01 2001 [St. Louis]
 *
 * Purpose:  Given a file number {fh} in an open index file
 *           {sfp}, retrieve file name {ret_filename} and
 *           the file format {ret_format}.
 *
 *           {ret_filename} is a pointer to a string maintained
 *           internally by {sfp}. It should not be free'd;
 *           SSIClose(sfp) takes care of it.
 *
 * Args:     sfp          - open index file
 *           fh           - handle on file to look up
 *           ret_filename - RETURN: name of file n
 *           ret_format   - RETURN: format of file n
 *
 * Returns:  0 on success, nonzero on failure.
 */
int
SSIFileInfo (SSIFILE * sfp, int fh, char **ret_filename, int *ret_format)
{
    if (fh < 0 || fh >= sfp->nfiles) {
        return SSI_ERR_BADARG;
    }
    *ret_filename = sfp->filename[fh];
    *ret_format = sfp->fileformat[fh];
    return 0;
}

/* Function: SSIClose()
 * Date:     SRE, Sun Dec 31 14:56:37 2000 [St. Louis]
 *
 * Purpose:  Close an open {SSIFILE *}.
 *
 * Args:     sfp - index file to close.
 *
 * Returns:  (void)
 */
void
SSIClose (SSIFILE * sfp)
{
    if (sfp != NULL) {
        clear_ssifile (sfp);
        if (sfp->fp != NULL) {
            fclose (sfp->fp);
        }
        free (sfp);
    }
}

/* clear_ssifile(): free the innards of SSIFILE, without
 * destroying the structure or closing the stream.
 */
static void
clear_ssifile (SSIFILE * sfp)
{
    int i;

    if (sfp->filename != NULL) {
        for (i = 0; i < sfp->nfiles; i++) {
            if (sfp->filename[i] != NULL) {
                free (sfp->filename[i]);
            }
				}		
        free (sfp->filename);
    }
    if (sfp->fileformat != NULL) {
        free (sfp->fileformat);
    }
    if (sfp->fileflags != NULL) {
        free (sfp->fileflags);
    }
    if (sfp->bpl != NULL) {
        free (sfp->bpl);
    }
    if (sfp->rpl != NULL) {
        free (sfp->rpl);
    }
}


/* Function: SSIRecommendMode()
 * Date:     SRE, Fri Feb 16 08:23:47 2001 [St. Louis]
 *
 * Purpose:  Examines the file and determines whether it should be
 *           indexed with large file support or not; returns
 *           SSI_OFFSET_I32 for most files, SSI_OFFSET_I64 for large
 *           files, or -1 on failure.
 *
 * Args:     file - name of file to check for size
 *
 * Returns:  -1 on failure (including case where file is too big)
 *           SSI_OFFSET_I32 for most files (<= 2^31-1 bytes)
 *           SSI_OFFSET_I64 for large files (> 2^31-1 bytes)
 */
int
SSIRecommendMode (char *file)
{
#ifdef SPEC_CPU
    return SSI_OFFSET_I32;
#else
#if HAVE_STAT64
    struct stat64 s1;
    if (stat64 (file, &s1) == 0) {
        if (s1.st_size <= 2146483647L) {
            return SSI_OFFSET_I32;
        }
        else {
            return SSI_OFFSET_I64;
        }
    }
#else
    struct stat s2;
    if (stat (file, &s2) == 0) {
        if (s2.st_size <= 2146483647L) {
            return SSI_OFFSET_I32;
        }
        else {
            return SSI_OFFSET_I64;
        }
    }
#endif
    return -1;
#endif
}


/* Function: SSICreateIndex()
 * Date:     SRE, Tue Jan  2 11:23:25 2001 [St. Louis]
 *
 * Purpose:  Creates and initializes a SSI index structure.
 *           Sequence file offset type is specified by {mode}.
 *
 * Args:     mode    - SSI_OFFSET_I32 or SSI_OFFSET_I64, sequence file index mode.
 *
 * Returns:  ptr to new index structure, or NULL on failure.
 *           Caller is responsible for free'ing the returned
 *           structure with SSIFreeIndex().
 */
SSIINDEX *
SSICreateIndex (int mode)
{
    SSIINDEX *g;

    g = NULL;
    if ((g = malloc (sizeof (SSIINDEX))) == NULL) {
        goto FAILURE;
    }
    g->smode = mode;
    g->imode = SSI_OFFSET_I32;	/* index always starts as 32-bit; may get upgraded later */
    g->external = FALSE;
    g->max_ram = SSI_MAXRAM;

#ifndef HAS_64BIT_FILE_OFFSETS
    if (mode == SSI_OFFSET_I64)
        Die ("\
Can't create a 64-bit SSI index on this system, sorry;\n\
I don't have 64-bit file offset functions available.\n");
#endif

    g->filenames = NULL;
    g->fileformat = NULL;
    g->bpl = NULL;
    g->rpl = NULL;
    g->flen = 0;
    g->nfiles = 0;

    g->pkeys = NULL;
    g->plen = 0;
    g->nprimary = 0;
    g->ptmpfile = "tmp.ssi.1";	/* hardcoded, for now. */
    g->ptmp = NULL;

    g->skeys = NULL;
    g->slen = 0;
    g->nsecondary = 0;
    g->stmpfile = "tmp.ssi.2";	/* hardcoded, for now. */
    g->stmp = NULL;

    /* All mallocs must go after NULL initializations, because of the cleanup strategy;
     * we'll try to free anything non-NULL if a malloc fails.
     */
    if ((g->filenames = malloc (sizeof (char *) * SSI_FILE_BLOCK)) == NULL) {
        goto FAILURE;
    }
    if ((g->fileformat = malloc (sizeof (sqd_uint32) * SSI_FILE_BLOCK)) == NULL) {
        goto FAILURE;
    }
    if ((g->bpl = malloc (sizeof (sqd_uint32) * SSI_FILE_BLOCK)) == NULL) {
        goto FAILURE;
    }
    if ((g->rpl = malloc (sizeof (sqd_uint32) * SSI_FILE_BLOCK)) == NULL) {
        goto FAILURE;
    }

    if ((g->pkeys = malloc (sizeof (struct ssipkey_s) * SSI_KEY_BLOCK)) == NULL) {
        goto FAILURE;
    }
    if ((g->skeys = malloc (sizeof (struct ssipkey_s) * SSI_KEY_BLOCK)) == NULL) {
        goto FAILURE;
    }

    return g;

FAILURE:
    SSIFreeIndex (g);		/* free the damaged structure */
    return NULL;
}

/* Function: SSIGetFilePosition()
 * Date:     SRE, Tue Jan  2 09:59:26 2001 [St. Louis]
 *
 * Purpose:  Fills {ret_offset} with the current disk
 *           offset of {fp}, relative to the start of the file.
 *           {mode} is set to either SSI_OFFSET_I32 or
 *           SSI_OFFSET_I64. If {mode} is _I32 (32 bit), just wraps
 *           a call to ftell(); otherwise deals with system-dependent
 *           details of 64-bit file offsets.
 *
 * Args:     fp         - open stream
 *           mode       - SSI_OFFSET_I32 or SSI_OFFSET_I64
 *           ret_offset - RETURN: file position
 *
 * Returns:  0 on success. nonzero on error.
 */
int
SSIGetFilePosition (FILE * fp, int mode, SSIOFFSET * ret_offset)
{
    if (mode == SSI_OFFSET_I32) {
        ret_offset->mode = SSI_OFFSET_I32;
        ret_offset->off.i32 = ftell (fp);
        if (ret_offset->off.i32 == -1) {
            return SSI_ERR_TELL_FAILED;
        }
    }
    else if (mode != SSI_OFFSET_I64) {
        abort ();    /* only happens on a coding error */
    }
    else {
        ret_offset->mode = SSI_OFFSET_I64;
#ifndef HAS_64BIT_FILE_OFFSETS
        return SSI_ERR_NO64BIT;
#elif defined HAVE_FTELLO && SIZEOF_OFF_T == 8
        if ((ret_offset->off.i64 = ftello (fp)) == -1) {
            return SSI_ERR_TELL_FAILED;
        }
#elif defined HAVE_FTELLO64 && SIZEOF_OFF64_T == 8
        if ((ret_offset->off.i64 = ftello64 (fp)) == -1) {
            return SSI_ERR_TELL_FAILED;
        }
#elif defined HAVE_FTELL64
        if ((ret_offset->off.i64 = ftell64 (fp)) == -1) {
            return SSI_ERR_TELL_FAILED;
        }
#elif defined ARITHMETIC_FPOS_T && SIZEOF_FPOS_T == 8
        if (fgetpos (fp, &(ret_offset->off.i64)) != 0) {
            return SSI_ERR_TELL_FAILED;
        }
#endif
    }
    return 0;
}

/* Function: SSIAddFileToIndex()
 * Date:     SRE, Tue Jan  2 12:54:36 2001 [St. Louis]
 *
 * Purpose:  Adds the sequence file {filename}, which is known to
 *           be in format {fmt}, to the index {g}. Creates and returns
 *           a unique filehandle {fh} for then associating primary keys
 *           with this file using SSIAddPrimaryKeyToIndex().
 *
 * Args:     g         - active index
 *           filename  - file to add
 *           fmt       - format code for this file (e.g. SQFILE_FASTA)
 *           ret_fh    - RETURN: unique handle for this file
 *
 * Returns:  0 on success; nonzero on error.
 */
int
SSIAddFileToIndex (SSIINDEX * g, char *filename, int fmt, int *ret_fh)
{
    int n;

    if (g->nfiles >= SSI_MAXFILES) {
        return SSI_ERR_TOOMANY_FILES;
    }

    n = strlen (filename);
    if ((n + 1) > g->flen) {
        g->flen = n + 1;
    }

    g->filenames[g->nfiles] = FileTail (filename, FALSE);
    g->fileformat[g->nfiles] = fmt;
    g->bpl[g->nfiles] = 0;
    g->rpl[g->nfiles] = 0;
    *ret_fh = g->nfiles;		/* handle is simply = file number */
    g->nfiles++;

    if (g->nfiles % SSI_FILE_BLOCK == 0) {
        g->filenames =
            realloc (g->filenames, sizeof (char *) * (g->nfiles + SSI_FILE_BLOCK));
        if (g->filenames == NULL) {
            return SSI_ERR_MALLOC;
        }
        g->fileformat =
            realloc (g->fileformat,
                     sizeof (sqd_uint32) * (g->nfiles + SSI_FILE_BLOCK));
        if (g->fileformat == NULL) {
            return SSI_ERR_MALLOC;
        }
        g->bpl =
            realloc (g->bpl, sizeof (sqd_uint32) * (g->nfiles + SSI_FILE_BLOCK));
        if (g->bpl == NULL) {
            return SSI_ERR_MALLOC;
        }
        g->rpl =
            realloc (g->rpl, sizeof (sqd_uint32) * (g->nfiles + SSI_FILE_BLOCK));
        if (g->rpl == NULL) {
            return SSI_ERR_MALLOC;
        }
    }
    return 0;
}


/* Function: SSISetFileForSubseq()
 * Date:     SRE, Tue Jan  9 10:02:05 2001 [St. Louis]
 *
 * Purpose:  Set SSI_FAST_SUBSEQ for the file indicated by
 *           filehandle {fh} in the index {g}, setting
 *           parameters {bpl} and {rpl} to the values given.
 *           {bpl} is the number of bytes per sequence data line.
 *           {rpl} is the number of residues per sequence data line.
 *           Caller must be sure that {bpl} and {rpl} do not change
 *           on any line of any sequence record in the file
 *           (except for the last data line of each record). If
 *           this is not the case in this file, SSI_FAST_SUBSEQ
 *           will not work, and this routine should not be
 *           called.
 *
 * Args:     g    - the active index
 *           fh   - handle for file to set SSI_FAST_SUBSEQ on
 *           bpl  - bytes per data line
 *           rpl  - residues per data line
 *
 * Returns:  0 on success; 1 on error.
 */
int
SSISetFileForSubseq (SSIINDEX * g, int fh, int bpl, int rpl)
{
    if (fh < 0 || fh >= g->nfiles) {
        return SSI_ERR_BADARG;
    }
    if (bpl <= 0 || rpl <= 0) {
        return SSI_ERR_BADARG;
    }
    g->bpl[fh] = bpl;
    g->rpl[fh] = rpl;
    return 0;
}


/* Function: SSIAddPrimaryKeyToIndex()
 * Date:     SRE, Tue Jan  2 11:50:54 2001 [St. Louis]
 *
 * Purpose:  Put primary key {key} in the index {g}, while telling
 *           the index this primary key is in the file associated
 *           with filehandle {fh} (returned by a previous call
 *           to SSIAddFileToIndex()), and its record starts at
 *           position {r_off} in the file.
 *
 *           {d_off} and {L} are optional; they may be left unset
 *           by passing NULL and 0, respectively. (If one is
 *           provided, both must be provided.) If they are provided,
 *           {d_off} gives the position of the first line of sequence
 *           data in the record, and {L} gives the length of
 *           the sequence in residues. They are used when
 *           SSI_FAST_SUBSEQ is set for this file. If SSI_FAST_SUBSEQ
 *           is not set for the file, {d_off} and {L} will be
 *           ignored by the index reading API even if they are stored
 *           by the index writing API, so it doesn't hurt for the
 *           indexing program to provide them; typically they
 *           won't know whether it's safe to set SSI_FAST_SUBSEQ
 *           for the whole file until the whole file has been
 *           read and every key has already been added to the index.
 *
 * Args:     g      - active index
 *           key    - primary key to add
 *           fh     - handle on file that this key's in
 *           r_off  - offset to start of record
 *           d_off  - offset to start of sequence data
 *           L      - length of sequence, or 0
 *
 * Returns:  0 on success, nonzero on error.
 */
int
SSIAddPrimaryKeyToIndex (SSIINDEX * g, char *key, int fh,
                         SSIOFFSET * r_off, SSIOFFSET * d_off, int L)
{
    int n;			/* a string length */

    if (fh >= SSI_MAXFILES) {
        return SSI_ERR_TOOMANY_FILES;
    }
    if (g->nprimary >= SSI_MAXKEYS) {
        return SSI_ERR_TOOMANY_KEYS;
    }
    if (L > 0 && d_off == NULL) {
        abort ();    /* need both. */
    }

    /* Before adding the key: check how big our index is.
     * If it's getting too large, switch to external mode.
     */
    if (!g->external && current_index_size (g) >= g->max_ram)
        if (activate_external_sort (g) != 0) {
            return SSI_ERR_NOFILE;
        }

    /* Update maximum pkey length, if needed.
     */
    n = strlen (key);
    if ((n + 1) > g->plen) {
        g->plen = n + 1;
    }

    /* External mode? Simply append to disk...
     */
    if (g->external) {
        if (g->smode == SSI_OFFSET_I32) {
            fprintf (g->ptmp, "%s\t%d\t%lu\t%lu\t%lu\n",
                     key, fh, (unsigned long) r_off->off.i32,
                     (unsigned long) (d_off == NULL ? 0 : d_off->off.i32),
                     (unsigned long) L);
        }
        else {
            fprintf (g->ptmp, "%s\t%d\t%llu\t%llu\t%lu\n",
                     key, fh, (unsigned long long) r_off->off.i64,
                     (unsigned long long) (d_off == NULL ? 0 : d_off->off.i64),
                     (unsigned long) L);
        }
        g->nprimary++;
        return 0;
    }

    /* Else: internal mode, keep keys in memory...
     */
    if ((g->pkeys[g->nprimary].key = sre_strdup (key, n)) == NULL) {
        return SSI_ERR_MALLOC;
    }
    g->pkeys[g->nprimary].fnum = (sqd_uint16) fh;
    g->pkeys[g->nprimary].r_off = *r_off;
    if (d_off != NULL && L > 0) {
        g->pkeys[g->nprimary].d_off = *d_off;
        g->pkeys[g->nprimary].len = L;
    }
    else {
        /* yeah, this looks stupid, but look: we have to give a valid
           looking, non-NULL d_off of some sort, or writes will fail.
           It's going to be unused anyway. */
        g->pkeys[g->nprimary].d_off = *r_off;
        g->pkeys[g->nprimary].len = 0;
    }
    g->nprimary++;

    if (g->nprimary % SSI_KEY_BLOCK == 0) {
        g->pkeys =
            realloc (g->pkeys,
                     sizeof (struct ssipkey_s) * (g->nprimary + SSI_KEY_BLOCK));
        if (g->pkeys == NULL) {
            return SSI_ERR_MALLOC;
        }
    }
    return 0;
}


/* Function: SSIAddSecondaryKeyToIndex()
 * Date:     SRE, Tue Jan  2 12:44:40 2001 [St. Louis]
 *
 * Purpose:  Puts secondary key {key} in the index {g}, associating
 *           it with primary key {pkey} that was previously
 *           registered by SSIAddPrimaryKeyToIndex().
 *
 * Args:     g    - active index
 *           key  - secondary key to add
 *           pkey - primary key to associate this key with
 *
 * Returns:  0 on success, nonzero on failure.
 */
int
SSIAddSecondaryKeyToIndex (SSIINDEX * g, char *key, char *pkey)
{
    int n;			/* a string length */

    if (g->nsecondary >= SSI_MAXKEYS) {
        return SSI_ERR_TOOMANY_KEYS;
    }

    /* Before adding the key: check how big our index is.
     * If it's getting too large, switch to external mode.
     */
    if (!g->external && current_index_size (g) >= g->max_ram)
        if (activate_external_sort (g) != 0) {
            return SSI_ERR_NOFILE;
        }

    /* Update maximum secondary key length, if necessary.
     */
    n = strlen (key);
    if ((n + 1) > g->slen) {
        g->slen = n + 1;
    }

    /* if external mode: write info to disk.
     */
    if (g->external) {
        fprintf (g->stmp, "%s\t%s\n", key, pkey);
        g->nsecondary++;
        return 0;
    }

    /* else, internal mode... store info in memory.
     */
    if ((g->skeys[g->nsecondary].key = sre_strdup (key, n)) == NULL) {
        return SSI_ERR_MALLOC;
    }
    if ((g->skeys[g->nsecondary].pkey = sre_strdup (pkey, -1)) == NULL) {
        return SSI_ERR_MALLOC;
    }
    g->nsecondary++;

    if (g->nsecondary % SSI_KEY_BLOCK == 0) {
        g->skeys =
            realloc (g->skeys,
                     sizeof (struct ssiskey_s) * (g->nsecondary + SSI_KEY_BLOCK));
        if (g->skeys == NULL) {
            return SSI_ERR_MALLOC;
        }
    }
    return 0;
}




/* Function: SSIWriteIndex()
 * Date:     SRE, Tue Jan  2 13:55:56 2001 [St. Louis]
 *
 * Purpose:  Writes complete index {g} in SSI format to a
 *           binary file {file}. Does all
 *           the overhead of sorting the primary and secondary keys,
 *           and maintaining the association of secondary keys
 *           with primary keys during and after the sort.
 *
 * Args:     file  - file to write to
 *           g     - index to sort & write out.
 *
 * Returns:  0 on success, nonzero on error.
 */
/* needed for qsort() */
static int
pkeysort (const void *k1, const void *k2)
{
    struct ssipkey_s *key1;
    struct ssipkey_s *key2;
    key1 = (struct ssipkey_s *) k1;
    key2 = (struct ssipkey_s *) k2;
    return strcmp (key1->key, key2->key);
}

static int
skeysort (const void *k1, const void *k2)
{
    struct ssiskey_s *key1;
    struct ssiskey_s *key2;
    key1 = (struct ssiskey_s *) k1;
    key2 = (struct ssiskey_s *) k2;
    return strcmp (key1->key, key2->key);
}

int
SSIWriteIndex (char *file, SSIINDEX * g)
{
    FILE *fp;
    int status;
    int i;
    sqd_uint32 header_flags, file_flags;
    sqd_uint32 frecsize, precsize, srecsize;
    sqd_uint64 foffset, poffset, soffset;
    char *s, *s2;

    if ((fp = fopen (file, "wb")) == NULL) {
        return SSI_ERR_NOFILE;
    }
    status = 0;

    /* How big is the index? If it's going to be > 2GB, we need
     * to flip to 64-bit index mode. 2047 (instead of 2048) gives us
     * some slop room.
     * die'ing here is pretty brutal - if we flip to 64-bit index
     * mode, we hve 100's of millions of keys, so we've processed
     * a long time before reaching this point. Ah well.
     */
    if (current_index_size (g) >= 2047) {
        g->imode = SSI_OFFSET_I64;
#ifndef HAS_64BIT_FILE_OFFSETS
        Die ("\
Can't switch to 64-bit SSI index mode on this system, sorry;\n\
I don't have 64-bit file offset functions available.\n");
#endif
    }

    /* Magic-looking numbers come from adding up sizes
     * of things in bytes
     */
    frecsize = 16 + g->flen;
    precsize = (g->smode == SSI_OFFSET_I64) ? 22 + g->plen : 14 + g->plen;
    srecsize = g->slen + g->plen;

    header_flags = 0;
    if (g->smode == SSI_OFFSET_I64) {
        header_flags |= SSI_USE64;
    }
    if (g->imode == SSI_OFFSET_I64) {
        header_flags |= SSI_USE64_INDEX;
    }

    /* Magic-looking numbers again come from adding up sizes
     * of things in bytes
     */
    foffset = (header_flags & SSI_USE64_INDEX) ? 66 : 54;
    poffset = foffset + frecsize * g->nfiles;
    soffset = poffset + precsize * g->nprimary;

    /* Sort the keys
     * If external mode, make system calls to UNIX/POSIX "sort" in place, then
     * open new sorted files for reading thru ptmp and stmp handles.
     * If internal mode, call qsort.
     *
     * Note that you'd better force a POSIX locale for the sort; else,
     * some silly distro (e.g. Mandrake Linux >=8.1) may have specified
     * LC_COLLATE=en_US, and this'll give a sort "bug" in which it doesn't
     * sort by byte order.
     */
    if (g->external) {
        char cmd[1024];

        fclose (g->ptmp);
        g->ptmp = NULL;
        sprintf (cmd, "env LC_ALL=POSIX sort -o %s %s\n", g->ptmpfile,
                 g->ptmpfile);
        if ((status = system (cmd)) != 0) {
            return SSI_ERR_EXTERNAL_SORT;
        }
        if ((g->ptmp = fopen (g->ptmpfile, "r")) == NULL) {
            return SSI_ERR_EXTERNAL_SORT;
        }

        fclose (g->stmp);
        g->stmp = NULL;
        sprintf (cmd, "env LC_ALL=POSIX sort -o %s %s\n", g->stmpfile,
                 g->stmpfile);
        if ((status = system (cmd)) != 0) {
            return SSI_ERR_EXTERNAL_SORT;
        }
        if ((g->stmp = fopen (g->stmpfile, "r")) == NULL) {
            return SSI_ERR_EXTERNAL_SORT;
        }
    }
    else {
        specqsort ((void *) g->pkeys, g->nprimary, sizeof (struct ssipkey_s),
                   pkeysort);
        specqsort ((void *) g->skeys, g->nsecondary, sizeof (struct ssiskey_s),
                   skeysort);
    }

    /* Write the header
     */
    if (!write_i32 (fp, v20magicssi)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, header_flags)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i16 (fp, g->nfiles)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, g->nprimary)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, g->nsecondary)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, g->flen)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, g->plen)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, g->slen)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, frecsize)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, precsize)) {
        return SSI_ERR_FWRITE;
    }
    if (!write_i32 (fp, srecsize)) {
        return SSI_ERR_FWRITE;
    }
    if (g->imode == SSI_OFFSET_I32) {
        if (!write_i32 (fp, foffset)) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i32 (fp, poffset)) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i32 (fp, soffset)) {
            return SSI_ERR_FWRITE;
        }
    }
    else {
        if (!write_i64 (fp, foffset)) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i64 (fp, poffset)) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i64 (fp, soffset)) {
            return SSI_ERR_FWRITE;
        }
    }

    /* The file section
     */
    if ((s = malloc (sizeof (char) * g->flen)) == NULL) {
        return SSI_ERR_MALLOC;
    }
    for (i = 0; i < g->nfiles; i++) {
        file_flags = 0;
        if (g->bpl[i] > 0 && g->rpl[i] > 0) {
            file_flags |= SSI_FAST_SUBSEQ;
        }

        strcpy (s, g->filenames[i]);
        if (fwrite (s, sizeof (char), g->flen, fp) != g->flen) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i32 (fp, g->fileformat[i])) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i32 (fp, file_flags)) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i32 (fp, g->bpl[i])) {
            return SSI_ERR_FWRITE;
        }
        if (!write_i32 (fp, g->rpl[i])) {
            return SSI_ERR_FWRITE;
        }
    }
    free (s);

    /* The primary key section
     */
    if ((s = malloc (sizeof (char) * g->plen)) == NULL) {
        return SSI_ERR_MALLOC;
    }
    if (g->external) {
        char *buf = NULL;
        int buflen = 0;
        struct ssipkey_s pkey;
        for (i = 0; i < g->nprimary; i++) {
            if (sre_fgets (&buf, &buflen, g->ptmp) == NULL) {
                return SSI_ERR_NODATA;
            }
            if (parse_pkey_info (buf, g->smode, &pkey) != 0) {
                return SSI_ERR_BADFORMAT;
            }
            strcpy (s, pkey.key);
            if (fwrite (s, sizeof (char), g->plen, fp) != g->plen) {
                return SSI_ERR_FWRITE;
            }
            if (!write_i16 (fp, pkey.fnum)) {
                return SSI_ERR_FWRITE;
            }
            if (!write_offset (fp, &(pkey.r_off))) {
                return SSI_ERR_FWRITE;
            }
            if (!write_offset (fp, &(pkey.d_off))) {
                return SSI_ERR_FWRITE;
            }
            if (!write_i32 (fp, pkey.len)) {
                return SSI_ERR_FWRITE;
            }
        }
        free (buf);
    }
    else {
        for (i = 0; i < g->nprimary; i++) {
            strcpy (s, g->pkeys[i].key);
            if (fwrite (s, sizeof (char), g->plen, fp) != g->plen) {
                return SSI_ERR_FWRITE;
            }
            if (!write_i16 (fp, g->pkeys[i].fnum)) {
                return SSI_ERR_FWRITE;
            }
            if (!write_offset (fp, &(g->pkeys[i].r_off))) {
                return SSI_ERR_FWRITE;
            }
            if (!write_offset (fp, &(g->pkeys[i].d_off))) {
                return SSI_ERR_FWRITE;
            }
            if (!write_i32 (fp, g->pkeys[i].len)) {
                return SSI_ERR_FWRITE;
            }
        }
    }

    /* The secondary key section
     */
    if (g->nsecondary > 0) {
        if ((s2 = malloc (sizeof (char) * g->slen)) == NULL) {
            return SSI_ERR_MALLOC;
        }

        if (g->external) {
            struct ssiskey_s skey;
            char *buf = NULL;
            int n = 0;

            for (i = 0; i < g->nsecondary; i++) {
                if (sre_fgets (&buf, &n, g->stmp) == NULL) {
                    return SSI_ERR_NODATA;
                }
                if (parse_skey_info (buf, &skey) != 0) {
                    return SSI_ERR_BADFORMAT;
                }
                strcpy (s2, skey.key);
                strcpy (s, skey.pkey);
                if (fwrite (s2, sizeof (char), g->slen, fp) != g->slen) {
                    return SSI_ERR_FWRITE;
                }
                if (fwrite (s, sizeof (char), g->plen, fp) != g->plen) {
                    return SSI_ERR_FWRITE;
                }
            }
            free (buf);
        }
        else {
            for (i = 0; i < g->nsecondary; i++) {
                strcpy (s2, g->skeys[i].key);
                strcpy (s, g->skeys[i].pkey);
                if (fwrite (s2, sizeof (char), g->slen, fp) != g->slen) {
                    return SSI_ERR_FWRITE;
                }
                if (fwrite (s, sizeof (char), g->plen, fp) != g->plen) {
                    return SSI_ERR_FWRITE;
                }
            }
        }
        free (s2);
    }

    free (s);
    fclose (fp);
    return status;
}


/* Function: SSIFreeIndex()
 * Date:     SRE, Tue Jan  2 11:44:08 2001 [St. Louis]
 *
 * Purpose:  Free an index structure {g}.
 *
 * Args:     g  - ptr to an open index.
 *
 * Returns:  (void)
 */
void
SSIFreeIndex (SSIINDEX * g)
{
    int i;
    if (g != NULL) {
        if (g->external == FALSE) {
            for (i = 0; i < g->nprimary; i++) {
                free (g->pkeys[i].key);
            }
            for (i = 0; i < g->nsecondary; i++) {
                free (g->skeys[i].key);
            }
            for (i = 0; i < g->nsecondary; i++) {
                free (g->skeys[i].pkey);
            }
            if (g->pkeys != NULL) {
                free (g->pkeys);
            }
            if (g->skeys != NULL) {
                free (g->skeys);
            }
        }
        else {
            if (g->ptmp != NULL) {
                fclose (g->ptmp);
            }
            if (g->stmp != NULL) {
                fclose (g->stmp);
            }
#if DEBUGLEVEL == 0
            remove (g->ptmpfile);
            remove (g->stmpfile);
#endif
        }
        for (i = 0; i < g->nfiles; i++) {
            free (g->filenames[i]);
        }
        if (g->filenames != NULL) {
            free (g->filenames);
        }
        if (g->fileformat != NULL) {
            free (g->fileformat);
        }
        if (g->bpl != NULL) {
            free (g->bpl);
        }
        if (g->rpl != NULL) {
            free (g->rpl);
        }
        free (g);
    }
}


/* Function: SSIErrorString()
 * Date:     SRE, Tue Jan  2 10:38:10 2001 [St. Louis]
 *
 * Purpose:  Returns a ptr to an internal string corresponding
 *           to error {n}, a code returned from any of the
 *           functions in the API that return non-zero on error.
 *
 * Args:     n - error code
 *
 * Returns:  ptr to an internal string.
 */
char *
SSIErrorString (int n)
{
    switch (n) {
    case SSI_ERR_OK:
        return "ok (no error)";
    case SSI_ERR_NODATA:
        return "no data, fread() failed";
    case SSI_ERR_NO_SUCH_KEY:
        return "no such key";
    case SSI_ERR_MALLOC:
        return "out of memory, malloc() failed";
    case SSI_ERR_NOFILE:
        return "file not found, fopen() failed";
    case SSI_ERR_BADMAGIC:
        return "not a SSI file? (bad magic)";
    case SSI_ERR_BADFORMAT:
        return "corrupt format? unexpected data";
    case SSI_ERR_NO64BIT:
        return "no large file support for this system";
    case SSI_ERR_SEEK_FAILED:
        return "failed to reposition on disk";
    case SSI_ERR_TELL_FAILED:
        return "failed to get file position on disk";
    case SSI_ERR_NO_SUBSEQS:
        return "no fast subseq support for this seqfile";
    case SSI_ERR_RANGE:
        return "subseq start is out of range";
    case SSI_ERR_BADARG:
        return "an argument is out of range";
    case SSI_ERR_TOOMANY_FILES:
        return "number of files exceeds limit";
    case SSI_ERR_TOOMANY_KEYS:
        return "number of keys exceeds limit";
    case SSI_ERR_FWRITE:
        return "an fwrite() failed";
    case SSI_ERR_EXTERNAL_SORT:
        return "some problem with external sorting";
    default:
        return "unrecognized code";
    }
    /*NOTREACHED*/

}

static int
read_i16 (FILE * fp, sqd_uint16 * ret_result)
{
    sqd_uint16 result;
    if (fread (&result, sizeof (sqd_uint16), 1, fp) != 1) {
        return 0;
    }
    *ret_result = sre_ntoh16 (result);
    return 1;
}

static int
write_i16 (FILE * fp, sqd_uint16 n)
{
    n = sre_hton16 (n);
    if (fwrite (&n, sizeof (sqd_uint16), 1, fp) != 1) {
        return 0;
    }
    return 1;
}

static int
read_i32 (FILE * fp, sqd_uint32 * ret_result)
{
    sqd_uint32 result;
    if (fread (&result, sizeof (sqd_uint32), 1, fp) != 1) {
        return 0;
    }
    *ret_result = sre_ntoh32 (result);
    return 1;
}

static int
write_i32 (FILE * fp, sqd_uint32 n)
{
    n = sre_hton32 (n);
    if (fwrite (&n, sizeof (sqd_uint32), 1, fp) != 1) {
        return 0;
    }
    return 1;
}

static int
read_i64 (FILE * fp, sqd_uint64 * ret_result)
{
    sqd_uint64 result;
    if (fread (&result, sizeof (sqd_uint64), 1, fp) != 1) {
        return 0;
    }
    *ret_result = sre_ntoh64 (result);
    return 1;
}

static int
write_i64 (FILE * fp, sqd_uint64 n)
{
    n = sre_hton64 (n);
    if (fwrite (&n, sizeof (sqd_uint64), 1, fp) != 1) {
        return 0;
    }
    return 1;
}

static int
read_offset (FILE * fp, char mode, SSIOFFSET * ret_offset)
{
    if (mode == SSI_OFFSET_I32) {
        ret_offset->mode = SSI_OFFSET_I32;
        if (!read_i32 (fp, &(ret_offset->off.i32))) {
            return 0;
        }
    }
    else if (mode == SSI_OFFSET_I64) {
        ret_offset->mode = SSI_OFFSET_I64;
        if (!read_i64 (fp, &(ret_offset->off.i64))) {
            return 0;
        }
    }
    else {
        return 0;
    }

    return 1;
}

static int
write_offset (FILE * fp, SSIOFFSET * offset)
{
    if (offset->mode == SSI_OFFSET_I32) {
        return write_i32 (fp, offset->off.i32);
    }
    else if (offset->mode == SSI_OFFSET_I64) {
        return write_i64 (fp, offset->off.i64);
    }
    else {
        abort ();
    }
    /*UNREACHED*/ return 1;	/* silence bitchy compilers */
}

static int
parse_pkey_info (char *buf, char mode, struct ssipkey_s *pkey)
{
    char *s, *tok;
    int n;

    s = buf;
    if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
        return SSI_ERR_BADFORMAT;
    }
    pkey->key = tok;
    if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
        return SSI_ERR_BADFORMAT;
    }
    pkey->fnum = (sqd_uint16) atoi (tok);

    if (mode == SSI_OFFSET_I32) {
        if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
            return SSI_ERR_BADFORMAT;
        }
        pkey->r_off.mode = mode;
        pkey->r_off.off.i32 = (sqd_uint32) strtoul (tok, NULL, 10);
        if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
            return SSI_ERR_BADFORMAT;
        }
        pkey->d_off.mode = mode;
        pkey->d_off.off.i32 = (sqd_uint32) strtoul (tok, NULL, 10);
    }
#ifdef HAS_64BIT_FILE_OFFSETS
    else {
        if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
            return SSI_ERR_BADFORMAT;
        }
        pkey->r_off.mode = mode;
        pkey->r_off.off.i64 = (sqd_uint64) strtoull (tok, NULL, 10);
        if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
            return SSI_ERR_BADFORMAT;
        }
        pkey->d_off.mode = mode;
        pkey->d_off.off.i64 = (sqd_uint64) strtoull (tok, NULL, 10);
    }
#else
    else {
        return SSI_ERR_NO64BIT;
    }
#endif
    if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
        return SSI_ERR_BADFORMAT;
    }
    pkey->len = (sqd_uint32) strtoul (tok, NULL, 10);

    return 0;
}

static int
parse_skey_info (char *buf, struct ssiskey_s *skey)
{
    char *s, *tok;
    int n;

    s = buf;
    if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
        return SSI_ERR_BADFORMAT;
    }
    skey->key = tok;
    if ((tok = sre_strtok (&s, "\t\n", &n)) == NULL) {
        return SSI_ERR_BADFORMAT;
    }
    skey->pkey = tok;
    return 0;
}

/* Function: binary_search()
 * Date:     SRE, Sun Dec 31 16:05:03 2000 [St. Louis]
 *
 * Purpose:  Find a key in a SSI index, by a binary search
 *           in an alphabetically sorted list of keys. If successful,
 *           return 0, and the index file is positioned to read
 *           the rest of the data for that key. Else returns nonzero.
 *
 * Args:     sfp    - an open SSIFILE
 *           key    - key to find
 *           klen   - key length to allocate (plen or slen from sfp)
 *           base   - base offset (poffset or soffset)
 *           recsize - size of each key record in bytes (precsize or srecsize)
 *           maxidx  - # of keys (nprimary or nsecondary)
 *
 * Returns:  0 on success, and leaves file positioned for reading remaining
 *           data for the key.
 *           Nonzero on failure:
 *                SSI_ERR_NO_SUCH_KEY  - that key's not in the index
 *                SSI_ERR_MALLOC       - a memory allocation failure
 *                SSI_ERR_NODATA       - an fread() failed
 */
static int
binary_search (SSIFILE * sfp, char *key, int klen, SSIOFFSET * base,
               sqd_uint32 recsize, sqd_uint32 maxidx)
{
    char *name;
    sqd_uint32 left, right, mid;
    int cmp;
    int status;

    if (maxidx == 0) {
        return SSI_ERR_NO_SUCH_KEY;    /* special case: empty index */
    }
    if ((name = malloc (sizeof (char) * klen)) == NULL) {
        return SSI_ERR_MALLOC;
    }
    left = 0;
    right = maxidx - 1;
    while (1) {			/* A binary search: */
        mid = (left + right) / 2;	/* careful here. only works because
				   we limit unsigned vars to signed ranges. */
        if ((status = indexfile_position (sfp, base, recsize, mid)) != 0) {
            free (name);
            return status;
        }
        if (fread (name, sizeof (char), klen, sfp->fp) != klen) {
            free (name);
            return SSI_ERR_NODATA;
        }
        cmp = strcmp (name, key);
        if (cmp == 0) {
            break;    /* found it!              */
        }
        else if (left >= right) {	/* oops, missed it; fail  */
            free (name);
            return SSI_ERR_NO_SUCH_KEY;
        }
        else if (cmp < 0) {
            left = mid + 1;    /* it's right of mid     */
        }
        else if (cmp > 0) {
            if (mid == 0) {
                free (name);
                return SSI_ERR_NO_SUCH_KEY;
            }				/* special case, beware */
            else {
                right = mid - 1;    /* it's left of mid      */
            }
        }
    }
    free (name);
    return 0;			/* and sfp->fp is positioned... */
}

/* Function: indexfile_position()
 * Date:     SRE, Mon Jan  1 19:32:49 2001 [St. Louis]
 *
 * Purpose:  Position the open index file {sfp} at the start
 *           of record {n} in a list of records that starts at
 *           base offset {base}, where each record takes up {l}
 *           bytes. (e.g. the position is byte (base + n*l)).
 *
 * Args:     sfp - open SSIFILE
 *           base  - offset of record 0 (e.g. sfp->foffset)
 *           len   - size of each record in bytes (e.g. sfp->frecsize)
 *           n     - which record to get (e.g. 0..sfp->nfiles)
 *
 * Returns:  0 on success, non-zero on failure.
 */
static int
indexfile_position (SSIFILE * sfp, SSIOFFSET * base, sqd_uint32 len,
                    sqd_uint32 n)
{
    SSIOFFSET pos;
    int status;

    if (base->mode == SSI_OFFSET_I32) {
        pos.mode = SSI_OFFSET_I32;
        pos.off.i32 = base->off.i32 + n * len;
    }
    else if (base->mode == SSI_OFFSET_I64) {
        pos.mode = SSI_OFFSET_I64;
        pos.off.i64 = base->off.i64 + n * len;
    }
    else {
        return 0;
    }
    if ((status = SSISetFilePosition (sfp->fp, &pos)) != 0) {
        return status;
    }
    return 0;
}

/* Function: current_index_size()
 * Date:     SRE, Tue Feb 20 18:23:30 2001 [St. Louis]
 *
 * Purpose:  Calculates the size of the current index,
 *           in megabytes.
 */
static sqd_uint64
current_index_size (SSIINDEX * g)
{
    sqd_uint64 frecsize, precsize, srecsize;
    sqd_uint64 total;

    /* Magic-looking numbers come from adding up sizes
     * of things in bytes
     */
    frecsize = 16 + g->flen;
    precsize = (g->smode == SSI_OFFSET_I64) ? 22 + g->plen : 14 + g->plen;
    srecsize = g->plen + g->slen;
    total = (66L +		/* header size, if 64bit index offsets */
             frecsize * g->nfiles +	/* file section size                   */
             precsize * g->nprimary +	/* primary key section size            */
             srecsize * g->nsecondary) /	/* secondary key section size          */
            1048576L;
    return total;
}

/* Function: activate_external_sort()
 * Date:     SRE, Mon Feb  4 09:08:08 2002 [St. Louis]
 *
 * Purpose:  Switch to external sort mode.
 *           Open file handles for external index files (ptmp, stmp).
 *           Flush current index information to these files.
 *           Free current memory, turn over control to the tmpfiles.
 *
 * Return:   0 on success; non-zero on failure.
 */
static int
activate_external_sort (SSIINDEX * g)
{
    int i;
    /* it's a bit late to be checking this, but... */
    if (g->external) {
        return 0;    /* we already are external, fool */
    }
    if (FileExists (g->ptmpfile)) {
        return 1;
    }
    if (FileExists (g->stmpfile)) {
        return 1;
    }
    if ((g->ptmp = fopen (g->ptmpfile, "w")) == NULL) {
        return 1;
    }
    if ((g->stmp = fopen (g->stmpfile, "w")) == NULL) {
        return 1;
    }

    /* Flush the current indices.
     */
    SQD_DPRINTF1 (("Switching to external sort - flushing ssiindex to disk...\n"));
    for (i = 0; i < g->nprimary; i++) {
        if (g->smode == SSI_OFFSET_I32) {
            fprintf (g->ptmp, "%s\t%u\t%lu\t%lu\t%lu\n",
                     g->pkeys[i].key, (unsigned int) g->pkeys[i].fnum,
                     (unsigned long) g->pkeys[i].r_off.off.i32,
                     (unsigned long) g->pkeys[i].d_off.off.i32,
                     (unsigned long) g->pkeys[i].len);
        }
        else {
            fprintf (g->ptmp, "%s\t%u\t%llu\t%llu\t%lu\n",
                     g->pkeys[i].key, (unsigned int) g->pkeys[i].fnum,
                     (unsigned long long) g->pkeys[i].r_off.off.i64,
                     (unsigned long long) g->pkeys[i].d_off.off.i64,
                     (unsigned long) g->pkeys[i].len);
        }
    }
    for (i = 0; i < g->nsecondary; i++) {
        fprintf (g->stmp, "%s\t%s\n", g->skeys[i].key, g->skeys[i].pkey);
    }

    /* Free the memory now that we've flushed our lists to disk
     */
    for (i = 0; i < g->nprimary; i++) {
        free (g->pkeys[i].key);
    }
    for (i = 0; i < g->nsecondary; i++) {
        free (g->skeys[i].key);
    }
    for (i = 0; i < g->nsecondary; i++) {
        free (g->skeys[i].pkey);
    }
    if (g->pkeys != NULL) {
        free (g->pkeys);
    }
    if (g->skeys != NULL) {
        free (g->skeys);
    }
    g->pkeys = NULL;
    g->skeys = NULL;

    /* Turn control over to external accumulation mode.
     */
    g->external = TRUE;
    return 0;
}


/*****************************************************************
 * Debugging API
 *****************************************************************/
void
SSIForceExternalSort (SSIINDEX * g)
{
    if (activate_external_sort (g) != 0) {
        Die ("failed to turn external sorting on.");
    }
}


/*****************************************************************
 * Test driving mode
 *****************************************************************/
#ifdef MUGGINS_LETS_ME_SLEEP
/* Minimally:
   cc -g -Wall -o shiva -DDEBUGLEVEL=1 -DMUGGINS_LETS_ME_SLEEP ssi.c sqerror.c sre_string.c types.c sre_ctype.c sre_math.c file.c -lm
*/

int
main (int argc, char **argv)
{
    char name[32], accession[32];
    SSIINDEX *ssi;
    int mode;
    SSIOFFSET r_off, d_off;
    FILE *ofp;
    int i;
    int fh;			/* a file handle */
    int status;			/* return status from a SSI call */

    mode = SSI_OFFSET_I32;
    if ((ssi = SSICreateIndex (mode)) == NULL) {
        Die ("Failed to allocate SSI index");
    }

    /* Generate two FASTA files, tmp.0 and tmp.1, and index them.
     */
    if ((ofp = fopen ("tmp.0", "w")) == NULL) {
        Die ("failed to open tmp.0");
    }
    if ((status = SSIAddFileToIndex (ssi, "tmp.0", SQFILE_FASTA, &fh)) != 0) {
        Die ("SSIAddFileToIndex() failed: %s", SSIErrorString (status));
    }
    for (i = 0; i < 10; i++) {
        if ((status = SSIGetFilePosition (ofp, mode, &r_off)) != 0) {
            Die ("SSIGetFilePosition() failed: %s", SSIErrorString (status));
        }
        sprintf (name, "seq%d", i);
        sprintf (accession, "ac%d", i);
        fprintf (ofp,
                 ">%s [%s] Description? we don't need no steenking description.\n",
                 name, accession);
        if ((status = SSIGetFilePosition (ofp, mode, &d_off)) != 0) {
            Die ("SSIGetFilePosition() failed: %s", SSIErrorString (status));
        }
        fprintf (ofp, "AAAAAAAAAA\n");
        fprintf (ofp, "CCCCCCCCCC\n");
        fprintf (ofp, "GGGGGGGGGG\n");
        fprintf (ofp, "TTTTTTTTTT\n");

        if ((status =
                    SSIAddPrimaryKeyToIndex (ssi, name, fh, &r_off, &d_off, 40)) != 0) {
            Die ("SSIAddPrimaryKeyToIndex() failed: %s", SSIErrorString (status));
        }
        if ((status = SSIAddSecondaryKeyToIndex (ssi, accession, name)) != 0) {
            Die ("SSIAddSecondaryKeyToIndex() failed: %s", SSIErrorString (status));
        }
    }
    SSISetFileForSubseq (ssi, fh, 11, 10);
    fclose (ofp);

    if ((ofp = fopen ("tmp.1", "w")) == NULL) {
        Die ("failed to open tmp.1");
    }
    if ((status = SSIAddFileToIndex (ssi, "tmp.1", SQFILE_FASTA, &fh)) != 0) {
        Die ("SSIAddFileToIndex() failed: %s", SSIErrorString (status));
    }
    for (i = 10; i < 20; i++) {
        if ((status = SSIGetFilePosition (ofp, mode, &r_off)) != 0) {
            Die ("SSIGetFilePosition() failed: %s", SSIErrorString (status));
        }
        sprintf (name, "seq%d", i);
        sprintf (accession, "ac%d", i);
        fprintf (ofp, ">%s [%s] i/o, i/o, it's off to disk we go.\n",
                 name, accession);
        if ((status = SSIGetFilePosition (ofp, mode, &d_off)) != 0) {
            Die ("SSIGetFilePosition() failed: %s", SSIErrorString (status));
        }
        fprintf (ofp, "AAAAAAAAAA 10\n");
        fprintf (ofp, "CCCCCCCCCC 20\n");
        fprintf (ofp, "GGGGGGGGGG 30\n");
        fprintf (ofp, "TTTTTTTTTT 40\n");

        if ((status =
                    SSIAddPrimaryKeyToIndex (ssi, name, fh, &r_off, &d_off, 40)) != 0) {
            Die ("SSIAddPrimaryKeyToIndex() failed: %s", SSIErrorString (status));
        }
        if ((status = SSIAddSecondaryKeyToIndex (ssi, accession, name)) != 0) {
            Die ("SSIAddSecondaryKeyToIndex() failed: %s", SSIErrorString (status));
        }
    }
    SSISetFileForSubseq (ssi, fh, 14, 10);
    fclose (ofp);

    /* Write the index to tmp.ssi
     */
    if ((status = SSIWriteIndex ("tmp.ssi", ssi)) != 0) {
        Die ("SSIWriteIndex() failed: %s", SSIErrorString (status));
    }
    SSIFreeIndex (ssi);

    /* Now reopen the index and run some tests.
     */
    exit (0);
}


#endif /* test driving code */



/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* stack.c
 * SRE, Thu Mar  3 10:08:48 1994
 *
 * Implementation of generic stack structures.
 * RCS $Id: stack.c,v 1.3 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdlib.h>
#include "squid.h"


/************************************************************
 * intstack_s implementation.
 *
 * Functions: InitIntStack() - returns ptr to new stack
 *            PushIntStack() - (void)
 *            PopIntStack()  - returns 1 on success, 0 if stack empty
 *            FreeIntStack() - returns number of elements free'd, or 0 if
 *                             stack was empty.
 *
 * Implementation of the pushdown stack for storing single
 * integers.
 *************************************************************/
struct intstack_s *
InitIntStack (void)
{
    struct intstack_s *stack;

    if ((stack =
                (struct intstack_s *) malloc (sizeof (struct intstack_s))) == NULL) {
        Die ("Memory allocation failure at %s line %d", __FILE__, __LINE__);
    }
    stack->nxt = NULL;
    return stack;
}

void
PushIntStack (struct intstack_s *stack, int data)
{
    struct intstack_s *new;

    if ((new =
                (struct intstack_s *) malloc (sizeof (struct intstack_s))) == NULL) {
        Die ("Memory allocation failure at %s line %d", __FILE__, __LINE__);
    }
    new->data = data;

    new->nxt = stack->nxt;
    stack->nxt = new;
}

int
PopIntStack (struct intstack_s *stack, int *ret_data)
{
    struct intstack_s *old;

    if (stack->nxt == NULL) {
        return 0;
    }

    old = stack->nxt;
    stack->nxt = old->nxt;

    *ret_data = old->data;
    free (old);
    return 1;
}

void
ReverseIntStack (struct intstack_s *stack)
{
    struct intstack_s *old;
    struct intstack_s *new;

    old = stack->nxt;
    stack->nxt = NULL;
    while (old != NULL) {
        new = old;			/* remove one from top of old stack */
        old = old->nxt;
        new->nxt = stack->nxt;	/* push it onto new stack */
        stack->nxt = new;
    }
}

int
FreeIntStack (struct intstack_s *stack)
{
    int data;
    int count = 0;

    while (PopIntStack (stack, &data)) {
        count++;
    }
    free (stack);
    return count;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* stockholm.c
 * SRE, Fri May 28 15:46:41 1999
 *
 * Reading/writing of Stockholm format multiple sequence alignments.
 *
 * example of API:
 *
 * MSA     *msa;
 * FILE    *fp;        -- opened for write with fopen()
 * MSAFILE *afp;       -- opened for read with MSAFileOpen()
 *
 * while ((msa = ReadStockholm(afp)) != NULL)
 *   {
 *      WriteStockholm(fp, msa);
 *      MSAFree(msa);
 *   }
 *
 * RCS $Id: stockholm.c,v 1.8 2003/04/14 16:00:16 eddy Exp $
 */
#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include "squid.h"
#include "msa.h"

static int parse_gf (MSA * msa, char *buf);
static int parse_gs (MSA * msa, char *buf);
static int parse_gc (MSA * msa, char *buf);
static int parse_gr (MSA * msa, char *buf);
static int parse_comment (MSA * msa, char *buf);
static int parse_sequence (MSA * msa, char *buf);
static void actually_write_stockholm (FILE * fp, MSA * msa, int cpl);

#ifdef TESTDRIVE_STOCKHOLM
/*****************************************************************
 * stockholm.c test driver:
 * cc -DTESTDRIVE_STOCKHOLM -g -O2 -Wall -o test stockholm.c msa.c gki.c sqerror.c sre_string.c file.c hsregex.c sre_math.c sre_ctype.c -lm
 *
 */
int
main (int argc, char **argv)
{
    MSAFILE *afp;
    MSA *msa;
    char *file;

    file = argv[1];

    if ((afp = MSAFileOpen (file, MSAFILE_STOCKHOLM, NULL)) == NULL) {
        Die ("Couldn't open %s\n", file);
    }

    while ((msa = ReadStockholm (afp)) != NULL) {
        WriteStockholm (stdout, msa);
        MSAFree (msa);
    }

    MSAFileClose (afp);
    exit (0);
}

/******************************************************************/
#endif /* testdriver */


/* Function: ReadStockholm()
 * Date:     SRE, Fri May 21 17:33:10 1999 [St. Louis]
 *
 * Purpose:  Parse the next alignment from an open Stockholm
 *           format alignment file. Return the alignment, or
 *           NULL if there are no more alignments in the file.
 *
 * Args:     afp  - open alignment file
 *
 * Returns:  MSA *   - an alignment object.
 *                     caller responsible for an MSAFree()
 *           NULL if no more alignments
 *
 * Diagnostics:
 *           Will Die() here with a (potentially) useful message
 *           if a parsing error occurs
 */
MSA *
ReadStockholm (MSAFILE * afp)
{
    MSA *msa;
    char *s;
    int status;

    if (feof (afp->f)) {
        return NULL;
    }

    /* Initialize allocation of the MSA.
     */
    msa = MSAAlloc (10, 0);

    /* Check the magic Stockholm header line.
     * We have to skip blank lines here, else we perceive
     * trailing blank lines in a file as a format error when
     * reading in multi-record mode.
     */
    do {
        if ((s = MSAFileGetLine (afp)) == NULL) {
            MSAFree (msa);
            return NULL;
        }
    } while (IsBlankline (s));

    if (strncmp (s, "# STOCKHOLM 1.", 14) != 0)
        Die ("\
File %s doesn't appear to be in Stockholm format.\n\
Assuming there isn't some other problem with your file (it is an\n\
alignment file, right?), please either:\n\
  a) use the Babelfish format autotranslator option (-B, usually);\n\
  b) specify the file's format with the --informat option; or\n\
  a) reformat the alignment to Stockholm format.\n", afp->fname);

    /* Read the alignment file one line at a time.
     */
    while ((s = MSAFileGetLine (afp)) != NULL) {
        while (*s == ' ' || *s == '\t') {
            s++;    /* skip leading whitespace */
        }

        if (*s == '#') {
            if (strncmp (s, "#=GF", 4) == 0) {
                status = parse_gf (msa, s);
            }
            else if (strncmp (s, "#=GS", 4) == 0) {
                status = parse_gs (msa, s);
            }
            else if (strncmp (s, "#=GC", 4) == 0) {
                status = parse_gc (msa, s);
            }
            else if (strncmp (s, "#=GR", 4) == 0) {
                status = parse_gr (msa, s);
            }
            else {
                status = parse_comment (msa, s);
            }
        }
        else if (strncmp (s, "//", 2) == 0) {
            break;
        }
        else if (*s == '\n') {
            continue;
        }
        else {
            status = parse_sequence (msa, s);
        }

        if (status == 0)
            Die
            ("Stockholm format parse error: line %d of file %s while reading alignment %s",
             afp->linenumber, afp->fname, msa->name == NULL ? "" : msa->name);
    }

    if (s == NULL && msa->nseq != 0)
        Die ("Didn't find // at end of alignment %s",
             msa->name == NULL ? "" : msa->name);

    if (s == NULL && msa->nseq == 0) {
        /* probably just some junk at end of file */
        MSAFree (msa);
        return NULL;
    }

    MSAVerifyParse (msa);
    return msa;
}


/* Function: WriteStockholm()
 * Date:     SRE, Mon May 31 19:15:22 1999 [St. Louis]
 *
 * Purpose:  Write an alignment in standard multi-block
 *           Stockholm format to an open file. A wrapper
 *           for actually_write_stockholm().
 *
 * Args:     fp  - file that's open for writing
 *           msa - alignment to write
 *
 * Returns:  (void)
 */
void
WriteStockholm (FILE * fp, MSA * msa)
{
    actually_write_stockholm (fp, msa, 50);	/* 50 char per block */
}

/* Function: WriteStockholmOneBlock()
 * Date:     SRE, Mon May 31 19:15:22 1999 [St. Louis]
 *
 * Purpose:  Write an alignment in Pfam's single-block
 *           Stockholm format to an open file. A wrapper
 *           for actually_write_stockholm().
 *
 * Args:     fp  - file that's open for writing
 *           msa - alignment to write
 *
 * Returns:  (void)
 */
void
WriteStockholmOneBlock (FILE * fp, MSA * msa)
{
    actually_write_stockholm (fp, msa, msa->alen);	/* one big block */
}


/* Function: actually_write_stockholm()
 * Date:     SRE, Fri May 21 17:39:22 1999 [St. Louis]
 *
 * Purpose:  Write an alignment in Stockholm format to
 *           an open file. This is the function that actually
 *           does the work. The API's WriteStockholm()
 *           and WriteStockholmOneBlock() are wrappers.
 *
 * Args:     fp    - file that's open for writing
 *           msa   - alignment to write
 *           cpl   - characters to write per line in alignment block
 *
 * Returns:  (void)
 */
static void
actually_write_stockholm (FILE * fp, MSA * msa, int cpl)
{
    int i, j;
    int len = 0;
    int namewidth;
    int typewidth = 0;		/* markup tags are up to 5 chars long */
    int markupwidth = 0;		/* #=GR, #=GC are four char wide + 1 space */
    char *buf;
    int currpos;
    char *s, *tok;

    /* Figure out how much space we need for name + markup
     * to keep the alignment in register. Required by Stockholm
     * spec, even though our Stockholm parser doesn't care (Erik's does).
     */
    namewidth = 0;
    for (i = 0; i < msa->nseq; i++) {
        if ((len = strlen (msa->sqname[i])) > namewidth) {
            namewidth = len;
        }
		}
    /* Figure out how much space we need for markup tags
     *   markupwidth = always 4 if we're doing markup:  strlen("#=GR")
     *   typewidth   = longest markup tag
     */
    if (msa->ss != NULL) {
        markupwidth = 4;
        typewidth = 2;
    }
    if (msa->sa != NULL) {
        markupwidth = 4;
        typewidth = 2;
    }
    for (i = 0; i < msa->ngr; i++) {
        if ((len = strlen (msa->gr_tag[i])) > typewidth) {
            typewidth = len;
        }
		}
    if (msa->rf != NULL) {
        markupwidth = 4;
        if (typewidth < 2) {
            typewidth = 2;
        }
    }
    if (msa->ss_cons != NULL) {
        markupwidth = 4;
        if (typewidth < 7) {
            typewidth = 7;
        }
    }
    if (msa->sa_cons != NULL) {
        markupwidth = 4;
        if (typewidth < 7) {
            typewidth = 7;
        }
    }
    for (i = 0; i < msa->ngc; i++) {
        if ((len = strlen (msa->gc_tag[i])) > typewidth) {
            typewidth = len;
        }
		}
    buf =
        MallocOrDie (sizeof (char) *
                     (cpl + namewidth + typewidth + markupwidth + 61));

    /* Magic Stockholm header
     */
    fprintf (fp, "# STOCKHOLM 1.0\n");

    /* Free text comments
     */
    for (i = 0; i < msa->ncomment; i++) {
        fprintf (fp, "# %s\n", msa->comment[i]);
    }
    if (msa->ncomment > 0) {
        fprintf (fp, "\n");
    }

    /* GF section: per-file annotation
     */
    if (msa->name != NULL) {
        fprintf (fp, "#=GF ID    %s\n", msa->name);
    }
    if (msa->acc != NULL) {
        fprintf (fp, "#=GF AC    %s\n", msa->acc);
    }
    if (msa->desc != NULL) {
        fprintf (fp, "#=GF DE    %s\n", msa->desc);
    }
    if (msa->au != NULL) {
        fprintf (fp, "#=GF AU    %s\n", msa->au);
    }

    /* Thresholds are hacky. Pfam has two. Rfam has one.
     */
    if (msa->cutoff_is_set[MSA_CUTOFF_GA1]
            && msa->cutoff_is_set[MSA_CUTOFF_GA2])
        fprintf (fp, "#=GF GA    %.1f %.1f\n", msa->cutoff[MSA_CUTOFF_GA1],
                 msa->cutoff[MSA_CUTOFF_GA2]);
    else if (msa->cutoff_is_set[MSA_CUTOFF_GA1]) {
        fprintf (fp, "#=GF GA    %.1f\n", msa->cutoff[MSA_CUTOFF_GA1]);
    }
    if (msa->cutoff_is_set[MSA_CUTOFF_NC1]
            && msa->cutoff_is_set[MSA_CUTOFF_NC2])
        fprintf (fp, "#=GF NC    %.1f %.1f\n", msa->cutoff[MSA_CUTOFF_NC1],
                 msa->cutoff[MSA_CUTOFF_NC2]);
    else if (msa->cutoff_is_set[MSA_CUTOFF_NC1]) {
        fprintf (fp, "#=GF NC    %.1f\n", msa->cutoff[MSA_CUTOFF_NC1]);
    }
    if (msa->cutoff_is_set[MSA_CUTOFF_TC1]
            && msa->cutoff_is_set[MSA_CUTOFF_TC2])
        fprintf (fp, "#=GF TC    %.1f %.1f\n", msa->cutoff[MSA_CUTOFF_TC1],
                 msa->cutoff[MSA_CUTOFF_TC2]);
    else if (msa->cutoff_is_set[MSA_CUTOFF_TC1]) {
        fprintf (fp, "#=GF TC    %.1f\n", msa->cutoff[MSA_CUTOFF_TC1]);
    }

    for (i = 0; i < msa->ngf; i++) {
        fprintf (fp, "#=GF %-5s %s\n", msa->gf_tag[i], msa->gf[i]);
    }
    fprintf (fp, "\n");


    /* GS section: per-sequence annotation
     */
    if (msa->flags & MSA_SET_WGT) {
        for (i = 0; i < msa->nseq; i++) {
            fprintf (fp, "#=GS %-*.*s WT    %.2f\n", namewidth, namewidth,
                     msa->sqname[i], msa->wgt[i]);
				}						 
        fprintf (fp, "\n");
    }
    if (msa->sqacc != NULL) {
        for (i = 0; i < msa->nseq; i++) {
            if (msa->sqacc[i] != NULL)
                fprintf (fp, "#=GS %-*.*s AC    %s\n", namewidth, namewidth,
                         msa->sqname[i], msa->sqacc[i]);
				}								 
        fprintf (fp, "\n");
    }
    if (msa->sqdesc != NULL) {
        for (i = 0; i < msa->nseq; i++) {
            if (msa->sqdesc[i] != NULL)
                fprintf (fp, "#=GS %*.*s DE    %s\n", namewidth, namewidth,
                         msa->sqname[i], msa->sqdesc[i]);
				}								 
        fprintf (fp, "\n");
    }
    for (i = 0; i < msa->ngs; i++) {
        /* Multiannotated GS tags are possible; for example,
         *     #=GS foo DR PDB; 1xxx;
         *     #=GS foo DR PDB; 2yyy;
         * These are stored, for example, as:
         *     msa->gs[0][0] = "PDB; 1xxx;\nPDB; 2yyy;"
         * and must be decomposed.
         */
        for (j = 0; j < msa->nseq; j++) {
            if (msa->gs[i][j] != NULL) {
                s = msa->gs[i][j];
                while ((tok = sre_strtok (&s, "\n", NULL)) != NULL)
                    fprintf (fp, "#=GS %*.*s %5s %s\n", namewidth, namewidth,
                             msa->sqname[j], msa->gs_tag[i], tok);
            }
				}		
        fprintf (fp, "\n");
    }

    /* Alignment section:
     * contains aligned sequence, #=GR annotation, and #=GC annotation
     */
    for (currpos = 0; currpos < msa->alen; currpos += cpl) {
        if (currpos > 0) {
            fprintf (fp, "\n");
        }
        for (i = 0; i < msa->nseq; i++) {
            strncpy (buf, msa->aseq[i] + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "%-*.*s  %s\n", namewidth + typewidth + markupwidth,
                     namewidth + typewidth + markupwidth, msa->sqname[i], buf);

            if (msa->ss != NULL && msa->ss[i] != NULL) {
                strncpy (buf, msa->ss[i] + currpos, cpl);
                buf[cpl] = '\0';
                fprintf (fp, "#=GR %-*.*s SS     %s\n", namewidth, namewidth,
                         msa->sqname[i], buf);
            }
            if (msa->sa != NULL && msa->sa[i] != NULL) {
                strncpy (buf, msa->sa[i] + currpos, cpl);
                buf[cpl] = '\0';
                fprintf (fp, "#=GR %-*.*s SA     %s\n", namewidth, namewidth,
                         msa->sqname[i], buf);
            }
            for (j = 0; j < msa->ngr; j++) {
                if (msa->gr[j][i] != NULL) {
                    strncpy (buf, msa->gr[j][i] + currpos, cpl);
                    buf[cpl] = '\0';
                    fprintf (fp, "#=GR %-*.*s %5s  %s\n",
                             namewidth, namewidth, msa->sqname[i], msa->gr_tag[j], buf);
                }
						}		
        }
        if (msa->ss_cons != NULL) {
            strncpy (buf, msa->ss_cons + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "#=GC %-*.*s %s\n", namewidth + typewidth,
                     namewidth + typewidth, "SS_cons", buf);
        }

        if (msa->sa_cons != NULL) {
            strncpy (buf, msa->sa_cons + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "#=GC %-*.*s %s\n", namewidth + typewidth,
                     namewidth + typewidth, "SA_cons", buf);
        }

        if (msa->rf != NULL) {
            strncpy (buf, msa->rf + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "#=GC %-*.*s %s\n", namewidth + typewidth,
                     namewidth + typewidth, "RF", buf);
        }
        for (j = 0; j < msa->ngc; j++) {
            strncpy (buf, msa->gc[j] + currpos, cpl);
            buf[cpl] = '\0';
            fprintf (fp, "#=GC %-*.*s %s\n", namewidth + typewidth,
                     namewidth + typewidth, msa->gc_tag[j], buf);
        }
    }
    fprintf (fp, "//\n");
    free (buf);
}





/* Format of a GF line:
 *    #=GF <featurename> <text>
 */
static int
parse_gf (MSA * msa, char *buf)
{
    char *gf;
    char *featurename;
    char *text;
    char *s;

    s = buf;
    if ((gf = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((featurename = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((text = sre_strtok (&s, "\n", NULL)) == NULL) {
        return 0;
    }
    while (*text && (*text == ' ' || *text == '\t')) {
        text++;
    }

    if (strcmp (featurename, "ID") == 0) {
        msa->name = sre_strdup (text, -1);
    }
    else if (strcmp (featurename, "AC") == 0) {
        msa->acc = sre_strdup (text, -1);
    }
    else if (strcmp (featurename, "DE") == 0) {
        msa->desc = sre_strdup (text, -1);
    }
    else if (strcmp (featurename, "AU") == 0) {
        msa->au = sre_strdup (text, -1);
    }
    else if (strcmp (featurename, "GA") == 0) {	/* Pfam has GA1, GA2. Rfam just has GA1. */
        s = text;
        if ((text = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
            return 0;
        }
        msa->cutoff[MSA_CUTOFF_GA1] = atof (text);
        msa->cutoff_is_set[MSA_CUTOFF_GA1] = TRUE;
        if ((text = sre_strtok (&s, WHITESPACE, NULL)) != NULL) {
            msa->cutoff[MSA_CUTOFF_GA2] = atof (text);
            msa->cutoff_is_set[MSA_CUTOFF_GA2] = TRUE;
        }
    }
    else if (strcmp (featurename, "NC") == 0) {
        s = text;
        if ((text = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
            return 0;
        }
        msa->cutoff[MSA_CUTOFF_NC1] = atof (text);
        msa->cutoff_is_set[MSA_CUTOFF_NC1] = TRUE;
        if ((text = sre_strtok (&s, WHITESPACE, NULL)) != NULL) {
            msa->cutoff[MSA_CUTOFF_NC2] = atof (text);
            msa->cutoff_is_set[MSA_CUTOFF_NC2] = TRUE;
        }
    }
    else if (strcmp (featurename, "TC") == 0) {
        s = text;
        if ((text = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
            return 0;
        }
        msa->cutoff[MSA_CUTOFF_TC1] = atof (text);
        msa->cutoff_is_set[MSA_CUTOFF_TC1] = TRUE;
        if ((text = sre_strtok (&s, WHITESPACE, NULL)) != NULL) {
            msa->cutoff[MSA_CUTOFF_TC2] = atof (text);
            msa->cutoff_is_set[MSA_CUTOFF_TC2] = TRUE;
        }
    }
    else {
        MSAAddGF (msa, featurename, text);
    }

    return 1;
}


/* Format of a GS line:
 *    #=GS <seqname> <featurename> <text>
 */
static int
parse_gs (MSA * msa, char *buf)
{
    char *gs;
    char *seqname;
    char *featurename;
    char *text;
    int seqidx;
    char *s;

    s = buf;
    if ((gs = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((seqname = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((featurename = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((text = sre_strtok (&s, "\n", NULL)) == NULL) {
        return 0;
    }
    while (*text && (*text == ' ' || *text == '\t')) {
        text++;
    }

    /* GS usually follows another GS; guess lastidx+1
     */
    seqidx = MSAGetSeqidx (msa, seqname, msa->lastidx + 1);
    msa->lastidx = seqidx;

    if (strcmp (featurename, "WT") == 0) {
        msa->wgt[seqidx] = atof (text);
        msa->flags |= MSA_SET_WGT;
    }

    else if (strcmp (featurename, "AC") == 0) {
        MSASetSeqAccession (msa, seqidx, text);
    }

    else if (strcmp (featurename, "DE") == 0) {
        MSASetSeqDescription (msa, seqidx, text);
    }

    else {
        MSAAddGS (msa, featurename, seqidx, text);
    }

    return 1;
}

/* Format of a GC line:
 *    #=GC <featurename> <text>
 */
static int
parse_gc (MSA * msa, char *buf)
{
    char *gc;
    char *featurename;
    char *text;
    char *s;
    int len;

    s = buf;
    if ((gc = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((featurename = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((text = sre_strtok (&s, WHITESPACE, &len)) == NULL) {
        return 0;
    }

    if (strcmp (featurename, "SS_cons") == 0) {
        sre_strcat (&(msa->ss_cons), -1, text, len);
    }
    else if (strcmp (featurename, "SA_cons") == 0) {
        sre_strcat (&(msa->sa_cons), -1, text, len);
    }
    else if (strcmp (featurename, "RF") == 0) {
        sre_strcat (&(msa->rf), -1, text, len);
    }
    else {
        MSAAppendGC (msa, featurename, text);
    }

    return 1;
}

/* Format of a GR line:
 *    #=GR <seqname> <featurename> <text>
 */
static int
parse_gr (MSA * msa, char *buf)
{
    char *gr;
    char *seqname;
    char *featurename;
    char *text;
    int seqidx;
    int len;
    int j;
    char *s;

    s = buf;
    if ((gr = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((seqname = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((featurename = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((text = sre_strtok (&s, WHITESPACE, &len)) == NULL) {
        return 0;
    }

    /* GR usually follows sequence it refers to; guess msa->lastidx */
    seqidx = MSAGetSeqidx (msa, seqname, msa->lastidx);
    msa->lastidx = seqidx;

    if (strcmp (featurename, "SS") == 0) {
        if (msa->ss == NULL) {
            msa->ss = MallocOrDie (sizeof (char *) * msa->nseqalloc);
            msa->sslen = MallocOrDie (sizeof (int) * msa->nseqalloc);
            for (j = 0; j < msa->nseqalloc; j++) {
                msa->ss[j] = NULL;
                msa->sslen[j] = 0;
            }
        }
        msa->sslen[seqidx] =
            sre_strcat (&(msa->ss[seqidx]), msa->sslen[seqidx], text, len);
    }
    else if (strcmp (featurename, "SA") == 0) {
        if (msa->sa == NULL) {
            msa->sa = MallocOrDie (sizeof (char *) * msa->nseqalloc);
            msa->salen = MallocOrDie (sizeof (int) * msa->nseqalloc);
            for (j = 0; j < msa->nseqalloc; j++) {
                msa->sa[j] = NULL;
                msa->salen[j] = 0;
            }
        }
        msa->salen[seqidx] =
            sre_strcat (&(msa->sa[seqidx]), msa->salen[seqidx], text, len);
    }
    else {
        MSAAppendGR (msa, featurename, seqidx, text);
    }

    return 1;
}


/* comments are simply stored verbatim, not parsed
 */
static int
parse_comment (MSA * msa, char *buf)
{
    char *s;
    char *comment;

    s = buf + 1;			/* skip leading '#' */
    if (*s == '\n') {
        *s = '\0';
        comment = s;
    }				/* deal with blank comment */
    else if ((comment = sre_strtok (&s, "\n", NULL)) == NULL) {
        return 0;
    }

    MSAAddComment (msa, comment);
    return 1;
}

static int
parse_sequence (MSA * msa, char *buf)
{
    char *s;
    char *seqname;
    char *text;
    int seqidx;
    int len;

    s = buf;
    if ((seqname = sre_strtok (&s, WHITESPACE, NULL)) == NULL) {
        return 0;
    }
    if ((text = sre_strtok (&s, WHITESPACE, &len)) == NULL) {
        return 0;
    }

    /* seq usually follows another seq; guess msa->lastidx +1 */
    seqidx = MSAGetSeqidx (msa, seqname, msa->lastidx + 1);
    msa->lastidx = seqidx;

    msa->sqlen[seqidx] =
        sre_strcat (&(msa->aseq[seqidx]), msa->sqlen[seqidx], text, len);
    return 1;
}



/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* tophits.c
 *
 * Routines for storing, sorting, displaying high scoring hits
 * and alignments.
 *
 *****************************************************************************
 *
 * main API:
 *
 * AllocTophits()       - allocation
 * FreeTophits()        - free'ing
 * RegisterHit()        - put information about a hit in the list
 * GetRankedHit()       - recovers information about a hit
 * FullSortTophits()    - sorts the top H hits.
 *
 *****************************************************************************
 * Brief example of use:
 *
 *   struct tophit_s   *yourhits;   // list of hits
 *   struct fancyali_s *ali;        // (optional structure) alignment of a hit
 *
 *   yourhits = AllocTophits(200);
 *   (for every hit in a search) {
 *        if (do_alignments)
 *           ali = Trace2FancyAli();  // You provide a function/structure here
 *        if (score > threshold)
 *           RegisterHit(yourhits, ...)
 *     }
 *
 *   FullSortTophits(yourhits);      // Sort hits by evalue
 *   for (i = 0; i < 100; i++)       // Recover hits out in ranked order
 *     {
 *       GetRankedHit(yourhits, i, ...);
 *                                   // Presumably you'd print here...
 *     }
 *   FreeTophits(yourhits);
 ***************************************************************************
 *
 * Estimated storage per hit:
 *        coords:   16 bytes
 *        scores:    8 bytes
 * name/acc/desc:  192 bytes
 *     alignment: 1000 bytes   total = ~1200 bytes with alignment;
 *                                   = ~200 bytes without
 *     Designed for: 10^5 hits (20 MB) or 10^4 alignments (10 MB)
 */

#include "config.h"
#include "squidconf.h"

#include <string.h>
#include <float.h>
#include <limits.h>

#include "structs.h"
#include "funcs.h"

/* Function: AllocTophits()
 *
 * Purpose:  Allocate a struct tophit_s, for maintaining
 *           a list of top-scoring hits in a database search.
 *
 * Args:     lumpsize - allocation lumpsize
 *
 * Return:   An allocated struct hit_s. Caller must free.
 */
struct tophit_s *
AllocTophits (int lumpsize)
{
    struct tophit_s *hitlist;

    hitlist = MallocOrDie (sizeof (struct tophit_s));
    hitlist->hit = NULL;
    hitlist->unsrt = MallocOrDie (lumpsize * sizeof (struct hit_s));
    hitlist->alloc = lumpsize;
    hitlist->num = 0;
    hitlist->lump = lumpsize;
    return hitlist;
}

void
GrowTophits (struct tophit_s *h)
{
    h->unsrt =
        ReallocOrDie (h->unsrt, (h->alloc + h->lump) * sizeof (struct hit_s));
    h->alloc += h->lump;
}

void
FreeTophits (struct tophit_s *h)
{
    int pos;
    for (pos = 0; pos < h->num; pos++) {
        if (h->unsrt[pos].ali != NULL) {
            FreeFancyAli (h->unsrt[pos].ali);
        }
        if (h->unsrt[pos].name != NULL) {
            free (h->unsrt[pos].name);
        }
        if (h->unsrt[pos].acc != NULL) {
            free (h->unsrt[pos].acc);
        }
        if (h->unsrt[pos].desc != NULL) {
            free (h->unsrt[pos].desc);
        }
    }
    free (h->unsrt);
    if (h->hit != NULL) {
        free (h->hit);
    }
    free (h);
}

struct fancyali_s *
AllocFancyAli (void)
{
    struct fancyali_s *ali;

    ali = MallocOrDie (sizeof (struct fancyali_s));
    ali->rfline = ali->csline = ali->model = ali->mline = ali->aseq = NULL;
    ali->query = ali->target = NULL;
    ali->sqfrom = ali->sqto = 0;
    return ali;
}

void
FreeFancyAli (struct fancyali_s *ali)
{
    if (ali != NULL) {
        if (ali->rfline != NULL) {
            free (ali->rfline);
        }
        if (ali->csline != NULL) {
            free (ali->csline);
        }
        if (ali->model != NULL) {
            free (ali->model);
        }
        if (ali->mline != NULL) {
            free (ali->mline);
        }
        if (ali->aseq != NULL) {
            free (ali->aseq);
        }
        if (ali->query != NULL) {
            free (ali->query);
        }
        if (ali->target != NULL) {
            free (ali->target);
        }
        free (ali);
    }
}

/* Function: RegisterHit()
 *
 * Purpose:  Add a new hit to a list of top hits.
 *
 *           "ali", if provided, is a pointer to allocated memory
 *           for an alignment output structure.
 *           Management is turned over to the top hits structure.
 *           Caller should not free them; they will be free'd by
 *           the FreeTophits() call.
 *
 *           In contrast, "name", "acc", and "desc" are copied, so caller
 *           is still responsible for these.
 *
 *           Number of args is unwieldy.
 *
 * Args:     h        - active top hit list
 *           key      - value to sort by: bigger is better
 *           pvalue   - P-value of this hit
 *           score    - score of this hit
 *           motherp  - P-value of parent whole sequence
 *           mothersc - score of parent whole sequence
 *           name     - name of target
 *           acc      - accession of target (may be NULL)
 *           desc     - description of target (may be NULL)
 *           sqfrom   - 1..L pos in target seq  of start
 *           sqto     - 1..L pos; sqfrom > sqto if rev comp
 *           sqlen    - length of sequence, L
 *           hmmfrom  - 0..M+1 pos in HMM of start
 *           hmmto    - 0..M+1 pos in HMM of end
 *           hmmlen   - length of HMM, M
 *           domidx   - number of this domain
 *           ndom     - total # of domains in sequence
 *           ali      - optional printable alignment info
 *
 * Return:   (void)
 *           hitlist is modified and possibly reallocated internally.
 */
void
RegisterHit (struct tophit_s *h, double key,
             double pvalue, float score, double motherp, float mothersc,
             char *name, char *acc, char *desc,
             int sqfrom, int sqto, int sqlen,
             int hmmfrom, int hmmto, int hmmlen,
             int domidx, int ndom, struct fancyali_s *ali)
{
    /* Check to see if list is full and we must realloc.
     */
    if (h->num == h->alloc) {
        GrowTophits (h);
    }

    h->unsrt[h->num].name = Strdup (name);
    h->unsrt[h->num].acc = Strdup (acc);
    h->unsrt[h->num].desc = Strdup (desc);
    h->unsrt[h->num].sortkey = key;
    h->unsrt[h->num].pvalue = pvalue;
    h->unsrt[h->num].score = score;
    h->unsrt[h->num].motherp = motherp;
    h->unsrt[h->num].mothersc = mothersc;
    h->unsrt[h->num].sqfrom = sqfrom;
    h->unsrt[h->num].sqto = sqto;
    h->unsrt[h->num].sqlen = sqlen;
    h->unsrt[h->num].hmmfrom = hmmfrom;
    h->unsrt[h->num].hmmto = hmmto;
    h->unsrt[h->num].hmmlen = hmmlen;
    h->unsrt[h->num].domidx = domidx;
    h->unsrt[h->num].ndom = ndom;
    h->unsrt[h->num].ali = ali;
    h->num++;
    return;
}

/* Function: GetRankedHit()
 * Date:     SRE, Tue Oct 28 10:06:48 1997 [Newton Institute, Cambridge UK]
 *
 * Purpose:  Recover the data from the i'th ranked hit.
 *           Any of the data ptrs may be passed as NULL for fields
 *           you don't want. hitlist must have been sorted first.
 *
 *           name, acc, desc, and ali are returned as pointers, not copies;
 *           don't free them!
 */
void
GetRankedHit (struct tophit_s *h, int rank,
              double *r_pvalue, float *r_score,
              double *r_motherp, float *r_mothersc,
              char **r_name, char **r_acc, char **r_desc,
              int *r_sqfrom, int *r_sqto, int *r_sqlen,
              int *r_hmmfrom, int *r_hmmto, int *r_hmmlen,
              int *r_domidx, int *r_ndom, struct fancyali_s **r_ali)
{
    if (r_pvalue != NULL) {
        *r_pvalue = h->hit[rank]->pvalue;
    }
    if (r_score != NULL) {
        *r_score = h->hit[rank]->score;
    }
    if (r_motherp != NULL) {
        *r_motherp = h->hit[rank]->motherp;
    }
    if (r_mothersc != NULL) {
        *r_mothersc = h->hit[rank]->mothersc;
    }
    if (r_name != NULL) {
        *r_name = h->hit[rank]->name;
    }
    if (r_acc != NULL) {
        *r_acc = h->hit[rank]->acc;
    }
    if (r_desc != NULL) {
        *r_desc = h->hit[rank]->desc;
    }
    if (r_sqfrom != NULL) {
        *r_sqfrom = h->hit[rank]->sqfrom;
    }
    if (r_sqto != NULL) {
        *r_sqto = h->hit[rank]->sqto;
    }
    if (r_sqlen != NULL) {
        *r_sqlen = h->hit[rank]->sqlen;
    }
    if (r_hmmfrom != NULL) {
        *r_hmmfrom = h->hit[rank]->hmmfrom;
    }
    if (r_hmmto != NULL) {
        *r_hmmto = h->hit[rank]->hmmto;
    }
    if (r_hmmlen != NULL) {
        *r_hmmlen = h->hit[rank]->hmmlen;
    }
    if (r_domidx != NULL) {
        *r_domidx = h->hit[rank]->domidx;
    }
    if (r_ndom != NULL) {
        *r_ndom = h->hit[rank]->ndom;
    }
    if (r_ali != NULL) {
        *r_ali = h->hit[rank]->ali;
    }
}

/* Function: TophitsMaxName()
 *
 * Purpose:  Returns the maximum name length in a top hits list;
 *           doesn't need to be sorted yet.
 */
int
TophitsMaxName (struct tophit_s *h)
{
    int i;
    int len, maxlen;

    maxlen = 0;
    for (i = 0; i < h->num; i++) {
        len = strlen (h->unsrt[i].name);
        if (len > maxlen) {
            maxlen = len;
        }
    }
    return maxlen;
}

/* Function: FullSortTophits()
 *
 * Purpose:  Completely sort the top hits list. Calls
 *           qsort() to do the sorting, and uses
 *           hit_comparison() to do the comparison.
 *
 * Args:     h - top hits structure
 */
int
hit_comparison (const void *vh1, const void *vh2)
{
    /* don't ask. don't change. Don't Panic. */
    struct hit_s *h1 = *((struct hit_s **) vh1);
    struct hit_s *h2 = *((struct hit_s **) vh2);

    if (h1->sortkey < h2->sortkey) {
        return 1;
    }
    else if (h1->sortkey > h2->sortkey) {
        return -1;
    }
    else if (h1->sortkey == h2->sortkey) {
        return 0;
    }
    /*NOTREACHED*/ return 0;
}

void
FullSortTophits (struct tophit_s *h)
{
    int i;

    /* If we don't have /any/ hits, then don't
     * bother.
     */
    if (h->num == 0) {
        return;
    }

    /* Assign the ptrs in h->hit.
     */
    h->hit = MallocOrDie (h->num * sizeof (struct hit_s *));
    for (i = 0; i < h->num; i++) {
        h->hit[i] = &(h->unsrt[i]);
    }

    /* Sort the pointers. Don't bother if we've only got one.
     */
    if (h->num > 1) {
        specqsort (h->hit, h->num, sizeof (struct hit_s *), hit_comparison);
    }
}



/* Function: TophitsReport()
 * Date:     Thu Dec 18 13:19:18 1997
 *
 * Purpose:  Generate a printout summarizing how much
 *           memory is used by a tophits structure,
 *           how many hits are stored, and how much
 *           waste there is from not knowing nseqs.
 *
 * Args:     h    - the sorted tophits list
 *           E    - the cutoff in Evalue
 *           nseq - the final number of seqs used for Eval
 *
 * Return:   (void)
 *           Prints information on stdout
 */
void
TophitsReport (struct tophit_s *h, double E, int nseq)
{
    int i;
    int memused;
    int x;
    int n;

    /* Count up how much memory is used
     * in the whole list.
     */
    memused = sizeof (struct hit_s) * h->alloc + sizeof (struct tophit_s);
    for (i = 0; i < h->num; i++) {
        if (h->unsrt[i].name != NULL) {
            memused += strlen (h->unsrt[i].name) + 1;
        }
        if (h->unsrt[i].acc != NULL) {
            memused += strlen (h->unsrt[i].acc) + 1;
        }
        if (h->unsrt[i].desc != NULL) {
            memused += strlen (h->unsrt[i].desc) + 1;
        }
        if (h->unsrt[i].ali != NULL) {
            memused += sizeof (struct fancyali_s);
            x = 0;
            if (h->unsrt[i].ali->rfline != NULL) {
                x++;
            }
            if (h->unsrt[i].ali->csline != NULL) {
                x++;
            }
            if (h->unsrt[i].ali->model != NULL) {
                x++;
            }
            if (h->unsrt[i].ali->mline != NULL) {
                x++;
            }
            if (h->unsrt[i].ali->aseq != NULL) {
                x++;
            }
            memused += x * (h->unsrt[i].ali->len + 1);

            if (h->unsrt[i].ali->query != NULL) {
                memused += strlen (h->unsrt[i].ali->query) + 1;
            }
            if (h->unsrt[i].ali->target != NULL) {
                memused += strlen (h->unsrt[i].ali->target) + 1;
            }
        }
    }

    /* Count how many hits actually satisfy the E cutoff.
     */
    n = 0;
    for (i = 0; i < h->num; i++) {
        if (h->hit[i]->pvalue * (double) nseq >= E) {
            break;
        }
        n++;
    }

    /* Format and print a summary
     */
    printf ("tophits_s report:\n");
    printf ("     Total hits:           %d\n", h->num);
    printf ("     Satisfying E cutoff:  %d\n", n);
    printf ("     Total memory:         %dK\n", memused / 1000);
}

/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* trace.c
 * SRE, Sat Nov 16 12:34:57 1996
 * RCS $Id: trace.c,v 1.19 2003/04/14 16:00:17 eddy Exp $
 *
 * Support for Plan 7 traceback data structure, p7trace_s.
 */

#include "config.h"
#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "squid.h"
#include "structs.h"
#include "funcs.h"

static void rightjustify (char *s, int n);

/* Function: P7AllocTrace(), P7ReallocTrace(), P7FreeTrace()
 *
 * Purpose:  allocation and freeing of traceback structures
 */
void
P7AllocTrace (int tlen, struct p7trace_s **ret_tr)
{
    struct p7trace_s *tr;

    tr = MallocOrDie (sizeof (struct p7trace_s));
    tr->statetype = MallocOrDie (sizeof (char) * tlen);
    tr->nodeidx = MallocOrDie (sizeof (int) * tlen);
    tr->pos = MallocOrDie (sizeof (int) * tlen);
    *ret_tr = tr;
}

void
P7ReallocTrace (struct p7trace_s *tr, int tlen)
{
    tr->statetype = ReallocOrDie (tr->statetype, tlen * sizeof (char));
    tr->nodeidx = ReallocOrDie (tr->nodeidx, tlen * sizeof (int));
    tr->pos = ReallocOrDie (tr->pos, tlen * sizeof (int));
}

void
P7FreeTrace (struct p7trace_s *tr)
{
    if (tr == NULL) {
        return;
    }
    free (tr->pos);
    free (tr->nodeidx);
    free (tr->statetype);
    free (tr);
}

/* Function: TraceSet()
 * Date:     SRE, Sun Mar  8 12:39:00 1998 [St. Louis]
 *
 * Purpose:  Convenience function; set values at position tpos
 *           in a trace.
 *
 *
 * Args:     tr   - trace object to write to
 *           tpos - ptr to position in trace to set
 *           type - statetype e.g. STS, etc.
 *           idx  - nodeidx 1..M or 0
 *           pos  - seq position 1..L or 0
 *
 * Returns:  void
 */
void
TraceSet (struct p7trace_s *tr, int tpos, char type, int idx, int pos)
{
    tr->statetype[tpos] = type;
    tr->nodeidx[tpos] = idx;
    tr->pos[tpos] = pos;
}


/* Function: MergeTraceArrays()
 * Date:     SRE, Sun Jul  5 15:09:10 1998 [St. Louis]
 *
 * Purpose:  Combine two arrays of traces into a single array.
 *           Used in hmmalign to merge traces from a fixed alignment
 *           with traces from individual unaligned seqs.
 *
 *           t1 traces always precede t2 traces in the resulting array.
 *
 * Args:     t1 - first set of traces
 *           n1 - number of traces in t1
 *           t2 - second set of traces
 *           n2 - number of traces in t2
 *
 * Returns:  pointer to new array of traces.
 *           Both t1 and t2 are free'd here! Do not reuse.
 */
struct p7trace_s **
MergeTraceArrays (struct p7trace_s **t1, int n1, struct p7trace_s **t2,
                  int n2)
{
    struct p7trace_s **tr;
    int i;			/* index in traces */

    tr = MallocOrDie (sizeof (struct p7trace_s *) * (n1 + n2));
    for (i = 0; i < n1; i++) {
        tr[i] = t1[i];
    }
    for (i = 0; i < n2; i++) {
        tr[n1 + i] = t2[i];
    }
    free (t1);
    free (t2);
    return tr;
}



/* Function: P7ReverseTrace()
 * Date:     SRE, Mon Aug 25 12:57:29 1997; Denver CO.
 *
 * Purpose:  Reverse the arrays in a traceback structure.
 *           Tracebacks from Forward() and Viterbi() are
 *           collected backwards, and call this function
 *           when they're done.
 *
 *           It's possible to reverse the arrays in place
 *           more efficiently; but the realloc/copy strategy
 *           has the advantage of reallocating the trace
 *           into the right size of memory. (Tracebacks
 *           overallocate.)
 *
 * Args:     tr - the traceback to reverse. tr->tlen must be set.
 *
 * Return:   (void)
 *           tr is modified.
 */
void
P7ReverseTrace (struct p7trace_s *tr)
{
    char *statetype;
    int *nodeidx;
    int *pos;
    int opos, npos;

    /* Allocate
     */
    statetype = MallocOrDie (sizeof (char) * tr->tlen);
    nodeidx = MallocOrDie (sizeof (int) * tr->tlen);
    pos = MallocOrDie (sizeof (int) * tr->tlen);

    /* Reverse the trace.
     */
    for (opos = tr->tlen - 1, npos = 0; npos < tr->tlen; npos++, opos--) {
        statetype[npos] = tr->statetype[opos];
        nodeidx[npos] = tr->nodeidx[opos];
        pos[npos] = tr->pos[opos];
    }

    /* Swap old, new arrays.
     */
    free (tr->statetype);
    free (tr->nodeidx);
    free (tr->pos);
    tr->statetype = statetype;
    tr->nodeidx = nodeidx;
    tr->pos = pos;
}



/* Function: P7TraceCount()
 *
 * Purpose:  Count a traceback into a count-based HMM structure.
 *           (Usually as part of a model parameter re-estimation.)
 *
 * Args:     hmm   - counts-based HMM
 *           dsq   - digitized sequence that traceback aligns to the HMM (1..L)
 *           wt    - weight on the sequence
 *           tr    - alignment of seq to HMM
 *
 * Return:   (void)
 */
void
P7TraceCount (struct plan7_s *hmm, char *dsq, float wt, struct p7trace_s *tr)
{
    int tpos;			/* position in tr */
    int i;			/* symbol position in seq */

    for (tpos = 0; tpos < tr->tlen; tpos++) {
        i = tr->pos[tpos];

        /* Emission counts.
         * Don't bother counting null states N,J,C.
         */
        if (tr->statetype[tpos] == STM) {
            P7CountSymbol (hmm->mat[tr->nodeidx[tpos]], dsq[i], wt);
        }
        else if (tr->statetype[tpos] == STI) {
            P7CountSymbol (hmm->ins[tr->nodeidx[tpos]], dsq[i], wt);
        }

        /* State transition counts
         */
        switch (tr->statetype[tpos]) {
        case STS:
            break;			/* don't bother; P=1 */
        case STN:
            switch (tr->statetype[tpos + 1]) {
            case STB:
                hmm->xt[XTN][MOVE] += wt;
                break;
            case STN:
                hmm->xt[XTN][LOOP] += wt;
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STB:
            switch (tr->statetype[tpos + 1]) {
            case STM:
                hmm->begin[tr->nodeidx[tpos + 1]] += wt;
                break;
            case STD:
                hmm->tbd1 += wt;
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STM:
            switch (tr->statetype[tpos + 1]) {
            case STM:
                hmm->t[tr->nodeidx[tpos]][TMM] += wt;
                break;
            case STI:
                hmm->t[tr->nodeidx[tpos]][TMI] += wt;
                break;
            case STD:
                hmm->t[tr->nodeidx[tpos]][TMD] += wt;
                break;
            case STE:
                hmm->end[tr->nodeidx[tpos]] += wt;
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STI:
            switch (tr->statetype[tpos + 1]) {
            case STM:
                hmm->t[tr->nodeidx[tpos]][TIM] += wt;
                break;
            case STI:
                hmm->t[tr->nodeidx[tpos]][TII] += wt;
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STD:
            switch (tr->statetype[tpos + 1]) {
            case STM:
                hmm->t[tr->nodeidx[tpos]][TDM] += wt;
                break;
            case STD:
                hmm->t[tr->nodeidx[tpos]][TDD] += wt;
                break;
            case STE:		/* ignore; p(D->E) = 1.0 */
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STE:
            switch (tr->statetype[tpos + 1]) {
            case STC:
                hmm->xt[XTE][MOVE] += wt;
                break;
            case STJ:
                hmm->xt[XTE][LOOP] += wt;
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STJ:
            switch (tr->statetype[tpos + 1]) {
            case STB:
                hmm->xt[XTJ][MOVE] += wt;
                break;
            case STJ:
                hmm->xt[XTJ][LOOP] += wt;
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STC:
            switch (tr->statetype[tpos + 1]) {
            case STT:
                hmm->xt[XTC][MOVE] += wt;
                break;
            case STC:
                hmm->xt[XTC][LOOP] += wt;
                break;
            default:
                Die ("illegal state transition %s->%s in traceback",
                     Statetype (tr->statetype[tpos]),
                     Statetype (tr->statetype[tpos + 1]));
            }
            break;
        case STT:
            break;			/* T is the last. It makes no transitions. */
        default:
            Die ("illegal state %s in traceback", Statetype (tr->statetype[tpos]));
        }
    }
}


/* Function: P7TraceScore()
 *
 * Purpose:  Score a traceback and return the score in scaled bits.
 *
 * Args:     hmm   - HMM with valid log odds scores.
 *           dsq   - digitized sequence that traceback aligns to the HMM (1..L)
 *           tr    - alignment of seq to HMM
 *
 * Return:   (void)
 */
float
P7TraceScore (struct plan7_s *hmm, char *dsq, struct p7trace_s *tr)
{
    int score;			/* total score as a scaled integer */
    int tpos;			/* position in tr */
    int sym;			/* digitized symbol in dsq */

    /*  P7PrintTrace(stdout, tr, hmm, dsq); */
    score = 0;
    for (tpos = 0; tpos < tr->tlen - 1; tpos++) {
        sym = (int) dsq[tr->pos[tpos]];

        /* Emissions.
         * Don't bother counting null states N,J,C.
         */
        if (tr->statetype[tpos] == STM) {
            score += hmm->msc[sym][tr->nodeidx[tpos]];
        }
        else if (tr->statetype[tpos] == STI) {
            score += hmm->isc[sym][tr->nodeidx[tpos]];
        }

        /* State transitions.
         */
        score += TransitionScoreLookup (hmm,
                                        tr->statetype[tpos], tr->nodeidx[tpos],
                                        tr->statetype[tpos + 1],
                                        tr->nodeidx[tpos + 1]);
    }
    return Scorify (score);
}



/* Function: P7Traces2Alignment()
 *
 * Purpose:  Convert an array of traceback structures for a set
 *           of sequences into a new multiple alignment.
 *
 *           Insertions are put into lower case and
 *           are not aligned; instead, Nterm is right-justified,
 *           Cterm is left-justified, and internal insertions
 *           are split in half and the halves are justified in
 *           each direction (the objective being to increase
 *           the chances of getting insertions aligned well enough
 *           for them to become a match). SAM gap char conventions
 *           are used: - in match columns, . in insert columns
 *
 * NOTE:     Does not recognize J state.
 *
 *           Can handle traces with D->I and I->D transitions;
 *           though these are disallowed by Plan7, they might be
 *           generated by aligning an alignment to a model, in
 *           the ImposeMasterTrace() step. Thus, --withali might
 *           generate alignments that are inconsistent with Plan7,
 *           that would have to be trace_doctor()'ed.
 *           xref STL6/p.117.
 *
 * Args:     dsq        - digitized unaligned sequences
 *           sqinfo     - array of info about the sequences
 *           wgt        - weights on seqs
 *           nseq       - number of sequences
 *           mlen       - length of model (number of match states)
 *           tr         - array of tracebacks
 *           matchonly  - TRUE if we don't print insert-generated symbols at all
 * Return:   MSA structure; NULL on failure.
 *           Caller responsible for freeing msa with MSAFree(msa);
 */
MSA *
P7Traces2Alignment (char **dsq, SQINFO * sqinfo, float *wgt, int nseq,
                    int mlen, struct p7trace_s ** tr, int matchonly)
{
    MSA *msa;			/* RETURN: new alignment */
    int idx;			/* counter for sequences */
    int alen;			/* width of alignment */
    int *inserts;			/* array of max gaps between aligned columns */
    int *matmap;			/* matmap[k] = apos of match k [1..M] */
    int nins;			/* counter for inserts */
    int apos;			/* position in aligned sequence (0..alen-1) */
    int rpos;			/* position in raw digital sequence (1..L) */
    int tpos;			/* position counter in traceback */
    int statetype;		/* type of current state, e.g. STM */
    int k;			/* counter over states in model */

    /* Here's the problem. We want to align the match states in columns,
     * but some sequences have inserted symbols in them; we need some
     * sort of overall knowledge of where the inserts are and how long
     * they are in order to create the alignment.
     *
     * Here's our trick. inserts[] is a 0..hmm->M array; inserts[i] stores
     * the maximum number of times insert substate i was used. This
     * is the maximum number of gaps to insert between canonical
     * column i and i+1.  inserts[0] is the N-term tail; inserts[M] is
     * the C-term tail.
     *
     * Remember that N and C emit on transition, hence the check for an
     * N->N or C->C transition before bumping nins.
     */
    inserts = (int *) MallocOrDie (sizeof (int) * (mlen + 1));
    for (k = 0; k <= mlen; k++) {
        inserts[k] = 0;
    }
    for (idx = 0; idx < nseq; idx++) {
        nins = 0;
        for (tpos = 0; tpos < tr[idx]->tlen; tpos++) {
            switch (tr[idx]->statetype[tpos]) {
            case STI:
                nins++;
                break;
            case STN:
                if (tr[idx]->statetype[tpos - 1] == STN) {
                    nins++;
                }
                break;
            case STC:
                if (tr[idx]->statetype[tpos - 1] == STC) {
                    nins++;
                }
                break;
            case STM:
            case STD:		/* M,D: record max. reset ctr. */
                if (nins > inserts[tr[idx]->nodeidx[tpos] - 1]) {
                    inserts[tr[idx]->nodeidx[tpos] - 1] = nins;
                }
                nins = 0;
                break;
            case STB:		/* B; record N-tail max, reset ctr */
                if (nins > inserts[0]) {
                    inserts[0] = nins;
                }
                nins = 0;
                break;
            case STT:		/* T: record C-tail max */
                if (nins > inserts[mlen]) {
                    inserts[mlen] = nins;
                }
                break;
            case STS:
            case STE:
                break;			/* ignore other states */
            case STJ:
                Die ("yo! you don't support J in Traces2Alignment(), remember?");
            default:
                Die ("Traces2Alignment reports unrecognized statetype %c",
                     Statetype (tr[idx]->statetype[tpos]));
            }
        }
    }

    /* Insert compression option. */
    if (matchonly)
        for (k = 0; k <= mlen; k++) {
            if (inserts[k] > 1) {
                inserts[k] = 1;
            }
				}
    /***********************************************
     * Construct the alignment
     ***********************************************/
    /* calculate alignment length and matmap */
    matmap = (int *) MallocOrDie (sizeof (int) * (mlen + 1));
    matmap[0] = -1;
    alen = inserts[0];
    for (k = 1; k <= mlen; k++) {
        matmap[k] = alen;
        alen += inserts[k] + 1;
    }
    /* allocation for new alignment */
    msa = MSAAlloc (nseq, alen);

    for (idx = 0; idx < nseq; idx++) {
        /* blank an aseq */
        for (apos = 0; apos < alen; apos++) {
            msa->aseq[idx][apos] = '.';
        }
        for (k = 1; k <= mlen; k++) {
            msa->aseq[idx][matmap[k]] = '-';
        }
        msa->aseq[idx][alen] = '\0';
        /* align the sequence */
        apos = 0;
        for (tpos = 0; tpos < tr[idx]->tlen; tpos++) {
            statetype = tr[idx]->statetype[tpos];	/* just for clarity */
            rpos = tr[idx]->pos[tpos];
            k = tr[idx]->nodeidx[tpos];

            if (statetype == STM) {
                apos = matmap[k];
                msa->aseq[idx][apos] = Alphabet[(int) dsq[idx][rpos]];
                apos++;
            }
            else if (statetype == STD) {
                apos = matmap[k] + 1;	/* need for handling D->I; xref STL6/p.117 */
            }
            else if (statetype == STI) {
                if (matchonly) {
                    msa->aseq[idx][apos] = '*';    /* insert compression option */
                }
                else {
                    msa->aseq[idx][apos] =
                        (char) tolower ((int) Alphabet[(int) dsq[idx][rpos]]);
                    apos++;
                }
            }
            else if ((statetype == STN || statetype == STC) && rpos > 0) {
                if (matchonly) {
                    msa->aseq[idx][apos] = '*';    /* insert compression option */
                }
                else {
                    msa->aseq[idx][apos] =
                        (char) tolower ((int) Alphabet[(int) dsq[idx][rpos]]);
                    apos++;
                }
            }
            else if (statetype == STE) {
                apos = matmap[mlen] + 1;    /* set position for C-term tail */
            }
        }

        /* N-terminal extension is right-justified.
         * Internal inserts are split in half, and C-term is right-justified.
         * C-terminal extension remains left-justified.
         */
        if (!matchonly) {
            rightjustify (msa->aseq[idx], inserts[0]);

            for (k = 1; k < mlen; k++) {
                if (inserts[k] > 1) {
                    for (nins = 0, apos = matmap[k] + 1; islower ((int) (msa->aseq[idx][apos])); apos++) {
                        nins++;
                    }
                    nins /= 2;		/* split the insertion in half */
                    rightjustify (msa->aseq[idx] + matmap[k] + 1 + nins,
                                  inserts[k] - nins);
                }
						}		
        }

    }

    /***********************************************
     * Build the rest of the MSA annotation.
     ***********************************************/

    msa->nseq = nseq;
    msa->alen = alen;
    msa->au = MallocOrDie (sizeof (char) * (strlen (PACKAGE_VERSION) + 7));
    sprintf (msa->au, "HMMER %s", PACKAGE_VERSION);
    /* copy sqinfo array and weights */
    for (idx = 0; idx < nseq; idx++) {
        msa->sqname[idx] = sre_strdup (sqinfo[idx].name, -1);
        if (sqinfo[idx].flags & SQINFO_ACC) {
            MSASetSeqAccession (msa, idx, sqinfo[idx].acc);
        }
        if (sqinfo[idx].flags & SQINFO_DESC) {
            MSASetSeqDescription (msa, idx, sqinfo[idx].desc);
        }

        if (sqinfo[idx].flags & SQINFO_SS) {
            if (msa->ss == NULL) {
                msa->ss = MallocOrDie (sizeof (char *) * nseq);
            }
            MakeAlignedString (msa->aseq[idx], alen,
                               sqinfo[idx].ss, &(msa->ss[idx]));
        }
        if (sqinfo[idx].flags & SQINFO_SA) {
            if (msa->sa == NULL) {
                msa->sa = MallocOrDie (sizeof (char *) * nseq);
            }
            MakeAlignedString (msa->aseq[idx], alen,
                               sqinfo[idx].sa, &(msa->sa[idx]));
        }
        msa->wgt[idx] = wgt[idx];
    }

    /* #=RF annotation: x for match column, . for insert column
     */
    msa->rf = (char *) MallocOrDie (sizeof (char) * (alen + 1));
    for (apos = 0; apos < alen; apos++) {
        msa->rf[apos] = '.';
    }
    for (k = 1; k <= mlen; k++) {
        msa->rf[matmap[k]] = 'x';
    }
    msa->rf[alen] = '\0';

    /* Currently, we produce no consensus structure.
     * #=CS, generated from HMM structural annotation, would go here.
     */

    free (inserts);
    free (matmap);
    return msa;
}

/* Function: TransitionScoreLookup()
 *
 * Purpose:  Convenience function used in PrintTrace() and TraceScore();
 *           given state types and node indices for a transition,
 *           return the integer score for that transition.
 */
int
TransitionScoreLookup (struct plan7_s *hmm, char st1, int k1,
                       char st2, int k2)
{
    switch (st1) {
    case STS:
        return 0;			/* S never pays */
    case STN:
        switch (st2) {
        case STB:
            return hmm->xsc[XTN][MOVE];
        case STN:
            return hmm->xsc[XTN][LOOP];
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STB:
        switch (st2) {
        case STM:
            return hmm->bsc[k2];
        case STD:
            return Prob2Score (hmm->tbd1, 1.);
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STM:
        switch (st2) {
        case STM:
            return hmm->tsc[TMM][k1];
        case STI:
            return hmm->tsc[TMI][k1];
        case STD:
            return hmm->tsc[TMD][k1];
        case STE:
            return hmm->esc[k1];
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STI:
        switch (st2) {
        case STM:
            return hmm->tsc[TIM][k1];
        case STI:
            return hmm->tsc[TII][k1];
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STD:
        switch (st2) {
        case STM:
            return hmm->tsc[TDM][k1];
        case STD:
            return hmm->tsc[TDD][k1];
        case STE:
            return 0;			/* D_m->E has probability 1.0 by definition in Plan7 */
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STE:
        switch (st2) {
        case STC:
            return hmm->xsc[XTE][MOVE];
        case STJ:
            return hmm->xsc[XTE][LOOP];
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STJ:
        switch (st2) {
        case STB:
            return hmm->xsc[XTJ][MOVE];
        case STJ:
            return hmm->xsc[XTJ][LOOP];
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STC:
        switch (st2) {
        case STT:
            return hmm->xsc[XTC][MOVE];
        case STC:
            return hmm->xsc[XTC][LOOP];
        default:
            Die ("illegal %s->%s transition", Statetype (st1), Statetype (st2));
        }
        break;
    case STT:
        return 0;			/* T makes no transitions */
    default:
        Die ("illegal state %s in traceback", Statetype (st1));
    }
    /*NOTREACHED*/ return 0;
}


/* Function: CreateFancyAli()
 * Date:     SRE, Mon Oct 27 06:49:44 1997 [Sanger Centre UK]
 *
 * Purpose:  Output of an HMM/sequence alignment, using a
 *           traceback structure. Deliberately similar to
 *           the output of BLAST, to make it easier for
 *           people to adapt their Perl parsers (or what have
 *           you) from BLAST to HMMER.
 *
 * Args:     tr  - traceback structure that gives the alignment
 *           hmm - the model
 *           dsq - the sequence (digitized form)
 *           name- name of the sequence
 *
 * Return:   allocated, filled fancy alignment structure.
 */
struct fancyali_s *
CreateFancyAli (struct p7trace_s *tr, struct plan7_s *hmm,
                char *dsq, char *name)
{
    struct fancyali_s *ali;	/* alignment to create                */
    int tpos;			/* position in trace and alignment    */
    int bestsym;			/* index of best symbol at this pos   */
    float mthresh;		/* above this P(x), display uppercase */

    /* Allocate and initialize the five lines of display
     */
    ali = AllocFancyAli ();
    ali->rfline = NULL;
    ali->csline = NULL;
    ali->model = MallocOrDie (sizeof (char) * (tr->tlen + 1));
    ali->mline = MallocOrDie (sizeof (char) * (tr->tlen + 1));
    ali->aseq = MallocOrDie (sizeof (char) * (tr->tlen + 1));

    memset (ali->model, ' ', tr->tlen);
    memset (ali->mline, ' ', tr->tlen);
    memset (ali->aseq, ' ', tr->tlen);

    if (hmm->flags & PLAN7_RF) {
        ali->rfline = (char *) MallocOrDie (sizeof (char) * (tr->tlen + 1));
        memset (ali->rfline, ' ', tr->tlen);
    }
    if (hmm->flags & PLAN7_CS) {
        ali->csline = (char *) MallocOrDie (sizeof (char) * (tr->tlen + 1));
        memset (ali->csline, ' ', tr->tlen);
    }

    ali->query = Strdup (hmm->name);
    ali->target = Strdup (name);

    if (Alphabet_type == hmmAMINO) {
        mthresh = 0.5;
    }
    else {
        mthresh = 0.9;
    }

    /* Find first, last seq position
     * HMM start/end positions currently not recorded, because there
     * might be multiple HMM hits per sequence.
     */
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        if (tr->pos[tpos] > 0) {
            ali->sqfrom = tr->pos[tpos];
            break;
        }
		}		
    for (tpos = tr->tlen - 1; tpos >= 0; tpos--) {
        if (tr->pos[tpos] > 0) {
            ali->sqto = tr->pos[tpos];
            break;
        }
		}
    /* Fill in the five lines of display
     */
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        switch (tr->statetype[tpos]) {
        case STS:
        case STT:
            ali->model[tpos] = '*';
            break;

        case STN:
        case STJ:
        case STC:
            ali->model[tpos] = '-';
            if (tr->pos[tpos] > 0) {
                ali->aseq[tpos] = tolower (Alphabet[(int) dsq[tr->pos[tpos]]]);
            }
            break;

        case STB:
            ali->model[tpos] = '>';
            break;

        case STE:
            ali->model[tpos] = '<';
            break;

        case STM:
            if (hmm->flags & PLAN7_RF) {
                ali->rfline[tpos] = hmm->rf[tr->nodeidx[tpos]];
            }
            if (hmm->flags & PLAN7_CS) {
                ali->csline[tpos] = hmm->cs[tr->nodeidx[tpos]];
            }
            bestsym = FArgMax (hmm->mat[tr->nodeidx[tpos]], Alphabet_size);
            ali->model[tpos] = Alphabet[bestsym];
            if (hmm->mat[tr->nodeidx[tpos]][bestsym] < mthresh) {
                ali->model[tpos] = tolower (ali->model[tpos]);
            }
            if (dsq[tr->pos[tpos]] == bestsym) {
                ali->mline[tpos] = Alphabet[(int) dsq[tr->pos[tpos]]];
                if (hmm->mat[tr->nodeidx[tpos]][bestsym] < mthresh) {
                    ali->mline[tpos] = tolower (ali->mline[tpos]);
                }
            }
            else if (hmm->msc[(int) dsq[tr->pos[tpos]]][tr->nodeidx[tpos]] > 0) {
                ali->mline[tpos] = '+';
            }
            ali->aseq[tpos] = Alphabet[(int) dsq[tr->pos[tpos]]];
            break;

        case STD:
            if (hmm->flags & PLAN7_RF) {
                ali->rfline[tpos] = hmm->rf[tr->nodeidx[tpos]];
            }
            if (hmm->flags & PLAN7_CS) {
                ali->csline[tpos] = hmm->cs[tr->nodeidx[tpos]];
            }
            bestsym = FArgMax (hmm->mat[tr->nodeidx[tpos]], Alphabet_size);
            ali->model[tpos] = Alphabet[bestsym];
            if (hmm->mat[tr->nodeidx[tpos]][bestsym] < mthresh) {
                ali->model[tpos] = tolower (ali->model[tpos]);
            }
            ali->aseq[tpos] = '-';
            break;

        case STI:
            ali->model[tpos] = '.';
            if (hmm->isc[(int) dsq[tr->pos[tpos]]][tr->nodeidx[tpos]] > 0) {
                ali->mline[tpos] = '+';
            }
            ali->aseq[tpos] =
                (char) tolower ((int) Alphabet[(int) dsq[tr->pos[tpos]]]);
            break;

        default:
            Die ("bogus statetype");
        }				/* end switch over statetypes */
    }				/* end loop over tpos */

    ali->len = tpos;
    if (hmm->flags & PLAN7_RF) {
        ali->rfline[tpos] = '\0';
    }
    if (hmm->flags & PLAN7_CS) {
        ali->csline[tpos] = '\0';
    }
    ali->model[tpos] = '\0';
    ali->mline[tpos] = '\0';
    ali->aseq[tpos] = '\0';
    return ali;
}


/* Function: PrintFancyAli()
 * Date:     SRE, Mon Oct 27 06:56:42 1997 [Sanger Centre UK]
 *
 * Purpose:  Print an HMM/sequence alignment from a fancyali_s
 *           structure. Line length controlled by ALILENGTH in
 *           config.h (set to 50).
 *
 * Args:     fp  - where to print it (stdout or open FILE)
 *           ali - alignment to print
 *
 * Return:   (void)
 */
void
PrintFancyAli (FILE * fp, struct fancyali_s *ali)
{
    char buffer[ALILENGTH + 1];	/* output line buffer                 */
    int starti, endi;
    int pos;
    int i;

    buffer[ALILENGTH] = '\0';
    endi = ali->sqfrom - 1;
    for (pos = 0; pos < ali->len; pos += ALILENGTH) {
        /* coords of target seq for this line */
        starti = endi + 1;
        for (i = pos; ali->aseq[i] != '\0' && i < pos + ALILENGTH; i++) {
            if (!isgap (ali->aseq[i])) {
                endi++;
            }
				}
        if (ali->csline != NULL) {
            strncpy (buffer, ali->csline + pos, ALILENGTH);
            fprintf (fp, "  %16s %s\n", "CS", buffer);
        }
        if (ali->rfline != NULL) {
            strncpy (buffer, ali->rfline + pos, ALILENGTH);
            fprintf (fp, "  %16s %s\n", "RF", buffer);
        }
        if (ali->model != NULL) {
            strncpy (buffer, ali->model + pos, ALILENGTH);
            fprintf (fp, "  %16s %s\n", " ", buffer);
        }
        if (ali->mline != NULL) {
            strncpy (buffer, ali->mline + pos, ALILENGTH);
            fprintf (fp, "  %16s %s\n", " ", buffer);
        }
        if (ali->aseq != NULL) {
            strncpy (buffer, ali->aseq + pos, ALILENGTH);
            if (endi >= starti)
                fprintf (fp, "  %10.10s %5d %s %-5d\n\n", ali->target, starti, buffer,
                         endi);
            else
                fprintf (fp, "  %10.10s %5s %s %-5s\n\n", ali->target, "-", buffer,
                         "-");
        }
    }

    /* Cleanup and return
     */
    fflush (fp);
    return;
}



/* Function: TraceDecompose()
 * Date:     Sat Aug 30 11:18:40 1997 (Denver CO)
 *
 * Purpose:  Decompose a long multi-hit trace into zero or more
 *           traces without N,C,J transitions: for consistent
 *           scoring and statistical evaluation of single domain
 *           hits.
 *
 * Args:     otr    - original trace structure
 *           ret_tr - RETURN: array of simpler traces
 *           ret_ntr- RETURN: number of traces.
 *
 * Return:   (void)
 *           ret_tr alloc'ed here; free individuals with FreeTrace().
 */
void
TraceDecompose (struct p7trace_s *otr, struct p7trace_s ***ret_tr,
                int *ret_ntr)
{
    struct p7trace_s **tr;	/* array of new traces          */
    int ntr;			/* number of traces             */
    int i, j;			/* position counters in traces  */
    int idx;			/* index over ntr subtraces     */

    /* First pass: count begin states to get ntr.
     */
    for (ntr = 0, i = 0; i < otr->tlen; i++) {
        if (otr->statetype[i] == STB) {
            ntr++;
        }
		}
    /* Allocations.
     */
    if (ntr == 0) {
        *ret_ntr = 0;
        *ret_tr = NULL;
        return;
    }
    tr = (struct p7trace_s **) MallocOrDie (sizeof (struct p7trace_s *) * ntr);

    for (idx = 0, i = 0; i < otr->tlen; i++) {
				/* i = position in old trace */
        if (otr->statetype[i] == STB) {
            for (j = i + 1; j < otr->tlen; j++) {
								/* j = tmp; get length of subtrace */
                if (otr->statetype[j] == STE) {
                    break;
                }
						}		
            /* trace = S-N-(B..E)-C-T : len + 4 : j-i+1 + 4 */
            P7AllocTrace (j - i + 5, &(tr[idx]));
            tr[idx]->tlen = j - i + 5;

            tr[idx]->statetype[0] = STS;
            tr[idx]->nodeidx[0] = 0;
            tr[idx]->pos[0] = 0;
            tr[idx]->statetype[1] = STN;
            tr[idx]->nodeidx[1] = 0;
            tr[idx]->pos[1] = 0;
            j = 2;			/* now j = position in new subtrace */
            while (1) {		/* copy subtrace */
                tr[idx]->statetype[j] = otr->statetype[i];
                tr[idx]->nodeidx[j] = otr->nodeidx[i];
                tr[idx]->pos[j] = otr->pos[i];
                if (otr->statetype[i] == STE) {
                    break;
                }
                i++;
                j++;
            }
            j++;
            tr[idx]->statetype[j] = STC;
            tr[idx]->nodeidx[j] = 0;
            tr[idx]->pos[j] = 0;
            j++;
            tr[idx]->statetype[j] = STT;
            tr[idx]->nodeidx[j] = 0;
            tr[idx]->pos[j] = 0;
            idx++;
        }
		}	

    *ret_tr = tr;
    *ret_ntr = ntr;
    return;
}


/* Function: TraceDomainNumber()
 *
 * Purpose:  Count how many times we traverse the
 *           model in a single Plan7 trace -- equivalent
 *           to counting the number of domains.
 *
 *           (A weakness is that we might discard some of
 *           those domains because they have low scores
 *           below E or T threshold.)
 */
int
TraceDomainNumber (struct p7trace_s *tr)
{
    int i;
    int ndom = 0;

    for (i = 0; i < tr->tlen; i++) {
        if (tr->statetype[i] == STB) {
            ndom++;
        }
		}		
    return ndom;
}


/* Function: TraceSimpleBounds()
 *
 * Purpose:  For a trace that contains only a single
 *           traverse of the model (i.e. something that's
 *           come from TraceDecompose(), or a global
 *           alignment), determine the bounds of
 *           the match on both the sequence [1..L] and the
 *           model [1..M].
 *
 * Args:     tr   - trace to look at
 *           i1   - RETURN: start point in sequence [1..L]
 *           i2   - RETURN: end point in sequence [1..L]
 *           k1   - RETURN: start point in model [1..M]
 *           k2   - RETURN: end point in model [1..M]
 */
void
TraceSimpleBounds (struct p7trace_s *tr, int *ret_i1, int *ret_i2,
                   int *ret_k1, int *ret_k2)
{
    int i1, i2, k1, k2, tpos;

    i1 = k1 = i2 = k2 = -1;

    /* Look forwards to find start of match */
    for (tpos = 0; tpos < tr->tlen; tpos++) {
        if (k1 == -1
                && (tr->statetype[tpos] == STM || tr->statetype[tpos] == STD)) {
            k1 = tr->nodeidx[tpos];
        }
        if (tr->statetype[tpos] == STM) {
            i1 = tr->pos[tpos];
            break;
        }
    }
    if (tpos == tr->tlen || i1 == -1 || k1 == -1) {
        Die ("sanity check failed: didn't find a match state in trace");
    }

    /* Look backwards to find end of match */
    for (tpos = tr->tlen - 1; tpos >= 0; tpos--) {
        if (k2 == -1
                && (tr->statetype[tpos] == STM || tr->statetype[tpos] == STD)) {
            k2 = tr->nodeidx[tpos];
        }
        if (tr->statetype[tpos] == STM) {
            i2 = tr->pos[tpos];
            break;
        }
    }
    if (tpos == tr->tlen || i2 == -1 || k2 == -1) {
        Die ("sanity check failed: didn't find a match state in trace");
    }

    *ret_k1 = k1;
    *ret_i1 = i1;
    *ret_k2 = k2;
    *ret_i2 = i2;
}


/* Function: MasterTraceFromMap()
 * Date:     SRE, Tue Jul  7 18:51:11 1998 [St. Louis]
 *
 * Purpose:  Convert an alignment map (e.g. hmm->map) to
 *           a master trace. Used for mapping an alignment
 *           onto an HMM. Generally precedes a call to
 *           ImposeMasterTrace(). Compare P7ViterbiAlignAlignment(),
 *           which aligns an alignment to the model using a
 *           Viterbi algorithm to get a master trace.
 *           MasterTraceFromMap() only works if the alignment
 *           is exactly the one used to train the model.
 *
 * Args:     map  - the map (usually hmm->map is passed) 1..M
 *           M    - length of map (model; usually hmm->M passed)
 *           alen - length of alignment that map refers to
 *
 * Returns:  ptr to master trace
 *           Caller must free: P7FreeTrace().
 */
struct p7trace_s *
MasterTraceFromMap (int *map, int M, int alen)
{
    struct p7trace_s *tr;		/* RETURN: master trace */
    int tpos;			/* position in trace */
    int apos;			/* position in alignment, 1..alen */
    int k;			/* position in model */

    /* Allocate for the trace.
     * S-N-B- ... - E-C-T  : 6 states + alen is maximum trace,
     * because each of alen columns is an N*, M*, I*, or C* metastate.
     * No D* metastates possible.
     */
    P7AllocTrace (alen + 6, &tr);

    /* Initialize the trace
     */
    tpos = 0;
    TraceSet (tr, tpos, STS, 0, 0);
    tpos++;
    TraceSet (tr, tpos, STN, 0, 0);
    tpos++;

    /* Leading N's
     */
    for (apos = 1; apos < map[1]; apos++) {
        TraceSet (tr, tpos, STN, 0, apos);
        tpos++;
    }				/* now apos == map[1] */
    TraceSet (tr, tpos, STB, 0, 0);
    tpos++;

    for (k = 1; k < M; k++) {
        TraceSet (tr, tpos, STM, k, apos);
        tpos++;
        apos++;

        for (; apos < map[k + 1]; apos++) {
            TraceSet (tr, tpos, STI, k, apos);
            tpos++;
        }
    }				/* now apos == map[M] and k == M */

    TraceSet (tr, tpos, STM, M, apos);
    tpos++;
    apos++;

    /* Trailing C's
     */
    TraceSet (tr, tpos, STE, 0, 0);
    tpos++;
    TraceSet (tr, tpos, STC, 0, 0);
    tpos++;
    for (; apos <= alen; apos++) {
        TraceSet (tr, tpos, STC, 0, apos);
        tpos++;
    }

    /* Terminate and return
     */
    TraceSet (tr, tpos, STT, 0, 0);
    tpos++;
    tr->tlen = tpos;
    return tr;
}



/* Function: ImposeMasterTrace()
 * Date:     SRE, Sun Jul  5 14:27:16 1998 [St. Louis]
 *
 * Purpose:  Goes with P7ViterbiAlignAlignment(), which gives us
 *           a "master trace" for a whole alignment. Now, given
 *           the alignment and the master trace, construct individual
 *           tracebacks for each sequence. Later we'll hand these
 *           (and presumably other traces) to P7Traces2Alignment().
 *
 *           It is possible to generate individual traces that
 *           are not consistent with Plan7 (e.g. D->I and I->D
 *           transitions may be present). P7Traces2Alignment()
 *           can handle such traces; other functions may not.
 *           See modelmaker.c:trace_doctor() if this is a problem.
 *
 *           Akin to modelmaker.c:fake_tracebacks().
 *
 * Args:     aseq  - aligned seqs
 *           nseq  - number of aligned seqs
 *           mtr   - master traceback
 *           ret_tr- RETURN: array of individual tracebacks, one for each aseq
 *
 * Returns:  (void)
 */
void
ImposeMasterTrace (char **aseq, int nseq, struct p7trace_s *mtr,
                   struct p7trace_s ***ret_tr)
{
    struct p7trace_s **tr;
    int idx;			/* counter over sequences */
    int i;			/* position in raw sequence (1..L) */
    int tpos;			/* position in traceback           */
    int mpos;			/* position in master trace        */

    tr = (struct p7trace_s **) MallocOrDie (sizeof (struct p7trace_s *) * nseq);

    for (idx = 0; idx < nseq; idx++) {
        P7AllocTrace (mtr->tlen, &tr[idx]);	/* we're guaranteed that individuals len < master len */

        tpos = 0;
        i = 1;
        for (mpos = 0; mpos < mtr->tlen; mpos++) {
            switch (mtr->statetype[mpos]) {
            case STS:		/* straight copies w/ no emission: S, B, D, E, T */
            case STB:
            case STD:
            case STE:
            case STT:
                TraceSet (tr[idx], tpos, mtr->statetype[mpos], mtr->nodeidx[mpos], 0);
                tpos++;
                break;

            case STM:		/* M* implies M or D */
                if (isgap (aseq[idx][mtr->pos[mpos] - 1])) {
                    TraceSet (tr[idx], tpos, STD, mtr->nodeidx[mpos], 0);
                }
                else {
                    TraceSet (tr[idx], tpos, STM, mtr->nodeidx[mpos], i);
                    i++;
                }
                tpos++;
                break;

            case STI:		/* I* implies I or nothing */
                if (!isgap (aseq[idx][mtr->pos[mpos] - 1])) {
                    TraceSet (tr[idx], tpos, STI, mtr->nodeidx[mpos], i);
                    i++;
                    tpos++;
                }
                break;

            case STJ:		/* N,J,C: first N* -> N. After that, N* -> N or nothing. */
            case STN:
            case STC:
                if (mtr->pos[mpos] == 0) {
                    TraceSet (tr[idx], tpos, mtr->statetype[mpos], 0, 0);
                    tpos++;
                }
                else if (!isgap (aseq[idx][mtr->pos[mpos] - 1])) {
                    TraceSet (tr[idx], tpos, mtr->statetype[mpos], 0, i);
                    i++;
                    tpos++;
                }
                break;

            case STBOGUS:
                Die ("never happens. Trust me.");
            }
        }
        tr[idx]->tlen = tpos;
    }
    *ret_tr = tr;
}


/* Function: rightjustify()
 *
 * Purpose:  Given a gap-containing string of length n,
 *           pull all the non-gap characters as far as
 *           possible to the right, leaving gaps on the
 *           left side. Used to rearrange the positions
 *           of insertions in HMMER alignments.
 */
static void
rightjustify (char *s, int n)
{
    int npos;
    int opos;

    npos = n - 1;
    opos = n - 1;
    while (opos >= 0) {
        if (isgap (s[opos])) {
            opos--;
        }
        else {
            s[npos--] = s[opos--];
        }
    }
    while (npos >= 0) {
        s[npos--] = '.';
    }
}


/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/*
 * translate.c - functions for translating nucleic acid sequence
 * created Tue Jan 12 11:27:29 1993, SRE
 *
 * RCS $Id: translate.c,v 1.3 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdio.h>
#include <string.h>
#include "squid.h"



/* Function: Translate(char *seq, char **code)
 *
 * Given a ptr to the start of a nucleic acid sequence,
 * and a genetic code, translate the sequence into
 * amino acid sequence.
 *
 * code is an array of 65 strings, representing
 * the translations of the 64 codons, arranged
 * in order AAA, AAC, AAG, AAU, ..., UUA, UUC, UUG, UUU.
 * '*' or '***' is used to represent termination
 * codons, usually. The final string, code[64],
 * is the code for an ambiguous amino acid.
 *
 * Because of the way space is allocated for the amino
 * acid sequence, the amino acid strings cannot be
 * longer than 3 letters each. (I don't foresee using
 * anything but the single- and triple- letter codes.)
 *
 * Returns a ptr to the translation string on success,
 * or NULL on failure.
 */
char *
Translate (char *seq, char **code)
{
    int codon;			/* index for codon         */
    char *aaseq;			/* RETURN: the translation */
    char *aaptr;			/* ptr into aaseq */
    int i;

    if (seq == NULL) {
        squid_errno = SQERR_NODATA;
        return NULL;
    }
    if ((aaseq = (char *) calloc (strlen (seq) + 1, sizeof (char))) == NULL) {
        Die ("calloc failed");
    }

    aaptr = aaseq;
    for (; *seq != '\0' && *(seq + 1) != '\0' && *(seq + 2) != '\0'; seq += 3) {
        /* calculate the lookup value for
           this codon */
        codon = 0;
        for (i = 0; i < 3; i++) {
            codon *= 4;
            switch (*(seq + i)) {
            case 'A':
            case 'a':
                break;
            case 'C':
            case 'c':
                codon += 1;
                break;
            case 'G':
            case 'g':
                codon += 2;
                break;
            case 'T':
            case 't':
                codon += 3;
                break;
            case 'U':
            case 'u':
                codon += 3;
                break;
            default:
                codon = 64;
                break;
            }
            if (codon == 64) {
                break;
            }
        }

        strcpy (aaptr, code[codon]);
        aaptr += strlen (code[codon]);
    }
    return aaseq;
}

/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* file: types.c
 *
 * Finicky type checkers for strings. Return 1 (TRUE) if ok, 0 elsewise.
 * Also, finicky type converters (sre_ntoh32() and friends)
 *
 * CVS $Id: types.c,v 1.6 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <string.h>
#include <ctype.h>
#include "squid.h"

/* Function: IsInt()
 *
 * Returns TRUE if s points to something that atoi() will parse
 * completely and convert to an integer.
 */
int
IsInt (char *s)
{
    int hex = 0;

    if (s == NULL) {
        squid_errno = SQERR_PARAMETER;
        return 0;
    }

    /* skip whitespace */
    while (isspace ((int) (*s))) {
        s++;
    }
    /* skip leading sign */
    if (*s == '-' || *s == '+') {
        s++;
    }
    /* skip leading conversion signals */
    if ((strncmp (s, "0x", 2) == 0 && (int) strlen (s) > 2) ||
            (strncmp (s, "0X", 2) == 0 && (int) strlen (s) > 2)) {
        s += 2;
        hex = 1;
    }
    else if (*s == '0' && (int) strlen (s) > 1) {
        s++;
    }
    /* examine remainder for garbage chars */
    if (!hex)
        while (*s != '\0') {
            if (!isdigit ((int) (*s))) {
                return 0;
            }
            s++;
        }
    else
        while (*s != '\0') {
            if (!isxdigit ((int) (*s))) {
                return 0;
            }
            s++;
        }

    return 1;
}


/* Function: IsReal()
 *
 * Purpose:  Returns TRUE if s is a string representation
 *           of a valid floating point number.
 */
int
IsReal (char *s)
{
    int gotdecimal = 0;
    int gotexp = 0;
    int gotreal = 0;

    if (s == NULL) {
        return 0;
    }

    while (isspace ((int) (*s))) {
        s++;    /* skip leading whitespace */
    }
    if (*s == '-' || *s == '+') {
        s++;    /* skip leading sign */
    }

    /* Examine remainder for garbage. Allowed one '.' and
     * one 'e' or 'E'; if both '.' and e/E occur, '.'
     * must be first.
     */
    while (*s != '\0') {
        if (isdigit ((int) (*s))) {
            gotreal++;
        }
        else if (*s == '.') {
            if (gotdecimal) {
                return 0;    /* can't have two */
            }
            if (gotexp) {
                return 0;    /* e/E preceded . */
            }
            else {
                gotdecimal++;
            }
        }
        else if (*s == 'e' || *s == 'E') {
            if (gotexp) {
                return 0;    /* can't have two */
            }
            else {
                gotexp++;
            }
        }
        else if (isspace ((int) (*s))) {
            break;
        }

        s++;
    }

    while (isspace ((int) (*s))) {
        s++;    /* skip trailing whitespace */
    }
    if (*s == '\0' && gotreal) {
        return 1;
    }
    else {
        return 0;
    }
}


/* Function: Byteswap()
 *
 * Purpose:  Swap between big-endian and little-endian.
 *           For example:
 *               int foo = 0x12345678;
 *               byteswap((char *) &foo, sizeof(int));
 *               printf("%x\n", foo)
 *           gives 78563412.
 *
 *           I don't fully understand byte-swapping issues.
 *           However, I have tested this on chars through floats,
 *           on various machines:
 *               SGI IRIX 4.0.5, SunOS 4.1.3, DEC Alpha OSF/1, Alliant
 *
 * Date: Sun Feb 12 10:26:22 1995
 */
void
Byteswap (char *swap, int nbytes)
{
    int x;
    char byte;

    for (x = 0; x < nbytes / 2; x++) {
        byte = swap[nbytes - x - 1];
        swap[nbytes - x - 1] = swap[x];
        swap[x] = byte;
    }
}



/* Functions: sre_ntoh16(), etc.
 * Date:      SRE, Sun Dec 31 11:26:53 2000 [St. Louis]
 *
 * Purpose:   Provide functionality of ntohs(), etc; extended
 *            to 64-bit unsigned ints, and explicitly provided
 *            in case a machine doesn't have the ntohs()
 *            family.
 *
 *            If we're using the host functions,
 *            USE_HOST_BYTESWAP_FUNCTIONS was set to 1 in
 *            squidconf.h, and we #define'd sre_hton16(x)=hton(x), etc.
 *            in squid.h. In doing this, we assumed that the
 *            host functions work on 16- and 32-bit unsigned quantities.
 *            If for some reason that's not true, set
 *            USE_HOST_BYTESWAP_FUNCTIONS to 0.
 */
#ifndef USE_HOST_BYTESWAP_FUNCTIONS
sqd_uint16
sre_ntoh16 (sqd_uint16 netshort)
{
#ifdef WORDS_BIGENDIAN
    return netshort;
#else
    Byteswap ((char *) &netshort, 2);
    return netshort;
#endif
}

sqd_uint32
sre_ntoh32 (sqd_uint32 netlong)
{
#ifdef WORDS_BIGENDIAN
    return netlong;
#else
    Byteswap ((char *) &netlong, 4);
    return netlong;
#endif
}

sqd_uint16
sre_hton16 (sqd_uint16 hostshort)
{
#ifdef WORDS_BIGENDIAN
    return hostshort;
#else
    Byteswap ((char *) &hostshort, 2);
    return hostshort;
#endif
}

sqd_uint32
sre_hton32 (sqd_uint32 hostlong)
{
#ifdef WORDS_BIGENDIAN
    return hostlong;
#else
    Byteswap ((char *) &hostlong, 4);
    return hostlong;
#endif
}
#endif /*USE_HOST_BYTESWAP_FUNCTIONS */

sqd_uint64
sre_ntoh64 (sqd_uint64 net_int64)
{
#ifdef WORDS_BIGENDIAN
    return net_int64;
#else
    Byteswap ((char *) &net_int64, 8);
    return net_int64;
#endif
}

sqd_uint64
sre_hton64 (sqd_uint64 host_int64)
{
#ifdef WORDS_BIGENDIAN
    return host_int64;
#else
    Byteswap ((char *) &host_int64, 8);
    return host_int64;
#endif
}




/* vectorops.c
 * Operations on vectors of floats or doubles.
 *
 * DSet(), FSet()       - set all items in vector to value.
 * DScale(), FScale()   - multiply all items in vector by scale
 * DSum(), FSum()       - return sum of values in vector
 * DAdd(), FAdd()       - add vec2 to vec1.
 * DCopy(), FCopy()     - set vec1 to be same as vec2.
 * DDot(), FDot()       - return dot product of two vectors.
 * DMax(), FMax()       - return value of maximum element in vector
 * DMin(), FMin()       - return value of minimum element in vector
 * DArgMax(), FArgMax() - return index of maximum element in vector
 * DArgMin(), FArgMin() - return index of minimum element in vector
 *
 * DNorm(), FNorm()     - normalize a probability vector of length n.
 * DLog(), FLog()       - convert to log probabilities
 * DExp(), FExp()       - convert log p's back to probabilities
 * DLogSum(), FLogSum() - given vector of log p's; return log of summed p's.
 *
 * SRE, Tue Oct  1 15:23:25 2002 [St. Louis]
 * CVS $Id: vectorops.c,v 1.4 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <stdlib.h>
#include <math.h>
#include <float.h>
#include "vectorops.h"

void
DSet (double *vec, int n, double value)
{
    int x;
    for (x = 0; x < n; x++) {
        vec[x] = value;
    }
}

void
FSet (float *vec, int n, float value)
{
    int x;
    for (x = 0; x < n; x++) {
        vec[x] = value;
    }
}

void
DScale (double *vec, int n, double scale)
{
    int x;
    for (x = 0; x < n; x++) {
        vec[x] *= scale;
    }
}

void
FScale (float *vec, int n, float scale)
{
    int x;
    for (x = 0; x < n; x++) {
        vec[x] *= scale;
    }
}

double
DSum (double *vec, int n)
{
    double sum = 0.;
    int x;
    for (x = 0; x < n; x++) {
        sum += vec[x];
    }
    return sum;
}

float
FSum (float *vec, int n)
{
    float sum = 0.;
    int x;
    for (x = 0; x < n; x++) {
        sum += vec[x];
    }
    return sum;
}

void
DAdd (double *vec1, double *vec2, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        vec1[x] += vec2[x];
    }
}

void
FAdd (float *vec1, float *vec2, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        vec1[x] += vec2[x];
    }
}

void
DCopy (double *vec1, double *vec2, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        vec1[x] = vec2[x];
    }
}

void
FCopy (float *vec1, float *vec2, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        vec1[x] = vec2[x];
    }
}

double
DDot (double *vec1, double *vec2, int n)
{
    double result = 0.;
    int x;
    for (x = 0; x < n; x++) {
        result += vec1[x] * vec2[x];
    }
    return result;
}

float
FDot (float *vec1, float *vec2, int n)
{
    float result = 0.;
    int x;
    for (x = 0; x < n; x++) {
        result += vec1[x] * vec2[x];
    }
    return result;
}

double
DMax (double *vec, int n)
{
    int i;
    double best;

    best = vec[0];
    for (i = 1; i < n; i++) {
        if (vec[i] > best) {
            best = vec[i];
        }
		}		
    return best;
}

float
FMax (float *vec, int n)
{
    int i;
    float best;

    best = vec[0];
    for (i = 1; i < n; i++) {
        if (vec[i] > best) {
            best = vec[i];
        }
		}		
    return best;
}

double
DMin (double *vec, int n)
{
    int i;
    double best;

    best = vec[0];
    for (i = 1; i < n; i++) {
        if (vec[i] < best) {
            best = vec[i];
        }
		}		
    return best;
}

float
FMin (float *vec, int n)
{
    int i;
    float best;

    best = vec[0];
    for (i = 1; i < n; i++) {
        if (vec[i] < best) {
            best = vec[i];
        }
		}		
    return best;
}

int
DArgMax (double *vec, int n)
{
    int i;
    int best = 0;

    for (i = 1; i < n; i++) {
        if (vec[i] > vec[best]) {
            best = i;
        }
		}		
    return best;
}

int
FArgMax (float *vec, int n)
{
    int i;
    int best = 0;

    for (i = 1; i < n; i++) {
        if (vec[i] > vec[best]) {
            best = i;
        }
		}		
    return best;
}

int
DArgMin (double *vec, int n)
{
    int i;
    int best = 0;
    for (i = 1; i < n; i++) {
        if (vec[i] < vec[best]) {
            best = i;
        }
		}		
    return best;
}

int
FArgMin (float *vec, int n)
{
    int i;
    int best = 0;

    for (i = 1; i < n; i++) {
        if (vec[i] < vec[best]) {
            best = i;
        }
		}		
    return best;
}

void
DNorm (double *vec, int n)
{
    int x;
    double sum;

    sum = DSum (vec, n);
    if (sum != 0.0)
        for (x = 0; x < n; x++) {
            vec[x] /= sum;
        }
    else
        for (x = 0; x < n; x++) {
            vec[x] = 1. / (double) n;
        }
}

void
FNorm (float *vec, int n)
{
    int x;
    float sum;

    sum = FSum (vec, n);
    if (sum != 0.0)
        for (x = 0; x < n; x++) {
            vec[x] /= sum;
        }
    else
        for (x = 0; x < n; x++) {
            vec[x] = 1. / (float) n;
        }
}

void
DLog (double *vec, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        if (vec[x] > 0.) {
            vec[x] = log (vec[x]);
        }
        else {
            vec[x] = -DBL_MAX;
        }
		}		
}

void
FLog (float *vec, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        if (vec[x] > 0.) {
            vec[x] = log (vec[x]);
        }
        else {
            vec[x] = -FLT_MAX;
        }
		}		
}

void
DExp (double *vec, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        vec[x] = exp (vec[x]);
    }
}

void
FExp (float *vec, int n)
{
    int x;
    for (x = 0; x < n; x++) {
        vec[x] = exp (vec[x]);
    }
}

double
DLogSum (double *vec, int n)
{
    int x;
    double max, sum;

    max = DMax (vec, n);
    sum = 0.0;
    for (x = 0; x < n; x++) {
        if (vec[x] > max - 50.) {
            sum += exp (vec[x] - max);
        }
		}		
    sum = log (sum) + max;
    return sum;
}

float
FLogSum (float *vec, int n)
{
    int x;
    float max, sum;

    max = FMax (vec, n);
    sum = 0.0;
    for (x = 0; x < n; x++) {
        if (vec[x] > max - 50.) {
            sum += exp (vec[x] - max);
        }
		}		
    sum = log (sum) + max;
    return sum;
}



/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* weight.c
 * SRE, Thu Mar  3 07:56:01 1994
 *
 * Calculate weights for sequences in an alignment.
 * RCS $Id: weight.c,v 1.11 2003/04/14 16:00:16 eddy Exp $
 */

#include "squidconf.h"

#include <ctype.h>
#include <string.h>
#include "squid.h"
#include "sre_random.h"

static void upweight (struct phylo_s *tree, int nseq, float *lwt, float *rwt,
                      int node);
static void downweight (struct phylo_s *tree, int nseq, float *lwt,
                        float *rwt, float *fwt, int node);
static float simple_distance (char *s1, char *s2);
static int simple_diffmx (char **aseqs, int num, float ***ret_dmx);

/* Function: GSCWeights()
 *
 * Purpose:  Use Erik's tree-based algorithm to set weights for
 *           sequences in an alignment. upweight() and downweight()
 *           are derived from Graeme Mitchison's code.
 *
 * Args:     aseq        - array of (0..nseq-1) aligned sequences
 *           nseq        - number of seqs in alignment
 *           alen        - length of alignment
 *           wgt         - allocated [0..nseq-1] array of weights to be returned
 *
 * Return:   (void)
 *           wgt is filled in.
 */
void
GSCWeights (char **aseq, int nseq, int alen, float *wgt)
{
    float **dmx;			/* distance (difference) matrix */
    struct phylo_s *tree;
    float *lwt, *rwt;		/* weight on left, right of this tree node */
    float *fwt;			/* final weight assigned to this node */
    int i;

    /* Sanity check first
     */
    if (nseq == 1) {
        wgt[0] = 1.0;
        return;
    }

    /* I use a simple fractional difference matrix derived by
     * pairwise identity. Perhaps I should include a Poisson
     * distance correction.
     */
    MakeDiffMx (aseq, nseq, &dmx);
    if (!Cluster (dmx, nseq, CLUSTER_MIN, &tree)) {
        Die ("Cluster() failed");
    }

    /* Allocations
     */
    lwt = MallocOrDie (sizeof (float) * (2 * nseq - 1));
    rwt = MallocOrDie (sizeof (float) * (2 * nseq - 1));
    fwt = MallocOrDie (sizeof (float) * (2 * nseq - 1));

    /* lwt and rwt are the total branch weight to the left and
     * right of a node or sequence. They are 0..2N-2.  0..N-1 are
     * the sequences; these have weight 0. N..2N-2 are the actual
     * tree nodes.
     */
    for (i = 0; i < nseq; i++) {
        lwt[i] = rwt[i] = 0.0;
    }
    /* recursively calculate rwt, lwt, starting
       at node nseq (the root) */
    upweight (tree, nseq, lwt, rwt, nseq);

    /* recursively distribute weight across the
       tree */
    fwt[nseq] = nseq;
    downweight (tree, nseq, lwt, rwt, fwt, nseq);
    /* collect the weights */
    for (i = 0; i < nseq; i++) {
        wgt[i] = fwt[i];
    }

    FMX2Free (dmx);
    FreePhylo (tree, nseq);
    free (lwt);
    free (rwt);
    free (fwt);
}

static void
upweight (struct phylo_s *tree, int nseq, float *lwt, float *rwt, int node)
{
    int ld, rd;

    ld = tree[node - nseq].left;
    if (ld >= nseq) {
        upweight (tree, nseq, lwt, rwt, ld);
    }
    rd = tree[node - nseq].right;
    if (rd >= nseq) {
        upweight (tree, nseq, lwt, rwt, rd);
    }
    lwt[node] = lwt[ld] + rwt[ld] + tree[node - nseq].lblen;
    rwt[node] = lwt[rd] + rwt[rd] + tree[node - nseq].rblen;
}


static void
downweight (struct phylo_s *tree, int nseq, float *lwt, float *rwt,
            float *fwt, int node)
{
    int ld, rd;
    float lnum, rnum;

    ld = tree[node - nseq].left;
    rd = tree[node - nseq].right;
    if (lwt[node] + rwt[node] > 0.0) {
        fwt[ld] = fwt[node] * (lwt[node] / (lwt[node] + rwt[node]));
        fwt[rd] = fwt[node] * (rwt[node] / (lwt[node] + rwt[node]));
    }
    else {
        lnum = (ld >= nseq) ? tree[ld - nseq].incnum : 1.0;
        rnum = (rd >= nseq) ? tree[rd - nseq].incnum : 1.0;
        fwt[ld] = fwt[node] * lnum / (lnum + rnum);
        fwt[rd] = fwt[node] * rnum / (lnum + rnum);
    }

    if (ld >= nseq) {
        downweight (tree, nseq, lwt, rwt, fwt, ld);
    }
    if (rd >= nseq) {
        downweight (tree, nseq, lwt, rwt, fwt, rd);
    }
}




/* Function: VoronoiWeights()
 *
 * Purpose:  Calculate weights using the scheme of Sibbald &
 *           Argos (JMB 216:813-818 1990). The scheme is
 *           slightly modified because the original algorithm
 *           actually doesn't work on gapped alignments.
 *           The sequences are assumed to be protein.
 *
 * Args:     aseq  - array of (0..nseq-1) aligned sequences
 *           nseq  - number of sequences
 *           alen  - length of alignment
 *           wgt   - allocated [0..nseq-1] array of weights to be returned
 *
 * Return:   void
 *           wgt is filled in.
 */
void
VoronoiWeights (char **aseq, int nseq, int alen, float *wgt)
{
    float **dmx;			/* distance (difference) matrix    */
    float *halfmin;		/* 1/2 minimum distance to other seqs */
    char **psym;			/* symbols seen in each column     */
    int *nsym;			/* # syms seen in each column      */
    int symseen[27];		/* flags for observed syms         */
    char *randseq;		/* randomly generated sequence     */
    int acol;			/* pos in aligned columns          */
    int idx;			/* index in sequences              */
    int symidx;			/* 0..25 index for symbol          */
    int i;			/* generic counter                 */
    float min;			/* minimum distance                */
    float dist;			/* distance between random and real */
    float challenge, champion;	/* for resolving ties              */
    int itscale;			/* how many iterations per seq     */
    int iteration;
    int best;			/* index of nearest real sequence  */

    /* Sanity check first
     */
    if (nseq == 1) {
        wgt[0] = 1.0;
        return;
    }

    itscale = 50;

    /* Precalculate 1/2 minimum distance to other
     * sequences for each sequence
     */
    if (!simple_diffmx (aseq, nseq, &dmx)) {
        Die ("simple_diffmx() failed");
    }
    halfmin = MallocOrDie (sizeof (float) * nseq);
    for (idx = 0; idx < nseq; idx++) {
        for (min = 1.0, i = 0; i < nseq; i++) {
            if (i == idx) {
                continue;
            }
            if (dmx[idx][i] < min) {
                min = dmx[idx][i];
            }
        }
        halfmin[idx] = min / 2.0;
    }
    Free2DArray ((void **) dmx, nseq);

    /* Set up the random sequence generating model.
     */
    psym = MallocOrDie (alen * sizeof (char *));
    nsym = MallocOrDie (alen * sizeof (int));
    for (acol = 0; acol < alen; acol++) {
        psym[acol] = MallocOrDie (27 * sizeof (char));
    }

    /* ifdef ORIGINAL_SIBBALD_ALGORITHM_IS_BROKEN */
    for (acol = 0; acol < alen; acol++) {
        memset (symseen, 0, sizeof (int) * 27);
        for (idx = 0; idx < nseq; idx++) {
            if (!isgap (aseq[idx][acol])) {
                if (isupper ((int) aseq[idx][acol])) {
                    symidx = aseq[idx][acol] - 'A';
                }
                else {
                    symidx = aseq[idx][acol] - 'a';
                }
                if (symidx >= 0 && symidx < 26) {
                    symseen[symidx] = 1;
                }
            }
            else {
                symseen[26] = 1;    /* a gap */
            }
				}
        for (nsym[acol] = 0, i = 0; i < 26; i++) {
            if (symseen[i]) {
                psym[acol][nsym[acol]] = 'A' + i;
                nsym[acol]++;
            }
				}		
        if (symseen[26]) {
            psym[acol][nsym[acol]] = ' ';
            nsym[acol]++;
        }
    }
    /* #endif ORIGINAL_SIBBALD_ALGORITHM_IS_BROKEN */

    /* Note: the original Sibbald&Argos algorithm calls for
     * bounding the sampled space using a template-like random
     * sequence generator. However, this leads to one minor
     * and one major problem. The minor problem is that
     * exceptional amino acids in a column can have a
     * significant effect by altering the amount of sampled
     * sequence space; the larger the data set, the worse
     * this problem becomes. The major problem is that
     * there is no reasonable way to deal with gaps.
     * Gapped sequences simply inhabit a different dimensionality
     * and it's pretty painful to imagine calculating Voronoi
     * volumes when the N in your N-space is varying.
     * Note that all the examples shown by Sibbald and Argos
     * are *ungapped* examples.
     *
     * The best way I've found to circumvent this problem is
     * just not to bound the sampled space; count gaps as
     * symbols and generate completely random sequences.
     */
#ifdef ALL_SEQUENCE_SPACE
    for (acol = 0; acol < alen; acol++) {
        strcpy (psym[acol], "ACDEFGHIKLMNPQRSTVWY ");
        nsym[acol] = 21;
    }
#endif

    /* Sibbald and Argos algorithm:
     *   1) assign all seqs weight 0.
     *   2) generate a "random" sequence
     *   3) calculate distance to every other sequence
     *      (if we get a distance < 1/2 minimum distance
     *       to other real seqs, we can stop)
     *   4) if unique closest sequence, increment its weight 1.
     *      if multiple closest seq, choose one randomly
     *   5) repeat 2-4 for lots of iterations
     *   6) normalize all weights to sum to nseq.
     */
    randseq = MallocOrDie ((alen + 1) * sizeof (char));

    best = 42.;			/* solely to silence GCC uninit warnings. */
    FSet (wgt, nseq, 0.0);
    for (iteration = 0; iteration < itscale * nseq; iteration++) {
        for (acol = 0; acol < alen; acol++) {
            randseq[acol] =
                (nsym[acol] == 0) ? ' ' : psym[acol][CHOOSE (nsym[acol])];
				}				
        randseq[acol] = '\0';

        champion = sre_random ();
        for (min = 1.0, idx = 0; idx < nseq; idx++) {
            dist = simple_distance (aseq[idx], randseq);
            if (dist < halfmin[idx]) {
                best = idx;
                break;
            }
            if (dist < min) {
                champion = sre_random ();
                best = idx;
                min = dist;
            }
            else if (dist == min) {
                challenge = sre_random ();
                if (challenge > champion) {
                    champion = challenge;
                    best = idx;
                    min = dist;
                }
            }
        }
        wgt[best] += 1.0;
    }

    for (idx = 0; idx < nseq; idx++) {
        wgt[idx] = wgt[idx] / (float) itscale;
    }

    free (randseq);
    free (nsym);
    free (halfmin);
    Free2DArray ((void **) psym, alen);
}


/* Function: simple_distance()
 *
 * Purpose:  For two identical-length null-terminated strings, return
 *           the fractional difference between them. (0..1)
 *           (Gaps don't count toward anything.)
 */
static float
simple_distance (char *s1, char *s2)
{
    int diff = 0;
    int valid = 0;

    for (; *s1 != '\0'; s1++, s2++) {
        if (isgap (*s1) || isgap (*s2)) {
            continue;
        }
        if (*s1 != *s2) {
            diff++;
        }
        valid++;
    }
    return (valid > 0 ? ((float) diff / (float) valid) : 0.0);
}

/* Function: simple_diffmx()
 *
 * Purpose:  Given a set of flushed, aligned sequences, construct
 *           an NxN fractional difference matrix using the
 *           simple_distance rule.
 *
 * Args:     aseqs        - flushed, aligned sequences
 *           num          - number of aseqs
 *           ret_dmx      - RETURN: difference matrix (caller must free)
 *
 * Return:   1 on success, 0 on failure.
 */
static int
simple_diffmx (char **aseqs, int num, float ***ret_dmx)
{
    float **dmx;			/* RETURN: distance matrix           */
    int i, j;			/* counters over sequences           */

    /* Allocate
     */
    if ((dmx = (float **) malloc (sizeof (float *) * num)) == NULL) {
        Die ("malloc failed");
    }
    for (i = 0; i < num; i++) {
        if ((dmx[i] = (float *) malloc (sizeof (float) * num)) == NULL) {
            Die ("malloc failed");
        }
		}
    /* Calculate distances, symmetric matrix
     */
    for (i = 0; i < num; i++) {
        for (j = i; j < num; j++) {
            dmx[i][j] = dmx[j][i] = simple_distance (aseqs[i], aseqs[j]);
        }
		}
    /* Return
     */
    *ret_dmx = dmx;
    return 1;
}



/* Function: BlosumWeights()
 * Date:     SRE, Fri Jul 16 17:33:59 1999 (St. Louis)
 *
 * Purpose:  Assign weights to a set of aligned sequences
 *           using the BLOSUM rule:
 *             - do single linkage clustering at some pairwise identity
 *             - in each cluster, give each sequence 1/clustsize
 *               total weight.
 *
 *           The clusters have no pairwise link >= maxid.
 *
 *           O(N) in memory. Probably ~O(NlogN) in time; O(N^2)
 *           in worst case, which is no links between sequences
 *           (e.g., values of maxid near 1.0).
 *
 * Args:     aseqs - alignment
 *           nseq  - number of seqs in alignment
 *           alen  - # of columns in alignment
 *           maxid - fractional identity (e.g. 0.62 for BLOSUM62)
 *           wgt   - [0..nseq-1] array of weights to be returned
 */
void
BlosumWeights (char **aseqs, int nseq, int alen, float maxid, float *wgt)
{
    int *c, nc;
    int *nmem;			/* number of seqs in each cluster */
    int i;			/* loop counter */

    SingleLinkCluster (aseqs, nseq, alen, maxid, &c, &nc);

    FSet (wgt, nseq, 1.0);
    nmem = MallocOrDie (sizeof (int) * nc);

    for (i = 0; i < nc; i++) {
        nmem[i] = 0;
    }
    for (i = 0; i < nseq; i++) {
        nmem[c[i]]++;
    }
    for (i = 0; i < nseq; i++) {
        wgt[i] = 1. / (float) nmem[c[i]];
    }

    free (nmem);
    free (c);
    return;
}


/* Function: PositionBasedWeights()
 * Date:     SRE, Fri Jul 16 17:47:22 1999 [St. Louis]
 *
 * Purpose:  Implementation of Henikoff and Henikoff position-based
 *           weights (JMB 243:574-578, 1994) [Henikoff94b].
 *
 *           A significant advantage of this approach that Steve and Jorja
 *           don't point out is that it is O(N) in memory, unlike
 *           many other approaches like GSC weights or Voronoi.
 *
 *           A potential disadvantage that they don't point out
 *           is that in the theoretical limit of infinite sequences
 *           in the alignment, weights go flat: eventually every
 *           column has at least one representative of each of 20 aa (or 4 nt)
 *           in it.
 *
 *           They also don't give a rule for how to handle gaps.
 *           The rule used here seems the obvious and sensible one
 *           (ignore them). This means that longer sequences
 *           initially get more weight; hence a "double
 *           normalization" in which the weights are first divided
 *           by sequence length (to compensate for that effect),
 *           then normalized to sum to nseq.
 *
 * Limitations:
 *           Implemented in a way that's alphabet-independent:
 *           it uses the 26 upper case letters as "residues".
 *           Any alphabetic character in aseq is interpreted as
 *           a unique "residue" (case insensitively; lower case
 *           mapped to upper case). All other characters are
 *           interpreted as gaps.
 *
 *           This way, we don't have to pass around any alphabet
 *           type info (DNA vs. RNA vs. protein) and don't have
 *           to deal with remapping IUPAC degenerate codes
 *           probabilistically. However, on the down side,
 *           a sequence with a lot of degenerate IUPAC characters
 *           will get an artifactually high PB weight.
 *
 * Args:     aseq   - sequence alignment to weight
 *           nseq   - number of sequences in alignment
 *           alen   - length of alignment
 *           wgt    - RETURN: weights filled in (pre-allocated 0..nseq-1)
 *
 * Returns:  (void)
 *           wgt is allocated (0..nseq-1) by caller, and filled in here.
 */
void
PositionBasedWeights (char **aseq, int nseq, int alen, float *wgt)
{
    int rescount[26];		/* count of A-Z residues in a column   */
    int nres;			/* number of different residues in col */
    int idx, pos;			/* indices into aseq                   */
    int x;
    float norm;

    FSet (wgt, nseq, 0.0);
    for (pos = 0; pos < alen; pos++) {
        for (x = 0; x < 26; x++) {
            rescount[x] = 0;
        }
        for (idx = 0; idx < nseq; idx++) {
            if (isalpha ((int) aseq[idx][pos])) {
                rescount[toupper ((int) aseq[idx][pos]) - 'A']++;
            }
				}
        nres = 0;
        for (x = 0; x < 26; x++) {
            if (rescount[x] > 0) {
                nres++;
            }
				}
        for (idx = 0; idx < nseq; idx++) {
            if (isalpha ((int) aseq[idx][pos]))
                wgt[idx] +=
                    1. / (float) (nres *
                                  rescount[toupper ((int) aseq[idx][pos]) - 'A']);
				}													
    }

    for (idx = 0; idx < nseq; idx++) {
        wgt[idx] /= (float) DealignedLength (aseq[idx]);
    }
    norm = (float) nseq / FSum (wgt, nseq);
    FScale (wgt, nseq, norm);
    return;
}




/* Function: FilterAlignment()
 * Date:     SRE, Wed Jun 30 09:19:30 1999 [St. Louis]
 *
 * Purpose:  Constructs a new alignment by removing near-identical
 *           sequences from a given alignment (where identity is
 *           calculated *based on the alignment*).
 *           Does not affect the given alignment.
 *           Keeps earlier sequence, discards later one.
 *
 *           Usually called as an ad hoc sequence "weighting" mechanism.
 *
 * Limitations:
 *           Unparsed Stockholm markup is not propagated into the
 *           new alignment.
 *
 * Args:     msa      -- original alignment
 *           cutoff   -- fraction identity cutoff. 0.8 removes sequences > 80% id.
 *           ret_new  -- RETURN: new MSA, usually w/ fewer sequences
 *
 * Return:   (void)
 *           ret_new must be free'd by caller: MSAFree().
 */
void
FilterAlignment (MSA * msa, float cutoff, MSA ** ret_new)
{
    int nnew;			/* number of seqs in new alignment */
    int *list;
    int *useme;
    float ident;
    int i, j;
    int remove;

    /* find which seqs to keep (list) */
    /* diff matrix; allow ragged ends */
    list = MallocOrDie (sizeof (int) * msa->nseq);
    useme = MallocOrDie (sizeof (int) * msa->nseq);
    for (i = 0; i < msa->nseq; i++) {
        useme[i] = FALSE;
    }

    nnew = 0;
    for (i = 0; i < msa->nseq; i++) {
        remove = FALSE;
        for (j = 0; j < nnew; j++) {
            ident = PairwiseIdentity (msa->aseq[i], msa->aseq[list[j]]);
            if (ident > cutoff) {
                remove = TRUE;
                printf ("removing %12s -- fractional identity %.2f to %s\n",
                        msa->sqname[i], ident, msa->sqname[list[j]]);
                break;
            }
        }
        if (remove == FALSE) {
            list[nnew++] = i;
            useme[i] = TRUE;
        }
    }

    MSASmallerAlignment (msa, useme, ret_new);
    free (list);
    free (useme);
    return;
}


/* Function: SampleAlignment()
 * Date:     SRE, Wed Jun 30 10:13:56 1999 [St. Louis]
 *
 * Purpose:  Constructs a new, smaller alignment by sampling a given
 *           number of sequences at random. Does not change the
 *           alignment nor the order of the sequences.
 *
 *           If you ask for a sample that is larger than nseqs,
 *           it silently returns the original alignment.
 *
 *           Not really a weighting method, but this is as good
 *           a place as any to keep it, since it's similar in
 *           construction to FilterAlignment().
 *
 * Args:     msa      -- original alignment
 *           sample   -- number of sequences in new alignment (0 < sample <= nseq)
 *           ret_new  -- RETURN: new MSA
 *
 * Return:   (void)
 *           ret_new must be free'd by caller: MSAFree().
 */
void
SampleAlignment (MSA * msa, int sample, MSA ** ret_new)
{
    int *list;			/* array for random selection w/o replace */
    int *useme;			/* array of flags 0..nseq-1: TRUE to use */
    int i, idx;
    int len;

    /* Allocations
     */
    list = (int *) MallocOrDie (sizeof (int) * msa->nseq);
    useme = (int *) MallocOrDie (sizeof (int) * msa->nseq);
    for (i = 0; i < msa->nseq; i++) {
        list[i] = i;
        useme[i] = FALSE;
    }

    /* Sanity check.
     */
    if (sample >= msa->nseq) {
        sample = msa->nseq;
    }

    /* random selection w/o replacement */
    for (len = msa->nseq, i = 0; i < sample; i++) {
        idx = CHOOSE (len);
        printf ("chose %d: %s\n", list[idx], msa->sqname[list[idx]]);
        useme[list[idx]] = TRUE;
        list[idx] = list[--len];
    }

    MSASmallerAlignment (msa, useme, ret_new);
    free (list);
    free (useme);
    return;
}


/* Function: SingleLinkCluster()
 * Date:     SRE, Fri Jul 16 15:02:57 1999 [St. Louis]
 *
 * Purpose:  Perform simple single link clustering of seqs in a
 *           sequence alignment. A pairwise identity threshold
 *           defines whether two sequences are linked or not.
 *
 *           Important: runs in O(N) memory, unlike standard
 *           graph decomposition algorithms that use O(N^2)
 *           adjacency matrices or adjacency lists. Requires
 *           O(N^2) time in worst case (which is when you have
 *           no links at all), O(NlogN) in "average"
 *           case, and O(N) in best case (when there is just
 *           one cluster in a completely connected graph.
 *
 *           (Developed because hmmbuild could no longer deal
 *           with GP120, a 16,013 sequence alignment.)
 *
 * Limitations:
 *           CASE-SENSITIVE. Assumes aseq have been put into
 *           either all lower or all upper case; or at least,
 *           within a column, there's no mixed case.
 *
 * Algorithm:
 *           I don't know if this algorithm is published. I
 *           haven't seen it in graph theory books, but that might
 *           be because it's so obvious that nobody's bothered.
 *
 *           In brief, we're going to do a breadth-first search
 *           of the graph, and we're going to calculate links
 *           on the fly rather than precalculating them into
 *           some sort of standard adjacency structure.
 *
 *           While working, we keep two stacks of maximum length N:
 *                a : list of vertices that are still unconnected.
 *                b : list of vertices that we've connected to
 *                    in our current breadth level, but we haven't
 *                    yet tested for other connections to a.
 *           The current length (number of elements in) a and b are
 *           kept in na, nb.
 *
 *           We store our results in an array of length N:
 *                c : assigns each vertex to a component. for example
 *                    c[4] = 1 means that vertex 4 is in component 1.
 *                    nc is the number of components. Components
 *                    are numbered from 0 to nc-1. We return c and nc
 *                    to our caller.
 *
 *           The algorithm is:
 *
 *           Initialisation:
 *                a  <-- all the vertices
 *                na <-- N
 *                b  <-- empty set
 *                nb <-- 0
 *                nc <-- 0
 *
 *           Then:
 *                while (a is not empty)
 *                  pop a vertex off a, push onto b
 *                  while (b is not empty)
 *                    pop vertex v off b
 *                    assign c[v] = nc
 *                    for each vertex w in a:
 *                       compare v,w. If w is linked to v, remove w
 *                       from a, push onto b.
 *                  nc++
 *           q.e.d. :)
 *
 * Args:     aseq   - aligned sequences
 *           nseq   - number of sequences in aseq
 *           alen   - alignment length
 *           maxid  - fractional identity threshold 0..1. if id >= maxid, seqs linked
 *           ret_c  - RETURN: 0..nseq-1 assignments of seqs to components (clusters)
 *           ret_nc - RETURN: number of components
 *
 * Returns:  void.
 *           ret_c is allocated here. Caller free's with free(*ret_c)
 */
void
SingleLinkCluster (char **aseq, int nseq, int alen, float maxid,
                   int **ret_c, int *ret_nc)
{
    int *a, na;			/* stack of available vertices */
    int *b, nb;			/* stack of working vertices   */
    int *c;			/* array of results            */
    int nc;			/* total number of components  */
    int v, w;			/* index of a working vertices */
    int i;			/* loop counter */

    /* allocations and initializations
     */
    a = MallocOrDie (sizeof (int) * nseq);
    b = MallocOrDie (sizeof (int) * nseq);
    c = MallocOrDie (sizeof (int) * nseq);
    for (i = 0; i < nseq; i++) {
        a[i] = i;
    }
    na = nseq;
    nb = 0;
    nc = 0;

    /* Main algorithm
     */
    while (na > 0) {
        v = a[na - 1];
        na--;			/* pop a vertex off a, */
        b[nb] = v;
        nb++;			/* and push onto b     */
        while (nb > 0) {
            v = b[nb - 1];
            nb--;			/* pop vertex off b          */
            c[v] = nc;		/* assign it to component nc */
            for (i = na - 1; i >= 0; i--)	{
								/* backwards, becase of deletion/swapping we do */
                if (simple_distance (aseq[v], aseq[a[i]]) <= 1. - maxid) {	/* linked? */
                    w = a[i];
                    a[i] = a[na - 1];
                    na--;			/* delete w from a (note swap) */
                    b[nb] = w;
                    nb++;			/* push w onto b */
                }
						}		
        }
        nc++;
    }

    /* Cleanup and return
     */
    free (a);
    free (b);
    *ret_c = c;
    *ret_nc = nc;
    return;
}
